/*
** Module   : progress.g
** Abstract : Progress 4GL Lexer and Parser Grammar
**
** Copyright (c) 2004-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 GES 20041116   @18747 First version implementing a complete lexer
**                           and a prototype parser.  The parser provides 
**                           the following features:
**                           - full expression processing for string,
**                             boolean, date and numeric expressions
**                             (both integer and decimal), recid, rowid,
**                             raw, and handle (all supported data types)
**                           - full support for tree creation including
**                             the proper/useful structuring of all 
**                             major structural language features such
**                             as blocks, procedures and functions
**                           - symbol resolution for variables, labels,
**                             procedures, functions and keywords
**                             (including handling abbreviations and
**                             reserved keywords)
** 002 GES 20041207   @18938 Added built-in function/variable support.
**                           Functions that take no parameters are
**                           implemented as variables and all others as
**                           functions.  Built-in variables that are not
**                           called functions can also be added in the
**                           future, with no infrastructure changes. 
**                           Changes were made to the following:
**                           - token types and keyword definitions added
**                             for all the new function/variable names
**                           - new FUNC_POLY type for functions that can
**                             return more than 1 data type depending on
**                             the input parameters (only built-ins can
**                             do this)
**                           - new initialization methods to populate the
**                             symbol resolver function/variable
**                             namespaces
**                           - modify constructors to call new initializer
**                             methods
**                           - added new "pull" methods to match reserved 
**                             variable or function keyword token types
**                           - modified lvalue and func_call to include
**                             these reserved keywords as a pull
** 003 GES 20041209   @18944 Modified the function and func_stmt rules to
**                           eliminate the block rule reference and tree
**                           node creation in in FORWARD declarations.
** 004 GES 20041209   @18950 Added array processing support. This caused
**                           changes in how lvalues work (new subscript
**                           operator [ ]) and in how variables are 
**                           defined (to handle extents and 
**                           initialization).  These changes required
**                           modifications to lvalue() and def_var_stmt()
**                           but in addition, to create a tree that is
**                           well structured for processing, structural
**                           changes were needed in these rules.  Subrules
**                           were added to ensure that subtrees are 
**                           created instead of leaving everything flat.
**                           Due to the testcase being used, an early
**                           version of additional language statements
**                           assign and display were added. Note that the
**                           addition of assign causes ANTLR to sense
**                           ambiguity with the assignment() rule. This
**                           is not actually a problem so warnings had
**                           to be suppressed.
** 005 GES 20041210   @18968 Modified the lexer to support date constants
**                           and a new token for the @ sign.  The date
**                           constants are implemented as modifications
**                           to the NUM_LITERAL rule since they need to
**                           be unambiguous in handling leading digits
**                           (just as decimal literals). In addition,
**                           slight modifications to the parser were made
**                           to add the token type DATE_LITERAL and to
**                           add this as a new alternative to the literal
**                           rule.
** 006 GES 20041221   @19042 Modified the lexer's mSYMBOL rule to
**                           recognize and match partially or fully
**                           qualified schema names. Such matches are
**                           given a DB_SYMBOL type and bypass the normal
**                           keyword lookup processing.  The parser was 
**                           modified to recognize and match schema
**                           fields, tables and database names. The core
**                           changes for this were implemented in the
**                           lvalue (for fields), record (for tables) and
**                           database rules. To ensure that this support
**                           was properly designed, some other major
**                           changes were made that could not previously
**                           be done without the schema support. The
**                           DO, FOR and REPEAT block statements were all
**                           added and/or completed (except for frame
**                           phrase support).  These all and DEF_VAR_STMT
**                           required a very large number of new rules
**                           of which many are used in more than 1
**                           location.  Some of the most important new
**                           rules include like_clause and record_phrase.
**                           Finally, the lookup side of label processing
**                           was added, which was required for the block
**                           processing (see label_reference and its
**                           primary user on_event_phrase).
** 007 NVS 20041220   @19032 Changed SSTRING/DSTRING and WS rules in the
**                           lexer. All three no longer refer to '\r' 
**                           character as it can't be seen after preproce-
**                           ssing. SSTRING and DSTRING keep newlines 
**                           because they only can see newlines that have 
**                           to be retained. The preprocessor deletes the 
**                           ordinary newlines from strings.
** 008 GES 20041223   @19056 Added IF THEN ELSE (as a language statement
**                           and as a non-standard, hard coded function 
**                           in the expression processing) and CASE
**                           language statement support.  In order to
**                           properly implement the THEN and ELSE clauses
**                           of these rules, it was necessary to separate
**                           the top-level alternatives of the block
**                           rule into a separate rule called
**                           single_block. This rule matches one and only
**                           one top-level alternative.  It is called
**                           in a loop by block and directly by the
**                           THEN and ELSE clauses.  Making this change
**                           does not cause any problems or changes to
**                           how the rest of the tree is built.  It simply
**                           refactors where the function is placed to
**                           allow the top-level alternatives to be used
**                           more flexibly.
** 009 GES 20041223   @19068 Added the substring( , , , ) = expression
**                           language statement with its function-like
**                           syntax and its built-in assignment.
** 010 NVS 20041223   @19076 Unary logical NOT operation has to have pre-
**                           cedence above AND and below comparisons.
**                           Created log_not_expr rule that fits between 
**                           log_and_expr and compare_expr rules. Deleted
**                           KW_NOT from the un_type rule, where it was
**                           put by mistake. The result is that expressio-
**                           ns like "true < not false" don't parse, but
**                           "true < (not false)" do, but this is exactly
**                           what Progress does.
** 011 GES 20041224   @19082 Refactored DEFINE statements into a single
**                           left-factored rule that handles the DEFINE
**                           keyword, the options (for object sharing and
**                           for parameter types) and then switches into
**                           the various alternatives (DEF VAR, DEF PARM,
**                           DEF BUF are all implemented at this time).
** 012 GES 20041224   @19092 Added FIND statement support.  Enhanced and
**                           reused support from the record_phrase.  Also
**                           added support for BELL, CLEAR, LEAVE, NEXT,
**                           PAUSE, QUIT, STATUS, STOP and UNDO.
** 013 NVS 20041224   @19101 Changed both ProgressLexer constructors. Now 
**                           they use DotKludgeStream and DotKludgeReader
**                           classes respectively. Those classes take care
**                           of missing dot after the last statement in
**                           the Progress source files.
** 014 NVS 20041225   @19110 Implemented block label scoping. Block labels
**                           are created in scopes so there is one label
**                           per scope. The scope and the only label in it
**                           are removed when the block terminates.
** 015 NVS 20041227   @19124 AT phrase support has been added. The DEFINE
**                           rule now has an embryonic FRAME definition
**                           rule which calls AT phrase.
** 016 NVS 20041227   @19127 Added color-specification and color-phrase
**                           support. The DEFINE FRAME has been extended
**                           to include form-item and frame-phrase (both
**                           still embryonic). Frame-phrase calls color-
**                           spec rule which in turn calls color-phrase
**                           rule. The latter works OK except for proterm-
**                           cap attributes which don't look like a symbol
** 017 GES 20041228   @19129 Added RUN statement support which required
**                           support for filenames, library references,
**                           the VALUE() construct and many clauses.
**                           Major lexer changes were made to the mSYMBOL
**                           rule (rewritten) and the DIVIDE rule was
**                           deleted (it is now handled in MSYMBOL).
** 018 NVS 20041228   @19130 Completed color-phrase implementation. Due to
**                           a lexer extension for filenames, the obscure 
**                           protermcap entries can now be parsed as file-
**                           names and the color-phrase rule handles this 
**                           case correctly.
** 019 GES 20041229   @19136 Added parameter and preprocessor argument
**                           processing to the RUN statement and updated
**                           the RUN statement to test filenames to
**                           see if they are actually internal procedure
**                           references.  Minor changes to the MESSAGE
**                           and DISPLAY statements to include multiple
**                           expressions.  These are otherwise still
**                           quite incomplete.
** 020 GES 20041230   @19140 Took some NVS work and massively modified it
**                           to implement complete frame phrase support.
**                           As part of this change, size phrase, at
**                           phrase, color specification, COLOR statement
**                           and DEFINE FRAME statement were all
**                           implemented.  The frame phrase support was
**                           added to the DO, REPEAT and FOR statements
**                           which completes these.
** 021 GES 20041231   @19141 Added shell command support to lexer and 
**                           parser. There is a new UNKNOWN_TOKEN type
**                           for the common special chars that are not
**                           normally supported in Progress, but that
**                           can be easily generated on the US keyboard 
**                           and which can be inserted as tokens in a
**                           command token list for a shell command
**                           (without putting them in a valid Progress
**                           string). The UNIX, DOS, OS2, BTOS, VMS and
**                           OS-COMMAND statements are all supported.
** 022 GES 20050103   @19150 Added format phrase, aggregate phrase and
**                           completed DISPLAY and MESSAGE statements.
** 023 GES 20050103   @19176 Added support for the widget phrase and the
**                           following language statements: DISABLE,
**                           ENABLE, DOWN, UP, ASSIGN (finished this),
**                           PUT, PUT SCREEN, DELETE, CREATE, HIDE, PAGE,
**                           VIEW, VALIDATE, UNDERLINE, READKEY, APPLY.
** 024 GES 20050103   @19190 Added view-as phrase support with all
**                           possible sub-phrases and clauses.  Linked
**                           this rule to def_var_stmt and format_phrase.
** 025 GES 20050103   @19194 Integrated schema processing and resolved
**                           issues related to precedence of records
**                           and fields, as well as making the semantic
**                           predicates (used to disambiguate) safe from 
**                           exceptions.
** 026 GES 20050106   @19257 Merged func_parameter and run_parameter into
**                           a single parameter rule. Actually the run
**                           form is a superset of the function form and
**                           the only real change is renaming the rules
**                           where they are referenced, renaming the 
**                           run_parameter, removing the func_parameter
**                           and updating the javadoc. This change causes
**                           no real change in parsing or AST build at
**                           this time.
** 027 GES 20050107   @19272 Added the DEFINE BUTTON statement which
**                           required the image phrase and a new widget
**                           namespace to be added. All widget references
**                           and other locations that widgets can be
**                           defined (in a view-as phrase) were also
**                           updated to consistently handle the namespace.
** 028 GES 20050111   @19314 Added the DEFINE BROWSE and DEFINE QUERY
**                           language statements and associated rules.
** 029 GES 20050112   @19317 Added the DEFINE MENU, DEFINE SUB-MENU and
**                           DEFINE STREAM language statements and
**                           associated rules.
** 030 GES 20050112   @19318 Added the SET and UPDATE language statements
**                           to the parser and modified the lexer to
**                           support a new CARET token that matches '^'.
**                           Also added the editing phrase which is
**                           a special kind of block that cannot be used
**                           at the top level in a Progress program.
**                           Also included PROMPT-FOR support since it is
**                           basically identical to SET and UPDATE.
** 031 GES 20050114   @19326 Added trigger blocks, cleaned up event
**                           processing, added WAIT-FOR and ON statements.
**                           In this pass, the most critical language
**                           support is now available, allowing most
**                           files to be processed!  Initial testing
**                           against real non-trivial examples was made
**                           and several bug fixes and minor additions
**                           were made to handle both documented and 
**                           undocumented features.
** 032 GES 20050117   @19342 Added more schema/database support including
**                           the update of the schema namespace when
**                           buffers, temp-tables and work-tables are
**                           defined.  Added the following language
**                           statements: CONNECT, DISCONNECT, CREATE
**                           ALIAS, DELETE ALIAS, DEFINE TEMP-TABLE,
**                           DEFINE WORK-TABLE/WORKFILE. 
** 033 GES 20050119   @19345 Added the rest of the database related
**                           language statements need for the 1st pass
**                           parser: CREATE DATABASE, CREATE QUERY,
**                           CREATE WIDGET-POOL, CREATE TEMP-TABLE,
**                           CREATE BUFFER, DELETE WIDGET-POOL, OPEN
**                           QUERY, CLOSE QUERY, REPOSITION, GET.
** 034 GES 20050119   @19347 Added record scoping support.  This simple
**                           implementation (scopes are added at the
**                           beginning of all inner_blocks and removed
**                           at the end AND all record rule references
**                           promote the table name to the current scope)
**                           solves all testcases found so far. There
**                           may be an obscure case of ambiguity that
**                           this does not resolve, but it will likely
**                           be VERY RARE.  At this time it seems that
**                           the schema support is complete.
** 035 GES 20050120   @19348 Implemented (all?) stream I/O related
**                           language statements: INPUT, INPUT-OUTPUT and
**                           OUTPUT CLOSE; INPUT CLEAR; INPUT FROM;
**                           OUTPUT TO; INPUT, INPUT-OUTPUT, OUTPUT
**                           THROUGH/THRU.
** 036 GES 20050120   @19355 Added many misc (less often used) language
**                           statements: ACCUMULATE, BUFFER-COMPARE,
**                           BUFFER-COPY, CHOOSE, CREATE widget, DELETE
**                           OBJECT, DELETE PROCEDURE, DELETE WIDGET,
**                           DICTIONARY, ENTRY (not function, only the
**                           language statement), EXPORT, FRAME-VALUE,
**                           IMPORT, NEXT-PROMPT, OS-DELETE, OVERLAY,
**                           PUT CURSOR, RAW_TRANSFER, RELEASE, SAVE
**                           CACHE, TERMINAL and COMPILE.
** 037 GES 20050121   @19362 Added all other language statements that
**                           appear to be common: CURRENT-LANGUAGE,
**                           CURRENT-VALUE, PROCESS EVENTS, CREATE BROWSE,
**                           CREATE X-DOCUMENT, CREATE X-NODEREF, CREATE
**                           SERVER, CREATE SERVER-SOCKET, CREATE SOCKET,
**                           EMPTY TEMP-TABLE, INSERT, OS-APPEND, OS-COPY,
**                           OS-CREATE-DIR, OS-RENAME, PROMSGS, PROPATH,
**                           SEEK, SHOW-STATS.
** 038 GES 20050121   @19365 Added all other documented functions
**                           including the special cases of ACCUM, CAN-DO
**                           and SUPER.  This represents a *probably*
**                           feature complete parser, although it is
**                           likely that many bugs or mistaken codings
**                           will need to be fixed AND the AST will need
**                           some polish.
** 039 GES 20050126   @19410 Major modifications and minor fixes as found
**                           by scanning real code.  Major mods include
**                           consolidation of system handles, widgets and
**                           variables/fields in the common lvalue rule
**                           as well as the movement of method/attribute
**                           support into lvalue.  This allows methods
**                           and attributes to be used in a much wider
**                           range of situations and reduced the
**                           ambiguity level of the parser in many places.
** 040 GES 20050127   @19451 Additional major modifications for handling
**                           widget (frame, menu, sub-menu, menu-item,
**                           browse) and query references.  Such 
**                           references occur in lvalues, to qualify
**                           method or attribute references.  These
**                           changes make lvalue a superset of what is
**                           needed for virtually all of the widget
**                           phrase.  This allowed a massive
**                           simplification of various reference rules
**                           and the widget_phrase while a massive
**                           set of changes (including new namespace
**                           processing) were added to lvalue.  Additional
**                           namespace changes and a rewrite of the
**                           assignment rule (to disambiguate some
**                           really nasty situations primarily caused by
**                           the addition of methods and attributes
**                           support.  The lexer had issues with matching
**                           inside of strings, including a charset that
**                           was missing 0x00 - 0x02 (and needed these)
**                           as well as a need to handle tildes
**                           differently.
** 041 GES 20050128   @19461 Added streams namespace support, new buffer 
**                           and stream references in lvalue, more refined
**                           schema scoping support for temp-table
**                           processing, added new undocumented features
**                           and fixed many bugs.
** 042 GES 20050131   @19466 Added many fixes, new MEMPTR data type 
**                           support, new language statements, new
**                           undocumented features based on customer
**                           source code.
** 043 GES 20050201   @19514 More fixes and modifications to support
**                           undocumented behavior based on scanning
**                           customer testcases. 
** 044 GES 20050204   @19568 More fixes and modifications to support
**                           undocumented behavior based on scanning
**                           customer testcases. 
** 045 GES 20050207   @19623 Minor modification to remove a deprecated
**                           schema dictionary call.
** 046 GES 20050207   @19742 More fixes based on customer testcases.
**                           Major modifications were made to handle
**                           schema lookups more correctly (especially
**                           in regards to scopes).  The DO statement
**                           only adds scopes on a DO FOR and buffers
**                           are added in either the global scope or in
**                           an internal procedure or function scope
**                           depending on the current context. Another 
**                           major change is the full implementation of
**                           special abbreviation processing for the
**                           as_clause, where undocumented shorter
**                           abbreviations are possible that are not
**                           normally possible elsewhere.
** 047 GES 20050217   @19858 Added many fixes, support for new hints
**                           (which allow us to trigger certain variable
**                           and schema overrides in form items and in
**                           the format phrase), major consolidation of
**                           assignment type language statements into the
**                           lvalue rule, addition of the PUBLISH
**                           language statement, cleanup of form items,
**                           keyword cleanups (keywords missing from the 
**                           dictionary or parser matching on invalid
**                           keyword names), added _cbit() and user()
**                           undocumented functions, color statement
**                           rework, easing restrictions on scope
**                           processing (procs and funcs can be nested
**                           inside inner blocks), premature end
**                           detection in external_proc (it returns a
**                           boolean now that tells the caller if EOF
**                           was encountered or not) and a lexer change
**                           to "eat" standalone "\" chars since these
**                           are ignored by Progress and customer files
**                           with these extra chars have been
**                           encountered.
** 048 GES 20050221   @19906 Added lookupTokenType() static method as a
**                           common helper to allow a conversion of a
**                           Progress-specific symbolic token type name
**                           to the associated integer token type. A
**                           reverse name from type lookup was also
**                           added.  This is intended to simplify the
**                           use of this information from multiple
**                           external locations.
** 049 GES 20050228   @20127 Tree building improvements to ensure that
**                           all language statements can be easily
**                           matched on a single, unique token beneath
**                           a statement token. In addition, improved
**                           the tree structure of the format phrase,
**                           frame phrase and form_item to include a
**                           manufactured root node.  The header_item 
**                           rule was removed since this is really an
**                           option keyword not the root of a form_item. 
**                           Functions that have special processing now
**                           return a root node that have the proper
**                           FUNC_ token type.
** 050 GES 20050307   @20219 Tree modifications to add an artificial
**                           EXPRESSION root when the expr rule is
**                           called downstream of the external_proc.
**                           Note that only the first entry to expr
**                           will generate such a token, recursion will
**                           not.  Added schema and other annotations
**                           to store data known at parse time which is
**                           difficult/impossible to recover later.
**                           Started major removal of unnecessary tokens
**                           such as DOT, COLON and COMMA.  Function
**                           definitions were modified to use a common
**                           rule reference to run_stmt_parms which 
**                           reduces the code duplication and also makes
**                           the parameters build with a subtree rooted
**                           at the LPARENS which is a better result.
**                           Undo language statement and the 
**                           on_event_phrase processing was rewritten to
**                           more properly build the resulting tree.
**                           After some analysis it was concluded that
**                           the end_stmt really only exists to delimit 
**                           for parsing purposes and is unnecessary in
**                           the final tree.  This statement is now
**                           "eaten" and won't appear in the tree.
** 051 ECF 20050322   @20456 Added artificial token to identify a record
**                           phrase. This makes downstream identification
**                           of this type of AST branch easier.
** 052 GES 20050328   @20510 Moved AST class processing into parser to
**                           centralize this function and make clients
**                           simpler and more independent.
** 053 GES 20050329   @20523 Tree building modifications to ease variable
**                           and field definition processing.
** 054 GES 20050331   @20560 Major modifications to store options at the
**                           time that variables and temp-tables are
**                           defined. These options are saved in the
**                           variable dictionary in the case of vars and
**                           are pushed out to the schema dictionary in
**                           the case of temp-tables. All variable and
**                           temp-table definitions that are LIKE another
**                           variable, field or table now honor and
**                           bidirectionally pass the options as needed.
**                           Variable and field references (the lvalue
**                           rule) store annotations for non-default
**                           options. WARNING: INITIAL, LABEL and
**                           VIEW-AS are 3 common options that are not
**                           processed or are only partially processed
**                           at this time.
** 055 GES 20050401   @20595 Minor changes to write variable option 
**                           annotations into the DEFINE_VARIABLE,
**                           DEFINE_PARAMETER, FORMAT_PHRASE, PARAMETER, 
**                           KW_SET or KW_UPDATE tokens when a variable
**                           is defined.  These annotations now include
**                           a "tempidx" which is a cross-reference
**                           between var defs and var refs. Node IDs
**                           can't be used because these are not known
**                           at parse time.  Post-parse fixups will need
**                           to convert the temp index into an AST ID.
** 056 GES 20050406   @20650 Additional changes related to reading and
**                           writing options and annotations for variables
**                           and fields.
** 057 ECF 20050412   @20681 New metadata type: "time".  Added "time"
**                           keyword to schema keyword list.  Added
**                           FIELD_TIME token type to METATYPES range.
** 058 GES 20050414   @20721 Modifications to support function annotations
**                           and rework of variable/field annotations
**                           to process based on references where
**                           possible. Changed initializer processing to
**                           cleanup undocumented deviations where a
**                           constant is defined inside quotes when it
**                           should not be (e.g. a NUM_LITERAL specified
**                           in source as "0").
** 059 GES 20050426   @20877 Modified string processing in the lexer to
**                           handle the fact that Progress has an
**                           undocumented escape char (backslash) that
**                           can do everything that the tilde can do.
** 060 GES 20050430   @20948 Token type and tree modifications to simplify 
**                           conversion of blocks and control flow
**                           constructs such as if/then/else, labels and
**                           case/when/then/otherwise.
** 061 ECF 20050502   @20987 Added token types used by schema table
**                           relation analysis:  MANY_TO_ONE, ONE_TO_ONE
**                           and KEY_FIELD.
** 062 GES 20050502   @2104x Fix for null pointer exception in 
**                           single_block.
** 063 GES 20050506   @21120 Final disposition of the use of KW_TODAY as
**                           a DATE_LITERAL.
** 064 NVS 20050519   @21242 Added new member evaluatingExpression and 2
**                           methods (is... and set...). This boolean is
**                           a flag for special preprocessor expression
**                           evaluation mode. lvalue rule is the only one
**                           that tests this flag. In expression evaluati-
**                           on mode, when DEFINED() built-in function is
**                           used to test an undefined preprocessor symbol
**                           , the lvalue rule simply returns SYMBOL token
**                           instead of throwing an exception. The regular
**                           mode requires that all SYMBOL type tokens are
**                           resolved into more specific types. If it can
**                           not be done, NoViableAltException is thrown.
** 065 GES 20050608   @21449 Function definition tree and annotation
**                           improvements.
** 066 GES 20050614   @21486 Modifications to lvalue post-processing to
**                           fix annotation placement in methods and
**                           attributes.
** 067 GES 20050614   @21491 Tree changes to simplify control flow
**                           conversion.  Also added annotations
**                           processing to an obscure location that 
**                           rewrites a token as a function call.
** 068 ECF 20050615   @21530 Additional token type for schema relation
**                           analysis. ONE_TO_MANY is used to describe the
**                           inverse end of the MANY_TO_ONE relation.
** 069 GES 20050621   @21542 Add a "parmtype" annotation when a parameter 
**                           is defined without a qualifier (this means
**                           the type is KW_INPUT).
** 070 GES 20050622   @21543 Modified the run statement to convert an
**                           internal or external procedure name (which
**                           is sometimes specified as a STRING) into
**                           a FILENAME or INT_PROC, which is more
**                           correct.  Note that this STRING form is
**                           undocumented, so we are simply interpreting
**                           things more accurately.
** 071 GES 20050630   @21578 Modified COMMAND_TOKENS processing to rewrite
**                           any non-STRING and non-KW_VALUE children as
**                           a COMMAND_TEXT token which eliminates
**                           downstream problems.
** 072 GES 20050706   @21668 Added missing oldtype annotation for entered
**                           and not entered builtin functions.
** 073 GES 20050719   @21737 Fixed minor bug in the printer rule and
**                           implemented major tree cleanup for the
**                           io_stmt rule and it's referenced rules.
**                           Many extraneous nodes are now dropped and
**                           some processing is improved. In addition,
**                           some important changes occurred in the lexer
**                           to provide support for relative . and ..
**                           entries in filenames.
** 074 GES 20050725   @21824 Added a case into the NUM_LITERAL to handle
**                           certain missing negative DEC_LITERAL cases.
** 075 ECF 20050725   @21825 Added an "export" mode to the lexer. Supports
**                           parsing schema data export files (*.d).
** 076 GES 20050725   @21827 Shifted to using a TokenStream filter to
**                           hide lexer tokens that used to be skipped.
**                           This allows certain places in the parser
**                           to query things like whitespace without
**                           these tokens being visible normally. Both
**                           the lexer and the parser were changed, though
**                           direct users of the lexer won't see any
**                           differences and any user of the
**                           ProgressParser will get the results
**                           transparently.
** 077 ECF 20050726   @21829 Enhanced export mode added in 075 (@21825).
**                           Added raw string processing mode to lexer.
**                           Allows scanning of "raw" Progress binary data
**                           which has been encoded into a string format
**                           during an export. Added RAW_STRING token type
**                           for such strings. Currently, no decoding is
**                           performed and the string is embedded in a
**                           token exactly as read, except that enclosing
**                           double quotes are discarded.
** 078 GES 20050811   @22116 Improved oldtype processing for methods and
**                           attributes (which previously were being
**                           forced to SYMBOL and losing keyword
**                           knowledge).
** 079 GES 20050819   @22142 Added token types to be used only downstream
**                           during annotations (as special constants).
**                           This is a handy way to encode values as
**                           constant names which can be shared in parts
**                           of the process that are separate but which
**                           must have a common understanding of options.
**                           These changes are provided to support
**                           transactions and block properties.  Other
**                           changes were made to root all triggers at
**                           a TRIGGER_BLOCK node and all editing blocks
**                           EDITING_BLOCK.  This dedicates INNER_BLOCK
**                           to DO, REPEAT and FOR blocks.
** 080 GES 20050822   @22188 Added 'recordtype' annotations to all fields
**                           and records.  This allows each database
**                           reference to be distinguished from temp-table
**                           and work-table references (this has an
**                           impact upon whether transaction support is
**                           activated or not, among other things).
** 081 GES 20050822   @22190 Lexer change to increase lookahead depth to
**                           4 and use this to disambiguate NUM_LITERAL
**                           from COMMENT when there is a ./* or ../*
**                           construct.
** 082 GES 20050822   @22194 Converted retry and transaction global vars
**                           into FUNC_LOGICAL.
** 083 GES 20050823   @22200 Added many method and attribute keywords
**                           and rewrote the rule to match these 
**                           constructs so that all method parameters are
**                           properly recognized (brackets can be used
**                           in place of parens) and these parameters are
**                           attached as children of the method node.
** 084 GES 20050830   @22432 Implemented specific data types for methods
**                           and attributes.
** 085 GES 20050907   @22603 Added length language statement (one of the
**                           assignment syntax types).
** 086 GES 20050922   @22779 Added token types for the different buffer
**                           references.  These can be used as constants
**                           in conversion processing, though they are
**                           not used in the parser.  Added a new
**                           annotation "bufname" to all schema nodes.
**                           Many tree structure changes were made to
**                           rules related to database processing.
** 087 GES 20051003   @22929 Updates to cleanup buffer processing. Most
**                           importantly, we now use the record rule
**                           from lvalue in the case where something
**                           is a buffer.  In addition, the non-define
**                           forms of buffer parameters (e.g. run
**                           statements) use lvalue instead of the full
**                           def_buf_stmt rule. This allows the buffer
**                           reference to be emitted more easily.
** 088 GES 20051004   @22932 Added logical database name lookups to all
**                           record references.  The results are stored
**                           as annotations.
** 089 GES 20051013   @23056 Moved field annotation logic to symbol
**                           resolver to facilitate reuse. Fixed OR
**                           operator processing (tree build made the
**                           top level EXPRESSION node the child of the
**                           KW_OR). Removed the stream_reference rule
**                           and converted all usage into a semantic
**                           predicate protected call to lvalue.  lvalue
**                           was enhanced to handle stream references
**                           with the KW_STREAM prefix.  This eliminates
**                           a bug where stream references could be
**                           mistaken for field names due to abbreviation
**                           conflicts (since the KW_STREAM was not used
**                           in lvalue to disambiguate.  By moving this
**                           processing back into lvalue, the problem is
**                           solved, BUT at the cost of having to limit
**                           certain usage locations by a semantic
**                           predicate (which only calls lvalue if the
**                           next token is KW_STREAM).
** 090 GES 20051020   @23077 Added token types for index matching
**                           annotations.
** 091 SIY 20051024   @23092 Added FRAME_ALLOC token type.
** 092 GES 20051026   @23109 Added QUERY_SUBST token type to root query
**                           substitution expressions.
** 093 GES 20051030   @23146 Added variable annotations to widgets since
**                           they are really a special kind of variable
**                           reference.
** 094 GES 20051031   @23150 Added table promotion for format phrase
**                           (really needed for the validate rule) and
**                           reworked the def_frame_stmt to drop the
**                           KW_FRAME since it is not needed.  Also made
**                           the field_clause return the widget and
**                           format_phrase nodes as peers rather than
**                           having the widget as the parent.  This is
**                           compatible with other rules so subsequent 
**                           processing is easier.
** 095 GES 20051101   @23152 Fixed a long-standing issue where record
**                           names get precedence over variable names.
**                           This was previously done to ensure that
**                           records are preferred over fields, but it
**                           turns out that this was wrong in that
**                           variable names will hide a record that 
**                           would be matched by the same name
**                           (abbreviated or not).
** 096 GES 20051102   @23200 Added COPY_TO_SB and COPY_FROM_SB types,
**                           removed unneeded KW_BROWSE, KW_CLEAR.
** 097 GES 20051107   @23279 Cleanup and reparenting of COLOR_PHRASE.
**                           Tree building changes for the message
**                           statement, forcing known ordering where
**                           possible. Added new type for key functions.
**                           Massive changes to fix entered/not entered
**                           and some uses of the INPUT function. This
**                           caused major disruption of the way in which
**                           frame references are sometimes prefixed to
**                           lvalues and had downstream impact on all
**                           types of assignments.  The result is a much
**                           cleaner tree.
** 098 GES 20051109   @23281 More tree cleanup and annotations for UI.
**                           In particular, button and browse references
**                           are cross-referenced to their original
**                           defining node (DEFINE_BUTTON or 
**                           DEFINE_BROWSE).  The underline stmt now
**                           parents its field list in a manufactured
**                           node.
** 099 GES 20051118   @23413 Added a VALIDATION token type.
** 100 GES 20051121   @23431 Made INPUT parm processing more restrictive. 
** 101 GES 20051122   @23432 Reworked GO-ON clause to allow re-merge of
**                           tokens that were separated because of our
**                           lexer approach but really should have been a
**                           single token (based on whitespace).
** 102 GES 20051123   @23442 Tree cleanup for column format phrases (in
**                           browse), now these are parented in a
**                           FORMAT_PHRASE node. Also fixed bug in
**                           format_phrase processing for field_clause.
** 103 GES 20051124   @23457 Added WHEN_LIST type.
** 104 GES 20051128   @23542 Excluded widget preference from subscripts.
** 105 ECF 20051206   @23673 Refined CONNECT_OPTIONS processing. Normalize
**                           children of this node from various token
**                           types to CONNECT_TEXT to allow downstream
**                           merge text processing.
** 106 GES 20051208   @23674 Merged up_stmt and down_stmt into the new
**                           up_or_down_stmt since both have the same
**                           exact structure and the old down_stmt had
**                           a sub-optimal coding for the stream ref.
** 107 GES 20060112   @23870 Added RUNTIME_FRAME_OPTIONS token.
** 108 GES 20060112   @23888 KW_OR processing fixed to not drop middle
**                           ORs when there are more than 2 top level
**                           ORs.
** 109 ECF 20060125   @24055 Drop FOR keyword from temp table parameter
**                           definition AST. It is unnecessary filler.
** 110 ECF 20060128   @24256 Added support for temp table record scoping
**                           downstream processing. Added 'schemaname',
**                           'bufname', and 'dbname' annotations to the
**                           define temp table AST.
** 111 GES 20060205   @24279 New token types for frame element processing.
** 112 GES 20060208   @24363 Allow arbitrary order for selection list
**                           options. Fixed mis-spelling of CARET.
** 113 ECF 20060216   @24611 Allow zero or more IS clauses on a temp table
**                           index. Previously set to zero or one.
** 114 GES 20060220   @24616 Added new browse attribute and method.
** 115 ECF 20060221   @24704 Support client-side where clauses. New token
**                           types CLIENT_WHERE and CLIENT_WHERE_SUBST.
** 116 ECF 20060226   @24820 Support field-level validation. Added new
**                           token type SCHEMA_VALIDATIONS.
** 117 GES 20060306   @24914 Added a new token type for customer-specific
**                           use.
** 118 GES 20060401   @25297 Tree building improvement for literals in
**                           set, update or prompt-for statements.
** 119 GES 20060401   @25304 Reduce WIDGET matches in FORM_ITEMs by 
**                           allowing standalone vars and fields to be
**                           matched as expressions.
** 120 GES 20060404   @25340 Reduce WIDGET matches in text clauses and
**                           any other user of field_clause rule (includes
**                           the ENABLE statement.
** 121 ECF 20060406   @25389 Change to def_buf_stmt. Look up qualified
**                           buffer name and store it in the bufname
**                           annotation of the buffer name symbol node.
** 122 GES 20060407   @25411 UP/DOWN stmts can take the stream reference
**                           as a 2nd parameter.
** 123 ECF 20060407   @25427 Remove regression in form_item. Use of
**                           SymbolResolver's lookupField() instead of
**                           isField() was causing an ambiguity error.
** 124 GES 20060420   @25631 Trigger blocks must add and remove scopes in
**                           the symbol and schema dictionaries.
** 125 GES 20060426   @25724 Moved the stream reference in an underline 
**                           stmt out of the content array.
** 126 GES 20060426   @25736 Handle vars are not added to the widget
**                           dictionary since they can never be direct
**                           widgets.
** 127 GES 20060713   @28004 Fixed widget_list loop closure issue.
** 128 ECF 20060818   @28686 Added boolean parameter to record rule. If
**                           true, record's table is promoted, else not.
** 129 GES 20061101   @30861 space_or_skip needed to allow a numeric
**                           constant not just integer.
** 130 ECF 20070219   @32177 Added new token types for schema parsing and
**                           lexing.
** 131 GES 20070220   @32193 Support for private access in procedures.
** 132 GES 20070221   @32203 Fix for FUNCTION IN SUPER which must be
**                           treated as another type of forward
**                           declaration.
** 133 GES 20070221   @32214 Tolerate (and ignore) syntactic sugar in
**                           function call parameter lists (where a
**                           parameter can be prefaced with an INPUT,
**                           INPUT-OUTPUT or OUTPUT keyword that is
**                           meaningless).
** 134 GES 20070308   @32292 Added NO-OVERRIDE keyword for schema support.
** 135 GES 20070308   @32346 Added FUNCTION char attribute.
** 136 GES 20070309   @32347 Function definitions IN proc_handle must be
**                           treated as forward declarations.
** 137 GES 20070309   @32351 Added FORMAT and FILE-TYPE char attributes.
** 138 GES 20070309   @32352 Added various editor widget attributes and
**                           methods.
** 139 GES 20070315   @32410 Added browse attributes. Fixed issue with
**                           LAST-KEY used as a function call parameter.
**                           Removed trash in frame phrase that is 
**                           silently ignored in Progress. Now honors
**                           seek(input) and seek(output) built-in 
**                           function calls (where input and output are
**                           specially named references to the unnamed
**                           streams).  Allowed PUT SCREEN to have the
**                           KW_COLUMNS used instead of KW_COL since
**                           the 4GL is tolerant of this.
** 140 GES 20070319   @32500 Javadoc fix.
** 141 GES 20070321   @32534 Added attributes for window, frame and other
**                           widgets. Fixed menubar keyword processing. 
**                           Fixed widget-pool processing (they take
**                           expressions not symbols and the construct
**                           in widget-pool needed to be excluded from
**                           lvalue widget qualifier processing). Cleaned
**                           up widget-pool tree building.
** 142 GES 20070323   @32562 Added undocumented feature to referencing
**                           phrase (used in a database trigger). Fix
**                           for regression in put screen (cut and paste
**                           mistake where the KW_COL was being set as the
**                           root of the subtree. Added lower as a synonym
**                           for lc since this is an undocumented form of
**                           the same built-in function.
** 143 GES 20070329   @32656 Added a special mode for subscript processing
**                           when called from within a like clause. This
**                           allows the Progress behavior of treating all
**                           like array[expr] subscript expressions as if
**                           they where the number 1.
** 144 GES 20070402   @32717 Support space and skip as column specs in
**                           define browse (this is undocumented). Added
**                           support for more undocumented junk as empty
**                           language stmts. Allow PROMPT as a pseudonym
**                           for PROMPT-FOR. Removed the approach in
**                           put screen and frame phrase for handling the
**                           use of COLUMNS as COLUMN and made a generic
**                           approach which is implemented in ui_stuff.
**                           This fixes a regression in frame phrase
**                           since the prior fix was not right.
** 145 GES 20070404   @32792 Fixed a problem with nested triggers where
**                           schema scopes would be improperly marked
**                           as global instead of simply restoring the
**                           prior value at the end of the trigger. Added
**                           support to notify the schema dictionary of
**                           the type of a scope when a scope is added.
**                           This is needed to differentiate internal
**                           procedure, user-defined functions and
**                           triggers from other scopes.
** 146 GES 20070405   @32794 Fixed sequence support (both the lang stmt
**                           and the functions) to match the sequence
**                           name as a symbol but convert it to a string.
** 147 GES 20070405   @32795 Allow the when clause to be optionally used
**                           in an underline stmt. This is an undocumented 
**                           feature.
** 148 GES 20070405   @32830 Added more attributes. Added some simple
**                           support for embedded SQL.
** 149 GES 20070409   @32873 Handle frame name quirk. Frame names are
**                           symbols that can start with characters other
**                           than a letter and underscore.
** 150 GES 20070411   @32925 SPACE and SKIP constructs are allowed but
**                           silently ignored in export stmts (this is
**                           an undocumented feature).
** 151 GES 20070411   @32926 Added more browse methods.
** 152 GES 20070411   @32937 Added support for undocumented feature of
**                           put stmt where space() or skip() can contain
**                           an expression rather than a literal. This
**                           means that there would have to be a runtime
**                           calc of the result. Note that display stmts
**                           can't do this same thing so it is limited in
**                           application (a comprehensive set of tests was
**                           not run so the possible list of stmts that
**                           are affected is unknown).
** 153 GES 20070411   @32941 Added support for undocumented feature of
**                           the underline stmt where it has an empty
**                           frame field list.
** 154 GES 20070411   @32942 Added undocumented synonym upper for caps.
**                           Also added an undocumented feature to allow
**                           any number of trailing NO-ERROR keywords
**                           on a run statement.
** 155 GES 20070412   @32957 Support undocumented schema name quirk where
**                           a field name can have whitespace injected
**                           between the record name portion and the dot.
** 156 GES 20070413   @32975 Added support for postfixed unary minus and
**                           plus in the lexer.
** 157 GES 20070416   @33007 Moved trigger scope processing out of 
**                           single_block and into the rules for trigger
**                           processing since buffers can be referenced
**                           before the single_block rule is matched and
**                           thus ambiguity can occur.
** 158 GES 20070416   @33011 Added support for undocumented trash in a
**                           widget phrase.
** 159 ECF 20070417   @33038 Fixed NPE. Added safety code to parser's
**                           followedByWhitespace() method.
** 160 GES 20070417   @33058 Added support for undocumented trash in a
**                           using clause (the INPUT keyword like H158).
** 161 GES 20070418   @33096 Support for empty THEN at the end of a block
**                           or the file.
** 162 GES 20070418   @33109 Added support for COM-HANDLE data type,
**                           create/release stmts for active-x object
**                           access and general support for active-x
**                           methods and properties in expressions and
**                           assignments.
** 163 GES 20070419   @33158 Added support for DOM-based XML processing
**                           (methods and attributes).
** 164 GES 20070420   @33176 Made a custom rule to match dynamic-function
**                           function calls since it has non-expression
**                           syntax in its parameter list.
** 165 GES 20070423   @33181 Added support for all browse methods and
**                           attributes.
** 166 GES 20070423   @33206 Added support for an undocumented browse
**                           option and an undocumented key function.
**                           Added other lock types to empty stmt. Added
**                           Progress 4GL where clauses to embedded SQL
**                           stmts. Added support for the def image stmt.
**                           Added all possible image attrs and methods.
** 167 GES 20070426   @33290 Implemented manual arbitrary lookahead to
**                           disambiguate labels that start with a number,
**                           hyphen or other non alpha character. Based 
**                           on this, manual matching and tree building
**                           may occur. COLON processing is now cleaned
**                           up to use the hidden token stream instead
**                           of appending a space to the end of the text.
**                           Some changes were made to the prediction
**                           logic by inserting fake pull entries at the
**                           top level of the parser (single_block).
**                           Label references are now updated to handle
**                           the malformed symbols, just like the label
**                           definitions. All of this malformed label
**                           processing is undocumented Progress gorp.
** 168 GES 20070430   @33305 Included all non-printable ASCII codes as
**                           whitespace in the lexer.
** 169 GES 20070501   @33322 Tweaked frame phrase to avoid the transaction
**                           keyword that may follow it in a block header.
** 170 GES 20070502   @33361 Support procedure names that contain a '.'
**                           character (or multiple instances of it). 
**                           Fixed a prediction problem in callers of
**                           label_reference by checking for a match at
**                           the beginning of the rule and matching the
**                           empty alternative if a properly formed or
**                           malformed label reference is not present.
** 171 GES 20070502   @33363 Fixes for attribute/method matching and
**                           the semantic predicate for the create stmt.
** 172 GES 20070503   @33404 Doc updates and some sample code added.
**                           Improved support for internal procedures
**                           which wrapper native calls to a shared lib.
**                           Add all remaining methods/attributes for
**                           this-procedure, buffers, buffer-fields.
**                           Added undocumented support for no-error
**                           in some language stmts. Added the release
**                           external stmt.
** 173 GES 20070504   @33405 Added undocumented use of a string as a
**                           filename. Added remaining meths/attrs for
**                           sockets and file-info.
** 174 GES 20070504   @33406 Added trigger procedure stmt.
** 175 GES 20070511   @33468 Fix for empty_stmt. Added undocumented
**                           feature to export stmt (it can be empty).
**                           Added undocumented support for internal
**                           procedure names that are reserved keywords.
** 176 GES 20070521   @33742 Added remaining raw/memptr features (assign
**                           type lang stmts where the only ones
**                           remaining). Changed source/target codepage
**                           spec in I/O options to take an expression
**                           (it is always a character type) instead of
**                           a filename (which was not valid in that
**                           case).
** 177 GES 20070524   @33744 Finished named events support (subscribe and
**                           unsubscribe).
** 178 GES 20070528   @33787 Modify support for H133 to only ignore the
**                           INPUT keyword in parameters for user-defined
**                           functions. Built-in functions don't ever use
**                           modifying keywords since there is no
**                           definition in the application source. So for
**                           built-ins, Progress assumes that any INPUT
**                           keyword is really the INPUT built-in.
** 179 GES 20070529   @33881 Added a large number of methods/attributes
**                           relating to the session handle, procedure
**                           handle usage, transactions and appservers.
**                           Added run super and stored procedure related
**                           statements.
** 180 GES 20070601   @33950 Added all remaining language stmts from 9.1C.
**                           This includes call, dde*, disable triggers,
**                           all the Windows registry stmts, load picture,
**                           transaction mode automatic, system-help and
**                           the Windows system-dialog* stmts.
** 181 GES 20070604   @33951 Improved tree structure of procedure stmt.
** 182 GES 20070608   @33997 Added some methods/attrs, fixed javadocs and
**                           added some comments.
** 183 EVL 20070617   @34070 Added new attributes to parse the clipboard
**                           system handle.
** 184 EVL 20070618   @34084 Fixed spelling for available-formats keyword.
**                           Adding the keyword "value".
** 185 EVL 20070618   @34093 Added parsing ability for COLOR-TABLE,
**                           COMPILER, DEBUGGER, FONT-TABLE, LAST-EVENT
**                           system handles.
** 186 EVL 20070619   @34119 Added parsing ability for Query object,
**                           Temp-Table object and RCODE-INFO system
**                           handles.
** 187 EVL 20070620   @34188 Added parsing ability for WEB-CONTEXT system
**                           handle.
** 188 EVL 20070621   @34195 Added button, combo-box widgets parsing
**                           handling.
** 189 EVL 20070622   @34205 Added parsing support for CONTROL-FRAME,
**                           DIALOG-BOX, EDITOR widgets.
** 190 EVL 20070622   @34218 Added parsing support for FIELD-GROUP,
**                           FILL-IN, FRAME, LITERAL, MENU widgets.
** 191 EVL 20070624   @34225 Added parsing support for MENU-ITEM, SLIDER,
**                           RADIO-SET, RECTANGLE, SELECTION-LIST widgets.
** 192 EVL 20070624   @34226 Added parsing support for SUB-MENU, TEXT,
**                           TOGGLE-BOX, WINDOW widgets.
** 193 EVL 20070626   @34271 Adding parsing support for keywords from A to
**                           DISCONNECT.
** 194 EVL 20070627   @34292 Adding parsing support for keywords from after
**                           DISCONNECT up to Q.
** 195 EVL 20070628   @34298 Adding parsing support for keywords from R to
**                           Y(end of Language Reference manual).
** 196 GES 20070628   @34319 Additional cleanup based on a review of the
**                           new keywords. The result of the recent
**                           keyword and attribute/method work is that
**                           the parser is now complete in these areas up
**                           through the 9.1E level. This includes 
**                           documented constants, SQL keywords and other
**                           documented keywords that are not otherwise
**                           documented as part of the 4GL.
** 197 GES 20070629   @34345 Changed symbol resolver calls to pass through
**                           ASTs to allow the schema dictionary to set
**                           line and column info for temp-table ASTs
**                           created during parsing.
** 198 GES 20070717   @34589 Fix for filename (it needed better pull
**                           rules for some cases where the "rebuild" of
**                           a windows filename is active). Fix for
**                           save cache stmt where the 2nd parameter is
**                           a string instead of a filename.
** 199 GES 20070720   @34648 Moved the keyword list in the lexer into a
**                           static member. This allows non-conversion
**                           code to use this list.
** 200 GES 20070720   @34650 Add undocumented feature to view stmt to
**                           allow a list of widgets rather than just
**                           one.
** 201 GES 20070730   @34729 Keyword cleanup. At this point, all keywords
**                           documented (or even referenced) in the 9.1E
**                           language reference are present in the lexer.
**                           Minor parser fixes and added documentation.
** 202 EVL 20070731   @34746 Adding keywords support for 10.1B attribute
**                           and methods completely documented from
**                           ADD-SCHEMA-LOCATION up to DECLARE-NAMESPACE.
** 203 EVL 20070801   @34756 Adding keywords support for 10.1B attribute
**                           and methods completely documented from
**                           DELETE-CHARACTER up to GET-INDEX-BY-QNAME.
**                           Several new data type token were also added.
** 204 GES 20070802   @34759 Added more complete embedded SQL support
**                           for full SELECT syntax.
** 205 GES 20070802   @34759 Added SQL support for ALTER TABLE, CLOSE
**                           (cursor) and CREATE INDEX.
** 206 EVL 20070802   @34781 Adding keywords support for 10.1B attribute
**                           and methods completely documented from
**                           GET-LOCALNAME-BY-INDEX up to LOG-AUDIT-EVENT.
**                           PLus new data type tokens again.
** 207 GES 20070803   @34783 Added SQL support for CREATE TABLE, CREATE
**                           VIEW, DELETE FROM, DROP INDEX, DROP TABLE,
**                           DROP VIEW, FETCH AND GRANT.
** 208 EVL 20070803   @34784 Adding keywords support for 10.1B attribute
**                           and methods completely documented from
**                           LOGIN-EXPIRATION-TIMESTAMP up to
**                           SAX-PARSE-NEXT. Adding new datatype tokens
**                           ATTR_DATETIME_TZ and METH_DATETIME_TZ.
** 209 EVL 20070806   @34785 Adding keywords support for 10.1B attribute
**                           and methods completely documented from
**                           SCHEMA-LOCATION up to USER-ID.
** 210 EVL 20070806   @34786 Adding keywords support for 10.1B attribute
**                           and methods completely documented from
**                           VALIDATE-SEAL up to
**                           XML-SUPPRESS-NAMESPACE-PROCESSING.
** 211 GES 20070806   @34789 Added SQL support for DECLARE CURSOR, INSERT
**                           INTO, OPEN CURSOR, REVOKE, UNION and SQL
**                           UPDATE statements.
** 212 GES 20070806   @34794 Added SQL operators to expression processing
**                           and support for where clause sub-selects.
**                           SQL support is 100% code-complete, yet very
**                           lightly tested at this time.
** 213 EVL 20070806   @34795 Adding keywords support for 10.1B attributes
**                           and methods partially documented in language
**                           reference 10.1B(not in index cases) from
**                           ACCEPT-CHANGES up to FORWARD-ONLY.
** 214 EVL 20070807   @34802 Adding keywords support for 10.1B attributes
**                           and methods partially documented in language
**                           reference 10.1B(not in index cases) from
**                           GET-CALLBACK-PROC-CONTEXT up to
**                           MUST-UNDERSTAND.
** 215 EVL 20070808   @34817 Adding keywords support for 10.1B attributes
**                           and methods partially documented in language
**                           reference 10.1B(not in index cases) from
**                           NESTED up to WHERE-STRING.
** 216 EVL 20070808   @34823 Adding keywords support for 10.1B new
**                           language statements, handles and events.
** 217 EVL 20070809   @34831 Adding keywords support for 10.1B new
**                           built-in functions. Also added several new
**                           data type tokens for return values.
** 218 EVL 20070809   @34838 Adding keywords support for 10.1B new items
**                           can not be found in language reference
**                           outside the global keyword Index.
** 219 EVL 20070810   @34840 Several fixes and additions after
**                           verification the global keywords index in
**                           10.1B language reference book.
** 220 GES 20070813   @34844 Fix for regression caused by embedded SQL
**                           expression support. Fix for misplaced
**                           artificial token KW_TAG (in reserved section
**                           when it should have been unreserved).
** 221 GES 20070816   @34892 Preload webspeed API facilities as built-in
**                           resources (functions, procedures, streams).
**                           This allows webspeed code to parse without
**                           using the PSC 4GL code that defines and
**                           implements these APIs. Since the webspeed
**                           API is documented but is implemented in 4GL
**                           code, this approach is quite different from
**                           real built-ins that are in the 4GL runtime.
**                           Added undocumented parsing support for the
**                           GO-ON clause.
** 222 GES 20070907   @35008 Added support for all 10.1B datatypes to
**                           functions, attributes, methods, variables
**                           fields and the respective language stmts
**                           that define these resources.
** 223 ECF 20070911   @35047 Added schema keywords. Support field triggers
**                           and table dump and load permissions.
** 224 ECF 20070912   @35051 Added support for obscure Progress quirks:
**                           - an editing block label definition can be
**                             embedded in a frame phrase
**                           - the UNDO keyword is optional in an ON QUIT
**                             phrase
**                           Also fixed using_clause to move the optional
**                           INPUT keyword before the optional frame
**                           reference.
** 225 GES 20070911   @35137 Added support for all 10.1B object oriented
**                           features. This required massive changes to
**                           add language statements, enhance expression
**                           processing, protect static data, resource
**                           resolution, token types and top level
**                           structure. A few minor fixes have been added
**                           as well: GROUP is a synonym for BREAK in
**                           block headers, AS clause was refactored to
**                           match undocumented syntax in forward function
**                           parameter definitions and some tokens were
**                           incorrectly defined in the wrong token type
**                           ranges.
** 226 GES 20070923   @35185 Added keyword ignore support to the lexer
**                           which is equivalent to the Progress -k
**                           startup parameter (honors a keyword "forget"
**                           file).
** 227 GES 20070924   @35200 Reworked the filename processing to match
**                           all possible valid filenames. The previous
**                           approach was very limited. The new approach
**                           parses based on whitespace and the DOT.
**                           Also added 10.1B SAX related lang stmts.
** 228 GES 20070925   @35242 Allowed DOT tokens that are not followed by
**                           whitespace to be embedded in OS command
**                           strings. Rewrite of frame phrase down clause
**                           disambiguation to be more complete. Browse
**                           column lists can be empty now (undocumented
**                           feature). String lexing must properly ignore
**                           backslash if UNIX escape processing is not
**                           supported.
** 229 GES 20070927   @35256 The assign statement needed to honor the
**                           qualified field name quirk (where whitespace
**                           can be between the record name and the
**                           following .fieldname portion. Changes to
**                           promotion behavior in the schema dictionary
**                           forced additional information to be passed
**                           on the method calls.
** 230 GES 20070927   @35311 Removed program_arguments and moved to a
**                           common command_token_list implementation.
**                           This was required to allow the matching
**                           of DOT tokens that aren't followed by
**                           whitespace. Note that the core loop had to
**                           be made greedy again for that to work. That
**                           change may cause breakage elsewhere. Index
**                           definition in temp-tables and use-index
**                           processing can now accept index names that
**                           have embedded slash characters (which are
**                           not allowed in field or table names but are
**                           allowed in index names). Embedded SQL select
**                           stmt processing was fixed in many ways. Of
**                           most importance is that table aliases can
**                           be referenced in field lists before the
**                           alias is defined in the from clause. These
**                           symbols are fixed up as fields after the
**                           aliases have been setup. Aggregate phrases
**                           are now silently ignored in the export stmt.
**                           Temp tables can be accessed as an lvalue for
**                           dereferencing attributes or methods.
** 231 GES 20071009   @35403 Added language stmt support from 10.1B for
**                           security, SOAP, 64-bit sequences, I18N and
**                           the new SYSTEM-DIALOG GET-DIR.
** 232 GES 20071011   @35528 Final 10.1B support. COPY-LOB, CREATE CALL,
**                           extent stmt, all ProDataSet features and
**                           handle usage/parameter usage updates. Some
**                           other options in INPUT FROM, OUTPUT TO and
**                           format phrase.
** 233 GES 20071030   @35899 Added lexer constructors to allow the dot
**                           kludge to be disabled. Fixed relex() to
**                           use the new constructor and also removed
**                           a problem where we assumed that the given
**                           string could only have double quotes. Some
**                           fixes for NOW support and the start of
**                           undocumented datetime and datetime-tz literal
**                           support. Added additional "pull" tokens into
**                           the filename rule to handle cases where a
**                           filename that needed merging would not be
**                           matched because of the lookahead calcs in the
**                           calling rule. Fixes for delete object and
**                           run statements. Modified column specs in
**                           select statements to be expressions rather
**                           than lvalues. Eased position based parsing
**                           for AS and LIKE clauses in a define parm
**                           stmt. These can appear after other options.
**                           Added lexer support for '.' as a date
**                           literal separator. Allow multiple frame
**                           phrases including any number of empty ones
**                           which are intermixed with non-empty ones.
**                           MTIME can be used without parenthesis and
**                           FRAME-COL/FRAME-DOWN/FRAME-LINE/FRAME-ROW can
**                           be called with an invalid frame name.
** 234 GES 20071126   @36025 Fix for regression with frame functions (all
**                           of them had their oldtype incorrectly forced
**                           to KW_FR_DOWN).
** 235 GES 20071129   @36079 Fixed incomplete form item parsing to always
**                           properly differentiate (and handle) inline
**                           var defs (where a symbol or an expression
**                           are peer alternatives). The previous approach
**                           could not match such cases without a hint or
**                           the LA(1) being AS or LIKE.
** 236 GES 20071204   @36186 Misc fixes and undocumented features. Added
**                           missing keywords in buffer compare/copy and
**                           codepage processing. Added implicit var def
**                           during accumulate stmt for count and total.
**                           The rowid keyword is not reserved even though
**                           it is documented as reserved. An aggregate
**                           phrase can be embedded inside a format phrase
**                           (it would be best to have it be included in
**                           the format phrase rule but that would require
**                           downstream conversion changes). Changed 
**                           lvalue to handle the full set of parm options
**                           such as "TABLE record".
** 237 GES 20071205   @36209 Certain record references (FOR stmt, DO
**                           PRESELECT and REPEAT PRESELECT) force
**                           re-promotion. Normal record references only
**                           promote once but these cases are all queries
**                           associated with blocks and they have special
**                           behavior. Note that OPEN QUERY and FIND have
**                           been tested and have no such behavior. The
**                           other record references have not been 
**                           explicitly tested, but generally it is
**                           assumed that they do not force re-promotion.
** 238 GES 20071207   @36287 can-find disables all promotion while it
**                           is active (even in the where clause...).
**                           Weak scopes that have a buffer associated
**                           (as a weak reference) must not propagate any
**                           references to that buffer. Strong scopes
**                           force promotion even if the table was 
**                           previously promoted.
** 239 GES 20080318   @37666 Changed approach to read the entire unicode
**                           charset instead of dealing with bytes. This
**                           makes it safe to process files with I18N 
**                           requirements.
** 240 GES 20080404   @37843 Some annotations were inappropriately being
**                           set when a display stmt had an inline var
**                           definition. Resetting initial values each
**                           time through the loop resolved the issue.
** 241 GES 20080701   @39056 Added NEW-LINE keyword and key-function.
**                           Added compiler constants for window-state
**                           as literals (they can be used freely). Added
**                           more undocumented keywords. Added DELETE-LINE
**                           and APPEND-LINE key-functions. Moved
**                           undocumented INPUT keyword into widget_phrase
**                           from a caller since that bogus keyword seems
**                           to be a more generic "feature" than
**                           previously thought.
** 242 CA  20080915   @39805 Added RUNTIME_BROWSE_OPTIONS token.
** 243 GES 20081007   @40356 Added new undocumented keywords and global
**                           vars. Added unquoted text support for
**                           certain cases. Fix for in procedure clause,
**                           widget phrase, view-as slider, title clause.
**                           Added undocumented key functions.
** 244 GES 20090428   @42024 Moved lexer and parser test drivers into main()
**                           for each class respectively.
** 245 GES 20100119   @44542 Provide a hook method to close resources.
** 246 GES 20100809          Added schema keywords. Added an option to allow
**                           calling code to force the parser to abort and
**                           expose any exceptions instead of attempting to
**                           auto-restart parsing.
** 247 SIY 20101002          Added proper handling for RECTANGLE widget.
** 248 GES 20110303          Cleaned up documentation and added comments.
** 249 GES 20110815          Added INACTIVE schema keyword. Added undocumented
**                           abbreviation support. Fixed VALIDATE option in
**                           temp-table field definitions. Fixed ambiguity
**                           between SUPER function and SUPER stmt. Fixed the
**                           IN WINDOW clause to match an lvalue instead of
**                           a symbol (otherwise things like ACTIVE-WINDOW
**                           cannot be used in this clause). Added undocumented
**                           use of VALUE() as a peer alternative to expr
**                           in RUN stmt preprocessor arguments. Allowed decimal
**                           literals in some places that only integer literals
**                           were previously allowed. Added the undocumented
**                           MARGIN-EXTRA option to the define button stmt.
**                           Fixed parsing of some cases for dynamic-function
**                           built-in when the first parm is an lvalue. Fixed
**                           com_parameter to support some optional keywords.
**                           Added support for the undocumented NOP mode in
**                           enable and disable stmts. This is where there is
**                           no ALL keyword or field list. Added missing IMAGE
**                           keyword to create_widget_stmt. Fixed the attribute
**                           assignment section of the create_browse_stmt.
**                           Added undocumented synonym reposition-backwards.
**                           Fixed undocumented expression usage in 
**                           delete_object_stmt. Added to the disambiguating
**                           semantic predicate in the assignment rule to avoid
**                           an lvalue followed by a binary operator. This case
**                           was not being handled naturally by the ANTLR
**                           lookahead. Changed in_procedure_context_clause
**                           to match an lvalue instead of a symbol. Added an
**                           undocumented keyword match to image_phrase. Added
**                           an undocumented logical attribute (used on browse
**                           widgets) called CREATE-ON-ADD. Query, proc and func
**                           names can be a malformed_symbol. The handle rule's
**                           semantic predicate test was enhanced to add other
**                           valid values. The in_window_clause was changed from
**                           using an lvalue to an expr since real code has
**                           shown that complex expressions are possible. Added
**                           an undocumented feature to the editor phrase (the
**                           inner-chars keyword can appear without the
**                           following inner-lines. For create_widget_pool_stmt
**                           the widget pool name was made independent of the
**                           KW_PERSIST. This is an undocumented feature. Made
**                           KW_V6FRAME independent of the other keywords as
**                           previously used in the version_6_frame_clause rule.
**                           This is an undocumented feature found in working
**                           4GL code. Added a synonym for CONTROL-FRAME which
**                           is CONTROL-CONTAINER. That is an undocumented
**                           keyword found in working 4GL code.  Added a quick
**                           and dirty check to see if a handle was ever
**                           assigned as an active-x control.  If so, save the
**                           variable or field name in a list of active-x
**                           handles. All subsequent use of the COLON operator
**                           on that lvalue will match com_property_or_method
**                           instead of attribute_or_method.  This is not
**                           foolproof, but it is expected to work in the vast
**                           majority of cases. create_xml_stmt now supports
**                           no-error (undocumented option found in working
**                           code). def_browse_stmt now modifies the widget
**                           namespace earlier so that later portions of the
**                           same statement can properly lookup the widget name.
**                           Reused a flag to control widget_qualifier matching
**                           of procedure handles for subscribe/unsubscribe.
**                           Support reposition-forwards which is an
**                           undocumented synonym for reposition-forward. Allow
**                           an undocumented form of the connected() built-in
**                           which takes a symbol (for the dbname) instead of
**                           a valid expression. Allow an undocumented for of
**                           the value() clause which takes unquoted text as
**                           if it was a string. Allow the undocumented usage
**                           of an aggregate_phrase in a form_item (mixed in
**                           with the format phrase the same way that it is
**                           possible in an export_field or display_stmt. The
**                           temp_table_field_options are now allowed before
**                           the AS or LIKE clause in a def temp table. This
**                           is an undocumented feature. Added key functions
**                           in the long list in the old programming handbook
**                           which is now gone. These are now known to be
**                           assignable to keys using the on stmt. That was
**                           unknown before because of the confusing way the
**                           docs were written.
** 250 GES 20110816          Added support for some 10.2x features as needed by
**                           customer code. Support the DB_REF_NON_STATIC type
**                           and processing for the new :: syntax. Added the
**                           default-string and default-value attributes.
** 251 GES 20110819          Added support for more 10.2x features as needed by
**                           customer code. Updated using_clause to support the
**                           FROM syntax. Added ABSTRACT option to class_stmt.
**                           Added STATIC option to all forms of DEFINE stmts,
**                           method and constructor.  Added pre_scan_class
**                           top-level entry point for a simple first pass to
**                           load the resolver with the class and method names.
**                           This enables out-of-sequence references to methods
**                           (in expressions) before their definitions appear
**                           in the file. Better disambiguation of THIS-OBJECT
**                           reference/THIS-OBJECT stmt and the NEW object
**                           func vs NEW record func. Added the finally_stmt, 
**                           catch_stmt, routine_level_stmt and throw support
**                           for undo_stmt and on_event_phrase. These features
**                           provide the core of the new structured error
**                           handling. Allow the string form of the using_stmt.
**                           Added define event and the class based event
**                           publish/subscribe/unsubscribe. Added first-form
**                           and last-form attributes. Property, class, event
**                           and method names can be reserved keywords. Added
**                           active-form system reference (like a global var).
**                           Added support for unqualified and classname-
**                           qualified static references. Added box, unbox,
**                           dynamic-cast and dynamic-invoke functions. Added
**                           dynamic-new stmt. Added support for the new form of
**                           wait-for with new .NET method and SET option.
**                           Added db-remote-host function. Added new CLOB
**                           option keywords for schema processing.
** 252 GES 20111128          Added the undocumented PASSWORD-FIELD option to the
**                           format_phrase. Matched undocumented behavior in
**                           def_menu_stmt where Progress will silently drop
**                           an extra "." character if the period is followed
**                           by the MENU-ITEM keyword. Convert any non-quoted
**                           format to a valid string in format_string. Allow
**                           FILENAME to be matched as an lvalue in symbol
**                           matching/expression eval modes. Added undocumented
**                           NO-TAB-STOP to browse options phrase. Added an
**                           undocumented abbreviation to SET-WAIT-STATE that
**                           was found in customer source. Made the
**                           DATA-RELATION optional in a DEFINE DATASET stmt
**                           (another undocumented feature). Added a missing
**                           IN WIDGET-POOL to create_xml_stmt. Added an
**                           undocumented server object attribute LOG-ID which
**                           returns the process ID of the appserver. Added an
**                           undocumented feature where unmatched trailing
**                           right parenthesis are silently dropped in RUN stmt
**                           parameters. Allow the undocumented behavior of
**                           BUFFER parameter names to be unspecified in
**                           forward function declarations. Added support for
**                           an undocumented feature, where PROMPT-FOR is
**                           accepted as a synonym to PROMPT in a color stmt.
**                           TERMINATE keyword had no artificial token (in the
**                           unreserved section) so it was treated as not
**                           unreserved (which disabled its use as a symbol).
**                           Added SAX parser related compiler constants. Moved
**                           browse_options_phrase into def_browse_stmt and made
**                           query_reference and column_spec optional (all of
**                           this is to support undocumented behavior).
**                           Malformed symbols can start with a decimal literal
**                           (found in customer code). Fixed absolute pathname
**                           support for Windows pathnames in run_stmt.
** 253 OM  20121116          Fixed parser for sequences dynamic forms. All 
**                           sequence manipulations use 64bit integers.
** 254 ECF 20121119          Added keywords and attributes for IDE-PARENT-HWND
**                           and IDE-WINDOW-TYPE. These are undocumented
**                           attributes to the CREATE WINDOW statement found
**                           in customer code. These are unreserved keywords
**                           with no possible abbreviations. Added constants
**                           to parser's literal rule for SAX-related compiler
**                           constants.
** 255 CA  20130115          Added support for persistent procedures, 
**                           procedure handles and super procedures.
**                           SEARCH-SELF and SEARCH-TARGET are emitted as
**                           literals.
**                           Fixed parsing of FUNCTION ... IN SUPER and
**                           FUNCTION ... IN handle cases.
** 256 VMN 20130127          Added setting literal type for KW_NO_LOCK,
**                           KW_SH_LOCK, KW_EXC_LOCK, KW_NO_WAIT
** 257 ECF 20130212          Added semantic predicate to record_phrase rule to not match on any of
**                           the lock keywords (KW_NO_LOCK, KW_SH_LOCK, KW_EXC_LOCK, KW_NO_WAIT)
**                           as the optional literal immediately after the record name. This fixes
**                           a regression in #256 which caused all lock types to be dropped from
**                           converted record phrases.
** 258 CA  20130215          Added support for FUNCTION ... RETURN EXTENT clause.
** 259 GES 20130214          Fix to calling location for map_to_actual_name and
**                           added a malformed symbol annotation.
** 260 CS  20130214          Added support for RUN unsupported options.
**                           Added keyword support for remove-requests-after.
**                           Fixed conversion of IN handle clauses, as the handle is a rvalue.
** 261 VMN 20130218          Bug fixed related to the parser changes for making
**                           the lock types into literals:
**                           FIND FIRST whatever WHERE  NO-LOCK.
**                           (issue #1947, comment #53-#58)
** 262 ECF 20130224          Added limited support for embedded SQL.
** 263 OM  20130220          Fixed parsing for DB_REF_NON_STATIC to accept any symbol (including
**                           keywords and progress reserved words in the right side of ::).
** 264 SVL 20130226          Added an optional APPEND keyword after a function call TABLE
**                           parameter. It can also match a but wider, but hopefully it is
**                           not a problem. Also, treat handle declared by DEFINE PARAMETER
**                           TABLE-HANDLE or used as a TABLE-HANDLE parameter as a variable. Can
**                           match wider too.
** 265 CA  20130304          CAN-FIND needs to create its own scope, where to add the target 
**                           buffers, but this scope must not propagate.
** 266 GES 20130304          Force BUFFER-COPY and BUFFER-COMPARE to promote their table even
**                           it was already promoted further up the stack. This solved problems
**                           with matching the right buffer during parsing real app code. Fixed
**                           GET-ATTRIBUTE() to be METH_CHAR (found to be different from the
**                           documented METH_LOGICAL based on real code).
** 267 CA  20130305          The format for a PUT field can be any dynamic expression, which
**                           gets evaluated at runtime.
** 268 CA  20130307          Tables referenced in the record phrase have precedence for 
**                           abbreviated field lookup.
** 269 CA  20130307          Emit warnings if there is more than one AS clause for a var def.
** 270 CA  20130307          Fixed honoring of VALIDATE clause for temp-table fields defined using
**                           the LIKE clause and tables defined using the LIKE clause.
** 271 GES 20130308          Literal lexing was too lenient. Date literals were being matched
**                           when they should be dec literals (as a general case). The 4GL allows
**                           these same forms (that should normally be dec literals) to be
**                           matched as a date literal during variable definition initializers.
**                           The parser was changed to rewrite the type of the literal in that
**                           case. Also fixed an attribute that was mistakenly marked as a
**                           method.
** 272 GES 20130308          Empty string as a date var initializer must be treated as the
**                           unknown value literal. This is just one more undocumented 4GL
**                           quirk.
** 273 SVL 20130311          Emit FOR option for DEFINE PARAMETER TABLE-HANDLE.
** 274 CA  20130312          Fixed initializer_constant to handle char fields/vars too.
** 275 CA  20130317          Removed the SCHEMA_VALIDATIONS node, as the schema VALEXP need to be
**                           kept at the field widget always.
** 276 CA  20130319          Fixed var processing for DB_REF_NON_STATIC.
** 277 GES 20130317          Fixed USING statement which was not properly adding the package.
** 278 OM  20130325          Added socket read const literals: READ_AVAILABLE_LITERAL and 
**                           READ_EXACT_NUM_LITERAL.
** 279 OM  20130430          Added KW_TIMEZONE as SESSION attribute.
** 280 OM  20130507          Added DATETIME_TZ_LITERAL lexer rules.
** 281 GES 20130517          Added byte ordering literals (4GL compiler constants).
** 282 OM  20130516          Added DATETIME_LITERAL lexer rules.
** 283 VIG 20130918          Added OUTPUT TO PRINTER "printerName" support (printerName as
**                           string).
** 284 CA  20131208          Emit SEARCH-TARGET and SEARCH-SELF literals as integer.
** 285 VIG 20131130          Added IN WINDOW clause support to DISPLAY statement
** 286 GES 20130919          Fixed max_rows_clause to accept an expression instead of requiring a
**                           numeric literal (found in customer source).  Added an implicit like
**                           clause to assign_trigger_var_def (this is an undocumented feature
**                           found in customer code). Added NO-ERROR support to create temp-table
**                           (found in customer code).
** 287 OM  20131128          Improved support for database triggers (both schema and session, 
**                           with emphasis on ASSIGN and WRITE). Added replication trigger
**                           support.
** 288 GES 20140116          Improved/fixed native API support in type specifications with
**                           proper HANDLE TO support (the previous implementation was completely
**                           wrong), the addition of VAR_ULONG and the use of a string literal
**                           for a procedure name.  Added NO-ERROR support to release external.
**                           Fixed malformed_symbol to merge until next whitespace (with some
**                           exceptions). Fixed the string form of procedure external to rewrite
**                           the string as a symbol.
** 289 GES 20140127          Fix for when a malformed symbol is followed by EOF instead of
**                           whitespace.
** 290 GES 20140130          Excluded finally from being a valid label name. Added NO-UNDERLINE
**                           as a new undocumented define browse option.
** 291 EVL 20140215          Fix for SSTRING and DSTRING sections to allow to handle Windows
**                           style new line separator '\r' while lexing DF schema file.
** 292 CA  20140313          Add line/column numbers for the process IO placeholder nodes.
**                           Fixed parsing of CREATE automation object com-handle statements.
** 293 EVL 20140314          When processing RUN FILENAME statements we need to take into account
**                           how the "\" character is considered in different OS.  In Windows it
**                           is a file separator for Progress, not escape character.
** 294 GES 20140319          Fixed set_or_update_or_prompt_for_stmt to allow format phrase
**                           variable definitions. Added XML options to define temp-table.
**                           Fixed syntax and added options for def_dataset_stmt. Allowed
**                           ascending() as a synonym for asc(). Allow colon to end a function
**                           definition even in forward cases. Added support for undocumented,
**                           old-school options in next-prompt. Added options to record phrase.
**                           Switched CREATE SERVER to use an attribute assignment instead of
**                           a simple assignment (the assignment is implicitly referencing
**                           attributes). Added buffer attributes/methods related to serializing
**                           to JSON/XML. Rewrote the run_stmt rule to allow an undocumented
**                           merging of text after a string. The use of malformed_symbol was
**                           eliminated because filename is now a superset of the functionality.
**                           Added the error-stack-trace attribute. Added undocumented no-error
**                           support for the I/O THROUGH stmts. Added undocumented 
**                           SCROLLBAR-VERTICAL to frame phrase. Allow database events to be
**                           used for widget triggers. Fixed OO property getter_setter support.
**                           Fixed dataset-handle parameter definitions in methods. Refactored
**                           attribute/method chaining to allow this off primary_expr as well as
**                           in lvalue. Added ROW-STATE compiler constants as literals. Complete
**                           rework of the chaining implementation for OO 4GL, handle attrs/meths
**                           and com-handle support. The previous implementation was flawed
**                           because it used recursion. This caused the trees to be reversed in
**                           structure from what they need to be to emit properly, leading to
**                           complex and unnecessary processing downstream (to fix this up).
**                           Instead, all chaining has been left-factored into a single location
**                           that uses a loop instead of recursion, leaving the resulting trees
**                           un-reversed (and in good shape for downstream processing). Also
**                           included in this set of changes is the ability to "mix and match"
**                           invocation types within the same chain.  All of this is valid in the
**                           4GL and was encountered in customer source code. The result is much
**                           simpler and much cruft was removed. Part of the changes include the
**                           replacement of lvalue as the first match in an assignment. Instead,
**                           the chained_object_members should be used in any location where
**                           chaining could be used on the left side of an assignment operator.
**                           Added the AS BUTTON quirk to the format phrase for frame definition
**                           statements (undocumented feature of the 4GL). Disambiguated the NEW
**                           record function from the NEW OO function in the case where the
**                           class name is also a table name. Added FIRST-OF/LAST-OF as methods.
**                           Reworked the parameter rule to split it into parameter passing and
**                           parameter definition rules. Although there is much overlap, in the
**                           end it was found that the two usages have enough differences to
**                           warrant separate rules. The result is much simpler, especially for
**                           the parameter passing case. Fixed before-table clause processing.
**                           Fixed sql aggregate function support to allow references to field
**                           names that cannot be resolved until after the from clause. Fixed
**                           function parameters to allow some keyword suffixes for table parms.
**                           Made editor_phrase options non-mandatory (undocumented feature).
**                           Added stream-handle support. Added more operators to the exclusion
**                           list for malformed_symbol. Changed the compile_stmt languages list
**                           to match expressions instead of symbols. Added the FIND-* compiler
**                           constants for the REPLACE() method. Fixed error() builtin func.
**                           Improved keyword ignore support at the parser level (not just in
**                           the lexer).
** 295 CA  20140701          The indices in a COLOR or FONT phrase can be expressions, not just 
**                           static values.
** 296 OM  20140711          Fixed ASSIGN database trigger conversion.
** 297 VIG 20141020          Added IN WINDOW clause support to ENABLE statement.
** 298 SVL 20141102          NO-COLUMN-SCROLLING is added as a logical attribute.
** 299 GES 20140820          Rationalization of color processing, creation of artificial 
**                           AS_BUTTON node. Added some token types. Added DB-CONTEXT attribute
**                           and DSLOG-MANAGER system handle.
** 300 ECF 20141121          Drop optional BUFFER and FIELD keywords from RAW TRANSFER AST.
** 301 EVL 20141208          Fix for DEFINE IMAGE statement.  The statement must return variable
**                           in order to reference the widget by name on conversion stage.
** 302 GES 20141125          Initial changes to handle INPUT-VALUE attr better. Improvements in
**                           INPUT builtin func differentiation from INPUT mode in function call
**                           parameter parsing. Support for user-defined function implicit
**                           parameter quirk.  An implicit prefer widgets mode must be enabled
**                           when lvalue is parsing an attribute/method referent.  Added bufpool
**                           and buffer-pool schema keywords. Added write-json() method. Extended
**                           the menu-item embedded DOT quirk to more cases. Added top-only,
**                           three-d and color_spec undocumented browse options. COLON and TO
**                           in format phrases can now be decimal literals (not just integers).
** 303 GES 20141231          Changes to support an undocumented 4GL quirk where "dotted" trash is
**                           silently ignored. Added the undocumented PROFILER system handle and
**                           some attributes found in customer source code.
** 304 VIG 20141104          Changed DEFINE MENU parsing, submenu_descriptor rule.
** 305 ECF 20150210          In format_phrase, promote buffer associated with widget instead of
**                           default buffer.
** 306 GES 20150218          Extended the INPUT builtin function disabiguation to RUN parameters.
** 307 CA  20150729          Added FRAME_LOCK and FRAME_UNLOCK tokens, needed for locking and 
**                           unlocking the frame in a CREATE <widget> ASSIGN statement.
** 308 OM  20150805          Checking for additional space after some keyword operators.
** 309 VIG 20150629          Fixed order of SUB-MENU options.
** 310 GES 20150814          Added undocumented support for an empty trigger phrase. Added attrs
**                           and methods for the undocumented PROFILER system handle, based on
**                           customer code.
** 311 OM  20150830          Added annotation for NEW, GLOBAL and SHARED attributes to definition
**                           of temp-tables records.
** 312 OM  20150924          ASC function declared as int64. Enabled INPUT-VALUE attribute.
** 313 ECF 20151030          Moved schema-level validation parsing from SchemaLoader to Progress
**                           parser. This parsing must happen within the context of the business
**                           logic in which the validation occurs. Parsed validation expressions
**                           and messages are attached to the statements which enable the
**                           corresponding widgets (enable, set, update, insert, prompt-for).
** 314 EVL 20160224          Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 315 GES 20160206          Bypass parameter processing in method_stmt when executing as a first
**                           pass. This avoids failures caused by trying to parse table parms
**                           (since temp-tables are not parsed yet). Also, avoid adding a widget
**                           for variables that are objects, otherwise the object invocation
**                           parsing can fail (because it thinks it is seeing a widget instead
**                           of an object instance).
** 316 EVK 20141212          Properly handle terminal streams when specified as a string literal.
** 317 GES 20160410          Added a P2J-specific extension to the the language for URL loading
**                           in the user interface (P2J-OPEN-URL).
**     CA  20160421          Added P2J-REMOTE-CALL function (language extension, not real 4GL).
** 318 GES 20160309          Differentiate function statements from the use of function as a
**                           user-defined symbol. Added read-json() method. Added ATTR_RAW and
**                           METH_RAW to attribute_method rule. Added serialize-name and 
**                           serialize-hidden as tt field options.  Made import fields optional
**                           (undocumented feature).  Stream names can be malformed. Import
**                           unformatted can reference a record (and it works too!), this is an
**                           undocumented feature of the 4GL. Added code to properly engage the
**                           prediction logic when encountering a malformed label that starts
**                           with a reserved keyword. Allow an optional format string to follow
**                           the aggregate phrase in an accumulate stmt (undocumented feature).
**                           Reduced overuse of the qualified name quirk code by checking the
**                           simple field match first. Reworked query, buffer, temp-table,
**                           dataset and data-source definitions to add these as resources in
**                           OO classes.  Lookups changed to implement class hierarchy resolution
**                           (of protected members).  Added UNDO keyword to define temp-table
**                           (this is an undocumented feature in the 4GL). Added value() as
**                           an undocumented option for KW_MAP in io_options. Made the format
**                           phrase in an underline statement something that is a full option
**                           we previously only allowed it for WHEN clauses but customer code
**                           has shown it is a more generic thing (there was a format string).
**                           transa and transac are not valid abbreviations of transaction (but
**                           trans is). Allow input function to be matched as the initial token
**                           of a standalone expression. Set text values to empty string for
**                           manufactured validation nodes (avoids NPE downstream when trying to
**                           use text from those nodes). Improved static OO resource handling.
**                           Treat members at the class and interface def level like an ext
**                           proc (the resources go into the schema dict global scope). Fixed
**                           dynamic-new parsing (it was profoundly broken and could never work).
**                           Variable names take precedence over class names. Index field
**                           references should not use lvalue (in the 4GL there are very strict
**                           syntax rules that allow no abbreviations, no qualifiers and no
**                           subscripted or unsubscripted extent field references. This avoid
**                           unnecessary ambiguity since the field reference can only be linked
**                           to a field in the current table. Added block-level statement.
** 319 ECF 20160428          Fixed attachment of table-level validation message.
** 320 CA  20160721          Empty string initializer for numeric types is emitted as 0.
** 321 ECF 20170130          Normalize token type of image_phrase parent to KW_FILE when
**                           KW_FIL_NAME is used (undocumented feature), to simplify downstream
**                           use.
** 322 EVL 20170215          Fix for _msg(kw_u_msg) to be considered as built-in function.  Also
**                           the value for keyword hidden-field-list is changed to KW_HID_FLDL.
**     SVL 20170227          Added some tokens.
**     GES 20170303          Added the undocumented update_logical_clause to the system-dialog-
**                           get-dir stmt.  Added the (trash) APPEND keyword to the input from
**                           stmt.  This is allowed (undocumented) but it seems to mean nothing.
**                           Added the PARENT-ID-RELATION clause to define dataset. Added an
**                           undocumented abbreviation for the row-markers keyword. Added a
**                           quirk of single statement body triggers that can have an end stmt
**                           without a corresponding block open.
**     ECF 20170303          Honor FOR TEMP-TABLE clause in DEFINE BUFFER/PARAMETER statements.
** 323 GES 20170314          Implemented better token types for embedded sql open_cursor,
**                           close_cursor and update_sql.
** 324 OM  20170509          Added KW_FIND_SEL to literal list. Few typos fixed.
** 325 GES 20170428          Changes related to call graph analysis v3.
**                           Fixed NO-ERROR match for CREATE <widget> statement (match before
**     CA  20170705          Fixed NO-ERROR match for CREATE <widget> statement (match before
**                           ASSIGN).
**     GES 20170705          Added IN WINDOW undocumented clause to SYSTEM-DIALOG-GET-DIR. Added
**                           the (undocumented) parenthesis-optional form of GUID. Added the
**                           undocumented CARET to the IMPORT UNFORMATTED statement. Added the
**                           undocumented WIDTH-CHARS option to column_format_phrase. Removed
**                           the direct use of record_funcs from the assign_stmt which is not
**                           needed because the 2nd alternative will already match record_funcs.
**                           It was causing problems when the input function was used for
**                           attribute access (meaningless, but it works).
**     GES 20170706          Refined the CARET in the IMPORT UNFORMATTED statement to make it a
**                           more generic result, after testing different forms that are
**                           possible in the 4GL. Added some undocumented names to key_function.
**                           Added the MESSAGE-DIGEST() built-in function.
**     GES 20170707          Added the FIRST-OBJECT attribute. Modified key_function to support
**                           an undocumented (but fully functional) STRING form. Changed
**                           incorrect usage of lvalue to expr in lob_spec. Added KW_BREAK to
**                           the open_query_stmt. Unified the list_items_clause and the
**                           list_item_pairs_clause because the 4GL compiler actually silently
**                           matches the same syntax for both. The pairs version will generate
**                           a runtime warning 4041 when used but for definition purposes there
**                           is no checking.
**     GES 20170708          The triggers phrase (used in widget and variable definitions or
**                           create statements) has an undocumented feature where it allows the
**                           embedded END statement (or END TRIGGERS) to be closed with etiher
**                           a DOT or a COLON.  Other uses of the END statement seem to only
**                           allow DOT but we have not tested all possible cases. The code was
**                           modified to allow this matching in the triggers phrase case only.
**                           This is tricky because the matching closes the containing language
**                           statement since this is just a phrase. This is referered to as the
**                           allowColon undocumented feature.
**     GES 20170710          Added undocumented abbreviation for ICFPARAMETER. Added
**                           XML-NODE-NAME as an attribute.
**     CA  20170710          column_spec (used by DEFINE BROWSE ... DISPLAY ... statement) must
**                           check for isQualifiedFieldNameQuirk before deciding if LT is a 
**                           record or not (as between the table, dot and field might be
**                           whitespace or comments).
**                           embedded_sql: a "CLOSE cursor" must be disambiguated from a 
**                           "CLOSE()" function call.
**     CA  20170711          Added parse for "STOP-AFTER expr" option at DO, FOR and REPEAT.
**                           Added undocumented LABEL option at DEFINE TEMP-TABLE/WORK-TABLE.
**                           Fixed problems in pre_scan_class: non-method members were not 
**                           included and also it was not differentiating properly the recursive
**                           calls.
**     GES 20170719          Added ADD-PARENT-ID-RELATION and TOP-NAV-QUERY.
**     HC  20170725          "COMPILE" statement must be disambiguated from a "COMPILE()" function
**                           call.
**     HC  20170726          Added number of errors counter.
**     ECF 20170731          Fixed connect_stmt to make physical database name optional.
**     GES 20170730          BUFFER-COMPARE is reserved and can be abbreviated. BUFFER-COPY is
**                           reserved.  Both of these changes were undocumented.
**     CA  20170804          IN SUPER or IN handle functions can be defined inside a class.
**     HC  20170804          Added ROW-STATE built-in function handling to record_funcs to fix
**                           parsing of ROW-STATE function call.
**     GES 20170806          Added undocumented string option X which can be present in any
**                           case where U would be present (but not together with U).
** 326 SVL 20170809          Implemented validation for browse columns.
** 327 GES 20170817          Added the TERM reserved keyword and removed the abbreviation for
**                           TERMINAL (the 4GL docs are incorrect on this point).
** 328 OM  20170802          Added support for TIMER FWD extension. Moved ENABLED attribute to
**                           PSTimer FWD extension. Removed possibleActiveX support.
**     OM  20170724          Rewrote position_clause_arg. All coordinates are mandatory.
** 329 CA  20170825          Initial support for USE-DICT-EXPS: not fully complete for the HELP
**                           behaviour.
** 330 HC  20170629          Implemented support of WINDOW:DISABLE-REDRAW, the FWD extension to
**                           4GL.
**     HC  20170819          Added extension widget method GET-MOUSE-POSITION.
** 331 GES 20170825          Improved DISPLAY stmt disambiguation of table/field name conflicts
**                           in the case where a there is a following form item (the table name
**                           no longer "wins").
**     CA  20170828          end_stmt_with_no_beginning must create a TRIGGER_BLOCK, not 
**                           INNER_BLOCK.
**     OM  20170914          When processing VALEXP for a field, its unqualified name takes
**                           precedence over fields of a possible buffer from a CAN-FIND function.
**     CA  20170915          Fixed a regression in H329: DEFINE FRAME statement must process its
**                           setUseDictExps call in define_stmt instead of def_frame_stmt, as in
**                           def_frame stmt we don't have access to the full statement AST.
** 332 CA  20170929          Added MNEMONIC and PREPROCESSED-LABEL FWD extension attributes.
** 333 CA  20171006          Finished USE-DICT-EXPS support, for both HELP and VALEXP.
** 334 CA  20171019          Fixed a problem in H333 - HELP node created for schema help must not
**                           use progressToJavaString.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

header
{
/*
** Module   : ProgressParser.java
**            ProgressParserTokenTypes.java
**            ProgressLexer.java
** Abstract : lex and parse Progress 4GL source code
**
** Copyright (c) 2004-2016, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 GES 20041116 ADD   @18747 WARNING, THIS IS A GENERATED FILE!!!
**                               DO NOT EDIT THIS FILE. The original source
**                               file is progress.g!
*/

package com.goldencode.p2j.uast;

import java.io.*;
import java.util.*;
import java.util.function.*;
import com.goldencode.ast.*;
import com.goldencode.util.*;
import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.preproc.*;
import com.goldencode.p2j.util.*;
import antlr.*;
import antlr.debug.misc.ASTFrame;
}

/**
 * Creates an Abstract Syntax Tree (AST) representation of a Progress 4GL
 * source file from an input stream of tokens (provided by the
 * <code>{@link ProgressLexer}</code>).  More specifically, the parser does
 * the following:
 * <ul>
 *    <li> Drives the lexer by calling its' <code>nextToken</code> method.
 *         This method tokenizes the lexer's input stream of characters and
 *         returns the next found <code>Token</code> object.
 *    <li> Provides a set of methods designed as entry points.  The entry 
 *         points control the type of parsing that will occur. At this time
 *         there are 2 valid entry points:
 *       <ul>
 *          <li> <code>{@link #external_proc}</code> - this is designed to
 *               parse an entire <b>(already preprocessed)</b> Progress 4GL
 *               source file
 *          <li> <code>{@link #expr}</code> - designed to parse any valid 
 *               Progress 4GL expression
 *       </ul>
 *    <li> Recognizes the following Progress 4GL language features:
 *       <ul>
 *          <li> full expression processing for all data types
 *          <li> support for the data types character, logical, date, integer,
 *               decimal, handle, widget-handle (this type is actually
 *               rewritten by the lexer as a handle), raw, recid, rowid, 
 *               memptr and com-handle
 *          <li> structural features such as blocks, external procedures,
 *               internal procedures and functions
 *          <li> variables, labels, procedure names, widgets, function names
 *               streams, menus, sub-menus, menu-items and language keywords
 *               (including handling abbreviations and reserved keywords)
 *          <li> schema object support (database, table and field names)
 *          <li> assignment statements including lvalue processing
 *          <li> language statements
 *          <li> attribute and method support
 *          <li> built-in functions (including those that act like variables
 *               or which have unusual syntax like postfixed function names
 *               or no use of parenthesized parameter lists)
 *          <li> COM (Active-X) property and method support
 *          <li> ProDataSets (datasets, data-sources, data-relations)
 *          <li> object-oriented support (classes, interfaces, methods,
 *               properties, packages, data members)
 *          <li> directly embedded SQL
 *          <li> "fake" interfaces to duplicate the features that WebSpeed
 *               supports (PSC encodes the documented API for WebSpeed as
 *               4GL source code which is included in WebSpeed applications);
 *               by preloading the definitions for these features directly
 *               in the parser, we will treat these as real built-ins and
 *               eliminate the need for the PSC 4GL source code to be present
 *               in order to parse and convert WebSpeed applications; this
 *               is just an extra precaution since the WebSpeed code was
 *               open sourced and it is probably safe to convert but we will
 *               try to avoid it; 
 *       </ul>
 *    <li> Builds an AST (&quot;intermediate form representation&quot;) of
 *         all the recognized tokens.  This AST is designed for subsequent
 *         processing.  For example, an expression AST is evaluated by the
 *         <code>{@link ExpressionEvaluator}</code> and the related helper
 *         class <code>{@link ExpressionEvaluator}</code>.  The
 *         resulting AST matches the block structured nature of the Progress
 *         4GL, including the provision of nesting and the refusal to nest
 *         in certain cases (e.g. Progress 4GL does not allow nesting of
 *         procedure or function definitions, but it does allow for the
 *         nesting of blocks such as <code>do</code> or <code>repeat</code>).
 *    <li> Handles the context sensitive aspects of symbol resolution.
 *    <li> To the extent that is possible while still generating a well
 *         structured tree, the parser has been structured to enforce
 *         syntactic correctness.  This is typically well implemented in
 *         specific language statements (their structure enforces 100%
 *         syntactic correctness in more cases or it enforces nearly 100%
 *         correctness).  Due to ambiguities in the Progress 4GL language and
 *         constraints in the ANTLR parser generation, it is not always
 *         possible to make the parser's structure (i.e. the manner in which
 *         its rules nest and how the matching of each rule processes)
 *         enforce syntax checks. For some language features, these checks
 *         must be implemented using ANTLR's semantic or syntactic predicates
 *         and in other cases ANTLR's actions must be used for a completely
 *         custom set of checks.
 * </ul>
 * <p>
 * This grammar is highly specific to ANTLR 2.7.4 and it may not be valid
 * in future versions of ANTLR.  This is due to the fact that the generated
 * code and the grammar syntax of ANTLR can change between versions.  These
 * factors are the most important constraints defining the structure of the
 * resulting grammar.  As a result, it is possible that the grammar will
 * require changes to run in future versions of ANTLR.
 * <p> 
 * <b>Please note that this is a generated file using ANTLR 2.7.4 and a
 * grammar specified in <code>progress.g</code>.</b>
 * <p>
 * The generated parser is an LL(k) type parser.  This means that it is
 * top-down parser than recursively descends from top level rules through 
 * more and more specific rules until matches are found, in which case the
 * token is consumed from the input stream and the rules return up one level
 * to their calling rule and so on until processing completes a full top
 * level match and returns all the way up to the top level entry point.
 * <p>
 * The top level entry points are designed to either match a single construct
 * (e.g. <code>expr</code> which parses and matches a single expression) or 
 * to iteratively match a set of constructs that might comprise an entire
 * source file (e.g. <code>external_proc</code>).
 * <p>
 * Starting at the top level, the parser requests the next tokens from the
 * lexer in order to have the proper amount of lookahead.  The lookahead
 * depth defined as the &quot;k&quot; of the parser, and in this case the k 
 * = 2.  These lookahead tokens are compared to sets for each alternative
 * that is defined at the top level.  Where there is ambiguity, the first
 * match will trigger a call to the rule specified in the alternative (and/or
 * it will match any hard coded tokens in the alternative). Once that 
 * alternative is entered, the parser expects that it will begin matching
 * specified constructs at the top level or in one of the rules that will be
 * recursively called.  This matching consumes tokens from the lexer's stream
 * so a call to these rules is a one way process which must end in a complete 
 * match to the set of rules that can be reached from the top level
 * alternative that was predicted by the lookahead.  This means that the 
 * prediction logic is generated at the top level (outside of the rules)
 * but based on the contents of the rules.  So the consolidated lookahead of
 * all possible paths through the rules is rolled up into the prediction
 * logic at the top level while the matching doesn't occur until the
 * specific rules are recursively entered.  This roll up can create sets
 * of predictions that cannot actually occur in a given grammar, but which
 * appear to be possible based on the rolled up prediction logic.  This
 * can be the source of ambiguity and hence ANTLR warnings.  It is important
 * to keep the top level alternatives very distinct and to organize them
 * in a decreasing specificity order so that the order will naturally
 * resolve ambiguity using lookahead.
 * <p>
 * For details on how the context sensitive symbol resolution occurs, please
 * see {@link #primary_expr}, {@link #func_call}, {@link #lvalue},
 * {@link #record}, {@link #label_reference} and {@link #symbol}.  In
 * addition, it is important to review the lexer's role in symbol resolution 
 * (see {@link ProgressLexer#mSYMBOL}).
 * <p>
 * Implementation WARNING: <b>ANTLR syntactic predicates cannot be used</b> 
 * because this changes how the all code is generated such that all actions
 * are made conditional on the state of the guessing mode. The problem with
 * this is that the parser implementation is heavily designed based on an
 * assumption that all actions (especially init actions) are not conditional.
 * For example, many local variables are defined in init actions. As soon as
 * a single syntactic predicate is added to the parser, all init actions are
 * wrapped in an if () { } block which makes any defined local variables
 * immediately go out of scope and thus they are not available to the rest
 * of the method.  This breaks the parser.
 * <p>
 * Notes on understanding keyword usage: All keywords have a specific token
 * type. That type is an artificially created token which exists in one of
 * two ranges: <code>BEGIN_RESERVED</code> through <code>END_RESERVED</code>
 * (for the reserved keywords) and <code>BEGIN_UNRESERVED</code> through
 * <code>END_UNRESERVED</code> (for unreserved keywords). Then the lexer is
 * modified to have a {@link Keyword} instance definition in it's list of
 * keywords. These keywords are used for lookups and that processing handles
 * abbreviations, case-insensitivity and other features of Progress keyword
 * processing. The lexer converts symbols into keyword tokens when this
 * keyword lookup is successful. Then the parser matches on the keyword
 * token type as needed to match the 4GL language syntax. There are keywords
 * in the list which are not represented in parser rules.  This is a normal
 * case for the following: events, constants, built-in function names and
 * those built-in functions that act like global variables. All of those
 * cases provide the keyword so that downstream tree processing can easily
 * compare events, constants or the <code>oldtype</code> annotation on
 * built-in functions/variables.  Do NOT be disturbed by the lack of any
 * usage of these tokens in the parser itself.
 * <p>
 * Open TODOs/Issues:
 * <p>
 * <ul>
 *    <li> stored procedure support is present but not complete
 *    <ul>
 *       <li> procedures are schema objects that act like tables in some
 *            ways, but have additional behavior as well, this schema support
 *            does not exist
 *       <li> the stored procedure rules must be updated to dynamically
 *            create the special stored procedure names when needed, likewise
 *            there are buffer and field definitions that must be dynamically
 *            created during parsing (this will use the functionality noted
 *            above for the schema)
 *       <li> see the rules {@link #run_stored_proc_stmt} and
 *            {@link #close_stored_proc_stmt} for more details
 *    </ul>
 *    <li> schema name lookups are documented to give preference to tables
 *         over temp-tables in the case where both share the same name (this
 *         has not necessarily been explicitly tested and it isn't obvious if
 *         we have intentionally or unintentionally implemented this at all);
 *         but in addition, the DEFINE BUFFER statement now allows one to
 *         specify the TEMP-TABLE keyword to force temp-tables to be preferred
 *         over a table of the same name (we now have a forceTemp flag but
 *         we need to honor it in the {@link #record} rule)
 *    <li> any use of the <code>NAMESPACE-URI</code> and/or 
 *         <code>NAMESPACE-PREFIX</code> options (see 
 *         {@link #namespace_clause}) is not honored at this time
 *    <li> 4GL aggregate support seems to be implemented by using hidden
 *         variables that are created on the fly and which can be referenced
 *         and assigned directly from user code (without any explicit variable
 *         definition code); for example the integer variable COUNT can be
 *         directly referenced by user code when the COUNT aggregate is in
 *         use; these variables, their types and the exact moment of creation
 *         need to be investigated and added (it is all undocumented of
 *         course)
 *    <li> {@link #form_item} has an incomplete implementation to support
 *         inline AS or LIKE clauses; this will not match all possible cases
 *         where AS or LIKE appears later in the format phrase (since they
 *         can be arbitrarily ordered); see <code>isFormVarDefinition()</code>
 *         which is some example code for a partial solution but it does not
 *         work completely enough to be used;
 *    <li> {@link #constant_form_item}, {@link #column_ref} and
 *         {@link #browse_options_phrase} matches a subset of the format
 *         phrase but these options are not rooted as a FORMAT_PHRASE node;
 *         they should be rooted at such a node and when that happens, there
 *         will be "downstream" reporting code and conversion code that needs
 *         to be reworked (but that code will be much simpler); as a related
 *         question: should the options for the various variable definition
 *         mechanisms be rooted at a FORMAT_PHRASE?
 *    <li> {@link #def_var_stmt} may need to exclude RAW and MEMPTR fields
 *         from adding a WID_FILL_IN definition (like it does today for
 *         handle types)
 *    <li> should INPUT-VALUE be a synonym for SCREEN-VALUE or are these
 *         really different values?
 *    <li> should we support VAR_LONGCHAR and FIELD_CLOB in
 *         {@link #choose_keys_clause}?
 *    <li> confirm that properties and class data members (variables) share
 *         the same namespace as is suggested by the documentation
 *    <li> buffers, queries, temp-tables, prodatasets and data-sources can
 *         be protected resources of a parent class that are accessed from
 *         a child class; this is not implemented at this time and some
 *         example code is needed to fully demonstrate the requirements
 *         before implementation (this will affect the schema dictionary and
 *         the symbol resolver's resource lookup processing and loadViaWalk()
 *         method)
 *    <li> since the parser now provides "fake" interfaces to duplicate
 *         the features that WebSpeed supports, that support may not have
 *         been completely documented by PSC and as such it may not be
 *         complete
 *    <ul>
 *       <li> there are references in the WebSpeed documentation to variable
 *            names like "SelfURL", "AppURL" and CGI environment variables
 *            but there are no details on the types or usage of these, so at
 *            this time these are not yet added but may subsequently be found
 *            in real source code
 *       <li> it is likely that the features of HTML mapping and session
 *            management may provide procedures and functions that are not
 *            documented but which are used in real source code 
 *    </ul>
 * </ul> 
 */
class ProgressParser extends Parser;

options
{   
   k = 3;
   buildAST = true;
   ASTLabelType = "Aast";
}

// WARNING!!! There are several contiguous blocks of related token types
//            that MUST be kept contiguous in this "tokens" section. This
//            enables rules to compare and lookahead based on a start and
//            end token instead of individually listing each token. This
//            greatly simplifies this grammar. In addition, at least two of
//            the blocks must actually stay adjacent (VARTYPES and
//            FIELDTYPES) since they are actually part of a larger contiguous
//            group called LVALUE.  Any token starting BEGIN_ or END_ is
//            a marker for one of these contiguous sections.  Do not add
//            unrelated tokens into these blocks.  Also note that the
//            reserved and unreserved keywords have 2 separate blocks.  It is
//            critical that any new keywords are categorized properly.
// Format: All token types for keywords are prefixed by KW_ and up to 8
//         descriptive characters for the keyword itself. If possible it is
//         the same as the keyword.  Otherwise, make an abbreviation as
//         needed. Hyphens are converted to underscores.
tokens
{
   DOT;
   FILEROOT;
   DB_SYMBOL;
   ANY_LOCK;
   FILENAME;
   BACKSLASH;
   LIBRARY_REF;
   BEGIN_RECORDTYPES;
      TABLE;
      BUFFER;
      TEMP_TABLE;
      WORK_TABLE;
   END_RECORDTYPES;
   BEGIN_BUFFERTYPES;
      NO_REFERENCE;
      WEAK_REFERENCE;
      STRONG_REFERENCE;
      FREE_REFERENCE;
   END_BUFFERTYPES;
   EXPANDED_SCOPE;
   BUFFER_SCOPE;
   BEGIN_SCHEMA;
      ABBREVIATED;
      AREA;
      CYCLE_ON_LIMIT;
      DESCRIPTION;
      DUMP_NAME;
      FROZEN;
      HIDDEN;
      INCREMENT;
      MANDATORY;
      ORDER;
      POSITION;
      SQL_WIDTH;
      TRAILER;
      VALEXP;
      VALMSG;
      BEGIN_SCHEMAKW;
         KW_ABBV;
         KW_AREA;
         KW_BIGINT;
         KW_BUFRPOOL;
         KW_BUFPOOL;
         KW_CLOB_CP;
         KW_CLOB_COL;
         KW_CLOB_TYP;
         KW_CRC;
         KW_CYCLE;
         KW_DMP_NAME;
         KW_FLD_TRG;
         KW_FIXCHAR;
         KW_FROZEN;
         KW_IDX_FLD;
         KW_INACTIVE;
         KW_INCR;
         KW_LOB_AREA;
         KW_LOB_BYTE;
         KW_LOB_SIZE;
         KW_NO_OVRRD;
         KW_NOT_CS;
         KW_NULLALWD;
         KW_PSC;
         KW_SEQUENCE;
         KW_SQL_WID;
         KW_TAB_TRG;
         KW_TIMESTMP;
         KW_VALEXP;
         KW_VALMSG;
      END_SCHEMAKW;
      BEGIN_METATYPES;
         FIELD_BIGINT;
         FIELD_BYTE;
         FIELD_DOUBLE;
         FIELD_FIXCHAR;
         FIELD_FLOAT;
         FIELD_SHORT;
         FIELD_TIMESTAMP;
         FIELD_TIME;
      END_METATYPES;
   END_SCHEMA;
   DATABASE;
   INDEX;
   INDEX_FIELD;
   BEGIN_INDEX_MATCH_TYPES;
      NO_MATCH;
      SELF_REFERENCE;
      EQUALITY_MATCH;
      BEGINS_MATCH;
      RANGE_MATCH;
      WORD_MATCH;
      SORT_MATCH;
   END_INDEX_MATCH_TYPES;
   SEQUENCE;
   PROPERTIES;
   MANY_TO_ONE;
   ONE_TO_MANY;
   ONE_TO_ONE;
   KEY_FIELD;
   RAW_STRING;
   QUERY;
   QUERY_SUBST;
   CLIENT_WHERE;
   CLIENT_WHERE_SUBST;
   EMBEDDED_SQL;
   COLUMN_LIST;
   BLOCK;
   PROCEDURE;
   CLASS_DEF;
   INTERFACE_DEF;
   METHOD_DEF;
   CONSTRUCTOR;
   DESTRUCTOR;
   CLASS_NAME;
   OBJECT_INVOCATION;
   METHOD_INVOCATION;
   PROPERTY_INVOCATION;
   EVENT_SIGNATURE;
   DOTNET_DELEGATE;
   INT_PROC;
   WEB_PROC;
   FUNCTION;
   INNER_BLOCK;
   TRIGGER_BLOCK;
   EDITING_BLOCK;
   FRAME_SCOPE;
   FRAME_ALLOC;
   COPY_TO_SB;
   COPY_FROM_SB;
   CUSTOMER_SPECIFIC;
   VALIDATION;
   TRANSACTION;
   SUB_TRANSACTION;
   NO_TRANSACTION;
   STATEMENT;
   LABEL_DEF;
   LABEL;
   EXPRESSION;
   EVENT;
   DB_EVENT;
   CLASS_EVENT;
   EVENT_LIST;
   KEY_FUNCTION;
   WIDGET_LIST;
   WIDGET_POOL;
   ASSIGNMENT;
   ASSIGN;
   ARGUMENTS;
   PARAMETER;
   HANDLE_TO;
   COMMAND_TOKENS;
   COMMAND_TEXT;
   CONNECT_OPTIONS;
   CONNECT_TEXT;
   IO_OPTIONS;
   PUT_FIELD;
   IMPORT_FIELD;
   EXPORT_FIELD;
   AGGREGATE;
   FRAME_PHRASE;
   FORMAT_PHRASE;
   TRIGGER_PHRASE;
   RECORD_PHRASE;
   COLOR_PHRASE;
   INLINE_VAR_DEF_ARRAY;
   INLINE_VAR_DEF;
   CONTENT_ARRAY;
   FRAME_ELEMENT;
   WHEN_LIST;
   EMBEDDED_ASSIGNMENT;
   FORM_ITEM;
   COLUMN_REF;
   RUNTIME_FRAME_OPTIONS;
   RUNTIME_BROWSE_OPTIONS;
   PARENT_CHILD_RELATION;
   DATA_SET;
   DATA_SOURCE;
   DATA_RELATION;
   PARENT_ID_RELATION;
   LANGUAGES;
   DYNAMIC;
   MINUS;
   PLUS;
   UN_MINUS;
   UN_PLUS;
   BOOL_TRUE;
   BOOL_FALSE;
   DEC_LITERAL;
   DATE_LITERAL;
   DATETIME_LITERAL;
   DATETIME_TZ_LITERAL;
   NO_LOCK_LITERAL;
   SHARE_LOCK_LITERAL;
   EXCLUSIVE_LOCK_LITERAL;
   NO_WAIT_LITERAL;
   READ_AVAILABLE_LITERAL;
   READ_EXACT_NUM_LITERAL;
   SEARCH_SELF_LITERAL;
   SEARCH_TARGET_LITERAL;
   OPTION;
   COM_METHOD;
   COM_PROPERTY;
   COM_PARAMETER;
   COM_INVOCATION;
   FILTER_SPEC;
   SUPPORT_LVL_NONE;
   SUPPORT_LVL_PARTIAL;
   SUPPORT_LVL_FULL_WITH_LIMITATIONS;
   SUPPORT_LVL_FULL;
   BOGUS;
   FRAME_LOCK;
   FRAME_UNLOCK;
   BEGIN_MULTIWORD;
      ALTER_TABLE;
      AS_BUTTON;
      CLOSE_QUERY;
      CLOSE_CURSOR;
      CLOSE_STORED_PROCEDURE;
      CREATE_ALIAS;
      CREATE_BROWSE;
      CREATE_BUFFER;
      CREATE_CALL;
      CREATE_CLIENT_PRINCIPAL;
      CREATE_OBJECT;
      CREATE_DATABASE;
      CREATE_DATASET;
      CREATE_DATA_SOURCE;
      CREATE_INDEX;
      CREATE_QUERY;
      CREATE_SAX_ATTRIBUTES;
      CREATE_SAX_READER;
      CREATE_SAX_WRITER;
      CREATE_SERVER;
      CREATE_SRV_SOCKET;
      CREATE_SOAP_HEADER;
      CREATE_SOAP_HEADER_ENTRYREF;
      CREATE_SOCKET;
      CREATE_TABLE;
      CREATE_TEMP_TABLE;
      CREATE_VIEW;
      CREATE_WIDGET;
      CREATE_WIDGET_POOL;
      CREATE_X_DOCUMENT;
      CREATE_X_NODEREF;
      DDE_ADVISE;
      DDE_EXECUTE;
      DDE_GET;
      DDE_INITIATE;
      DDE_REQUEST;
      DDE_SEND;
      DDE_TERMINATE;
      DECLARE_CURSOR;
      DEFINE_BROWSE;
      DEFINE_BUFFER;
      DEFINE_BUTTON;
      DEFINE_DATASET;
      DEFINE_DATA_SOURCE;
      DEFINE_EVENT;
      DEFINE_FIELD;
      DEFINE_FRAME;
      DEFINE_IMAGE;
      DEFINE_MENU;
      DEFINE_PARAMETER;
      DEFINE_PROPERTY;
      DEFINE_QUERY;
      DEFINE_RECTANGLE;
      DEFINE_STREAM;
      DEFINE_SUB_MENU;
      DEFINE_TEMP_TABLE;
      DEFINE_VARIABLE;
      DEFINE_WORK_TABLE;
      DELETE_ALIAS;
      DELETE_FROM;
      DELETE_OBJECT;
      DELETE_PROCEDURE;
      DELETE_WIDGET;
      DELETE_WIDGET_POOL;
      DISABLE_TRIGGERS;
      DROP_INDEX;
      DROP_TABLE;
      DROP_VIEW;
      EMPTY_TEMP_TABLE;
      GROUP_BY;
      IN_WIDGET_POOL;
      INNER_JOIN;
      INPUT_CLEAR;
      INPUT_CLOSE;
      INPUT_FROM;
      INPUT_THRU;
      INPUT_OUTPUT_CLOSE;
      INPUT_OUTPUT_THRU;
      INSERT_INTO;
      IS_NULL;
      IS_NOT_NULL;
      LEFT_OUTER_JOIN;
      OPEN_CURSOR;
      OPEN_QUERY;
      ORDER_BY;
      OUTPUT_CLOSE;
      OUTPUT_THRU;
      OUTPUT_TO;
      NOT_BETWEEN;
      NOT_ENTERED;
      NOT_EXISTS;
      NOT_IN;
      NOT_LIKE;
      NOT_NULL;
      PROCESS_EVENTS;
      PUT_CURSOR;
      PUT_SCREEN;
      RELEASE_EXTERNAL;
      RELEASE_OBJECT;
      RIGHT_OUTER_JOIN;
      RUN_SUPER;
      RUN_STORED_PROCEDURE;
      SAVE_CACHE;
      SYSTEM_DIALOG_COLOR;
      SYSTEM_DIALOG_FONT;
      SYSTEM_DIALOG_GET_DIR;
      SYSTEM_DIALOG_GET_FILE;
      SYSTEM_DIALOG_PRINTER_SETUP;
      TRANSACTION_DISTINCT;
      TRIGGER_PROCEDURE;
      TRANSACTION_MODE_AUTO;
      UPDATE_SQL;
      WHERE_CURRENT_OF;
      WITH_CHECK_OPTION;
      WITH_GRANT_OPTION;
   END_MULTIWORD;
   BEGIN_FUNCTYPES;
      FUNC_CHAR;
      FUNC_CLASS;
      FUNC_COM_HANDLE;
      FUNC_DATE;
      FUNC_DATETIME;
      FUNC_DATETIME_TZ;
      FUNC_DEC;
      FUNC_INT;
      FUNC_INT64;
      FUNC_LOGICAL;
      FUNC_LONGCHAR;
      FUNC_RECID;
      FUNC_ROWID;
      FUNC_HANDLE;
      FUNC_RAW;
      FUNC_MEMPTR;
      FUNC_POLY;
   END_FUNCTYPES;
   BEGIN_LVALUE;
      BEGIN_VARTYPES;
         VAR_CHAR;
         VAR_CLASS;
         VAR_COM_HANDLE;
         VAR_DATE;
         VAR_DATETIME;
         VAR_DATETIME_TZ;
         VAR_DEC;
         VAR_HANDLE;
         VAR_INT;
         VAR_INT64;
         VAR_LOGICAL;
         VAR_LONGCHAR;
         VAR_MEMPTR;
         VAR_RAW;
         VAR_RECID;
         VAR_ROWID;
         SYS_HANDLE;
         BEGIN_NATIVE_VARTYPES;
            VAR_BYTE;
            VAR_DOUBLE;
            VAR_FLOAT;
            VAR_LONG;
            VAR_SHORT;
            VAR_ULONG;
            VAR_USHORT;
         END_NATIVE_VARTYPES;
      END_VARTYPES;   
      BEGIN_FIELDTYPES;
         FIELD_BLOB;
         FIELD_CHAR;
         FIELD_CLASS;
         FIELD_CLOB;
         FIELD_COM_HANDLE;
         FIELD_DATE;
         FIELD_DATETIME;
         FIELD_DATETIME_TZ;
         FIELD_DEC;
         FIELD_INT;
         FIELD_INT64;
         FIELD_LOGICAL;
         FIELD_RECID;
         FIELD_ROWID;
         FIELD_HANDLE;
         FIELD_RAW;
      END_FIELDTYPES;   
      BEGIN_WIDGETS;
         WID_BROWSE;
         WID_BROWSE_COL;
         WID_BUTTON;
         WID_COMBO;
         WID_DIALOG;
         WID_EDITOR;
         WID_FILL_IN;
         WID_FRAME;
         WID_IMAGE;
         WID_LITERAL;
         WID_MENU;
         WID_MENU_ITM;
         WID_RADIO;
         WID_RECT;
         WID_SEL_LST;
         WID_SLIDER;
         WID_SUB_MENU;
         WID_TEXT;
         WID_TOGGLE;
         WID_WINDOW;
      END_WIDGETS;
      STREAM;
   END_LVALUE;
   BEGIN_ATTR_METH;
      BEGIN_ATTR;
         ATTR_CHAR;
         ATTR_CLASS;
         ATTR_COM_HANDLE;
         ATTR_DATE;
         ATTR_DATETIME;
         ATTR_DATETIME_TZ;
         ATTR_DEC;
         ATTR_HANDLE;
         ATTR_INT;
         ATTR_INT64;
         ATTR_LOGICAL;
         ATTR_LONGCHAR;
         ATTR_MEMPTR;
         ATTR_RAW;
         ATTR_RECID;
         ATTR_ROWID;
         ATTR_POLY;
      END_ATTR;
      BEGIN_METH;
         METH_CHAR;
         METH_CLASS;
         METH_COM_HANDLE;
         METH_DATE;
         METH_DATETIME;
         METH_DATETIME_TZ;
         METH_DEC;
         METH_HANDLE;
         METH_INT;
         METH_INT64;
         METH_LOGICAL;
         METH_LONGCHAR;
         METH_MEMPTR;
         METH_RAW;
         METH_RECID;
         METH_ROWID;
         METH_POLY;
         METH_VOID;
      END_METH;
   END_ATTR_METH;
   BEGIN_RESERVED;
      KW_U_CTRL;
      KW_U_MSG;
      KW_U_PCTRL;
      KW_U_SERIAL;
      KW_ACCUM;
      KW_ADD;
      KW_ASCEND;
      KW_ASC;
      KW_ACT_FORM;
      KW_ACT_WIN;
      KW_ALIAS;
      KW_ALL;
      KW_ALTER;
      KW_AMBIG;
      KW_ANALYZ;
      KW_AND;
      KW_ANY;
      KW_APPLY;
      KW_AS;
      KW_ASYNC;
      KW_ASSIGN;
      KW_AT;
      KW_ATTR;
      KW_AUD_CTRL;
      KW_AUD_POL;
      KW_AUTHORZN;
      KW_AUTO_RET;
      KW_AVAIL;
      KW_B_ENDIAN;
      KW_BACKGRND;
      KW_B4_HIDE;
      KW_BEGINS;
      KW_BELL;
      KW_BETWEEN;
      KW_BLANK;
      KW_BREAK;
      KW_BTOS;
      KW_BUF_COMP;
      KW_BUF_COPY;
      KW_BY;
      KW_CALL;
      KW_CAN_DO;
      KW_CAN_FIND;
      KW_CASE;
      KW_CASE_SEN;
      KW_CAST;
      KW_CENTER;
      KW_CHR;
      KW_CHECK;
      KW_CLEAR;
      KW_CLIP;
      KW_CODEBASE;
      KW_COL_LAB;
      KW_COLON;
      KW_COLOR;
      KW_COLUMNS;
      KW_COMPILER;
      KW_CONN_ED;
      KW_CONTROL;
      KW_COUNT_OF;
      KW_CPY_LOB;
      KW_CREATE;
      KW_CTOS;
      KW_CURRENT;
      KW_CUR_CHG;
      KW_CUR_LANG;
      KW_CUR_WIN;
      KW_CURSOR;
      KW_DATA_SRC;
      KW_DATABASE;
      KW_DATASRV;
      KW_DATASET;
      KW_DB_CTXT;
      KW_DB_REF;
      KW_DB_REM_H;
      KW_DBCP;
      KW_DBCOLL;
      KW_DBNAME;
      KW_DBPARAM;
      KW_DBREST;
      KW_DBTASKID;
      KW_DBTYPE;
      KW_DBVERS;
      KW_DDE;
      KW_DEBLANK;
      KW_DEBUGGER;
      KW_DBG_LST;
      KW_DECIMALS;
      KW_DECLARE;
      KW_DEFAULT;
      KW_DEF_NOXL;
      KW_DEF_WIN;
      KW_DEFINE;
      KW_DELETE;
      KW_DELIMIT;
      KW_DESCEND;
      KW_DICT;
      KW_DISABLE;
      KW_DIS_A_ZA;
      KW_DISCONN;
      KW_DISP;
      KW_DISTINCT;
      KW_DO;
      KW_DOS;
      KW_DOWN;
      KW_DROP;
      KW_DSET_HND;
      KW_DSL_MGR;
      KW_DYN_FUNC;
      KW_DYN_INVK;
      KW_EACH;
      KW_EDITING;
      KW_ELSE;
      KW_ENABLE;
      KW_ENCODE;
      KW_END;
      KW_ENTRY;
      KW_ERR_STAT;
      KW_ESCAPE;
      KW_EVT_PROC;
      KW_EXC_LOCK;
      KW_EXISTS;
      KW_ETIME;
      KW_EXCEPT;
      KW_EXPORT;
      KW_FETCH;
      KW_FIELD;
      KW_FIL_INFO;
      KW_FILL;
      KW_FIND;
      KW_FIND_CS;
      KW_FIND_GLO;
      KW_FIND_NO;
      KW_FIND_PO;
      KW_FIND_SEL;
      KW_FIND_WA;
      KW_FIRST;
      KW_FIRST_OF;
      KW_FOCUS;
      KW_FONT;
      KW_FOR;
      KW_FORM;
      KW_FORMAT;
      KW_FRAME;
      KW_FR_COL;
      KW_FR_DB;
      KW_FR_DOWN;
      KW_FR_FIELD;
      KW_FR_FILE;
      KW_FR_INDEX;
      KW_FR_LINE;
      KW_FR_NAME;
      KW_FR_ROW;
      KW_FR_VAL;
      KW_FROM;
      KW_FROM_CHR;
      KW_FROM_PIX;
      KW_FUNC_C_T;
      KW_GW;
      KW_GET_A_CT;
      KW_GET_BUFH;
      KW_GET_BYTE;
      KW_GET_CP;
      KW_GET_COLL;
      KW_GET_ERRC;
      KW_GET_ERRR;
      KW_GET_FNAM;
      KW_GET_FOFF;
      KW_GET_K_V;
      KW_GLOBAL;
      KW_GO_ON;
      KW_GO_PEND;
      KW_GRANT;
      KW_GRAPHIC;
      KW_GROUP;
      KW_HEADER;
      KW_HAVING;
      KW_HELP;
      KW_HIDE;
      KW_HOST_B_O;
      KW_IF;
      KW_IMPORT;
      KW_IN;
      KW_INDEX;
      KW_INDICAT;
      KW_INPUT;
      KW_IN_OUT;
      KW_INSERT;
      KW_INTO;
      KW_IS;
      KW_IS_ATTR;
      KW_IS_LEAD;
      KW_JOIN;
      KW_KBLABEL;
      KW_KEYCODE;
      KW_KEYFUNC;
      KW_KEYLAB;
      KW_KEYS;
      KW_KW;
      KW_L_ENDIAN;
      KW_LABEL;
      KW_LAST;
      KW_LASTKEY;
      KW_LAST_EVT;
      KW_LAST_OF;
      KW_LDBNAME;
      KW_LEAVE;
      KW_LIB;
      KW_LIKE;
      KW_LINE_CNT;
      KW_LISTING;
      KW_LOCKED;
      KW_LOG_MGR;
      KW_LOOKUP;
      KW_MACH_CLS;
      KW_MAP;
      KW_MEMBER;
      KW_MOUSE;
      KW_MPE;
      KW_MSG;
      KW_MSG_LINE;
      KW_NEW;
      KW_NEXT;
      KW_NEXT_PMT;
      KW_NO_ATTR;
      KW_NO_ATTRL;
      KW_NO_ERROR;
      KW_NO_FILL;
      KW_NO_FOCUS;
      KW_NO_HIDE;
      KW_NO_LABEL;
      KW_NO_LOBS;
      KW_NO_LOCK;
      KW_NO_MAP;
      KW_NO_MSG;
      KW_NO_PAUSE;
      KW_NO_PRE;
      KW_NO_RET_V;
      KW_NO_UNDO;
      KW_NO_VALID;
      KW_NO_WAIT;
      KW_NOT;
      KW_NOW;
      KW_NULL;
      KW_NUM_ALIA;
      KW_NUM_DBS;
      KW_NUM_ENT;
      KW_OF;
      KW_OFF;
      KW_OLD;
      KW_ON;
      KW_OPEN;
      KW_OR;
      KW_OUTPUT;
      KW_OPSYS;
      KW_OPTION;
      KW_OS2;
      KW_OS_APPND;
      KW_OS_COPY;
      KW_OS_CMD;
      KW_OS_MKDIR;
      KW_OS_DEL;
      KW_OS_DIR;
      KW_OS_REN;
      KW_OTHER;
      KW_OVERLAY;
      KW_PAGE;
      KW_PAGE_B;
      KW_PAGE_NUM;
      KW_PAGE_T;
      KW_PARM;
      KW_PASSWD_F;
      KW_PAUSE;
      KW_PDBNAME;
      KW_PERSIST;
      KW_PIXELS;
      KW_PREPROC;
      KW_PRIVILEG;
      KW_PROCESS;
      KW_PROC_C_T;
      KW_PROC_HND;
      KW_PROC_ST;
      KW_PROFILER;
      KW_PROGNAME;
      KW_PROGRESS;
      KW_PRMT_FOR;
      KW_PROMSGS;
      KW_PROPATH;
      KW_PROVER;
      KW_PUT;
      KW_PUT_BYTE;
      KW_PUT_K_V;
      KW_QUERY;
      KW_QRY_CLOS;
      KW_QRY_OFF;
      KW_QRY_TUNE;
      KW_QUIT;
      KW_R_INDEX;
      KW_RCOD_INF;
      KW_READ_AVL;
      KW_READ_E_N;
      KW_READKEY;
      KW_RECID;
      KW_RECT;
      KW_RELEASE;
      KW_REPEAT;
      KW_REPL_CRE;
      KW_REPL_DEL;
      KW_REPL_WRI;
      KW_REPOS;
      KW_REPOS_B;
      KW_REPOS_F;
      KW_REPOS_2R;
      KW_REPOS_2I;
      KW_RETAIN;
      KW_RETRY;
      KW_RETURN;
      KW_RETURNS;
      KW_REVERT;
      KW_REVOKE;
      KW_ROW_CRTD;
      KW_ROW_DELD;
      KW_ROW_MODD;
      KW_ROW_UMOD;
      KW_RUN;
      KW_SAVE;
      KW_SAX_COMP;
      KW_SAX_PARE;
      KW_SAX_RUNN;
      KW_SAX_UNIN;
      KW_SAX_WBEG;
      KW_SAX_WCOM;
      KW_SAX_WCON;
      KW_SAX_WELM;
      KW_SAX_WERR;
      KW_SAX_WIDL;
      KW_SAX_WTAG;
      KW_SCHEMA;
      KW_SCREEN;
      KW_SCRN_IO;
      KW_SCRN_LNS;
      KW_SCROLL;
      KW_SDBNAME;
      KW_SEAR_SLF;
      KW_SEAR_TRG;
      KW_SEARCH;
      KW_SECUR_P;
      KW_SEEK;
      KW_SELECT;
      KW_SELF;
      KW_SESSION;
      KW_SET;
      KW_SET_A_CT;
      KW_SETUSER;
      KW_SH_LOCK;
      KW_SHARED;
      KW_SHOW_ST;
      KW_SKIP;
      KW_SKIP_D_R;
      KW_SOME;
      KW_SPACE;
      KW_STATUS;
      KW_STREAM;
      KW_STRM_HND;
      KW_STRM_IO;
      KW_SUPER;
      KW_SYS_DLG;
      KW_TABLE;
      KW_TAB_NUM;
      KW_TERM;
      KW_TEXT;
      KW_THEN;
      KW_THIS_OBJ;
      KW_THIS_PRC;
      KW_TIME;
      KW_TITLE;
      KW_TO;
      KW_TOP_ONLY;
      KW_TO_ROWID;
      KW_TRANS;
      KW_TRIGGER;
      KW_TRIGGERS;
      KW_TRIM;
      KW_TXT_CURS;
      KW_UNDO;
      KW_UNION;
      KW_UNIQUE;
      KW_UNIX;
      KW_UNDERLIN;
      KW_UNFORMAT;
      KW_UP;
      KW_UPDATE;
      KW_USE_IDX;
      KW_USERID;
      KW_USING;
      KW_V6FRAME;
      KW_VALUE;
      KW_VAL_CHG;
      KW_VALUES;
      KW_VIEW;
      KW_VIEW_AS;
      KW_VMS;
      KW_WAIT_FOR;
      KW_WHEN;
      KW_WHERE;
      KW_WHILE;
      KW_WINDOW;
      KW_WIN_DMIN;
      KW_WIN_MAX;
      KW_WIN_MIN;
      KW_WIN_NORM;
      KW_WITH;
      KW_WORK_TAB;
      KW_WRITE;
      KW_XCODE;
      KW_XREF;
      KW_XREF_XML;
   END_RESERVED;
   BEGIN_UNRESERVED;
      KW_ABORT;
      KW_ABS;
      KW_ABSTRACT;
      KW_ACC_CHG;
      KW_ACC_RCHG;
      KW_ACCEL;
      KW_ACTIVE;
      KW_ACTOR;
      KW_ADD_BUF;
      KW_ADD_C_C;
      KW_ADD_C_F;
      KW_ADD_EVTP;
      KW_ADD_F_F;
      KW_ADD_1ST;
      KW_ADD_HENT;
      KW_ADD_INVL;
      KW_ADD_IDXF;
      KW_ADD_LAST;
      KW_ADD_L_C;
      KW_ADD_LIKF;
      KW_ADD_LIKI;
      KW_ADD_NEWF;
      KW_ADD_NEWI;
      KW_ADD_PREL;
      KW_ADD_REL;
      KW_ADD_SLOC;
      KW_ADD_SRCB;
      KW_ADD_SUP;
      KW_ADM_DATA;
      KW_ADM_TO;
      KW_ADVISE;
      KW_AFT_BUFF;
      KW_AFT_FILL;
      KW_AFT_R_F;
      KW_AFT_ROID;
      KW_AFT_TBL;
      KW_ALERT_BX;
      KW_ALLW_C_S;
      KW_ALLW_REP;
      KW_ALT_KEY;
      KW_ALW_ON_T;
      KW_ANSI_ONL;
      KW_ANY_KEY;
      KW_ANY_PRT;
      KW_ANYWHERE;
      KW_APPEND;
      KW_APPEND_C;
      KW_APPEND_L;
      KW_APPL_A_B;
      KW_APPL;
      KW_APPL_CID;
      KW_APPL_CBK;
      KW_APP_INFO;
      KW_APP_PW;
      KW_APP_UID;
      KW_ARR_MSG;
      KW_ASCII_2H;
      KW_ASK_OVER;
      KW_ASSEMBLY;
      KW_ASYNC_RC;
      KW_ASYNC_RH;
      KW_ATT_DSRC;
      KW_ATT_PLST;
      KW_ATTR_NAM;
      KW_AUD_ENAB;
      KW_AUD_EV_C;
      KW_AUTHEN_F;
      KW_AUTO_COM;
      KW_AUTO_DEL;
      KW_AUTO_D_X;
      KW_AUTO_END;
      KW_AUTO_GO;
      KW_AUTO_IND;
      KW_AUTO_RES;
      KW_AUTO_SYN;
      KW_AUTO_VAL;
      KW_AUTO_ZAP;
      KW_AUTOMATC;
      KW_AVERAGE;
      KW_AVG;
      KW_AVL_FMTS;
      KW_AVL_MSGS;
      KW_BACKSP;
      KW_BACKWARD;
      KW_BACK_TAB;
      KW_BAS_LOGG;
      KW_BASE64_D;
      KW_BASE64_E;
      KW_BASE_ADE;
      KW_BASE_KEY;
      KW_BATCH_MO;
      KW_BATCH_SZ;
      KW_B4_BUFF;
      KW_B4_FILL;
      KW_B4_R_F;
      KW_B4_ROWID;
      KW_B4_TABLE;
      KW_BEG_EV_G;
      KW_BGCOLOR;
      KW_BINARY;
      KW_BIND;
      KW_BIND_WH;
      KW_BLOB;
      KW_BLOCK;
      KW_BLK_LVL;
      KW_BLK_IT_D;
      KW_BORD_B_C;
      KW_BORD_B_P;
      KW_BORD_L_C;
      KW_BORD_L_P;
      KW_BORD_R_C;
      KW_BORD_R_P;
      KW_BORD_T_C;
      KW_BORD_T_P;
      KW_BOTH;
      KW_BOTTOM;
      KW_BOTTOM_C;
      KW_BOX;
      KW_BOX_SEL;
      KW_BREAK_L;
      KW_BROWSE;
      KW_BUFFER;
      KW_BUFFER_C;
      KW_BUF_CREA;
      KW_BUF_DEL;
      KW_BUF_FLD;
      KW_BUF_HNDL;
      KW_BUF_NAME;
      KW_BUFFER_L;
      KW_BUF_REL;
      KW_BUF_VLID;
      KW_BUF_VAL;
      KW_BUTTON;
      KW_BY_PTR;
      KW_BY_REF;
      KW_BY_VALUE;
      KW_BY_VAR_P;
      KW_BYTE;
      KW_BYTES_R;
      KW_BYTES_W;
      KW_CACHE;
      KW_CACHE_SZ;
      KW_CALL_NAM;
      KW_CALL_TYP;
      KW_CANC_BRK;
      KW_CANCEL_B;
      KW_CANCEL_P;
      KW_CANCEL_R;
      KW_CANCELLD;
      KW_CAN_CREA;
      KW_CAN_DEL;
      KW_CAN_DUMP;
      KW_CAN_LOAD;
      KW_CAN_QRY;
      KW_CAN_READ;
      KW_CAN_SET;
      KW_CAN_WRT;
      KW_CAPS;
      KW_CARE_PNT;
      KW_CBIT;
      KW_CATCH;
      KW_CDECL;
      KW_CFG_NAME;
      KW_CHAINED;
      KW_CHAR;
      KW_CHAR_LEN;
      KW_CHARSET;
      KW_CHECKED;
      KW_CHECK_AM;
      KW_CHLD_BUF;
      KW_CHLD_NUM;
      KW_CHOICES;
      KW_CHOOSE;
      KW_CLASS;
      KW_CLOB;
      KW_CLR_SEL;
      KW_CLR_AP_C;
      KW_CLR_LOG;
      KW_CLR_S_AR;
      KW_CLS_TYPE;
      KW_CLNT_C_I;
      KW_CLNT_PRL;
      KW_CLNT_TTY;
      KW_CLNT_TYP;
      KW_CLNT_WS;
      KW_CLONE_ND;
      KW_CLOSE;
      KW_CLOSE_AP;
      KW_CLOSE_LG;
      KW_CNCL_R_A;
      KW_CNTRL_BX;
      KW_CODE;
      KW_CP_CVT;
      KW_CP;
      KW_COLLATE;
      KW_COLON_AL;
      KW_COLR_TAB;
      KW_COL;
      KW_COL_BGC;
      KW_COL_CP;
      KW_COL_DC;
      KW_COL_FGC;
      KW_COL_FONT;
      KW_COL_MOV;
      KW_COL_PFC;
      KW_COL_OF;
      KW_COL_R_O;
      KW_COL_RES;
      KW_COL_SCR;
      KW_COM_HNDL;
      KW_COM_SELF;
      KW_COMBO_BX;
      KW_COMMAND;
      KW_COMPARE;
      KW_COMPARES;
      KW_COMPILE;
      KW_COMPLETE;
      KW_CONN;
      KW_CONSTRUC;
      KW_CTX;
      KW_CTX_H;
      KW_CTX_H_F;
      KW_CTX_H_ID;
      KW_CTX_POP;
      KW_CONTAINS;
      KW_CONTENTS;
      KW_CNTRL_FR;
      KW_CNTRL_NM;
      KW_CONTROLS;
      KW_CVT_DT;
      KW_CONVERT;
      KW_CVT_2OFF;
      KW_CVT_3D_C;
      KW_COOKIE_D;
      KW_COPY;
      KW_CPY_DSET;
      KW_CPY_SATR;
      KW_CPY_TTBL;
      KW_COUNT;
      KW_COVERAGE;
      KW_CPCASE;
      KW_CPCOLL;
      KW_CPINTERN;
      KW_CPLOG;
      KW_CPPRINT;
      KW_CPRCODEI;
      KW_CPRCODEO;
      KW_CPSTREAM;
      KW_CPTERM;
      KW_CRC_VAL;
      KW_CREAT_LK;
      KW_CREAT_OA;
      KW_CREAT_ND;
      KW_CREAT_NN;
      KW_CREAT_RL;
      KW_CREAT_TF;
      KW_CURRENCY;
      KW_CUR_COL;
      KW_CUR_DATE;
      KW_CUR_RES;
      KW_CUR_R_M;
      KW_CUR_VAL;
      KW_CUR_CHAR;
      KW_CUR_ENV;
      KW_CUR_DOWN;
      KW_CUR_ITER;
      KW_CUR_LEFT;
      KW_CUR_LINE;
      KW_CUR_OFF;
      KW_CUR_RGHT;
      KW_CUR_UP;
      KW_CUT;
      KW_DATA_BND;
      KW_DATA_E_R;
      KW_DATA_REL;
      KW_DATA_SCM;
      KW_DATA_SM;
      KW_DATATYPE;
      KW_DATE;
      KW_DATETIME;
      KW_DATE_TZ;
      KW_DATE_FMT;
      KW_DAY;
      KW_DCOLOR;
      KW_DDE_ERR;
      KW_DDE_ID;
      KW_DDE_ITEM;
      KW_DDE_NAME;
      KW_DDE_NOTI;
      KW_DDE_TOPI;
      KW_DEBUG;
      KW_DECRYPT;
      KW_DBG_ALRT;
      KW_DEC;
      KW_DECL_NSP;
      KW_DEF_ACTN;
      KW_DEFLT_BN;
      KW_DEF_BUFH;
      KW_DEF_COMM;
      KW_DEF_EXTN;
      KW_DEF_STR;
      KW_DEF_POP;
      KW_DEF_VAL;
      KW_DEF_UEVM;
      KW_DELEGATE;
      KW_DEL_CHAR;
      KW_DEL_COL;
      KW_DEL_COOK;
      KW_DEL_C_R;
      KW_DEL_E_L;
      KW_DEL_FLD;
      KW_DEL_H_EN;
      KW_DEL_LINE;
      KW_DEL_NODE;
      KW_DEL_PROC;
      KW_DEL_R_L;
      KW_DEL_S_R;
      KW_DEL_S_RS;
      KW_DEL_WORD;
      KW_DESCR;
      KW_DESEL_FR;
      KW_DESEL_R;
      KW_DESEL_SR;
      KW_DESELCTN;
      KW_DESTRUCT;
      KW_DET_DSRC;
      KW_DIALOG;
      KW_DIR;
      KW_DIRECTRY;
      KW_DIS_D_TR;
      KW_DIS_L_TR;
      KW_DIS_REDR; // FWD extension, not real 4GL!
      KW_DISABL_C;
      KW_DISABLED;
      KW_DISPATCH;
      KW_DISP_MSG;
      KW_DISP_TZ;
      KW_DISP_TYP;
      KW_DOMAIN_D;
      KW_DOMAIN_N;
      KW_DOMAIN_T;
      KW_DOUBLE;
      KW_DRAG_EN;
      KW_DROP_DWN;
      KW_DROP_FN;
      KW_DROP_LST;
      KW_DROP_TAR;
      KW_DUMP;
      KW_DUMP_LGN;
      KW_DYN_CAST;
      KW_DYN_CURV;
      KW_DYN_NEW;
      KW_DYN_NEXV;
      KW_DYNAMIC;
      KW_ECHO;
      KW_EDGE_C;
      KW_EDGE_P;
      KW_EDIT_C_P;
      KW_EDIT_C_U;
      KW_EDIT_CLR;
      KW_EDIT_CPY;
      KW_EDIT_CUT;
      KW_EDIT_PAS;
      KW_EDIT_UND;
      KW_EDITOR;
      KW_EDIT_B_T;
      KW_EDIT_TAB;
      KW_EMPTY;
      KW_EMPTY_DS;
      KW_EMPTY_SN;
      KW_EMPTY_TT;
      KW_ENABLE_C;
      KW_ENABLE_E;
      KW_ENC_AMK;
      KW_ENC_SALT;
      KW_ENCODING;
      KW_ENCRYPT;
      KW_END_B_SN;
      KW_END_DOC;
      KW_END_ELEM;
      KW_END_EV_G;
      KW_END_SEAR;
      KW_END_USER;
      KW_END_ERR;
      KW_END_F_D;
      KW_END_MOV;
      KW_END_RESZ;
      KW_END_RRES;
      KW_ENDKEY;
      KW_ENT_TLST;
      KW_ENTERED;
      KW_ENTER_MB;
      KW_ERROR;
      KW_ERR_CODE;
      KW_ERR_COL;
      KW_ERR_OBJD;
      KW_ERR_ROW;
      KW_ERR_S_T;
      KW_ERR_STR;
      KW_EQ;
      KW_EVENTS;
      KW_EVT_GRID;
      KW_EVT_TYPE;
      KW_E_PROC_C;
      KW_EXCL_ID;
      KW_EXCL_WEB;
      KW_EXECUTE;
      KW_EXEC_LOG;
      KW_EXIT;
      KW_EXP;
      KW_EXPAND;
      KW_EXPANDBL;
      KW_EXPORT_P;
      KW_EXPLICIT;
      KW_EXTENDED;
      KW_EXTENT;
      KW_EXTERN;
      KW_FETCH_SR;
      KW_FGCOLOR;
      KW_FILE;
      KW_FIL_C_D;
      KW_FIL_C_T;
      KW_FIL_NAME;
      KW_FIL_OFF;
      KW_FIL_M_D;
      KW_FIL_M_T;
      KW_FIL_SIZE;
      KW_FIL_TYPE;
      KW_FILL_IN;
      KW_FILL_MOD;
      KW_FILL_WST;
      KW_FILLED;
      KW_FILTERS;
      KW_FINAL;
      KW_FINALLY;
      KW_FIND_BR;
      KW_FIND_CUR;
      KW_FIND_FD;
      KW_FIND_1ST;
      KW_FIND_LST;
      KW_FIND_NXT;
      KW_FIND_PRV;
      KW_FIND_UNI;
      KW_FINDER;
      KW_FIR_DSET;
      KW_FIRST_AR;
      KW_FIRST_BU;
      KW_FIRST_CH;
      KW_FIRST_CO;
      KW_FIRST_DS;
      KW_FIRST_FM;
      KW_FIRST_OB;
      KW_FIRST_PR;
      KW_FIRST_QR;
      KW_FIRST_SR;
      KW_FIRST_SS;
      KW_FIRST_SO;
      KW_FIRST_TI;
      KW_FIT_LCOL;
      KW_FIX_CP;
      KW_FIXD_ONL;
      KW_FLAT_BUT;
      KW_FLOAT;
      KW_FONT_TAB;
      KW_FOCUS_R;
      KW_FOCUS_RS;
      KW_FORCE_F;
      KW_FORE;
      KW_F_KEY_H;
      KW_FORM_INP;
      KW_FORM_LIN;
      KW_FMT_DT;
      KW_FORMATTE;
      KW_FORWARD;
      KW_FRAGMENT;
      KW_FREQ;
      KW_FROM_CUR;
      KW_FR_SPACE;
      KW_FR_X;
      KW_FR_Y;
      KW_FUNCT;
      KW_FULL_H_C;
      KW_FULL_H_P;
      KW_FULL_W_C;
      KW_FULL_W_P;
      KW_FULLPATH;
      KW_FWD_ONLY;
      KW_GEN_MD5;
      KW_GEN_PBEK;
      KW_GEN_PBES;
      KW_GEN_RNDK;
      KW_GEN_UUID;
      KW_GET;
      KW_GETATTRI;
      KW_GET_ATTR;
      KW_GET_A_N;
      KW_GET_BDAT;
      KW_GET_BITS;
      KW_GET_B_OR;
      KW_GET_BLUE;
      KW_GET_BR_C;
      KW_GET_BYTS;
      KW_GET_B_A;
      KW_GET_CFG;
      KW_GET_CFGV;
      KW_GET_CBPC;
      KW_GET_CBPN;
      KW_GET_CGIL;
      KW_GET_CGIV;
      KW_GET_CGL;
      KW_GET_CGLV;
      KW_GET_CHG;
      KW_GET_CHLD;
      KW_GET_CLL;
      KW_GET_CODP;
      KW_GET_CREL;
      KW_GET_CUR;
      KW_GET_DIR;
      KW_GET_DBL;
      KW_GET_DS_B;
      KW_GET_D_E;
      KW_GET_D_F;
      KW_GET_DYN;
      KW_GET_FILE;
      KW_GET_1ST;
      KW_GET_FLT;
      KW_GET_GRN;
      KW_GET_HD_E;
      KW_GET_IBNN;
      KW_GET_IBQN;
      KW_GET_I64;
      KW_GET_ITER;
      KW_GET_LAST;
      KW_GET_LNBI;
      KW_GET_LONG;
      KW_GET_L_V;
      KW_GET_MOP; // FWD extension, not real 4GL!
      KW_GET_MSG;
      KW_GET_MSGG;
      KW_GET_MSGS;
      KW_GET_NEXT;
      KW_GET_NODE;
      KW_GET_NUM;
      KW_GET_PAR;
      KW_GET_PTR;
      KW_GET_PREV;
      KW_GET_PROP;
      KW_GET_PRT;
      KW_GET_QNBI;
      KW_GET_R_R;
      KW_GET_RED;
      KW_GET_REL;
      KW_GET_RGB;
      KW_GET_SHRT;
      KW_GET_SELW;
      KW_GET_SER;
      KW_GET_SIG;
      KW_GET_SRCB;
      KW_GET_SZ;
      KW_GET_S_O;
      KW_GET_STR;
      KW_GET_T_S;
      KW_GET_TBI;
      KW_GET_TBNN;
      KW_GET_TBQN;
      KW_GET_TI;
      KW_GET_THCH;
      KW_GET_THPX;
      KW_GET_TOPB;
      KW_GET_TWCH;
      KW_GET_TWPX;
      KW_GET_UBI;
      KW_GET_UL;
      KW_GET_USHT;
      KW_GET_U_F;
      KW_GET_VAL;
      KW_GET_VBI;
      KW_GET_VBNN;
      KW_GET_VBQN;
      KW_GET_WAIT;
      KW_GO;
      KW_GOTO;
      KW_GRD_F_H;
      KW_GRD_F_V;
      KW_GRD_SNAP;
      KW_GRD_UHC;
      KW_GRD_UHP;
      KW_GRD_UWC;
      KW_GRD_UWP;
      KW_GRD_VIS;
      KW_GROUP_BX;
      KW_GET_CGI;
      KW_GET_COOK;
      KW_GET_FLD;
      KW_GT;
      KW_GTE;
      KW_GUID;
      KW_HANDLE;
      KW_HANDLER;
      KW_HAS_LOBS;
      KW_HAS_REC;
      KW_HEIGHT_C;
      KW_HEIGHT_P;
      KW_HELP_TOP;
      KW_HEX_DECD;
      KW_HEX_ENCD;
      KW_HID_FLD;
      KW_HID_FLDL;
      KW_HIDDEN;
      KW_HINT;
      KW_HOME;
      KW_HORIZ;
      KW_HTML_CHS;
      KW_HTML_ENC;
      KW_HTML_EOL;
      KW_HTML_EOP;
      KW_HTML_FRB;
      KW_HTML_FRE;
      KW_HTML_H_B;
      KW_HTML_H_E;
      KW_HTML_T_B;
      KW_HTML_T_E;
      KW_HTMLERR;
      KW_HWND;
      KW_HONOR_PK;
      KW_HONOR_RK;
      KW_ICON;
      KW_ICFPARM;
      KW_IDEPHWND;
      KW_IDEWNTYP;
      KW_IDX_HINT;
      KW_IDX_INFO;
      KW_IGN_CMOD;
      KW_IMAGE;
      KW_IMG_DOWN;
      KW_IMG_INS;
      KW_IMG_SZ;
      KW_IMG_SZ_C;
      KW_IMG_SZ_P;
      KW_IMG_UP;
      KW_IDX_REPO;
      KW_IMM_DISP;
      KW_IMP_NODE;
      KW_IMP_PRNC;
      KW_IMPLEMTS;
      KW_IN_HNDL;
      KW_INC_EX_I;
      KW_INFO;
      KW_INH_BGC;
      KW_INH_FGC;
      KW_INHERITS;
      KW_INIT_SES;
      KW_INIT;
      KW_INIT_DIR;
      KW_INIT_FLT;
      KW_INIT_D_T;
      KW_INITIATE;
      KW_INNER_C;
      KW_INNER_L;
      KW_INNER;
      KW_INPT_VAL;
      KW_INS_ATTR;
      KW_INS_B4;
      KW_INS_BTAB;
      KW_INS_COL;
      KW_INS_FLD;
      KW_INS_FLDD;
      KW_INS_FLDL;
      KW_INS_MODE;
      KW_INS_FILE;
      KW_INS_ROW;
      KW_INS_STR;
      KW_INS_TAB;
      KW_INST_PRC;
      KW_INT;
      KW_INT64;
      KW_INT_ENT;
      KW_INTERFAC;
      KW_INTERVAL;
      KW_INVOKE;
      KW_IS_CLASS;
      KW_IS_COLCP;
      KW_IS_CP_FX;
      KW_IS_OPEN;
      KW_IS_P_SET;
      KW_IS_R_SEL;
      KW_IS_SEL;
      KW_IS_XML;
      KW_ISO_DATE;
      KW_ITEM;
      KW_ITEM_ROW;
      KW_ITER_CHG;
      KW_IUNKNOWN;
      KW_JOIN_BY;
      KW_KEEP_CON;
      KW_KEEP_ZOR;
      KW_KEEP_MSG;
      KW_KEEP_SEC;
      KW_KEEP_TAB;
      KW_KEY;
      KW_KW_ALL;
      KW_LAB_BGC;
      KW_LAB_DC;
      KW_LAB_FGC;
      KW_LAB_FONT;
      KW_LAB_PFC;
      KW_LABELS;
      KW_LANDSCAP;
      KW_LANGUAGE;
      KW_LARGE;
      KW_LAST_BAT;
      KW_LAST_FRM;
      KW_LAST_OBJ;
      KW_LAST_TI;
      KW_LG_2_SM;
      KW_LAST_AR;
      KW_LAST_CH;
      KW_LAST_PRC;
      KW_LAST_SRV;
      KW_LAST_SS;
      KW_LAST_SOC;
      KW_LCHR_2NV;
      KW_LEFT;
      KW_LEFT_AL;
      KW_LEFT_END;
      KW_LEFT_MC;
      KW_LEFT_MDC;
      KW_LEFT_MD;
      KW_LEFT_MU;
      KW_LENGTH;
      KW_LINE;
      KW_L_TRIM;
      KW_LISTINGS;
      KW_LST_PAIR;
      KW_LIST_ITM;
      KW_LIST_PNM;
      KW_LIT_QSTN;
      KW_LST_EVNT;
      KW_LST_QRY;
      KW_LST_SET;
      KW_LST_WID;
      KW_LOAD;
      KW_LOAD_DMN;
      KW_LOAD_ICO;
      KW_LOAD_IMG;
      KW_LOAD_I_D;
      KW_LOAD_I_I;
      KW_LOAD_I_U;
      KW_LOAD_M_P;
      KW_LOAD_PIC;
      KW_LOAD_S_I;
      KW_LOADCTRL;
      KW_LOB_DIR;
      KW_LOC_C_N;
      KW_LOC_HOST;
      KW_LOC_L_N;
      KW_LOC_NAME;
      KW_LOC_PORT;
      KW_LOC_P_ID;
      KW_LOC_S_ID;
      KW_LOC_TYPE;
      KW_LOCK_REG;
      KW_LOG;
      KW_LOG_A_EV;
      KW_LOG_E_TS;
      KW_LOG_EN_T;
      KW_LOG_ID;
      KW_LOG_HOST;
      KW_LOG_STAT;
      KW_LOG_THRS;
      KW_LOGF_NAM;
      KW_LOGG_LEV;
      KW_LOGICAL;
      KW_LOGOUT;
      KW_LOOKAHD;
      KW_LONG;
      KW_LONGCHAR;
      KW_LC;
      KW_LT;
      KW_LTE;
      KW_MAINMENU;
      KW_MAND;
      KW_MAN_HIGH;
      KW_MARG_EX;
      KW_MARG_H_C;
      KW_MARG_H_P;
      KW_MARG_W_C;
      KW_MARG_W_P;
      KW_MATCHES;
      KW_MAX;
      KW_MAX_BTN;
      KW_MAX_CHAR;
      KW_MAX_D_G;
      KW_MAX_HT;
      KW_MAX_H_C;
      KW_MAX_H_P;
      KW_MAX_W_C;
      KW_MAX_W_P;
      KW_MAX_ROWS;
      KW_MAX_SZ;
      KW_MAX_VAL;
      KW_MAX_WID;
      KW_MAXIMIZE;
      KW_MD5_VAL;
      KW_MD5_DIG;
      KW_MEMPTR;
      KW_MPTR_2NV;
      KW_MENU;
      KW_MENU_BAR;
      KW_MENU_DRP;
      KW_MENU_ITM;
      KW_MENU_KEY;
      KW_MENU_MOU;
      KW_MERGE_BF;
      KW_MERGE_CH;
      KW_MERGE_RC;
      KW_METHOD;
      KW_MSG_AREA;
      KW_MSG_AFNT;
      KW_MSG_DIG;
      KW_MESSAGES;
      KW_MDL_MC;
      KW_MDL_MDC;
      KW_MDL_MD;
      KW_MDL_MU;
      KW_MIN;
      KW_MIN_BTN;
      KW_MIN_CWCH;
      KW_MIN_CWPX;
      KW_MIN_H_C;
      KW_MIN_H_P;
      KW_MIN_SCHM;
      KW_MIN_SZ;
      KW_MIN_VAL;
      KW_MIN_W_C;
      KW_MIN_W_P;
      KW_MNEMON; // FWD extension, not real 4GL!
      KW_MODIFIED;
      KW_MOD;
      KW_MONTH;
      KW_MOU_XC;  
      KW_MOU_XDC; 
      KW_MOU_XD;  
      KW_MOU_XU;  
      KW_MOU_MC;  
      KW_MOU_MDC; 
      KW_MOU_MD;  
      KW_MOU_MU;  
      KW_MOU_MVC; 
      KW_MOU_MVDC;
      KW_MOU_MVD; 
      KW_MOU_MVU;
      KW_MOU_PTR;
      KW_MOU_SC;  
      KW_MOU_SDC; 
      KW_MOU_SD;  
      KW_MOU_SU;  
      KW_MOVABLE;
      KW_MOVE;
      KW_MOV_A_T;
      KW_MOV_B_T;
      KW_MOV_COL;
      KW_MOV_2_B;
      KW_MOV_2_T;
      KW_MOV_2EOF;
      KW_MTIME;
      KW_MULT_CMP;
      KW_MULTIPLE;
      KW_MULT_KEY;
      KW_MULTI_IN;
      KW_MUST_EXI;
      KW_MUST_UND;
      KW_NAME;
      KW_NAMESP_P;
      KW_NAMESP_U;
      KW_NATIVE;
      KW_NE;
      KW_NEEDS_AP;
      KW_NEEDS_PR;
      KW_NESTED;
      KW_NEXT_COL;
      KW_NEXT_ERR;
      KW_NEXT_FR;
      KW_NEXT_RID;
      KW_NEXT_SIB;
      KW_NEXT_TAB;
      KW_NEXT_VAL;
      KW_NEXTWORD;
      KW_NEW_INST;
      KW_NEW_LINE;
      KW_NEW_ROW;
      KW_NNMSP_SL;
      KW_NORMAL;
      KW_NORMALZE;
      KW_NO_APPLY;
      KW_NO_ARMSG;
      KW_NO_ASSGN;
      KW_NO_AUTOV;
      KW_NO_BIND;
      KW_NO_BOX;
      KW_NO_COLS;
      KW_NO_CONS;
      KW_NO_CVT;
      KW_NO_CV_3D;
      KW_NO_CUR_V;
      KW_NO_DEBUG;
      KW_NO_DRAG;
      KW_NO_ECHO;
      KW_NO_EM_SP;
      KW_NO_HELP;
      KW_NO_IDX_H;
      KW_NO_INHBG;
      KW_NO_INHFG;
      KW_NO_JOIN;
      KW_NO_LOOKA;
      KW_NO_ROW_M;
      KW_NO_SCH_M;
      KW_NO_SEP_C;
      KW_NO_SEPS;
      KW_NO_SCR_V;
      KW_NO_TAB_S;
      KW_NO_UNDL;
      KW_NO_WRAP;
      KW_NODE_VAL;
      KW_NODE_V2M;
      KW_NODV_2LC;
      KW_NONE;
      KW_NUM_BUFF;
      KW_NUM_BUTT;
      KW_NUM_CHLN;
      KW_NUM_CH_R;
      KW_NUM_COL;
      KW_NUM_COPY;
      KW_NUM_DROP;
      KW_NUM_FLD;
      KW_NUM_FMTS;
      KW_NUM_HD_E;
      KW_NUM_ITMS;
      KW_NUM_ITER;
      KW_NUM_LOGF;
      KW_NUM_LK_C;
      KW_NUM_LNS;
      KW_NUM_MSG;
      KW_NUM_PARM;
      KW_NUM_REF;
      KW_NUM_REL;
      KW_NUM_REPL;
      KW_NUM_RES;
      KW_NUM_SR;
      KW_NUM_SRCB;
      KW_NUM_TOPB;
      KW_NUM_SW;
      KW_NUM_TABS;
      KW_NUM_2RTN;
      KW_NUM_V_C;
      KW_NUM_D_P;
      KW_NUMERIC;
      KW_NUM_FMT;
      KW_NUM_SEP;
      KW_OBJECT;
      KW_OCT_LEN;
      KW_OFF_END;
      KW_OFF_HOME;
      KW_OK;
      KW_OK_CAN;
      KW_OLE_INVL;
      KW_OLE_NAML;
      KW_ON_FR_B;
      KW_ONLY;
      KW_ORG_HAND;
      KW_ORG_ROID;
      KW_ORDER;
      KW_ORD_JOIN;
      KW_OPEN_L_A;
      KW_OPEN_URL;
      KW_OPTIONS;
      KW_ORDINAL;
      KW_OS_DRV;
      KW_OS_ERR;
      KW_OS_G_ENV;
      KW_OUT_JOIN;
      KW_OUTER;
      KW_OUTPUTCT;
      KW_OUT_HDR;
      KW_OUT_HH;
      KW_OUT_MSGS;
      KW_OVERRIDE;
      KW_OWNER;
      KW_OWN_DOC;
      KW_P2J_RC; // P2J-extension, not real 4GL!
      KW_PAGE_DWN;
      KW_PAGE_LFT;
      KW_PAGE_RT;
      KW_PAGE_SZ;
      KW_PAGE_UP;
      KW_PAGE_WID;
      KW_PAGED;
      KW_PAR_BUFF;
      KW_PAR_FLDA;
      KW_PAR_FLDB;
      KW_PAR_IFLD;
      KW_PAR_IREL;
      KW_PAR_REL;
      KW_PAR_W_C;
      KW_PARENT;
      KW_PART_KEY;
      KW_PATHNAME;
      KW_PARSE_ST;
      KW_PASCAL;
      KW_PASTE;
      KW_PBE_H_AL;
      KW_PBE_KEYR;
      KW_PERF;
      KW_PREF_DS;
      KW_PRE_LBL; // FWD extension, not real 4GL!
      KW_PRS_PROC;
      KW_PRS_C_D;
      KW_PFCOLOR;
      KW_PICK;
      KW_PICK_ARE;
      KW_PICK_BTH;
      KW_PIX_COL;
      KW_PIX_ROW;
      KW_POP_MENU;
      KW_POP_ONLY;
      KW_PORTRAIT;
      KW_POS;
      KW_PRECISN;
      KW_PREP_STR;
      KW_PREPARED;
      KW_PRESEL;
      KW_PREV;
      KW_PREV_COL;
      KW_PREV_FR;
      KW_PREV_SIB;
      KW_PREV_T_I;
      KW_PREVWORD;
      KW_PRIMARY;
      KW_PRINTER;
      KW_PRT_C_H;
      KW_PRT_HDC;
      KW_PRT_NAME;
      KW_PRT_PORT;
      KW_PRT_SET;
      KW_PRIVATE;
      KW_PRIV_DAT;
      KW_PROCNAME;
      KW_PROC_COM;
      KW_PROC;
      KW_PROC_W_R;
      KW_PROC_TXT;
      KW_PROC_T_B;
      KW_PRODATAS;
      KW_PROFILNG;
      KW_PROG_SRC;
      KW_PROMPT;  
      KW_PROPERTY;
      KW_PROTECTD;
      KW_PROXY;
      KW_PROX_PWD;
      KW_PROX_UID;
      KW_PUBLIC;
      KW_PUB_ID;
      KW_PUBLISH;
      KW_PUB_EVTS;
      KW_PUT_BITS;
      KW_PUT_BYTS;
      KW_PUT_I64;
      KW_PUT_DBL;
      KW_PUT_FLT;
      KW_PUT_LONG;
      KW_PUT_SHT;
      KW_PUT_STR;
      KW_PUT_UL;
      KW_PUT_USHT;
      KW_QRY_OPEN;
      KW_QRY_PREP;
      KW_QUEST;
      KW_QUE_MSG;
      KW_QUOTER;
      KW_RADIO_B;
      KW_RADIO_S;
      KW_RANDOM;
      KW_RAW;
      KW_RAW_TRAN;
      KW_READ;
      KW_REAL;
      KW_READ_FIL;
      KW_READ_JSN;
      KW_READ_ONL;
      KW_READ_RES;
      KW_READ_XML;
      KW_READ_XSC;
      KW_RECALL;
      KW_REC_LEN;
      KW_RECURSE;
      KW_REF_ONLY;
      KW_REFR_A_P;
      KW_REFRESH;
      KW_REFRABLE;
      KW_REG_DMN;
      KW_REJ_CHGS;
      KW_REJ_RCHG;
      KW_REJECTED;
      KW_REL_FLDS;
      KW_RELS_ACT;
      KW_REM_HOST;
      KW_REM_PORT;
      KW_REMOTE;
      KW_REM_ATTR;
      KW_REM_EVTP;
      KW_REM_CHLD;
      KW_REM_SUP;
      KW_REPLACE;
      KW_REPOS_M;
      KW_REP_CHLD;
      KW_REP_STXT;
      KW_REPORTS;
      KW_REQUEST;
      KW_RESET;
      KW_RESIZE;
      KW_RESIZABL;
      KW_REST_RID;
      KW_RESULT;
      KW_RESUME_D;
      KW_RET_SHAP;
      KW_RETRY_C;
      KW_RET_INS;
      KW_RET_VAL;
      KW_RET_VDT;
      KW_RET_2SD;
      KW_REV_FROM;
      KW_RGB_VAL;
      KW_RIGHT;
      KW_RIGHT_AL;
      KW_RT_END;
      KW_RT_MC; 
      KW_RT_MDC;
      KW_RT_MD; 
      KW_RT_MU; 
      KW_R_TRIM;
      KW_ROLES;
      KW_ROUND;
      KW_ROUNDED;
      KW_ROUTINEL;
      KW_ROW;
      KW_ROW_CRT;
      KW_ROW_DEL;
      KW_ROW_DISP;
      KW_ROW_ENTR;
      KW_ROW_H_C; 
      KW_ROW_H_P; 
      KW_ROW_LEAV;
      KW_ROW_OF;
      KW_ROW_MARK;
      KW_ROW_RESZ;
      KW_ROW_STAT;
      KW_ROW_UPD;
      KW_ROWID;
      KW_RUN_PROC;
      KW_RUN_W_O;
      KW_RULE;
      KW_RULE_ROW;
      KW_RULE_Y;
      KW_SAVE_AS;
      KW_SAVE_FIL;
      KW_SAX_ATTR;
      KW_SAX_PARF;
      KW_SAX_PARN;
      KW_SAX_PARS;
      KW_SAX_READ;
      KW_SAX_WRIT;
      KW_SAVE_RCH;
      KW_SAVE_WST;
      KW_SCH_CHG;
      KW_SCH_LOC;
      KW_SCH_MARS;
      KW_SCH_PATH;
      KW_SCROLLBL;
      KW_SCROLLBA;
      KW_SCR_DELT;
      KW_SCR_LEFT;
      KW_SCROLL_H;
      KW_SCR_MODE;
      KW_SCR_NOT;
      KW_SCR_OFFS;
      KW_SCR_RT;
      KW_SCR_2CR;
      KW_SCR_2ITM;
      KW_SCR_2SR;
      KW_SCRN_VAL;
      KW_SCROLL_V;
      KW_SCR_RPOS;
      KW_SCROLLIN;
      KW_SEAL;
      KW_SEAL_TST;
      KW_SECTION;
      KW_SEL_ALL;
      KW_SEL_END;
      KW_SEL_FOCR;
      KW_SEL_NEXT;
      KW_SEL_PREV;
      KW_SEL_ROW;
      KW_SEL_STRT;
      KW_SEL_TXT;
      KW_SELECTBL;
      KW_SELECTED;
      KW_SELECTN;
      KW_SEL_LST;
      KW_SEND;
      KW_SEND_SQL;
      KW_SENSITIV;
      KW_SEPS;
      KW_SEP_CONN;
      KW_SEP_FGC;
      KW_SERIALZH;
      KW_SERIALZN;
      KW_SERIALZR;
      KW_SERVER;
      KW_SESS_END;
      KW_SESSN_ID;
      KW_SRV_C_B;
      KW_SRV_C_BR;
      KW_SRV_C_C;
      KW_SRV_C_I;
      KW_SRV_OP_M;
      KW_SETTINGS;
      KW_SET_ACTR;
      KW_SET_ACTX;
      KW_SET_ATTR;
      KW_SET_A_L;
      KW_SET_A_N;
      KW_SET_BLUE;
      KW_SET_BRK;
      KW_SET_BUF;
      KW_SET_B_OR;
      KW_SET_CB_P;
      KW_SET_CBAC;
      KW_SET_CLNT;
      KW_SET_COMM;
      KW_SET_C_P;
      KW_SET_CONT;
      KW_SET_COOK;
      KW_SET_DBCL;
      KW_SET_DYN;
      KW_SET_EVMO;
      KW_SET_GRN;
      KW_SET_ISRC;
      KW_SET_M_UN;
      KW_SET_N_F;
      KW_SET_NODE;
      KW_SET_ODST;
      KW_SET_OPT;
      KW_SET_PARM;
      KW_SET_PROP;
      KW_SET_PTR;
      KW_SET_RED;
      KW_SET_RGB;
      KW_SET_RRP;
      KW_SET_RPOS;
      KW_SET_ROLL;
      KW_SET_SEL;
      KW_SET_SERD;
      KW_SET_SZ;
      KW_SET_S_AR;
      KW_SET_T_S;
      KW_SET_U_F;
      KW_SET_S_O;
      KW_SET_WAIT;
      KW_SET_W_S;
      KW_SRV_SOCK;
      KW_SHA1_DIG;
      KW_SHORT;
      KW_SHOW_ITB;
      KW_SIDE_L;
      KW_SIDE_L_H;
      KW_SIGNATUR;
      KW_SILENT;
      KW_SIMPLE;
      KW_SINGLE;
      KW_SIZE;
      KW_SIZE_C;
      KW_SIZE_P;
      KW_SLIDER;
      KW_SMAL_ICO;
      KW_SMAL_TTL;
      KW_SMALLINT;
      KW_SRC_PROC;
      KW_SOAP_HDR;
      KW_SOAP_HER;
      KW_SOAP_F;
      KW_SOAP_F_A;
      KW_SOAP_F_C;
      KW_SOAP_F_D;
      KW_SOAP_F_S;
      KW_SOCKET;
      KW_SORT;
      KW_SORT_ASC;
      KW_SORT_NUM;
      KW_SOURCE;
      KW_SQL;
      KW_SQRT;
      KW_SSL_SRVN;
      KW_STANDALN;
      KW_START;
      KW_START_BS;
      KW_START_DC;
      KW_START_EL;
      KW_START_MV;
      KW_START_RS;
      KW_START_RR;
      KW_START_SC;
      KW_STARTING;
      KW_STAT_DET;
      KW_STATIC;
      KW_STATUS_A;
      KW_STAT_A_F;
      KW_STDCALL;
      KW_STOP;
      KW_STOP_AFT;
      KW_STOP_D;
      KW_STOP_PRS;
      KW_STOPPED;
      KW_STORPROC;
      KW_STUP_PAR;
      KW_ST_2_FIT;
      KW_STR_VAL;
      KW_STR_XREF;
      KW_STRICT;
      KW_STRING;
      KW_SUB_AVG;
      KW_SUB_CNT;
      KW_SUB_MAX;
      KW_SUB_MENU;
      KW_SUB_M_H;
      KW_SUB_MIN;
      KW_SUB_TOT;
      KW_SUBSTIT;
      KW_SUBSTR;
      KW_SUBSCRIB;
      KW_SUBTYPE;
      KW_SUM;
      KW_SUP_NS_P;
      KW_SUP_PROC;
      KW_SUP_WARN;
      KW_SYM_EN_A;
      KW_SYM_EN_I;
      KW_SYM_EN_K;
      KW_SYM_SUPP;
      KW_SYNCHRON;
      KW_SYS_A_B;
      KW_SYS_HELP;
      KW_SYS_ID;
      KW_TAB;
      KW_TAB_CRCL;
      KW_TAB_HAND;
      KW_TAB_LIST;
      KW_TAB_POS;
      KW_TAB_SCAN;
      KW_TAB_STOP;
      KW_TAG;
      KW_TARGET;
      KW_TAR_PROC;
      KW_TEMP_DIR;
      KW_TEMP_TAB;
      KW_TRC_FILT;
      KW_TRAC_CHG;
      KW_TT_PREP;
      KW_TERMINAT;
      KW_TXT_SEG;
      KW_TXT_SEL;
      KW_3D;
      KW_THROUGH;
      KW_THROW;
      KW_TIC_MARK;
      KW_TIMEZONE;
      KW_TIME_SRC;
      KW_TITL_BGC;
      KW_TITL_DC;
      KW_TITL_FGC;
      KW_TITL_FON;
      KW_TODAY;
      KW_TOGGL_BX;
      KW_TOOLTIP;
      KW_TOOLTIPS;
      KW_TOP;
      KW_TOP_COL;
      KW_TOP_NAVQ;
      KW_TOPIC;
      KW_TOTAL;
      KW_TRAILING;
      KW_TRAN_I_P;
      KW_TRAN_MOD;
      KW_TRANSPAR;
      KW_TRUNC;
      KW_TTCP;
      KW_TYPE;
      KW_TYPE_OF;
      KW_UNBOX;
      KW_UNBUF;
      KW_UNIQ_ID;
      KW_UNIQ_MAT;
      KW_UNIX_END;
      KW_UNL_HID;
      KW_UNLOAD;
      KW_UNS_BYTE;
      KW_UNS_SHRT;
      KW_UNS_LONG;
      KW_UNSUBSCR;
      KW_UPD_ATTR;
      KW_URL;
      KW_URL_DECO;
      KW_URL_ENCO;
      KW_URL_FLD;
      KW_URL_FLDL;
      KW_URL_FMT;
      KW_URL_PW;
      KW_URL_UID;
      KW_USE;
      KW_USE_DCT;
      KW_USE_FIL;
      KW_USE_REV;
      KW_USE_TXT;
      KW_USE_UND;
      KW_USE_WIDP;
      KW_USR_DATA;
      KW_USER_ID;
      KW_UTC_OFF;
      KW_V6DISP;
      KW_VAL_ENAB;
      KW_VAL_EXPR;
      KW_VAL_MSG;
      KW_VAL_SEAL;
      KW_VAL_XML;
      KW_VALIDATE;
      KW_VAL_EVT;
      KW_VAL_HND;
      KW_VAL_OBJ;
      KW_VAR;
      KW_VERBOSE;
      KW_VERSION;
      KW_VERT;
      KW_VIEW_FCR;
      KW_VIRT_HC;
      KW_VIRT_HP;
      KW_VIRT_WC;
      KW_VIRT_WP;
      KW_VISIBLE;
      KW_VOID;
      KW_WARN;
      KW_WEB;
      KW_WEB_CTX;
      KW_WEB_OUT;
      KW_WEEK;
      KW_WHERE_ST;
      KW_WIDGET;
      KW_WID_ENT;
      KW_WID_HAND;
      KW_WID_ID;
      KW_WID_LEAV;
      KW_WID_POOL;
      KW_WIDTH;
      KW_WIDTH_C;
      KW_WIDTH_P;
      KW_WIN_CLOS;
      KW_WIN_NAME;
      KW_WIN_RESI;
      KW_WIN_REST;
      KW_WIN_STAT;
      KW_WIN_SYS;
      KW_WORD_IDX;
      KW_WORD_WRP;
      KW_WA_H_P;
      KW_WA_W_P;
      KW_WA_X;
      KW_WA_Y;
      KW_WR_CDATA;
      KW_WR_CHARS;
      KW_WR_CMNT;
      KW_WR_DATA;
      KW_WR_D_ELM;
      KW_WR_E_ELM;
      KW_WR_ENT_R;
      KW_WR_EXDTD;
      KW_WR_FRAGM;
      KW_WR_JSON;
      KW_WR_MSG;
      KW_WR_PRINS;
      KW_WR_STAT;
      KW_WR_XML;
      KW_WR_XMLSC;
      KW_X;
      KW_XML_DTYP;
      KW_XML_NNAM;
      KW_XML_NTYP;
      KW_XML_SCHP;
      KW_XML_SNSP;
      KW_X_DOC;
      KW_X_NODE;
      KW_X_OF;
      KW_YEAR;
      KW_YEAR_OFF;
      KW_YES_NO;
      KW_YES_NO_C;
      KW_Y;
      KW_Y_OF;
      BEGIN_FWD_EXTENSION;
         KW_TIMER;
         KW_CALLBACK;
         KW_ENABLED;
         KW_CEASE;
         CREATE_TIMER;
      END_FWD_EXTENSION;
   END_UNRESERVED;
   BEGIN_CALLGRAPH;
      FILE_RESOURCE;
      AST_NODE;
      IPC_RESOURCE;
      NATIVE_PROGRAM_RESOURCE;
      UNKNOWN_RESOURCE;
      INCLUDES;
      RUN_FILENAME;
      RUN_FILENAME_ON_SERVER;
      RUN_INT_PROC;
      RUN_INT_PROC_ON_SERVER;
      RUN_VALUE;
      RUN_VALUE_ON_SERVER;
      RUN_PORT_TYPE_ON_SERVER;
      RUN_PORT_TYPE_VALUE_ON_SERVER;
      RUN_LIBRARY_REF;
      RUN_LIBRARY_REF_ON_SERVER;
      FUNCTION_CALL;
      AMBIGUOUS;
      MISSING;
      PROCEDURE_FILE;
      INCLUDE_FILE;
      SCHEMA_FILE;
      EXTERNAL_PROCEDURE;
      INTERNAL_PROCEDURE;
      LIBRARY_PROCEDURE;
      PORT_TYPE;
      SHARED_LIBRARY;
      NATIVE_API;
      NATIVE_PROCESS;
      CHILD_PROCESS;
      NETWORK_CONNECTION;
      COM_OBJECT;
      DDE_SERVER;
      OCX_CONTROL;
      DEFINE_PROPERTY_GET;
      DEFINE_PROPERTY_SET;
      MORE;
   END_CALLGRAPH;
}

// helper members and methods
{
   /** Name of the specialized AST class to use when parsing. */
   private static final String AST_CLASS = ProgressAst.class.getName();

   /** Provides an efficient reverse lookup of names to token types. */
   private static Map tokenLookup = null;
   
   /** Debug mode for the parser. */
   private boolean debug = false;
   
   /** Tracks the nesting level of procedures and functions. */
   private int nestLevel = 0;
   
   /** Tracks the number of statements within a constructor. */
   private int constStmts = -1;
   
   /**
    * Contains all symbol resolution dictionaries and provides the 
    * maintenance/lookup methods.
    */
   private SymbolResolver sym = null;

   /**
    * Sets the bias in {@link #lvalue} toward (if <code>true</code>) or
    * away from (if <code>false</code>) widgets.  Defaults to 
    * <code>false</code>.
    */
   private boolean preferWidgets = false;
   
   /** Class name to use in {@link #lvalue} for data member lookups. */
   private String currentClassName = null;
   
   /** Flag to use in {@link #lvalue} for static-only data member lookups. */
   private boolean currentStaticFlag = false;
   
   /**
    * Modify {@link #lvalue} processing when called from inside the
    * {@link #primary_expr} rule.
    */
   // private boolean inPrimaryExpr = false;
   
   /**
    * Modify {@link #lvalue} field annotations and {@link #subscript}
    * processing when inside a <code>LIKE</code> clause.
    */
   private boolean inLikeClause = false;
   
   /** Modify {@link #expr} process when inside embedded SQL statements. */
   private boolean embeddedSql = false;
   
   /** Tracks the number of times an SQL UNION statement is nested. */
   private int sqlNestLevel = 0;
   
   /** Tracks the number of times an SQL SELECT statement is nested. */
   private int selectNestLevel = 0;
   
   /** <code>true</code> if processing an SQL UNION statement. */
   private boolean inUnion = false;
   
   /**
    * The name of the database field being currently processed for possible
    * validation.
    */
   private String validateFieldName = null;
   
   /** <code>true</code> if a validate expression is being processed. */
   private boolean validation = false;
   
   /**
    * Set to <code>false</code> when the parser is currently in an internal 
    * procedure, function or trigger, otherwise is set <code>true</code>.
    */
   private boolean bufferScope = true;
   
   /** Prefer temp-tables where the name conflicts with a table. */
   private boolean forceTemp = false;
   
   /** <code>true</code> if the parser was entered via external_proc. */
   private boolean topLevelEntry = false;
   
   /** Greater than 0 if the parser was entered via pre_scan_class. */
   private int preScanPass = 0;
   
   /** Tracks the level of recursion in the {@link #expr} rule. */
   private int exprLvl = 0;
   
   /** <code>true</code> if parsing a WebSpeed program. */
   private boolean webspeed = false;
   
   /** <code>true</code> to consume errors and automatically restart parsing. */
   private boolean consumeErr = true;
   
   /**
    * Determines if the parser was entered for preprocessor expression
    * evaluation.
    */
   private boolean evaluatingExpression = false;
   
   /** Flag to allow {@link #lvalue} to match a <code>SYMBOL</code>. */
   private boolean allowSymbolMatch = false;

   /** Allows matching certain unquoted text as strings. */
   private boolean allowStringMatch = false;

   /** Filter implementation to allow access to hidden tokens. */
   private TokenStreamHiddenTokenFilter hidden = null;
   
   /** Allow matching an unquoted method name in special cases. */
   private boolean insideClassEvent = false;
   
   /** Flag indicating we are processing the class body. */
   private boolean insideClass = false;
   
   /** <code>true</code> if we are parsing a built-in class/interface. */
   private boolean builtInCls = false;
   
   /** <code>true</code> if we are parsing a .NET class/interface. */
   private boolean dotNetCls = false;
   
   /** <code>true</code> if we are parsing inside a static method. */
   private boolean inStaticCtxt = false;
   
   /** The number of errors encountered during parsing */
   private int errorCount = 0;
   
   /**
    * Constructs a parser using a lexer and a user-defined symbol resolver.
    * This method ensures that the symbol resolver's function and variable
    * dictionaries are preloaded with the built-in symbols of the Progress
    * language.  This preloading must occur before parsing can begin.
    *
    * @param    lexer
    *           An instance of a <code>TokenStream</code> object.
    * @param    sr
    *           The user-defined (and optionally preloaded) symbol resolver.
    */
   public ProgressParser(TokenStream lexer, SymbolResolver sr)
   {
      // init our parent class
      this(lexer, 3);
      
      // shift into filter mode
      hidden = new TokenStreamHiddenTokenFilter(lexer);
      setTokenBuffer(new TokenBuffer(hidden));
      
      // setup the lexer
      ((ProgressLexer)lexer).activateHiddenProcessing();
      setupHiddenFilter();
      
      // setup our resolver
      sym = sr;
      initializeProcedureDictionary();
      initializeStreamDictionary();
      initializeFunctionDictionary();
      initializeVariableDictionary();
      initializeAttributesDictionary();
      
      honorKeywordIgnoreList(((ProgressLexer)lexer).getIgnoredKeywords());
      
      // set the AST class
      setASTNodeClass(AST_CLASS);      
   }
   
   /**
    * Writes error data to <code>System.err</code>, including a full stack
    * trace.
    *
    * @param    re
    *           The error on which to report.
    */
   public void reportError(RecognitionException re)
   {
      // keep the total error count
      errorCount++;

      if (consumeErr)
      {
         // print the error info, consume the exception then restart parsing
         re.printStackTrace(System.err);
      }
      else
      {
         // don't consume the error and don't restart parsing
         throw new RuntimeException(re);
      }
   }
   
   /**
    * Helper to package up an error message and the no-viable-alternative location details
    * into a single exception instance.
    *
    * @param    msg
    *           Error message for the containing <code>RuntimeException</code>
    * @param    node
    *           AST node which provides context for the no-viable-alternative location details.
    *
    * @return   The exception.
    */
   public RuntimeException genExc(String msg, Aast node)
   {
      AnnotatedAst ast = (AnnotatedAst) node;
      return new RuntimeException(msg, new NoViableAltException(ast));
   }
    
   /**
    * Helper to package up an error message and the no-viable-alternative location details
    * into a single exception instance.
    *
    * @param    msg
    *           Error message for the containing <code>RuntimeException</code>
    * @param    node
    *           Token node which provides context for the no-viable-alternative location details.
    *
    * @return   The exception.
    */
   public RuntimeException genExc(String msg, Token node)
   {
      return new RuntimeException(msg, new NoViableAltException(node, getFilename()));
   }
    
   /**
    * Tests if debug is turned on.
    */
   public boolean isDebug()
   {
      return debug;
   }
   
   /**
    * Sets the status of the debug mode member.
    *
    * @param    debug
    *           <code>true</code> to enable debug output.
    */
   public void setDebug(boolean debug)
   {
      this.debug = debug;
   }
   
   /**
    * Tests if the consume error flag is turned on.
    */
   public boolean isConsumeError()
   {
      return consumeErr;
   }
   
   /**
    * Sets the status of the consume error flag.
    *
    * @param    consume
    *           <code>true</code> to enable debug output.
    */
   public void setConsumeError(boolean consume)
   {
      this.consumeErr = consume;
   }
   
   /**
    * Gets the status of the expression evaluation member.
    */
   public boolean isEvaluatingExpression()
   {
      return evaluatingExpression;
   }
   
   /**
    * Sets the status of the expression evaluation member.
    *
    * @param    eval
    *           <code>true</code> if the current code is used to evaluate
    *           an expression.
    */
   public void setEvaluatingExpression(boolean eval)
   {
      evaluatingExpression = eval;
   }
   
   /**
    * Tests if this is parsing a WebSpeed program.
    */
   public boolean isWebSpeed()
   {
      return webspeed;
   }
   
   /**
    * Sets the status of the WebSpeed parsing flag.
    *
    * @param    webspeed
    *           <code>true</code> if the current program is a WebSpeed
    *           application.
    */
   public void setWebSpeed(boolean webspeed)
   {
      this.webspeed = webspeed;
   }
   
   /**
    * Tests if this is parsing a built-in 4GL class/interface.
    */
   public boolean isBuiltInCls()
   {
      return builtInCls;
   }
   
   /**
    * Sets the status of the built-in 4GL parsing flag.
    *
    * @param    builtInCls
    *           <code>true</code> if the current program is a built-in 4GL
    *           class/interface.
    */
   public void setBuiltInCls(boolean builtInCls)
   {
      this.builtInCls = builtInCls;
   }
   
   /**
    * Tests if this is parsing a .NET class/interface.
    */
   public boolean isDotNetCls()
   {
      return dotNetCls;
   }
   
   /**
    * Sets the status of the .NET parsing flag.
    *
    * @param    dotNetCls
    *           <code>true</code> if the current program is a .NET
    *           class/interface.
    */
   public void setDotNetCls(boolean dotNetCls)
   {
      this.dotNetCls = dotNetCls;
   }
   
   /**
    * Returns the number of errors encountered during parsing.
    *
    * @return  See above.
    */
   public int getErrorCount()
   {
      return errorCount;
   }
   
   /**
    * Translates a text representation of a parser token type into the
    * actual integer value.
    *
    * @param    tokenName
    *           The text representation of the parser token type.
    *
    * @return   A valid parser-defined integer token type value or -1 if
    *           no match was found.
    */
   public static synchronized int lookupTokenType(String tokenName)
   {
      // one-time init when we are first called
      if (tokenLookup == null)
      {
         int len = ProgressParser._tokenNames.length;
         
         tokenLookup = new HashMap();
         
         // load the map with our lowercased symbolic name as the key and
         // the wrappered int token type as the value
         for ( int i = 0; i < len; i++ )
         {
            tokenLookup.put(ProgressParser._tokenNames[i].toLowerCase(),
                            new Integer(i));
         }
      }
      
      Integer result = (Integer) tokenLookup.get(tokenName.toLowerCase());
      
      if (result == null)
      {
         return -1;
      }
      
      return result.intValue();
   }
   
   /**
    * Translates an integer value of a parser token type into the associated 
    * human readable text representation.
    *
    * @param    type
    *           The integer token type.
    *
    * @return   A valid parser-defined symbolic name or <code>null</code> if
    *           no match was found.
    */
   public static String lookupTokenName(int type)
   {
      if (type < 0 || type >= ProgressParser._tokenNames.length)
      {
         return null;
      }
      
      return ProgressParser._tokenNames[type];
   }
   
   /**
    * Create a schema validation sub-tree if the given field reference has schema-level assign
    * validation defined, and attach it to the given anchor point. If no validation expression
    * exists in the schema for the given field, no change is made.
    *
    * @param   anchor
    *          AST to which a KW_VALIDATE and/or KW_HELP sub-tree will be attached as a child.
    * @param   fieldRef
    *          The AST for a business logic reference to a database field. This AST is expected
    *          to have <code>schemaname</code> and <code>bufname</code> annotations.
    * @param   sym
    *          Symbol resolver which will be used to parse the schema validation expression, if
    *          any, for the given field.
    *
    * @return  Schema validation node that was attached, if any, or <code>null</code> if nothing
    *          was attached.
    */
   private List<Aast> attachAssignSchemaValidation(Aast anchor, Aast fieldRef, SymbolResolver sym)
   {
      String schemaName = (String) fieldRef.getAnnotation("schemaname");
      String bufName = (String) fieldRef.getAnnotation("bufname");
      Aast schemaRef = sym.lookupFieldAst(fieldRef.getText());
      
      validateFieldName = schemaName;
      List<Aast> ret = attachAssignSchemaValidation(anchor, schemaRef, schemaName, bufName, sym);
      validateFieldName = null;
      return ret;
   }
   
   /**
    * Create a schema validation sub-tree if the given field reference has schema-level assign
    * validation defined, and attach it to the given anchor point. If no validation expression
    * exists in the schema for the given field, no change is made.
    *
    * @param   anchor
    *          AST to which a KW_VALIDATE and/or KW_HELP sub-tree will be attached as a child.
    * @param   schemaRef
    *          The AST for a field provided by the schema dictionary, from which the validation
    *          expression and message are retrieved.
    * @param   schemaName
    *          Fully qualified schema name of the field for which validation is being processed.
    * @param   bufName
    *          Qualified name of the buffer referencing the table which contains the field for
    *          which validation is being processed.
    * @param   sym
    *          Symbol resolver which will be used to parse the schema validation expression, if
    *          any, for the given field.
    *
    * @return  Schema validation node that was attached, if any, or <code>null</code> if nothing
    *          was attached.
    */
   private List<Aast> attachAssignSchemaValidation(Aast anchor,
                                                   Aast schemaRef,
                                                   String schemaName,
                                                   String bufName,
                                                   SymbolResolver sym)
   {
      Aast valExp = schemaRef.getImmediateChild(VALEXP, null);
      Aast valMsg = schemaRef.getImmediateChild(VALMSG, null);
      String help = null;
      
      if (schemaRef.isAnnotation("help"))
      {
         help = (String) schemaRef.getAnnotation("help");
      }
      else if (schemaRef.downPath("KW_HELP"))
      {
         Aast helpRef = schemaRef.getImmediateChild(KW_HELP, null);
         helpRef = (Aast) helpRef.getFirstChild();
         help = helpRef.getText();
      }
      
      return attachSchemaValidation(anchor, valExp, valMsg, schemaName, bufName, sym, help);
   }
   
   /**
    * Given a record reference, find all fields in the corresponding table for which schema-level
    * assign validation is defined; create a schema validation sub-tree for each, and attach
    * these sub-trees to the given anchor point. If no validation expressions exist for any of
    * the fields in the table, no change is made.
    *
    * @param   anchor
    *          AST to which zero or more KW_VALIDATE and/or KW_HELP sub-trees will be attached as
    *          children.
    * @param   recordRef
    *          The AST for a business logic reference to a database record. This AST is expected
    *          to have <code>schemaname</code> and <code>bufname</code> annotations.
    * @param   sym
    *          Symbol resolver which will be used to parse the schema validation expressions, if
    *          any, for the fields in the given table.
    * @param   createFormatPhrase
    *          <code>true</code> if format phrase node should be created to attach KW_VALIDATE
    *          nodes. <code>false</code> if format phrase node should not be created, e.g. for
    *          browse columns.
    */
   private void attachAllAssignSchemaValidation(Aast anchor,
                                                Aast recordRef,
                                                SymbolResolver sym,
                                                boolean createFormatPhrase)
   {
      String tableSchemaName = (String) recordRef.getAnnotation("schemaname");
      String bufName = (String) recordRef.getAnnotation("bufname");
      
      Iterator<Aast> fieldIter = sym.fields(tableSchemaName);
      while (fieldIter.hasNext())
      {
         Aast schemaRef = fieldIter.next();

         Aast fpRef = null;
         if (createFormatPhrase)
         {
            // build intermediate format phrase node to which to attach KW_VALIDATE nodes, if any
            fpRef = new ProgressAst();
            fpRef.setType(FORMAT_PHRASE);
         }
         else
         {
            fpRef = anchor;
         }

         String fieldSchemaName = tableSchemaName + "." + schemaRef.getText().toLowerCase();

         attachAssignSchemaValidation(fpRef, schemaRef, fieldSchemaName, bufName, sym);

         // if any sub-trees were attached to format phrase node, attach it to anchor, just after
         // TABLE or BUFFER node
         if (createFormatPhrase && (fpRef.getFirstChild() != null))
         {
            anchor.addChild(fpRef);
         }
      }
   }
   
   /**
    * Create a schema validation sub-tree if the given record reference has schema-level delete
    * validation defined, and attach it to the given anchor point. If no validation expression
    * exists in the schema for the given table, no change is made.
    *
    * @param   anchor
    *          AST to which a KW_VALIDATE and/or HW_HELP sub-tree will be attached as a child.
    * @param   recordRef
    *          The AST for a business logic reference to a database record. This AST is expected
    *          to have <code>schemaname</code> and <code>bufname</code> annotations.
    * @param   sym
    *          Symbol resolver which will be used to parse the schema validation expression, if
    *          any, for the given field.
    */
   private void attachDeleteSchemaValidation(Aast anchor, Aast recordRef, SymbolResolver sym)
   {
      Aast schemaRef = sym.lookupTableAst(recordRef.getText());
      
      Aast props = schemaRef.getImmediateChild(PROPERTIES, null);
      if (props == null)
      {
         // some temp-/work-tables will not have a PROPERTIES node
         return;
      }
      
      Aast valExp = props.getImmediateChild(VALEXP, null);
      if (valExp == null)
      {
         return;
      }
      
      Aast valMsg = props.getImmediateChild(VALMSG, null);
      String schemaName = (String) recordRef.getAnnotation("schemaname");
      String bufName = (String) recordRef.getAnnotation("bufname");
      
      attachSchemaValidation(anchor, valExp, valMsg, schemaName, bufName, sym, null);
   }
   
   /**
    * Create a schema validation sub-tree for the field or record indicated by the given schema
    * and buffer names, and attach it to the given anchor point. If no validation expression
    * exists in the schema for the given field or record, no change is made.
    *
    * @param   anchor
    *          AST to which a KW_VALIDATE and/or KW_HELP sub-tree will be attached as a child.
    * @param   valExp
    *          Validation expression AST.
    * @param   valMsg
    *          Validation message AST.
    * @param   schemaName
    *          Fully qualified schema name of database field for which validation is needed.
    * @param   bufName
    *          Name of buffer referenced in business logic, which contains field.
    * @param   sym
    *          Symbol resolver which will be used to parse the schema validation expression, if
    *          any, for the given field or record.
    * @param   help
    *          The schema-level help text.
    *
    * @return  Schema validation node that was attached, if any, or <code>null</code> if nothing
    *          was attached.
    */
   private List<Aast> attachSchemaValidation(Aast anchor,
                                             Aast valExp,
                                             Aast valMsg,
                                             String schemaName,
                                             String bufName,
                                             SymbolResolver sym,
                                             String help)
   {
      // TODO: the anchor in is not in all cases a FORMAT_PHRASE, when BROWSE DISPLAY/ENABLE is
      // involved; if this gets changed, then TRPL conversion rules need to be adjusted, to work 
      // with the new AST structure.
      boolean isFP = anchor.getType() == FORMAT_PHRASE;
   
      List<Aast> res = new ArrayList<>();
      
      // do not attach if there is already a KW_VALIDATE in this FORMAT_PHRASE
      if (valExp != null && (!isFP || !anchor.downPath("KW_VALIDATE")))
      {
         // expression text is at VALEXP->STRING:text
         String expText = valExp.getFirstChild().getText();
         
         String msgText = "\"\"";
         if (valMsg != null)
         {
            // message text is at VALMSG->EXPRESSION->STRING:text
            msgText = valMsg.getFirstChild().getFirstChild().getText();
         }
         
         try
         {
            // preprocess text, in case there are any includes that need to be resolved
            String[] paths = AstGenerator.getPropath(null, null);
            Options options = AstGenerator.buildOptions(paths, null, null);
            StringWriter out = new StringWriter();
            FileScope fs = new FileScope("<inline>", new StringReader(expText));
            new Preprocessor(fs, out, new PrintWriter(System.err), null, options);
            expText = out.toString();
            
            if (expText.length() > 0)
            {
               // parse expText using a new Progress parser, but with the provided symbol resolver
               StringReader reader = new StringReader(expText);
               ProgressLexer lexer = new ProgressLexer(reader, sym);
               ProgressParser parser = new ProgressParser(lexer, sym);
               
               // make the new parser aware of the field being validated. In case of CAN-FIND,
               // when resolving an unqualified field name, the [validateFieldName] takes
               // precedence during parsing of this expression.
               parser.validateFieldName = this.validateFieldName;
               parser.validation = true;
               parser.setConsumeError(false);
               parser.expr();
               
               // add KW_VALIDATE child to anchor node
               Aast valParent = new ProgressAst();
               valParent.setType(KW_VALIDATE);
               valParent.setText("");
               valParent.putAnnotation("schema_validation", true);
               valParent.putAnnotation("schemaname", schemaName);
               valParent.putAnnotation("bufname", bufName);
               anchor.addChild(valParent);
               
               // add first EXPRESSION child to KW_VALIDATE node (for condition)
               Aast exp = new ProgressAst();
               exp.setType(EXPRESSION);
               exp.setText("");
               valParent.addChild(exp);
               
               // add validation expression subtree to condition EXPRESSION node
               Aast ve = (Aast) parser.getAST();
               exp.addChild(ve);
               
               // add second EXPRESSION child to KW_VALIDATE node (for message)
               exp = new ProgressAst();
               exp.setType(EXPRESSION);
               valParent.addChild(exp);
               
               // add STRING child to message EXPRESSION node
               Aast str = new ProgressAst();
               str.setType(STRING);
               str.setText(msgText);
               exp.addChild(str);
               
               res.add(valParent);
            }
         }
         catch (Exception exc)
         {
            System.err.println("WARNING: ignoring invalid schema validation expression for field '" +
                               schemaName +
                               "':");
            System.err.println(expText);
            System.err.println(exc.getMessage());
         }
      }

      // do not attach if there is already a KW_HELP in this FORMAT_PHRASE; skip COLUMN_REF for
      // BROWSE enable option, as HELP will be pulled automatically by conversion rules, if needed
      if (help != null                   &&
          anchor.getType() != COLUMN_REF &&
          (!isFP || !anchor.downPath("KW_HELP")))
      {
         Aast helpAst = new ProgressAst();
         helpAst.setType(KW_HELP);
         helpAst.setText("help");
         helpAst.putAnnotation("schema_help", true);
         anchor.addChild(helpAst);
         
         Aast str = new ProgressAst();
         str.setType(STRING);
         str.setText(help);
         helpAst.addChild(str);
         
         res.add(helpAst);
      }

      return res;
   }
   
   /**
    * Matches and consumes a token that must be within a specific range of
    * token types.  This code is based on other matching methods provided
    * by ANTLR.  The ANTLR generator will properly create code to call
    * this method using the &quot;..&quot; range operator.  While the 
    * generated lexer does contain a <code>matchRange</code> method, the
    * parser does not!  This method was added manually to resolve this
    * condition.
    * <p>
    * Note that to make this useful, one must maintain a list of the
    * artificial token types (using the tokens { } section in the parser's
    * grammar). More importantly, all needed tokens to be matched in a range
    * need to be kept contiguous so that the range neither includes token
    * types it should not, nor excludes types it should not.  Since ANTLR
    * generates the token types in order found in that section, one trick
    * used is to add a special token to mark the begin and end of any such
    * range.  These can be used instead of the actual first and last tokens.
    * This makes it easier to edit without introducing an error.
    */
   public void matchRange( int t1, int t2 )
   throws MismatchedTokenException,
          TokenStreamException
   {
      if ( LA(1) < t1 || LA(1) > t2 )
      {
         throw new MismatchedTokenException(tokenNames,
                                            LT(1),
                                            t1,
                                            false,
                                            getFilename());
      }
      else
      {
         // mark token as consumed -- fetch next token deferred until LA/LT
         consume();
      }
   }
   
   /**
    * Remove any built-in functions or global variables that are defined with the given
    * keywords.
    *
    * @param    ignored
    *           The list to process.
    */
   private void honorKeywordIgnoreList(Keyword[] ignored)
   {
      if (ignored != null)
      {
         for (Keyword word : ignored)
         {
            sym.removeBuiltinFunction(word);
            sym.removeGlobalVariable(word);
         }
      }
   }
   
   /**
    * Configure the hidden filter to ensure that token types that are not
    * normally recognized by the parser, are hidden (but still accessible)
    * using the filter.
    */
   private void setupHiddenFilter()
   {
      hidden.hide(WS);
      hidden.hide(TILDE);
      hidden.hide(COMMENT);
      hidden.hide(BACKSLASH);
   }
   
   /**
    * Scan forward to determine if the current node is an inline variable 
    * definition.  For this to be the case, the current node must be a
    * symbol which is not already a variable or field.  In addition, it must
    * be followed by an <code>AS</code> or <code>LIKE</code> keyword.
    * <p>
    * <b>At this time, this method will not work 100% properly so DON'T use
    * it! WARNING: this code needs to know when to stop looking ahead and
    * this test is far too permissive! We really need to stop when the format
    * phrase is done. But to predict when that might occur is not easy!</b>
    *
    * @return   <code>true</code> if this is an inline variable definition.
    */
   private boolean isFormVarDefinition()
   throws TokenStreamException
   {
      Token candidate = LT(1);
      
      int    typ = candidate.getType(); 
      String txt = candidate.getText();
      
      if (typ == SYMBOL || (typ > BEGIN_UNRESERVED && typ < END_UNRESERVED))
      {
         if (sym.lookupVariable(txt) == -1 && !sym.isField(txt))
         {
            int idx = 2;
            
            while (true)
            {
               int next = LA(idx);
               
               // WARNING: this code needs to know when to stop looking ahead
               //          and this test is far too permissive! We really
               //          need to stop when the format phrase is done. But
               //          to predict when that might occur is not easy!
               if (next == KW_WITH || next == DOT)
               {
                  break;
               }
               
               if (next == KW_AS || next == KW_LIKE)
               {
                  return true;
               }
               
               idx++;
            }
         }
      }
      
      return false;
   }
   
   /**
    * Check the tokens starting at the given index to see if they match a malformed label (that
    * starts with a reserved keyword), a colon and then is followed by the beginning of an inner
    * block.  If found, rewrite the leading reserved keyword's token type to a SYMBOL so that
    * the prediction logic for a downstream <code>single_block</code> will be engaged. This
    * should only ever be called just before a <code>single_block</code> is entered.
    *
    * @param    idx
    *           The index of the token that would start the label (and it would be a reserved
    *           keyword).
    */
   private void checkAndFixLabelWithReservedKeywordPrefix(int idx)
   throws TokenStreamException,
          MismatchedTokenException
   {
      int rtype = LA(idx);
      
      if (rtype > BEGIN_RESERVED && rtype < END_RESERVED)
      {
         Predicate<CommonToken> follow = (CommonToken tok) ->
         {
            int type = tok.getType();
            
            return (type == KW_DO || type == KW_REPEAT || type == KW_FOR || type == KW_EDITING);
         };
         
         if (isMalformedLabel(true, idx, follow))
         {
            LT(idx).setType(SYMBOL);
         }
      }
   }
   
   /**
    * Manually looks ahead an 2 tokens to determine if the next tokens represent a properly
    * formed symbol.  Normal symbols must start with an alphabetic character or an underscore.
    * They must then be followed by a colon and then whitespace.  The problem is that it is
    * legal to use nearly all unreserved keywords as labels. This method matches these cases
    * except for the one case that is unreserved but is still excluded from being a label
    * (the FINALLY keyword).
    * <p>
    * A malformed label will NOT be matched by this rule (see {@link #isMalformedLabel}).
    *
    * @param    typ
    *           The token type of the first lookahead token. This is a performance optimization
    *           only since the same value could have been queried too.
    *
    * @return   <code>true</code> if the following tokens represent a match as a properly
    *           formed label.
    */
   private boolean isProperLabel(int typ)
   throws TokenStreamException
   {
      // must be a valid symbol or unreserved keyword (except finally is unreserved
      // but cannot be used as a label!)
      if (typ == SYMBOL || ((typ > BEGIN_UNRESERVED && typ < END_UNRESERVED) &&
                             typ != KW_FINALLY))
      {
         // possible label
         if (LA(2) == COLON && followedByWhitespace(LT(2)))
         {
            // OK, this is a properly formed label
            return true;
         }
      }
      
      return false;
   }
   
   /**
    * Manually looks ahead an arbitrary number of tokens to determine if the next tokens 
    * represent a malformed symbol.  Normal symbols must start with an alphabetic character
    * or an underscore.  Labels have been found to possibly start with a pretty much any kind
    * of text including @#$%^, hyphens, operators and other "second and later" symbol
    * characters. The lexer cannot know when to lex a symbol in such a manner so it must be
    * detected and put back together in the parser.  This method detects the malformed label
    * case.
    * <p>
    * A normal label will NOT be matched by this rule so any testing for normal labels must be
    * done outside of this rule. See {@link #isProperLabel}.
    *
    * @param    def
    *           <code>true</code> if this should test for a label definition (the malformed
    *           symbol must be followed by a colon). When <code>false</code>, no colon is
    *           expected BUT the resulting text must appear in the scoped label dictionary.
    * @param    start
    *           The index of the token at which to start checking.
    * @param    follow
    *           If non-null AND a malformed label has been detected, this predicate will be
    *           checked against the AST node that immediately follows the COLON.  This is only
    *           checked in label definition mode.
    *
    * @return   <code>true</code> if the following tokens represent a match as a malformed
    *           label.
    */
   private boolean isMalformedLabel(boolean def, int start, Predicate<CommonToken> follow)
   throws TokenStreamException,
          MismatchedTokenException
   {
      int[] end = new int[] { COLON, COMMA, LPARENS, RPARENS, EOF };
      int   num = start;
   
      CommonToken current  = (CommonToken) LT(num);
      CommonToken next     = (CommonToken) LT(++num);
      int         nextType = next.getType();   
      
      StringBuilder sb = new StringBuilder();
      
      while (!followedByWhitespace(current)       &&
             !matchesList(nextType, end)          &&
             !(nextType == DOT && followedByWhitespace(next)))
      {
         // save the text of the current token
         if (!def)
            sb.append(current.getText());
         
         // update the state for next iteration
         current  = next;
         next     = (CommonToken) LT(++num);
         nextType = next.getType();
      }
      
      // did more than 1 token get processed?
      if (num > (start + 1))
      {
         if (!def)
         {
            // put the last text into the buffer
            sb.append(current.getText());
            
            // possible label reference
            if (sym.lookupLabel(sb.toString()) != -1)
            {
               return true;
            }
         }
         else
         {
            // possible new label definition
            
            // we have a candidate label (text that could be a label) but it
            // must be followed by a COLON and whitespace to fully match
            if (next.getType() == COLON && followedByWhitespace(next))
            {
               if (follow != null)
               {
                  return follow.test((CommonToken) LT(num + 1));
               }
            
               return true;
            }
         }
      }
      
      return false;
   }
   
   /**
    * The P4GL requires an additional space after keyword operators. This method checks if this
    * exists and if it is missing a parsing exception is thrown. Note that the space is not
    * required BEFORE the keyword operator and not required at either ends in case of symbol
    * operator.
    * <p>
    * Known affected keyword operators are: {@code kw_eq}, {@code kw_ne}, {@code kw_lt},
    * {@code kw_gt}, {@code kw_gte}, {@code kw_lte}, {@code kw_matches}, {@code kw_begins},
    * {@code kw_contains}, {@code kw_or} and {@code kw_and}.
    *
    * @param   binOperator
    *          Aast node of the operator to be checked.
    *
    * @throws  MismatchedTokenException
    *          if no space is encounter after this token.
    * @throws  MismatchedTokenException
    *          if fails to extract the next token.
    */
   private void checkExtraSpace(Aast binOperator)
   throws MismatchedTokenException,
          TokenStreamException
   {
      if (followedByWhitespace(LT(0)))
      {
         // this is how P4GL expects
         return;
      }
      
      // second chance: if the second operand is parenthesized the P4GL validates it
      if (LT(1).getType() == LPARENS) 
      {
         return;
      }
      
      MismatchedTokenException exc =
         new MismatchedTokenException(new String[] { "Unknown Field or Variable name" },
                                      binOperator, -1, -1, true);
      exc.token = LT(0); 
      throw exc;
   }
   
   /**
    * Manually looks ahead an arbitrary number of tokens to determine if the 
    * next tokens represent a malformed symbol as a function name followed by
    * an <code>LPARENS</code>.  Normal function names must start
    * with an alphabetic character or an underscore.  Other names have been
    * found which can possibly start with a number, hyphen and other
    * "second and later" symbol characters. The lexer cannot know when to
    * lex a symbol in such a manner so it must be detected and put back
    * together in the parser.
    * <p>
    * This does not match any tokens, it just reads ahead to check.
    *
    * @return   <code>true</code> if and only if the following tokens represent
    *           a match as a malformed function call name.
    */
   private boolean isMalformedSymbolFuncCall()
   throws TokenStreamException
   {
      int   idx   = 1;
      Token next  = LT(1);
      int   ttype = next.getType();
      
      StringBuilder sb = new StringBuilder();
      
      // a malformed symbol must start with one of these tokens
      if (ttype != NUM_LITERAL          &&
          ttype != DATE_LITERAL         &&
          ttype != DATETIME_LITERAL     &&
          ttype != DATETIME_TZ_LITERAL  &&
          ttype != MINUS                &&
          ttype != UNKNOWN_TOKEN)
      {
         return false;
      }
      
      while (true)
      {
         // any intervening whitespace before we hit text means this is not a
         // match
         if (followedByWhitespace(next))
         {
            break;
         }
         
         // store the text of the current token
         sb.append(next.getText());
         
         // get the next token (which we already know is contiguous)
         idx++;
         next  = LT(idx);
         ttype = next.getType();
         
         // when the next token falls in this range, it means there is some
         // alphabetic text following the other trash (this will cause the
         // symbol rule to be matched in the lexer, which can produce one of
         // these types), which tells us that this is a valid symbol
         if ((ttype > BEGIN_RESERVED   && ttype < END_RESERVED)   ||
             (ttype > BEGIN_UNRESERVED && ttype < END_UNRESERVED) ||
             ttype == SYMBOL)
         {
            // this should be the last bit of the function name
            sb.append(next.getText());
         
            // this is a malformed symbol, BUT is it a function call?
         
            idx++;
            next = LT(idx);
            
            boolean func = (next.getType() == LPARENS) && 
                           (sym.lookupFunction(sb.toString()) != -1);
            
            // if this is followed by an LPARENS (regardless of whitespace in
            // between) and the name can be looked up in our function dictionary
            // then this is a function call
            return func ? true : false;
         }
         
         // iterate as long as there is valid preceding crap to read through
         if (ttype == NUM_LITERAL          ||
             ttype == DATE_LITERAL         ||
             ttype == DATETIME_LITERAL     ||
             ttype == DATETIME_TZ_LITERAL  ||
             ttype == MINUS                ||
             ttype == UNKNOWN_TOKEN)
         {
            continue;
         }
         
         // we must not have matched anything valid, this isn't a malformed
         // symbol
         break;
      }
      
      return false;
   }
   
   /**
    * Detect if the given token type is a symbol or unreserved keyword.
    *
    * @param    ttype
    *           The token type to check;
    *
    * @return   <code>true</code> if the type is some kind of symbol.
    */
   private boolean isSomeSymbol(int ttype)
   {
      return (ttype == SYMBOL || (ttype > BEGIN_UNRESERVED && ttype < END_UNRESERVED));
   }
   
   /**
    * Manually looks ahead an arbitrary number of tokens to determine if the 
    * next tokens represent a malformed symbol.  Normal symbols must start
    * with an alphabetic character or an underscore.  Other names have been
    * found which can possibly start with a number, hyphen and other
    * "second and later" symbol characters. The lexer cannot know when to
    * lex a symbol in such a manner so it must be detected and put back
    * together in the parser.
    * <p>
    * This does not match any tokens, it just reads ahead to check.
    *
    * @param    start
    *           Starting token to check.
    *
    * @return   <code>true</code> if and only if the following tokens represent
    *           a match as a malformed symbol.
    */
   private boolean isMalformedSymbol(int start)
   throws TokenStreamException
   {
      int   idx   = start;
      Token next  = LT(idx);
      int   ttype = next.getType();
      
      // a malformed symbol must start with one of these tokens
      if (ttype != NUM_LITERAL          &&
          ttype != DATE_LITERAL         &&
          ttype != DATETIME_LITERAL     &&
          ttype != DATETIME_TZ_LITERAL  &&
          ttype != MINUS                &&
          ttype != UNKNOWN_TOKEN)
      {
         return false;
      }
      
      while (true)
      {
         // any intervening whitespace before we hit text means this is not a
         // match
         if (followedByWhitespace(next))
         {
            break;
         }
         
         // get the next token (which we already know is contiguous)
         idx++;
         next  = LT(idx);
         ttype = next.getType();
         
         // when the next token falls in this range, it means there is some
         // alphabetic text following the other trash (this will cause the
         // symbol rule to be matched in the lexer, which can produce one of
         // these types), which tells us that this is a valid symbol
         if ((ttype > BEGIN_RESERVED   && ttype < END_RESERVED)   ||
             (ttype > BEGIN_UNRESERVED && ttype < END_UNRESERVED) ||
             ttype == SYMBOL)
         {
            // this is a malformed symbol
            return true;
         }
         
         // iterate as long as there is valid preceding crap to read through
         if (ttype == NUM_LITERAL          ||
             ttype == DATE_LITERAL         ||
             ttype == DATETIME_LITERAL     ||
             ttype == DATETIME_TZ_LITERAL  ||
             ttype == MINUS                ||
             ttype == UNKNOWN_TOKEN)
         {
            continue;
         }
         
         // we must not have matched anything valid, this isn't a malformed
         // symbol
         break;
      }
      
      return false;
   }
   
   /**
    * Manually looks ahead an arbitrary number of tokens, reassembling the text 
    * and type of a malformed symbol into a single token while dropping the
    * extraneous tokens.
    * <p>
    * This discards bogus intermedite tokens and fixes up the last token in the
    * stream. You MUST already know that the following tokens ARE a malformed
    * symbol by using {@link #isMalformedSymbolFuncCall}!
    *
    * @return   The text of the token that was reassembled.
    */
   private String reassembleMalformedSymbol()
   throws TokenStreamException,
          NoViableAltException,
          MismatchedTokenException
   {
      int         idx   = 1;
      CommonToken first = (CommonToken) LT(1);
      CommonToken next  = first;
      int         ttype = first.getType();
      int         line  = first.getLine();
      int         col   = first.getColumn();
      
      // a malformed symbol must start with one of these tokens
      if (ttype != NUM_LITERAL          &&
          ttype != DATE_LITERAL         &&
          ttype != DATETIME_LITERAL     &&
          ttype != DATETIME_TZ_LITERAL  &&
          ttype != MINUS                &&
          ttype != UNKNOWN_TOKEN)
      {
         throw new NoViableAltException(LT(1), getFilename());
      }
      
      StringBuilder sb = new StringBuilder();
      
      while (true)
      {
         // any intervening whitespace before we hit text means this is not a
         // match
         if (followedByWhitespace(next))
         {
            throw new NoViableAltException(LT(1), getFilename());
         }
         
         // add the text to our buffer
         sb.append(next.getText());
         
         // this is a token we are discarding, match it
         match(ttype);
         
         // get the next token
         next  = (CommonToken) LT(1);
         ttype = next.getType();
         
         // when the next token falls in this range, it means there is some
         // alphabetic text following the other trash (this will cause the
         // symbol rule to be matched in the lexer, which can produce one of
         // these types), which tells us that this is a valid symbol
         if ((ttype > BEGIN_RESERVED   && ttype < END_RESERVED)   ||
             (ttype > BEGIN_UNRESERVED && ttype < END_UNRESERVED) ||
             ttype == SYMBOL)
         {
            sb.append(next.getText());
            
            // done, save off results in the last token
            next.setText(sb.toString());
            next.setType(SYMBOL);
            next.setLine(line);
            next.setColumn(col);
            break;
         }
         
         // iterate as long as there is valid preceding crap to read through
         if (ttype != NUM_LITERAL          &&
             ttype != DATE_LITERAL         &&
             ttype != DATETIME_LITERAL     &&
             ttype != DATETIME_TZ_LITERAL  &&
             ttype != MINUS                &&
             ttype != UNKNOWN_TOKEN)
         {
            throw new NoViableAltException(LT(1), getFilename());
         }
      }
      
      return sb.toString();
   }
   
   /**
    * Core lookahead processing for the lvalue rule which checks for a match to a variable,
    * widget, class data member/property or database field.  It will NOT match qualifiers
    * frames, streams or assignment style statements.  This is useful for detecting if a
    * following token stream is an lvalue or not.  When token stream changes are allowed,
    * the text can be rewritten and tokens can even be matched.  DO NOT allow changes when
    * passing an index other than 1, as any matching done will not match on the tokens
    * expected.
    * <p>
    * This is split off from the lvalue rule itself so that other code can call this for
    * lookahead purposes (no changes allowed).
    *
    * @param    idx
    *           The token at which to start looking for an lvalue match. This is a 1-based
    *           index which is how ANTLR works.
    * @param    allowChange
    *           <code>true</code> to enable token text modification and matching.  This only
    *           occurs with certain validation and field naming quirks.
    *
    * @return   The token type if the tokens are an lvalue or -1 if no match was found.
    */
   private int resolveLvalueCoreType(int idx, boolean allowChange)
   throws TokenStreamException,
          MismatchedTokenException
   {
      String text = LT(idx).getText();
      
      // variables by the same name as a field, will hide that field 
      int newtype = sym.lookupVariable(text);
      
      // there are some cases where calling rules (possibly far up
      // the call stack) need to match a widget in preference to
      // a variable, in these cases a global state variable is
      // set and this will override any variable found with a
      // widget instead (if the widget exists)
      if (preferWidgets)
      {
         int widtype = sym.lookupWidget(text);
         
         if (widtype != -1)
            newtype = widtype;
      }
      else
      {
         // in normal processing, we might not prefer widgets but we
         // still need to check for a match
         if (newtype == -1)
         {
            // this must be a widget name
            newtype = sym.lookupWidget(text);
         }
      }
      
      // check if this is an object data member or property (may
      // be defined somewhere in the class hierarchy); calling rules
      // may set the class name based on context to control the
      // lookup here
      if (newtype == -1)
      {
         newtype = sym.lookupDataMember(currentClassName, text, currentStaticFlag);
      }
      
      // special processing for validation (if there is a match to the short name only, we
      // override with the full schema name to make it unambiguous)
      if (newtype == -1 && validation && validateFieldName != null)
      {
         int    sidx   = validateFieldName.lastIndexOf('.');
         String ambig  = validateFieldName.substring(sidx + 1).toLowerCase();
         String lcText = text.toLowerCase();
         
         // test in ignore-case and also handle abbreviations
         if (ambig.startsWith(lcText))
         {
            text = validateFieldName;
            
            // force the text to the full name
            if (allowChange)
            {
               LT(idx).setText(validateFieldName);
            }
         }
      }
      
      int idx2 = idx + 1;
      int idx3 = idx + 2;
      
      // handle a qualified database naming quirk in which there 
      // is whitespace between the record and the DOT and/or 
      // whitespace between the DOT and the field portion such that
      // Progress matches "record    .field" or "record.  field" or
      // "record  .  field" all as the single qualified field name
      // "record.field";
      if (newtype == -1 && isQualifiedFieldNameQuirk(LT(idx), LT(idx2), LT(idx3)))
      {                    
         StringBuilder sb = new StringBuilder();
         sb.append(text).append('.').append(LT(idx3).getText());
         
         String total = sb.toString();
         
         newtype = sym.lookupField(total);
         
         if (newtype != -1 && allowChange)
         {
            // we have the special case of a whitespace-infested 
            // qualified database name, put it "back together"
            
            CommonToken tok = (CommonToken) LT(idx);
            int line = tok.getLine();
            int colm = tok.getColumn();
            
            // manually match the next 2 tokens
            match(LA(idx));
            match(DOT);
            
            // force the text of the field piece of the token to
            // the reassembled text
            tok = (CommonToken) LT(idx);
            tok.setText(total);
            tok.setLine(line);
            tok.setColumn(colm);
         }
      }
      
      // lookup a field if the variable/widget reference wasn't found the call to isField()
      // is needed because otherwise the schema dictionary may report an ambiguous name even
      // though we don't care about such reports; this is done before the name quirk processing
      // so that a simple field reference is honored in preference
      if (newtype == -1 && sym.isField(text))
      {
         // this may be a field name
         newtype = sym.lookupField(text);
      }
      
      return newtype;
   }
   
   /**
    * Common code to add a scope (symbol dictionary and schema scopes) for a
    * trigger.  See <code>clearTriggerScope()</code>.
    * <p>
    * Triggers can be nested inside procedures, functions and other triggers
    * so we must save the value found here so that it can be restored later.
    * Otherwise we would force everything else to global scope while still in
    * a nested scope.
    *
    * @return   The previous state of the <code>bufferScope</code> variable
    *           which will be set to <code>false</code> by this method and
    *           which must be reset to this returned value when the scope
    *           is popped.
    */
   private boolean setupTriggerScope()
   {
      boolean oldBufScope = bufferScope;
      
      sym.addScope();
      sym.addSchemaScope(true);
      bufferScope = false;
      
      return oldBufScope;
   }
   
   /**
    * Common code to remove a scope (symbol dictionary and schema scopes) for
    * a trigger.  See <code>setupTriggerScope()</code>.
    * <p>
    * Triggers can be nested inside procedures, functions and other triggers
    * so we must restore the value found at entry so that we don't force
    * everything else to global scope while still in a nested scope.
    *
    * @param    oldBufScope
    *           The previous state of the <code>bufferScope</code> flag 
    *           which will be restored in this call.
    */
   private void clearTriggerScope(boolean oldBufScope)
   {
      bufferScope = oldBufScope;
      
      // triggers propagate
      sym.deleteSchemaScope(true);
      sym.deleteScope();
   }
   
   /**
    * Defines a new buffer for the given record name based on the given symbol node which
    * specifies the buffer name. The resulting symbol node is annotated as a result and the
    * symbol resolver and schema dictionary are updated as needed to be aware of the new buffer.
    *
    * @param    symbol
    *           The node that defines the buffer's name.
    * @param    recordName
    *           The record that defines the buffer's structure.
    * @param    am 
    *           Access modifiers if not null. Only will be non-null if this is a resource defined
    *           as a member of a class definition.
    * @param    st
    *           Static specifier if not null. Only will be non-null if this is a resource defined
    *           as a member of a class definition.
    * @param    forceTemp
    *           if {@code true}, assume {@code recordName} represents a temp-table, and do not
    *           search for it in the global schema scope when adding fields from it to the buffer
    *           being define.
    */
   private void defineBufferFromSymbol(Aast symbol,
                                       String recordName,
                                       Aast am,
                                       Aast st,
                                       boolean forceTemp)
   {
      String name = symbol.getText();
   
      // the target scope is determined by whether we are currently
      // executing in an internal procedure, function or trigger in which
      // case bufferScope is false, otherwise it is true and the
      // buffer is created in a global scope
      Object table = sym.addTable(name, BUFFER, bufferScope, symbol, am, st);
      
      sym.addFieldsFrom(table, recordName, forceTemp);
      
      String bufname = sym.lookupBufferName(name, false);
      symbol.putAnnotation("bufname", bufname);
   }
   
   /**
    * Examine any following hidden tokens and if there is any whitespace 
    * then an annotation named 'merge' will be set to <code>true</code> which
    * can be used later to properly build the converted command text back as
    * it was in the original source (where the lexing should be based ONLY
    * on whitespace).  If the AST node is a <code>KW_VALUE</code> or a
    * <code>STRING</code>, <code>merge</code> will be marked as 
    * <code>false</code> since no merging should be done in these cases.
    * Note that the <code>KW_VALUE</code> support is only enabled based on
    * the <code>honorValue</code>.
    * <p>
    * This is needed because the normal Progress parsing/lexing rules are
    * suspended when processing command text, BUT there is no clean way of
    * sharing state between the parser and lexer due to the lookahead design.
    * In Progress, it is likely that there is a single monolithic parser/lexer
    * which would allow for shared state to change lexing.  We have to put
    * this fixup in to merge tokens back together (because many command line
    * constructs have meaning in Progress and thus are lexed as specific 
    * tokens when really they are just text.  For example, &gt; is often used
    * and may need to be part of a larger construct like 2&gt;&amp;1 on the
    * command line, but it would be lexed as 4 tokens by our lexer.  We
    * used this method to detect when we need to merge the resulting AST
    * nodes back together at conversion time. 
    *
    * @param    tok
    *           The token from the token stream (the lexer) which is needed
    *           in order to examine if there are following hidden tokens.
    * @param    ast
    *           The AST node created from the given token.  This is the node
    *           that will be annotated.
    * @param    honorValue
    *           If <code>true</code>, the Progress VALUE() construct is
    *           recognized and no merge is implemented in this case.
    */
   private void processHiddenText(Token   tok,
                                  Aast    ast,
                                  boolean honorValue)
   {   
      // default to merging, unless we find following whitespace
      boolean merge = true;
      
      int type = tok.getType();
      
      if ((type == KW_VALUE && honorValue) || type == STRING)
      {
         merge = false;
      }
      else
      {
         // find first hidden token
         CommonHiddenStreamToken space = (CommonHiddenStreamToken) tok;
         space = hidden.getHiddenAfter(space);      
        
         // collect any hidden text
         while (space != null)
         {
            if (space.getType() == WS)
            {
               merge = false;
               break;
            }
            
            space = hidden.getHiddenAfter(space);
         }
      }
      
      // mark the node's merge requirement
      ast.putAnnotation("merge", new Boolean(merge));
   }
   
   /**
    * Merge the following non-hidden token text until (and including) a token
    * followed by whitespace or one of a list of following tokens that cannot be
    * matched.  There are 2 special tokens added to the list which is passed as 
    * a parameter: <code>DOT</code> and <code>EOF</code>.  The <code>DOT</code>
    * is special, in that it is allowed to continue the merging IF it is NOT
    * followed by whitespace.  IF the <code>DOT</code> is followed by whitespace
    * then the merging process ends and the <code>DOT</code> is NOT included
    * in the merged text.
    * <p>
    * All the merged tokens will be consumed and the merged text will be stored
    * in the last token before the token that triggered the end of merging. The
    * line and column numbers of the merged token will be forced to the same
    * values as the first token on entry.
    *
    * @param    end
    *           List of token types which trigger end of merging (in addition
    *           to <code>DOT</code> and <code>EOF</code>). May be
    *           <code>null</code> if there are no additional token types.
    *
    * @return   <code>true</code> if merging was done, <code>false</code> if
    *           no changes were made.
    */
   private boolean mergeUntilWS(int[] end)
   throws TokenStreamException,
          MismatchedTokenException
   {
      if (end == null)
      {
         end = new int[1];
      }
      else
      {
         int[] list = new int[end.length + 1];
         
         for (int i = 0; i < end.length; i++)
         {
            list[i + 1] = end[i];
         }
         
         end = list;
      }
      
      end[0] = EOF;
   
      boolean merged = false;
   
      CommonToken current = (CommonToken) LT(1);
      CommonToken next    = (CommonToken) LT(2);
      CommonToken after   = (CommonToken) LT(3);
      
      int line = current.getLine();
      int col  = current.getColumn();
      int type = next.getType();
      
      StringBuilder sb = new StringBuilder();
      
      while (!followedByWhitespace(current)     && 
             !matchesList(type, end)            &&
             !(type == DOT &&
               (followedByWhitespace(next) || (after != null && after.getType() == EOF))))
      {
         merged = true;
      
         // save the text of the current token
         sb.append(current.getText());
         
         // match the current token but leave the next token unmatched
         if (current.getType() == DOT)
         {
            match(DOT);
         }
         else
         {
            matchNot(DOT);
         }
         
         // update the state for next iteration
         current = next;
         next    = after;
         after   = (CommonToken) LT(3);
         type    = next.getType();
      }
      
      if (merged)
      {
         // put the last text into the buffer
         sb.append(current.getText());
      
         // override the last token's contents and positioning with the
         // merged results
         current.setText(sb.toString());
         current.setLine(line);
         current.setColumn(col);
         
         // we don't match here, so that the calling code can match on the
         // resulting token
      }
      
      return merged;
   }
   
   /**
    * Tests if the given <code>type</code> matches any in the given list.
    *
    * @param    type
    *           The token type to test for.
    * @param    list
    *           The list to check against.
    *
    * @return   <code>true</code> if the given type is in the list.
    */
   private boolean matchesList(int type, int[] list)
   {
      for (int i = 0; i < list.length; i++)
      {
         if (type == list[i])
            return true;
      }
      
      return false;
   }
   
   /**                          
    * Check if the given token is followed by whitespace in the original
    * source code (by reading from the hidden token stream).
    *
    * @param    tok
    *           The token from the token stream (the lexer) which is needed
    *           in order to examine if there are following hidden tokens.
    *
    * @return   <code>true</code> if the given token is followed by whitespace
    *           in the hidden token stream.
    */
   private boolean followedByWhitespace(Token tok)
   {
      CommonHiddenStreamToken space = null;
      space = hidden.getHiddenAfter((CommonHiddenStreamToken) tok);      
      
      return (space != null && space.getType() == WS);
   }
   
   /**
    * Check if the given token is followed by whitespace in the original
    * source code which includes a newline character. The check is done by
    * reading from the hidden token stream.
    *
    * @param    tok
    *           The token from the token stream (the lexer) which is needed
    *           in order to examine if there are following hidden tokens.
    *
    * @return   <code>true</code> if the given token is the last content on
    *           the line.
    */
   private boolean followedByLineEnd(Token tok)
   {
      CommonHiddenStreamToken space = null;
      space = hidden.getHiddenAfter((CommonHiddenStreamToken) tok);      
      
      if (space != null && space.getType() == WS)
      {
         if (space.getText().indexOf("\n") != -1)
         {
            return true;
         }
      }
      
      return false;
   }
   
   /**
    * Check if the given token is prefaced by whitespace in the original
    * source code (by reading from the hidden token stream).
    *
    * @param    tok
    *           The token from the token stream (the lexer) which is needed
    *           in order to examine any prefacing hidden tokens.
    *
    * @return   <code>true</code> if the given token is prefaced by whitespace
    *           in the hidden token stream.
    */
   private boolean prefacedByWhitespace(Token tok)
   {
      CommonHiddenStreamToken space = null;
      space = hidden.getHiddenBefore((CommonHiddenStreamToken) tok);      
      
      return (space != null && space.getType() == WS);
   }
   
   /**
    * Helper function to convert a <code>STRING</code> node into a node of the specified type.
    *
    * @param    node
    *           The <code>STRING</code> node to modify.
    * @param    newtype
    *           The new type of the target node. If passed as -1, the type will be looked up as
    *           a keyword.
    */
   private void convertStringTo(Aast node, int newtype)
   throws TokenStreamException,
          NoViableAltException
   {
      if (node.getType() != STRING)
      {
         throw new RuntimeException("Node must be a STRING type!");
      }
      
      // remove the outer quotes, handle escapes etc...
      String clean = character.progressToJavaString(node.getText());
      
      if (newtype == -1)
      {
         // simulate the keyword processing that would have happened if this wasn't a string
         Keyword found = sym.lookupKeyword(clean);
         
         if (found != null)
         {
            newtype = found.getTokenType();
         }
         else
         {
            throw new NoViableAltException(LT(1), getFilename());
         }
      }
                       
      // set the new token type and text
      node.setText(clean);
      node.setType(newtype);
   }
              
   /**
    * Helper function to replace the current token type with a new token
    * type, but first saving off the old type and then storing that old
    * token type in an annotation named 'oldtype'.
    *
    * @param    node
    *           The AST to annotate and modify.
    * @param    newtype
    *           The new type of the target node.
    */
   private void saveAndReplaceType(Aast node, int newtype)
   {
      // save off the old token type
      int oldtype = node.getType();
      
      // set the new token type
      node.setType(newtype);
      
      // save any original token type that rewriting may have erased
      // as an annotation (unless the oldtype is already set)
      if (node.getAnnotation("oldtype") == null)
      {
         node.putAnnotation("oldtype", new Long(oldtype));      
      }
   }
   
   /**
    * The 4GL seems to disambiguate assignment statements from usage of the EQUALS operator, by
    * examining the lvalue.  If the lvalue is of a type that cannot be assigned to, then the
    * operator is treated as an EQUALS not an ASSIGN.  This code attempts to do the same 
    * detection and will rewrite the operator node to an EQUALS (it will be passed in as an
    * ASSIGN).
    *
    * @param    lnode
    *           The lvalue AST to examine.
    * @param    opnode
    *           The operator AST to edit.
    *
    * @return   <code>true</code> if the operator was edited.
    */
   private boolean restoreEqualsOperator(Aast lnode, Aast opnode)
   {
      if (lnode != null && opnode != null)
      {
         int type = lnode.getType();
         
         // object methods can be "bare" (within a class definition a method call can
         // be made without an object reference and colon operator) but normally they
         // use the COLON operator; handle references and COM automation cannot be
         // bare, they always use an operator
         if (type == COLON || type == OBJECT_INVOCATION || type == COM_INVOCATION)
         {
            // in these cases we find the result type of the sub-expression by looking
            // at the 2nd child of the operator (even in a deeply nested chaining case)
            lnode = lnode.getChildAt(1);
            type  = lnode.getType();
         }
         
         if ((type >= BEGIN_FUNCTYPES && type <= END_FUNCTYPES) ||
             (type >= BEGIN_METH && type <= END_METH)           ||
             type == COM_METHOD                                 ||
             isLiteral(lnode))
         {
            opnode.setType(EQUALS);
            
            return true;
         }
      }
      
      return false;
   }
   
   /**
    * Report if the node is a literal.
    *
    * @param    node
    *           The node AST to examine.
    *
    * @return   <code>true</code> if the node is a literal.
    */
   private boolean isLiteral(Aast node)
   {
      return node.isAnnotation("is-literal") && (Boolean) node.getAnnotation("is-literal");
   }
   
   /**
    * Some parts of the parser can match a function when it should have matched the same name
    * as an "lvalue-like" language statement. Such matches would be a mistake (since we can't
    * properly lookahead the arbitrary amount needed to detect if there is an EQUALS token.
    * Instead, we detect the problem here and rewrite the node to "undo" the damage. 
    *
    * @param    node
    *           The AST to restore.
    *
    * @return   <code>true</code> if the node was edited.
    */
   private boolean restoreAssignStyleStmt(Aast node)
   {
      if (node.isAnnotation("oldtype"))
      {
         Long otype = (Long) node.getAnnotation("oldtype");
      
         if (otype == KW_CUR_VAL || otype == KW_DYN_CURV || 
             otype == KW_ENTRY   || otype == KW_SUBSTR   ||
             otype == KW_EXTENT  || otype == KW_LENGTH)
         {
            node.setType(otype.intValue());
            node.removeAnnotation("oldtype");
            node.removeAnnotation("builtin");
            node.removeAnnotation("returnsunknown");
            
            return true;
         }
      }
      
      return false;
   }
   
   /**
    * Helper to instantiate a new (limited use) lexer and re-lex the input.
    * The problem is that some initializers have non-strings inside quotes
    * and the lexer doesn't know when to remove the quotes and when to leave
    * them.  This results in the <code>INITIAL</code> having children that
    * are of the <code>STRING</code> type even when they should be some other
    * kind of literal.  <b>This method removes the unnecessary enclosing
    * quote characters and then calls the lexer to obtain the resulting token. 
    * It is assumed at this time that only one token will be present in the
    * passed input string.  If this is not the case, we will have to make
    * more serious changes!</b>
    *
    * @param    input
    *           The single or double quoted enclosed string containing a
    *           token to lex. MUST NOT be <code>null</code>. MUST have
    *           enclosing quote characters as the first AND last characters.
    *
    * @return   The token of the proper token type and text OR 
    *           <code>null</code> if the input text is an empty string.
    */
   private Token relex(String input) 
   throws TokenStreamException
   {
      String cleaner = input.substring(1, input.length() - 1);
      
      if ("".equals(cleaner))
         return null;
         
      SymbolResolver sr = new SymbolResolver(false);
      
      // disable the dot kludge as the following DOT will sometimes break
      // lexing, this needs to be a single token only
      ProgressLexer relex = new ProgressLexer(new StringReader(cleaner), sr, false);
                                               
      return relex.nextToken();
   }   
   
   /**
    * Loads the <code>SymbolResolver</code> with Progress 4GL built-in
    * functions (which take parameters). This is called during construction 
    * to ensure that the functions are available before parsing begins.
    * <p>
    * Each function is added with the full name (even if the keyword supports
    * abbreviations) because the {@link #func_call} method will detect when
    * a keyword is being used and it will use the keyword dictionary to
    * handle the abbreviation support.  This lookup will convert any
    * abbreviated text to the full text before lookup in the function
    * dictionary.
    * <p>
    * All functions are added with the token type associated with their
    * respective return value.  Some built-in functions may have the special
    * type of <code>FUNC_POLY</code>. This special type indicates that the
    * function can return multiple different return types. The built-in
    * functions typically implement the return type based on the type of the 
    * parameters passed.  User-defined functions cannot provide the same
    * feature, this can only occur in built-in functions.
    * <p>
    * Built-in functions that have no parameters are treated as variables.
    * See {@link #initializeVariableDictionary}.
    * <p>
    * Note that there are some built-in functions that can act like variables
    * OR as a normal parenthesized function.  In other words, for the
    * following list, the parenthesis are optional:
    * <p>
    * <ul>
    *    <li>ETIME
    *    <li>FRAME-COL
    *    <li>FRAME-DOWN
    *    <li>FRAME-LINE
    *    <li>FRAME-ROW
    *    <li>LINE-COUNTER
    *    <li>PAGE-NUMBER
    *    <li>PAGE-SIZE
    *    <li>SUPER (handled in the super_function rule, not here)
    *    <li>USERID (and it's synonym USER which is not an abbreviation)
    * </ul>
    * <p>
    * Such variables/functions are included in both this rule and in
    * <code>initializeVariableDictionary</code>. 
    */
   private void initializeFunctionDictionary()
   {
      sym.addBuiltinFunction("_cbit"                , FUNC_LOGICAL    , true );
      sym.addBuiltinFunction("_msg"                 , FUNC_INT        , true );
      sym.addBuiltinFunction("absolute"             , FUNC_POLY       , true ); // INT or DEC
      sym.addBuiltinFunction("accumulate"           , FUNC_POLY       , false); // used for annotations only, matches keyword text not actual function name
      sym.addBuiltinFunction("add-interval"         , FUNC_POLY       , false); // DATE, DATETIME or DATETIME-TZ
      sym.addBuiltinFunction("alias"                , FUNC_CHAR       , true );
      sym.addBuiltinFunction("ambiguous"            , FUNC_LOGICAL    , false); // used for annotations only
      sym.addBuiltinFunction("ascending"            , FUNC_INT64      , true ); // this is a synonym for ASC
      sym.addBuiltinFunction("audit-enabled"        , FUNC_LOGICAL    , false);
      sym.addBuiltinFunction("available"            , FUNC_LOGICAL    , false); // used for annotations only
      sym.addBuiltinFunction("base64-decode"        , FUNC_MEMPTR     , false);
      sym.addBuiltinFunction("base64-encode"        , FUNC_LONGCHAR   , false);
      sym.addBuiltinFunction("box"                  , FUNC_CLASS      , true , "System.Object"); // TODO: this qualified class is wrong but we need a placeholder
      sym.addBuiltinFunction("can-do"               , FUNC_LOGICAL    , true );
      sym.addBuiltinFunction("can-find"             , FUNC_LOGICAL    , false); // used for annotations only
      sym.addBuiltinFunction("can-query"            , FUNC_LOGICAL    , true );
      sym.addBuiltinFunction("can-set"              , FUNC_LOGICAL    , true );
      sym.addBuiltinFunction("caps"                 , FUNC_CHAR       , true );
      sym.addBuiltinFunction("cast"                 , FUNC_CLASS      , true ); // used for annotations only
      sym.addBuiltinFunction("chr"                  , FUNC_CHAR       , true );
      sym.addBuiltinFunction("codepage-convert"     , FUNC_CHAR       , true );
      sym.addBuiltinFunction("compare"              , FUNC_LOGICAL    , true );
      sym.addBuiltinFunction("connected"            , FUNC_LOGICAL    , true );
      sym.addBuiltinFunction("count-of"             , FUNC_INT        , false);
      sym.addBuiltinFunction("current-changed"      , FUNC_LOGICAL    , false); // used for annotations only
      sym.addBuiltinFunction("current-result-row"   , FUNC_INT        , true );
      sym.addBuiltinFunction("current-value"        , FUNC_INT64      , true ); // used for annotations only
      sym.addBuiltinFunction("data-source-modified" , FUNC_LOGICAL    , false);
      sym.addBuiltinFunction("date"                 , FUNC_DATE       , true );
      sym.addBuiltinFunction("datetime"             , FUNC_DATETIME   , false);
      sym.addBuiltinFunction("datetime-tz"          , FUNC_DATETIME_TZ, false);
      sym.addBuiltinFunction("day"                  , FUNC_INT        , true );
      sym.addBuiltinFunction("dbcodepage"           , FUNC_CHAR       , true );
      sym.addBuiltinFunction("dbcollation"          , FUNC_CHAR       , true );
      sym.addBuiltinFunction("dbparam"              , FUNC_CHAR       , true );
      sym.addBuiltinFunction("db-remote-host"       , FUNC_CHAR       , true );
      sym.addBuiltinFunction("dbrestrictions"       , FUNC_CHAR       , true );
      sym.addBuiltinFunction("dbtaskid"             , FUNC_INT        , true );
      sym.addBuiltinFunction("dbtype"               , FUNC_CHAR       , true );
      sym.addBuiltinFunction("dbversion"            , FUNC_CHAR       , true );
      sym.addBuiltinFunction("decimal"              , FUNC_DEC        , true );
      sym.addBuiltinFunction("decrypt"              , FUNC_MEMPTR     , false);
      sym.addBuiltinFunction("dynamic-cast"         , FUNC_CLASS      , true , "Progress.Lang.Object");  // TODO: this qualified class is wrong but we need a placeholder
      sym.addBuiltinFunction("dynamic-current-value", FUNC_INT64      , false);
      sym.addBuiltinFunction("dynamic-function"     , FUNC_POLY       , true ); // like called func, uses for annotations only
      sym.addBuiltinFunction("dynamic-invoke"       , FUNC_POLY       , true );
      sym.addBuiltinFunction("dynamic-new"          , FUNC_CLASS      , true , "Progress.Lang.Object");  // used for annotations only
      sym.addBuiltinFunction("dynamic-next-value"   , FUNC_INT64      , false);
      sym.addBuiltinFunction("encode"               , FUNC_CHAR       , true );
      sym.addBuiltinFunction("encrypt"              , FUNC_MEMPTR     , false);
      sym.addBuiltinFunction("entered"              , FUNC_LOGICAL    , false); // used for annotations only
      sym.addBuiltinFunction("entry"                , FUNC_CHAR       , true );
      sym.addBuiltinFunction("error"                , FUNC_LOGICAL    , false); // used for annotations only
      sym.addBuiltinFunction("etime"                , FUNC_INT        , false);
      sym.addBuiltinFunction("exp"                  , FUNC_DEC        , true );
      sym.addBuiltinFunction("extent"               , FUNC_INT        , false);
      sym.addBuiltinFunction("fill"                 , FUNC_CHAR       , true );
      sym.addBuiltinFunction("first"                , FUNC_LOGICAL    , false);
      sym.addBuiltinFunction("first-of"             , FUNC_LOGICAL    , false);
      sym.addBuiltinFunction("frame-col"            , FUNC_DEC        , false);
      sym.addBuiltinFunction("frame-down"           , FUNC_INT        , false);
      sym.addBuiltinFunction("frame-line"           , FUNC_INT        , false);
      sym.addBuiltinFunction("frame-row"            , FUNC_DEC        , false);
      sym.addBuiltinFunction("generate-pbe-key"     , FUNC_RAW        , false);
      sym.addBuiltinFunction("get-bits"             , FUNC_INT        , true );
      sym.addBuiltinFunction("get-byte"             , FUNC_INT        , true );
      sym.addBuiltinFunction("get-byte-order"       , FUNC_INT        , false);
      sym.addBuiltinFunction("get-bytes"            , FUNC_POLY       , true ); // RAW or MEMPTR
      sym.addBuiltinFunction("get-codepage"         , FUNC_CHAR       , false);
      sym.addBuiltinFunction("get-collation"        , FUNC_CHAR       , false);
      sym.addBuiltinFunction("get-collations"       , FUNC_CHAR       , true );
      sym.addBuiltinFunction("get-double"           , FUNC_DEC        , true );
      sym.addBuiltinFunction("get-float"            , FUNC_DEC        , true );
      sym.addBuiltinFunction("get-int64"            , FUNC_INT64      , false);
      sym.addBuiltinFunction("get-long"             , FUNC_INT        , true );
      sym.addBuiltinFunction("get-pointer-value"    , FUNC_INT64      , false);
      sym.addBuiltinFunction("get-short"            , FUNC_INT        , true );
      sym.addBuiltinFunction("get-size"             , FUNC_INT64      , false);
      sym.addBuiltinFunction("get-string"           , FUNC_CHAR       , true );
      sym.addBuiltinFunction("get-unsigned-short"   , FUNC_INT        , true );
      sym.addBuiltinFunction("get-unsigned-long"    , FUNC_INT64      , false);
      sym.addBuiltinFunction("guid"                 , FUNC_CHAR       , false);
      sym.addBuiltinFunction("handle"               , FUNC_HANDLE     , false);
      sym.addBuiltinFunction("hex-decode"           , FUNC_RAW        , false);
      sym.addBuiltinFunction("hex-encode"           , FUNC_CHAR       , false);
      sym.addBuiltinFunction("if"                   , FUNC_POLY       , true ); // used for annotations only
      sym.addBuiltinFunction("index"                , FUNC_INT        , true );
      sym.addBuiltinFunction("input"                , FUNC_POLY       , false); // used for annotations only
      sym.addBuiltinFunction("integer"              , FUNC_INT        , true );
      sym.addBuiltinFunction("int64"                , FUNC_INT64      , false);
      sym.addBuiltinFunction("interval"             , FUNC_INT64      , false);
      sym.addBuiltinFunction("is-codepage-fixed"    , FUNC_LOGICAL    , false);
      sym.addBuiltinFunction("is-column-codepage"   , FUNC_LOGICAL    , false);
      sym.addBuiltinFunction("is-lead-byte"         , FUNC_LOGICAL    , true );
      sym.addBuiltinFunction("iso-date"             , FUNC_CHAR       , false);
      sym.addBuiltinFunction("kblabel"              , FUNC_CHAR       , true );
      sym.addBuiltinFunction("keycode"              , FUNC_INT        , true );
      sym.addBuiltinFunction("key-code"             , FUNC_INT        , true );
      sym.addBuiltinFunction("keyfunction"          , FUNC_CHAR       , true );
      sym.addBuiltinFunction("key-function"         , FUNC_CHAR       , true );
      sym.addBuiltinFunction("keylabel"             , FUNC_CHAR       , true );
      sym.addBuiltinFunction("key-label"            , FUNC_CHAR       , true );
      sym.addBuiltinFunction("keyword"              , FUNC_CHAR       , true );
      sym.addBuiltinFunction("keyword-all"          , FUNC_CHAR       , true );
      sym.addBuiltinFunction("last"                 , FUNC_LOGICAL    , false);
      sym.addBuiltinFunction("last-of"              , FUNC_LOGICAL    , false);
      sym.addBuiltinFunction("lc"                   , FUNC_CHAR       , true );
      sym.addBuiltinFunction("ldbname"              , FUNC_CHAR       , true );
      sym.addBuiltinFunction("left-trim"            , FUNC_CHAR       , true );
      sym.addBuiltinFunction("length"               , FUNC_INT        , true );
      sym.addBuiltinFunction("library"              , FUNC_CHAR       , true );
      sym.addBuiltinFunction("line-counter"         , FUNC_INT        , false);
      sym.addBuiltinFunction("list-events"          , FUNC_CHAR       , true );
      sym.addBuiltinFunction("list-query-attrs"     , FUNC_CHAR       , true );
      sym.addBuiltinFunction("list-set-attrs"       , FUNC_CHAR       , true );
      sym.addBuiltinFunction("list-widgets"         , FUNC_CHAR       , true );
      sym.addBuiltinFunction("load-picture"         , FUNC_COM_HANDLE , true ); // supposed to be a "statement" but really acts like a function
      sym.addBuiltinFunction("locked"               , FUNC_LOGICAL    , false); // used for annotations only
      sym.addBuiltinFunction("log"                  , FUNC_DEC        , true );
      sym.addBuiltinFunction("logical"              , FUNC_LOGICAL    , false);
      sym.addBuiltinFunction("lookup"               , FUNC_INT        , true );
      sym.addBuiltinFunction("lower"                , FUNC_CHAR       , true );
      sym.addBuiltinFunction("maximum"              , FUNC_POLY       , true ); // same as parms
      sym.addBuiltinFunction("md5-digest"           , FUNC_RAW        , false);
      sym.addBuiltinFunction("member"               , FUNC_CHAR       , true );
      sym.addBuiltinFunction("message-digest"       , FUNC_RAW        , false);
      sym.addBuiltinFunction("minimum"              , FUNC_POLY       , true ); // same as parms
      sym.addBuiltinFunction("month"                , FUNC_INT        , true );
      sym.addBuiltinFunction("mtime"                , FUNC_INT        , false);
      sym.addBuiltinFunction("new"                  , FUNC_LOGICAL    , false); // used for annotations only
      sym.addBuiltinFunction("next-value"           , FUNC_INT64      , true ); // used for annotations only
      sym.addBuiltinFunction("normalize"            , FUNC_POLY       , false); // CHAR or LONGCHAR
      sym.addBuiltinFunction("not entered"          , FUNC_LOGICAL    , false); // used for annotations only
      sym.addBuiltinFunction("now"                  , FUNC_DATETIME_TZ, true );
      sym.addBuiltinFunction("num-entries"          , FUNC_INT        , true );
      sym.addBuiltinFunction("num-results"          , FUNC_INT        , true );
      sym.addBuiltinFunction("os-getenv"            , FUNC_CHAR       , true );
      sym.addBuiltinFunction("p2j-remote-call"      , FUNC_CHAR       , true); // P2J-extension, not real 4GL!
      sym.addBuiltinFunction("page-number"          , FUNC_INT        , false);
      sym.addBuiltinFunction("page-size"            , FUNC_INT        , false);
      sym.addBuiltinFunction("pdbname"              , FUNC_CHAR       , true );
      sym.addBuiltinFunction("program-name"         , FUNC_CHAR       , true );
      sym.addBuiltinFunction("query-off-end"        , FUNC_LOGICAL    , true );
      sym.addBuiltinFunction("quoter"               , FUNC_CHAR       , false);
      sym.addBuiltinFunction("r-index"              , FUNC_INT        , true );
      sym.addBuiltinFunction("raw"                  , FUNC_RAW        , false);
      sym.addBuiltinFunction("random"               , FUNC_INT        , true );
      sym.addBuiltinFunction("recid"                , FUNC_RECID      , true ); // used for annotations only
      sym.addBuiltinFunction("record-length"        , FUNC_INT        , false);
      sym.addBuiltinFunction("rejected"             , FUNC_LOGICAL    , false);
      sym.addBuiltinFunction("replace"              , FUNC_CHAR       , true );
      sym.addBuiltinFunction("retry"                , FUNC_LOGICAL    , true ); // used for annotations only
      sym.addBuiltinFunction("rgb-value"            , FUNC_INT        , true );
      sym.addBuiltinFunction("right-trim"           , FUNC_CHAR       , true );
      sym.addBuiltinFunction("round"                , FUNC_DEC        , true );
      sym.addBuiltinFunction("rowid"                , FUNC_ROWID      , true ); // used for annotations only
      sym.addBuiltinFunction("row-state"            , FUNC_INT        , false); // used for annotations only
      sym.addBuiltinFunction("sdbname"              , FUNC_CHAR       , true );
      sym.addBuiltinFunction("search"               , FUNC_CHAR       , true );
      sym.addBuiltinFunction("seek"                 , FUNC_INT        , true );
      sym.addBuiltinFunction("setuserid"            , FUNC_LOGICAL    , true );
      sym.addBuiltinFunction("set-db-client"        , FUNC_LOGICAL    , false);
      sym.addBuiltinFunction("sha1-digest"          , FUNC_RAW        , false);
      sym.addBuiltinFunction("sqrt"                 , FUNC_DEC        , true );
      sym.addBuiltinFunction("ssl-server-name"      , FUNC_CHAR       , false);
      sym.addBuiltinFunction("string"               , FUNC_CHAR       , true );
      sym.addBuiltinFunction("substitute"           , FUNC_CHAR       , true );
      sym.addBuiltinFunction("substring"            , FUNC_CHAR       , true );
      sym.addBuiltinFunction("super"                , FUNC_POLY       , true ); // used for annotations only
      sym.addBuiltinFunction("time"                 , FUNC_INT        , false); // used for annotations only
      sym.addBuiltinFunction("timezone"             , FUNC_INT        , false);
      sym.addBuiltinFunction("to-rowid"             , FUNC_ROWID      , true );
      sym.addBuiltinFunction("today"                , FUNC_DATE       , false); // used for annotations only
      sym.addBuiltinFunction("trans"                , FUNC_LOGICAL    , true ); // used for annotations only
      sym.addBuiltinFunction("transaction"          , FUNC_LOGICAL    , true ); // used for annotations only
      sym.addBuiltinFunction("trim"                 , FUNC_CHAR       , true );
      sym.addBuiltinFunction("truncate"             , FUNC_DEC        , true );
      sym.addBuiltinFunction("type-of"              , FUNC_LOGICAL    , false);
      sym.addBuiltinFunction("upper"                , FUNC_CHAR       , true );
      sym.addBuiltinFunction("user"                 , FUNC_CHAR       , false);
      sym.addBuiltinFunction("userid"               , FUNC_CHAR       , false);
      sym.addBuiltinFunction("valid-event"          , FUNC_LOGICAL    , true );
      sym.addBuiltinFunction("valid-handle"         , FUNC_LOGICAL    , false);
      sym.addBuiltinFunction("valid-object"         , FUNC_LOGICAL    , false);
      sym.addBuiltinFunction("weekday"              , FUNC_INT        , true );
      sym.addBuiltinFunction("widget-handle"        , FUNC_HANDLE     , true );
      sym.addBuiltinFunction("year"                 , FUNC_INT        , true );
      
      if (webspeed)
      {
         // add our "fake" webspeed built-in functions
         sym.addBuiltinFunction("available-messages" , FUNC_LOGICAL , false);
         sym.addBuiltinFunction("check-agent-mode"   , FUNC_LOGICAL , false);
         sym.addBuiltinFunction("convert-datetime"   , FUNC_CHAR    , false);
         sym.addBuiltinFunction("delete-cookie"      , FUNC_CHAR    , true );
         sym.addBuiltinFunction("format-datetime"    , FUNC_CHAR    , false);
         sym.addBuiltinFunction("get-cgi"            , FUNC_CHAR    , false);
         sym.addBuiltinFunction("get-cgi-long"       , FUNC_LONGCHAR, false);
         sym.addBuiltinFunction("get-config"         , FUNC_CHAR    , false);
         sym.addBuiltinFunction("get-cookie"         , FUNC_CHAR    , false);
         sym.addBuiltinFunction("get-field"          , FUNC_CHAR    , false);
         sym.addBuiltinFunction("get-long-value"     , FUNC_LONGCHAR, false);
         sym.addBuiltinFunction("get-message-groups" , FUNC_CHAR    , false);
         sym.addBuiltinFunction("get-messages"       , FUNC_CHAR    , false);
         sym.addBuiltinFunction("get-user-field"     , FUNC_CHAR    , false);
         sym.addBuiltinFunction("get-value"          , FUNC_CHAR    , false);
         sym.addBuiltinFunction("hidden-field"       , FUNC_CHAR    , false);
         sym.addBuiltinFunction("hidden-field-list"  , FUNC_CHAR    , false);
         sym.addBuiltinFunction("html-encode"        , FUNC_CHAR    , false);
         sym.addBuiltinFunction("output-content-type", FUNC_LOGICAL , false);
         sym.addBuiltinFunction("output-http-header" , FUNC_CHAR    , false);
         sym.addBuiltinFunction("output-messages"    , FUNC_INT     , false);
         sym.addBuiltinFunction("queue-message"      , FUNC_INT     , false);
         sym.addBuiltinFunction("set-cookie"         , FUNC_CHAR    , false);
         sym.addBuiltinFunction("set-user-field"     , FUNC_LOGICAL , false);
         sym.addBuiltinFunction("setwebstate"        , FUNC_LOGICAL , false);
         sym.addBuiltinFunction("unbox"              , FUNC_POLY    , true );
         sym.addBuiltinFunction("url-decode"         , FUNC_CHAR    , false);
         sym.addBuiltinFunction("url-encode"         , FUNC_CHAR    , false);
         sym.addBuiltinFunction("url-field"          , FUNC_CHAR    , false);
         sym.addBuiltinFunction("url-field-list"     , FUNC_CHAR    , false);
         sym.addBuiltinFunction("url-format"         , FUNC_CHAR    , false);
      }
   }
   
   /**
    * Loads the <code>SymbolResolver</code> with Progress 4GL attributes and
    * methods. This is called during construction to ensure that resolution 
    * is available before parsing begins.
    * <p>
    * Each attribute or method is added with the keyword's token type as the
    * key and the data type (attributes) or return type (methods) as the
    * mapped value.
    */
   private void initializeAttributesDictionary()
   {
      sym.addAttributeOrMethod( KW_ACCEL,    ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_ACC_CHG , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_ACC_RCHG, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_ACTIVE  , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_ACTOR   , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_ADD_BUF , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_ADD_C_C , METH_HANDLE  );
      sym.addAttributeOrMethod( KW_ADD_C_F , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_ADD_EVTP, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_ADD_F_F , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_ADD_1ST , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_ADD_HENT, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_ADD_IDXF, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_ADD_LAST, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_ADD_L_C , METH_HANDLE  );
      sym.addAttributeOrMethod( KW_ADD_LIKF, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_ADD_LIKI, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_ADD_NEWF, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_ADD_NEWI, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_ADD_PREL, METH_HANDLE  );
      sym.addAttributeOrMethod( KW_ADD_REL , METH_HANDLE  );
      sym.addAttributeOrMethod( KW_ADD_SLOC, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_ADD_SRCB, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_ADD_SUP , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_ADM_DATA, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_AFT_BUFF, ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_AFT_ROID, ATTR_ROWID   );
      sym.addAttributeOrMethod( KW_AFT_TBL , ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_ALLW_C_S, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_ALW_ON_T, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_AMBIG   , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_APPEND_C, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_APPL_A_B, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_APPL_CID, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_APPL_CBK, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_APP_INFO, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_APP_PW  , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_APP_UID , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_ASYNC   , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_ASYNC_RC, ATTR_INT     );
      sym.addAttributeOrMethod( KW_ASYNC_RH, ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_ATT_DSRC, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_ATT_PLST, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_ATTR    , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_ATTR_NAM, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_AUD_EV_C, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_AUTHEN_F, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_AUTO_COM, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_AUTO_DEL, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_AUTO_D_X, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_AUTO_END, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_AUTO_GO , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_AUTO_IND, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_AUTO_RES, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_AUTO_RET, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_AUTO_SYN, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_AUTO_VAL, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_AUTO_ZAP, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_AVAIL   , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_AVL_FMTS, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_BACKGRND, ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_BAS_LOGG, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_BASE_ADE, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_BATCH_MO, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_BATCH_SZ, ATTR_INT     );
      sym.addAttributeOrMethod( KW_B4_BUFF , ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_B4_ROWID, ATTR_ROWID   );
      sym.addAttributeOrMethod( KW_B4_TABLE, ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_BEG_EV_G, METH_CHAR    );
      sym.addAttributeOrMethod( KW_BGCOLOR , ATTR_INT     );
      sym.addAttributeOrMethod( KW_BLANK   , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_BLK_IT_D, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_BORD_B_C, ATTR_DEC     );
      sym.addAttributeOrMethod( KW_BORD_B_P, ATTR_INT     );
      sym.addAttributeOrMethod( KW_BORD_L_C, ATTR_DEC     );
      sym.addAttributeOrMethod( KW_BORD_L_P, ATTR_INT     );
      sym.addAttributeOrMethod( KW_BORD_R_C, ATTR_DEC     );
      sym.addAttributeOrMethod( KW_BORD_R_P, ATTR_INT     );
      sym.addAttributeOrMethod( KW_BORD_T_C, ATTR_DEC     );
      sym.addAttributeOrMethod( KW_BORD_T_P, ATTR_INT     );
      sym.addAttributeOrMethod( KW_BOX     , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_BOX_SEL , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_BUFFER_C, ATTR_INT     );
      sym.addAttributeOrMethod( KW_BUFFER_L, ATTR_INT     );
      sym.addAttributeOrMethod( KW_BUF_COMP, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_BUF_COPY, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_BUF_CREA, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_BUF_DEL , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_BUF_FLD , ATTR_HANDLE  ); // warning: this can have LPARENS-based subscripting!
      sym.addAttributeOrMethod( KW_BUF_HNDL, ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_BUF_NAME, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_BUF_REL , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_BUF_VLID, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_BUF_VAL , ATTR_POLY    );
      sym.addAttributeOrMethod( KW_BYTES_R , ATTR_INT     );
      sym.addAttributeOrMethod( KW_BYTES_W , ATTR_INT     );
      sym.addAttributeOrMethod( KW_CACHE,    ATTR_INT     );
      sym.addAttributeOrMethod( KW_CALL_NAM, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_CALL_TYP, ATTR_INT     );
      sym.addAttributeOrMethod( KW_CALLBACK, ATTR_CHAR    ); // FWD TIMER extension
      sym.addAttributeOrMethod( KW_CANC_BRK, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_CANCEL_B, ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_CANCEL_R, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_CANCELLD, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_CAN_CREA, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_CAN_DEL , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_CAN_READ, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_CAN_WRT , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_CARE_PNT, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_CASE_SEN, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_CEASE,    METH_VOID    ); // FWD TIMER extension
      sym.addAttributeOrMethod( KW_CENTER  , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_CFG_NAME, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_CHARSET , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_CHECKED , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_CHLD_BUF, ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_CHLD_NUM, ATTR_INT     );
      sym.addAttributeOrMethod( KW_CLEAR   , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_CLR_AP_C, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_CLR_LOG , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_CLR_SEL , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_CLR_S_AR, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_CLS_TYPE, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_CLNT_C_I, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_CLNT_TTY, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_CLNT_TYP, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_CLNT_WS , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_CLONE_ND, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_CLOSE_LG, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_CNCL_R_A, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_CNTRL_BX, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_CP      , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_CPY_DSET, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_CPY_SATR, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_CPY_TTBL, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_CODE    , ATTR_INT     );
      sym.addAttributeOrMethod( KW_COL     , ATTR_DEC     );
      sym.addAttributeOrMethod( KW_COL_BGC , ATTR_INT     );
      sym.addAttributeOrMethod( KW_COL_DC  , ATTR_INT     );
      sym.addAttributeOrMethod( KW_COL_FGC , ATTR_INT     );
      sym.addAttributeOrMethod( KW_COL_FONT, ATTR_INT     );
      sym.addAttributeOrMethod( KW_COL_LAB , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_COL_MOV , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_COL_PFC , ATTR_INT     );
      sym.addAttributeOrMethod( KW_COL_R_O , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_COL_RES , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_COL_SCR , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_COM_HNDL, ATTR_COM_HANDLE );
      sym.addAttributeOrMethod( KW_COMPLETE, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_CONN    , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_CONN_ED , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_CTX_H   , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_CTX_H_F , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_CTX_H_ID, ATTR_INT     );
      sym.addAttributeOrMethod( KW_CVT_3D_C, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_COVERAGE, ATTR_LOGICAL ); // undocumented, found in customer code
      sym.addAttributeOrMethod( KW_CPCASE  , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_CPCOLL  , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_CPINTERN, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_CPLOG   , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_CPPRINT , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_CPRCODEI, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_CPRCODEO, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_CPSTREAM, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_CPTERM  , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_CRC_VAL , ATTR_INT     );
      sym.addAttributeOrMethod( KW_CREAT_LK, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_CREAT_OA, ATTR_LOGICAL );  // undocumented, found in customer code
      sym.addAttributeOrMethod( KW_CREAT_ND, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_CREAT_NN, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_CREAT_RL, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_CUR_CHG , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_CUR_CHAR, ATTR_INT     );
      sym.addAttributeOrMethod( KW_CUR_COL , ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_CUR_ITER, ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_CUR_LINE, ATTR_INT     );
      sym.addAttributeOrMethod( KW_CUR_ENV , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_CUR_OFF , ATTR_INT     );
      sym.addAttributeOrMethod( KW_CUR_RES , ATTR_INT     );
      sym.addAttributeOrMethod( KW_CUR_R_M , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_CUR_WIN , ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_CVT_2OFF, METH_INT     );
      sym.addAttributeOrMethod( KW_DATA_E_R, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_DATA_SCM, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_DATA_SM , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_DATA_SRC, ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_DATASET , ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_DATATYPE, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_DATE_FMT, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_DB_CTXT , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_DB_REF  , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_DBNAME  , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_DCOLOR  , ATTR_INT     );
      sym.addAttributeOrMethod( KW_DDE_ERR , ATTR_INT     );
      sym.addAttributeOrMethod( KW_DDE_ID  , ATTR_INT     );
      sym.addAttributeOrMethod( KW_DDE_ITEM, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_DDE_NAME, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_DDE_TOPI, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_DBG_ALRT, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_DEBLANK , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_DEBUG   , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_DECIMALS, ATTR_INT     );
      sym.addAttributeOrMethod( KW_DECL_NSP, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_DEFAULT , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_DEF_BUFH, ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_DEF_COMM, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_DEFLT_BN, ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_DEF_STR , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_DEF_VAL , ATTR_POLY    );
      sym.addAttributeOrMethod( KW_DEL_CHAR, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_DEL_C_R , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_DEL_H_EN, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_DEL_LINE, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_DEL_NODE, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_DEL_R_L , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_DEL_S_R , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_DEL_S_RS, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_DELETE  , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_DELIMIT , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_DESCR   , ATTR_CHAR    ); // undocumented, found in customer code
      sym.addAttributeOrMethod( KW_DESEL_R , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_DESEL_FR, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_DESEL_SR, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_DET_DSRC, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_DIRECTRY, ATTR_CHAR    ); // undocumented, found in customer code
      sym.addAttributeOrMethod( KW_DIS_A_ZA, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_DIS_D_TR, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_DIS_L_TR, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_DIS_REDR, ATTR_LOGICAL ); // FWD extension, not real 4GL!
      sym.addAttributeOrMethod( KW_DISABL_C, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_DISABLE , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_DISCONN , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_DISP_MSG, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_DISP_TZ , ATTR_INT     );
      sym.addAttributeOrMethod( KW_DISP_TYP, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_DOMAIN_D, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_DOMAIN_N, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_DOMAIN_T, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_DOWN    , ATTR_INT     );
      sym.addAttributeOrMethod( KW_DRAG_EN , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_DROP_TAR, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_DUMP_LGN, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_DYNAMIC , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_EDGE_C  , ATTR_DEC     );
      sym.addAttributeOrMethod( KW_EDGE_P  , ATTR_INT     );
      sym.addAttributeOrMethod( KW_EDIT_C_P, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_EDIT_C_U, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_EDIT_CLR, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_EDIT_CPY, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_EDIT_CUT, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_EDIT_PAS, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_EDIT_UND, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_EMPTY   , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_EMPTY_DS, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_EMPTY_TT, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_ENABLE  , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_ENABLE_C, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_ENABLE_E, METH_CHAR    );
      sym.addAttributeOrMethod( KW_ENABLED , ATTR_LOGICAL ); // undocumented, found in customer code
      sym.addAttributeOrMethod( KW_ENC_AMK , METH_CHAR    );
      sym.addAttributeOrMethod( KW_ENC_SALT, ATTR_RAW     ); 
      sym.addAttributeOrMethod( KW_ENCODING, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_END_DOC , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_END_ELEM, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_END_EV_G, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_END_F_D , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_END_USER, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_ENT_TLST, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_ENTRY   , METH_CHAR    );
      sym.addAttributeOrMethod( KW_ERROR   , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_ERR_COL , ATTR_INT     );
      sym.addAttributeOrMethod( KW_ERR_OBJD, ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_ERR_ROW , ATTR_INT     );
      sym.addAttributeOrMethod( KW_ERR_S_T , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_ERR_STR , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_EVT_GRID, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_EVT_PROC, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_EVT_TYPE, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_E_PROC_C, ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_EXEC_LOG, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_EXCL_ID , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_EXPAND  , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_EXPANDBL, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_EXPORT  , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_EXPORT_P, METH_RAW     );
      sym.addAttributeOrMethod( KW_EXTENT  , ATTR_INT     );
      sym.addAttributeOrMethod( KW_FETCH_SR, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_FGCOLOR , ATTR_INT     );
      sym.addAttributeOrMethod( KW_FIL_C_D , ATTR_DATE    );
      sym.addAttributeOrMethod( KW_FIL_C_T , ATTR_INT     );
      sym.addAttributeOrMethod( KW_FIL_NAME, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_FIL_OFF,  ATTR_INT     );
      sym.addAttributeOrMethod( KW_FIL_M_D , ATTR_DATE    );
      sym.addAttributeOrMethod( KW_FIL_M_T , ATTR_INT     );
      sym.addAttributeOrMethod( KW_FIL_SIZE, ATTR_INT     );
      sym.addAttributeOrMethod( KW_FIL_TYPE, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_FILL    , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_FILL_MOD, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_FILL_WST, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_FILLED  , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_FIND_BR , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_FIND_CUR, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_FIND_1ST, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_FIND_LST, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_FIND_UNI, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_FIR_DSET, ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_FIRST_AR, ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_FIRST_BU, ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_FIRST_CH, ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_FIRST_CO, ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_FIRST_DS, ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_FIRST_FM, ATTR_CLASS  , "Progress.Windows.IForm" );
      sym.addAttributeOrMethod( KW_FIRST_OB, ATTR_CLASS  , SymbolResolver.ROOT_OBJ_NAME );
      sym.addAttributeOrMethod( KW_FIRST_OF, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_FIT_LCOL, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_FIRST_PR, ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_FIRST_QR, ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_FIRST_SR, ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_FIRST_SS, ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_FIRST_SO, ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_FIRST_TI, ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_FLAT_BUT, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_FOCUS_R , ATTR_INT     );
      sym.addAttributeOrMethod( KW_FOCUS_RS, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_FONT    , ATTR_INT     );
      sym.addAttributeOrMethod( KW_FORE    , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_FORM_INP, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_FORM_LIN, ATTR_MEMPTR  );
      sym.addAttributeOrMethod( KW_FORMAT  , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_FORMATTE, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_FRAGMENT, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_FRAME   , ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_FR_COL  , ATTR_DEC     );
      sym.addAttributeOrMethod( KW_FR_NAME , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_FR_ROW  , ATTR_DEC     );
      sym.addAttributeOrMethod( KW_FR_SPACE, ATTR_INT     );
      sym.addAttributeOrMethod( KW_FR_X    , ATTR_INT     );
      sym.addAttributeOrMethod( KW_FR_Y    , ATTR_INT     );
      sym.addAttributeOrMethod( KW_FREQ    , ATTR_INT     );
      sym.addAttributeOrMethod( KW_FULL_H_C, ATTR_DEC     );
      sym.addAttributeOrMethod( KW_FULL_H_P, ATTR_INT     );
      sym.addAttributeOrMethod( KW_FULL_W_C, ATTR_DEC     );
      sym.addAttributeOrMethod( KW_FULL_W_P, ATTR_INT     );
      sym.addAttributeOrMethod( KW_FULLPATH, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_FUNCT   , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_FWD_ONLY, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_GET_ATTR, METH_CHAR    );  // the 4GL docs say this should return LOGICAL, but that is not actually correct
      sym.addAttributeOrMethod( KW_GET_A_N , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_GET_BDAT, METH_MEMPTR  );
      sym.addAttributeOrMethod( KW_GET_BLUE, METH_INT     );
      sym.addAttributeOrMethod( KW_GET_BR_C, METH_HANDLE  );
      sym.addAttributeOrMethod( KW_GET_BUFH, METH_HANDLE  );
      sym.addAttributeOrMethod( KW_GET_B_A , ATTR_INT     );
      sym.addAttributeOrMethod( KW_GET_CBPC, METH_HANDLE  );
      sym.addAttributeOrMethod( KW_GET_CBPN, METH_CHAR    );
      sym.addAttributeOrMethod( KW_GET_CFGV, METH_CHAR    );
      sym.addAttributeOrMethod( KW_GET_CGIL, METH_CHAR    );
      sym.addAttributeOrMethod( KW_GET_CGIV, METH_CHAR    );
      sym.addAttributeOrMethod( KW_GET_CGLV, METH_LONGCHAR);
      sym.addAttributeOrMethod( KW_GET_CHG , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_GET_CHLD, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_GET_CREL, METH_HANDLE  );
      sym.addAttributeOrMethod( KW_GET_CUR , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_GET_D_E , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_GET_D_F , METH_CHAR    );
      sym.addAttributeOrMethod( KW_GET_DS_B, METH_HANDLE  );
      sym.addAttributeOrMethod( KW_GET_DYN , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_GET_ERRC, METH_INT     );
      sym.addAttributeOrMethod( KW_GET_ERRR, METH_INT     );
      sym.addAttributeOrMethod( KW_GET_FNAM, METH_CHAR    );
      sym.addAttributeOrMethod( KW_GET_FOFF, METH_INT     );
      sym.addAttributeOrMethod( KW_GET_1ST , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_GET_GRN , METH_INT     );
      sym.addAttributeOrMethod( KW_GET_HD_E, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_GET_IBNN, METH_INT     );
      sym.addAttributeOrMethod( KW_GET_IBQN, METH_INT     );
      sym.addAttributeOrMethod( KW_GET_ITER, METH_HANDLE  );
      sym.addAttributeOrMethod( KW_GET_LAST, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_GET_LNBI, METH_CHAR    );
      sym.addAttributeOrMethod( KW_GET_MOP , METH_POLY    ); // FWD extension, not real 4GL!
      sym.addAttributeOrMethod( KW_GET_MSG , METH_CHAR    );
      sym.addAttributeOrMethod( KW_GET_NEXT, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_GET_NODE, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_GET_NUM , METH_INT     );
      sym.addAttributeOrMethod( KW_GET_PAR , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_GET_PREV, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_GET_PROP, METH_CHAR    );
      sym.addAttributeOrMethod( KW_GET_PRT , METH_CHAR    );
      sym.addAttributeOrMethod( KW_GET_QNBI, METH_CHAR    );
      sym.addAttributeOrMethod( KW_GET_RED , METH_INT     );
      sym.addAttributeOrMethod( KW_GET_REL , METH_HANDLE  );
      sym.addAttributeOrMethod( KW_GET_R_R , METH_INT     );
      sym.addAttributeOrMethod( KW_GET_RGB , METH_INT     );
      sym.addAttributeOrMethod( KW_GET_SELW, METH_HANDLE  );
      sym.addAttributeOrMethod( KW_GET_SER , METH_LONGCHAR);
      sym.addAttributeOrMethod( KW_GET_SIG , METH_CHAR    );
      sym.addAttributeOrMethod( KW_GET_S_O , METH_CHAR    );
      sym.addAttributeOrMethod( KW_GET_SRCB, METH_HANDLE  );
      sym.addAttributeOrMethod( KW_GET_TBI , METH_CHAR    );
      sym.addAttributeOrMethod( KW_GET_TBNN, METH_CHAR    );
      sym.addAttributeOrMethod( KW_GET_TBQN, METH_CHAR    );
      sym.addAttributeOrMethod( KW_GET_TI  , METH_HANDLE  );
      sym.addAttributeOrMethod( KW_GET_THCH, METH_DEC     );
      sym.addAttributeOrMethod( KW_GET_THPX, METH_INT     );
      sym.addAttributeOrMethod( KW_GET_TOPB, METH_HANDLE  );
      sym.addAttributeOrMethod( KW_GET_TWCH, METH_DEC     );
      sym.addAttributeOrMethod( KW_GET_TWPX, METH_INT     );
      sym.addAttributeOrMethod( KW_GET_UBI , METH_CHAR    );
      sym.addAttributeOrMethod( KW_GET_VBI , METH_CHAR    );
      sym.addAttributeOrMethod( KW_GET_VBNN, METH_CHAR    );
      sym.addAttributeOrMethod( KW_GET_VBQN, METH_CHAR    );
      sym.addAttributeOrMethod( KW_GET_WAIT, METH_CHAR    );
      sym.addAttributeOrMethod( KW_GRAPHIC , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_GRD_F_H , ATTR_INT     );
      sym.addAttributeOrMethod( KW_GRD_F_V , ATTR_INT     );
      sym.addAttributeOrMethod( KW_GRD_SNAP, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_GRD_UHC , ATTR_DEC     );
      sym.addAttributeOrMethod( KW_GRD_UHP , ATTR_INT     );
      sym.addAttributeOrMethod( KW_GRD_UWC , ATTR_DEC     );
      sym.addAttributeOrMethod( KW_GRD_UWP , ATTR_INT     );
      sym.addAttributeOrMethod( KW_GRD_VIS , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_GROUP_BX, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_HANDLE  , ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_HANDLER , ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_HAS_LOBS, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_HAS_REC , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_HEIGHT_C, ATTR_DEC     );
      sym.addAttributeOrMethod( KW_HEIGHT_P, ATTR_INT     );
      sym.addAttributeOrMethod( KW_HELP    , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_HIDDEN  , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_HORIZ   , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_HTML_CHS, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_HTML_EOL, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_HTML_EOP, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_HTML_FRB, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_HTML_FRE, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_HTML_H_B, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_HTML_H_E, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_HTML_T_B, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_HTML_T_E, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_HWND    , ATTR_INT     );
      sym.addAttributeOrMethod( KW_ICON    , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_ICFPARM , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_IDEPHWND, ATTR_INT     );  // undocumented, found in customer code as INT (not HANDLE)
      sym.addAttributeOrMethod( KW_IDEWNTYP, ATTR_INT     );  // undocumented
      sym.addAttributeOrMethod( KW_IGN_CMOD, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_IMAGE   , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_IMG_DOWN, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_IMG_INS , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_IMG_UP  , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_IMM_DISP, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_IMP_NODE, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_IMP_PRNC, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_IDX_INFO, METH_CHAR    );
      sym.addAttributeOrMethod( KW_IN_HNDL , ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_INC_EX_I, METH_INT     );
      sym.addAttributeOrMethod( KW_INDEX   , ATTR_INT     );
      sym.addAttributeOrMethod( KW_INH_BGC , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_INH_FGC , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_INIT    , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_INIT_D_T, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_INITIATE, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_INNER_C , ATTR_INT     );
      sym.addAttributeOrMethod( KW_INNER_L , ATTR_INT     );
      sym.addAttributeOrMethod( KW_INPT_VAL, ATTR_POLY    );
      sym.addAttributeOrMethod( KW_INS_ATTR, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_INS_B4  , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_INS_BTAB, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_INS_FILE, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_INS_ROW , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_INS_STR , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_INS_TAB , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_INSERT  , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_INST_PRC, ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_INT_ENT , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_INTERVAL, ATTR_INT64   ); // FWD TIMER extension
      sym.addAttributeOrMethod( KW_INVOKE  , METH_VOID    );
      sym.addAttributeOrMethod( KW_IS_CLASS, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_IS_OPEN , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_IS_P_SET, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_IS_R_SEL, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_IS_SEL  , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_IS_XML  , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_ITEM_ROW, ATTR_INT     );
      sym.addAttributeOrMethod( KW_KEEP_CON, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_KEEP_ZOR, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_KEEP_SEC, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_KEY     , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_KEYS    , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_LABEL   , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_LAB_BGC , ATTR_INT     );
      sym.addAttributeOrMethod( KW_LAB_DC  , ATTR_INT     );
      sym.addAttributeOrMethod( KW_LAB_FGC , ATTR_INT     );
      sym.addAttributeOrMethod( KW_LAB_FONT, ATTR_INT     );
      sym.addAttributeOrMethod( KW_LABELS  , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_LANGUAGE, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_LARGE   , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_LAST_AR , ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_LAST_BAT, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_LAST_CH , ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_LAST_FRM, ATTR_CLASS  , "Progress.Windows.IForm" );
      sym.addAttributeOrMethod( KW_LAST_OBJ, ATTR_CLASS  , SymbolResolver.ROOT_OBJ_NAME );
      sym.addAttributeOrMethod( KW_LAST_OF , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_LAST_PRC, ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_LAST_SRV, ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_LAST_SS , ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_LAST_SOC, ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_LAST_TI , ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_LCHR_2NV, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_LENGTH  , ATTR_INT     );
      sym.addAttributeOrMethod( KW_LG_2_SM , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_LINE    , ATTR_INT     );
      sym.addAttributeOrMethod( KW_LISTINGS, ATTR_LOGICAL );  // undocumented, found in customer code
      sym.addAttributeOrMethod( KW_LIST_PNM, METH_CHAR    );
      sym.addAttributeOrMethod( KW_LIT_QSTN, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_LOAD    , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_LOAD_DMN, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_LOAD_ICO, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_LOAD_IMG, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_LOAD_I_D, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_LOAD_I_I, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_LOAD_I_U, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_LOAD_M_P, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_LOAD_S_I, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_LOC_C_N , ATTR_INT     );
      sym.addAttributeOrMethod( KW_LOC_HOST, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_LOC_L_N , ATTR_INT     );
      sym.addAttributeOrMethod( KW_LOC_NAME, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_LOC_PORT, ATTR_INT     );
      sym.addAttributeOrMethod( KW_LOC_P_ID, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_LOC_S_ID, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_LOC_TYPE, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_LOCKED  , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_LOCK_REG, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_LOOKUP  , METH_INT     );
      sym.addAttributeOrMethod( KW_LOG_A_EV, METH_CHAR    );
      sym.addAttributeOrMethod( KW_LOG_E_TS, ATTR_DATETIME_TZ );
      sym.addAttributeOrMethod( KW_LOG_EN_T, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_LOG_ID  , ATTR_INT     );
      sym.addAttributeOrMethod( KW_LOG_HOST, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_LOG_STAT, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_LOG_THRS, ATTR_INT     );
      sym.addAttributeOrMethod( KW_LOGF_NAM, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_LOGG_LEV, ATTR_INT     );
      sym.addAttributeOrMethod( KW_LOGOUT  , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_LST_PAIR, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_LIST_ITM, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_MAND    , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_MAN_HIGH, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_MAX_BTN , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_MAX_CHAR, ATTR_INT     );
      sym.addAttributeOrMethod( KW_MAX_D_G , ATTR_INT     );
      sym.addAttributeOrMethod( KW_MAX_HT  , ATTR_DEC     );
      sym.addAttributeOrMethod( KW_MAX_H_C , ATTR_DEC     );
      sym.addAttributeOrMethod( KW_MAX_H_P , ATTR_INT     );
      sym.addAttributeOrMethod( KW_MAX_VAL , ATTR_INT     );
      sym.addAttributeOrMethod( KW_MAX_WID , ATTR_DEC     );
      sym.addAttributeOrMethod( KW_MAX_W_C , ATTR_DEC     );
      sym.addAttributeOrMethod( KW_MAX_W_P , ATTR_INT     );
      sym.addAttributeOrMethod( KW_MD5_VAL , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_MPTR_2NV, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_MENU_BAR, ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_MENU_KEY, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_MENU_MOU, ATTR_INT     );
      sym.addAttributeOrMethod( KW_MERGE_BF, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_MERGE_CH, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_MERGE_RC, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_MIN_BTN , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_MIN_CWCH, ATTR_DEC     );
      sym.addAttributeOrMethod( KW_MIN_CWPX, ATTR_INT     );
      sym.addAttributeOrMethod( KW_MIN_H_C , ATTR_DEC     );
      sym.addAttributeOrMethod( KW_MIN_H_P , ATTR_INT     );
      sym.addAttributeOrMethod( KW_MIN_SCHM, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_MIN_VAL , ATTR_INT     );
      sym.addAttributeOrMethod( KW_MIN_W_C , ATTR_DEC     );
      sym.addAttributeOrMethod( KW_MIN_W_P , ATTR_INT     );
      sym.addAttributeOrMethod( KW_MSG_AREA, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_MSG_AFNT, ATTR_INT     );
      sym.addAttributeOrMethod( KW_MNEMON,   ATTR_INT     ); // FWD extension, not real 4GL!
      sym.addAttributeOrMethod( KW_MODIFIED, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_MOU_PTR , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_MOVABLE , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_MOV_A_T , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_MOV_B_T , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_MOV_COL , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_MOV_2_B , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_MOV_2_T , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_MOV_2EOF, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_MULT_CMP, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_MULTIPLE, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_MULTI_IN, ATTR_INT     );
      sym.addAttributeOrMethod( KW_MUST_UND, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_OWN_DOC , ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_OVERLAY , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_NAME    , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_NAMESP_P, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_NAMESP_U, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_NEEDS_AP, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_NEEDS_PR, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_NESTED  , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_NEW     , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_NEW_ROW , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_NEXT_COL, ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_NEXT_RID, ATTR_ROWID   );
      sym.addAttributeOrMethod( KW_NEXT_SIB, ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_NEXT_TAB, ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_NNMSP_SL, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_NO_COLS,  ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_NO_FOCUS, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_NO_CUR_V, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_NO_EM_SP, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_NO_SCH_M, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_NO_VALID, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_NODE_VAL, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_NODE_V2M, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_NODV_2LC, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_NORMALZE, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_NUM_BUFF, ATTR_INT     );
      sym.addAttributeOrMethod( KW_NUM_BUTT, ATTR_INT     );
      sym.addAttributeOrMethod( KW_NUM_CH_R, ATTR_INT     );
      sym.addAttributeOrMethod( KW_NUM_DROP, ATTR_INT     );
      sym.addAttributeOrMethod( KW_NUM_ENT , ATTR_INT     );
      sym.addAttributeOrMethod( KW_NUM_FLD , ATTR_INT     );
      sym.addAttributeOrMethod( KW_NUM_FMTS, ATTR_INT     );
      sym.addAttributeOrMethod( KW_NUM_HD_E, ATTR_INT     );
      sym.addAttributeOrMethod( KW_NUM_ITMS, ATTR_INT     );
      sym.addAttributeOrMethod( KW_NUM_ITER, ATTR_INT     );
      sym.addAttributeOrMethod( KW_NUM_LK_C, ATTR_INT     );
      sym.addAttributeOrMethod( KW_NUM_LNS,  ATTR_INT     );
      sym.addAttributeOrMethod( KW_NUM_LOGF, ATTR_INT     );
      sym.addAttributeOrMethod( KW_NUM_MSG , ATTR_INT     );
      sym.addAttributeOrMethod( KW_NUM_CHLN, ATTR_INT     );
      sym.addAttributeOrMethod( KW_NUM_COL , ATTR_INT     );
      sym.addAttributeOrMethod( KW_NUM_PARM, ATTR_INT     );
      sym.addAttributeOrMethod( KW_NUM_REF , ATTR_INT     );
      sym.addAttributeOrMethod( KW_NUM_REL , ATTR_INT     );
      sym.addAttributeOrMethod( KW_NUM_REPL, ATTR_INT     );
      sym.addAttributeOrMethod( KW_NUM_RES , ATTR_INT     );
      sym.addAttributeOrMethod( KW_NUM_SR  , ATTR_INT     );
      sym.addAttributeOrMethod( KW_NUM_SW  , ATTR_INT     );
      sym.addAttributeOrMethod( KW_NUM_TABS, ATTR_INT     );
      sym.addAttributeOrMethod( KW_NUM_2RTN, ATTR_INT     );
      sym.addAttributeOrMethod( KW_NUM_V_C , ATTR_INT     );
      sym.addAttributeOrMethod( KW_NUM_D_P , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_NUM_FMT , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_NUM_SEP , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_NUM_SRCB, ATTR_INT     );
      sym.addAttributeOrMethod( KW_NUM_TOPB, ATTR_INT     );
      sym.addAttributeOrMethod( KW_OLE_INVL, ATTR_INT     );
      sym.addAttributeOrMethod( KW_OLE_NAML, ATTR_INT     );
      sym.addAttributeOrMethod( KW_ON_FR_B , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_ORG_HAND, ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_ORG_ROID, ATTR_ROWID   );
      sym.addAttributeOrMethod( KW_OWNER   , ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_PAGE_B  , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_PAGE_T  , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_PAR_BUFF, ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_PAR_IREL, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_PAR_REL , ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_PARENT  , ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_PARM    , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_PARSE_ST, ATTR_INT     );
      sym.addAttributeOrMethod( KW_PASSWD_F, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_PATHNAME, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_PBE_H_AL, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_PBE_KEYR, ATTR_INT     );
      sym.addAttributeOrMethod( KW_PREF_DS , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_PRS_PROC, ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_PRS_C_D , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_PERSIST , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_PFCOLOR , ATTR_INT     );
      sym.addAttributeOrMethod( KW_PIX_COL , ATTR_INT     );
      sym.addAttributeOrMethod( KW_PIX_ROW , ATTR_INT     );
      sym.addAttributeOrMethod( KW_POP_MENU, ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_POP_ONLY, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_POS     , ATTR_INT     );
      sym.addAttributeOrMethod( KW_PREP_STR, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_PREPARED, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_PREV_COL, ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_PREV_SIB, ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_PREV_T_I, ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_PRE_LBL,  ATTR_CHAR    ); // FWD extension, not real 4GL!
      sym.addAttributeOrMethod( KW_PRIMARY , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_PROFILNG, ATTR_LOGICAL ); // undocumented, found in customer code
      sym.addAttributeOrMethod( KW_PROG_SRC, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_PRT_C_H , ATTR_INT     );
      sym.addAttributeOrMethod( KW_PRT_HDC , ATTR_INT     );
      sym.addAttributeOrMethod( KW_PRT_NAME, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_PRT_PORT, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_PRIV_DAT, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_PROCNAME, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_PROXY   , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_PROX_PWD, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_PROX_UID, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_PUB_ID  , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_PUB_EVTS, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_QUERY   , ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_QRY_CLOS, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_QRY_OFF , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_QRY_OPEN, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_QRY_PREP, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_QUIT    , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_RADIO_B , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_RAW_TRAN, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_READ    , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_READ_FIL, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_READ_JSN, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_READ_ONL, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_READ_XML, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_READ_XSC, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_RECID   , ATTR_RECID   );
      sym.addAttributeOrMethod( KW_REC_LEN , ATTR_INT     );
      sym.addAttributeOrMethod( KW_REFRESH , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_REFR_A_P, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_REFRABLE, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_REG_DMN , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_REJ_CHGS, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_REJ_RCHG, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_REJECTED, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_REL_FLDS, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_RELS_ACT, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_REM_EVTP, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_REM_HOST, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_REM_PORT, ATTR_INT     );
      sym.addAttributeOrMethod( KW_REMOTE  , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_REM_ATTR, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_REM_CHLD, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_REM_SUP , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_REPLACE , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_REP_CHLD, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_REP_STXT, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_REPOS   , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_REPOS_B , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_REPOS_F , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_REPOS_2R, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_REPOS_2I, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_RESET   , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_RESIZE  , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_RESIZABL, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_REST_RID, ATTR_ROWID   );
      sym.addAttributeOrMethod( KW_RET_INS , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_RET_SHAP, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_RET_VAL , ATTR_POLY    );
      sym.addAttributeOrMethod( KW_RET_VDT , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_ROLES   , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_ROUNDED , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_ROW     , ATTR_DEC     );
      sym.addAttributeOrMethod( KW_ROW_H_C , ATTR_DEC     );
      sym.addAttributeOrMethod( KW_ROW_H_P , ATTR_INT     );
      sym.addAttributeOrMethod( KW_ROWID   , ATTR_ROWID   );
      sym.addAttributeOrMethod( KW_ROW_MARK, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_ROW_RESZ, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_ROW_STAT, ATTR_INT     );
      sym.addAttributeOrMethod( KW_SAVE    , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_SAVE_FIL, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_SAVE_RCH, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_SAVE_WST, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_SAX_PARF, METH_VOID    );
      sym.addAttributeOrMethod( KW_SAX_PARN, METH_VOID    );
      sym.addAttributeOrMethod( KW_SAX_PARS, METH_VOID    );
      sym.addAttributeOrMethod( KW_SCH_CHG , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_SCH_LOC , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_SCH_MARS, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_SCH_PATH, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_SCRN_VAL, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_SCRN_LNS, ATTR_DEC     );
      sym.addAttributeOrMethod( KW_SCROLLBA, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_SCR_2CR , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_SCR_2ITM, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_SCR_2SR , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_SCROLLBL, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_SCROLL_H, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_SCROLL_V, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_SEAL    , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_SEAL_TST, ATTR_DATETIME_TZ );
      sym.addAttributeOrMethod( KW_SEARCH  , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_SELECTBL, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_SEL_ALL , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_SEL_END , ATTR_INT     );
      sym.addAttributeOrMethod( KW_SEL_FOCR, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_SEL_NEXT, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_SEL_PREV, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_SEL_ROW , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_SEL_STRT, ATTR_INT     );
      sym.addAttributeOrMethod( KW_SEL_TXT , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_SELECTED, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_SENSITIV, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_SEPS    , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_SEP_FGC , ATTR_INT     );
      sym.addAttributeOrMethod( KW_SERIALZH, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_SERIALZN, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_SERIALZR, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_SERVER  , ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_SESS_END, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_SESSN_ID, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_SRV_C_B , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_SRV_C_BR, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_SRV_C_C , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_SRV_C_I , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_SRV_OP_M, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_SET_ACTR, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_SET_ACTX, METH_CHAR    );
      sym.addAttributeOrMethod( KW_SET_ATTR, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_SET_A_N , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_SET_BLUE, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_SET_BRK , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_SET_BUF , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_SET_CB_P, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_SET_CBAC, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_SET_CLNT, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_SET_COMM, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_SET_C_P , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_SET_DYN , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_SET_GRN , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_SET_ISRC, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_SET_M_UN, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_SET_N_F , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_SET_NODE, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_SET_ODST, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_SET_PARM, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_SET_PROP, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_SET_RED , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_SET_RGB , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_SET_RRP , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_SET_RPOS, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_SET_ROLL, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_SET_SEL , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_SET_SERD, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_SET_S_AR, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_SET_S_O , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_SET_WAIT, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_SHOW_ITB, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_SIDE_L  , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_SIDE_L_H, ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_SKIP_D_R, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_SMAL_ICO, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_SMAL_TTL, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_SOAP_F_A, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_SOAP_F_C, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_SOAP_F_D, ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_SOAP_F_S, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_SORT    , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_SORT_ASC, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_SORT_NUM, ATTR_INT     );
      sym.addAttributeOrMethod( KW_SSL_SRVN, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_STANDALN, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_STATUS_A, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_START_DC, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_START_EL, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_STAT_DET, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_STAT_A_F, ATTR_INT     );
      sym.addAttributeOrMethod( KW_STREAM  , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_START,    METH_VOID    ); // FWD TIMER extension
      sym.addAttributeOrMethod( KW_STOP    , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_STOP_PRS, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_STOPPED , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_STRICT  , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_STR_VAL , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_STUP_PAR, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_ST_2_FIT, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_SUBTYPE , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_SUP_NS_P, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_SUP_PROC, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_SUP_WARN, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_SYM_EN_A, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_SYM_EN_I, ATTR_RAW     );
      sym.addAttributeOrMethod( KW_SYM_EN_K, ATTR_RAW     );
      sym.addAttributeOrMethod( KW_SYM_SUPP, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_SYNCHRON, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_SYS_A_B , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_SYS_ID  , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_3D      , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_TAB_POS , ATTR_INT     );
      sym.addAttributeOrMethod( KW_TAB_STOP, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_TABLE   , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_TAB_CRCL, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_TAB_HAND, ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_TAB_LIST, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_TAB_NUM , ATTR_INT     );
      sym.addAttributeOrMethod( KW_TEMP_DIR, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_TRC_FILT, ATTR_CHAR    ); // undocumented, found in customer code
      sym.addAttributeOrMethod( KW_TRAC_CHG, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_TXT_SEL , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_TIC_MARK, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_TITLE   , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_TIMEZONE, ATTR_INT     );
      sym.addAttributeOrMethod( KW_TIME_SRC, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_TITL_BGC, ATTR_INT     );
      sym.addAttributeOrMethod( KW_TITL_DC , ATTR_INT     );
      sym.addAttributeOrMethod( KW_TITL_FGC, ATTR_INT     );
      sym.addAttributeOrMethod( KW_TITL_FON, ATTR_INT     );
      sym.addAttributeOrMethod( KW_TT_PREP , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_TOGGL_BX, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_TOOLTIP , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_TOOLTIPS, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_TOP_NAVQ, ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_TOP_ONLY, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_TRANS   , ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_TRAN_I_P, ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_TRANSPAR, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_TYPE    , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_UNDO    , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_UNIQ_ID , ATTR_INT     );
      sym.addAttributeOrMethod( KW_UNIQ_MAT, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_UPD_ATTR, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_URL     , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_URL_DECO, METH_CHAR    );
      sym.addAttributeOrMethod( KW_URL_ENCO, METH_CHAR    );
      sym.addAttributeOrMethod( KW_URL_PW  , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_URL_UID , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_USR_DATA, METH_LOGICAL ); // undocumented, found in customer code
      sym.addAttributeOrMethod( KW_USER_ID , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_UTC_OFF , ATTR_INT     );
      sym.addAttributeOrMethod( KW_V6DISP  , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_VALUE   , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_VAL_ENAB, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_VAL_EXPR, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_VAL_MSG , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_VAL_SEAL, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_VAL_XML , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_VALIDATE, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_VERSION , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_VIEW_AS , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_VIEW_FCR, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_VIRT_HC , ATTR_DEC     );
      sym.addAttributeOrMethod( KW_VIRT_HP , ATTR_INT     );
      sym.addAttributeOrMethod( KW_VIRT_WC , ATTR_DEC     );
      sym.addAttributeOrMethod( KW_VIRT_WP , ATTR_INT     );
      sym.addAttributeOrMethod( KW_VISIBLE , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_WHERE_ST, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_WID_ENT , ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_WID_ID  , ATTR_INT     );
      sym.addAttributeOrMethod( KW_WID_LEAV, ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_WIDTH_C , ATTR_DEC     );
      sym.addAttributeOrMethod( KW_WIDTH_P , ATTR_INT     );
      sym.addAttributeOrMethod( KW_WIN_STAT, ATTR_INT     );
      sym.addAttributeOrMethod( KW_WIN_SYS , ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_WINDOW  , ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_WA_H_P  , ATTR_INT     );
      sym.addAttributeOrMethod( KW_WA_W_P  , ATTR_INT     );
      sym.addAttributeOrMethod( KW_WA_X    , ATTR_INT     );
      sym.addAttributeOrMethod( KW_WA_Y    , ATTR_INT     );
      sym.addAttributeOrMethod( KW_WARN    , ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_WORD_WRP, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_WRITE   , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_WR_CDATA, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_WR_CHARS, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_WR_CMNT , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_WR_DATA , METH_LOGICAL ); // undocumented, found in customer code
      sym.addAttributeOrMethod( KW_WR_D_ELM, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_WR_E_ELM, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_WR_ENT_R, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_WR_EXDTD, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_WR_FRAGM, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_WR_JSON , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_WR_MSG  , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_WR_PRINS, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_WR_STAT , ATTR_INT     );
      sym.addAttributeOrMethod( KW_WR_XML  , METH_LOGICAL );
      sym.addAttributeOrMethod( KW_WR_XMLSC, METH_LOGICAL );
      sym.addAttributeOrMethod( KW_X       , ATTR_INT     );
      sym.addAttributeOrMethod( KW_XML_DTYP, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_XML_NNAM, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_XML_NTYP, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_XML_SCHP, ATTR_CHAR    );
      sym.addAttributeOrMethod( KW_XML_SNSP, ATTR_LOGICAL );
      sym.addAttributeOrMethod( KW_X_DOC   , ATTR_HANDLE  );
      sym.addAttributeOrMethod( KW_Y       , ATTR_INT     );
      sym.addAttributeOrMethod( KW_YEAR_OFF, ATTR_INT     );
   }
   
   /**
    * Loads the <code>SymbolResolver</code> with Progress 4GL built-in
    * variables and any built-in functions which DO NOT take parameters. This 
    * is called during construction to ensure that the functions are 
    * available before parsing begins.
    * <p>
    * Built-in functions that have no parameters are treated as variables.    
    * This is done because this is an exact syntactic and semantic match
    * for such constructions.  The fact that the Progress language names
    * such items functions has no meaning.  For this reason, this function is 
    * used to handle such built-ins.
    * <p>
    * Each variable is added with the full name (even if the keyword supports
    * abbreviations) because the {@link #lvalue} method will detect when
    * a keyword is being used and it will use the keyword dictionary to
    * handle the abbreviation support.  This lookup will convert any
    * abbreviated text to the full text before lookup in the variable
    * dictionary.
    * <p>
    * All variables are added with the token type associated with their
    * respective return value.
    * <p>
    * For built-in functions that take parameters, see
    * {@link #initializeFunctionDictionary}.
    * <p>
    * This method also enables the matching of all valid system handles:
    * <p>
    * <pre>
    * ACTIVE_WINDOW
    * CLIPBOARD
    * CODEBASE-LOCATOR
    * COLOR-TABLE
    * COM-SELF
    * COMPILER
    * CURRENT-WINDOW
    * DEBUGGER
    * DEFAULT-WINDOW
    * ERROR-STATUS
    * FILE-INFO
    * FOCUS
    * FONT-TABLE
    * LAST-EVENT
    * RCODE-INFO
    * SELF
    * SESSION
    * SOURCE-PROCEDURE
    * TARGET-PROCEDURE
    * THIS-PROCEDURE
    * WEB-CONTEXT
    * </pre>
    * <p> 
    * Note that there are some built-in functions that can act like variables
    * OR as a normal parenthesized function.  In other words, for the
    * following list, the parenthesis are optional:
    * <p>
    * <ul>
    *    <li>ETIME
    *    <li>FRAME-COL
    *    <li>FRAME-DOWN
    *    <li>FRAME-LINE
    *    <li>FRAME-ROW
    *    <li>GUID
    *    <li>LINE-COUNTER
    *    <li>PAGE-NUMBER
    *    <li>PAGE-SIZE
    *    <li>SUPER (handled in super_function rule instead of here)
    *    <li>TRANS or TRANSACTION
    *    <li>USERID (and it's synonym USER which is not an abbreviation)
    * </ul>
    * <p>
    * Such variables/functions are included in both this rule and in
    * <code>initializeFunctionDictionary</code>.  The token type for such
    * constructs is set to the function version <code>FUNC_</code> instead of
    * the variable version <code>VAR_</code>.
    */
   private void initializeVariableDictionary()
   {
      sym.addGlobalVariable("_control"           , VAR_CHAR         );
      sym.addGlobalVariable("_msg"               , FUNC_INT         );
      sym.addGlobalVariable("_pcontrol"          , VAR_CHAR         );
      sym.addGlobalVariable("_serial-num"        , VAR_INT          );
      sym.addGlobalVariable("active-form"        , VAR_CLASS        , "Progress.Windows.IForm" );
      sym.addGlobalVariable("active-window"      , SYS_HANDLE       );
      sym.addGlobalVariable("audit-control"      , SYS_HANDLE       );
      sym.addGlobalVariable("audit-policy"       , SYS_HANDLE       );
      sym.addGlobalVariable("clipboard"          , SYS_HANDLE       );
      sym.addGlobalVariable("codebase-locator"   , SYS_HANDLE       );
      sym.addGlobalVariable("color-table"        , SYS_HANDLE       );
      sym.addGlobalVariable("com-self"           , SYS_HANDLE       );
      sym.addGlobalVariable("compiler"           , SYS_HANDLE       );
      sym.addGlobalVariable("current-language"   , VAR_CHAR         );
      sym.addGlobalVariable("current-window"     , SYS_HANDLE       );
      sym.addGlobalVariable("dataservers"        , VAR_CHAR         );
      sym.addGlobalVariable("dbname"             , VAR_CHAR         );
      sym.addGlobalVariable("debugger"           , SYS_HANDLE       );
      sym.addGlobalVariable("default-window"     , SYS_HANDLE       );
      sym.addGlobalVariable("dslog-manager"      , SYS_HANDLE       );
      sym.addGlobalVariable("error-status"       , SYS_HANDLE       );
      sym.addGlobalVariable("etime"              , FUNC_INT         );
      sym.addGlobalVariable("exclusive-lock"     , VAR_INT          ); // used as a symbolic constant in handle-based method calls as a valid integer expression
      sym.addGlobalVariable("file-information"   , SYS_HANDLE       );
      sym.addGlobalVariable("focus"              , SYS_HANDLE       );
      sym.addGlobalVariable("font-table"         , SYS_HANDLE       );
      sym.addGlobalVariable("frame-col"          , FUNC_DEC         );
      sym.addGlobalVariable("frame-db"           , VAR_CHAR         );
      sym.addGlobalVariable("frame-down"         , FUNC_INT         );
      sym.addGlobalVariable("frame-field"        , VAR_CHAR         );
      sym.addGlobalVariable("frame-file"         , VAR_CHAR         );
      sym.addGlobalVariable("frame-index"        , VAR_INT          );
      sym.addGlobalVariable("frame-line"         , FUNC_INT         );
      sym.addGlobalVariable("frame-name"         , VAR_CHAR         );
      sym.addGlobalVariable("frame-row"          , FUNC_DEC         );
      sym.addGlobalVariable("frame-value"        , VAR_CHAR         );
      sym.addGlobalVariable("gateways"           , VAR_CHAR         );
      sym.addGlobalVariable("generate-pbe-salt"  , VAR_RAW          );
      sym.addGlobalVariable("generate-random-key", VAR_RAW          );
      sym.addGlobalVariable("generate-uuid"      , VAR_RAW          );
      sym.addGlobalVariable("get-codepages"      , VAR_CHAR         );
      sym.addGlobalVariable("go-pending"         , VAR_LOGICAL      );
      sym.addGlobalVariable("guid"               , FUNC_CHAR        ); // built-in function with optional parens, this is the undocumented form
      sym.addGlobalVariable("is-attr-space"      , VAR_LOGICAL      );
      sym.addGlobalVariable("lastkey"            , VAR_INT          );
      sym.addGlobalVariable("last-key"           , VAR_INT          );
      sym.addGlobalVariable("last-event"         , SYS_HANDLE       );
      sym.addGlobalVariable("line-counter"       , FUNC_INT         );
      sym.addGlobalVariable("log-manager"        , SYS_HANDLE       );
      sym.addGlobalVariable("message-lines"      , VAR_INT          );
      sym.addGlobalVariable("mtime"              , FUNC_INT         ); // entry here with a FUNC_ type suggests this can have optional parens, is that correct?
      sym.addGlobalVariable("no-lock"            , VAR_INT          ); // used as a symbolic constant in handle-based method calls as a valid integer expression
      sym.addGlobalVariable("no-wait"            , VAR_INT          ); // used as a symbolic constant in handle-based method calls as a valid integer expression
      sym.addGlobalVariable("now"                , FUNC_DATETIME_TZ ); // entry here with a FUNC_ type suggests this can have optional parens, is that correct?
      sym.addGlobalVariable("num-aliases"        , VAR_INT          );
      sym.addGlobalVariable("num-dbs"            , VAR_INT          );
      sym.addGlobalVariable("opsys"              , VAR_CHAR         );
      sym.addGlobalVariable("os-drives"          , VAR_CHAR         );
      sym.addGlobalVariable("os-error"           , VAR_INT          );
      sym.addGlobalVariable("page-number"        , FUNC_INT         );
      sym.addGlobalVariable("page-size"          , FUNC_INT         );
      sym.addGlobalVariable("proc-handle"        , VAR_INT          );
      sym.addGlobalVariable("proc-status"        , VAR_INT          );
      sym.addGlobalVariable("profiler"           , SYS_HANDLE       );
      sym.addGlobalVariable("progress"           , VAR_CHAR         );
      sym.addGlobalVariable("promsgs"            , VAR_CHAR         );
      sym.addGlobalVariable("propath"            , VAR_CHAR         );
      sym.addGlobalVariable("proversion"         , VAR_CHAR         );
      sym.addGlobalVariable("rcode-information"  , SYS_HANDLE       );
      sym.addGlobalVariable("retry"              , FUNC_LOGICAL     ); // entry here with a FUNC_ type suggests this can have optional parens, is that correct?
      sym.addGlobalVariable("return-value"       , VAR_CHAR         );
      sym.addGlobalVariable("screen-lines"       , VAR_INT          );
      sym.addGlobalVariable("security-policy"    , SYS_HANDLE       );
      sym.addGlobalVariable("self"               , SYS_HANDLE       );
      sym.addGlobalVariable("session"            , SYS_HANDLE       );
      sym.addGlobalVariable("share-lock"         , VAR_INT          ); // used as a symbolic constant in handle-based method calls as a valid integer expression
      sym.addGlobalVariable("source-procedure"   , SYS_HANDLE       );
      sym.addGlobalVariable("target-procedure"   , SYS_HANDLE       );
      sym.addGlobalVariable("terminal"           , VAR_CHAR         );
      sym.addGlobalVariable("this-procedure"     , SYS_HANDLE       );
      sym.addGlobalVariable("time"               , FUNC_INT         ); // entry here with a FUNC_ type suggests this can have optional parens, is that correct?
      sym.addGlobalVariable("timezone"           , FUNC_INT         ); // entry here with a FUNC_ type suggests this can have optional parens, is that correct?
      sym.addGlobalVariable("today"              , FUNC_DATE        ); // entry here with a FUNC_ type suggests this can have optional parens, is that correct?
      sym.addGlobalVariable("trans"              , FUNC_LOGICAL     );
      sym.addGlobalVariable("transaction"        , FUNC_LOGICAL     );
      sym.addGlobalVariable("user"               , FUNC_CHAR        );
      sym.addGlobalVariable("userid"             , FUNC_CHAR        );
      sym.addGlobalVariable("web-context"        , SYS_HANDLE       );
   }
   
   /**
    * Loads the <code>SymbolResolver</code> with Progress 4GL internal
    * procedures that are well known. This is called during construction 
    * to ensure that the functions are available before parsing begins.
    * <p>
    * These procedures are normally implemented by 4GL code in WebSpeed
    * but their definitions are being placed here so that code can parse
    * without any PSC 4GL code. Most of these are old WebSpeed 1.0 interfaces
    * that are deprecated but still supported for backwards compatibility. 
    */
   private void initializeProcedureDictionary()
   {
      if (webspeed)
      {
         sym.addProcedure("adm-timing-out"       , WEB_PROC );
         sym.addProcedure("AsciiToHtml"          , WEB_PROC ); // html-encode()
         sym.addProcedure("CookieDate"           , WEB_PROC ); // format-cookie()
         sym.addProcedure("DeleteCookie"         , WEB_PROC ); // delete-cookie()
         sym.addProcedure("dispatch"             , WEB_PROC );
         sym.addProcedure("get-transaction-state", WEB_PROC );
         sym.addProcedure("getAttribute"         , WEB_PROC );
         sym.addProcedure("GetCGI"               , WEB_PROC ); // get-cgi()
         sym.addProcedure("GetCookie"            , WEB_PROC ); // get-cookie()
         sym.addProcedure("GetField"             , WEB_PROC ); // get-field()
         sym.addProcedure("HtmlError"            , WEB_PROC );
         sym.addProcedure("init-session"         , WEB_PROC );
         sym.addProcedure("OutputContentType"    , WEB_PROC ); // output-content-type()
         sym.addProcedure("outputHeader"         , WEB_PROC );
         sym.addProcedure("outputHeaders"        , WEB_PROC ); // same as outputHeader
         sym.addProcedure("OutputHttpHeader"     , WEB_PROC ); // output-http-header()
         sym.addProcedure("process-web-request"  , WEB_PROC );
         sym.addProcedure("run-web-object"       , WEB_PROC );
         sym.addProcedure("set-attribute-list"   , WEB_PROC );
         sym.addProcedure("set-transaction-state", WEB_PROC );
         sym.addProcedure("SetCookie"            , WEB_PROC ); // set-cookie()
         sym.addProcedure("UrlDecode"            , WEB_PROC ); // url-decode()
         sym.addProcedure("UrlEncode"            , WEB_PROC ); // url-encode()
         sym.addProcedure("web.output"           , WEB_PROC );
      }
   }
   
   /**
    * Loads the <code>SymbolResolver</code> with Progress 4GL internal
    * streams that are well known. This is called during construction 
    * to ensure that the streams are available before parsing begins.
    * <p>
    * These streams are normally implemented by 4GL code in WebSpeed
    * but their definitions are being placed here so that code can parse
    * without any PSC 4GL code. 
    */
   private void initializeStreamDictionary()
   {
      if (webspeed)
      {
         sym.addStream("WebStream", STREAM);
      }
   }
   
   /**
    * Common routine to rewrite the data, field type based on a requirement
    * for additional abbreviation support (beyond that which can normally be
    * used in the rest of the Progress language.
    * <p>
    * The 2nd lookahead token is read for its text. If it is one of the
    * valid (expected) keywords, then this method returns immediately.
    * Otherwise, a check is made for any of the special abbreviations
    * that follow:
    * <pre>
    *    Data Type Keyword        Actual Minimum Abbrev      Lexer Abbrev
    *    -----------------        ---------------------      -------------
    *    KW_BLOB                  n/a                        n/a
    *    KW_CHAR                  c                          char
    *    KW_CLASS                 n/a                        n/a
    *    KW_CLOB                  n/a                        n/a
    *    KW_COM_HNDL              n/a                        n/a
    *    KW_DATE                  da                         n/a
    *    KW_DATETIME              n/a                        n/a
    *    KW_DATE_TZ               n/a                        n/a
    *    KW_DEC                   de                         dec
    *    KW_HANDLE                handle                     n/a
    *    KW_INT                   i                          int
    *    KW_INT64                 n/a                        n/a
    *    KW_LOGICAL               l                          n/a
    *    KW_LONGCHAR              n/a                        n/a
    *    KW_MEMPTR                m                          mem
    *    KW_RAW                   ra                         n/a
    *    KW_RECID                 re                         n/a
    *    KW_ROWID                 row                        n/a
    *    KW_WID_HAND              widg                       widget-h
    *    The following are supported in native interfaces:
    *    KW_BYTE                  n/a                        n/a
    *    KW_DOUBLE                n/a                        n/a
    *    KW_FLOAT                 n/a                        n/a
    *    KW_LONG                  n/a                        n/a
    *    KW_SHORT                 n/a                        n/a
    *    KW_UNS_SHRT              n/a                        n/a
    * </pre>
    * <p>
    * If one of these (from the actual column above) is matched, the token
    * type of the 2nd lookahead token is rewritten with the associated
    * data/field type keyword.  Otherwise an exception is thrown.
    * <p>
    * This method exists simply to centralize the processing that is used
    * in both {@link #as_clause} and {@link #as_field_clause}.  NOTE that
    * callers MUST ensure that any necessary pull methods (such as 
    * {@link #reserved_or_symbol}) are included to properly generate the
    * predication logic that will allow the full set of possibilities to
    * actually be encountered.
    *
    * @param   lookahead
    *          Number of tokens to look ahead.
    */
   private void specialDataTypeKeywordAbbreviations(int lookahead)
   throws TokenStreamException
   {
      int sec = LA(lookahead);
      
      // override token type for 2nd token IF it already is NOT a match
      // to one of our required data type keywords
      if (sec != KW_BLOB     &&
          sec != KW_CHAR     &&
          sec != KW_CLASS    &&
          sec != KW_CLOB     &&
          sec != KW_COM_HNDL &&
          sec != KW_DATE     &&
          sec != KW_DATETIME &&
          sec != KW_DATE_TZ  &&
          sec != KW_DEC      &&
          sec != KW_INT      &&
          sec != KW_INT64    &&
          sec != KW_LOGICAL  &&
          sec != KW_LONGCHAR &&
          sec != KW_RECID    &&
          sec != KW_ROWID    &&
          sec != KW_HANDLE   &&
          sec != KW_WID_HAND &&
          sec != KW_RAW      &&
          sec != KW_MEMPTR   &&
          sec != KW_BYTE     &&
          sec != KW_DOUBLE   &&
          sec != KW_FLOAT    &&
          sec != KW_LONG     &&
          sec != KW_SHORT    &&
          sec != KW_UNS_SHRT)
      {
         String text    = LT(lookahead).getText().toLowerCase();
         int    newtype = -1;
         
         if (text.equals("c") || text.equals("ch") || text.equals("cha"))
         {
            newtype = KW_CHAR; 
         }
         else if (text.equals("da") || text.equals("dat"))
         {
            newtype = KW_DATE; 
         }
         else if (text.equals("de"))
         {
            newtype = KW_DEC; 
         }            
         else if (text.equals("i") || text.equals("in"))
         {
            newtype = KW_INT; 
         }
         else if (text.equals("l")     || text.equals("lo")   ||
                  text.equals("log")   || text.equals("logi") ||
                  text.equals("logic") || text.equals("logica"))
         {
            newtype = KW_LOGICAL; 
         }
         else if (text.equals("m") || text.equals("me"))
         {
            newtype = KW_MEMPTR; 
         }
         else if (text.equals("ra"))
         {
            newtype = KW_RAW; 
         }
         else if (text.equals("re") || text.equals("rec") ||
                  text.equals("reci"))
         {
            newtype = KW_RECID; 
         }
         else if (text.equals("row") || text.equals("rowi"))
         {
            newtype = KW_ROWID; 
         }
         else if (text.equals("widg")   || text.equals("widge") || 
                  text.equals("widget") || text.equals("widget-"))
         {
            newtype = KW_WID_HAND; 
         }            
         
         // at this point, there should always be a match in valid
         // Progress 4GL source so we throw an exception if this is
         // not the case
         if (newtype != -1)
         {
            LT(lookahead).setType(newtype);
         }
      }
   }
   
   /**
    * Check if this token is a user-defined function call. The method checks whether the first
    * token is marked as type <code>KW_CLOSE</code> or <code>KW_COMPILE</code> with next token
    * being a <code>LPARENS</code>: in this case, transform it to <code>SYMBOL</code>, as it
    * will be a function or OO method call.
    *
    * @param    tok1
    *           First token, typically LT(1).
    * @param    tok2
    *           Second token, typically LT(2).
    *
    * @return    <code>true</code> if the type of this token was changed.
    */
   private boolean isUserDefFunctionCall(Token tok1, Token tok2)
   {
      if (tok2.getType() != LPARENS)
      {
        return false;
      }

      // CLOSE or COMPILE is not a reserved keyword, thus can be used in a function name;
      // we need to disambiguate between a close or compile function call and close or
      // compile statement.

      // this needs to be disambiguated done here, otherwise it wouldn't be converted
      // properly.
      if (tok1.getType() == KW_CLOSE ||
          tok1.getType() == KW_COMPILE)
      {
         tok1.setType(SYMBOL);
         
         return true;
      }
      
      return false;
   }
   
   /**
    * Simple helper to return the given node if it is a field type or if it
    * is an expression node, its first child will be returned if that child
    * is a field type.
    *
    * @param    possible
    *           The given node to test, cannot be <code>null</code>.
    *
    * @return   The field node found or <code>null</code> if no field node
    *           was found.
    */
   private Aast findFieldNode(Aast possible)
   {
      int ttype = possible.getType();
      
      // dereference first child if this is an expression node
      if (ttype == EXPRESSION)
      {
         possible = (Aast) possible.getFirstChild();
         ttype    = possible.getType();
      }      
      
      // is the node a field type
      if (ttype > BEGIN_FIELDTYPES && ttype < END_FIELDTYPES)
      {
         return possible;
      }
      
      return null;
   }
   
   /**
    * Check if the USE-DICT-EXPS frame option is set for this frame.  If is not active, then also
    * remove all validation and help options pulled from the schema by this statement.
    *
    * @param    frame
    *           The frame name.
    * @param    framePhrase
    *           The FRAME_PHRASE AST.
    * @param    stmt
    *           The current UI statement.
    */
   private void setUseDictExps(String frame, Aast framePhrase, Aast stmt)
   {
      // clear the "schema-processed" annotation for all fields
      Set<Aast> fields = sym.getFrameFields(frame);
      
      if (fields != null)
      {
         for (Aast fld : fields)
         {
            fld.removeAnnotation("schema-processed");
         }
      }

      boolean prevUseDictExps = sym.isUseDictExps(frame);
      
      if (framePhrase != null && framePhrase.downPath("KW_USE_DCT"))
      {
         sym.markUseDictExps(frame);
      }

      // if USE-DICT-EXPS is not set for this frame and this statement is not one which pulls the
      // schema validation automatically, then remove all KW_VALIDATE[schema_validation=true] and
      // KW_HELP[schema_help=true] nodes from this statement.
      
      int type = stmt.getType();
      
      if (!sym.isUseDictExps(frame) &&
          !(type == KW_ENABLE ||
            type == KW_SET    ||
            type == KW_UPDATE ||
            type == KW_PRMT_FOR))
      {
         // fixup the parents so nodes can be removed
         ((ProgressAst) stmt).fixupParent(null);
         
         List<Aast> remove = new ArrayList<>();
         Iterator<Aast> iter = stmt.iterator();
         while (iter.hasNext())
         {
            Aast ref = iter.next();
            
            int rtype = ref.getType();
            
            if (rtype == KW_VALIDATE                  &&
                ref.isAnnotation("schema_validation") &&
                (Boolean) ref.getAnnotation("schema_validation"))
            {
               remove.add(ref);
            }
            
            // schema-level help is allowed to remain only if we are an UI statement which
            // enables the field.  otherwise, both the KW_HELP node and the field-level "help"
            // annotation will be removed.
            if (rtype == KW_HELP                &&
                ref.isAnnotation("schema_help") &&
                (Boolean) ref.getAnnotation("schema_help"))
            {
               remove.add(ref);
            }
            
            if (rtype > BEGIN_FIELDTYPES && rtype < END_FIELDTYPES)
            {
               ref.removeAnnotation("help");
            }
         }
         
         Set<Aast> formatPhrases = new HashSet<>();

         for (Aast ref : remove)
         {
            formatPhrases.add(ref.getParent());

            ref.remove();
         }
         
         for (Aast ref : formatPhrases)
         {
            if (ref.getNumImmediateChildren() == 0)
            {
               ref.remove();
            }
         }
      }
      
      // if USE-DICT-EXPS is set before this statement, then remove the VALEXP and HELP nodes for
      // all field widgets
      if (prevUseDictExps)
      {
         // fixup the parents so nodes can be removed
         ((ProgressAst) stmt).fixupParent(null);
         
         List<Aast> remove = new ArrayList<>();
         Iterator<Aast> iter = stmt.iterator();
         while (iter.hasNext())
         {
            Aast ref = iter.next();
            
            int rtype = ref.getType();
            
            if (rtype == KW_HELP || rtype == KW_VALIDATE)
            {
               // check if the format_phrase is associated with a widget...
               Aast formatPhrase = ref.getParent();
               Aast widgetRef = formatPhrase.getPrevSibling();
               if (widgetRef.getType() == EXPRESSION)
               {
                  widgetRef = (Aast) widgetRef.getFirstChild();
               }
               int wtype = widgetRef.getType();
               
               if (wtype > BEGIN_FIELDTYPES && wtype < END_FIELDTYPES)
               {
                  // this is a HELP or VALEXP associated with a field, so remove it...
                  remove.add(ref);
               }
            }

            if (rtype > BEGIN_FIELDTYPES && rtype < END_FIELDTYPES)
            {
               ref.removeAnnotation("help");
            }
         }
         
         Set<Aast> formatPhrases = new HashSet<>();

         for (Aast ref : remove)
         {
            formatPhrases.add(ref.getParent());

            ref.remove();
         }
         
         for (Aast ref : formatPhrases)
         {
            if (ref.getNumImmediateChildren() == 0)
            {
               ref.remove();
            }
         }
      }
   }
   
   /**
    * Simple helper to extract a frame name from a frame phrase node. If a null frame phrase node
    * is provided, or the given node has no KW_FRAME child, an empty string is returned.
    *
    * @param   framePhrase
    *          Frame phrase node, or <code>null</code>.
    *
    * @return  The frame name extracted from FRAME_PHRASE/KW_FRAME/WID_FRAME, or empty string if
    *          a name could not be found (this is the equivalent to the default frame).
    */
   private String getFrameName(Aast framePhrase)
   {
      if (framePhrase == null)
      {
         return "";
      }
      
      Aast kwFrame = framePhrase.getImmediateChild(KW_FRAME, null);
      if (kwFrame == null)
      {
         return "";
      }
      
      return kwFrame.getFirstChild().getText();
   }
   
   /**
    * Disambiguates a record reference from an {@link #lvalue} or an
    * {@link #expr} by checking that there is no exact match with a
    * variable name and that the name would match a table name in the
    * schema dictionary.  The variable dictionary lookup must be handled
    * here since variable names hide table names AND since we match record
    * references in a different rule ({@link #record}) than where we match
    * variables and fields (<code>lvalue</code>).  In particular, this
    * rule must force a match on records before fields, but since the
    * matching of variables and fields is hidden in <code>lvalue</code>
    * we must disambiguate here.
    *
    * @param    text
    *           The name (possibly abbreviated) of the potential record 
    *           reference.
    *
    * @return   <code>true</code> if this is a record reference.
    */
   private boolean isRecord(String text)
   {
      return ((sym.lookupVariable(text) == -1) && sym.isTable(text));
   }
   
   /**
    * Use lookahead to determine if the qualified database name quirk is
    * present in the given tokens. In such a case, the source code has
    * whitespace between the record name and the DOT and/or whitespace
    * between the DOT and the fieldname portions but Progress treats this as
    * a properly qualified field name.
    *
    * @param    tok1
    *           First token, typically LT(1).
    * @param    tok2
    *           Second token, typically LT(2).
    * @param    tok3
    *           Third token, typically LT(3).
    *
    * @return   <code>true</code> if the name quirk is detected.
    */
   private boolean isQualifiedFieldNameQuirk(Token tok1,
                                             Token tok2,
                                             Token tok3)
   {
      boolean result = false;
   
      String text = tok1.getText();
      
      if (isRecord(text) && tok2.getType() == DOT)
      {
         if (followedByWhitespace(tok1) || followedByWhitespace(tok2))
         {
            StringBuilder sb = new StringBuilder();
            sb.append(text).append('.').append(tok3.getText());
            
            result = (sym.lookupField(sb.toString()) != -1);
         }
      }
         
      return result;
   }
   
   /**
    * Returns <code>true</code> if the given type is <code>KW_CHAR</code>,
    * <code>KW_LONGCHAR</code>, <code>VAR_CHAR</code>,
    * <code>VAR_LONGCHAR</code>, <code>FIELD_CHAR</code> or
    * <code>FIELD_CLOB</code>.
    *
    * @param    type
    *           Token type to test.
    *
    * @return   <code>true</code> if the type represents character data.
    */
   private boolean isCharType(int type)
   {
      return (type == KW_CHAR      ||
              type == KW_LONGCHAR  ||
              type == VAR_CHAR     ||
              type == VAR_LONGCHAR ||
              type == FIELD_CHAR   ||
              type == FIELD_CLOB); 
   }
   
   /**
    * Returns <code>true</code> if the given type is <code>KW_DATE</code>,
    * <code>KW_DATETIME</code>, <code>KW_DATE_TZ</code>, <code>VAR_DATE</code>,
    * <code>VAR_DATETIME</code>, <code>VAR_DATETIME_TZ</code>, <code>FIELD_DATE</code>,
    * <code>FIELD_DATETIME</code> or <code>FIELD_DATETIME_TZ</code>.
    *
    * @param    type
    *           Token type to test.
    *
    * @return   <code>true</code> if the type represents date data.
    */
   private boolean isDateType(int type)
   {
      return (type == KW_DATE         ||
              type == KW_DATETIME     ||
              type == KW_DATE_TZ      ||
              type == VAR_DATE        ||
              type == VAR_DATETIME    ||
              type == VAR_DATETIME_TZ ||
              type == FIELD_DATE      ||
              type == FIELD_DATETIME  ||
              type == FIELD_DATETIME_TZ); 
   }
   
   /**
    * Returns <code>true</code> if the given type is <code>KW_INT</code>,
    * <code>KW_INT64</code>, <code>KW_RECID</code>, <code>VAR_INT</code>,
    * <code>VAR_INT64</code>, <code>VAR_RECID</code>, <code>FIELD_INT</code>,
    * <code>FIELD_INT64</code> or <code>FIELD_RECID</code>.
    *
    * @param    type
    *           Token type to test.
    *
    * @return   <code>true</code> if the type represents date data.
    */
   private boolean isNumericType(int type)
   {
      return (type == KW_INT      ||
              type == KW_INT64    ||
              type == KW_RECID    ||
              type == KW_DEC      ||
              type == VAR_INT     ||
              type == VAR_INT64   ||
              type == VAR_RECID   ||
              type == VAR_DEC     ||
              type == FIELD_INT   ||
              type == FIELD_INT64 ||
              type == FIELD_RECID ||
              type == FIELD_DEC); 
   }
   
   /**
    * Returns <code>true</code> if the given type is a variable, field,
    * function, method, attribute that refers to a handle or if it is a
    * system handle.
    *
    * @param    node
    *           AST node to test.
    *
    * @return   <code>true</code> if the type represents handle data.
    */
   private boolean isHandleType(Aast node)
   {
      int type = node.getType();
      
      return (type == VAR_HANDLE   ||
              type == FIELD_HANDLE ||
              type == FUNC_HANDLE  ||
              type == ATTR_HANDLE  ||
              type == METH_HANDLE  ||
              type == SYS_HANDLE); 
   }
   
   /**
    * Returns <code>true</code> if the given type is a variable, field, function, method,
    * attribute or system handle that refers to a COM-HANDLE.
    *
    * @param    lvalue
    *           AST node to test.
    *
    * @return   <code>true</code> if the type represents COM-HANDLE data.
    */
   private boolean isComHandleType(Aast lvalue)
   {
      int     type = lvalue.getType();
      String  text = null;
      
      if (type == FIELD_HANDLE)
      {
         text = (String) lvalue.getAnnotation("schemaname");
      }
      
      if (type == VAR_HANDLE)
      {
         text = lvalue.getText();
      }
      
      if (type == SYS_HANDLE)
      {
         Long otype = (Long) lvalue.getAnnotation("oldtype");
         
         if (otype != null && otype.longValue() == KW_COM_SELF)
            return true;
      }
      
      return (type == VAR_COM_HANDLE   ||
              type == FIELD_COM_HANDLE ||
              type == FUNC_COM_HANDLE  ||
              type == ATTR_COM_HANDLE  ||
              type == METH_COM_HANDLE);
   }              
   
   /**
    * Returns <code>true</code> if the given subtree resolves to a <code>VAR_CLASS</code>,
    * <code>FIELD_CLASS</code>, <code>FUNC_CLASS</code>, <code>METH_CLASS</code>,
    * <code>CLASS_NAME</code> or if this is a built-in attribute or non-user defined method call
    * <code>COLON</code> that has a <code>ATTR_CLASS</code> or <code>METH_CLASS</code> in the
    * index position 1.
    * <p>
    * Parenthesis will be ignored and the contained sub-expression will be checked.
    *
    * @param    node
    *           AST node to test.
    *
    * @return   <code>true</code> if the type represents an object reference.
    */
   private boolean isObjectSubtree(Aast node)
   {
      int  type = node.getType();
      Aast ref  = node;
   
      while (ref != null && type == LPARENS)
      {
         ref = ref.getChildAt(0);
         
         if (ref == null)
         {
            break;
         }
         
         type = ref.getType();
      }
   
      return isObjectType(type);
   }
   
   /**
    * Returns <code>true</code> if the given type is <code>VAR_CLASS</code>,
    * <code>FIELD_CLASS</code>, <code>FUNC_CLASS</code>, <code>METH_CLASS</code>,  
    * <code>CLASS_NAME</code> or <code>ATTR_CLASS</code>.
    *
    * @param    type
    *           The type to test.
    *
    * @return   <code>true</code> if the type represents an object reference.
    */
   private boolean isObjectType(int type)
   {
      return (type == CLASS_NAME   ||
              type == VAR_CLASS    ||
              type == FIELD_CLASS  ||
              type == FUNC_CLASS   ||
              type == ATTR_CLASS   ||
              type == METH_CLASS);
   }
   
   /**
    * Returns <code>true</code> if the given subtree resolves to a <code>CLASS_NAME</code>.
    * Parenthesis will be ignored and the contained sub-expression will be checked.
    *
    * @param    node
    *           AST node to test.
    *
    * @return   <code>true</code> if the type represents an class reference.
    */
   private boolean isStaticObject(Aast node)
   {
      int  type = node.getType();
      Aast ref  = node;
   
      while (ref != null && type == LPARENS)
      {
         ref = ref.getChildAt(0);
         
         if (ref == null)
         {
            break;
         }
         
         type = ref.getType();
      }
   
      return (type == CLASS_NAME);
   }
   
   /**
    * Find the class name referenced by the given object reference.
    *
    * @param    node
    *           AST node which is the object reference.
    *
    * @return   The fully qualified class name of the object this node
    *           references or <code>null</code> if the reference is not
    *           to an object.
    */
   private String getObjectClassName(Aast node)
   {
      int  type = node.getType();
      Aast ref  = node;
      
      while (ref != null && type == LPARENS)
      {
         ref = ref.getChildAt(0);
         
         if (ref == null)
         {
            break;
         }
         
         type = ref.getType();
      }
   
      return (String) ref.getAnnotation("qualified");
   }
   
   /**
    * Handle annotation of variables.
    *
    * @param    name
    *           Variable name.
    * @param    type
    *           Variable type.
    * @param    node
    *           AST node to be annotated.
    */
   private void annotateVariable(String name, int type, Aast node)
   {
      // write any non-standard options into annotations
      sym.annotateVariableOptions(name, node, false);
      
      if (type == VAR_CLASS && !node.isAnnotation("qualified"))
      {
         // this must be a class-based data member (var or property)
         // so we need to lookup the class name and force the
         // annotation
         String qname = sym.lookupDataMemberClass(currentClassName,
                                                  name,
                                                  currentStaticFlag);
                                                  
         // make our best effort to recover from an unknown class ref
         if (qname == null)
         {
            ClassDefinition cls = sym.getCurrentClassDef();
            
            int otype = -1;
            
            if (node.isAnnotation("oldtype"))
            {
               Long ot = (Long) node.getAnnotation("oldtype");
               otype = ot.intValue();
            }
         
            // this should only be possible for two special cases
            if (otype == KW_THIS_OBJ)
            {
               qname = cls.getName();
            }
            else if (otype == KW_SUPER)
            {
               // super cannot be used unless there is a parent class, but we
               // are going to be safe here
               cls = cls.getParent();
               qname = cls != null ? cls.getName() : null;
            }
         }
         
         // should always be non-null, but we are being safe here
         if (qname != null)
            sym.annotateClassRef(qname, node);
      }
   }
   
   /**                      
    * Provides a command line interface for an end user to drive and/or test
    * this class.
    * <p>
    * Syntax:
    * <pre>
    *   java ProgressParser [-v] &lt;Progress_4GL_source_file&gt;
    * </pre>
    * Where:
    * <ul>
    *   <li> [-v] if present will enable visual mode, causing the tree to be
    *        displayed inside a Swing GUI panel
    *   <li> &lt;Progress_4GL_source_file&gt; is the filename of a
    *        preprocessed Progress 4GL source file (enclose the name in
    *        double quotes if spaces or other reserved command line 
    *        characters are part of the filename)
    * </ul>
    *
    * @param   args 
    *          List of command line arguments.
    */
   public static void main(String[] args) 
   {
      String syntax = "Syntax: java ProgressParser [-v] <4gl_source_file>";
      
      // filename is required
      if (args.length < 1)
      {
         System.err.println(syntax);
         System.exit(-1);
      }
      
      boolean visual = false;
      String  fname  = args[0];
      
      // honor visual mode
      if (args.length == 2 && "-v".equals(args[0]))
      {
         visual = true;
         fname  = args[1];
      }
         
      try
      {
         // setup
         SymbolResolver  sym    = new SymbolResolver(true);
         FileReader      fr     = new FileReader(fname);
         BufferedReader  br     = new BufferedReader(fr);
         ProgressLexer   lexer  = new ProgressLexer(br, sym);
         ProgressParser  parser = new ProgressParser(lexer, sym);
         
         // parse the input file and create a tree
         parser.external_proc();
         
         AnnotatedAst result = (AnnotatedAst) parser.getAST();
         
         // this is needed to ensure the parents are set right, or DumpTree
         // will be unable to calculate indents AND features like the
         // Aast.iterator() will also fail to operate properly
         result.fixups((Aast) null, null);
         
         // print a report to stdout
         DumpTree visitor = new DumpTree();
         visitor.visit(result);         
         
         if (visual)
         {
            // display the tree in a frame window if debug is on
            ASTFrame frame = new ASTFrame("Progress 4GL Source File", result);                                                
            frame.setVisible(true);
         }
      }
       
      catch(Exception excpt)
      {
         excpt.printStackTrace();
      }
   }
}

/** 
 * Pre-scan (1st pass) at reading a class definition to pre-load the symbol
 * resolver with the class' internal resources.  Class resources can be
 * referenced before their definitions appear which is why this pre-scan
 * is needed.  This does not fully parse the code.  Instead it takes a
 * highly abbreviated scan which ignores most content.  This should only
 * be called on a class definition.  The AST that results should be ignored.
 * <p>
 * At this time, the only things processed here are the following (everything
 * else is ignored):
 * <p>
 * <ul>
 *    <li> {@link #using_stmt}
 *    <li> {@link #class_stmt}
 *    <li> {@link #interface_stmt}
 *    <li> {@link #method_stmt}
 * </ul>
 * <p>
 * The above are reparsed on the 2nd pass but some of the configuration is only
 * processed on the 1st pass.  The <code>method_stmt</code> is an exception
 * since the method AST nodes need to be annotated.
 */
pre_scan_class!
   :
      {
         preScanPass++;
         
         boolean all = false;
      }
      all = external_proc
      {
         preScanPass--;
      }
   ;

/** 
 * Top-level program definition (and entry point) of a Progress 4GL external
 * procedure.  The only real purpose of this entry point is to create an
 * artificial root node for the external procedure itself.  Since there
 * is no prediction or looping occurring, there will only ever be a single
 * root node of an external procedure.  This entry point is simply used
 * for tree creation but no real parsing ever happens here.  The real core
 * implementation of the top-level parsing starts with {@link #block}.
 *
 * @return    <code>true</code> if all tokens in the input stream have been
 *            processed and the next token is <code>EOF</code>, otherwise
 *            some unexpected or invalid combination caused the top-level
 *            processing to end prematurely and without warnings or errors.
 */
external_proc! returns [boolean complete = false]
   :
      {
         topLevelEntry = true;
      }
      b:block
      {
         // this rule doesn't consume EOF, so if we have properly parsed the 
         // entire file, then the next token *should* be EOF;  test for this
         // and return the result of the test to the caller
         complete = (LA(1) == EOF);
         
         ## = #([BLOCK, "block"], b);
         
         ##.putAnnotation("builtin-cls", builtInCls);
         ##.putAnnotation("dotnet-cls", dotNetCls);
      }
   ;

/**
 * This rule is used to be able to mark blocks which are associated with OO members 
 * (methods, getters, setters).  It will disable OO dereferencing.
 *
 * @see   #block
 */
oo_block
   :
      {
         if (preScanPass > 0)
         {
            sym.setIgnoreOORefs(true);
         }
      }
      block
      {
         sym.setIgnoreOORefs(false);
      }
   ;

/**
 * This is the primary top-level rule that recursively handles Progress 4GL 
 * source files.  Note that these source files <b>must</b> be already
 * preprocessed (Progress 4GL Preprocessor statements are not handled by
 * this parser).
 * <p>
 * The core top-level processing is actually in {@link #single_block} which
 * is the rule that matches a one and only one top-level alternative. By
 * separating the alternatives from this top-level rule, it is possible to 
 * implement recursive calls that only accept a single block, statement or
 * assignment.  The <code>THEN or ELSE</code> clauses in an <code>IF</code>
 * language statement is an example of this requirement.  Other places in the
 * code that require a match to multiple blocks, statements and assignments
 * must call this rule which implements 0 or more matches to the
 * <code>single_block</code>.  This rule implements the looping and the
 * <code>single_block</code> implements the matching.
 * <p>
 * This rule loops until end of input which makes it ideal for processing an
 * entire external procedure (see {@link #external_proc}).  This rule will
 * match a top-level alternative iteratively occurs until there are no more
 * tokens to process, in which case it returns to its caller.
 * <p>
 * Note that there are spurious ambiguity warnings generated for this rule,
 * due to the fact that many of the language constructs are simply ambiguous
 * by design. This ambiguity is resolved by ordering (precedence), token
 * rewriting and semantic predicates in the lower level rules.  For this
 * reason we disable ambiguity warnings.
 * <p>
 * This rule is also used as an entry point for {@link #procedure},
 * {@link #function} and {@link #inner_block} which allows each of these
 * rules to contain nested blocks by using recursion.  Note that the fact
 * that procedure and function definitions cannot be nested is handled by
 * semantic predicates in the <code>procedure</code> and 
 * <code>function</code> rules respectively.
 */
block
   :
      // we don't think the inner block can be labeled in this case, but if it can (and if the
      // label can be malformed, then checkAndFixLabelWithReservedKeywordPrefix(1) will have to
      // be called here to properly engage the prediction logic for single_block
      (
         options { generateAmbigWarnings = false; }
         :
         { LA(1) != EOF }?
         single_block[false, false]
      )*
   ;
   
/**
 * This rule implements the core matching of the top-level language
 * constructs.  It is called by {@link #block} which is a top-level entry 
 * point for processing an external procedure as well as the recursively
 * called entry points for implementing nesting.  This rule is also called
 * directly from any parser location that needs to implement a match for 
 * any single procedure, class definition, function, trigger, block, statement
 * or assignment. The key here is that this rule does not loop.  It matches
 * one of the 12 top-level constructs and then returns.  This makes it ideal
 * to handle the <code>THEN or ELSE</code> clauses in an <code>IF</code>
 * language statement.
 * <p>
 * There are 12 possible alternatives for a top-level Progress language
 * construct:
 * <ul>
 *    <li> {@link #procedure} (definition) - rooted at a
 *         <code>PROCEDURE</code> node
 *    <li> {@link #class_def} (definition) - rooted at a
 *         <code>CLASS_DEF</code> node
 *    <li> {@link #interface_def} (definition) - rooted at an
 *         <code>INTERFACE_DEF</code> node
 *    <li> {@link #constructor} (definition) - rooted at an
 *         <code>CONSTRUCTOR</code> node
 *    <li> {@link #destructor} (definition) - rooted at an
 *         <code>DESTRUCTOR</code> node
 *    <li> {@link #user_defined_method} (definition) - rooted at an
 *         <code>METHOD_DEF</code> node
 *    <li> trigger (definition) - rooted at a <code>TRIGGER_BLOCK</code> node
 *    <li> {@link #function} (forward declaration or definition)  - rooted at
 *         a <code>FUNCTION</code> node
 *    <li> {@link #inner_block} (DO, REPEAT or FOR) - rooted at an
 *         <code>INNER_BLOCK</code> node
 *    <li> other language {@link #statement} - rooted at a
 *         <code>STATEMENT</code> node
 *    <li> directly embedded SQL processed by the {@link #statement} rule -
 *         rooted at an <code>EMBEDDED_SQL</code> node
 *    <li> {@link #assignment} - rooted at an <code>ASSIGNMENT</code> node
 * </ul>
 * <p>
 * <b>Trigger support is implemented as the {@link #on_stmt} which is
 * matched as a language statement in the <code>statement</code> rule.</b>
 * So while this rule can return a <code>TRIGGER_BLOCK</code>, this is not
 * possible from the true top level of a Progress program (even though
 * Progress seems to consider triggers as top-level constructs), this parser
 * will match such constructs through <code>on_stmt</code>.
 * <p>
 * The order of evaluation is very important.  Of special significance is the
 * fact that assignments are processed last (have the lowest precedence).
 * This is important since the leftmost token of an assignment is an optional
 * <code>lvalue</code> which could be a variable or field with a name that
 * matches an unreserved keyword.  Since the lexer is biased toward setting
 * token types to a keyword when there are conflicts, ambiguity is removed by
 * allowing all the keyword processing to occur first and then overriding the 
 * token type with a variable, field or function type.
 * <p>
 * When entered, this rule predicts a match based on the 2 lookahead tokens
 * and the rolled up prediction sets from each alternative, it calls the
 * alternative which matches and consumes all tokens so described by the
 * rules and the tree of rules which can be reached from the top-level
 * alternative. At that point, the flow of control returns back to this
 * method and this method returns to its caller.
 * <p>
 * In ANTLR, any rule references that impact the leftmost k tokens (where
 * k is the lookahead depth) cause a vertical 'rollup' of all possible sets
 * of leftmost tokens.  Rollup refers to the aggregation of possibilities
 * from called rules into a set (often a BitSet) that can be checked against
 * lookahead token 1, a 2nd set that can be checked against lookahead token
 * 2 etc... The core problem is that ANTLR's rollup mechanism is not
 * sophisticated enough to detect the <b>dependencies</b> between the sets.
 * This means that a rule reference that should only be called with a very
 * specific, ordered pair of tokens (e.g. X and Y) might end up matching
 * when A, Y or X, B are encountered in the token stream, depending on the
 * grammar.  Even though the grammer encodes these dependencies, they are
 * not considered when the top-level prediction logic is written.  This leads
 * false positives that would never normally be possible in the grammar. To
 * resolve these ambiguities, the following techniques are used: 
 * <ul>
 *    <li> proper ordering of alternatives to put more specific matches
 *         before more general matches
 *    <li> token type rewriting in called rules (based on context) to
 *         ensure that:
 *    <ul>
 *       <li> more generic token types like <code>SYMBOL</code> can be
 *            matched against more specific requirements
 *       <li> more specific token types (like unreserved keywords) are
 *            matched by other possible token types that have lower
 *            precedence but would normally still be possible depending on
 *            context
 *    </ul>
 *    <li> semantic predicates are used to add specific, targeted tests into
 *         top level rules wherever the predictive logic is too broad
 *         (eliminating the false positive by limiting the matches to the
 *          strict set of possible matches based on the dependencies)
 * </ul>
 * <p>
 * These tricks don't stop ANTLR from detecting the ambiguity but as long as
 * the grammar author inspects the resulting generated code and runs tests
 * to provide there is no ambiguity, then it is proper to disable the ANTLR
 * ambiguity warnings.
 * <p>
 * At this time, the rule is structured such that ANTLR properly predicts
 * the difference between a <code>procedure</code> keyword that is being
 * followed by a colon and thus is a label (the optional beginning of a
 * do, repeat or for block) versus the use of the <code>procedure</code>
 * keyword that starts a procedure definition (and must be followed by a
 * symbol which is the procedure name).  It can do the same with the
 * <code>function</code> keyword.  There is no ambiguity due to this.
 * <p>
 * A semantic predicate is used here to force the prediction for an inner
 * block to be more narrow than ANTLR would generate on its own.  This
 * avoids some mistaken matches.  In addition to the more specific prediction,
 * this rule needs to check whether the colon that is possibly signifying
 * a label was followed by whitespace or not.  This is needed to differentiate
 * between labels (which must always be followed by whitespace) and
 * attributes/methods where the colon cannot be followed by whitespace.
 * This part of the test relies upon the lexer to store an extra space after
 * the colon (in the token's text itself) if there was any kind of following
 * whitespace.  Only the lexer can do that because the whitespace is
 * discarded before the parser encounters it (changing that is just not
 * feasible).  So this approach is lame but it is better than the alternative
 * (which is to allow the {@link ProgressLexer#mSYMBOL} rule to consume
 * embedded colons and force the parser to split up the combined token into
 * its constituent parts.  See {@link #primary_expr} which is the rule that
 * should be matched if the label's colon does not have following whitespace. 
 * <p>
 * What makes this rule (and any looping calling rule like <code>block</code>)
 * work is that the <code>END</code> keyword cannot EVER be matched in the
 * leftmost token position.  This allows each nested block to end when such
 * a construct is encountered, which allows the calling rule to return or
 * to consume the <code>END</code> and then return.
 *
 * @param    blockRoot
 *           If <code>true</code>, root all child nodes at an artificial
 *           <code>BLOCK</code> node.
 * @param    trigger
 *           If <code>true</code> then the result will be rooted at a node
 *           of type <code>TRIGGER_BLOCK</code>. Ignored if 
 *           <code>blockRoot</code> is <code>false</code>.
 */
single_block [boolean blockRoot, boolean trigger]
   :
      {
         isUserDefFunctionCall(LT(1), LT(2));

         int typ1 = LA(1);
         int typ2 = LA(2);
         boolean blk = (typ1 == KW_DO || typ1 == KW_REPEAT || typ1 == KW_FOR);
         boolean proper = false;
         boolean mal = false;
         Aast malformedLabel = null;
         
         boolean inputFunc = (typ1 == KW_INPUT &&
                              (typ2 != KW_CLEAR && typ2 != KW_STREAM && typ2 != KW_STRM_HND &&
                               typ2 != KW_CLOSE && typ2 != KW_FROM   && typ2 != KW_THROUGH));
                               
         // for every stmt or other feature used in a constructor, increment
         // the counter
         if (constStmts > -1)
         {
            constStmts++;
         }
         
         if (!blk)
         {
            // not an unlabeled inner block, check for a possible label  
            proper = isProperLabel(typ1);
         
            if (!proper)
            {
               // not a proper label, check for a malformed label
               mal = isMalformedLabel(true, 1, null); 
               
               // HARD CODED call that can't be placed directly in our rule
               // below (inner_block) without very bad results on prediction
               // logic generation
               if (mal)
               {
                  malformed_symbol();
                  malformedLabel = (Aast)returnAST;
                  match(COLON);
               }
            }
         }
      }
      (  
         options { generateAmbigWarnings = false; }
         :
           procedure         
         | class_def
         | interface_def
         | constructor
         | destructor
         | user_defined_method
         
           // the keyword "function" is not reserved, so it can be used as a user-defined symbol
           // like a variable name, we must manually differentiate a real function statement
           // by detecting if there is a function name following the keyword
         | { LA(1) == KW_FUNCT && (isSomeSymbol(LA(2)) || isMalformedSymbol(2)) }?
           function
         
           // the rollup is too broad leading to false positives, we force the
           // real test (as ANTLR should have created it) to limit the
           // prediction to the correct set (the problem stems from the
           // use of the (symbol COLON)? to enable labels... this conflicts 
           // with just about everything because of the unreserved keywords
         | { blk || proper || mal }?
           inner_block[true, malformedLabel]
           
         | { LA(1) == DOT && (!followedByWhitespace(LT(1)) && LA(2) != EOF) }?
           crap_to_ignore
                                       
           // the SUPER statement (which must be matched here) and SUPER
           // function (which should be matched in the assignment rule below)
           // are ambiguous since their syntax can match in most cases; this
           // disambiguates the two; likewise the THIS-OBJECT stmt and 
           // THIS-OBJECT reference must similarly be disambiguated
         | { ((LA(1) != KW_SUPER && LA(1) != KW_THIS_OBJ) || constStmts == 1) && !inputFunc }?
           statement
           
         | assignment
           
         | oea_annotations
         
         | // this is a bogus entry to add the proper rollup to cause our
           // generated predictive logic to accept these tokens as possible
           // at this level, BUT we can never really match them since they
           // are disabled with a semantic predicate; so this performs a
           // "pull" function, but can't ever be matched here (all possible
           // LA(1) tokens for a malformed_symbol must be predicted in one
           // of these top-level rules (and all the others are predicted in
           // assignment at this time)
           { false }?
           (UNKNOWN_TOKEN)
      )
      {
         // when used in if/then/else or case/when/then/otherwise, this
         // rule can match a single statement or assignment, however this
         // flag specifies to root such statements or assignments at an 
         // artificial block node to aid at conversion time
         if (blockRoot && ## != null)
         {
            int    stype = ##.getType();
            int    ttype = TRIGGER_BLOCK;
            String ttxt  = "trigger block";
            
            // we don't expect function or procedure definitions here and
            // we MUST exclude inner_block definitions since these are
            // already rooted properly
            if (stype == STATEMENT || stype == ASSIGNMENT)
            {
               ## = #([BLOCK, "block"], ##);
               
               if (trigger)
               {
                  // for simple STATEMENTs or ASSIGNMENTs, we previously
                  // created a BLOCK root node, now this will be placed
                  // under a TRIGGER_BLOCK
                  ## = #([ttype, ttxt], ##);
               }
            }
            else if (stype == INNER_BLOCK)
            {
               if (trigger)
               {
                  // this is the case where there is a group of statements
                  // enclosed in a DO block (REPEAT and FOR cannot be used,
                  // even the DO block itself MUST be unqualified) so we
                  // override the type and text since this isn't a real 
                  // innerblock
                  ##.setType(ttype);
                  ##.setText(ttxt);
               }
            }
         }
      }
   ;

/**
 * Undocumented syntax found in working code that appears to be ignored in Progress.  Match the
 * text and anything else to the end of the line and drop all of it. This seems to match up with
 * the concept of annotations in the OpenEdge Architect.  For example:
 * <p>
 * <pre>
 * &#64;VisualDesigner
 * </pre>
 */
oea_annotations!
   :
      AT any_symbol_at_all
      (
         options { generateAmbigWarnings = false; }
         :
         { !followedByLineEnd(LT(1)) }?
         .
      )*
   ;
   
/**
 * Undocumented syntax found in working code that is ignored in Progress.  Match a DOT followed
 * by text with no intervening whitespace, then any amount of whitespace, text, numbers and
 * ended by a terminating DOT.  All of the matched tokens will be dropped.
 * <p>
 * The rules:
 * <p>
 * <ol>
 *    <li> must start with a DOT
 *    <li> must have a valid character immediately following the leading DOT without any
 *         intervening whitespace
 *    <li> valid characters are anything EXCEPT numbers, open curly braces, single, or double
 *         quote chars
 *    <li> after that any whitespace (inluding newlines), numbers or other text is matched
 *         until a following DOT
 *    <li> inclusion of unescaped single or double quote characters anywhere does not work
 *         properly but it doesn't cause an error (more testing will be needed to see how
 *         it responds)
 *    <li> unescaped opening curly braces cannot be embedded
 *    <li> can be at the beginning, middle or end of any line in the file, but must NOT be
 *         inside a statement or expression
 *    <li> this construct CANNOT be started by the ternminating DOT of a statement, it must
 *         stand on its own
 * </ol>
 * <p>
 * For example:
 * <p>
 * <pre>
 * .t  y757565gfhvbm hgj lk hkudu09sd
 * 
 * .
 * .?/.,&gt;&lt;;:~`\|][}&amp;=+-_)(*^%@!$#.
 * ./.
 * ..FDFDFD.
 * .......
 * .SDSD....SDSD.
 * .a......
 * ..dd..ddd   ...
 * .,.
 * .&gt;.
 * .&lt;.
 * .;.
 * .:.
 * .~.
 * .`.
 * .\.
 * .|.
 * .].
 * .[.
 * .}.
 * .=.
 * .+.
 * .-.
 * ._.
 * .).
 * .(.
 * .*.
 * .&amp;.
 * .^.
 * .%.
 * .@.
 * .!.
 * .$.
 * .#.
 * .~".
 * .~{.
 * .~'.
 * .s89s.
 * .hdhdhdgsgdhfdjfj.
 * .w.
 * .l87676.
 * </pre>
 */
crap_to_ignore!
   :
      d:DOT
      (
         options { generateAmbigWarnings = false; }
         :
           { LA(1) == DOT && !followedByWhitespace(LT(1)) }?
           DOT
         | ~DOT
      )*
      e:DOT
   ;
      
/**
 * This rule creates the block structure for a procedure. All procedure
 * definitions must start with the <code>PROCEDURE</code> language statement
 * (see {@link #proc_stmt}). This is followed by a recursive call to
 * {@link #block} which handles the ability to nest top-level statements
 * inside a procedure. Finally, the rule ends with an optional
 * <code>END</code> language statement. The Progress language states that it
 * is valid to not have and <code>END</code>. In this case, the procedure's
 * end is implicit at the end of the file.
 * <p>
 * The most important features of this rule are:
 * <ul>
 *    <li> The use of the <code>nestLevel</code> member to track the current
 *         nesting level of the tree.  This allows us to throw an exception
 *         (using a semantic predicate) when there is any attempt to nest
 *         procedure or function definitions (which is not allowed). The
 *         semantic predicate is used here instead of using the parser
 *         structure itself (the manner in which the rules call each other)
 *         since the ability to properly nest blocks is dependent upon the
 *         structure as implemented.  Any other approach limited the nesting
 *         in a manner that was more restrictive than the Progress language
 *         itself.
 *    <li> At the beginning of this method, a scope is added to the variable
 *         dictionary (this is the only part of the symbol dictionary that
 *         is scope aware).  At the end of this method, the current scope
 *         is removed.  This allows the definition of local variables that
 *         hide variables of the same name in higher level scopes. Note that
 *         since functions and procedures cannot be nested, there is no need
 *         to use scopes in the respective namespaces for functions and
 *         procedures.  Progress essentially implements a flat namespace for
 *         each of these respectively.
 *    <li> At the beginning of this method, a scope is added to the schema
 *         dictionary.  In addition, a flag is set to indicate that the
 *         parser is currently inside an internal procedure.  This flag is
 *         used during buffer creation (see {@link #def_buf_stmt}) to 
 *         determine the proper scope in which to create the buffer.  At the
 *         end of this method, the current scope is removed.
 * </ul>
 * <p>
 * This rule is a peer alternative to {@link #function} and
 * {@link #class_def} rules (among others).
 * <p>
 * Called by {@link #single_block}.
 */
procedure! 
   :
      {
         nestLevel++;
         sym.addScope();
         sym.addSchemaScope(true);
         bufferScope = false;
      }
      (
         // can't define an internal proc inside another internal proc
         p:proc_stmt { nestLevel < 2 }?
         b:block
         (
            options { generateAmbigWarnings = false; }
            :
            end_stmt[true]
         )?
      )
      {
         ## = #([PROCEDURE, "procedure"], p, ([BLOCK, "block"], b)); 
         
         // internal procedure and function definitions cannot be nested in
         // each other or within themselves, so whenever this exits the
         // buffer scope can be safely set to global
         bufferScope = true;
         
         // internal procedures propagate
         sym.deleteSchemaScope(true);
         sym.deleteScope();
         nestLevel--;
      }      
   ;
   
/**
 * Implements the structure of a <code>PROCEDURE</code> block definition
 * which is the manner in which Progress implements a form of subroutine. 
 * <p>
 * The most important feature of this rule is the direct usage of the
 * {@link #malformed_symbol} rule to force the second token to a 
 * <code>SYMBOL</code> token type. This allows reserved keywords and names that
 * start with numbers or other characters that usually cannot be used as the
 * first character of a symbol.  Additionally, a <code>DB_SYMBOL</code>
 * or a <code>FILENAME</code> can be matched there. This token is then used
 * in the exit action to add the defined procedure's name to the procedure 
 * namespace. 
 * <p>
 * Note that it is undocumented but possible to use reserved keywords as
 * internal procedure names in Progress 4GL. Otherwise only the rule
 * {@link #symbol} would have been needed.
 * <p>
 * Procedures do not have a return value.  There is a special result code to
 * optionally indicate success or failure but there is no way to call a
 * procedure and assign the result to an lvalue.  For this reason, there is
 * no need to associate the procedure name with a special token type that
 * is data type specific (i.e. this is different from functions and
 * variables).  A standard token type of <code>PROCEDURE</code> is associated
 * with all procedures in the namespace, but it is not really used for
 * anything.
 */
proc_stmt
   :
      {
         boolean privateAccess = false;
      }
      (
         KW_PROC^
         (
            options { generateAmbigWarnings = false; }
            :            
            t:STRING | s:malformed_symbol
         )
         (external_api_definition)?
         (KW_PRIVATE { privateAccess = true; })?
         (in_super_clause)?
         (DOT! | COLON!)
      )
      {
         if (#t != null)
         {
            #t.setType(SYMBOL);
            #t.setText(character.progressToJavaString(#t.getText()));
            #s = #t;
         }
         
         sym.addProcedure(#s.getText(), INT_PROC); 
         
         // remember if the method should be private
         if (privateAccess)
         {
            ##.putAnnotation("access", "private");
         }
      }
   ;
   
/**
 * Match the <code>EXTERNAL</code> keyword with a required following
 * <code>STRING</code> (shared library name). Uses {@link #ordinal_clause}.
 * Used by {@link #proc_stmt}.
 */
external_api_definition
   :
      KW_EXTERN^
      STRING
      (KW_CDECL | KW_PASCAL | KW_STDCALL)?
      (ordinal_clause)?
      (KW_PERSIST)?
   ;
   
/**
 * Match the <code>ORDINAL</code> keyword with a required following
 * <code>NUM_LITERAL</code>. Used by {@link #external_api_definition}.
 */
ordinal_clause
   :
      KW_ORDINAL^ NUM_LITERAL
   ;
   
/**
 * This rule creates the block structure for a class definition. All class
 * definitions must start with the {@link #class_stmt}.  This is followed by
 * a recursive call to {@link #block} which handles the ability to nest
 * top-level statements inside a class definition. Finally, the rule ends
 * with an <code>END</code> language statement. 
 * <p>
 * The most important features of this rule are:
 * <ul>
 *    <li> The use of the <code>nestLevel</code> member to track the current
 *         nesting level of the tree.  This allows us to throw an exception
 *         (using a semantic predicate) when there is any attempt to nest
 *         procedure or function definitions (which is not allowed). The
 *         semantic predicate is used here instead of using the parser
 *         structure itself (the manner in which the rules call each other)
 *         since the ability to properly nest blocks is dependent upon the
 *         structure as implemented.  Any other approach limited the nesting
 *         in a manner that was more restrictive than the Progress language
 *         itself.
 *    <li> At the beginning of this method, a scope is added to the variable
 *         dictionary (this is the only part of the symbol dictionary that
 *         is scope aware).  At the end of this method, the current scope
 *         is removed.  This allows the definition of local variables that
 *         hide variables of the same name in higher level scopes. Note that
 *         since functions and procedures cannot be nested, there is no need
 *         to use scopes in the respective namespaces for functions and
 *         procedures.  Progress essentially implements a flat namespace for
 *         each of these respectively.
 *    <li> At the beginning of this method, a scope is added to the schema
 *         dictionary.  In addition, a flag is set to indicate that the
 *         parser is currently inside an internal procedure.  This flag is
 *         used during buffer creation (see {@link #def_buf_stmt}) to 
 *         determine the proper scope in which to create the buffer.  At the
 *         end of this method, the current scope is removed.
 * </ul>
 * <p>
 * This rule is a peer alternative to {@link #procedure},
 * {@link #interface_def}, {@link #function} and other top-level blocks.
 * <p>
 * Called by {@link #single_block}.
 */
class_def! 
   :
      {
         insideClass = true;

         nestLevel++;
         sym.addScope();
         sym.addSchemaScope(true);
         
         // treat class definitions like external procedures, don't force bufferScope to false
      }
      (
         // can't define an internal proc inside another internal proc
         c:class_stmt { nestLevel < 2 }?
         b:block
         (
            options { generateAmbigWarnings = false; }
            :
            end_stmt[true]
         )?
      )
      {
         ## = #([CLASS_DEF, "class definition"], c, ([BLOCK, "block"], b)); 
         
         // don't know if class defs propagate, assume true for now
         sym.deleteSchemaScope(true);
         sym.deleteScope();
         nestLevel--;
         sym.classDefinitionComplete();

         insideClass = false;
      }
   ;
   
/**
 * Implements the structure of a <code>CLASS</code> language statement. 
 * <p>
 * The most important feature of this rule is the direct usage of the
 * {@link #any_symbol_at_all} rule to force the second token to a 
 * <code>SYMBOL</code> token type. This token is then used in the exit
 * action to add the class's definition. 
 */
class_stmt
   :
      (
         KW_CLASS^ s:any_symbol_at_all
         (
              inher:inherits_clause
            | impl:implements_clause
            | (KW_ABSTRACT | KW_FINAL)
            | KW_USE_WIDP
         )*
         (DOT! | COLON!)
      )
      {
         sym.addClass(#s, #inher, #impl);
         
         builtInCls = sym.isBuiltInClass(#s.getText()); 
         dotNetCls = sym.isDotNetClass(#s.getText()); 
      }
   ;
   
/**
 * Matches the <code>KW_INHERITS</code> with a following
 * {@link #user_defined_type_name}. This defines the superclass for a
 * {@link #class_stmt}.
 */
inherits_clause
   :
      KW_INHERITS^ user_defined_type_name
   ;

/**
 * Matches <code>KW_IMPLMTS</code> with one or more comma-separated following
 * {@link #user_defined_type_name}. Used by {@link #class_stmt}.
 */
implements_clause
   :
      KW_IMPLEMTS^ user_defined_type_name (COMMA! user_defined_type_name)*
   ;
   
/**
 * This rule creates the block structure for an interface definition. All
 * interface definitions must start with the {@link #interface_stmt}.
 * This is followed by a recursive call to {@link #block} which handles
 * the ability to nest top-level statements inside an interface. Finally,
 * the rule ends with an <code>END</code> language statement. 
 * <p>
 * The most important features of this rule are:
 * <ul>
 *    <li> The use of the <code>nestLevel</code> member to track the current
 *         nesting level of the tree.  This allows us to throw an exception
 *         (using a semantic predicate) when there is any attempt to nest
 *         procedure or function definitions (which is not allowed). The
 *         semantic predicate is used here instead of using the parser
 *         structure itself (the manner in which the rules call each other)
 *         since the ability to properly nest blocks is dependent upon the
 *         structure as implemented.  Any other approach limited the nesting
 *         in a manner that was more restrictive than the Progress language
 *         itself.
 *    <li> At the beginning of this method, a scope is added to the variable
 *         dictionary (this is the only part of the symbol dictionary that
 *         is scope aware).  At the end of this method, the current scope
 *         is removed.  This allows the definition of local variables that
 *         hide variables of the same name in higher level scopes. Note that
 *         since functions and procedures cannot be nested, there is no need
 *         to use scopes in the respective namespaces for functions and
 *         procedures.  Progress essentially implements a flat namespace for
 *         each of these respectively.
 *    <li> At the beginning of this method, a scope is added to the schema
 *         dictionary.  In addition, a flag is set to indicate that the
 *         parser is currently inside an internal procedure.  This flag is
 *         used during buffer creation (see {@link #def_buf_stmt}) to 
 *         determine the proper scope in which to create the buffer.  At the
 *         end of this method, the current scope is removed.
 * </ul>
 * <p>
 * This rule is a peer alternative to {@link #procedure}, {@link #class_def}
 * {@link #function} and other top-level blocks.
 * <p>
 * Called by {@link #single_block}.
 */
interface_def! 
   :
      {
         nestLevel++;
         sym.addScope();
         sym.addSchemaScope(true);
         
         // treat class definitions like external procedures, don't force bufferScope to false
      }
      (
         // can't define an internal proc inside another internal proc
         i:interface_stmt { nestLevel < 2 }?
         b:block 
         (
            options { generateAmbigWarnings = false; }
            :
            end_stmt[true]
         )?
      )
      {
         ## = #([INTERFACE_DEF, "interface definition"],
                i,
                ([BLOCK, "block"], b)); 
         
         // don't know if interfaces propagate, assume true for now
         sym.deleteSchemaScope(true);
         sym.deleteScope();
         nestLevel--;
         sym.classDefinitionComplete();
      }      
   ;
   
/**
 * Implements the structure of a <code>INTERFACE</code> language statement. 
 * <p>
 * The most important feature of this rule is the direct usage of the
 * {@link #any_symbol_at_all} rule to force the second token to a 
 * <code>SYMBOL</code> token type. This token is then used in the exit
 * action to add the interface's definition. 
 */
interface_stmt
   :
      (
         KW_INTERFAC^ s:any_symbol_at_all (DOT! | COLON!)
      )
      {
         sym.addInterface(#s); 
         
         builtInCls = sym.isBuiltInClass(#s.getText()); 
         dotNetCls = sym.isDotNetClass(#s.getText()); 
      }
   ;
   
/**
 * Matches an {@link #any_non_reserved_symbol} as a class name reference. This
 * is the same idea as {@link #user_defined_type_name} except that here we do
 * not match reserved keywords. This is required to provide the behavior of
 * how class names are matched in {@link #primary_expr}. If the name is
 * an unqualified class name, it will be resolved into the fully
 * qualified class name that is referenced. The resulting qualified name will
 * be stored in an annotation named "qualified". The resulting node type will
 * always be <code>CLASS_NAME</code>.
 */
non_reserved_user_defined_type_name
   :
      nam:any_non_reserved_symbol
      {
         String name = #nam.getText();
         String full = sym.loadClass(name);
         
         if (full == null)
         {
            String err = String.format("Unresolvable user defined type '%s'.", name);
            throw genExc(err, #nam);
         }
         
         sym.annotateClassRef(full, ##);
         ##.setType(CLASS_NAME);
      }
   ;
   
/**
 * Matches an {@link #any_symbol_at_all} or a <code>STRING</code> and if it
 * represents an unqualified class name, it will be resolved into the fully
 * qualified class name that is referenced. The resulting qualified name will
 * be stored in an annotation named "qualified". The resulting node type will
 * always be <code>CLASS_NAME</code>.
 * <p>
 * The 10.2x syntax which allows a main.class.Name+InnerClassName format is
 * supported here.
 */
user_defined_type_name
   :
      (
           sym:any_symbol_at_all
           (
              options { generateAmbigWarnings = false; }
              :
              PLUS! innerCls:any_symbol_at_all!
           )?
         | str:STRING
      )
      {
         String name = null;
         
         if (#sym != null)
         {
            name = #sym.getText();
            
            if (#innerCls != null)
            {
               name = name + "+" + #innerCls.getText();
            }
         }
         else
         {
            name = character.progressToJavaString(#str.getText());
            #str.setText(name);
         }
         
         String full = sym.loadClass(name);
         
         if (full == null)
         {
            String err = String.format("Unresolvable user defined type '%s'.", name);
            throw genExc(err, ##);
         }
         
         sym.annotateClassRef(full, ##);
         ##.setType(CLASS_NAME);
      }
   ;
   
/**
 * Matches the <code>USING</code> language statement which has a
 * <code>KW_USING</code> node followed by an {@link #any_symbol_at_all}.
 * After the symbol, because of the way the lexer works, it is possible to
 * match a <code>DOT</code> and then a <code>MULTIPLY</code>.  This text is
 * merged into the symbol if found.  Finally, an optional {@link #from_assembly}
 * clause can be matched. 
 * <p>
 * This updates the symbol resolver's package list with the symbol text. That
 * is subsequently used to resolve unqualified class names into fully
 * qualified package + class names.
 * <p>
 * Called by {@link #stmt_list}.
 */
using_stmt
   :
      // the search spec form of the package name will appear as 3 tokens
      // so we must optionally match that way
      KW_USING^ 
      (
           sy:any_symbol_at_all
           (
              options { generateAmbigWarnings = false; }
              :
              DOT! m:MULTIPLY!
           )?
         | st:STRING
      )
      (f:from_assembly)? DOT!
      {
         StringBuilder sb = new StringBuilder();
      
         if (#st != null)
         {
            String txt = character.progressToJavaString(#st.getText());
            sb.append(StringHelper.processEscapes(txt));
         }
         else
         {
            sb.append(#sy.getText());
            
            if (#m != null)
            {
               sb.append(".*");
               #sy.setText(sb.toString());
            }
         }
         
         sym.addPackage(sb.toString(), #f);
         
         if (preScanPass == 0)
         {
            ##.putAnnotation("search_spec", sb.toString());
         }
      }
   ;
   
/**
 * Matches the <code>KW_FROM</code> in a {@link #using_stmt}. There is a
 * required following <code>KW_ASSEMBLY</code> or <code>KW_PROPATH</code>.
 * Separated in order to build the tree properly.
 */
from_assembly
   :
      KW_FROM^ (KW_ASSEMBLY | KW_PROPATH)
   ;
   
/**
 * This rule creates the block structure for a constructor definition. All
 * interface definitions must start with the {@link #constructor_stmt}.
 * This is followed by a recursive call to {@link #block} which handles
 * the ability to nest top-level statements inside a constructor. Finally,
 * the rule ends with an <code>END</code> language statement. 
 * <p>
 * The most important features of this rule are:
 * <ul>
 *    <li> The use of the <code>nestLevel</code> member to track the current
 *         nesting level of the tree.  This allows us to throw an exception
 *         (using a semantic predicate) when this is not nested inside a
 *         class definition.
 *    <li> At the beginning of this method, a scope is added to the variable
 *         dictionary (this is the only part of the symbol dictionary that
 *         is scope aware).  At the end of this method, the current scope
 *         is removed.  This allows the definition of local variables that
 *         hide variables of the same name in higher level scopes. Note that
 *         since functions and procedures cannot be nested, there is no need
 *         to use scopes in the respective namespaces for functions and
 *         procedures.  Progress essentially implements a flat namespace for
 *         each of these respectively.
 *    <li> At the beginning of this method, a scope is added to the schema
 *         dictionary.  In addition, a flag is set to indicate that the
 *         parser is currently inside an internal procedure.  This flag is
 *         used during buffer creation (see {@link #def_buf_stmt}) to 
 *         determine the proper scope in which to create the buffer.  At the
 *         end of this method, the current scope is removed.
 * </ul>
 * <p>
 * This rule is a peer alternative to {@link #procedure}, {@link #class_def}
 * {@link #interface_def}, {@link #destructor}, {@link #function} and other
 * top-level blocks.
 * <p>
 * Called by {@link #single_block}.
 */
constructor!
   :
      {
         nestLevel++;
         sym.addScope();
         sym.addSchemaScope(true);
         boolean oldBufScope = bufferScope;
         bufferScope = false;
         constStmts = 0;
      }
      (
         // constructor must be defined inside a class definition
         c:constructor_stmt { nestLevel == 2 }?
         b:oo_block
         (
            options { generateAmbigWarnings = false; }
            :
            end_stmt[true]
         )?
      )
      {
         ## = #([CONSTRUCTOR, "constructor"], c, ([BLOCK, "block"], b)); 
         
         constStmts = -1;
         
         // constructors are always nested
         bufferScope = oldBufScope;
         
         // don't know if constructors propagate, assume true for now
         sym.deleteSchemaScope(true);
         sym.deleteScope();
         nestLevel--;
      }      
   ;
   
/*
 * Matches the <code>CONSTRUCTOR</code> language statement. This is a <code>KW_CONSTRUC</code>
 * followed by an optional {@link #access_mode}, a class name ({@link #any_symbol_at_all}), 
 * {@link #param_list_definition} and a <code>DOT</code> or <code>COLON</code>. The class name
 * is syntactic sugar and is dropped.
 */
constructor_stmt
   :
      KW_CONSTRUC^ (access_mode | KW_STATIC)* any_symbol_at_all!
      param_list_definition[true, false] (DOT! | COLON!)
   ;
   
/**
 * Matches <code>KW_PUBLIC</code>, <code>KW_PROTECTD</code> or
 * <code>KW_PRIVATE</code>.
 */
access_mode
   :
      KW_PUBLIC | KW_PROTECTD | KW_PRIVATE
   ;
   
/**
 * This rule creates the block structure for a user defined method. All
 * method definitions must start with the {@link #method_stmt}. If this
 * is a class definition rather than an interface definition, then this is
 * followed by a recursive call to {@link #block} which handles the ability
 * to nest top-level statements inside a method. If there is a block then
 * the rule ends with an <code>END</code> language statement. 
 * <p>
 * The most important features of this rule are:
 * <ul>
 *    <li> The use of the <code>nestLevel</code> member to track the current
 *         nesting level of the tree.  This allows us to throw an exception
 *         (using a semantic predicate) when this is not nested inside a
 *         class definition.
 *    <li> At the beginning of this method, a scope is added to the variable
 *         dictionary (this is the only part of the symbol dictionary that
 *         is scope aware).  At the end of this method, the current scope
 *         is removed.  This allows the definition of local variables that
 *         hide variables of the same name in higher level scopes. Note that
 *         since functions and procedures cannot be nested, there is no need
 *         to use scopes in the respective namespaces for functions and
 *         procedures.  Progress essentially implements a flat namespace for
 *         each of these respectively.
 *    <li> At the beginning of this method, a scope is added to the schema
 *         dictionary.  In addition, a flag is set to indicate that the
 *         parser is currently inside an internal procedure.  This flag is
 *         used during buffer creation (see {@link #def_buf_stmt}) to 
 *         determine the proper scope in which to create the buffer.  At the
 *         end of this method, the current scope is removed.
 * </ul>
 * <p>
 * This rule is a peer alternative to {@link #procedure}, {@link #class_def}
 * {@link #interface_def}, {@link #constructor}, {@link #destructor},
 * {@link #function} and other top-level blocks.
 * <p>
 * Called by {@link #single_block}.
 */
user_defined_method!
   :
      {
         nestLevel++;
         sym.addScope();
         sym.addSchemaScope(true);
         boolean oldBufScope = bufferScope;
         bufferScope = false;
         boolean iface = sym.getCurrentClassDef().isInterface();
      }
      (
         // method must be defined inside a class definition
         m:method_stmt { nestLevel == 2 }?
         (
              // method bodys are only matched inside a class definition, not
              // in an interface def (for now there are no abstract methods
              // in progress so this should be OK)
              { !iface }?
              b:oo_block
              (
                 options { generateAmbigWarnings = false; }
                 :
                 end_stmt[true]
              )?
              
            | // empty alternative
         )
      )
      {
         Aast blk = (#b == null) ? null : #([BLOCK, "block"], b);
         ## = #([METHOD_DEF, "method definition"], m, blk);
         
         ##.putAnnotation("prototype", new Boolean(blk == null));
         
         // methods are always nested
         bufferScope = oldBufScope;
         
         // don't know if methods propagate, assume true for now (since
         // methods that have not return values act like internal procs
         // and those that return values act like functions, there may be
         // 2 different cases here)
         sym.deleteSchemaScope(true);
         sym.deleteScope();
         inStaticCtxt = false;
         nestLevel--;
      }      
   ;
   
/*
 * Matches the <code>METHOD</code> language statement. This is a <code>KW_METHOD</code>
 * followed by an optional {@link #access_mode}, a method name ({@link #symbol}),
 * {@link #param_list_definition} and a <code>DOT</code> or <code>COLON</code>.
 */
method_stmt
   {
      int mtype = -1;
      int size  = -1;
   }
   :
      KW_METHOD^
      (
         options { generateAmbigWarnings = false; }
         :
           a:access_mode
         | st:KW_STATIC { inStaticCtxt = true; }
         | KW_ABSTRACT
         | KW_OVERRIDE
         | KW_FINAL
      )*
      mtype = cls:method_return
      (
         size = extent
         {
            if (size == -1 || size > 0) ##.putAnnotation("extent",  new Long(size));
         }
      )?
      s:any_symbol_at_all
      
      param_list_definition[true, false] (DOT! | COLON!)
      
      {
         int atype = (#a == null) ? KW_PUBLIC : #a.getType();
         
         String mname = #s.getText();
         String qname = null;
         
         // we need to find the 
         if (mtype == METH_CLASS)
         {
            qname = (String) #cls.getAnnotation("qualified");
         }
         
         // add the method to the current class def and annotate the root
         // node with the method level annotations
         sym.addObjectMethod(mname, mtype, atype, (#st != null), ##, qname);
      }
   ;

/**
 * Matches the return datatype of a user-defined method. This is used by 
 * {@link #method_stmt} and is only separated to make the AST build
 * correctly.  
 *
 * @return    The token type of the method's return value.
 */
method_return returns [int mtype = -1]      
   :
      ( 
         options { generateAmbigWarnings = false; }
         :
           KW_VOID                 { mtype = METH_VOID; }
         | KW_CHAR                 { mtype = METH_CHAR; }
         | KW_COM_HNDL             { mtype = METH_COM_HANDLE; }
         | KW_DATE                 { mtype = METH_DATE; }
         | KW_DATETIME             { mtype = METH_DATETIME; }
         | KW_DATE_TZ              { mtype = METH_DATETIME_TZ; }
         | KW_DEC                  { mtype = METH_DEC; }
         | KW_HANDLE               { mtype = METH_HANDLE; }
         | KW_WID_HAND             { mtype = METH_HANDLE; }
         | KW_INT                  { mtype = METH_INT; }
         | KW_INT64                { mtype = METH_INT64; }
         | (KW_LOGICAL | k:KW_LOG  { #k.setType(KW_LOGICAL); })
                                   { mtype = METH_LOGICAL; }
         | KW_LONGCHAR             { mtype = METH_LONGCHAR; }
         | KW_MEMPTR               { mtype = METH_MEMPTR; }
         | KW_RAW                  { mtype = METH_RAW; }
         | KW_RECID                { mtype = METH_RECID; }
         | KW_ROWID                { mtype = METH_ROWID; }
         | (
              options { generateAmbigWarnings = false; }
              :
              KW_CLASS!
           )?
           user_defined_type_name  { mtype = METH_CLASS; }
      )
   ;
   
/**
 * This rule creates the block structure for a destructor definition. All
 * interface definitions must start with the {@link #destructor_stmt}.
 * This is followed by a recursive call to {@link #block} which handles
 * the ability to nest top-level statements inside a destructor. Finally,
 * the rule ends with an <code>END</code> language statement. 
 * <p>
 * The most important features of this rule are:
 * <ul>
 *    <li> The use of the <code>nestLevel</code> member to track the current
 *         nesting level of the tree.  This allows us to throw an exception
 *         (using a semantic predicate) when this is not nested inside a
 *         class definition.
 *    <li> At the beginning of this method, a scope is added to the variable
 *         dictionary (this is the only part of the symbol dictionary that
 *         is scope aware).  At the end of this method, the current scope
 *         is removed.  This allows the definition of local variables that
 *         hide variables of the same name in higher level scopes. Note that
 *         since functions and procedures cannot be nested, there is no need
 *         to use scopes in the respective namespaces for functions and
 *         procedures.  Progress essentially implements a flat namespace for
 *         each of these respectively.
 *    <li> At the beginning of this method, a scope is added to the schema
 *         dictionary.  In addition, a flag is set to indicate that the
 *         parser is currently inside an internal procedure.  This flag is
 *         used during buffer creation (see {@link #def_buf_stmt}) to 
 *         determine the proper scope in which to create the buffer.  At the
 *         end of this method, the current scope is removed.
 * </ul>
 * <p>
 * This rule is a peer alternative to {@link #procedure}, {@link #class_def}
 * {@link #interface_def}, {@link #constructor}, {@link #function} and other
 * top-level blocks.
 * <p>
 * Called by {@link #single_block}.
 */
destructor!
   :
      {
         nestLevel++;
         sym.addScope();
         sym.addSchemaScope(true);
         boolean oldBufScope = bufferScope;
         bufferScope = false;
      }
      (
         // destructor must be defined inside a class definition
         d:destructor_stmt { nestLevel == 2 }?
         b:oo_block
         (
            options { generateAmbigWarnings = false; }
            :
            end_stmt[true]
         )?
      )
      {
         ## = #([DESTRUCTOR, "destructor"], d, ([BLOCK, "block"], b)); 
         
         // destructors are always nested
         bufferScope = oldBufScope;
         
         // don't know if destructors propagate, assume true for now
         sym.deleteSchemaScope(true);
         sym.deleteScope();
         nestLevel--;
      }      
   ;
   
/*
 * Matches the <code>DESTRUCTOR</code> language statement. This is a
 * <code>KW_DESTRUCT</code> followed by a class name using
 * {@link #any_symbol_at_all} and a <code>DOT</code> or <code>COLON</code>.
 * Note that all of the following tokens are syntactic sugar and are thus
 * dropped.
 */
destructor_stmt
   :
      KW_DESTRUCT^ (KW_PUBLIC!)? any_symbol_at_all!
      LPARENS! RPARENS! (DOT! | COLON!)
   ;
   
/**
 * This rule creates the block structure for a function or handles the case
 * of a forward function declaration (without a block). All function 
 * definitions must start with the <code>FUNCTION</code> language statement.
 * If this is not a forward declaration (identified by the use of the
 * <code>FORWARD</code> keyword), then the <code>FUNCTION</code> statement
 * is followed by a recursive call to {@link #block} which handles
 * the ability to nest top-level statements inside a function. Finally,
 * the rule ends with an optional <code>END</code> language statement. The
 * Progress language states that it is valid to not have and <code>END</code>.
 * In this case, the function's end is implicit at the end of the file.
 * <p>
 * The most important features of this rule are:
 * <ul>
 *    <li> The use of the <code>nestLevel</code> member to track the current
 *         nesting level of the tree.  This allows us to throw an exception
 *         (using a semantic predicate) when there is any attempt to nest
 *         procedure or function definitions (which is not allowed). The
 *         semantic predicate is used here instead of using the parser
 *         structure itself (the manner in which the rules call each other)
 *         since the ability to properly nest blocks is dependent upon the
 *         structure as implemented.  Any other approach limited the nesting
 *         in a manner that was more restrictive than the Progress language
 *         itself.
 *    <li> At the beginning of this method, a scope is added to the variable
 *         dictionary (this is the only part of the symbol dictionary that
 *         is scope aware).  At the end of this method, the current scope
 *         is removed.  This allows the definition of local variables that
 *         hide variables of the same name in higher level scopes. Note that
 *         since functions and procedures cannot be nested, there is no need
 *         to use scopes in the respective namespaces for functions and
 *         procedures.  Progress essentially implements a flat namespace for
 *         each of these respectively.
 *    <li> The subrule that implements the recursive call to 
 *         <code>block</code> cannot be made optional (using ANTLR syntax)
 *         without introducing ambiguity. Instead, a trick is used to create
 *         the same result. The subrule is split into two alternatives:
 *         <ul>
 *            <li> the original call to <code>block</code> (with its optional 
 *                 <code>END</code> statement)
 *            <li> the empty alternative (no rule or token references)
 *         </ul>
 *         Most importantly, the choice between these alternatives is made 
 *         using a semantic predicate which tests the return value from
 *         the {@link #func_stmt} rule.  This return value is a boolean that
 *         is <code>true</code> only in the case where this is a forward
 *         declaration.  This approach means that forward declarations cause
 *         no block to appear in the tree, but all other types of functions
 *         properly create the block as expected.
 *    <li> At the beginning of this method, a scope is added to the schema
 *         dictionary.  In addition, a flag is set to indicate that the
 *         parser is currently inside an internal procedure.  This flag is
 *         used during buffer creation (see {@link #def_buf_stmt}) to 
 *         determine the proper scope in which to create the buffer.  At the
 *         end of this method, the current scope is removed.
 * </ul>
 * <p>
 * This rule is a peer alternative to {@link #procedure}.
 * <p>
 * Called by {@link #single_block}.
 */
function! 
   :
      {
         boolean fwd = false;      
         nestLevel++;
         sym.addScope();
         sym.addSchemaScope(true);
         bufferScope = false;
      }
      (
         fwd=f:func_stmt
         ( 
              { fwd == false }?
              b:block 
              (
                 options { generateAmbigWarnings = false; }
                 :
                 end_stmt[true]
              )?
            | // the empty alternative! which makes this subrule optional
              // based on the semantic predicate { fwd == false }? above 
         )
      )
      {
         AST optional = null;
         
         // the #(,) ANTLR syntax unconditionally creates a node, BUT if
         // there is no subtree "b" then we don't need this; make this
         // creation conditional in the case where fwd == true         
         if (#b != null)
         {
            optional = #([BLOCK, "block"], b);
         }
            
         ## = #([FUNCTION, "function"], f, optional); 
         
         // internal procedure and function definitions cannot be nested in
         // each other or within themselves, so whenever this exits the
         // buffer scope can be safely set to global
         bufferScope = true;
         
         // functions propagate
         sym.deleteSchemaScope(true);
         sym.deleteScope();
         nestLevel--;
      }      
   ;
   
/**
 * Implements the structure of a <code>FUNCTION</code> block definition
 * which is the manner in which Progress implements a subroutine which can 
 * be called as part of an expression and which returns a value of a
 * prespecified data type.
 * <p>
 * The most important feature of this rule is the direct usage of the
 * {@link #symbol} rule to force the second token to a <code>SYMBOL</code> 
 * token type.  This token is then used in the exit action to add the
 * defined function's name to the function namespace. 
 * <p>
 * Since each function has a prespecified return type, this type is
 * stored in the function namespace as the token type associated with the
 * specific function name that is added.  This is then used as the token
 * type for the root node of the subtree that represents a function call.
 * In this manner, the expression evaluation can properly determine the
 * data type of the return value for any defined or declared function.
 * <p>
 * All supported data types can be used as return types. Since the return
 * type is provided back and stored an annotation, the tree associated with
 * the return type is not needed and thus it is dropped.
 * <p>
 * This rule calls {@link #param_list_definition} to process the function parameters
 * and add them to the variable dictionary.
 * <p>
 * Forward declarations are supported by returning <code>false</code> unless
 * a forward declaration is detected by the presence of the 
 * <code>FORWARD</code> keyword.  In such cases, the return value will change.
 * <p>
 * Note that in either case (normal definition or forward declaration), the 
 * function is added to the function namespace.  This is exactly as Progress
 * implements it, since this allows the function name to be referenced
 * even if the definition has not occurred (a forward declaration). Note that
 * it is perfectly legal to have the same function be forward declared AND
 * then defined later in the source file.  This causes the same function to
 * be added twice to the function namespace, however there is only ever a
 * single definition.  The second addition, just overwrites the original
 * forward declaration with the same information.  This does not cause any
 * problems but it may be implementation dependent.  If the namespace
 * implementation changes, this could be a problem in the future.
 *
 * @return    Specifies whether or not this statement was a forward
 *            declaration. A forward declaration is one that has an option
 *            such as <code>FORWARD</code>, <code>IN SUPER</code> or
 *            <code>IN proc_handle</code> specified.  Such declarations do
 *            not have associated blocks.
 */
func_stmt returns [boolean fwd = false]
   :
      {
         int     ftype = -1;
         int     size  = -1;
         boolean privateAccess = false;
         boolean save          = false;
         boolean in_super_or_h = false;
      }   
      (
         fct:KW_FUNCT^ s:malformed_symbol!
         ftype = fr:func_return!
         (
            size = extent
            {
               if (size == -1 || size > 0) ##.putAnnotation("extent",  new Long(size));
            }
         )?
         (KW_PRIVATE { privateAccess = true; })?
         (pl:param_list_definition[true, true])?
         
         (
              KW_FORWARD                                      { fwd = true; save = true; }
            | (map_to_actual_name)? in_super_or_in_handle     { fwd = true; in_super_or_h = true; }
         )? { nestLevel < 2 || (insideClass && nestLevel == 2 && in_super_or_h)}?
         
         // although the FORWARD, IN SUPER and IN handle are documented to always end in a DOT,
         // it turns out that using a COLON is accepted without problem
         (DOT! | COLON!)
      )
      {
         String qname = null;
         String fname = #s.getText();
         
         if (ftype == FUNC_CLASS)
         {
            qname = (String) #fr.getChildAt(0).getAnnotation("qualified");
         }
         
         sym.addFunction(fname, ftype, qname); 
         sym.annotateFunction(fname, ##, true); 
         
         // remember if the method should be private
         if (privateAccess)
         {
            ##.putAnnotation("access", "private");
         }
         
         if (save)
         {
            // this is a forward function def, cache the local variable defs that
            // correspond to the function parameter defs
            sym.saveLocalVariables(fname);
         }
         else
         {
            if (!fwd && (#pl == null || #pl.getNumImmediateChildren() == 0))
            {
               // this is a non-forward function definition that has no parameters
               // but it is possible that a previous matching forward definition
               // did have parameters and those are implicitly implemented and allowed
               // in the "concrete" definition; restore any cached vars here
               if (sym.restoreLocalVariables(fname))
               {
                  // remember that there was something to copy
                  ##.putAnnotation("implicit_parms", new Boolean(true));
               }
            }
            else
            {
               // we don't need any cached values by this name, cleanup
               sym.removeLocalVariables(fname);
            }
         }
      }
   ;

/**
 * Matches <code>IN SUPER</code> and <code>IN handle</code> 
 * clauses in a {@link #func_stmt}.
 */
in_super_or_in_handle
   :
      KW_IN^
      (  
         options { generateAmbigWarnings = false; }
         :
           KW_SUPER
         | expr
      )
   ;

/**      
 * Matches <code>IN SUPER</code> keywords in a {@link #func_stmt} and is
 * separated to improve the tree.
 */
in_super_clause      
   :
      KW_IN^ KW_SUPER
   ;
      
/**      
 * Matches the <code>MAP TO symbol</code> construct (the <code>TO</code>
 * keyword is optional) in a {@link #func_stmt}.  It is separated to improve 
 * the tree.
 */
map_to_actual_name      
   :
      KW_MAP^ (KW_TO!)? actual:symbol
   ;      
      
/**
 * Matches the optional <code>RETURNS</code> keyword and the required
 * function return datatype (which is specified whether the 
 * <code>RETURNS</code> keyword is there or not.  This is used by 
 * {@link #func_stmt} and is only separated to make the AST build
 * correctly.  
 * <p>
 * Undocumented is the fact that one can interchangeably use the 
 * <code>RETURN</code> here instead of the documented <code>RETURNS</code>. 
 * Both are matched but the token type is forced to <code>RETURNS</code>
 * to simplify tree processing.
 *
 * @return    The token type of the function's return value.
 */
func_return returns [int ftype = -1]      
   :
      (
         options { generateAmbigWarnings = false; }
         :
         KW_RETURNS^ | r:KW_RETURN^ { #r.setType(KW_RETURNS); } 
      )? 
      ( 
         options { generateAmbigWarnings = false; }
         :
           KW_CHAR                 { ftype = FUNC_CHAR; }
         | KW_COM_HNDL             { ftype = FUNC_COM_HANDLE; }
         | KW_DATE                 { ftype = FUNC_DATE; }
         | KW_DATETIME             { ftype = FUNC_DATETIME; }
         | KW_DATE_TZ              { ftype = FUNC_DATETIME_TZ; }
         | KW_DEC                  { ftype = FUNC_DEC; }
         | KW_INT                  { ftype = FUNC_INT; }
         | KW_INT64                { ftype = FUNC_INT64; }
         | (KW_LOGICAL | k:KW_LOG  { #k.setType(KW_LOGICAL); })
                                   { ftype = FUNC_LOGICAL; }
         | KW_LONGCHAR             { ftype = FUNC_LONGCHAR; }
         | KW_RECID                { ftype = FUNC_RECID; }
         | KW_ROWID                { ftype = FUNC_ROWID; }
         | KW_HANDLE               { ftype = FUNC_HANDLE; }
         | KW_WID_HAND             { ftype = FUNC_HANDLE; }
         | KW_RAW                  { ftype = FUNC_RAW; }
         | KW_MEMPTR               { ftype = FUNC_MEMPTR; }
         | (
              options { generateAmbigWarnings = false; }
              :
              KW_CLASS!
           )?
           user_defined_type_name  { ftype = FUNC_CLASS; }
      )
   ;
   
/**
 * Matches a {@link #symbol}, a <code>DB_SYMBOL</code> or a 
 * <code>FILENAME</code> and always returns a <code>SYMBOL</code> node.
 */
any_non_reserved_symbol
   :
      (
           symbol
         | d:DB_SYMBOL { #d.setType(SYMBOL); }
         | f:FILENAME  { #f.setType(SYMBOL); }
      )
   ;
      
/**
 * Matches a {@link #reserved_or_symbol}, a <code>DB_SYMBOL</code>,
 * <code>BOOL_TRUE</code>, <code>BOOL_FALSE</code> or a<code>FILENAME</code>
 * and always returns a <code>SYMBOL</code> node.
 */
any_symbol_at_all
   :
      (                    
           reserved_or_symbol
         | y:BOOL_TRUE  { #y.setType(SYMBOL); }
         | n:BOOL_FALSE { #n.setType(SYMBOL); }
         | d:DB_SYMBOL  { #d.setType(SYMBOL); }
         | f:FILENAME   { #f.setType(SYMBOL); }
      )
   ;
      
/**
 * Matches the <code>DEFINE</code> Progress 4GL language statement and
 * switches into the correct alternative based on the lookahead tokens.
 * <p>
 * This rule is called by {@link #stmt_list} and it handles the matching
 * for the <code>DEFINE</code> keyword, the optional shared scope keywords
 * <code>NEW, GLOBAL, SHARED</code> and the optional keywords for parameter
 * types <code>INPUT, OUTPUT, INPUT-OUTPUT and RETURN</code>.  It then
 * calls the appropriate alternative to allow the specific processing to
 * occur.  The alternatives:  
 * <ul>
 *    <li> BROWSE {@link #def_browse_stmt}
 *    <li> BUTTON {@link #def_button_stmt}
 *    <li> BUFFER {@link #def_buf_stmt}
 *    <li> DATASET {@link #def_dataset_stmt}
 *    <li> DATA-SOURCE {@link #def_datasrc_stmt}
 *    <li> EVENT {@link #def_event_stmt}
 *    <li> FRAME {@link #def_frame_stmt}
 *    <li> IMAGE {@link #def_image_stmt}
 *    <li> MENU and SUB-MENU {@link #def_menu_stmt}
 *    <li> PARAMETER {@link #def_parm_stmt}
 *    <li> PROPERTY {@link #def_prop_stmt}
 *    <li> RECTANGLE {@link #def_rect_stmt}
 *    <li> STREAM {@link #def_stream_stmt}
 *    <li> QUERY {@link #def_query_stmt}
 *    <li> TEMP-TABLE (and WORK-TABLE/WORKFILE) {@link #def_temp_table_stmt}
 *    <li> VARIABLE {@link #def_var_stmt}
 * </ul>
 * <p>
 * This design allows all <code>DEFINE</code> statements to be supported
 * without ambiguity because it left-factors the common keyword.
 */
define_stmt
   :
      {
         int     mtype      = -1;
         String  varname    = null;
         String  tablename  = null;
         boolean force      = false;
         boolean fakeBuffer = false;
         Aast    likeast    = null;
         String  likename   = null;
         boolean tt_new     = false;
         boolean tt_global  = false;
         boolean tt_shared  = false;
         
         boolean[] allowColon = new boolean[] { false };
         
         Object[] df = null;
      }
      d:KW_DEFINE^
      (
           am:access_mode
         | st:KW_STATIC
         | ab:KW_ABSTRACT  // ignored for now
         | ov:KW_OVERRIDE  // ignored for now
      )*
      (
           (
                { #am == null }?
                (KW_NEW (KW_GLOBAL { tt_global = true; } )? { tt_new = true; } )? 
                KW_SHARED { tt_shared = true; }
              | // empty alternative
           )
           (
                varname=def_browse_stmt             { #d.setType(DEFINE_BROWSE); }
              | def_buf_stmt[true, false, #am, #st] { #d.setType(DEFINE_BUFFER); }
              | def_dataset_stmt[#am, #st]          { #d.setType(DEFINE_DATASET); }
              | def_datasrc_stmt[#am, #st]          { #d.setType(DEFINE_DATA_SOURCE); }
              | df = def_frame_stmt
                {
                   #d.setType(DEFINE_FRAME);
                  
                   String frame = (String) df[0];
                   Aast fp = (Aast) df[1];
                   setUseDictExps(frame, fp, #d);
                }
              
                // also handles SUB-MENU
              | mtype=def_menu_stmt
                {
                   if (mtype == WID_MENU)
                   {
                      #d.setType(DEFINE_MENU);
                   }
                   else
                   {
                      #d.setType(DEFINE_SUB_MENU);
                   }
                   varname = ##.getImmediateChild(SYMBOL, null).getText();

                   likeast = ##.getImmediateChild(KW_LIKE, null);
                   if (likeast != null)
                   {
                      likename = likeast.getImmediateChild(SYMBOL, null).getText();
                      sym.addMenuLike(varname, likename, ##);
                   }
                }           
              | def_query_stmt[#am, #st]       { #d.setType(DEFINE_QUERY); }
              | def_stream_stmt      { #d.setType(DEFINE_STREAM); }
              | tablename = t:def_temp_table_stmt[tt_new, tt_global, tt_shared, #am, #st]
                // also handles WORK-TABLE/WORKFILE
                {
                   fakeBuffer = true;
                   if (#t.getType() == KW_TEMP_TAB)
                   {
                      #d.setType(DEFINE_TEMP_TABLE);
                   }
                   else
                   {
                      #d.setType(DEFINE_WORK_TABLE);
                   }
                }
              | varname=def_var_stmt[allowColon] { #d.setType(DEFINE_VARIABLE); }
           )
         | (k1:KW_INPUT | k2:KW_OUTPUT | k3:KW_IN_OUT | k4:KW_RETURN)? 
           varname=def_parm_stmt
           {
              #d.setType(DEFINE_PARAMETER);
              if (#k1 == null && #k2 == null && #k3 == null && #k4 == null)
              {
                 force = true;
              }
           }
         | varname=def_prop_stmt                 { #d.setType(DEFINE_PROPERTY); }
         | varname=def_event_stmt                { #d.setType(DEFINE_EVENT); }
         | varname=def_button_stmt[allowColon]   { #d.setType(DEFINE_BUTTON); }
         | varname=def_rect_stmt[allowColon]     { #d.setType(DEFINE_RECTANGLE); }
         | varname=def_image_stmt                { #d.setType(DEFINE_IMAGE); }
      )
      (
           // the trigger phrase that can appear in some of these statements can have
           // an END TRIGGERS statement that is closed with either a COLON or DOT 
           { allowColon[0] }?
           COLON!
         | DOT!
      )
      {
         if (varname != null)
         {
            // pass the AST with all options to the symbol resolver to set
            // those variable options accordingly
            sym.setVariableOptions(varname, ##);
            
            // write any non-standard options into annotations
            sym.annotateVariableOptions(varname, ##, true);
            
            // check if this is a define_parameter and if the parameter type
            // was specified
            if (force)
            {
               // no type specified, default to KW_INPUT
               ##.putAnnotation("parmtype", new Long(KW_INPUT));
            }
            
            int ttype = ##.getType();
            
            // force this as a data member of any active class definition (all
            // 3 types share the same namespace)
            if (ttype == DEFINE_VARIABLE || ttype == DEFINE_PROPERTY ||
                ttype == DEFINE_EVENT)
            {
               String qname = null;
            
               // calc the fully qualified name for class vars and props
               if (ttype == DEFINE_VARIABLE || ttype == DEFINE_PROPERTY)
               {
                  Aast as = ##.getImmediateChild(KW_AS, null);
                  
                  // look for an AS clause
                  if (as != null)
                  {
                     // if this is an object instance reference, get the class
                     // name of the type
                     Aast cname = as.getChildAt(0);
                     
                     if (cname.getType() == CLASS_NAME)
                     {
                        qname = (String) cname.getAnnotation("qualified");
                     }
                  }
                  else
                  {
                     // must be a LIKE clause
                     Aast like  = ##.getImmediateChild(KW_LIKE, null);
                     Aast lval  = like.getChildAt(0);
                     int  ltype = lval.getType();
                     
                     if (ltype == VAR_CLASS || ltype == FIELD_CLASS)
                     {
                        qname = (String) lval.getAnnotation("qualified");
                     }
                  }
               }
            
               sym.addDataMember(varname, #am, #st, qname);
               
               if (qname != null)
                  sym.annotateClassRef(qname, ##);
            }
         }
         
         if (fakeBuffer)
         {
            // now store an annotation with the fully qualified name and the
            // shorted unabbreviated unique buffer name
            String fullname = sym.lookupFullTableName(tablename);
            String bufname  = sym.lookupBufferName(tablename, false);
            String dbname   = sym.lookupDatabaseName(tablename);
                     
            if (fullname != null)         
               ##.putAnnotation("schemaname", fullname);
               
            if (bufname != null)
               ##.putAnnotation("bufname", bufname);
               
            if (dbname != null)
               ##.putAnnotation("dbname", dbname);
         }
      }
   ;
      
/**
 * Matches a <code>DEFINE PARAMETER</code> Progress 4GL language statement.
 * This is the statement that defines or imports new user-named parameters
 * for a procedure (internal or external).  
 * <p>
 * This rule is called by {@link #define_stmt} which is where the matching
 * on the <code>DEFINE</code> keyword and the optional shared scope for
 * variables, temp-tables and streams.  This rule supports shared, global
 * shared or local variables and the syntax for both declaring or defining
 * (<code>NEW</code> keyword) shared variables.  HOWEVER, all the matching
 * for these optional keywords is done in the <code>define_stmt</code> rule. 
 * This refactoring allows all <code>DEFINE</code> statements to be supported
 * without ambiguity.
 * <p> 
 * This rule matches the <code>PARAMETER</code> keyword and then calls
 * other rules to implement the different alternatives.
 * <p>
 * Calls {@link #regular_parm}, {@link #table_parm} and {@link #def_buf_stmt}.
 *
 * @return   The variable name of the variable that was defined or
 *           <code>null</code> if this is a record or buffer reference
 *           instead.
 */
def_parm_stmt returns [String varname = null]     
   :
      p:KW_PARM! 
      (
         options { generateAmbigWarnings = false; }
         :
           varname=table_parm
         | def_buf_stmt[false, false, null, null]
         | varname=regular_parm[#p, null]
      ) 
   ;
      
/**
 * Matches a <code>DEFINE BROWSE</code> Progress 4GL language statement.
 * This is the statement that defines or imports a new user-named browse
 * widget.
 * <p>
 * This rule is called by {@link #define_stmt} which is where the matching
 * on the <code>DEFINE</code> keyword and the optional shared scope for
 * browses.  This rule supports shared or local buffers and the syntax for
 * both declaring or defining (<code>NEW</code> keyword) shared buffers. 
 * HOWEVER, all the matching for these optional keywords is done in the
 * <code>define_stmt</code> rule. This refactoring allows all 
 * <code>DEFINE</code> statements to be supported without ambiguity.
 * <p>
 * This rule matches the <code>BROWSE</code> keyword and all subsequent
 * processing.  Calls {@link #symbol}, {@link #query_reference} and
 * {@link #column_spec} rules.
 * <p>
 * This rule adds the browse's name to the widget namespace.
 * <p>
 * Supports the Progress undocumented feature that the query_reference and
 * the column_spec can be optional (missing).  But even if they are missing,
 * the {@link #browse_options_phrase} can still be included.  For this reason,
 * the browse_options_phrase is directly referenced instead of being included
 * at the end of the column_spec.
 * <p>
 * @return   The variable name of the browse that was defined.
 */
def_browse_stmt returns [String varname = null]      
   :
      (
         KW_BROWSE! b:symbol
         {
            // this must be done here (as soon as the symbol is processed
            // since the browse name can be referenced later in the same
            // language stmt and so the namespace must include this widget
            // (e.g. a validate() option in the column_spec is one example
            // of how it could be referenced)
            varname = #b.getText();         
            sym.addWidget(varname, WID_BROWSE);
         }
         (
            options { generateAmbigWarnings = false; }
            :
              { LA(1) != KW_DISP && LA(1) != KW_WITH }?
              query_reference
              (
                   (KW_SH_LOCK | KW_EXC_LOCK | KW_NO_LOCK)
                 | KW_NO_WAIT
              )*
            | // empty alternative
         )
         (
            options { generateAmbigWarnings = false; }
            :
              { LA(1) == KW_DISP }?
              (column_spec)
            |
              (browse_options_phrase)*
         )
      )
   ;      
      
/**
 * Matches a <code>DEFINE QUERY</code> Progress 4GL language statement.
 * This is the statement that defines or imports a new user-named query.
 * <p>
 * This rule is called by {@link #define_stmt} which is where the matching
 * on the <code>DEFINE</code> keyword and the optional shared scope for
 * browses.  This rule supports shared or local buffers and the syntax for
 * both declaring or defining (<code>NEW</code> keyword) shared buffers. 
 * HOWEVER, all the matching for these optional keywords is done in the
 * <code>define_stmt</code> rule. This refactoring allows all 
 * <code>DEFINE</code> statements to be supported without ambiguity.
 * <p>
 * This rule matches the <code>QUERY</code> keyword and all subsequent
 * processing.  Calls {@link #malformed_symbol}, {@link #for_record_spec} and
 * {@link #cache_records} rules.
 *
 * @param    am
 *           Access modifiers if not null. Only will be non-null if this is a resource defined
 *           as a member of a class definition.
 * @param    st
 *           Static specifier if not null. Only will be non-null if this is a resource defined
 *           as a member of a class definition.
 */
def_query_stmt[Aast am, Aast st]
   :
      KW_QUERY^ q:malformed_symbol for_record_spec[true, false, false]
      (
           cache_records
         | KW_SCROLLIN
         | KW_RCOD_INF
      )*
      {
         sym.addQuery(#q.getText(), QUERY, am, st);
      }
   ;
      
/**      
 * Matches the <code>CACHE</code> option for the <code>DEFINE QUERY</code>
 * language statement.
 * <p>
 * Called by {@link #def_query_stmt}.
 * <p>
 * Separated from the main rule to properly build the AST without complexity.
 * This is the only reason such a simple rule is implemented separately.
 */ 
cache_records      
   :
      KW_CACHE^ NUM_LITERAL
   ;
      
/**      
 * Matches the manner of specifying a list of records (and optional include or 
 * exclude list of fields) for certain database related language statements.
 * Calls {@link #record} and {@link #field_list} rules. 
 * <p>
 * Called by {@link #def_query_stmt}, {@link #def_dataset_stmt},
 * {@link #def_datasrc_stmt} and {@link #do_repeat_stmt} rules.
 *
 * @param    flds
 *           <code>true</code> if the field list option should be allowed to
 *           be matched.
 * @param    keys
 *           <code>true</code> to match the keys clause after a record.
 * @param    force
 *           <code>true</code> to force promotion even if the record was
 *           already previously promoted.
 */ 
for_record_spec [boolean flds, boolean keys, boolean force]
   :
      KW_FOR^ record[true, force, false]
      ( { flds }? field_list | { keys }? datasrc_keys_clause )?
      (
         COMMA! record[true, force, false]
         ( { flds }? field_list | { keys }? datasrc_keys_clause )?
      )*
   ;
   
/**
 * Matches the <code>KEYS</code> clause in a source buffer specification of
 * the <code>DEFINE DATA-SOURCE</code> statement.  Called directly by
 * {@link #for_record_spec}. 
 */
datasrc_keys_clause
   :
      KW_KEYS^
      LPARENS!
      (
         options { generateAmbigWarnings = false; }
         :
         KW_ROWID | (lvalue (COMMA! lvalue)*)
      )
      RPARENS!
   ;
      
/**      
 * Matches a simple <code>FOR</code> and a record name for the
 * {@link #def_buf_stmt}. 
 * <p>
 * Calls {@link #record}.
 * <p>
 * Separated from the main rule to properly build the AST without complexity.
 * This is the only reason such a simple rule is implemented separately.
 *
 * @param    promote
 *           <code>true</code> to force promotion of the referenced table.
 *
 * @return   The name of the first record referenced.
 */ 
simple_for_record_spec [boolean promote] returns [String rtext = null]
   :
      {
         boolean old = forceTemp;
      }
      
      KW_FOR^
      (
         options { generateAmbigWarnings = false; }
         :
         KW_TEMP_TAB { forceTemp = true; }
      )?             
      r:record[promote, false, false]
      
      {
         forceTemp = old;
         rtext = #r.getText();
      }
   ;
   
/**
 * Matches a <code>DEFINE BUFFER</code> Progress 4GL language statement.
 * This is the statement that defines or imports new user-named buffers and
 * associates these with a specific database table, temp-table or  
 * work-table.
 * <p>
 * This rule is called by {@link #define_stmt} which is where the matching
 * on the <code>DEFINE</code> keyword and the optional shared scope for
 * buffers.  This rule supports shared or local buffers and the syntax for
 * both declaring or defining (<code>NEW</code> keyword) shared buffers. 
 * HOWEVER, all the matching for these optional keywords is done in the
 * <code>define_stmt</code> rule. This refactoring allows all 
 * <code>DEFINE</code> statements to be supported without ambiguity.
 * <p>
 * This rule matches the <code>BUFFER</code> keyword and all subsequent
 * processing.  Note that due to syntax compatibility, this rule is also
 * called from {@link #def_parm_stmt} and {@link #parameter} which allows 
 * this logic to be centralized.  The following <code>FOR</code> construct
 * is only optional to support the call from the <code>parameter</code>
 * rule when it is processing a <code>RUN</code> statement parameter.  When
 * the <code>parameter</code> rule is processing a <code>FUNCTION</code> 
 * parameter, the <code>FOR</code> processing will be matched (if the code
 * is valid Progress 4GL). Anytime the <code>FOR</code> processing is matched,
 * this means that this is a new or imported buffer definition and the
 * buffer will be added to the schema namespace.
 * <p>
 * This supports an undocumented Progress feature where function parameters
 * can be specified without the symbolic name (e.g. in a forward function
 * declaration it can look like <code>BUFFER FOR record</code>).
 *
 * @param    define_stmt
 *           If <code>true</code>, this rule has been called from the
 *           <code>DEFINE BUFFER</code> statement and the 
 *           <code>KW_BUFFER</code> node should be dropped (it is 
 *           unnecessary).  If <code>false</code>, the caller is some form
 *           of parameter processing and the <code>KW_BUFFER</code> node is 
 *           used as the resulting sub-tree root.
 * @param    optionalSym
 *           If <code>true</code>, this rule has been called from a function
 *           parameter and the buffer's symbolic name is optional. 
 * @param    am 
 *           Access modifiers if not null. Only will be non-null if this is a resource defined
 *           as a member of a class definition.
 * @param    st
 *           Static specifier if not null. Only will be non-null if this is a resource defined
 *           as a member of a class definition.
 */
def_buf_stmt [boolean define_stmt, boolean optionalSym, Aast am, Aast st]       
   :
      {
         boolean promote = !define_stmt;
         String rtext = null;
      }
      (
         (
              { define_stmt }?
              KW_BUFFER!
            | KW_BUFFER^
         )
         (
            options { generateAmbigWarnings = false; }
            :
              b:symbol
            | { optionalSym }?
              // empty alternative
         )         
         ( rtext=simple_for_record_spec[promote] )?
         (
              KW_PRESEL
            | label
            | namespace_clause
         )*
      )
      {
         if (#b != null && rtext != null)
         {
            boolean forceTemp = false;
            AST ns = #b.getNextSibling();
            if (ns != null && ns.getType() == KW_FOR)
            {
               AST a = ns.getFirstChild();
               if (a != null && a.getType() == KW_TEMP_TAB)
               {
                  forceTemp = true;
               }
            }
            defineBufferFromSymbol(#b, rtext, am, st, forceTemp);
         }
      }
   ;
   
/**
 * Matches the <code>DEFINE DATASET</code> language statement.
 * <p>
 * This rule is called by {@link #define_stmt} which is where the matching
 * on the <code>DEFINE</code> keyword and the optional access modes or sharing
 * options.
 * <p>
 * This rule matches the <code>DATASET</code> keyword and a following
 * {@link #symbol}. The {@link #namespace_clause} can be matched optionally.
 * Then there is always a match to {@link #for_record_spec}. At the end is
 * the use of {@link #data_relation}.
 * <p>
 * The dataset name will be added to a unique namespace such that the
 * dataset can be referenced as a handle and provided as a parameter.
 * <p>
 * Customer source code shows the DATA-RELATION is optional even though the
 * Progress docs state otherwise. This must be an undocumented feature.
 *
 * @param    am
 *           Access modifiers if not null. Only will be non-null if this is a resource defined
 *           as a member of a class definition.
 * @param    st
 *           Static specifier if not null. Only will be non-null if this is a resource defined
 *           as a member of a class definition.
 */
def_dataset_stmt[Aast am, Aast st]
   :
      {
         int idx = 1;
         int pidx = 1;
      }
      KW_DATASET! d:symbol
      (
           namespace_clause
         | xml_node_name
         | xml_node_type
         | serialize_name
         | KW_SERIALZH
         | KW_REF_ONLY
      )*
      for_record_spec[false, false, false]
      ( 
           data_relation[idx]       { idx++; }
         | parent_id_relation[pidx] { pidx++; }
      )*
      {
         // add the dataset to the namespace
         sym.addDataSet(#d.getText(), DATA_SET, am, st);
      }
   ;
   
/**
 * Matches the <code>DEFINE DATA-SOURCE</code> language statement.
 * <p>
 * This rule is called by {@link #define_stmt} which is where the matching
 * on the <code>DEFINE</code> keyword and the optional access modes or sharing
 * options.
 * <p>
 * This rule matches the <code>DATA-SOURCE</code> keyword and a following
 * {@link #symbol}. Then there is always a match to {@link #for_query_clause}
 * or a {@link #for_record_spec}.
 * <p>
 * The data source name will be added to a unique namespace such that the
 * data source can be referenced as a handle and provided as a parameter.
 *
 * @param    am
 *           Access modifiers if not null. Only will be non-null if this is a resource defined
 *           as a member of a class definition.
 * @param    st
 *           Static specifier if not null. Only will be non-null if this is a resource defined
 *           as a member of a class definition.
 */
def_datasrc_stmt[Aast am, Aast st]
   :
      KW_DATA_SRC! d:symbol
      (
           { LA(2) == KW_QUERY }?
           for_query_clause
         | for_record_spec[false, true, false]
      )
      {
         // add the data source to the namespace
         sym.addDataSource(#d.getText(), DATA_SOURCE, am, st);
      }
   ;
   
/**
 * Match the <code>FOR</code> keyword followed by a {@link #query_reference}.
 * Used by {@link #def_datasrc_stmt}.
 */
for_query_clause
   :
      KW_FOR^ query_reference
   ;
   
/**
 * Matches the <code>NAMESPACE-URI</code> or <code>NAMESPACE-PREFIX</code>
 * with a following <code>STRING</code>. Used by {@link #def_dataset_stmt}
 * {@link #def_temp_table_stmt} and {@link #def_buf_stmt}.
 */
namespace_clause
   :
      (KW_NAMESP_U^ | KW_NAMESP_P^) STRING 
   ;
   
/**
 * Matches a <code>DATA-RELATION</code> clause in a {@link #def_dataset_stmt}.
 * There is an optional {@link #symbol} which will be manufactured based on
 * the given index number if the relation name is not there. The naming
 * scheme is to use <code>Relation</code> and append the index number to the
 * end. This name (manufactured or provided) is added to a namespace for
 * data relations which can then be de-referenced in use as a handle.
 *
 * @param    idx
 *           1-based index number of the data relation in a given dataset
 *           definition.
 */
data_relation! [int idx]
   :
      dr:KW_DATA_REL (d:symbol)? f:for_data_relation_spec
      {
         if (#d == null)
         {
            // create an artificial node
            #d = #[SYMBOL, String.format("Relation%d", idx)];
         }
         
         String name = #d.getText();
         
         // add to the relation namespace
         sym.addDataRelation(name, DATA_RELATION);
         
         // build the tree
         ## = #(dr, d, f);
      }
   ;
   
/**
 * Matches a <code>FOR</code> followed by two {@link #record} references
 * and options including {@link #field_mapping_clause}. Used by
 * {@link #data_relation}.
 */
for_data_relation_spec
   :
      KW_FOR^ record[true, false, false] COMMA! record[true, false, false]
      (
           KW_REPOS
         | KW_RECURSE
         | KW_NOT_ACTV
         | nested_clause
         | field_mapping_clause
      )*
   ;
   
/**
 * Matches a <code>PARENT-ID-RELATION</code> clause in a {@link #def_dataset_stmt}.
 * There is an optional {@link #symbol} which will be manufactured based on
 * the given index number if the relation name is not there. The naming
 * scheme is to use <code>ParentRelation</code> and append the index number to the
 * end. This name (manufactured or provided) is added to a namespace for
 * data relations which can then be de-referenced in use as a handle.
 *
 * @param    idx
 *           1-based index number of the data relation in a given dataset
 *           definition.
 */
parent_id_relation! [int idx]
   :
      dr:KW_PAR_IREL (d:symbol)? f:for_parent_relation_spec
      {
         if (#d == null)
         {
            // create an artificial node
            #d = #[SYMBOL, String.format("ParentRelation%d", idx)];
         }
         
         // TODO: add namespace processing for parent id relations, add whatever lookup is needed
         //       in lvalue presumeably
         // String name = #d.getText();
         
         // add to the relation namespace
         /// sym.addParentIdRelation(name, PARENT_ID_RELATION);
         
         // build the tree
         ## = #(dr, d, f);
      }
   ;
   
/**
 * Matches a <code>FOR</code> followed by two {@link #record} references
 * and options including {@link #field_mapping_clause}. Used by
 * {@link #parent_id_relation}.
 */
for_parent_relation_spec
   :
      KW_FOR^ record[true, false, false] COMMA! record[true, false, false]
      parent_id_field_clause (parent_fields_ordering_clause)*
   ;
   
/**
 * Matches a <code>PARENT-ID-FIELD</code> followed by an {@link #lvalue} reference.
 * <p>
 * Used by {@link #for_parent_relation_spec}.
 */
parent_id_field_clause
   :
      KW_PAR_IFLD^ lvalue
   ;
   
/**
 * Matches a <code>PARENT-FIELDS-[BEFORE|AFTER]</code> followed by parenthesized list of
 * one or more comma separated {@link #lvalue} references.
 * <p>
 * Used by {@link #for_parent_relation_spec}.
 */
parent_fields_ordering_clause
   :
      (KW_PAR_FLDB^ | KW_PAR_FLDA^) LPARENS! lvalue (COMMA! lvalue)* RPARENS!
   ;
   
/**
 * Matches a <code>NESTED</code> keyword followed by an optional <code>FOREIGN-KEY-HIDDEN</code>
 * keyword. Called by {@link #for_data_relation_spec}. 
 */
nested_clause
   :
      KW_NESTED^ (KW_F_KEY_H)?
   ;
   
/**
 * Matches a <code>RELATION-FIELDS</code> keyword followed by a list of
 * one or more {@link #parent_child_field_relation} references. Called by
 * {@link #for_data_relation_spec}. 
 */
field_mapping_clause
   :
      KW_REL_FLDS^ LPARENS! (parent_child_field_relation (COMMA!)?)+ RPARENS!
   ;
   
/**
 * Matches a pair of {@link #lvalue} references and roots the pair at a
 * <code>PARENT_CHILD_RELATION</code> node. Called by 
 * {@link #field_mapping_clause}.
 */
parent_child_field_relation
   :
      {
         astFactory.makeASTRoot(currentAST, #[PARENT_CHILD_RELATION,""]);
      }
      lvalue COMMA! lvalue
   ;

/**
 * Matches a <code>DEFINE MENU or DEFINE SUB-MENU</code> Progress 4GL
 * language statement. This is the statement that defines user-named 
 * menus and sub-menus.  Menus are stored in their own namespace which is
 * separate from the other widgets.
 * <p>
 * This rule is called by {@link #define_stmt} which is where the matching
 * on the <code>DEFINE</code> keyword.
 * <p>
 * This rule matches the <code>MENU or SUB-MENU</code> keywords and all
 * subsequent processing.  Calls the following rules:
 * <p>
 * {@link #symbol}
 * {@link #ui_stuff}
 * {@link #simple_title_string}
 * {@link #like_menu_clause}
 * {@link #menu_element_descriptor}
 * <p>
 * All options can be processed in any order.
 * <p>
 * An undocumented "feature" (found in customer code) is that Progress will
 * ignore a <code>DOT</code> in between <code>MENU-ITEM</code> clauses. In
 * other words, it will silently drop that extra character if the period
 * is followed by the <code>MENU-ITEM</code> or <code>RULE</code> keywords.
 * Presumably, the <code>SUB-MENU</code> and <code>SKIP</code> keywords also
 * have this effect but it hasn't been confirmed yet.
 */
def_menu_stmt returns [int mtype = -1]
   :
      {
         mtype = WID_MENU;
      }
      (KW_MENU! | KW_SUB_MENU! { mtype = WID_SUB_MENU; } ) m:symbol
      (
         options { generateAmbigWarnings = false; }
         :
           ui_stuff
         | (KW_MENU_BAR | KW_SUB_M_H | simple_title_string) 
         | (like_menu_clause | menu_element_descriptor)
         
           // TODO: We would improve this exlusion code from just using COLON.
         | { (LA(2) == KW_MENU_ITM && LA(4) != COLON) || 
             (LA(2) == KW_SUB_MENU && LA(4) != COLON) ||
             LA(2) == KW_RULE     || LA(2) == KW_SKIP }?
           DOT!
      )*
      {
         sym.addMenu(#m.getText(), mtype);
      }      
   ;
      
/**   
 * Matches the <code>RULE or SKIP</code> keywords or the other possible
 * constructs {@link #menu_item_phrase} and {@link #submenu_descriptor} that
 * can be used as a menu element.  Called from {@link #def_menu_stmt}.
 * <p>
 * By making a separate rule for this, the logic is centralized and the
 * AST is kept simple because all tokens can be rooted at this option's
 * <code>SUB-MENU</code> keyword.
 */
menu_element_descriptor
   :
      (
           KW_RULE^
         | KW_SKIP^
         | submenu_descriptor
         | menu_item_phrase
      )
   ;
   
/**   
 * Matches the <code>SUB-MENU</code> keyword with following sub-menu name
 * and options.  This is used in a {@link #menu_element_descriptor}.
 * <p>
 * At this time, the <code>symbol</code> is a placeholder for proper
 * menu namespace processing.
 * <p>
 * This is very similar to the {@link #menu_reference} rule but it includes
 * a more specific keyword and has options.  The two rules can't be used as
 * alternatives because they are ambiguous.
 * <p>
 * By making a separate rule for this, the logic is centralized and the
 * AST is kept simple because all tokens can be rooted at this option's
 * <code>SUB-MENU</code> keyword.
 */
submenu_descriptor   
   :
      KW_SUB_MENU^ lvalue (label | KW_DISABLED)*
   ;
      
/**
 * Matches a <code>MENU-ITEM</code> definition in the Progress 4GL
 * <code>DEFINE MENU or DEFINE SUB-MENU</code> language statements. 
 * Menus, sub-menus and menu-items are stored in their own namespace which is
 * separate from the other widgets, though at this time there is no actual
 * menu namespace processing in this parser.
 * <p>
 * This rule is called by {@link #menu_element_descriptor}.
 * <p>
 * This rule matches the <code>MENU-ITEM</code> keyword and all
 * subsequent processing.  Calls the following rules:
 * <p>
 * {@link #symbol}
 * {@link #label}
 * {@link #ui_stuff}
 * <p>
 * All options can be processed in any order.
 * <p>
 * The <code>ui_stuff</code> rule generates bogus ambiguity warnings with
 * the exit branch of the loop. Ambiguity warnings have been disabled.
 */
menu_item_phrase      
   :
      KW_MENU_ITM^ itm:symbol
      (
         options { generateAmbigWarnings = false; }
         :
           KW_READ_ONL
         | KW_TOGGL_BX
         | KW_DISABLED
         | label
         | ui_stuff
         
           // elsewhere we are supporting the allowColon[] functionality but the use of colon
           // for menus has other implications and we haven't tested the combinations to see
           // what is valid; for now we are just disabling this behavior for menus
         | trigger_phrase
      )*
      {
         String txt = #itm.getText();
         sym.addMenuItem(txt, WID_MENU_ITM);

         // pass the AST with all options to the symbol resolver to set
         // those variable options accordingly
         sym.setVariableOptions(txt, ##);

         // write any non-standard options into annotations
         sym.annotateVariableOptions(txt, ##, true);
      }      
   ;
            
/**
 * Matches a <code>DEFINE STREAM</code> Progress 4GL language statement.
 * This is the statement that defines or imports new user-named stream.
 * <p>
 * This rule is called by {@link #define_stmt} which is where the matching
 * on the <code>DEFINE</code> keyword and the optional shared scope for
 * buffers.  This rule supports shared or local buffers and the syntax for
 * both declaring or defining (<code>NEW</code> keyword) shared buffers. 
 * HOWEVER, all the matching for these optional keywords is done in the
 * <code>define_stmt</code> rule. This refactoring allows all 
 * <code>DEFINE</code> statements to be supported without ambiguity.
 * <p>
 * This rule matches the <code>STREAM</code> keyword and the required
 * stream name. This rule is similar to the {@link #lvalue}
 * but it is separated such that namespace additions can be done without
 * impacting that rule.
 */
def_stream_stmt      
   :
      KW_STREAM^ stream:malformed_symbol
      { sym.addStream(#stream.getText(), STREAM); }
   ;
      
/**
 * Matches a <code>DEFINE BUTTON</code> Progress 4GL language statement.
 * This is the statement that defines user-named button widgets.
 * <p>
 * This rule is called by {@link #define_stmt} which is where the matching
 * on the <code>DEFINE</code> keyword.
 * <p>
 * This rule matches the <code>BUTTON</code> keyword and all subsequent
 * processing.  Calls the following rules:
 * <p>
 * {@link #symbol}
 * {@link #ui_stuff}
 * {@link #image_stuff}
 * {@link #label}
 * {@link #like_widget_clause}
 * {@link #size_phrase}
 * {@link #no_focus_clause}
 * {@link #tooltip_clause}
 * {@link #trigger_phrase}
 * <p>
 * All options can be processed in any order.
 * <p>
 * This supports the undocumented <code>MARGIN-EXTRA</code> option.
 * <p>
 * This rule updates the widget namespace with the new button widget however
 * there is no backing variable for this widget.
 *
 * @param    allowColon
 *           Set to <code>true</code> on return to tell the calling code to match either
 *           a DOT or a COLON to end the statement (this will be set only if a triggers
 *           phrase is matched).
 *
 * @return   The variable name of the button that was defined.
 */
def_button_stmt [boolean[] allowColon] returns [String varname = null]      
   :
      (
         KW_BUTTON!
         b:symbol
         (
              (KW_AUTO_GO | KW_AUTO_END)
            | KW_DEFAULT
            | KW_DROP_TAR
            | KW_MARG_EX
            | ui_stuff
            | image_stuff
            | label
            | like_widget_clause
            | size_phrase
            | no_focus_clause
            | KW_NO_CV_3D
            | tooltip_clause
            | trigger_phrase { allowColon[0] = true; }
         )*
      )
      {
         varname = #b.getText();
         sym.addWidget(varname, WID_BUTTON);
      }
   ;
   
/**
 * Matches a <code>DEFINE RECTANGLE</code> Progress 4GL language statement.
 * This is the statement that defines user-named rectangle widgets.
 * <p>
 * This rule is called by {@link #define_stmt} which is where the matching
 * on the <code>DEFINE</code> keyword.
 * <p>
 * This rule matches the <code>RECTANGLE</code> keyword and all subsequent
 * processing.  Calls the following rules:
 * <p>
 * {@link #symbol}
 * {@link #edge_clause}
 * {@link #ui_stuff}
 * {@link #like_widget_clause}
 * {@link #size_phrase}
 * {@link #tooltip_clause}
 * {@link #trigger_phrase}
 * <p>
 * All options can be processed in any order.
 * <p>
 * This rule updates the widget namespace with the new rectangle widget
 * however there is no backing variable for this widget.
 *
 * @param    allowColon
 *           Set to <code>true</code> on return to tell the calling code to match either
 *           a DOT or a COLON to end the statement (this will be set only if a triggers
 *           phrase is matched).
 *
 * @return   The rectangle name that is being defined.
 */
def_rect_stmt [boolean[] allowColon] returns [String varname = null]   
   :
      KW_RECT!
      r:symbol
      (
           KW_NO_FILL
         | KW_GRAPHIC
         | KW_ROUNDED
         | KW_GROUP_BX
         | edge_clause
         | ui_stuff
         | like_widget_clause
         | size_phrase
         | tooltip_clause
         | trigger_phrase { allowColon[0] = true; }
      )*
      {
         varname = #r.getText();
         sym.addWidget(#r.getText(), WID_RECT);
      }
   ;
   
/**
 * Matches a <code>DEFINE IMAGE</code> Progress 4GL language statement.
 * This is the statement that defines user-named image widgets.
 * <p>
 * This rule is called by {@link #define_stmt} which is where the matching
 * on the <code>DEFINE</code> keyword.
 * <p>
 * This rule matches the <code>IMAGE</code> keyword and all subsequent
 * processing.  Calls the following rules:
 * <p>
 * {@link #symbol}
 * {@link #ui_stuff}
 * {@link #stretch_to_fit_clause}
 * {@link #image_phrase}
 * {@link #like_widget_clause}
 * {@link #size_phrase}
 * {@link #tooltip_clause}
 * <p>
 * All options can be processed in any order. In addition to these rules,
 * the <code>CONVERT-3D-COLORS</code> and <code>TRANSPARENT</code> options
 * can be specified.
 * <p>
 * This rule updates the widget namespace with the new image widget however
 * there is no backing variable for this widget.
 *
 * @return   The variable name of the browse that was defined.
 */
def_image_stmt returns [String varname = null]
   :
      KW_IMAGE!
      i:symbol
      (
           KW_CVT_3D_C
         | KW_TRANSPAR
         | stretch_to_fit_clause
         | ui_stuff
         | image_phrase
         | like_widget_clause
         | size_phrase
         | tooltip_clause
      )*
      {
         varname = #i.getText();
         sym.addWidget(varname, WID_IMAGE);
      }
   ;
   
/**
 * Support the <code>STRETCH-TO-FIT</code> with an optional 
 * <code>RETAIN-SHAPE</code>. This is separated to properly build the tree
 * rooted at the <code>KW_ST_2_FIT</code> node.  Called from
 * {@link #def_image_stmt}
 */
stretch_to_fit_clause
   :
      KW_ST_2_FIT^ (KW_RET_SHAP)?
   ;
   
/**
 * Matches a <code>DEFINE FRAME</code> Progress 4GL language statement
 * which is nearly identical to the {@link #form_stmt}.
 * <p>
 * The tricky part of this statement is handling the options in any order and
 * resolving the ambiguity between the record and field specifications.  
 * Records are disambiguated from fields (records take precedence when there
 * is a name conflict) by trying to do a lookup, if it fails the record is
 * no matched (see {@link #record_spec}) and instead the {@link #form_item} 
 * is checked.  By carefully ordering of these options and using a semantic
 * predicate, the ambiguity is resolved.  For this reason, ambiguity warnings
 * have been disabled.
 * <p>
 * Called by {@link #define_stmt}. Uses {@link #record_spec}, 
 * {@link #form_item} and {@link #frame_phrase} to properly build the AST.
 *
 * @return    An array with index 0 the frame name and on index 1 the frame_phrase AST.
 */
def_frame_stmt returns [Object[] d = new Object[2]]
   :
      {
         Set<Aast> fields = new HashSet<>();
      }
      KW_FRAME! f:malformed_symbol
      (
         options { generateAmbigWarnings = false; }
         :
           KW_DROP_TAR
         | (KW_HEADER | KW_BACKGRND)
           // we must disambiguate a record from a field (records take 
           // precedence) by trying to do a lookup, if it fails we don't
           // match this alternative
         | { isRecord(LT(1).getText()) }?
           record_spec[fields]
         | form_item[fields]
         | fp:frame_phrase[false, false]
      )*
      {
         String frame = #f.getText();
         sym.addFrame(frame, WID_FRAME);
         if (!fields.isEmpty())
         {
            sym.addFrameFields(frame, fields);
         }
         
         d[0] = frame;
         d[1] = #fp;
      }
   ;
   
/**
 * Matches a <code>FORM</code> Progress 4GL language statement which is
 * nearly identical to the {@link #def_frame_stmt}.
 * <p>
 * The tricky part of this statement is handling the options in any order and
 * resolving the ambiguity between the record and field specifications.  
 * Records are disambiguated from fields (records take precedence when there
 * is a name conflict) by trying to do a lookup, if it fails the record is
 * not matched (see {@link #record_spec}) and instead the {@link #form_item} 
 * is checked.  By carefully ordering these options and using a semantic
 * predicate, the ambiguity is resolved.  For this reason, ambiguity warnings
 * have been disabled.
 * <p>
 * Called by {@link #stmt_list}. Uses {@link #record_spec}, 
 * {@link #form_item} and {@link #frame_phrase} to properly build the AST.
 * <p>
 * Progress is so ridiculously ambiguous about keywords that there is a
 * reserved <code>FORMAT</code> keyword that can be abbreviated down to
 * 4 characters.  Yes, the reserved <code>FORM</code> keyword could actually
 * be the <code>FORMAT</code> keyword and must be disambiguated by context.
 * Our keyword processing will always generate a match to <code>FORMAT</code>
 * and we override it to be a <code>FORM</code>, once we match here.
 */
form_stmt
   :
      {
         Set<Aast> fields = new HashSet<>();
         String frame = "";
      }
      f:KW_FORMAT^   { #f.setType(KW_FORM); }
      (
         options { generateAmbigWarnings = false; }
         :
           (KW_HEADER | KW_BACKGRND)
           
           // we must disambiguate a record from a field (records take 
           // precedence) by trying to do a lookup, if it fails we don't
           // match this alternative
         | { isRecord(LT(1).getText()) }?
           record_spec[fields]
         | form_item[fields]
         | fp:frame_phrase[false, false]
           {
              if (frame.length() == 0)
              {
                 String name = getFrameName(#fp);
                 if (name.length() > 0)
                 {
                    frame = name;
                 }
              }
           }
      )*
      DOT!
      {
         if (!fields.isEmpty())
         {
            sym.addFrameFields(frame, fields);
         }

         setUseDictExps(frame, #fp, ##);
      }
   ;
   
/**
 * Matches the shorthand of specifying a record (and optional exclude list
 * of fields) instead of an explicit field list, in the 
 * {@link #def_frame_stmt}, {@link #display_stmt} and {@link #column_spec}
 * rules.  Uses {@link #except_list}.
 * <p>
 * This rule is naturally greedy (in matching a list of field names to
 * exclude, and doing so in a loop. ANTLR warns that this causes ambiguity. 
 * The greedy option is used to tell ANTLR to disable the warning, though it
 * has no impact on the actual code generation.
 * <p>
 * Separated from the main rule to properly build the AST without complexity.
 * This is the only reason such a simple rule is implemented separately.
 * <p>
 * If a non-null <code>fields</code> argument is provided, a bogus AST will be
 * manufactured to represent a business logic reference to each field in the
 * database table represented by the record spec. These are used downstream for
 * schema-level field assign validation processing.
 * <p>
 * TBD:  we always promote the associated table here, though it is unclear
 * whether this is necessary in the {@link #form_stmt} case.
 *
 * @param   fields
 *          Set into which manufactured field references should be stored. May be
 *          <code>null</code>.
 */
record_spec [Set<Aast> fields]
   :
      r:record[true, false, false] (except_list)?
      {
         if (fields != null)
         {
            String sname = (String) #r.getAnnotation("schemaname");
            String bufname = (String) #r.getAnnotation("bufname");
            Iterator<Aast> iter = sym.fields(sname);
            while (iter.hasNext())
            {
               Aast fld = iter.next();
               String fname = fld.getText();
               
               // manufacture bogus field reference AST with bufname and schemaname annotations
               Aast fldRef = new ProgressAst();
               fldRef.setType(fld.getType());
               fldRef.setText(bufname + "." + fname);
               fldRef.putAnnotation("schemaname", sname + "." + fname);
               fldRef.putAnnotation("bufname", bufname);
               
               fields.add(fldRef);
            }
         }
      }
   ;
   
/**
 * Specifies an list of fields that are not excluded from the previous record
 * or query definition.
 * <p>
 * Used by {@link #record_spec} and {@link #browse_all_list}.
 * <p>
 * This rule is naturally greedy (in matching a list of field names to
 * exclude, and doing so in a loop. ANTLR warns that this causes ambiguity. 
 * The greedy option is used to tell ANTLR to disable the warning, though it
 * has no impact on the actual code generation.
 * <p>
 * Separated from the main rule to properly build the AST without complexity.
 * This is the only reason such a simple rule is implemented separately.
 */
except_list
   :
      KW_EXCEPT^ 
      (
         options { greedy = true; }
         :
         lvalue
      )*
   ;
   
/**
 * Matches the form item construct of the Progress 4GL language.  Calls
 * {@link #space_or_skip}, {@link #widget} and {@link #expr}. Called by 
 * {@link #def_frame_stmt} and {@link #form_stmt}.  The resulting tree will be
 * rooted at a <code>FORM_ITEM</code> token.
 * <p>
 * This is ambiguity between the {@link #widget} rule and the expression/
 * variable definition mode where the rule uses a {@link #symbol}
 * for matching.  Since both rules can can reference a symbol in the
 * leftmost position, lookups are done in the widget namespace to
 * disambiguate (in a semantic predicate) the widget option and the
 * expression path (with modified matching) is used for both the variable
 * definition and general expressions.
 * <p>
 * New variables can be defined using the the format phrase with an embedded
 * <code>AS or LIKE</code> clause!  The <code>AS or LIKE</code> clause can
 * be arbitrarily deep in the format phrase so we enable a special mode of
 * the {@link #expr} processing that allows {@link #lvalue} to match with
 * a <code>SYMBOL</code>.  This overcomes the limitations of fixed lookahead
 * cause (where the <code>AS or LIKE</code> keywords cannot be found in the
 * 2nd token).
 * <p>
 * Another unsupported feature of Progress is the implicit scoping of a
 * qualified variable name (when a field is referenced - this will be
 * matched by the expression alternative) to the unqualified version. Thus
 * a name like tablename.fieldname can be referenced in the following
 * format phrase as fieldname, even though it would normally be ambiguous.
 * To work around this, hints can be used to define a variable of the
 * appropriate name with the proper <code>FIELD_</code> token type.
 * <p>
 * The use of the {@link #format_phrase} following an expression is ambiguous
 * since it includes rule references that include options that include
 * expressions.  Thus, the lookahead set for the format phrase is quite
 * large and this naturally can conflict with following constructs in the
 * form item's calling rule(s).  A review of the code does not show any
 * problems and no false positives have been encountered yet. Ambiguity
 * warnings for this case have been disabled but this location bears
 * watching.   Similar ambiguities exist for the aggregate phrase, the use of
 * which is an undocumented feature in the form statement.  This rules does
 * allow this usage.
 *
 * @param   fields
 *          Set into which a field reference should be stored, if the form item
 *          represents a database field.
 */
form_item [Set<Aast> fields]
   :
      {
         String symbolName = null;
         Aast   fld        = null;
         Aast   child      = null;
         int    numFP      = 0;
         astFactory.makeASTRoot(currentAST, #[FORM_ITEM,"form item"]);
      }
      (
         options { generateAmbigWarnings = false; }
         :
           space_or_skip
         
           // we aren't sure that if this is a widget or some lvalue, so we
           // disambiguate here, otherwise we allow the match with the
           // following expression which could be an lvalue (field or var)
           // NOTE: this seems somewhat redundant (since the expr actually
           // calls the same lvalue rule as does widget) but it is necessary 
           // otherwise the expr processing would be changed. In this case,
           // if the widget match fails, normal expression processing is
           // maintained.
         | (
                {
                   sym.lookupWidget(LT(1).getText())   != -1 &&
                   sym.lookupVariable(LT(1).getText()) == -1 &&
                   !sym.isField(LT(1).getText())
                }?
                w:widget
                
                // an inline var def can occur here but the AS or LIKE
                // keywords can appear deeper in the format phrase than
                // normal lookahead tests can "see"; since some basic
                // attempts to duplicate the lookup processing of the
                // expression became too complicated to allow directly
                // matching a symbol here, and manually implementing 
                // guessing mode (simulate a syntactic predicate) without
                // hosing up the lookahead (see isFormVarDefinition() which
                // shows the possibility but is not quite right) is not
                // working, the solution is to temporarily allow lvalue to
                // match a symbol here instead of bombing out
                
                // note also that 2 or more unquoted - or + symbols can be
                // matched as a string in this place
              | { allowSymbolMatch = true; allowStringMatch = true; }
                e:expr
                { allowSymbolMatch = false; allowStringMatch = false; }
           )
           {
              // setup for the format phrase processing which may involve
              // a variable definition or 
              if (#e != null)
              {
                 child = #e.getChildAt(0);
                 
                 if (child != null && child.getType() == SYMBOL)
                 {
                    symbolName = child.getText();
                 }
                 else
                 {
                    fld = findFieldNode(#e);
                 }
              }
              else if (#w != null)
              {
                 fld = findFieldNode(#w);
              }
           }
           // WARNING: since format_phrase may be empty and this rule is an
           //          optional loop, the prediction logic for format_phrase
           //          includes all possible following constructs (e.g. DOT)
           //          even though these can't be matched in a format phrase;
           //          this leads to an infinite loop which we avoid by not
           //          allowing more than 2 matches to a format phrase (YES,
           //          there can be 2 format phrases, one before the aggregate
           //          phrase and one after --> just like in the 4GL)
           (
              options { generateAmbigWarnings = false; }
              :
                LPARENS! aggregate_phrase RPARENS!
              | { numFP < 2 }?
                format_phrase[symbolName, fld, true, false]  { numFP++; }
           )*
      )
      {
         // fixup the tree to remove the expression node if this was a
         // variable definition
         if (symbolName != null)
         {
            // the Aast helper methods can't be used here because they are
            // dependent upon parent nodes or try to fixup IDs that don't
            // exist yet
            child.setNextSibling(#e.getNextSibling());
            ##.setFirstChild(child);
         }
         
         if (fld != null)
         {
            fields.add(fld);
         }
      }
   ;
   
/**
 * Matches the <code>SPACE() or SKIP()</code> constructs and the optional
 * following numeric constant.  Note that although ANTLR reports ambiguity
 * in the optional clause, there is no actual ambiguity so warnings have
 * been disabled.
 * <p>
 * Used by {@link #form_item}, {@link #set_or_update_or_prompt_for_stmt },
 * {@link #display_stmt}, {@link #enable_stmt}, {@link #export_stmt},
 * {@link #column_spec} and {@link #content_array} (used in a
 * <code>MESSAGE</code> statement).
 * <p>
 * The subtree is rooted by the keyword.
 */
space_or_skip
   :
      (KW_SPACE^ | KW_SKIP^)
      (
         options { generateAmbigWarnings = false; }
         :
         LPARENS! numeric_literal RPARENS!
      )?
   ;
   
/**
 * Matches the <code>SPACE() or SKIP()</code> constructs and the optional
 * following {@link #expr}.  Note that although ANTLR reports ambiguity
 * in the optional clause, there is no actual ambiguity so warnings have
 * been disabled.
 * <p>
 * This is an undocumented feature of Progress which cannot be used in all
 * language statements.  The idea is that this construct requires runtime
 * resolution of the value in the places where it is allowed.
 * <p>
 * Used by {@link #put_stmt}.
 * <p>
 * The subtree is rooted by the keyword.
 */
runtime_space_or_skip
   :
      (KW_SPACE^ | KW_SKIP^)
      (
         options { generateAmbigWarnings = false; }
         :
         LPARENS! expr RPARENS!
      )?
   ;
   
/**
 * Matches the <code>NULL()</code> construct and the optional following 
 * integer constant.  Note that although ANTLR reports ambiguity
 * in the optional clause, there is no actual ambiguity so warnings have
 * been disabled.
 * <p>
 * Used by {@link #put_stmt}.
 * <p>
 * The subtree is rooted by the keyword.
 */
null_clause
   :
      KW_NULL^
      (
         options { generateAmbigWarnings = false; }
         :
         LPARENS! NUM_LITERAL RPARENS!
      )?
   ;
   
/**
 * Matches the <code>SET, UPDATE and PROMPT-FOR</code> language statements
 * and implements the (nearly identical) logic for all 3.
 * <p>
 * This rule is called by {@link #stmt_list}.
 * <p>
 * Calls the following rules:
 * <p>
 * {@link #lvalue} (this handles the field processing AND stream references)
 * {@link #space_or_skip}
 * {@link #text_clause}
 * {@link #at_phrase}
 * {@link #to_constant}
 * {@link #colon_constant} (undocumented but found in customer source)
 * {@link #constant_form_item}
 * {@link #record_spec}
 * {@link #go_on_clause}
 * {@link #frame_phrase}
 * {@link #inner_block}
 * <p>
 * All options can be processed in any order.
 * <p>
 * The list of fields, records or other values to set or update is
 * particularly tricky because it has very real ambiguity in allowing an
 * <code>lvalue</code> in the leftmost token of 2 peer alternatives. While
 * normally, a semantic predicate could be used to resolve this ambiguity,
 * the <code>lvalue</code> rule itself defeats this since it supports
 * array subscripting.  This form of <code>lvalue</code> means that the
 * <code>EQUALS</code> token (that disambiguates the assignment form from the
 * <code>field_clause</code>) can appear at an arbitrary point in the token
 * stream (not just in the k == 2 position).  For example, myVar[ i + 7 ] =
 * puts the <code>EQUALS</code> in the k == 7 position.  By splitting off
 * the subsequent processing and left factoring the <code>lvalue</code>,
 * this problem is resolved.
 * <p>
 * The order of the field list/record spec alternatives is also designed to
 * match more specific choices first.  This combined with a semantic
 * predicate to pick between a record and field, makes this work. 
 * <p>
 * The use of <code>inner_block</code> is to implement the editing clause,
 * which is a form that <code>inner_block</code> supports.  This also
 * matches <code>DO, REPEAT and FOR</code> keywords in addition to
 * <code>EDITING</code> but this should not be a problem in practice.
 * This can't be easily solved with a semantic validating rule since the
 * optional label that can prefix an editing phrase makes the editing
 * keyword appear in either the k == 1 or k == 3 positions.  The call to
 * <code>inner_block</code> uses {@link #end_stmt} to delimit the 
 * block.  Based on how we call <code>inner_block</code>, the subsequent
 * invocation of <code>end_stmt</code> does NOT eat the trailing
 * <code>DOT</code> since it is called inline in this language statement.
 * <p>
 * Use of some other unreserved keywords naturally conflicts with possible 
 * following keywords, but this is a bogus ambiguity warning as Progress  
 * is designed to match as a keyword first. Ambiguity warnings have been
 * disabled in some subrules as a result.
 * <p>
 * The use of the <code>constant_form_item</code> is required for the
 * <code>PROMPT-FOR</code> statement and it may not be valid for the
 * <code>SET or UPDATE</code> statements based on the Progress documentation.
 * Valid Progress code will not have a problem with this.
 * <p>
 * An undocumented feature exists where in-line variable definitions are possible using an
 * AS or LIKE clause in a format phrase (in all 3 forms of this rule).  The AS or LIKE can
 * appear later in the phrase. This was found in working customer code.
 */
set_or_update_or_prompt_for_stmt
   :
      {
         Aast      fld        = null;
         Set<Aast> fields     = new HashSet<>();
         String    symbolName = null;
      }
      (
         (    KW_SET^
            | KW_UPDATE^ 
            | KW_PRMT_FOR^
            | p:KW_PROMPT^ { #p.setType(KW_PRMT_FOR); }
         )
         (
            options { generateAmbigWarnings = false; }
            :
            
              KW_UNL_HID
            | CARET
            | space_or_skip
            | text_clause
            | literal (literal_format_phrase)?
            | KW_NO_ERROR
            | go_on_clause
            | fp:frame_phrase[false, true]
            
            | // this implements the editing phrase
              {
                 LA(1) == KW_EDITING                              ||
                 (LA(2) == COLON && followedByWhitespace(LT(2)))  ||
                 isMalformedLabel(true, 1, null)
              }?
              {
                 Aast malformedLabel = null;
              
                 // not a proper label, check for a malformed label
                 boolean mal = isMalformedLabel(true, 1, null); 
                 
                 // HARD CODED call that can't be placed directly in our rule
                 // below (inner_block) without very bad results on prediction
                 // logic generation
                 if (mal)
                 {
                    malformed_symbol();
                    malformedLabel = (Aast)returnAST;
                    match(COLON);
                 }
              }
              inner_block[false, malformedLabel]
              
            | { isRecord(LT(1).getText()) }?
              r:record_spec[fields]
              {
                 attachAllAssignSchemaValidation(##, #r, sym, true);
              }
              
              // this should really be lvalue for all cases except some forms of the
              // embedded assign which can reference complex expressions on the left
              // side of the assignment operator (e.g. var:SCREEN-VALUE = "something")
              // TODO: this is too permissive in the common case (not the embedded
              //       assignment)
            | { allowSymbolMatch = true; }
              l:chained_object_members
              {
                 // reset each loop iteration
                 symbolName = null;
                 fld        = null;
                 
                 if (#l.getType() == SYMBOL)
                 {
                    symbolName = #l.getText();
                 }
                 else
                 {
                    fld = findFieldNode(#l);
                    if (fld != null)
                    {
                       fields.add(fld);
                    }
                 }
                 
                 allowSymbolMatch = false;
              }
              (
                 options { generateAmbigWarnings = false; }
                 :
                   embedded_assign
                 | format_phrase[symbolName, fld, true, true]
              )? 
         )*
         DOT!
      )
      {
         String frame = getFrameName(#fp);

         if (!fields.isEmpty())
         {
            sym.addFrameFields(frame, fields);
         }

         setUseDictExps(frame, #fp, ##);
      }
   ;
   
/**
 * Matches the <code>INSERT</code> language statement.  Uses the
 * {@link #record_spec}, {@link #using_id_clause} and
 * {@link #frame_phrase} rules.
 * <p>
 * Used by {@link #stmt_list}.
 */
insert_stmt
   :
      {
         Set<Aast> fields = new HashSet<>();
      }
      KW_INSERT^
      r:record_spec[fields]
      (using_id_clause)?
      (fp:frame_phrase[false, false])?
      (KW_NO_ERROR)?
      DOT!
      {
         String frame = getFrameName(#fp);

         if (!fields.isEmpty())
         {
            sym.addFrameFields(frame, fields);
         }
         attachAllAssignSchemaValidation(##, #r, sym, true);

         setUseDictExps(frame, #fp, ##);
      }
   ;
   
/**
 * Matches the <code>RAW</code> language statement with the following
 * {@link #lvalue} and 1 or 2 optional integer {@link #expr} for position
 * and length.
 * <p>
 * Called by {@link #assign_type_syntax_stmt_list}.
 */
raw_stmt
   :
      KW_RAW^ LPARENS! lvalue (COMMA! expr (COMMA! expr)? )? RPARENS!
   ;
   
/**
 * Matches the <code>SET-SIZE</code> language statement with the following
 * {@link #lvalue}.
 * <p>
 * Called by {@link #assign_type_syntax_stmt_list}.
 */
set_size_stmt
   :
      KW_SET_SZ^ LPARENS! lvalue RPARENS!
   ;
   
/**
 * Matches the <code>SET-BYTE-ORDER</code> language statement with the
 * following {@link #lvalue}.
 * <p>
 * Called by {@link #assign_type_syntax_stmt_list}.
 */
set_byte_order_stmt
   :
      KW_SET_B_OR^ LPARENS! lvalue RPARENS!
   ;
   
/**
 * Matches the <code>SET-POINTER-VALUE</code> language statement with the
 * following {@link #lvalue}.
 * <p>
 * Called by {@link #assign_type_syntax_stmt_list}.
 */
set_pointer_value_stmt
   :
      KW_SET_PTR^ LPARENS! lvalue RPARENS!
   ;
   
/**
 * Matches the <code>PUT-BITS, PUT-BYTE, PUT-BYTES, PUT-DOUBLE,</code>
 * <code>PUT-FLOAT, PUT-INT64, PUT-LONG, PUT-SHORT, PUT-STRING,</code>
 * <code>PUT-UNSIGNED-LONG, PUT-UNSIGNED-SHORT</code> language statement
 * with the following {@link #expr} expressions.
 * <p>
 * Most of these statements are specific to the <code>MEMPTR</code> data type
 * but not all (contrary to what the name of this rule might suggest. The
 * format is essentially identical, so they were all collapsed into this
 * single rule.
 * <p>
 * Called by {@link #assign_type_syntax_stmt_list}.
 */
put_memptr_stmt   
   :
      (
           KW_PUT_BITS^
         | KW_PUT_BYTE^ 
         | KW_PUT_BYTS^
         | KW_PUT_DBL^
         | KW_PUT_FLT^
         | KW_PUT_I64^
         | KW_PUT_LONG^
         | KW_PUT_SHT^
         | KW_PUT_STR^
         | KW_PUT_UL^
         | KW_PUT_USHT^
      )
      LPARENS! expr COMMA! expr (COMMA! expr)? RPARENS!
   ;
   
/**
 * Matches the <code>NEXT-PROMPT</code> language statement.  Uses the {@link #lvalue} and
 * {@link #frame_phrase} rules. Undocumented support has been added for optional 
 * <code>INPUT</code> and <code>FRAME framename</code> qualifiers before the lvalue. This
 * was working syntax that was found in customer source code.
 * <p>
 * The 4GL has an undocumented feature that allows a {@link #format_phrase} to be specified
 * after the lvalue.  This is implemented here.
 * <p>
 * Used by {@link #stmt_list}.
 */
next_prompt_stmt
   :
      KW_NEXT_PMT^
      (KW_INPUT!)?
      (
         options { generateAmbigWarnings = false; }
         :
         frame_reference[false]
      )?
      l:lvalue
      (
         options { generateAmbigWarnings = false; }
         :
         format_phrase[null, null, false, false]
      )?
      (fp:frame_phrase[false, false])?
      DOT!
      {
         int type = #l.getType();
         if (type > BEGIN_FIELDTYPES && type < END_FIELDTYPES)
         {
            String frameName = getFrameName(#fp);
            sym.addFrameField(frameName, #l);
         }
      }
   ;
   
/**
 * Matches the <code>GO-ON()</code> construct and the list of key labels 
 * that follow in enclosing parenthesis.  Each key label is any token at all
 * with the loop ending on an <code>RPARENS</code>. Please note that the
 * Progress documentation states that these must be whitespace separated and
 * NOT comma separated. This is INCORRECT. As an undocumented feature, one
 * can mix or match whitespace and/or commas to separate event entries. This
 * rule supports this fully.
 * <p>
 * For certain event definitions such a CTRL-@, this code will initially
 * match on 2 or more tokens (in this case a SYMBOL (CTRL-) and an AT (@)).
 * These get converted to EVENT tokens and <code>processHiddenText()</code>
 * is used to save enough data regarding the whitespace to allow post-parse
 * fixups to properly merge these back into a single EVENT node.  
 * <p>
 * Used by {@link #set_or_update_or_prompt_for_stmt} and {@link #choose_stmt}.
 * <p>
 * Please note that the method <code>processHiddenText</code> is used to
 * detect and annotate if a merge of AST nodes must occur due to missing
 * whitespace between tokens.  This is required since in this place the
 * Progress lexer operates using whitespace only, but this mode cannot be
 * easily forced in the current approach since sharing state between the
 * parser and lexer is exceptionally difficult (not feasible).
 * <p>
 * The subtree is rooted by the keyword.
 */
go_on_clause   
   :
      KW_GO_ON^
      LPARENS! 
      (
         e:~RPARENS
         {
            if (LA(1) != RPARENS && LA(1) != COMMA)
            {
               processHiddenText(e, #e, false);
            }
            
            // convert the Progress string to a Java string if needed
            if (#e.getType() == STRING)
            {
               #e.setText(character.progressToJavaString(#e.getText()));
            }
            
            // set the event type
            #e.setType(EVENT);            
         }
         
         (
            options { generateAmbigWarnings = false; }
            :
            COMMA!
         )?
      )+
      RPARENS!
   ;
   
/**
 * Matches the format phrase construct of the Progress 4GL language which
 * optionally appears following a literal in a referencing rule but does not
 * begin with a specific keyword. Only a small subset of the standard 
 * options can be specified (see {@link #format_phrase}).
 * <p>
 * The following list of rules are called to properly implement the tree
 * building (which will be rooted at a <code>FORMAT_PHRASE</code> token):
 * <p>
 * <ul>
 *    <li> {@link #at_phrase}
 *    <li> {@link #colon_constant}
 *    <li> {@link #to_constant} 
 * </ul>
 * <p>
 * Called from {@link #set_or_update_or_prompt_for_stmt}.
 */
literal_format_phrase   
   :
      {
         astFactory.makeASTRoot(currentAST, #[FORMAT_PHRASE,"format phrase"]);
      }
      (at_phrase | colon_constant | to_constant)
   ;
   
/**
 * Matches the format phrase construct of the Progress 4GL language which
 * optionally appears in a referencing rule but does not begin with a specific
 * keyword. All of its options can be specified in any order.
 * <p>
 * The following list of rules are called to properly implement the tree
 * building (which will be rooted at a <code>FORMAT_PHRASE</code> token):
 * <p>
 * <ul>
 *    <li> {@link #at_phrase}
 *    <li> {@link #as_clause}
 *    <li> {@link #like_clause}
 *    <li> {@link #colon_constant}
 *    <li> {@link #to_constant} 
 *    <li> {@link #column_label}
 *    <li> {@link #format_string}
 *    <li> {@link #help_string}
 *    <li> {@link #label}
 *    <li> {@link #validate}
 *    <li> {@link #ui_stuff}
 *    <li> {@link #view_as_phrase}
 *    <li> {@link #simple_when_clause}
 *    <li> {@link #at_base_field_clause}
 * </ul>
 * <p>
 * Ambiguity is caused by the use of unreserved keywords in the leftmost 
 * position combined with a loop that has a mixture of single and multiple
 * token alternatives and a rule reference that includes an expression 
 * starting in the second token position.  Based on the complicated
 * loop closure rule needed to process these options in any order, the
 * k=2 lookahead is sometimes being checked to confirm that a following
 * token is seen before a match is made.  This logic has been reviewed and
 * many cases have been tested, however all possible cases cannot be tested.  
 * For this reason, one must watch this rule carefully.  We have disabled
 * ambiguity warnings and it is thought that this is safe, but great care
 * must be taken to confirm this over time.
 * <p>
 * Called from {@link #form_item}.
 * <p>
 * The <code>WHEN</code> clause (accessed via <code>simple_when_clause</code>)
 * is not documented as part of the format phrase, BUT customer examples
 * show that this can be embedded inside one.  Rather than try to handle
 * all the individual locations where such a highly variable construct
 * (format phrase -&gt; when phrase -&gt; format phrase), it is more natural to
 * simply include it in the format phrase as we have done (probably just as
 * Progress does this internally).
 * <p>
 * The '@' base_field clause (accessed via <code>at_base_field_clause</code>)
 * is not documented as part of this phrase but customer examples show that
 * this can be embedded inside one.  See the paragraph above on the when
 * clause.
 * <p>
 * The undocumented option <code>PASSWORD-FIELD</code> is matched here. It was
 * found in customer source code.
 * <p>
 * When called from frame definition statements, this will match the undocumented option
 * <code>AS BUTTON</code> which is the equivalent of an inline <code>DEFINE BUTTON</code>
 * statement. This is a quirk of the 4GL. The 4GL doesn't provide this same functionality
 * for any other widget types. The other configuration of the format phrase will modify the
 * new button widget, as you would expect.
 *
 * @param    symName
 *           If not <code>null</code> then a variable definition will be
 *           processed and the format phrase is expected to contain an
 *           <code>AS</code> or <code>LIKE</code> clause.
 * @param    fld
 *           If not <code>null</code> then the given field reference will
 *           define a temporary scope promotion for the field's table such
 *           that validation expressions will resolve unambiguously.
 * @param    frameDef
 *           <code>true</code> if the caller is a statement that would create or add to a
 *           frame definition.
 * @param    enableWidget
 *           <code>true</code> if this format phrase is associated with a field clause which
 *           represents a widget being enabled in a frame; else <code>false</code>.
 */
format_phrase [ String symName, Aast fld, boolean frameDef, boolean enableWidget ]
   :
      {
         astFactory.makeASTRoot(currentAST, #[FORMAT_PHRASE,"format phrase"]);
         
         String baseSymName = null;
         
         boolean matchSym = false;
         boolean schem    = false;
         
         // validate processing needs a special promotion of the table because the expression can
         // have otherwise ambiguous field references
         if (fld           != null             && 
             fld.getType()  > BEGIN_FIELDTYPES &&
             fld.getType()  < END_FIELDTYPES)
         {
            // add a schema scope
            schem = true;
            sym.addSchemaScope(false);
            
            // promote the field's associated buffer
            sym.promoteTableName((String) fld.getAnnotation("bufname"), false, false);
         }
      }
      (
         options { generateAmbigWarnings = false; }
         :
           at_phrase
         | { frameDef && symName != null }?
           as_button_quirk[symName]
         | (as_clause[symName, false] | like_clause[symName, null, false])
         | (KW_ATTR | KW_NO_ATTR)
         | KW_AUTO_RET
         | KW_BLANK
         | (colon_constant | to_constant)
         | column_label
         | KW_DEBLANK
         | KW_DIS_A_ZA
         | format_string[false]
         | h:help_string
         | (label | KW_NO_LABEL)
         | KW_NO_TAB_S
         | KW_PASSWD_F
         | v:validate[schem ? (String) fld.getAnnotation("schemaname") : null]
         | ui_stuff
         | view_as_phrase[symName]
         | simple_when_clause
         
           // we have to accommodate the fact that we can have an inline
           // var def where the symbol is not defined in the calling code
           // (and passed in via the symName parm) but is part of the
           // at base field clause
         | {
              matchSym = (symName == null);
           }
           baseSymName = at_base_field_clause[matchSym]
           {
              if (matchSym && baseSymName != null)
              {
                 symName = baseSymName;
              }
           }
      )*
      {
         // process schema-level validation for an any widget in a frame definition statement, if
         // no overriding validation expression was provided.  depending on USE-DICT-EXPS usage,
         // all validation exps which appear before this setting (in a statement which doesn't
         // enable the widget) will be removed.  this is required because at this time we don't
         // know the frame name - this is computed later, at the frame phrase parsing, so the 
         // approach is to add everything and remove as needed, depending on the USE-DICT-EXPS,
         // when the frame phrase is parsed
         if ((#v == null || #h == null) && 
             fld != null                && 
             frameDef                   && 
             !fld.isAnnotation("schema-processed"))
         {
            attachAssignSchemaValidation(##, fld, sym);
         }
         
         if (fld != null)
         {
            // do not allow the field to be processed on the second pass for format_phrase (used
            // by def_frame, display_stmt, form_stmt)
            fld.putAnnotation("schema-processed", true);
         }
         
         // if we previously added a scope, clear it now
         if (schem)
         {
            sym.deleteSchemaScope(false);
         }
         
         // safety first since this rule can have everything be optional
         if (##.getNumberOfChildren() == 0)
         {
            // get rid of the artificial root we previously created
            ## = null;
         }
         else
         {
            // we only can be defining a new variable if the symbol name is
            // not null
            if (symName != null)
            {
               // pass the AST with all options to the symbol resolver to set
               // those variable options accordingly
               sym.setVariableOptions(symName, ##);
               
               // write any non-standard options into annotations
               sym.annotateVariableOptions(symName, ##, true);
            }
         }
      }
   ;
   
/**
 * Matches an <code>AS BUTTON</code> clause and adds a button widget as if an inline
 * <code>DEFINE BUTTON</code> had been implemented. This is an undocumented quirk of the
 * {@link #format_phrase}. The root node will be of type <code>AS_BUTTON</code>.
 *
 * @param    symName
 *           The name of the new button. Must not be <code>null</code>.
 */ 
as_button_quirk [String symName]
   :
      a:KW_AS^ KW_BUTTON! { #a.setType(AS_BUTTON); }
      {
         sym.addWidget(symName, WID_BUTTON);
      }
   ;
   
/**
 * Matches an aggregate phrase, most often used in the <code>DISPLAY</code>
 * language statement.
 * <p>
 * Called by {@link #display_stmt}.  Uses {@link #label} and 
 * {@link #by_clause}.
 * <p>
 * <b>This rule uses a trick to create an artificial node as a root while
 * still allowing ANTLR to do normal tree building (no suppression of the
 * AST build).</b>  This is only useful in cases where there is no valid
 * token which can be used as a tree root AND where a loop is being used
 * such that the contents of the loop cannot be referenced in a manually
 * built #([,],,,) ANTLR construct. ANTLR doesn't support (as of 2.7.4) the
 * ability to label a loop such that the entire contents can be referenced.
 * If one labels the individual elements inside the loop, only the last one
 * is available when the loop concludes.  However, by leaving the normal
 * tree creation unchanged but manually creating an artificial node and
 * manually setting this node as the root of the returned tree BEFORE the
 * children are added, this problem can be solved.
 * <p>
 * When the {@link #accum_function} was added, ANTLR started showing bogus
 * ambiguity warnings.  These have been disabled.
 */
aggregate_phrase   
   :
      { astFactory.makeASTRoot(currentAST, #[AGGREGATE,"aggregate"]); }
      (
         options { generateAmbigWarnings = false; }
         :
         (
              KW_AVERAGE | KW_COUNT   | KW_MAX     | KW_MIN     | KW_TOTAL 
            | KW_SUB_AVG | KW_SUB_CNT | KW_SUB_MAX | KW_SUB_MIN | KW_SUB_TOT
         )
         (label)?
         (by_clause)*
      )*
      { if (##.getNumberOfChildren() == 0) ## = null; }
   ;
   
/**
 * Matches the constant prefixed subclause of the {@link #form_item}.
 * Calls {@link #view_as_text_clause}, {@link #ui_stuff}, {@link #at_phrase} 
 * and {@link #to_constant}.
 * <p>
 * ANTLR detects ambiguity in the loop of this rule due to conflict between
 * the unreserved keywords and the expression that follows in the 2nd token
 * position for <code>ui_stuff</code>.  This code should match on the
 * keywords first, so these ambiguity warnings have been disabled. 
 * <p>
 * Though undocumented, one can specify a {@link #format_string} as an
 * option.
 * <p>
 * Separated from the main rule to properly build the AST with less effort.
 * This is the only reason such a simple rule is implemented separately.
 */ 
constant_form_item
   :
      literal
      
      // TODO: root these options at a FORMAT_PHRASE node
      (at_phrase | to_constant)?
      (
         options { generateAmbigWarnings = false; }
         :
           ui_stuff
         | format_string[false]
         | view_as_text_clause
      )*
   ;
   
/**
 * Matches a specific <code>TO literal</code> clause used in the
 * {@link #constant_form_item} and {@link #format_phrase}.
 * <p>
 * Separated from the main rule to properly build the AST with less effort.
 * This is the only reason such a simple rule is implemented separately.
 */ 
to_constant   
   :
      KW_TO^ numeric_literal
   ;   
   
/**
 * Matches a specific <code>COLON literal</code> clause used in the
 * {@link #format_phrase}.
 * <p>
 * Separated from the main rule to properly build the AST with less effort.
 * This is the only reason such a simple rule is implemented separately.
 */ 
colon_constant
   :
      KW_COLON^ numeric_literal
   ;
   
/**
 * Matches a specific <code>VIEW-AS TEXT</code> clause used in the
 * {@link #constant_form_item}.
 * <p>
 * Separated from the main rule to properly build the AST with less effort.
 * This is the only reason such a simple rule is implemented separately.
 */
view_as_text_clause
   :
      KW_VIEW_AS^ KW_TEXT
   ;
   
/**
 * Matches the frame phrase construct of the Progress 4GL language which
 * always begins with the <code>WITH</code> keyword. All of its options
 * can be specified in any order. In fact, this can be an empty phrase
 * with nothing following the keyword (undocumented feature). As another
 * undocumented and useless feature, one can insert any number of whitespace
 * delimited <code>WITH</code> keywords in between any option and it will
 * parse and work as if these extra keywords did not exist. This is probably
 * just in fact allowing multiple sequential frame phrases (with any of them
 * being empty), but in this rule we just drop the extra <code>WITH</code>
 * keyword and parse everything into a single frame phrase. This should be
 * logically equivalent. 
 * <p>
 * The following list of rules are called to properly implement the tree
 * building (which will be rooted at a <code>FRAME_PHRASE</code> token):
 * <p>
 * <ul>
 *    <li> {@link #accum_clause}
 *    <li> {@link #at_phrase}
 *    <li> {@link #button_reference}
 *    <li> {@link #color_spec}
 *    <li> {@link #context_help_file}
 *    <li> {@link #frame_reference}
 *    <li> {@link #size_phrase}
 *    <li> {@link #lvalue} (only for stream and browse references, protected with a semantic predicate)
 *    <li> {@link #title_phrase}
 *    <li> {@link #version_6_frame_clause}
 *    <li> {@link #view_as_dialog_clause}
 *    <li> {@link #in_window_clause}
 *    <li> {@link #misc_frame_opts}
 *    <li> {@link #columns_clause}
 *    <li> {@link #down_clause}
 * </ul>
 * <p>
 * The order of the above list is important in the case of the last two
 * rule references {@link #columns_clause} and {@link #down_clause}. These
 * prefix a <code>NUM_LITERAL</code> and an expression respectively to
 * a keyword which is inherently ambiguous. The ambiguity is resolved by
 * the careful ordering of these alternatives. Based on the complicated
 * loop closure rule needed to process these options in any order, the
 * k=2 lookahead is often being checked to confirm that a following
 * token is seen before a match is made.  This logic has been reviewed and
 * many cases have been tested, however all possible cases cannot be tested.  
 * For this reason, one must watch this rule carefully.  We have disabled
 * ambiguity warnings and it is thought that this is safe, but great care
 * must be taken to confirm this over time.
 * <p>
 * Called from {@link #do_repeat_stmt}, {@link #for_stmt} and
 * {@link #def_frame_stmt} among other locations.
 *
 * @param    blockHdr
 *           <code>true</code> if this frame phrase is being called from
 *           within a block header (which will change its matching logic
 *           to allow it to avoid matching with possible following keywords
 *           in the calling rule).
 * @param    editing
 *           <code>true</code> if this frame phrase can be followed by an
 *           editing block.
 */
frame_phrase [boolean blockHdr, boolean editing]
   :
      w:KW_WITH^ { #w.setType(FRAME_PHRASE); }
      (
         options { generateAmbigWarnings = false; }
         :
           KW_WITH!   // undocumented "recursive" behavior
         | accum_clause
         | at_phrase
         | (KW_ATTR | KW_NO_ATTR) 
         | button_reference
         | KW_CENTER
         | color_spec
         | ui_stuff
         | KW_CTX_H 
         | context_help_file
         | KW_DROP_TAR
         | KW_EXPORT
         | frame_reference[true]
         | (KW_INH_BGC | KW_NO_INHBG)
         | (KW_INH_FGC | KW_NO_INHFG)
         | KW_KEEP_TAB 
         | KW_NO_BOX
         | KW_NO_HIDE 
         | KW_NO_LABEL 
         | KW_USE_DCT
         | KW_NO_VALID 
         | KW_NO_AUTOV 
         | KW_NO_HELP 
         | KW_NO_UNDL
         | KW_OVERLAY
         | (KW_PAGE_B | KW_PAGE_T) 
         | (KW_SCRN_IO | KW_STRM_IO)
         | KW_SCROLLBL
         | KW_SCROLL_V   // undocumented but found in customer source
         | KW_SIDE_L
         | size_phrase
         | KW_3D 
         | title_phrase
         | KW_TOP_ONLY
         | KW_USE_TXT
         
           // version_6_frame_clause: the documentation notes that these next 3
           // keywords are dependent on each other but working customer code
           // shows the usage of KW_USE_REV on its own; so instead of calling
           // the version_6_frame_clause sub-rule, it has been separated out
           // here
         | KW_V6FRAME
         | KW_USE_REV
         | KW_USE_UND
         
         | view_as_dialog_clause
         | in_window_clause
         | misc_frame_opts
         | { LA(1) == KW_STREAM || LA(1) == KW_STRM_HND || LA(1) == KW_BROWSE }?
           lvalue  // used to be stream_reference and browse_reference
         
           // these last two MUST be last and MUST be in this order since
           // the first one matches a NUM_LITERAL in the first token and
           // the second matches any expression
         | columns_clause
         
           // note: inner block headers can have a frame phrase followed by
           // the transaction keyword (which can be interpreted as the
           // transaction builtin function) or a loop increment expression
           // (both of which can follow the frame phrase in a block header
           // AND both of which are ambiguous with the down clause since that
           // option starts with an optional expression), so we avoid
           // matching the transaction keyword and anything else that matches
           // will be disambiguated by this rule that can handle both the
           // down clause and the loop increment
           // warning: this can continue to match frame phrase constructs
           // after a loop increment has been matched which can't happen in
           // valid 4GL code but as long as we don't see any such invalid code
           // that behavior won't trigger
         | { blockHdr && LA(1) != KW_TRANS }?
           down_clause_or_loop_incr
         
           // note: the COLON test is meant to exclude matching editing
           // block label definitions BUT it may inappropriately exclude valid
           // usage of the colon as a dereferencing operator (attrs, methods,
           // com props/methods, object members/props/methods) in a down
           // clause --> watch for this
         | { !blockHdr && (!editing || LA(2) != COLON) }?
           down_clause
      )*
   ;
   
/**
 * Matches the <code>V6FRAME</code> keyword and the optional following
 * keywords.  ANTLR detects ambiguity due to the use of non-reserved keywords 
 * and the potential to use symbols and expressions in the k=1 construct in
 * the calling rule {@link #frame_phrase}.  To eliminate the ambiguity
 * warning, ANTLR's greedy option had to be explicitly enabled.   
 * Since greedy is the default, this doesn't change the code generation, it
 * only suppresses the warning.  <b>Note that this rule checks the k=2 
 * lookahead before matching the <code>USE-REVERSE or USE-UNDERLINE</code>
 * keywords.  This may be OK but it may also bypass the match in a case where
 * something unexpected follows the optional keyword.  Since this cannot be
 * proven for all cases, watch this code!</b>
 * <p>
 * Separated from the main rule to properly build the AST without complexity.
 * This is the only reason such a simple rule is implemented separately.
 */
version_6_frame_clause   
   :
      KW_V6FRAME^        
      (
         options { greedy = true; }
         :
           KW_USE_REV
         | KW_USE_UND
      )?
   ;
   
/**
 * Matches the <code>COLUMNS</code> keyword and the required <b>preceding</b>
 * numeric literal.  In order to eliminate ambiguity, this option must be used
 * after more specific alternatives and before any alternatives that are
 * less specific (only one is).  This is currently used last in the calling
 * rule {@link #frame_phrase} for this reason.
 * <p>
 * It turns out that in Progress will accept the <code>COLUMN</code> keyword 
 * in place of the <code>COLUMNS</code> keyword, even though they are
 * supposed to be different keywords (and <code>COLUMN</code> has no
 * abbreviations).
 * <p>
 * Separated from the main rule to properly build the AST without complexity.
 * This is the only reason such a simple rule is implemented separately.
 */
columns_clause   
   :
      NUM_LITERAL (KW_COLUMNS^ | c:KW_COL^ { #c.setType(KW_COLUMNS); }) 
   ;
   
/**
 * Matches the <code>DOWN</code> keyword and the optional preceding
 * expression.  In order to eliminate ambiguity, this option must be used
 * after more specific alternatives and before any alternatives that are
 * less specific (none are).  This is currently used last in the calling
 * rule {@link #frame_phrase} for this reason.
 * <p>
 * Separated from the main rule to properly build the AST without complexity.
 * This is the only reason such a simple rule is implemented separately.
 */
down_clause   
   :
      (expr)? KW_DOWN^
   ;
   
/**
 * Matches the <code>ACCUM</code> keyword and the optional following maximum
 * length.  This rule conflicts with the use of a <code>NUM_LITERAL</code>
 * construct when k=1 in the calling rule {@link #frame_phrase}.  To eliminate 
 * the ambiguity warning, ANTLR's greedy option had to be explicitly enabled.   
 * Since greedy is the default, this doesn't change the code generation, it
 * only suppresses the warning.  <b>Note that this rule checks the k=2 
 * lookahead before matching the <code>NUM_LITERAL</code>.  This may be OK
 * but it may also bypass the match in a case where something unexpected
 * follows the <code>NUM_LITERAL</code>.  Since this cannot be proven for all
 * cases, watch this code!</b>
 * <p>
 * Separated from the main rule to properly build the AST without complexity.
 * This is the only reason such a simple rule is implemented separately.
 */
accum_clause   
   :
      KW_ACCUM^
      (
         options { greedy = true; }
         :
         NUM_LITERAL 
      )?
   ;
   
/**
 * Matches the <code>RETAIN, SCROLL and WIDTH</code> keywords and the required
 * following numeric or decimal literal. Note that it is undocumented that
 * Progress allows <code>WIDTH-CHARS</code> to be used as a synonym for
 * <code>WIDTH</code> in this case.  Also undocumented is the fact that the
 * <code>WIDTH-PIXELS</code> keyword can be used here.
 * <p>
 * Used by {@link #frame_phrase}.
 * <p>
 * Separated from the main rule to properly build the AST without complexity.
 * This is the only reason such a simple rule is implemented separately.
 */
misc_frame_opts   
   :
      (
           KW_RETAIN^
         | KW_SCROLL^
         | KW_WIDTH^
         | w:KW_WIDTH_C^ { #w.setType(KW_WIDTH); }
         | KW_WIDTH_P^
      )
      numeric_literal 
   ;
   
/**
 * Matches the <code>CANCEL-BUTTON or DEFAULT-BUTTON</code> keywords and the
 * required following button name.  No button namespace processing is done
 * to confirm that the following symbol is indeed a button.
 * <p>
 * Used by {@link #frame_phrase}.
 * <p>
 * Separated from the main rule to properly build the AST without complexity.
 * This is the only reason such a simple rule is implemented separately.
 */
button_reference   
   :
      (KW_CANCEL_B^ | KW_DEFLT_BN^) lvalue
   ;
   
/**      
 * Matches the <code>VIEW-AS DIALOG-BOX</code> clause.
 * <p>
 * Used by {@link #frame_phrase}.
 * <p>
 * Separated from the main rule to properly build the AST without complexity.
 * This is the only reason such a simple rule is implemented separately.
 */
view_as_dialog_clause   
   :
      KW_VIEW_AS^ KW_DIALOG
   ;
   
/**
 * Matches the <code>CONTEXT-HELP-FILE</code> keyword and the required
 * following {@link #filename}.
 * <p>
 * Used by {@link #frame_phrase}.
 * <p>
 * Separated from the main rule to properly build the AST without complexity.
 * This is the only reason such a simple rule is implemented separately.
 */
context_help_file   
   :
      KW_CTX_H_F^ filename
   ;
   
/**
 * Matches the size phrase construct of the Progress 4GL language.
 * This rule is called from {@link #frame_phrase} rule et al.
 */
size_phrase
   :
      (KW_SIZE^ | KW_SIZE_C^ | KW_SIZE_P^) 
      numeric_literal KW_BY! numeric_literal 
   ;   
   
/**
 * Matches the <code>EDGE-CHARS or EDGE-PIXELS</code> construct of the 
 * Progress 4GL language. This rule is called from the {@link #def_rect_stmt}
 * rule.
 */
edge_clause
   :
      (KW_EDGE_C^ | KW_EDGE_P^) numeric_literal 
   ;      
   
/**
 * Matches the title-phrase construct of the Progress 4GL language.
 * This implementation is relaxed compared to the 4GL.  In particular, the
 * {@link #color_spec} is used for the bulk of the function but it does
 * have options that cannot normally be constructed in a title phrase. In
 * addition, the {@link #ui_stuff} rule is used to add the optional
 * <code>FONT expression</code> construct that occurs just before the 
 * title string expression.
 * <p>
 * This rule is called from {@link #frame_phrase} and {@link #browse_options_phrase} rules.
 * <p>
 * ANTLR detects ambiguity in the use of <code>ui_stuff</code> and <code>color_spec</code>.
 * The code has been protected with semantic predicates and ambiguity warnings have been
 * disabled.
 */
title_phrase
   :
      KW_TITLE^
      (
         options { generateAmbigWarnings = false; }
         :
           { LA(1) == KW_COLOR }?
           color_spec
         | { 
             LA(1) == KW_BGCOLOR || LA(1) == KW_FGCOLOR || LA(1) == KW_DCOLOR || LA(1) == KW_FONT
           }?
           ui_stuff
      )*
      expr
   ;
   
/**   
 * Matches the <code>COLOR</code> language statement and the required
 * following color specification. This rule uses an inlined version of the
 * {@link #color_spec} with modifications to allow an empty color phrase.
 * The {@link #display_color_clause} and {@link #prompt_color_clause} rules
 * provide the bulk of the implementation.
 * <p>
 * Actual customer code shows that a {@link #format_phrase} can be placed
 * after the {@link #lvalue} used to describe each field.  In addition, at
 * the end of the statement an optional {@link #frame_phrase} is matched.
 * Although it is documented that at least 1 field (lvalue) is required,
 * customer code has shown this to be optional.
 * <p>
 * Bogus ANTLR ambiguity warnings have been disabled.
 */
color_stmt
   :
      {
         Set<Aast> fields = new HashSet<>();
      }
      (
         KW_COLOR^
         (
             options { generateAmbigWarnings = false; }
             :
             prompt_color_clause
             
             // an undocumented condition found in customer source, where there
             // is a display keyword but no following color phrase! we match
             // the display keyword and go on to the lvalue, only in the case
             // where we see that there is a following lvalue (WARNING: this
             // semantic predicate doesn't really test for all finds of lvalue!)
           | {
               (sym.lookupVariable(LT(2).getText()) != -1 || sym.isField(LT(2).getText())) &&
               LA(2) != KW_MESSAGES
             }?
             KW_DISP
             
           | display_color_clause
             (
                options { generateAmbigWarnings = false; }
                :
                prompt_color_clause
             )?
         )
         color_field_list[fields]
         (fp:frame_phrase[false, false])?
         DOT!
         {
            String frame = getFrameName(#fp);

            if (!fields.isEmpty())
            {
               sym.addFrameFields(frame, fields);
            }

            setUseDictExps(frame, #fp, ##);
         }
      )
   ;
   
/**
 * Aggregates (in a tree building sense) the list of all fields having 
 * color(s) set by a {@link #color_stmt}.  Calls {@link #lvalue} and
 * {@link #format_phrase} and roots everything at a node of type
 * <code>CONTENT_ARRAY</code>.
 *
 * @param   fields
 *          Set into which all database field references should be stored.
 */
color_field_list [Set<Aast> fields]
   :
      {
         Aast   fld        = null;
         String symbolName = null;
         astFactory.makeASTRoot(currentAST, #[CONTENT_ARRAY, ""]);
      }
      (
         { allowSymbolMatch = true; }
         l:lvalue
         {
            // reset each loop iteration
            symbolName = null;
            fld        = null;
            
            if (#l.getType() == SYMBOL)
            {
               symbolName = #l.getText();
            }
            else
            {
               fld = findFieldNode(#l);
               if (fld != null)
               {
                  fields.add(fld);
               }
            }
            
            allowSymbolMatch = false;
         }
         (
            options { generateAmbigWarnings = false; }
            :
            format_phrase[symbolName, fld, true, false]
         )?
      )*
      {
         if (##.getNumberOfChildren() == 0)
         {
            ## = null;
         }
      }
   ;
   
/**
 * Matches the color_spec construct of the Progress 4GL language. Uses the {@link #color_phrase}
 * rule downstream for the bulk of this rule's implementation. 
 * <p>
 * This rule is called from {@link #title_phrase} and {@link #frame_phrase}.
 * <p>
 * ANTLR detects ambiguity in each loop of this rule due to conflict between
 * unreserved keywords and the symbol that can be generated by the rule
 * reference to <code>color_phrase</code> or the expression that follows in
 * the 2nd token position for <code>ui_stuff</code>.  This code should match
 * on the keywords first, so these ambiguity warnings have been disabled.
 */
color_spec
   :
      KW_COLOR^ display_color_clause
      (
         options { generateAmbigWarnings = false; }
         :
         prompt_color_clause
      )?
   ;
   
/**
 * Matches the <code>DISPLAY color_phrase</code> construct used in a color
 * specification (actually the <code>DISPLAY</code> keyword is optional).
 * <p>
 * Used by {@link #color_spec}.
 * <p>
 * Separated from the main rule to properly build the AST without complexity.
 * This is the only reason such a simple rule is implemented separately.
 */
display_color_clause   
   :
     (
        options { generateAmbigWarnings = false; }
        :
        KW_DISP^
     )?
     color_phrase
   ;
   
/**
 * Matches the <code>PROMPT color_phrase</code> construct used in a color
 * specification. This supports an undocumented Progress feature, where
 * <code>PROMPT-FOR</code> is accepted as a synonym to <code>PROMPT</code>.
 * <p>
 * Used by {@link #color_spec}.
 * <p>
 * Separated from the main rule to properly build the AST without complexity.
 * This is the only reason such a simple rule is implemented separately.
 */
prompt_color_clause   
   :
      (KW_PROMPT^ | p:KW_PRMT_FOR^ { #p.setType(KW_PROMPT); }) color_phrase
   ;
   
/**
 * Matches the color_phrase construct of the Progress 4GL language.
 * This implementation is relaxed compared to the 4GL and has the following
 * limitations:
 * <p>
 * <ul>
 *     <li>dos-hex-attributes are not supported
 *     <li>Windows color attributes are not supported
 * </ul>
 * <p>
 * Valid color specification options are these keywords: NORMAL, INPUT and
 * MESSAGES. They have a unique capacity: they can be abbreviated down to 1
 * character even though there are conflicting keywords in 4GL! This is how it
 * works.
 * <p>
 * The NORMAL keyword can be coded as anything between N and NORMAL, excluding
 * NO, which is a reserved 4GL keyword and can't be used in color-phrase.
 * <p>
 * The INPUT keyword can be coded as anything between I and INPUT, although
 * normally this keyword can't be abbreviated. There is a conflicting 4GL
 * keyword IN, but in the color-phrase it assumes the role of INPUT.
 * <p>
 * The MESSAGES keyword can be coded as anything between M and MESSAGES. There
 * is a conflicting 4GL keyword MESSAGE, which can't be abbreviated, but it
 * assumes the role of MESSAGES when used in a color-phrase.
 * <p>
 * This explains why one sees KW_IN and KW_MSG as alternatives in this rule.
 * In both cases, the token type is converted to the originally required type.
 * The KW_INPUT, KW_NORMAL and KW_MESSAGES are defined as having no 
 * abbreviations, otherwise they would conflict with many other keywords.
 * Instead, this rule checks to see if SYMBOL type tokens (which abbreviations
 * are) represent a valid abbreviation and sets the token type explicitly.
 * <p>
 * Customer source has shown that reserved keywords such as 
 * <code>UNDERLINE</code> can be found in a color phrase (this is
 * undocumented by Progress and is currently supported on a case by case
 * basis.
 * <p>
 * Note that ANTLR reports ambiguity for this rule. This is due to the call to
 * protermcap, which ultimately calls the symbol rule. This conflicts with
 * the color_spec rule where KW_PROMPT designates the optional part. This
 * ambiguity is in ANTLR logic only. There is no ambiguity in the generated
 * code. For this reason, ambiguity warnings are suppressed for this rule.
 * <p>
 * This rule is called from {@link #color_spec} rule.
 */
color_phrase
   :
      { astFactory.makeASTRoot(currentAST, #[COLOR_PHRASE, ""]); }
      (
         options { generateAmbigWarnings = false; } 
         :
           value
         | KW_NORMAL
         | (KW_INPUT    | i:KW_IN  { #i.setType(KW_INPUT); })
         | (KW_MESSAGES | m:KW_MSG { #m.setType(KW_MESSAGES); })   
         | s:filename
           {
              String t = #s.getText().toLowerCase();
              
              if ("normal".startsWith(t))
              {
                 #s.setType(KW_NORMAL); 
              }
              else if ("input".startsWith(t))
              {
                 #s.setType(KW_INPUT); 
              }
              else if ("messages".startsWith(t))
              {
                 #s.setType(KW_MESSAGES);
              }
              else
              {
                 #s.setType(SYMBOL);
              }
           }
      )
   ;
   
/**
 * Matches the column specification for a <code>DEFINE BROWSE</code>   
 * language statement which is started by a <code>DISPLAY</code> keyword.
 * This is followed by either a record (and optional exclude list of fields)
 * or an expression and optional {@link #column_format_phrase} and
 * {@link #at_base_field_clause}. This record or expression is documented as
 * required but customer source has shown this to actually be optional.
 * <p>
 * Calls {@link #record_spec} and {@link #browse_enable_phrase} rules.
 * <p>
 * Note that Progress supports an undocumented use of {@link #space_or_skip}
 * as a column specification.
 * <p>
 * Called by {@link #def_browse_stmt}.
 * <p>
 * DISPLAY statement doens't pull schema VALEXP or HELP, so this will not use
 * {@link #attachAssignSchemaValidation}.
 */
column_spec
   :
      KW_DISP^
      (
           { 
              isRecord(LT(1).getText()) && !isQualifiedFieldNameQuirk(LT(1), LT(2), LT(3))
           }?
           r:record_spec[null]
           {
              attachAllAssignSchemaValidation(##, #r, sym, false);
           }
           
         | (
                space_or_skip
              | e:expr
                (
                   options { generateAmbigWarnings = false; }
                   :
                   column_format_phrase
                )?
           )*
      )
      (browse_enable_phrase)?
      (browse_options_phrase)*
   ;
   
/**
 * Matches the phrase which specifies which columns in the 
 * {@link #column_spec} are enabled for input (always is preceded by the
 * <code>ENABLE</code> keyword).
 * <p>
 * Calls {@link #browse_all_list}, {@link #lvalue}, {@link #help_string}
 * and {@link #validate}.
 */
browse_enable_phrase   
   :
      KW_ENABLE^ ( browse_all_list | (column_ref)+ )
   ;
   
/**
 * Defines one column element in a list of columns to enable in the 
 * {@link #browse_enable_phrase}.  Returns a node of type 
 * <code>COLUMN_REF</code> and uses {@link #lvalue}, {@link #help_string}
 * and {@link #validate} rule references.
 */
column_ref
   :
      { boolean schem = false;
        astFactory.makeASTRoot(currentAST, #[COLUMN_REF,"column ref"]); }
      (
         l:lvalue
         {
            schem = #l.getType() > BEGIN_FIELDTYPES && #l.getType() < END_FIELDTYPES;
         }
         // TODO: root these options at a FORMAT_PHRASE node; conversion rules will need to be
         // adjusted, too
         (
           options { generateAmbigWarnings = false; }
           :
             help_string
           | v:validate[schem ? (String) #l.getAnnotation("schemaname") : null]
           | KW_AUTO_RET
           | KW_DIS_A_ZA
         )*
         {
            if (#v == null)
            {
               attachAssignSchemaValidation(##, #l, sym);
            }
         }
      )
   ;
   
/**      
 * Matches the shorthand of specifying all fields (and optional exclude list
 * of fields) instead of an explicit field list, in the 
 * {@link #browse_enable_phrase} rule.  Uses {@link #except_list}.
 * <p>
 * This rule is naturally greedy (in matching a list of field names to
 * exclude, and doing so in a loop. ANTLR warns that this causes ambiguity. 
 * The greedy option is used to tell ANTLR to disable the warning, though it
 * has no impact on the actual code generation.
 * <p>
 * Separated from the main rule to properly build the AST without complexity.
 * This is the only reason such a simple rule is implemented separately.
 */ 
browse_all_list
   :
      KW_ALL^ (except_list)?
   ;
   
/**
 * Matches the column format phrase construct of the Progress 4GL language
 * which is used in a <code>DEFINE BROWSE</code>. All of its options can be
 * specified in any order.  This is generally a subset of the normal
 * {@link #format_phrase}, however there is one case that can be matched
 * here (<code>KW_WIDTH</code>) that cannot be matched in a format phrase.
 * <p>
 * This rule is separate since there is only a small subset of the possible
 * choices that can be matched, this reduces ambiguity.  Note that the result
 * is still rooted at a <code>FORMAT_PHRASE</code> node.
 * <p>
 * The following list of rules are called to properly implement the tree
 * building:
 * <p>
 * <ul>
 *    <li> {@link #label}
 *    <li> {@link #misc_frame_opts} (for the WIDTH, WIDTH-CHARS and WIDTH-PIXELS constructs only,
 *         the WIDTH-CHARS and WIDTH-PIXELS options are undocumented but were found in customer
 *         source)
 *    <li> {@link #ui_stuff}
 *    <li> {@link #column_label}
 *    <li> {@link #at_base_field_clause}
 *    <li> {@link #view_as_phrase}
 * </ul>
 * <p>
 * We have disabled ambiguity warnings because there is no real issue with
 * the <code>ui_stuff</code>.
 * <p>
 * The '@' base_field clause (accessed via <code>at_base_field_clause</code>)
 * is not documented as part of this phrase but customer examples show that
 * this can be embedded inside one.  To simplify logic, this has been merged.
 * <p> 
 * Called from {@link #column_spec}.
 */
column_format_phrase
   :
      {
         String bogus = null;
         astFactory.makeASTRoot(currentAST, #[FORMAT_PHRASE,""]);
      }
      (
         options { generateAmbigWarnings = false; }
         :
           (label | KW_NO_LABEL)
         | { LA(1) == KW_WIDTH || LA(1) == KW_WIDTH_C || LA(1) == KW_WIDTH_P }?
           misc_frame_opts
         | ui_stuff
         | format_string[false]
         | column_label
         | bogus = at_base_field_clause[false]
         | view_as_phrase[null]
      )*
      {
         // safety first since this rule can have everything be optional
         if (##.getNumberOfChildren() == 0)
         {
            // get rid of the artificial root we previously created
            ## = null;
         }
      }
   ;   
      
/**
 * Matches the browse options phrase that is almost the same as the frame
 * phrase construct (except it is only used in the defining browses). It
 * always begins with the <code>WITH</code> keyword. All of its options
 * can be specified in any order.  Called from {@link #column_spec}.
 * <p>
 * The following list of rules are called to properly implement the tree
 * building:
 * <p>
 * <ul>
 *    <li> {@link #ui_stuff}
 *    <li> {@link #row_height_clause}
 *    <li> {@link #size_phrase}
 *    <li> {@link #title_phrase}
 *    <li> {@link #tooltip_clause}
 *    <li> {@link #misc_frame_opts}
 *    <li> {@link #down_clause}
 * </ul>
 * <p>
 * Undocumented options <code>KEEP-TAB-ORDER</code>, <code>CENTERED</code>,
 * <code>KW_NO_COLS</code>, <code>KW_SCROLLBL</code>, <code>KW_NO_TAB_S</code> and
 * <code>KW_NO_UNDL</code> are supported here since Progress allows them.
 * <p>
 * The order of the above list is important in the case of the last
 * rule reference to {@link #down_clause}. This prefixes an expression
 * to a keyword which is inherently ambiguous. The ambiguity is resolved by
 * the careful ordering of these alternatives. Based on the complicated
 * loop closure rule needed to process these options in any order, the
 * k=2 lookahead is often being checked to confirm that a following
 * token is seen before a match is made.  This logic has been reviewed and
 * many cases have been tested, however all possible cases cannot be tested.  
 * For this reason, one must watch this rule carefully.  We have disabled
 * ambiguity warnings and it is thought that this is safe, but great care
 * must be taken to confirm this over time.
 */
browse_options_phrase
   :
      // TODO: root these options at a FORMAT_PHRASE or a BROWSE_OPTIONS node
      KW_WITH^
      (
         options { generateAmbigWarnings = false; }
         :
           ui_stuff
         | (KW_MULTIPLE | KW_SINGLE) 
         | (KW_SEPS | KW_NO_SEPS)
         | KW_NO_ASSGN
         | KW_NO_ROW_M
         | KW_NO_BOX
         | KW_NO_LABEL 
         | KW_NO_VALID 
         | KW_NO_AUTOV 
         | (KW_NO_SCR_V | KW_SCROLL_V) 
         | KW_DROP_TAR
         | KW_OVERLAY
         | KW_EXPANDBL
         | KW_FIT_LCOL
         | KW_NO_EM_SP
         | KW_NO_COLS           // undocumented but found in customer source
         | KW_CENTER            // undocumented but found in customer source
         | KW_KEEP_TAB          // undocumented but found in customer source
         | KW_SCROLLBL          // undocumented but found in customer source
         | KW_NO_TAB_S          // undocumented but found in customer source
         | KW_NO_UNDL           // undocumented but found in customer source
         | KW_TOP_ONLY          // undocumented but found in customer source
         | KW_3D                // undocumented but found in customer source
         | color_spec           // undocumented but found in customer source
         | row_height_clause 
         | size_phrase
         | title_phrase
         | tooltip_clause
         | misc_frame_opts
         
           // this last one MUST be last since it matches any expression
           // in the first token position
         | down_clause
      )*
   ;
   
/**
 * Matches the <code>ROW-HEIGHT-CHARS and ROW-HEIGHT-PIXELS</code> construct
 * with its required following integer constant.  Used in the
 * {@link #browse_options_phrase}.
 * <p>
 * Separating this allows the AST to be built cleanly and also centralizes
 * the logic for reuse.
 */
row_height_clause   
   :
      (KW_ROW_H_C^ | KW_ROW_H_P^) numeric_literal
   ;
   
/**
 * Matches the <code>VALUE( expression )</code> construct which converts
 * an expression into a symbol-like result (example: a type of name).  This is
 * used in many locations such as {@link #color_phrase} and {@link #run_stmt}.
 * <p>
 * Working 4GL code has shown that unquoted text can be present here instead
 * of a valid expression. Progress seems to treat that as if the text was
 * a quoted string literal.  This undocumented feature is supported here by
 * setting a flag to allow a symbol match in the expression processing.
 * <p>
 * Separating this allows the AST to be built cleanly and also centralizes
 * the logic for reuse.
 */
value
   :
      { allowSymbolMatch = true; }
      KW_VALUE^ LPARENS! expr RPARENS!
      { allowSymbolMatch = false; }
   ;
   
/**
 * Matches the <code>VALIDATE( expression )</code> construct which specifies
 * a boolean condition expression and a message string expression to display
 * when the condition expression evaluates to false.
 * <p>
 * Used by {@link #format_phrase}.
 * <p>
 * Separating this allows the AST to be built cleanly and also centralizes
 * the logic for reuse.
 *
 * @param   schemaName
 *          The schema name of the field being validated (may be null).
 */
validate [String schemaName]
   :
      { validation = true;
        validateFieldName = schemaName; }
      ( KW_VALIDATE^ LPARENS! cond:expr COMMA! tmsg:expr RPARENS! )
      { validation = false;
        validateFieldName = null; }
   ;
   
/**
 * Implements the reserved_or_symbol rule which matches symbols and reserved
 * or nonreserved keywords that have to be handled as symbols. It converts
 * matched tokens into SYMBOL type ones for consistency of the UAST. 
 */
reserved_or_symbol
   :
      {
         int oldtype = -1;
      
         if (LA(1) != SYMBOL)
         {
            oldtype = LA(1);
            LT(1).setType(SYMBOL);
         }
      }
      ( 
           BEGIN_RESERVED..END_RESERVED
         | symbol
      )
      {
         // save off original type from before token rewriting
         if (oldtype != -1)
            ##.putAnnotation("oldtype", new Long(oldtype));
      }
   ;

/**
 * Implements the <code>TABLE</code>, <code>DATASET</code>, <code>TABLE-HANDLE</code> and
 * <code>DATASET-HANDLE</code> logic for parameter definitions (see {@link #def_parm_stmt}).
 * <p>
 * One would expect that this method has a major namespace update role, since normal parameter
 * processing requires this.  In fact, only the form which references a table handle or dataset
 * handle updates the namespace (the variable namespace in this case).  In this case the
 * {@link #symbol} rule is called and the resulting name is added as a variable of type
 * <code>VAR_HANDLE</code>.
 * <p>
 * All forms (no matter the caller) that reference a temp-table or dataset name are in fact
 * referencing a resource that must have been statically defined previously in the procedure.
 * This means that the associated namespace updates have already occurred and this rule simply
 * accesses  this data via a lookup in the called {@link #record} rule (in the case of a table)
 * or directly (in the case of a dataset).
 *
 * @return   The variable name of the handle variable that was defined or
 *           <code>null</code> if this is a record or dataset name reference.
 */
table_parm returns [String varname = null]
   :
      {
         boolean handle = false;
         boolean dset   = false;
      }
      (
         (
              KW_TABLE^
            | KW_TAB_HAND^ { handle = true; }
            | KW_DATASET^  { dset = true; }
            | KW_DSET_HND^ { handle = true; }
         )
         // undocumented FOR keyword (seems like syntactic sugar so we drop it)
         (KW_FOR!)?
         (
              { handle }?
              t:symbol
              {
                 varname = #t.getText();
                 sym.addLocalVariable(varname, VAR_HANDLE, null);
              }
              
            | { dset }?
              d:symbol
              {
                 // this will be a problem later if it is not found
                 // TODO: need to check if explicit static/instance qualifiers can be used here
                 // TODO: need to check implicit static/instance (based on the surrounding code)
                 #d.setType(sym.lookupDataSet(#d.getText()));
              }
              
              // we haven't confirmed that the force flag should be false
            | record[true, false, false]
         )
         (KW_APPEND | KW_BY_VALUE | KW_BIND)*
      )
   ;
   
/**
 * Implements the core logic for the <code>DEFINE PARAMETER</code> Progress
 * 4GL language statement. This is the statement that defines named parameters
 * in procedures (external or internal).  The {@link #symbol} rule is used to
 * force the token type for the variable name to be a <code>SYMBOL</code>.
 *  Then this name is passed as a parameter to the {@link #as_clause} or
 * {@link #like_clause} rules, which add this variable to the variable
 * dictionary associating it with the correct data type.
 * <p>
 * This rule is called by {@link #def_parm_stmt} which is where the matching
 * on the <code>PARAMETER</code> keyword is done.
 * <p>
 * Multiple options are supported in any order (which is how Progress handles
 * these), however this code does not check for a repeated definition of
 * the options. Note that as an undocumented feature, Progress also allows
 * that the AS and LIKE clauses can be provided after other options. This is
 * supported.  In fact, these clauses can be optional in certain circumstances.
 * <p>
 * These option rules are all implemented as separate rules to force a clean 
 * tree structure for each option rather than accepting a flat structure
 * where all options and any associated data are all created as siblings.
 * Instead, each option subtree is created under its associated keyword's 
 * token and this is a child of the <code>PARAMETER</code> keyword in the
 * calling rule.
 *
 * @param    param
 *           The parameter AST where to save the extent annotation, if set. May be null.
 * @param    likeRef
 *           An <code>Aast</code> of a field/variable to be copyed in a LIKE -like in the event
 *           that the param does not contain any AS or LIKE clauses.
 *
 * @return   The new variable's name.
 */
regular_parm [Aast param, Aast likeRef] returns [String varname = null]
   :
      {
         int vtype = -1;
         int size  = -1;
      }
      s:symbol  { varname = #s.getText(); }
      (
         options { generateAmbigWarnings = false; }
         :
           a:as_clause[varname, true]
           {
              vtype = #a.getFirstChild().getType();
           }
           
         | l:like_clause[varname, null, false]
           {
              vtype = #l.getFirstChild().getType();
           }
           
         | initializer[vtype]
         | size = extent 
           {
              if ((size == -1 || size > 0) && param != null) 
                 param.putAnnotation("extent", new Long(size));
           }
         | label
         | format_string[false]
         | column_label
         | decimals_clause
         | case_sensitive
         | KW_NO_UNDO
      )*   
      {
         if (vtype == -1) // neither as_clause nor like_clause were processed
         {
            if (likeRef != null)
            {
                sym.addLocalVariableLike(varname, likeRef);
            }
            else
            {
               // throw incomplete parameter definition in regular_parm
            }
         }
      }
   ;
   
/**
 * Matches the <code>CASE-SENSITIVE</code> keyword with the optional
 * preceeding <code>NOT</code> keyword. Separated to enhance the tree build. 
 * <p>
 * Called by {@link #regular_parm}, {@link #temp_table_field_options} and
 * {@link #def_var_stmt}. 
 */
case_sensitive   
   :
      (KW_NOT^)? KW_CASE_SEN
   ;
   
/**
 * This rule creates the block structure for a Progress 4GL nestable block. 
 * The {@link #block_list} rule is called to start the block and this is
 * followed by a recursive call to {@link #block} which handles
 * the ability to nest top-level statements inside a block. Finally,
 * the rule ends with an optional <code>END</code> language statement. The
 * Progress language states that it is valid to omit the <code>END</code>.
 * In this case, the block's end is implicit at the end of the file.
 * <p>
 * The <code>nestLevel</code> instance member is incremented at the entry to
 * this rule and decremented on exit.  This allows the proper enforcement of
 * nesting limitations (procedures and functions cannot be nested, ever).
 * <p> 
 * In front of any of these constructs is an optional user-defined label.
 * The presence of this label is matched by the following <code>COLON</code> 
 * that <b>must precede</b> the Progress 4GL language statement that starts
 * the block.  The {@link #symbol} method is used to force the user-defined
 * name to a <code>SYMBOL</code> token type.  If matched, this symbol's text
 * is added into the label namespace and the token type of the symbol is
 * set to <code>LABEL</code>.
 * <p>
 * Schema record scopes are added before calling <code>block_list</code>
 * based on lookahead.  A scope is added in the case of
 * <code>FOR, REPEAT and DO FOR</code> statements.  In the case of 
 * <code>EDITING and normal DO</code> blocks, a scope is NOT added. If a
 * scope was added, then it is removed in the exit action.  This allows the
 * proper pushing and popping of record scopes in the same way as Progress
 * implements it. <b>Note that at this time, there is only a simple imitation
 * of the implicit expansion of a record scope from one block to another.
 * This is done by the {@link #record} rule by ALWAYS promoting a table to
 * the current scope. A more complicated solution may be necessary but it is
 * not yet known how much this is actually in use.</b>
 * <p>
 * This rule also creates an artificial node in the tree as the parent to
 * all children of the block.  This node is created as a token type of 
 * <code>INNER_BLOCK</code> if this is a <code>DO, FOR or REPEAT</code>
 * block and as a <code>EDITING_BLOCK</code> if this is an editing block.
 * <p>
 * Called by {@link #single_block}.
 *
 * @param    eat
 *           If <code>true</code>, this code will consume (and drop) a
 *           following <code>DOT</code> after the <code>END</code>.
 * @param    malformed
 *           If not <code>null</code>, this is a replacement for the match
 *           on a properly formed label but due to limits of branch prediction
 *           it had to be handled in the caller rather than here.
 */
inner_block! [boolean eat, Aast malformed]
   :
      {
         // procedures and functions CAN be defined inside inner blocks
         // so nestLevel should not be incremented here!
      
         boolean labeled  = false;
         boolean delScope = false;
         
         // the optional symbol section below will never trigger if this
         // is non-null so we don't have to disambiguate there
         if (malformed != null)
         {
            s_AST = malformed;
         }
      }   
      (
         // optional label processing
         ( s:symbol COLON! )?
         
         // this action MUST be done BEFORE any other rules are called since
         // these rules are allowed to reference the label that has just been
         // defined
         {
            if (#s != null)
            {
               #s.setType(LABEL_DEF);
               sym.addLabel(#s.getText(), LABEL);
               labeled = true;
            }
            
            // this code must always process next (before block_list is
            // called)
            if ( (LA(1) == KW_FOR) || (LA(1) == KW_REPEAT) || 
                 (LA(1) == KW_DO   &&  LA(2) == KW_FOR) )
            {
               delScope = true;
               sym.addSchemaScope(false);
            }
         }
         
         l:block_list b:block 
         (
            options { generateAmbigWarnings = false; }
            :
            end_stmt[eat]
         )?
      )
      {
         int    ttype = INNER_BLOCK;
         String ttxt  = "inner block";
         
         if (#l.getType() == KW_EDITING)
         {
            ttype = EDITING_BLOCK;
            ttxt  = "editing block";
         }
         
         ## = #([ttype, ttxt], s, l, ([BLOCK, "block"], b));
         
         // inner blocks propagate
         if (delScope)
            sym.deleteSchemaScope(true);
            
         if (labeled)
            sym.deleteLabel();
            
         // procedures and functions CAN be defined inside inner blocks
         // so nestLevel should not be decremented here!
      }      
   ;

/**
 * Matches all forms of inner blocks (nestable blocks - not procedures and
 * not functions) including <code>DO</code>, <code>REPEAT</code> and
 * <code>FOR</code>.
 * <p>
 * As all 3 of these language statements can end with a <code>COLON</code>
 * or a <code>DOT</code>, this rule implements this match in common code.
 * <p>
 * The core of this rule specifies multiple alternatives that match the
 * Progress language statements <code>DO, REPEAT</code>
 * ({@link #do_repeat_stmt}) and <code>FOR</code> ({@link #for_stmt}).
 * <p>
 * This rule also matches the <code>EDITING</code> keyword and allows this
 * rule to be used as the editing phrase.  This works because the calling
 * rule (<code>inner_block</code>) is identical in syntax to the editing
 * block. Note that a minor drawback to this approach is that no error
 * is thrown if someone tries to use an editing phrase in the top level of 
 * a Progress program, although this would not be valid Progress source.
 * <p>
 * These required keywords should provide the differentiation needed to
 * eliminate the ambiguity that is generated by having the optional 
 * <code>symbol</code> reference as the first token, <b>however it turns out
 * that the generated ANTLR code is not smart enough to handle this.</b>
 * Instead the {@link #block} top-level rule has a semantic predicate that
 * reimplements the dependencies implemented by this rule's logic.
 * <p>
 * Called by {@link #inner_block}.
 */
block_list  
   :
      (do_repeat_stmt | for_stmt | KW_EDITING^) (DOT! | COLON!) 
   ;
   
/**
 * Matches both forms of an assignment statement (explicit assignment to 
 * an <code>lvalue</code> and the alternate implicit discard of the
 * expression's return value).  More specifically, these two forms:
 * <ul>
 *    <li> any <code>lvalue</code> followed by an <code>EQUALS</code> token
 *         and an expression (this is the traditional concept of an
 *         assignment - one which stores the expression result in a variable
 *         or field)
 *    <li> any valid expression that stands on its own, with no 
 *         <code>lvalue</code> and no <code>EQUALS</code> operator (this
 *         is necessary to allow one to execute an arbitrary function and
 *         discard the result while any side effects of the function still
 *         occur)
 * </ul>
 * <p>
 * Note that the second form is a bit odd compared to most languages since
 * it is legal (though useless) to specify a standalone non-assignment
 * expression that is not a function call (whose return value is discarded
 * but which may have side effects that are useful).  An example of a useless
 * expression is a string token followed by a <code>DOT</code>.  This is a
 * valid expression and the Progress compiler and interpreter do not complain,
 * however there are no side effects that have any useful benefits.  The 
 * Progress environment is tolerant of these useless expressions.
 * <p>
 * When the first form is encountered, the <code>EQUALS</code> token type is
 * rewritten as an <code>ASSIGN</code> token type.  Saving this context is 
 * key to subsequent evaluation of an expression since the parser already
 * has determined the exact operation to be processed (otherwise there would
 * be ambiguity left behind in the resulting expression tree).
 * <p>
 * This rule is called from the top level {@link #single_block} and calls the
 * {@link #lvalue} and {@link #expr} rules as needed.  It also creates an
 * artificial tree node (of the <code>ASSIGNMENT</code> token type) to
 * contain the resulting tree.
 * <p>
 * Note that ANTLR reports ambiguity between this rule and the {@link #assign}
 * rule.  The ambiguity is resolved by the ordering of the top level rule
 * placing the <code>ASSIGN</code> language statement (which is unambiguous)
 * with higher precedence than the <code>assignment</code> rule. As long as
 * the <code>assign</code> rule is only used by the <code>assign_stmt</code>
 * rule, there is no real ambiguity.  For this reason, ambiguity warnings
 * are suppressed for this rule.
 * <p>
 * ANTLR detects ambiguity between this rule and <code>expr</code> because of 
 * the optional <code>lvalue</code> and <code>EQUALS</code> which conflict
 * with lookahead choices for <code>expr</code>.  What makes this OK is the
 * fact that there is an implicit precedence in choosing <code>EQUALS</code>
 * as an assignment operator over the equivalence operator. This makes the
 * ambiguity resolve naturally.  For this reason, ambiguity warnings
 * are suppressed for this rule.
 * <p>
 * This rule really needs a syntactic predicate (but we can't use these, see
 * the class overview for details) to work in all cases.  The problem is
 * that it inherently relies upon lookahead (of whether there is an 
 * <code>EQUALS</code> token in the 2nd position) to disambiguate between
 * the optional <code>lvalue</code> + <code>EQUALS</code> AND a standalone  
 * expression.  In practice, this works fine for the vast majority of cases,
 * but it breaks down in the following cases:
 * <p>
 * <ul>
 *    <li> <b>an <code>lvalue</code> with &gt; 1 token is standalone on a 
 *        line</b> (e.g. <code>i[8].</code> which is 4 tokens plus the
 *        terminating <code>DOT</code>) --&gt; note that as long as there
 *        actually IS an assignment on the right side of a complex 
 *        (multi-token) <code>lvalue</code>, there would not be a problem!
 *    <li> <b>an <code>lvalue</code> based expression is used (no assignment) 
 *         </b> (e.g. <code>queryHandle:prepare(someThing).</code> which is 6
 *         tokens plus the terminating <code>DOT</code> OR even simpler is 
 *         <code>queryHandle:query-open.</code> which is only 3 tokens and
 *         a terminating <code>DOT</code>)
 *    <li> language statements that look like functions (and often have
 *         functions with matching names!) but act like lvalues (see
 *         {@link #assign_type_syntax_stmt_list} which is called indirectly
 *         via the <code>lvalue</code> rule)
 * </ul>
 * <p>
 * All of these cases are resolved by matching the in the <code>lvalue</code>
 * rule and instrumenting the following <code>EQUALS</code> to be bypassed 
 * if a <code>DOT or NO-ERROR</code> are encountered.  This would mean that 
 * there is no following expression to match so in this case, the expression 
 * is bypassed too.
 * <p>
 * This approach also requires that the preceding <code>lvalue</code> be
 * bypassed if a left parenthesis is found in the second token position
 * (otherwise the <code>lvalue</code> would match the symbol in a function
 * call but then it would bomb out at the parenthesis.  <b>An exception to
 * this 'parenthesis avoidance' approach is made for those language statements
 * that look like function calls but act like lvalues.  An explicit test for
 * this situation bypasses the normal parenthesis avoidance mechanism.</b>
 * <p>
 * Another problem is disambiguation of the case where the <code>lvalue</code>
 * is followed by an operator (that is not EQUALS).  The ANTLR lookahead does
 * not naturally disambiguate this case, so an extra test for operators is
 * done, to avoid the assignment path in this case.  The result is usually a
 * useless expression whose evaluated result is discarded, but it is valid.
 * <p>
 * To implement all of this multiple disambiguating semantic predicates had
 * to be carefully inserted, empty alternatives (sometimes triggering 
 * associated actions) were added and the tree building was modified.
 * All-in-all this was a truly nasty solution.
 */
assignment!   
   :
      {
         boolean bypass = false;
         boolean lstmt  = false;
         boolean oper   = false;
         
         if (
              (LA(1) == KW_CUR_VAL  || LA(1) == KW_DYN_CURV ||
               LA(1) == KW_ENTRY    || LA(1) == KW_EXTENT   ||
               LA(1) == KW_FIX_CP   || LA(1) == KW_LENGTH   ||
               LA(1) == KW_OVERLAY  || LA(1) == KW_PUT_BITS ||
               LA(1) == KW_PUT_BYTE || LA(1) == KW_PUT_BYTS ||
               LA(1) == KW_PUT_DBL  || LA(1) == KW_PUT_FLT  ||
               LA(1) == KW_PUT_I64  || LA(1) == KW_PUT_LONG || 
               LA(1) == KW_PUT_SHT  || LA(1) == KW_PUT_STR  || 
               LA(1) == KW_PUT_UL   || LA(1) == KW_PUT_USHT ||
               LA(1) == KW_RAW      || LA(1) == KW_SET_B_OR ||
               LA(1) == KW_SET_PTR  || LA(1) == KW_SET_SZ   ||
               LA(1) == KW_SUBSTR)  && LA(2) == LPARENS)
         {
            lstmt = true;
         }
         
         // check for a 2nd token which is a binary operator following
         // an lvalue; such a case is not an assignment and must flow down
         // the standalone expression path (warning: EQUALS is deliberately
         // excluded from this test since the precedence must be given to
         // the EQUALS token that will be considered an ASSIGN operator)
         if (
              (LA(2) == KW_OR       || LA(2) == KW_AND      ||
               LA(2) == KW_NOT      || LA(2) == KW_EQ       || 
               LA(2) == NOT_EQ      || LA(2) == KW_NE       ||
               LA(2) == GT          || LA(2) == KW_GT       ||
               LA(2) == LT          || LA(2) == KW_LT       ||
               LA(2) == GTE         || LA(2) == KW_GTE      ||
               LA(2) == LTE         || LA(2) == KW_LTE      ||
               LA(2) == KW_BEGINS   || LA(2) == KW_MATCHES  ||
               LA(2) == KW_CONTAINS || LA(2) == PLUS        ||
               LA(2) == MINUS       || LA(2) == MULTIPLY    ||
               LA(2) == DIVIDE      || LA(2) == KW_MOD)
            )
         {
            oper = true;
         }
      }
      // This is how we would want to code this.  But the problem is that there
      // is ambiguity between the EQUALS inside the expr and the EQUALS here.
      // Instead, the preference must go to the ASSIGN operator (the EQUALS
      // here). That means that we must code this with more complexity. :(
      // (
      //    l:expr
      //    (eq:EQUALS^ { #eq.setType( ASSIGN ); } (rv:KW_NO_RET_V)? e:expr)?
      // )?
      (
         options { generateAmbigWarnings = false; }
         :
           { (LA(2) != LPARENS && !oper) || lstmt }?
           l:chained_object_members
           (
                eq:EQUALS^  { #eq.setType(ASSIGN); }
                
                // the empty alternative, it just triggers an action!
              | { LA(1) == DOT || LA(1) == KW_NO_ERROR }?
                { bypass = true; }
           )
         | // the empty alternative
      )?
      (
            { bypass == false }?
            (rv:KW_NO_RET_V)? e:expr
            
         |  // the empty alternative
      )
      (k:KW_NO_ERROR)?
      DOT!
      { 
         // in the init rule we determined that this looked like a language
         // statement, but if there was no EQUALS token, then bypass will
         // be true and this was really a function that should be an
         // assignment
         if (lstmt && !bypass)
         {
            // these are language statements, not assignments!
            ## = #([STATEMENT, "statement"], (eq, l, e), k);
            
            restoreAssignStyleStmt(#l);
         }
         else
         {
            // this is a true assignment statement
            if (bypass)
            {
               ## = #([ASSIGNMENT, "assignment"], l, k);
            }
            else
            {
               // standard form with the optional lvalue, optional assignment
               // operator and the required following expression
               ## = #([ASSIGNMENT, "assignment"], (eq, l, e), rv, k);
               
               restoreEqualsOperator(#l, #eq);
            }
         }
      }
   ;
   
/**
 * This rule creates an artificial tree node for any non-block language
 * statement (these are language statements that do not start or end
 * blocks).  The real work is delegated to {@link #stmt_list}.  
 * <p>
 * The only real work this rule does is create the tree node in a standard
 * manner.  The resulting language statement's subtree is a child of the
 * standard, artificial <code>STATEMENT</code> token type node except in
 * the case of embedded SQL where the node type of <code>EMBEDDED_SQL</code>
 * will be at the root.
 * <p>
 * Any statement that generates a <code>null</code> return AST (the only
 * one is {@link #empty_stmt} at this time), will bypass the creation of
 * the artificial <code>STATEMENT</code> root node.
 * <p>
 * Called by {@link #single_block}.
 */
statement!   
   :
      ( s:stmt_list )
      {
         if (#s == null)
         {
            // the empty statement can generate a null return AST
            ## = null;
         }
         else if (#s.getType() == EMBEDDED_SQL)
         {
            // don't root this at a STATEMENT node
            ## = #s;
         }
         else
         {
            // otherwise root everything at an artificial node
            ## = #([STATEMENT, "statement"], s);
         }
      }
   ;
   
/**
 * This is a simple (flat) list of alternatives for all language statements
 * that do not have anything to do with a starting or ending a Progress 4GL
 * block.  It is critical that each alternative in this list be properly
 * defined such that there is no unresolved ambiguity OR such that the
 * alternatives can be ordered so that ANTLR can eliminate ambiguity by
 * using standard lookahead (see options starting with create or delete). 
 * Any single word language statement should appear after any language
 * statement that starts with the same keyword but then has a 2nd required
 * keyword by which the two can be differentiated.
 * <p>
 * Although ANTLR properly handles the ambiguity of statements that start 
 * with common keywords like <code>CREATE</code> (assuming they are ordered
 * more specific to less specific), it still can detect ambiguity.  For this
 * reason, ambiguity warnings have been disabled.  <b>As statements are
 * added to this rule it is critical that the actual logic be reviewed to
 * ensure that no real ambiguity has been added.</b> 
 * <p>
 * Simple (single keyword, no options, no data or expressions) language
 * statements are implemented 'in-line' in this rule.  All statements
 * simple or rule references) are listed alphabetically unless there is a
 * need to avoid ambiguity.  
 * <p>
 * This rule seems to be where the statement-ending <code>DOT</code> should
 * be matched.  However, there are locations such as {@link #if_stmt} and
 * {@link #case_stmt} that don't easily fit into this plan.  So the end of
 * statement delimiter is coded in each statement separately.
 * <p>
 * Do NOT put an <code>END</code> statement into this list.  The nested
 * nature of the grammar's block structure is only possible because the
 * <code>END</code> statement is only found at the end of a block.  This
 * allows the ends of blocks to be identified since any loop that is
 * matching statements, assignments... will naturally stop when it hits
 * an end statement.  See {@link #single_block}.
 * <p>
 * Called by {@link #statement}.
 */
stmt_list   
   :
      {
         if (isUserDefFunctionCall(LT(1), LT(2)))
         {
            return;
         }

         CommonToken next  = (CommonToken) LT(1);
         CommonToken after = (CommonToken) LT(2);
      
         // RUN stmt filenames that start with a DOT need some processing here to allow
         // matching run_stmt below
         if (next  != null && next.getType()  == KW_RUN &&
             after != null && after.getType() == DOT    &&
             !followedByWhitespace(after))
         {
            // rewrite the type of the token to enable matching/merging in run_stmt
            after.setType(SYMBOL);
         }
      }
      (
         options { generateAmbigWarnings = false; }
         :
           accumulate_stmt
         | apply_stmt
         | assign_stmt
         | KW_BELL^ DOT!
         | block_level_stmt
         | buffer_compare_or_copy_stmt
         | call_stmt
         | case_stmt
         | catch_stmt
         | choose_stmt
         | clear_stmt
         | close_query_stmt
         | close_stored_proc_stmt
         | { LA(1) == KW_COLOR }? color_stmt
         | compile_stmt
         | connect_stmt
         | copy_lob_stmt
         | create_alias_stmt
         | create_browse_stmt
         | create_buffer_stmt
         | create_call_stmt
         | create_client_principal_stmt
         | create_database_stmt
         | create_dataset_stmt
         | create_datasource_stmt
         | create_query_stmt
         | create_sax_stmt
         | create_server_stmt
         | create_soap_header_stmt
         | create_soap_header_entryref_stmt
         | create_socket_stmt
         | create_temp_table_stmt
         | create_widget_pool_stmt
         | create_xml_stmt
         
           // can't use isRecord() here since in this location a record
           // name that conflicts with a variable name will match the record
           // first (and isRecord() does the reverse)
         | { sym.isTable(LT(2).getText()) }?
           create_stmt
           
           // must follow create_stmt due to ambiguity
         | create_widget_or_object_stmt
         | dde_advise_stmt
         | dde_execute_stmt
         | dde_get_stmt
         | dde_initiate_stmt
         | dde_request_stmt
         | dde_send_stmt
         | dde_terminate_stmt
         | define_stmt
         | delete_alias_stmt
         | delete_object_stmt
         | delete_procedure_stmt
         | delete_widget_stmt
         | delete_widget_pool_stmt
         | delete_stmt
         | KW_DICT^ DOT!
         | disable_triggers_stmt
         | disable_stmt
         | disconnect_stmt
         | display_stmt
         | empty_temp_table_stmt
         | enable_stmt
         | export_stmt
         | finally_stmt
         | find_stmt
         | form_stmt
         | get_stmt
         | get_key_value_stmt
         | hide_stmt
         | if_stmt
         | import_stmt
         | insert_stmt
         | io_stmt
         | load_stmt
         | msg_stmt
         | next_or_leave_stmt
         | next_prompt_stmt
         | on_stmt
         | open_query_stmt
         | os_misc_stmt
         | os_create_dir_stmt
         | os_delete_stmt
         | page_stmt
         | pause_stmt
         | p:KW_PROCESS^ { #p.setType(PROCESS_EVENTS); } KW_EVENTS! DOT!
         | publish_stmt
         | put_screen_stmt
         | put_cursor_stmt
         | put_key_value_stmt
         | put_stmt
         | KW_QUIT^ DOT!
         | raw_transfer_stmt
         | readkey_stmt
         | release_external_stmt
         | release_object_stmt
         | release_stmt
         | reposition_stmt
         | return_stmt
         | routine_level_stmt
         
           // these next 2 must be before run_stmt to eliminate ambiguity
         | run_stored_proc_stmt
         | run_super_stmt
         | run_stmt
         | save_cache_stmt
         | scroll_stmt
         | seek_stmt
         | set_or_update_or_prompt_for_stmt
         | shell_command_stmt
         | KW_SHOW_ST^ (KW_CLEAR)? DOT!
         | KW_STOP^ DOT!
         | status_stmt
         | subscribe_stmt
         
           // only match if this is the first stmt in a constructor
         | { constStmts == 1 }?
           super_stmt
           
         | system_dialog_color_stmt
         | system_dialog_font_stmt
         | system_dialog_get_dir_stmt
         | system_dialog_get_file_stmt
         | system_dialog_printer_setup_stmt
         | system_help_stmt
         | terminal_stmt
         
           // only match if this is the first stmt in a constructor
         | { constStmts == 1 }?
           this_object_stmt
           
         | trigger_procedure_stmt
         | transaction_mode_auto_stmt
         | underline_stmt
         | undo_stmt
         | unload_stmt
         | unsubscribe_stmt
         | up_or_down_stmt
         | use_stmt
         | using_stmt
         | validate_stmt
         | view_stmt
         
           // the wait-for stmts must be in this order
         | {
              sym.canLoadClass(LT(2).getText()) ||
              isObjectType(resolveLvalueCoreType(2, false))
           }?
           wait_for_dotnet_stmt
         | wait_for_stmt
         
         | embedded_sql
         | empty_stmt
         
           // P2J-extensions go here, these are not real 4GL!
         | p2j_open_url_stmt
      )
   ;

/**   
 * Matches both <code>DO</code> and <code>REPEAT</code> Progress 4GL language
 * statements.  This is possible because other than the primary (leftmost)
 * keyword being different (<code>DO</code> versus <code>REPEAT</code>), the  
 * structure of the 2 statements is syntactically identical. The meaning of
 * the 2 statements is different but this does not affect parsing at all.
 * <p>
 * All possible common clauses between this rule and {@link #for_stmt}
 * have been factored into {@link #common_block_options}.
 * <p>
 * In fact, if it were not for some small but meaningful differences, there
 * could have been a common rule for all 3 inner block statements.
 * <p>
 * Called by {@link #block_list}.
 * <p>
 * Bogus ANTLR ambiguity warnings have been disabled.
 * <p>
 * Uses the following rules:
 * {@link #for_record_spec}
 * {@link #preselect_phrase}
 * {@link #common_block_options}
 */
do_repeat_stmt      
   :
      {
         Aast node = null;
         Aast loop = null;
      }
      (KW_DO^ | KW_REPEAT^)
      (
         options { generateAmbigWarnings = false; }
         :
           // force promotion in a strong scope
           for_record_spec[false, false, true]
         | preselect_phrase
         | node = common_block_options
           {
              // node is replaced every time through the loop so we must
              // cache off the first non-null result for post-processing
              if (loop == null && node != null)
              {
                 loop = node;
              }
           }
      )*
      {
         // implement any deferred tree processing for a mis-matched loop
         // increment sub-tree
         if (loop != null)
         {
            ##.addChild(loop);
         }
      }
   ;

/**   
 * Matches <code>FOR</code> Progress 4GL language statement.
 * <p>
 * All possible common clauses between this rule and {@link #do_repeat_stmt}
 * have been factored into {@link #common_block_options}.
 * <p>
 * In fact, if it were not for some small but meaningful differences,
 * there could have been a common rule for all 3 inner block statements.
 * In particular, a variant of the {@link #preselect_phrase} rule that is
 * used by the <code>do_repeat_stmt</code> is inlined in here.  The
 * inline grammar differs in that there is no <code>preselect</code> keyword.
 * <p>
 * Called by {@link #block_list}.
 * <p>
 * Uses {@link #each_first_last_spec} to implement a common record phrase
 * list.
 */
for_stmt   
   :
      {
         Aast node = null;
         Aast loop = null;
      }
      // note: the FOR stmt forces re-promotion
      KW_FOR^ each_first_last_spec[true, true]
      (
         node = common_block_options
         {
            // node is replaced every time through the loop so we must
            // cache off the first non-null result for post-processing
            if (loop == null && node != null)
            {
               loop = node;
            }
         }
      )*
      {
         // implement any deferred tree processing for a mis-matched loop
         // increment sub-tree
         if (loop != null)
         {
            ##.addChild(loop);
         }
      }
   ;
   
/**   
 * Matches common options that a <code>FOR, DO or REPEAT</code> Progress 4GL
 * language statement can implement.
 * <p>
 * Note that the block control options are defined such that they can be
 * accepted in any order.  In addition, there is no check to ensure that
 * only one of each option is used, although this will not cause a problem
 * for correct Progress 4GL source.
 * <p>
 * Bogus ANTLR ambiguity warnings have been disabled.
 * <p>
 * Called by {@link #for_stmt} and {@link #do_repeat_stmt}.
 * <p>
 * Uses:
 * {@link #query_tuning_phrase}
 * {@link #while_clause}
 * {@link #on_event_phrase}
 * {@link #loop_incr}
 * {@link #collate_clause}
 * {@link #sort_order_clause}
 * {@link #frame_phrase}
 * {@link #down_clause_or_loop_incr}
 *
 * @return   A loop increment sub-tree that was matched in a frame phrase
 *           due to ambiguity and which needs to be moved to the caller's
 *           tree. If <code>null</code> there is nothing to do.
 */
common_block_options returns [Aast loop]
   {
      loop = null;
   }
   :
      (
         options { generateAmbigWarnings = false; }
         :            
           while_clause
         | KW_TRANS
         | KW_STOP_AFT^ expr
         | on_event_phrase
         | query_tuning_phrase
         | KW_BREAK
         
           // undocumented KW_BREAK synonym
         | g:KW_GROUP { #g.setType(KW_BREAK); }
         
         | (collate_clause | sort_order_clause)
         
           // matches a loop increment that appears before a frame phrase but
           // the frame phrase itself can match a loop increment too since
           // it is ambiguous with the down_clause
         | loop_incr
         
         | f:frame_phrase[true, false]
           {
              loop = #f.getImmediateChild(KW_TO, null);
              
              if (loop != null)
              {
                 // fixups haven't been done yet so we force the parent which
                 // is a prerequiisite for remove()
                 loop.setParent(#f);
                 loop.remove();
              }
           }
      )
   ;
   
/**   
 * Matches the loop increment definition defined by the <code>TO</code>
 * keyword and optionally the <code>BY</code> keyword.  Common rule references
 * were used to reuse working, tested code, however the resulting tree may
 * not be intuitive since the leftmost child of the root <code>TO</code>
 * keyword will be an {@link #assign} (which is rooted by an 
 * <code>ASSIGN</code> token that has an lvalue and expression as children.
 * The next child of the <code>TO</code> is the loop terminating expression
 * followed by the optional {@link #by_clause}.
 * <p>
 * Used by {@link #do_repeat_stmt} and {@link #for_stmt}.
 * <p>
 * The subtree is rooted by the keyword (which is the reason for using a 
 * separate rule for something so simple).
 * <p>
 * Bogus ANTLR ambiguity warnings have been disabled.      
 */
loop_incr   
   :
      assign KW_TO^ expr
      (
         options { generateAmbigWarnings = false; }
         :
         by_clause
      )?
   ;
   
/**
 * Disambiguate between a {@link #loop_incr} (which always starts with an
 * {@link #lvalue}) and a {@link #down_clause} (which optionally starts with
 * an {@link #expr}). The matching in this rule is in-lined from the original
 * rules in order to left-factor the common leftmost expression construct.
 */
down_clause_or_loop_incr!
   :
      (
           d0:KW_DOWN
         | e0:expr
           (
                d1:KW_DOWN
                
                // replacement portion for loop increment
              | t:KW_TO e1:expr
                (
                   options { generateAmbigWarnings = false; }
                   :
                   b:by_clause
                )?
           )
      )
      {
         if (#d0 != null)
         {
            ## = #d0;
         }
         else if (#d1 != null)
         {
            ## = #(d1, e0);
         }
         else
         {
            // normally we must explicitly match an "EQUALS expression"
            // after an lvalue, in this case this is matched by the
            // expr rule and we need to transform it into the sub-tree
            // that it should have been
                   
            Aast assn = #e0.getChildAt(0);
            assn.setType(ASSIGN);
            
            ## = #(t, assn, e1, b);
         }
      }
   ;

/**   
 * Matches the loop termination expression defined by the <code>WHILE</code>
 * keyword.  The keyword is always followed by an expression. 
 * <p>
 * Used by {@link #do_repeat_stmt} and {@link #for_stmt}.
 * <p>
 * The subtree is rooted by the keyword (which is the reason for using a 
 * separate rule for something so simple).
 */
while_clause   
   :
      KW_WHILE^ expr
   ;
   
/**   
 * Matches the <code>IF THEN ELSE</code> language statement.  This rule uses   
 * rule references {@link #then_clause} and {@link #else_clause} to properly
 * build the AST.
 * <p>
 * This language statement has inherent ambiguity with the built-in function
 * that uses the same <code>IF THEN ELSE</code> keywords (and has a nearly
 * identical meaning but is only for use in expressions).  Progress gives
 * precedence to matching the language statement first and then the function
 * form.  This parser implements this rule in the {@link #statement} portion
 * of the top-level {@link #single_block} rule.  Since this is matched before
 * expressions (which are in the {@link #assignment} rule which has lower
 * precedence than <code>statement</code>) the logic works properly and all
 * ambiguity is resolved.
 * <p>
 * Note that there is another inherent ambiguity in this construction.  It is
 * the classic problem with the optional <code>ELSE</code> portion of the
 * rule. The standard matching of ANTLR in this case already resolves the
 * detected ambiguity so there is nothing to do except disable the warnings.
 * <p>
 * Used by {@link #stmt_list}.
 */
if_stmt   
   :
      KW_IF^ expr then_clause
      (
         options { warnWhenFollowAmbig = false; }
         :
         else_clause
      )?
   ;
   
/**   
 * Matches the <code>THEN</code> block of an <code>IF THEN ELSE</code>
 * language statement.  This rule makes a single, recursive call to the
 * core rule {@link #single_block}.  This allows a single statement, inner
 * block (<code>DO, REPEAT or FOR</code>) or assignment to be included
 * after the <code>THEN</code>.  Note that no checking is done on whether
 * a function, trigger or procedure definition is used instead of the above
 * constructions.  While this seems wrong, it is exactly how Progress handles
 * this situation (without complaining)!
 * <p>
 * It is possible to have an empty then block if the next statement is 
 * <code>END</code> or if the end of file is reached. Progress tolerates
 * this syntax silently.
 * <p>
 * Used by {@link #if_stmt} and {@link #when_clause}.
 * <p>
 * The subtree is rooted by the keyword (which is the reason for using a 
 * separate rule for something so simple).
 */
then_clause   
   :
      {
         checkAndFixLabelWithReservedKeywordPrefix(2);
      }
      KW_THEN^
      (
            { LA(1) != EOF && LA(1) != KW_END }?
            single_block[true, false]
         | // empty alternative
      )
   ;
   
/**   
 * Matches the <code>ELSE</code> block of an <code>IF THEN ELSE</code>
 * language statement.  This rule makes a single, recursive call to the
 * core rule {@link #single_block}.  This allows a single statement, inner
 * block (<code>DO, REPEAT or FOR</code>) or assignment to be included
 * after the <code>ELSE</code>.  Note that no checking is done on whether
 * a function, trigger or procedure definition is used instead of the above
 * constructions.  While this seems wrong, it is exactly how Progress handles
 * this situation (without complaining)!
 * <p>
 * It is legal for this block (which follows the <code>ELSE</code> keyword)
 * to be missing.  For this reason, the call to <code>single_block</code>
 * was made optional.  As usual, bogus ANTLR ambiguity warnings are disabled.
 * <p>
 * Used by {@link #if_stmt}.
 * <p>
 * The subtree is rooted by the keyword (which is the reason for using a 
 * separate rule for something so simple).
 */
else_clause
   :
      {
         checkAndFixLabelWithReservedKeywordPrefix(2);
      }
      KW_ELSE^
      (
         options { generateAmbigWarnings = false; }
         :
         single_block[true, false]
      )?
   ;
   
/**   
 * Matches the <code>CASE</code> language statement.  This rule uses   
 * rule references {@link #when_clause}, {@link #otherwise} and
 * {@link #end_stmt} to properly build the AST.
 * <p>
 * Used by {@link #stmt_list}.
 */
case_stmt   
   :
      KW_CASE^ expr (COLON! | DOT!) (when_clause)* (otherwise)? end_stmt[true]
   ;
   
/**   
 * Matches the <code>WHEN</code> block of a <code>CASE</code>
 * language statement.  After matching the list of possible expressions using
 * the {@link #simple_when_clause} rule reference, this rule makes a single
 * call to the {@link #then_clause}.
 * <p>
 * Used by {@link #case_stmt}.
 * <p>
 * Note that there is an ambiguity caused between the <code>OR</code> keyword
 * implemented in {@link #expr} and by the use of the <code>OR</code>
 * keyword in a <code>WHEN</code> clause (additional WHEN conditions can
 * be added to the same statement using an <code>OR WHEN condition</code>
 * syntax). Since the following condition is an expression, one must detect
 * the use of the following <code>WHEN</code> as the next token after the
 * <code>OR</code> to disambiguate. The <code>expr</code> rule which 
 * implements the logical <code>OR</code> processing thus uses a semantic
 * predicate to terminate expression processing (and not match on the
 * <code>OR</code> in the case where lookahead 2 is the reserved keyword
 * <code>WHEN</code>.
 * <p>
 * The subtree is rooted by the keyword (which is the reason for using a 
 * separate rule for something so simple).
 */
when_clause   
   :
      simple_when_clause (KW_OR! simple_when_clause)* then_clause
   ;

/**   
 * Matches the <code>OTHERWISE</code> block of a <code>CASE</code>
 * language statement.  This rule makes a single, recursive call to the
 * core rule {@link #single_block}.  This allows a single statement, inner
 * block (<code>DO, REPEAT or FOR</code>) or assignment to be included
 * after the <code>THEN</code>.  Note that no checking is done on whether
 * a function, trigger or procedure definition is used instead of the above
 * constructions.  While this seems wrong, it is exactly how Progress handles
 * this situation (without complaining)!
 * <p>
 * Used by {@link #case_stmt}.
 * <p>
 * The subtree is rooted by the keyword (which is the reason for using a 
 * separate rule for something so simple).
 */
otherwise
   :
      {
         checkAndFixLabelWithReservedKeywordPrefix(2);
      }
      KW_OTHER^ single_block[true, false]
   ;
   
/**   
 * Matches an <code>END</code> Progress 4GL language statement.  This is the
 * rule that ends all Progress block definitions.
 * <p>
 * Note that Progress 4GL allows an optional keyword to be placed between the
 * <code>KW_END</code> and the <code>DOT</code> for some blocks.  This works
 * for <code>CASE</code>, <code>PROCEDURE</code>, <code>FUNCTION</code> and
 * <code>TRIGGERS</code> keywords.  It does not work for <code>DO</code>,
 * <code>REPEAT</code> or <code>FOR</code> keywords.  This feature is 
 * completely implemented except in Progress 4GL, abbreviations cannot be
 * used in this optional keyword, but in this implementation abbreviations
 * are tolerated.  Since they should not be encountered in practice, this is
 * an acceptable decision.
 * <p>
 * <b>DO NOT EVER include this directly in the {@link #stmt_list} rule
 * otherwise all block processing will break!</b>
 * <p>
 * Bogus ANTLR ambiguity warnings have been disabled.
 *
 * @param    eat
 *           Match (and drop) a following <code>DOT</code> if one exists.
 */
end_stmt [boolean eat]
   :
      // KW_DO, KW_REPEAT, KW_FOR are not supported as optional
      // documentation keywords after KW_END!
      KW_END!                             
      (
         options { generateAmbigWarnings = false; }
         :
           KW_CASE!
         | KW_CATCH!
         | KW_CLASS!
         | KW_CONSTRUC!
         | KW_DESTRUCT!
         | KW_FINALLY!
         | KW_FUNCT!
         | KW_GET!
         | KW_INTERFAC!
         | KW_METHOD!
         | KW_PROC!
         | KW_SET!
         | KW_TRIGGERS!
      )?
      (
           { eat }?
           DOT!
         | // empty alternative
      )
   ;

/**
 * Matches the <code>FIND</code> language statement and implements all
 * forms with most of the core processing delegated to the 
 * {@link #record_phrase} rule.
 * <p>
 * Note that one option <code>CURRENT</code> allows more possible record
 * phrase options than are actually valid in Progress, however it is
 * simply a superset of the valid options and as such is safe.
 * <p>
 * There is inherent ambiguity here with the use of unreserved Progress
 * keywords <code>PREV, LEFT and OUTER-JOIN</code> that is detected by
 * ANTLR.  The code has been reviewed and the conflicting cases either
 * can't occur in valid Progress code or they will be properly resolved
 * anyway.  For this reason, ambiguity warnings are disabled.
 */
find_stmt   
   :
      KW_FIND^
      (
         options { generateAmbigWarnings = false; }
         :      
         KW_FIRST | KW_LAST | KW_NEXT | KW_PREV | KW_CURRENT
      )?
      record_phrase[false, false]
      DOT!
   ;
   
/**
 * This is a simple placeholder to match the use of a SQL <code>SELECT</code>
 * statement used directly in 4GL code in order to obtain a count of records
 * from a given table.
 */
embedded_sql
   :
      {
         astFactory.makeASTRoot(currentAST, #[EMBEDDED_SQL,""]);
         embeddedSql = true;
         inUnion = false;
      }
      (
           alter_table_stmt
         | close_cursor_stmt
         | create_index_stmt
         | create_table_stmt
         | create_view_stmt
         | declare_cursor_stmt
         | delete_from_stmt
         | drop_index_stmt
         | drop_table_stmt
         | drop_view_stmt
         | fetch_stmt
         | grant_stmt
         | insert_into_stmt
         | open_cursor_stmt
         | revoke_stmt
         | union_stmt                // handles select_stmt too
         | sql_update_stmt
      )
      DOT!
      {
         embeddedSql = false;
      }
   ;
   
/**
 * Support SQL <code>ALTER TABLE</code> directly in 4GL code. Called from
 * {@link #embedded_sql} and uses the following rules:
 * <p>
 * <ul>
 *    <li> {@link #record}
 *    <li> {@link #sql_column_def}
 * </ul>
 */
alter_table_stmt
   :
      a:KW_ALTER^ { #a.setType(ALTER_TABLE); } KW_TABLE! record[true, false, false]
      sql_column_def
   ;
   
/**
 * Matches the <code>CLOSE</code> keyword with a following {@link #symbol}.
 * Called from {@link #embedded_sql}.
 */
close_cursor_stmt
   :
      c:KW_CLOSE^ { #c.setType(CLOSE_CURSOR); } symbol
   ;
   
/**
 * Matches the <code>CREATE INDEX</code> statement with a following
 * {@link #symbol}, {@link #record} and one or more {@link #lvalue}.
 * Called from {@link #embedded_sql}.
 */
create_index_stmt
   :
      c:KW_CREATE^ { #c.setType(CREATE_INDEX); } (KW_UNIQUE)? KW_INDEX!
      any_non_reserved_symbol KW_ON! record[true, false, false]
      LPARENS! lvalue (COMMA! lvalue)* RPARENS!
   ;
   
/**
 * Support SQL <code>CREATE TABLE</code> directly in 4GL code. Called from
 * {@link #embedded_sql} and uses the following rules:
 * <p>
 * <ul>
 *    <li> {@link #symbol}
 *    <li> {@link #sql_column_spec}
 *    <li> {@link #sql_unique_index}
 * </ul>
 */
create_table_stmt
   :
      c:KW_CREATE^ { #c.setType(CREATE_TABLE); } KW_TABLE! symbol
      (sql_column_spec[true] | sql_unique_index)+
   ;
   
/**
 * Support SQL <code>CREATE VIEW</code> directly in 4GL code. Called from
 * {@link #embedded_sql} and uses the following rules:
 * <p>
 * <ul>
 *    <li> {@link #symbol}
 *    <li> {@link #sql_column_spec}
 *    <li> {@link #sql_unique_index}
 * </ul>
 */
create_view_stmt
   :
      c:KW_CREATE^ { #c.setType(CREATE_VIEW); } KW_VIEW! symbol
      (sql_name_list[true])? KW_AS! select_stmt
      (w:KW_WITH { #w.setType(WITH_CHECK_OPTION); } KW_CHECK! KW_OPTION!)?
   ;
   
/**
 * Support SQL <code>DECLARE CURSOR</code> directly in 4GL code. Called from
 * {@link #embedded_sql} and uses the following rules:
 * <p>
 * <ul>
 *    <li> {@link #symbol}
 *    <li> {@link #union_stmt}
 * </ul>
 */
declare_cursor_stmt
   :
      d:KW_DECLARE^ { #d.setType(DECLARE_CURSOR); } symbol KW_CURSOR! KW_FOR!
      union_stmt 
      (
         KW_FOR! (
                      KW_READ_ONL
                    | (r:KW_READ (KW_ONLY!)? { #r.setType(KW_READ_ONL); })
                    | KW_UPDATE
                 )
      )? 
   ;
   
/**
 * Support SQL <code>DELETE FROM</code> directly in 4GL code. Called from
 * {@link #embedded_sql} and uses the following rules:
 * <p>
 * <ul>
 *    <li> {@link #record}
 *    <li> {@link #where_clause}
 *    <li> {@link #where_current_of}
 * </ul>
 */
delete_from_stmt
   :
      d:KW_DELETE^ { #d.setType(DELETE_FROM); } KW_FROM! record[true, false, false]
      (where_clause | where_current_of)?
   ;
   
/**
 * Support SQL <code>DROP INDEX</code> directly in 4GL code. Called from
 * {@link #embedded_sql} and uses {@link #symbol}.
 */
drop_index_stmt
   :
      d:KW_DROP^ { #d.setType(DROP_INDEX); } KW_INDEX! any_non_reserved_symbol
   ;
   
/**
 * Support SQL <code>DROP TABLE</code> directly in 4GL code. Called from
 * {@link #embedded_sql} and uses {@link #record}.
 */
drop_table_stmt
   :
      d:KW_DROP^ { #d.setType(DROP_TABLE); } KW_TABLE! record[false, false, false]
   ;
   
/**
 * Support SQL <code>DROP VIEW</code> directly in 4GL code. Called from
 * {@link #embedded_sql} and uses {@link #symbol}.
 */
drop_view_stmt
   :
      d:KW_DROP^ { #d.setType(DROP_VIEW); } KW_VIEW! symbol
   ;
   
/**
 * Support SQL <code>FETCH</code> directly in 4GL code. Called from
 * {@link #embedded_sql}. Uses {@link #symbol} and {@link #sql_into}.
 */
fetch_stmt
   :
      KW_FETCH^ symbol sql_into
   ;
   
/**
 * Support SQL <code>GRANT</code> directly in 4GL code. Called from
 * {@link #embedded_sql}. Uses {@link #sql_update_rights}, 
 * {@link #sql_on_table} and {@link #sql_to_from_users}.
 */
grant_stmt
   :
      KW_GRANT^
      (
           KW_ALL (KW_PRIVILEG!)?
         | KW_SELECT
         | KW_INSERT
         | KW_DELETE
         | sql_update_rights  
      )+
      sql_on_table sql_to_from_users
      (w:KW_WITH { #w.setType(WITH_GRANT_OPTION); } KW_GRANT! KW_OPTION!)?
   ;
   
/**
 * Support SQL <code>INSERT INTO</code> directly in 4GL code. Called from
 * {@link #embedded_sql}. Uses {@link #record}, {@link #select_stmt}, 
 * {@link #sql_simple_column_list} and {@link #sql_value_list}.
 */
insert_into_stmt
   :
      i:KW_INSERT { #i.setType(INSERT_INTO); } KW_INTO! record[true, false, false]
      (sql_simple_column_list)? (sql_value_list | select_stmt)
   ;
   
/**
 * Matches the <code>OPEN</code> keyword with a following {@link #symbol}.
 * Called from {@link #embedded_sql}.
 */
open_cursor_stmt
   :
      o:KW_OPEN^ { #o.setType(OPEN_CURSOR); } symbol
   ;
   
/**
 * Support SQL <code>REVOKE</code> directly in 4GL code. Called from
 * {@link #embedded_sql}. Uses {@link #sql_update_rights}, 
 * {@link #sql_on_table} and {@link #sql_to_from_users}.
 */
revoke_stmt
   :
      KW_REVOKE^
      (
           KW_ALL (KW_PRIVILEG!)?
         | KW_SELECT
         | KW_INSERT
         | KW_DELETE
         | sql_update_rights  
      )+
      sql_on_table sql_to_from_users
   ;
   
/**
 * Support SQL <code>SELECT</code> directly in 4GL code. Called from
 * {@link #embedded_sql} and uses the following rules:
 * <p>
 * <ul>
 *    <li> {@link #sql_column_list}
 *    <li> {@link #sql_into}
 *    <li> {@link #sql_from}
 *    <li> {@link #where_clause}
 *    <li> {@link #sql_group_by}
 *    <li> {@link #sql_having}
 *    <li> {@link #sql_order_by}
 *    <li> {@link #frame_phrase}
 * </ul>
 */
select_stmt
   :
      {
         selectNestLevel++;
      }
      KW_SELECT^
      (KW_ALL | KW_DISTINCT)?
      (MULTIPLY | cl:sql_column_list)
      (sql_into)?
      sql_from
      {
         // the field list may have symbols that couldn't be resolved because
         // they were qualified with SQL correlation names that are only
         // defined during the FROM clause processing; fix these up as proper
         // fields now that the FROM clause buffer processing is complete
         if (#cl != null)
         {
            Iterator iter = #cl.iterator();
            
            while (iter.hasNext())
            {
               Aast next = (Aast) iter.next();
               
               if (next.getType() == SYMBOL && sym.isField(next.getText()))
               {
                  sym.annotateField(next, null, true, true);
               }
            }
         }
      }
      (
         options { generateAmbigWarnings = false; }
         :
           where_clause
         | sql_group_by
         | sql_having
         | { !inUnion }?
           sql_order_by
         
           // this can conflict with the "with check option" clause of the
           // create_view_stmt so we disabiguate it here
         | { LA(2) != KW_CHECK && LA(3) != KW_OPTION }?
           frame_phrase[false, false]
      )*
      {
         selectNestLevel--;
      }
   ;
   
/**
 * Matches the SQL <code>UNION</code> statement which can be the same as a
 * simple {@link #select_stmt}, or it can contain a nested set of select
 * statements joined by the <code>UNION</code> keyword. There is no limit
 * to the number of select statements joined in a union. Note that if a
 * union is active, then any {@link #sql_order_by} clause will attach to
 * the top-level union statement node rather than to any of the select
 * statements. Called by {@link #embedded_sql}.
 */
union_stmt
   :
      {
         sqlNestLevel++;
      }
      
      select_stmt
      (
         options { generateAmbigWarnings = false; }
         :
         KW_UNION^ { inUnion = true; } (KW_ALL)? union_stmt
      )*
      
      // the ORDER BY clause is part of a UNION statement if the UNION
      // keyword is present; the inUnion flag disables matching on such
      // a construct in the select_stmt; in addition, only the top level
      // union stmt should match it (since it is a clause that affects the
      // entire union)
      (
           { inUnion && sqlNestLevel == 1 }?
           sql_order_by
           
         | // empty alternative
      )
      
      {
         sqlNestLevel--;
      }
   ;
   
/**
 * Matches the SQL <code>UPDATE</code> statement. Uses {@link #record},
 * {@link #sql_assign_list}, {@link #where_clause} and
 * {@link #where_current_of}. Called by {@link #embedded_sql}.
 */
sql_update_stmt
   :
      u:KW_UPDATE^ { #u.setType(UPDATE_SQL); } record[true, false, false] sql_assign_list
      (where_clause | where_current_of)?
   ;
   
/**
 * Matches a sub-select in a <code>WHERE</code> clause.
 */
sub_select_stmt
   :
      (
           KW_ANY^
         | KW_ALL^
         | KW_SOME^
         | KW_EXISTS^
         | n:KW_NOT^ KW_EXISTS! { #n.setType(NOT_EXISTS); }
      )? 
      LPARENS! select_stmt RPARENS!
   ;
   
/**
 * Matches a sub-select in a <code>WHERE</code> clause that is based on the
 * <code>IN</code> operator.
 */
in_operator_sub_select_stmt
   :
      (n:KW_NOT^ KW_IN! { #n.setType(NOT_IN); } | KW_IN^) in_operator_parms
   ;

/**
 * Matches the value list or sub-select portion of an <code>IN</code>
 * operator.
 */
in_operator_parms
   :
      LPARENS^ (select_stmt | sum_expr (COMMA! sum_expr)*) RPARENS!
   ;
   
/**
 * Matches the list of column assignments in an {@link #sql_update_stmt}
 * which are rooted at a <code>SET</code> keyword.
 */
sql_assign_list
   :
      KW_SET^ sql_assign (COMMA! sql_assign)*
   ;
   
/**
 * Matches the SQL assignment of an {@link #sql_value} to an {@link #lvalue}.
 * Called by {@link #sql_assign_list}.
 */
sql_assign
   :
      lvalue eq:EQUALS^ { #eq.setType(ASSIGN); } sql_value
   ;
   
/**
 * Matches the list of fields in a {@link #insert_into_stmt}. Uses
 * {@link #lvalue}.
 */
sql_simple_column_list
   :
      {
         astFactory.makeASTRoot(currentAST, #[COLUMN_LIST,""]);
      }
      LPARENS! lvalue (COMMA! lvalue)* RPARENS!
   ;
   
/**
 * Matches the list of values in a {@link #insert_into_stmt}. Uses
 * {@link #expr}.
 */
sql_value_list
   :
      KW_VALUES^ LPARENS! sql_value (COMMA! sql_value)* RPARENS!
   ;
   
/**
 * Matches a single {@link #expr} or <code>KW_NULL</code>. Used in
 * {@link #sql_value_list} and {@link #sql_update_stmt}.
 */
sql_value
   :
      KW_NULL | expr
   ;

/**
 * Matches SQL <code>UPDATE</code> rights option with a list of columns
 * using {@link #lvalue}. Used by {@link #grant_stmt}.
 */
sql_update_rights
   :
      KW_UPDATE^ LPARENS! lvalue (COMMA! lvalue)* RPARENS!
   ;
   
/**
 * Matches SQL <code>ON</code> followed by a {@link #record}. Used by
 * {@link #grant_stmt}.
 */
sql_on_table
   :
      KW_ON^ record[false, false, false]
   ;
   
/**
 * Matches SQL <code>TO</code> or <code>FROM</code> rights option with a list
 * of users as {@link #symbol} or the keyword <code>PUBLIC</code>. Used by
 * {@link #grant_stmt} and {@link #revoke_stmt}.
 */
sql_to_from_users
   :
      (KW_TO^ | KW_FROM^) 
      (
         options { generateAmbigWarnings = false; }
         :
         KW_PUBLIC | symbol (COMMA! symbol)*
      )
   ;
   
/**
 * Matches the <code>WHERE CURRENT OF</code> keyword sequence and a following
 * {@link #symbol}. Used by {@link #delete_from_stmt}.
 */
where_current_of
   :
      w:KW_WHERE^ { #w.setType(WHERE_CURRENT_OF); } KW_CURRENT! KW_OF! symbol
   ;

/**
 * Matches the <code>ADD</code>, <code>DROP</code> or <code>ALTER</code>
 * specification for an {@link #alter_table_stmt}. Uses the following:
 * <p>
 * <ul>
 *    <li> {@link #sql_data_type}
 *    <li> {@link #record}
 *    <li> {@link #sql_default_value}
 *    <li> {@link #label}
 *    <li> {@link #format_string}
 *    <li> {@link #column_label}
 *    <li> {@link #case_sensitive}
 * </ul>
 */
sql_column_def
   :
      {
         boolean add  = false;
         boolean drop = false;
      }
      (KW_ADD^ { add = true; } | KW_DROP^ { drop = true; } | KW_ALTER^)
      KW_COL!
      (
           { add }?
           sql_column_spec[false]
           
         | record[true, false, false]
      )
      (
          { !drop }?
          sql_column_options
          
         | // empty alternative
      )
   ;
   
/**
 * Matches a column name as a {@link #symbol} and the following data type
 * as {@link #sql_data_type}. Called by {@link #sql_column_def} and
 * {@link #create_table_stmt}.
 */
sql_column_spec [boolean root]
   :
      {
         if (root)
         {
            astFactory.makeASTRoot(currentAST, #[KW_COL,""]);
         }
      }
      (symbol | d:DB_SYMBOL { #d.setType(SYMBOL); }) sql_data_type
   ;
   
/**
 * Matches <code>UNIQUE</code> followed by a parenthesized list of column
 * names using {@link #sql_column_list}. This is a unique index definition.
 * Called by {@link #create_table_stmt}.
 */
sql_unique_index
   :
      KW_UNIQUE^ sql_name_list[false]
   ;
   
/**
 * Matches a parenthesized list of column names as {@link #symbol}. Called
 * by {@link #sql_unique_index} and {@link #create_view_stmt}.
 */
sql_name_list [boolean root]
   :
      {
         if (root)
         {
            astFactory.makeASTRoot(currentAST, #[KW_COLUMNS,""]);
         }
      }
      LPARENS! symbol (COMMA! symbol)* RPARENS!
   ;
   
/**
 * Matches the possible clauses to define column attributes or options.
 * Uses the following:
 * <p>
 * <ul>
 *    <li> {@link #sql_default_value}
 *    <li> {@link #label}
 *    <li> {@link #format_string}
 *    <li> {@link #column_label}
 *    <li> {@link #case_sensitive}
 *    <li> {@link #sql_not_null}
 * </ul>
 */
sql_column_options
   :
      (
           sql_default_value
         | label
         | format_string[false]
         | column_label
         | case_sensitive
         | sql_not_null
      )
   ;
   
/**
 * Matches the <code>DEFAULT</code> keyword and the following literal. Used
 * by {@link #sql_column_options}.
 */
sql_default_value
   :
      KW_DEFAULT^ literal
   ;
   
/**
 * Matches the <code>NOT NULL</code> keywords and the optional following
 * <code>UNIQUE</code>. Used by {@link #sql_column_options}.
 */
sql_not_null
   :
      n:KW_NOT^ { #n.setType(NOT_NULL); } KW_NULL! (KW_UNIQUE)?
   ;
   
/**
 * Matches all possible SQL data types.
 * <p>
 * <pre>
 * CHARACTER
 * DECIMAL
 * DATE
 * DOUBLE
 * FLOAT
 * INTEGER
 * LOGICAL
 * NUMERIC
 * REAL
 * SMALLINT
 * </pre>
 */
sql_data_type
   :
      (
           KW_CHAR
         | KW_DEC
         | KW_DATE
         | KW_DOUBLE
         | KW_FLOAT
         | KW_INT
         | KW_LOGICAL
         | KW_NUMERIC
         | KW_REAL
         | KW_SMALLINT
      )
      (sql_data_option)?
   ;
   
/**
 * Matches limited options (for column data type specifiers) as numeric
 * literals in parenthesis. Called by {@link #sql_data_type}.
 */
sql_data_option
   :
      LPARENS^ literal (COMMA! literal)? RPARENS!
   ;
   
/**
 * Matches the list of fields or SQL aggregate functions in a
 * {@link #select_stmt}.  Uses {@link #sql_individual_column}. 
 */
sql_column_list
   :
      {
         astFactory.makeASTRoot(currentAST, #[COLUMN_LIST,""]);
      }
      sql_individual_column (COMMA! sql_individual_column)*
   ;

/**
 * Matches a single field or expression (including SQL aggregate functions) in 
 * a {@link #sql_column_list}. Uses {@link #expr}, {@link #format_phrase} and
 * {@link #any_non_reserved_symbol}. The symbol processing is needed to
 * match on qualified field names that reference aliased table names
 * (a "SQL correlation name").
 */
sql_individual_column
   :
      (
         options { generateAmbigWarnings = false; }
         :
           { !sym.isField(LT(1).getText()) }?
           any_non_reserved_symbol
           
         | expr
      )
      (
         options { generateAmbigWarnings = false; }
         :
         format_phrase[null, null, false, false]
      )?
   ;
   
/**
 * Matches an embedded SQL aggregate function call. The following are
 * supported:
 * <p>
 * <pre>
 * AVG
 * COUNT
 * MAX
 * MIN
 * SUM
 * </pre>
 */
sql_aggregate_funcs
   :
      {
         boolean count = false;
      }
      (KW_AVG^ | KW_COUNT^ { count = true; } | KW_MAX^ | KW_MIN^ | KW_SUM^)
      LPARENS!
      (
           { count }?
           m:MULTIPLY
           {
               #m.setType(STRING);
               #m.setText("\"" + #m.getText() + "\"");
           }
         | (KW_DISTINCT)?
           (
              options { generateAmbigWarnings = false; }
              :
                { !sym.isField(LT(1).getText()) }?
                any_non_reserved_symbol
                
              | expr
           )
      )
      RPARENS!
      {
         saveAndReplaceType(##, count ? FUNC_INT : FUNC_POLY);
         
         // manually write our function-specific annotations
         ##.putAnnotation("builtin", new Boolean(true));
         ##.putAnnotation("sql", new Boolean(true));
      }
   ;
   
/**
 * Matches an embedded SQL <code>FROM</code> clause and the following
 * {@link #sql_table_spec}.
 */
sql_from
   :
      KW_FROM^
      sql_table_spec
      (
         options { generateAmbigWarnings = false; }
         :
           // this is an implicit join
           (COMMA! sql_table_spec)*
           
         | sql_explicit_join
      )?
   ;
   
/**
 * Matches all 3 kinds of explicit join (inner joins, left outer joins and
 * right outer joins). Uses {@link #sql_table_spec} and
 * {@link #sql_join_condition}.  Used by {@link #sql_from}. 
 */
sql_explicit_join
   :
      (
           // right outer joins (only 1 can be present)
           r:KW_RIGHT^ { #r.setType(RIGHT_OUTER_JOIN); }
           (KW_OUTER!)?
           KW_JOIN! sql_table_spec sql_join_condition
            
           // matches any number of left outer joins and inner joins (by
           // default or if the inner keyword is present, it is an inner
           // join, if the left keyword is present it is a left outer join)
         | (
              (KW_INNER! | l:KW_LEFT! (KW_OUTER!)? )? j:KW_JOIN^
              { #j.setType(#l == null ? INNER_JOIN : LEFT_OUTER_JOIN); }
              sql_table_spec sql_join_condition
           )+
      )
   ;

/**
 * Matches an <code>ON</code> followed by an {@link #expr} which is a SQL
 * join condition. Separated to build the tree properly.
 */
sql_join_condition
   :
      KW_ON^ expr
   ;
   
/**
 * Matches a {@link #record} and an optional following 
 * {@link #sql_correlation_name}. Used from {@link #sql_from}.
 */
sql_table_spec
   :
      r:record[true, false, false] 
      (
         options { generateAmbigWarnings = false; }
         :
         sql_correlation_name[#r.getText()] 
      )?
   ;
   
/**
 * Matches a correlation name ({@link #symbol}) in an {@link #sql_table_spec}
 * and creates a new buffer for the given table name.
 *
 * @param    rtext
 *           The table which is getting an alias.
 */
sql_correlation_name [String rtext]
   :
      c:symbol
      {
         defineBufferFromSymbol(#c, rtext, null, null, false);
      }
   ;
   
/**
 * Matches an embedded SQL <code>INTO</code> clause and the following list
 * of variables (uses {@link #lvalue} and {@link #sql_indicator_var}).
 */
sql_into
   :
      KW_INTO^
      lvalue (sql_indicator_var)? ( COMMA! lvalue (sql_indicator_var)? )*
   ;
   
/**
 * Matches the indicator variable specification in a {@link #sql_into}.
 * Uses {@link #lvalue}.
 */
sql_indicator_var
   :
      KW_INDICAT^ lvalue
   ;
   
/**
 * Matches the SQL <code>GROUP BY</code> specification and its following
 * list of comma separated {@link #expr}. Called by {@link #select_stmt}.
 */
sql_group_by
   :
      g:KW_GROUP^ { #g.setType(GROUP_BY); } KW_BY! expr (COMMA! expr)*
   ;
   
/**
 * Matches the SQL <code>HAVING</code> keyword and its following
 * {@link #expr}. Called by {@link #select_stmt}.
 */
sql_having
   :
      KW_HAVING^ expr
   ;
   
/**
 * Matches the SQL <code>ORDER BY</code> specification and its following
 * list of comma separated {@link #expr} and optional 
 * {@link #sql_direction_spec}. Called by {@link #select_stmt}.
 */
sql_order_by
   :
      o:KW_ORDER^ { #o.setType(ORDER_BY); } KW_BY!
      expr (sql_direction_spec)? ( COMMA! expr (sql_direction_spec)? )*
   ;
   
/**
 * Match the <code>ASCEND</code> or <code>DESCEND</code> keywords. Called
 * by {@link #sql_order_by}.
 */
sql_direction_spec
   :
      // we don't have separate keyword processing for KW_ASCEND, but we differentiate it here
      // to make the downstream processing more meaningful
      a:KW_ASC { #a.setType(KW_ASCEND); } | KW_DESCEND
   ;
   
/**
 * Matches the <code>CREATE</code> language statement and implements all
 * following options.  Calls the {@link #record} and {@link #using_id_clause}
 * rules. Called by the {@link #stmt_list} rule.
 */
create_stmt   
   :
      KW_CREATE^ record[true, false, false]
      (using_id_clause)? (KW_NO_ERROR)? DOT!
   ;
   
/**
 * Matches the <code>RELEASE</code> language statement.  Calls the 
 * {@link #record}.
 * <p>
 * <b>At this time, the record scope in the schema is not removed when this
 * rule is called.  This may be something that is needed in some obscure
 * case. To implement this, one would have to call 
 * {@link SymbolResolver#deleteSchemaScope} as well as notify the 
 * {@link #inner_block} rule so that it doesn't try to delete the same scope 
 * (this is non-trivial due to nesting). Watch for problems that might arise
 * from this choice to ignore the scope deletion.</b>
 * <p>
 * Called by the {@link #stmt_list} rule.
 */
release_stmt   
   :
      KW_RELEASE^ record[true, false, false] (KW_NO_ERROR)? DOT!
   ;   
   
/**
 * Matches a <code>RELEASE EXTERNAL</code> language statement, and the following required
 * {@link #expr}.  This also supports the use of NO-ERROR (undocumented 4GL feature).
 * <p>
 * Used by {@link #stmt_list}.
 */
release_external_stmt   
   :
      r:KW_RELEASE^  { #r.setType(RELEASE_EXTERNAL); }
      KW_EXTERN!
      (
         options { generateAmbigWarnings = false; }
         :
         KW_PROC!
      )?
      expr
      (KW_NO_ERROR)?
      DOT!
   ;   
   
/**
 * Matches the <code>USING</code> keyword and the following
 * <code>RECID or ROWID</code> construct as encoded by
 * {@link #recid_or_rowid}.  Called by {@link #create_stmt}.
 * <p>
 * The subtree is rooted by the keyword (which is the reason for using a 
 * separate rule for something so simple).
 */      
using_id_clause   
   :
      KW_USING^ recid_or_rowid
   ;
   
/**
 * Matches the <code>RECID or ROWID</code> keywords and the following
 * expression inside parenthesis.  Calls {@link #expr} and is called by
 * {@link #create_stmt}.
 * <p>
 * The subtree is rooted by the keyword (which is the reason for using a 
 * separate rule for something so simple).
 */   
recid_or_rowid   
   :
      (KW_RECID^ | KW_ROWID^) LPARENS! expr RPARENS!
   ;
   
/**
 * Matches the <code>DELETE</code> language statement and implements all
 * following options.  Calls the {@link #record} and {@link #validate} rules.
 * Called by the {@link #stmt_list} rule.
 */
delete_stmt
   :
      KW_DELETE^ r:record[true, false, false] (v:validate[null])? (KW_NO_ERROR)? DOT!
      {
         // if there is no explicit validation statement, create one using the schema-level
         // validation defined for this table, if any
         if (#v == null)
         {
            attachDeleteSchemaValidation(##, #r, sym);
         }
      }
   ;
   
/**
 * Matches an <code>ASSIGN</code> Progress 4GL language statement.
 * This is the statement that either moves data from screen buffers to an
 * lvalue OR it handles multiple regular assignments as a single block,  
 * eliminating any overhead associated with each assignment until the entire
 * set is complete (e.g. rebuilding an index).
 * <p>
 * It is undocumented, but the assign statement can be empty (just the
 * assign keyword followed by a dot).
 * <p>
 * The syntax is differentiated from a top level {@link #assignment} statement
 * by:
 * <ul>
 *    <li> the <code>ASSIGN</code> keyword as the first token
 *    <li> the requirement that there be an lvalue and assignment operator
 *         before any expression rather than being optional as in the 
 *         top level assignment rule
 *    <li> only one terminating <code>DOT</code>
 * </ul>
 * <p>
 * A rule reference called {@link #assign} is used to implement the        
 * assignment logic.  Note that this is <b>nearly identical</b> to the
 * <code>assignment</code> rule except for the non-mandatory lvalue and
 * assignment operator and the lack of the terminating <code>DOT</code>.
 * Due to these differences, the <code>assignment</code> rule could not be
 * reused here.
 * <p>
 * Uses a separate rule for core logic allows ANTLR's natural tree building to 
 * generate a proper result (the entire subtree created by <code>assign</code>
 * is added as a child of the <code>ASSIGN</code> keyword).
 * <p>
 * Uses {@link #record_spec}, {@link #assign} and {@link #simple_when_clause}.
 * <p>
 * This rule has inherent ambguity between the record form and the field
 * form.  This is resolved by precedence.  A test for whether the 1st token
 * is a table reference is made, if it matches a table reference, then the
 * record form is matched.
 * <p>
 * ANTLR reports bogus ambiguity in the use of the browse reference but
 * warnings have been disabled since this is not ambiguous.
 */
assign_stmt
   :
      (
         KW_ASSIGN^
         (
              {
                 isRecord(LT(1).getText()) &&
                 !isQualifiedFieldNameQuirk(LT(1), LT(2), LT(3))
              }?
              record_spec[null]
            | ( assign (simple_when_clause)? )+
         )?
         (KW_NO_ERROR)?
         DOT!
      )
   ;
   
/**
 * Implements the semantics of the assignment sub-expression in a Progress
 * <code>ASSIGN</code> language statement.
 * <p>
 * The syntax is differentiated from a top level {@link #assignment}
 * statement by:
 * <ul>
 *    <li> the requirement that there be an lvalue in the leftmost position
 *    <li> an expression cannot stand alone (without a prefixed lvalue and
 *         assignment operator)
 *    <li> no terminating <code>DOT</code>
 * </ul>
 * <p>
 * This is referenced from {@link #assign_stmt}, {@link #loop_incr} and
 * {@link #sub_assign_clause} rules.  It is used to implement the core
 * assignment logic.
 * <p>
 * A special mode exists that allows for an <code>lvalue</code> prefixed
 * with a frame or browse reference.
 * <p>
 * Creates both operands as children of the artifically created 
 * <code>ASSIGN</code> token type (which the <code>EQUALS</code> token is
 * replaced by).
 */   
assign
   :
      l:chained_object_members (eq:EQUALS^ { #eq.setType(ASSIGN); } expr)?
      {
         restoreAssignStyleStmt(#l);
         restoreEqualsOperator(#l, #eq);
      }
   ;
   
/**   
 * Matches all language statements that have assignment-like syntax. The
 * following statements are so simple that they don't have separate rules
 * but are instead treated as global variables that always exist (thus they
 * are not referenced here but are directly matched in the calling rule):
 * <p>
 * <pre>
 *   CURRENT-LANGUAGE
 *   FRAME-VALUE
 *   PROMSGS
 *   PROPATH
 * </pre>
 * <p>
 * These statements are referenced:
 * <p>
 * <ul>
 *    <li> {@link #current_value_stmt}
 *    <li> {@link #dynamic_current_value_stmt}
 *    <li> {@link #dynamic_new_stmt}
 *    <li> {@link #entry_stmt}
 *    <li> {@link #extent_stmt}
 *    <li> {@link #fix_codepage_stmt}
 *    <li> {@link #length_stmt}
 *    <li> {@link #overlay_stmt}
 *    <li> {@link #put_memptr_stmt}
 *    <li> {@link #raw_stmt}
 *    <li> {@link #set_byte_order_stmt}
 *    <li> {@link #set_pointer_value_stmt}
 *    <li> {@link #set_size_stmt}
 *    <li> {@link #substring_stmt}
 * </ul>
 * <p>
 * This rule is consolidated to allow these constructs to be accessed from
 * {@link #lvalue} which allows then to be in an {@link #assign_stmt} which
 * actually calls this rule via {@link #assign}.  Top level use of these as
 * a language statement is handled via the {@link #assignment} rule which
 * also handles the matching for the <code>EQUALS</code> and following
 * {@link #expr}.
 */
assign_type_syntax_stmt_list   
   :
      (
           current_value_stmt
         | dynamic_current_value_stmt
         | dynamic_new_stmt
         | entry_stmt
         | extent_stmt
         | fix_codepage_stmt
         | length_stmt
         | overlay_stmt
         | put_memptr_stmt
         | raw_stmt
         | set_byte_order_stmt
         | set_pointer_value_stmt
         | set_size_stmt
         | substring_stmt
      )
   ;
   
/**   
 * Matches the <code>SUBSTRING( , , , ) = expression</code> language 
 * statement.  This rule uses a rule reference to process expressions
 * for what appear to be comma-separated parameters (see {@link #expr}). 
 * Some tokens are suppressed and the <code>EQUALS</code> token is processed
 * in the calling rule.  The resulting AST should be easy to process.
 * <p>
 * This language statement has inherent ambiguity with the built-in function
 * that uses the same <code>SUBSTRING</code> keyword (and has a nearly
 * identical meaning but is only for use in expressions and it cannot be used
 * as an assignment).  Progress gives precedence to matching the language 
 * statement first and then the function form.  This parser implements this 
 * rule in the {@link #lvalue} portion of the top-level {@link #assignment}
 * rule.  Since this is matched before expressions (which is where a function
 * call is matched) the logic works properly and all ambiguity is resolved.
 * <p>
 * Used by {@link #assign_type_syntax_stmt_list}.
 */   
substring_stmt   
   :
      KW_SUBSTR^
      LPARENS! lvalue COMMA! expr (COMMA! expr (COMMA! expr)? )? RPARENS!
   ;
   
/**   
 * Matches the <code>COMPILE</code> language statement.
 * <p>
 * Calls {@link #external_name_spec} and {@link #compile_options_clause}
 * rules.
 * <p>
 * Used by {@link #stmt_list}.
 */
compile_stmt
   :
      KW_COMPILE^ external_name_spec (compile_options_clause)* DOT!
   ;
   
/**   
 * Matches the core set of options for the compile statement.
 * <p>
 * Calls:
 * <p>
 * <ul>
 *    <li> {@link #embedded_assign}
 *    <li> {@link #into_directory_clause}
 *    <li> {@link #external_name_spec}
 *    <li> {@link #optional_listing_clause}
 *    <li> {@link #expr}
 *    <li> {@link #languages_clause}
 * </ul>
 * <p>
 * Used by {@link #compile_stmt}.
 */
compile_options_clause   
   :
      (
           (KW_ATTR^ | KW_STRM_IO^ | KW_MIN_SZ^ | KW_GEN_MD5^)
           (embedded_assign)?
         | KW_NO_ATTR^     // undocumented but found in customer source
         | KW_SAVE^ (embedded_assign)? (into_directory_clause)?
         | KW_LISTING^ external_name_spec (optional_listing_clause)*
         | KW_XCODE^ expr
         | (KW_STR_XREF^ | KW_XREF^)
           external_name_spec
           (optional_listing_clause)?
         | languages_clause
         | (KW_DBG_LST^ | KW_PREPROC^)
           external_name_spec
         | KW_NO_ERROR^
         | KW_V6FRAME^ (embedded_assign)? (KW_USE_REV | KW_USE_UND)?
      )
   ;
   
/**   
 * Matches the <code>INTO</code> keyword followed by a directory name.
 * <p>
 * Calls {@link #external_name_spec}.
 * <p>
 * Used by {@link #compile_options_clause}.
 */
into_directory_clause   
   :
      KW_INTO^ external_name_spec 
   ;
   
/**   
 * Matches the <code>APPEND, PAGE-SIZE and PAGE-WIDTH</code> keywords
 * followed by an {@link #embedded_assign} or an {@link #expr}.
 * <p>
 * Used by {@link #compile_options_clause}.
 */
optional_listing_clause   
   :
        KW_APPEND^ (embedded_assign)?
      | KW_PAGE_SZ^ expr
      | KW_PAGE_WID^ expr
   ;
   
/**   
 * Matches the <code>LANGUAGES</code> keyword and a parenthesized, 
 * comma-separated list of {@link #language_list} entries.  This can be
 * followed by an {@link #embedded_assign} option.
 * <p>
 * Used by {@link #compile_options_clause}.
 */
languages_clause   
   :
      KW_LANGUAGE^
      LPARENS! language_list (COMMA! language_list)* RPARENS!
      (KW_TXT_SEG embedded_assign)?
   ;
   
/**   
 * Matches the a list of colon separated languages, all of which are made 
 * children of an artificial mode of type <code>LANGUAGES</code>.  Each
 * language is matched by a {@link #expr}.
 * <p>
 * Used by {@link #languages_clause}.
 */
language_list   
   :
      { astFactory.makeASTRoot(currentAST, #[LANGUAGES,"language list"]); }   
      expr (COLON! expr)*
      { if (##.getNumberOfChildren() == 0) ## = null; }
   ;
   
/**   
 * Simple rule to handle the common case where a {@link #filename}, a
 * <code>STRING</code> or a {@link #value} can be matched.
 * <p>
 * Called by {@link #compile_stmt}, {@link #into_directory_clause},
 * {@link #save_cache_stmt} and {@link #io_through_stmt}.
 */
external_name_spec
   :
      (
         options { generateAmbigWarnings = false; }
         :
           s:STRING { convertStringTo(#s, FILENAME); }
         | value
         | filename
      )
   ;
   
/**   
 * Matches the <code>ACCUMULATE</code> language statement.  Calls the {@link #expr} and
 * {@link #aggregate_phrase}.  An undocumented feature of the 4GL is also supported. This
 * allows an optional {@link #format_string} rule to follow the aggregate phrase.
 * <p>
 * If the accumulator type is <code>COUNT</code> or <code>TOTAL</code>, a
 * <code>VAR_INT</code> of name "count" or <code>VAR_DEC</code> of name
 * "total" will be created respectively. This is an undocumented feature
 * of the 4GL where the data can be read and even assigned directly as an
 * lvalue! Note that only <code>COUNT</code> and <code>TOTAL</code> have
 * this behavior. All other accumulator types (including all the 
 * <code>SUB-</code> variants) do NOT exhibit this behavior.
 * <p>
 * Used by {@link #stmt_list}.
 */   
accumulate_stmt
   :
      KW_ACCUM^ ( expr LPARENS! a:aggregate_phrase RPARENS! (format_string[false])? )+ DOT!
      {
         Aast child = #a.getChildAt(0);
         int  ttype = child.getType();
         
         if (ttype == KW_COUNT)
         {
            sym.addLocalVariable("count", VAR_INT, null);
         }
         else if (ttype == KW_TOTAL)
         {
            sym.addLocalVariable("total", VAR_DEC, null);
         }
      }
   ;
   
/**   
 * Matches the <code>STATUS</code> language statement.  Calls the
 * {@link #in_window_clause}.
 * <p>
 * Used by {@link #stmt_list}.
 */   
status_stmt   
   :
      KW_STATUS^
      (
           KW_DEFAULT (expr)?
         | KW_INPUT (KW_OFF | expr)?
      )
      (in_window_clause)?
      DOT!
   ;
   
/**   
 * Matches the <code>HIDE</code> language statement.  Calls the
 * {@link #lvalue} for stream references, {@link #widget_phrase} and 
 * {@link #in_window_clause} rules.
 * <p>
 * Used by {@link #stmt_list}.
 */   
hide_stmt   
   :
      KW_HIDE^ 
      (
           { LA(1) == KW_STREAM || LA(1) == KW_STRM_HND }? lvalue // used to be stream_reference
         | // empty alternative
      )
      (KW_MSG | KW_ALL | (widget_phrase)+)?
      (KW_NO_PAUSE)?
      (in_window_clause)?
      DOT!
   ;   
   
/**   
 * Matches the <code>PAUSE</code> language statement.  Calls the
 * {@link #in_window_clause} and {@link #simple_message_clause} rules.
 * <p>
 * Used by {@link #stmt_list}.
 */   
pause_stmt   
   :
      KW_PAUSE^ (expr)? (KW_B4_HIDE)? (simple_message_clause)?
      (in_window_clause)? DOT!
   ;
   
/**
 * Matches the <code>MESSAGE STRING or NO-MESSAGE</code> clause used in the
 * <code>PAUSE</code> language statement.  Called by {@link #pause_stmt}.
 */
simple_message_clause   
   :
      (KW_MSG^ STRING | KW_NO_MSG)
   ;
   
/**   
 * Matches the <code>DISABLE</code> language statement.  Calls the
 * {@link #all_clause}, {@link #field_clause} and {@link #frame_phrase}.
 * <p>
 * The <code>UNLESS-HIDDEN</code> unreserved keyword naturally conflicts with
 * the possible following lvalue.  This is not really ambiguous so warnings
 * have been disabled.
 * <p>
 * This supports an undocumented feature where there is no ALL keyword or field
 * list.  In such a case, this is thought to be a no-operation (at runtime).
 * <p>
 * Used by {@link #stmt_list}.
 */
disable_stmt
   :
      {
         Set<Aast> fields = new HashSet<>();
      }
      KW_DISABLE^
      (
         options { generateAmbigWarnings = false; }
         :
         KW_UNL_HID
      )?
      (
           all_clause
         | (
              fld:field_clause[false]
              {
                 int type = #fld.getType();
                 if (type > BEGIN_FIELDTYPES && type < END_FIELDTYPES)
                 {
                    fields.add(#fld);
                 }
              }
           )+
      )?
      (fp:frame_phrase[false, false])?
      DOT!
      {
         String frameName = getFrameName(#fp);

         if (!fields.isEmpty())
         {
            sym.addFrameFields(frameName, fields);
         }

         setUseDictExps(frameName, #fp, ##);
      }
   ;
   
/**
 * Matches the <code>ENABLE</code> language statement.  Calls the
 * {@link #all_clause}, {@link #field_clause} and {@link #frame_phrase}.
 * <p>
 * The <code>UNLESS-HIDDEN</code> unreserved keyword naturally conflicts with
 * the possible following lvalue.  This is not really ambiguous so warnings
 * have been disabled.
 * <p>
 * This supports an undocumented feature where there is no ALL keyword or field
 * list.  In such a case, this is thought to be a no-operation (at runtime).
 * <p>
 * Used by {@link #stmt_list}.
 */
enable_stmt
   :
      {
         Set<Aast> fields = new HashSet<>();
      }
      KW_ENABLE^
      (
         options { generateAmbigWarnings = false; }
         :
         KW_UNL_HID
      )?
      (
           a:all_clause
         | (
                space_or_skip
              | text_clause 
              | constant_form_item
              | fld:field_clause[true]
                {
                   int type = #fld.getType();
                   if (type > BEGIN_FIELDTYPES && type < END_FIELDTYPES)
                   {
                      fields.add(#fld);
                   }
                }
           )+
      )?
      (in_window_clause)?
      (fp:frame_phrase[false, false])?
      DOT!
      {
         String frameName = getFrameName(#fp);

         if (!fields.isEmpty())
         {
            sym.addFrameFields(frameName, fields);
         }

         if (#a != null)
         {
            String frame = getFrameName(#fp);
            fields = sym.getFrameFields(frame);
            if (fields != null)
            {
               Set<String> exceptFields = new HashSet<>();
               
               if (#a.downPath("KW_EXCEPT"))
               {
                  Aast ref = #a.getImmediateChild(CONTENT_ARRAY, null);
                  Iterator<Aast> iter = ref.iterator();
                  while (iter.hasNext())
                  {
                     ref = iter.next();
                     int type = ref.getType();
                     if (type > BEGIN_FIELDTYPES && type < END_FIELDTYPES)
                     {
                        // EXCEPT fields must not influence HELP
                        ref.removeAnnotation("help");

                        String sname = (String) ref.getAnnotation("schemaname");
                        String bname = (String) ref.getAnnotation("bufname");
                        
                        exceptFields.add(sname + "$" + bname);
                     }
                  }
               }

               ASTPair allPair = new ASTPair();
               astFactory.makeASTRoot(allPair, #a);
               for (Aast fld : fields)
               {
                  String sname = (String) fld.getAnnotation("schemaname");
                  String bname = (String) fld.getAnnotation("bufname");
                  
                  if (exceptFields.contains(sname + "$" + bname))
                  {
                     // do not pull fields which are referenced by EXCEPT
                     continue;
                  }
                  
                  Aast wid = sym.createField(sname);
                  wid.putAnnotation("enable_all", true);
                  wid.setHidden(true);
                  
                  // some annotations are specific to this field definition...
                  wid.putAnnotation("bufname", bname);
                  wid.putAnnotation("name", (String) fld.getAnnotation("name"));

                  Aast fmt = new ProgressAst();
                  List<Aast> asts = attachAssignSchemaValidation(fmt, wid, sym);
                  if (asts != null && !asts.isEmpty())
                  {
                     #a.addChild(wid);
                     #a.addChild(fmt);
                     wid.putAnnotation("enable_all", true);
                     fmt.putAnnotation("enable_all", true);
                     fmt.setHidden(true);
                     fmt.setType(FORMAT_PHRASE);
                  }
               }
            }
         }

         // do this last, as we need the fields for ENABLE ALL processed
         setUseDictExps(frameName, #fp, ##);
      }
   ;
   
/**
 * Matches the <code>PUT SCREEN</code> language statement.  Calls the
 * {@link #simple_color_clause}, {@link #ui_stuff} and {@link #expr}.
 * <p>
 * The <code>ui_stuff</code> rule uses unreserved keywords that naturally
 * conflict with the possible following expression.  This is not really 
 * ambiguous so warnings have been disabled.
 * <p>
 * Used by {@link #stmt_list}.
 */   
put_screen_stmt   
   :
      p:KW_PUT^  { #p.setType(PUT_SCREEN); }
      KW_SCREEN!
      (
         options { generateAmbigWarnings = false; }
         :
           (KW_ATTR | KW_NO_ATTR)
         | simple_color_clause
         | { LA(1) == KW_ROW || LA(1) == KW_COL || LA(1) == KW_COLUMNS }?
           ui_stuff
         | expr
      )*
      DOT!
   ;
   
/**   
 * Matches the <code>PAGE</code> language statement.  Optionally calls the
 * {@link #lvalue} rule for stream references.
 * <p>
 * Used by {@link #stmt_list}.
 */   
page_stmt   
   :
      KW_PAGE^
      (
           { LA(1) == KW_STREAM || LA(1) == KW_STRM_HND }? lvalue // used to be stream_reference
         | // empty alternative
      )
      DOT!
   ;
   
/**
 * Matches the <code>READKEY</code> language statement.  Optionally calls the
 * {@link #lvalue} (for stream references) and {@link #pause_clause} rules.
 * <p>
 * Used by {@link #stmt_list}.
 */
readkey_stmt
   :
      KW_READKEY^
      (
           { LA(1) == KW_STREAM || LA(1) == KW_STRM_HND }? lvalue // used to be stream_reference
         | // empty alternative
      )      
      (pause_clause)?
      DOT!
   ;
   
/**
 * Matches the <code>EXPORT</code> language statement.
 * <p>
 * Calls:
 * <p>
 * <ul>
 *    <li> {@link #lvalue} (for stream references)
 *    <li> {@link #delimiter_clause}
 *    <li> {@link #record_spec}
 *    <li> {@link #space_or_skip} (undocumented but silently ignored in
 *         Progress)
 *    <li> {@link #export_field}
 * </ul>
 * <p>
 * Used by {@link #stmt_list}.
 * <p>
 * In an undocumented twist, it is perfectly valid to omit any record or
 * field definition in this statement, though it seems to be a NOP in such
 * a case.
 * <p>
 * Bogus ANTLR ambiguity warnings have been disabled.
 */
export_stmt
   :
      KW_EXPORT^
      (
           { LA(1) == KW_STREAM || LA(1) == KW_STRM_HND }? lvalue // used to be stream_reference
         | // empty alternative
      )
      (delimiter_clause)?
      (
           // we must disambiguate a record from an expression (records take 
           // precedence) by trying to do a lookup, if it fails we don't
           // match this alternative
           { isRecord(LT(1).getText()) }?
           record_spec[null]
           
         | (
                // Progress silently tolerates (and ignores) any space or skip
                // constructs in the list, so we match but drop the result (this
                // is an undocumented "feature")
                space_or_skip!
              
                // common case of an expression
              | export_field
           )*
      )
      (KW_NO_LOBS)?
      DOT!
   ;
   
/**
 * Implements proper tree building for the most common field specification
 * in the {@link #export_stmt}.  Roots an {@link #expr} at a node of type
 * <code>EXPORT_FIELD</code>.
 * <p>
 * Progress has an undocumented feature where format and aggregate phrases
 * can be attached to any expression but although a {@link #format_phrase}
 * and {@link #aggregate_phrase} are matched, they are dropped from the
 * resulting tree since in Progress they are silently ignored.
 */
export_field
   :
      {
         int numFP = 0;
         astFactory.makeASTRoot(currentAST, #[EXPORT_FIELD,"output field"]);
      }
      (
         expr
         
         // may be followed by complete garbage that is used in display stmts
         // but is totally ignored in an export stmt
         // WARNING: since format_phrase may be empty and this rule is an
         //          optional loop, the prediction logic for format_phrase
         //          includes all possible following constructs (e.g. DOT)
         //          even though these can't be matched in a format phrase;
         //          this leads to an infinite loop which we avoid by not
         //          allowing more than 2 matches to a format phrase (YES,
         //          there can be 2 format phrases, one before the aggregate
         //          phrase and one after --> just like in the 4GL)
         (
            options { generateAmbigWarnings = false; }
            :
              LPARENS! aggregate_phrase! RPARENS!
            | { numFP < 2 }?
              format_phrase[null, null, false, false]!   { numFP++; }
         )*
      )
   ;
   
/**
 * Matches the <code>IMPORT</code> language statement.
 * <p>
 * Calls:
 * <p>
 * <ul>
 *    <li> {@link #lvalue} (for stream references)
 *    <li> {@link #unformatted_clause}
 *    <li> {@link #delimiter_clause}
 *    <li> {@link #record_spec}
 *    <li> {@link #import_field} (undocumented feature: this can be zero or more instead of
 *         one or more as the 4GL docs indicate)
 * </ul>
 * <p>
 * Used by {@link #stmt_list}.
 */   
import_stmt
   :
      KW_IMPORT^
      (
           { LA(1) == KW_STREAM || LA(1) == KW_STRM_HND }? lvalue // used to be stream_reference
         | // empty alternative
      )
      (
           unformatted_clause
         | (delimiter_clause)?
           (
                // we must disambiguate a record from an lvalue (records take 
                // precedence) by trying to do a lookup, if it fails we don't
                // match this alternative
                { isRecord(LT(1).getText()) }?
                record_spec[null]
              | (CARET | import_field)*
           )
      )
      (KW_NO_LOBS)?
      (KW_NO_ERROR)?
      DOT!
   ;
     
/**      
 * Implements proper tree building for the most common field specification
 * in the {@link #import_stmt}.  Roots an {@link #lvalue} at a node of type
 * <code>IMPORT_FIELD</code>.
 */
import_field
   :
      { astFactory.makeASTRoot(currentAST, #[IMPORT_FIELD,"input field"]); }
      (lvalue)
   ;
   
/**
  * Matches the <code>UNFORMATTED</code> keyword and the following {@link #record} or
  * {@link #lvalue}.  The record case is an undocumented feature of the 4GL.
  * <p>
  * Another undocumented feature (found in customer source code) is that a CARET can be inserted
  * before the <code>lvalue</code>.  It generates a warning message but still runs un the 4GL.
  * The meaning is questionable (it is probably ignored) since the idea is to read the entire
  * line into the lvalue and the caret means "ignore the content in this position", but in
  * "whole line" more it doesn't make sense.
  * <p>
  * Used by {@link #import_stmt}.
  */
unformatted_clause
   :
      KW_UNFORMAT^
      (
           // we must disambiguate a record from an lvalue (records take 
           // precedence) by trying to do a lookup, if it fails we don't
           // match this alternative
           { isRecord(LT(1).getText()) }?
           record[true, false, false]
           
            // the 4GL allows matching multiple lvalues and the CARAT (undocumented) but the
            // presence of more than one of these items generates this compile-time warning:
            // ** WARNING -- IMPORT UNFORMATTED statement references more than one field. (4958)
            // the CARAT by itself or in the first position ignores the line
            // the CARAT and/or any lvalues that are not in first position are ignored and only
            // any lvalue in the first position will be used
         | (
               CARET  // undocumented
             | lvalue
           )+
      )
   ;
   
/**
  * Matches the <code>DELIMITER</code> keyword and the following single
  * quoted character (a <code>STRING</code>).
  * <p>
  * Used by {@link #export_stmt} and {@link #import_stmt} rules.
  */
delimiter_clause
   :
      KW_DELIMIT^ STRING
   ;
   
/**
 * Matches the <code>SEEK</code> language statement.
 * <p>
 * Calls {@link #lvalue} (for stream references) and {@link #expr} rules.
 * <p>
 * Used by {@link #stmt_list}.
 */
seek_stmt
   :
      KW_SEEK^
      (KW_INPUT | KW_OUTPUT | { LA(1) == KW_STREAM || LA(1) == KW_STRM_HND }? lvalue )
      KW_TO! (KW_END | expr)
      DOT!
   ;
   
/**
 * Matches the <code>RAW-TRANSFER</code> language statement.
 * <p>
 * Calls {@link #buffer_or_field_spec}.
 * <p>
 * Used by {@link #stmt_list}.
 */   
raw_transfer_stmt
   :
      KW_RAW_TRAN^ buffer_or_field_spec KW_TO! buffer_or_field_spec
      (KW_NO_ERROR)?
      DOT!
   ;
   
/**
 * Matches the <code>COPY-LOB</code> language statement.
 * <p>
 * Calls {@link #from_lob_source}, {@link #to_lob_target} and {@link #codepage_convert_phrase}.
 * <p>
 * Used by {@link #stmt_list}.
 */   
copy_lob_stmt
   :
      KW_CPY_LOB^ from_lob_source to_lob_target 
      (codepage_convert_phrase)?
      (KW_NO_ERROR)?
      DOT!
   ;
   
/**
 * Match the <code>FROM</code> specification of {@link #copy_lob_stmt} using
 * {@link #lob_spec}, {@link #starting_clause} and {@link #for_clause}. The
 * result is always rooted at a <code>KW_FROM</code>, even though that keyword
 * is optional in the 4GL (a token will be manufactured if needed).
 */
from_lob_source
   :
      {
         if (LA(1) != KW_FROM)
         {
            astFactory.makeASTRoot(currentAST, #[KW_FROM,"from"]);
         }
      }
      (KW_FROM^)? lob_spec[false] (starting_clause | for_clause)*
   ;
   
/**
 * Match the <code>TO</code> specification of {@link #copy_lob_stmt} using
 * {@link #lob_spec}.  The result is always rooted at a <code>KW_TO</code>.
 */
to_lob_target
   :
      KW_TO^ lob_spec[true] 
   ;
   
/**
 * Match the specification of a single source or target LOB with is either an {@link #expr}
 * (optionally prepended with the <code>KW_OBJECT</code>) or a <code>KW_FILE</code> followed by
 *  an {@link #expr}. If <code>to</code> is <code>true</code> then {@link #overlay_clause} or
 * <code>KW_APPEND</code> can be optionally matched at the end. Used by {@link #from_lob_source}
 * and {@link #to_lob_target}.
 */
lob_spec [boolean to]
   :
      (
           { LA(1) != KW_FILE }?
           (
              options { generateAmbigWarnings = false; }
              :
              KW_OBJECT!
           )?
           expr ( { to }? overlay_clause )?
           
         | KW_FILE^ expr ( { to }? KW_APPEND )?
      )
   ;
   
/**
 * Matches the <code>OVERLAY</code> keyword followed by <code>AT</code>
 * (which is dropped from the tree) and an {@link #expr}. <code>KW_TRIM</code>
 * is optionally matched at the end. Used by {@link #lob_spec}.
 */
overlay_clause
   :
      KW_OVERLAY^ KW_AT! expr (KW_TRIM)?
   ;
   
/**
 * Matches the <code>STARTING</code> keyword followed by <code>AT</code>
 * (which is dropped from the tree) and an {@link #expr}. Used by
 * {@link #from_lob_source}.
 */
starting_clause
   :
      KW_STARTING^ KW_AT! expr
   ;
   
/**
 * Matches the <code>FOR</code> keyword followed by an {@link #expr}. Used by
 * {@link #from_lob_source}.
 */
for_clause
   :
      KW_FOR^ expr
   ;
   
/**   
 * Matches the optional prefix <code>BUFFER or FIELD</code> keyword and
 * the required following buffer ({@link #record}) or field {@link #lvalue}.
 * These two alternatives are disambiguated by semantic predicate.
 * <p>
 * Bogus ANTLR ambiguity warnings have been disabled.
 * <p>
 * Used by {@link #raw_transfer_stmt}.
 */   
buffer_or_field_spec
   :
      (
         options { generateAmbigWarnings = false; }
         :
         KW_BUFFER! | KW_FIELD!
      )?
      (
           // we must disambiguate a record from a field (records take 
           // precedence) by trying to do a lookup, if it fails we don't
           // match this alternative
           { isRecord(LT(1).getText()) }?
           record[true, false, false]
         | lvalue
      )
   ;
   
/**   
 * Matches the <code>FIX-CODEPAGE</code> language statement.  Calls the
 * {@link #lvalue} rule.
 * <p>
 * Used by {@link #assign_type_syntax_stmt_list}.
 */   
fix_codepage_stmt   
   :   
      KW_FIX_CP^ LPARENS! lvalue RPARENS!
   ;
   
/**   
 * Matches the <code>ENTRY</code> language statement.  Calls {@link #expr},
 * and {@link #lvalue} rules.
 * <p>
 * Used by {@link #assign_type_syntax_stmt_list}.
 */   
entry_stmt   
   :   
      KW_ENTRY^ LPARENS! expr COMMA! lvalue (COMMA! expr)? RPARENS!
   ;
   
/**   
 * Matches the <code>OVERLAY</code> language statement.  Calls {@link #expr},
 * and {@link #lvalue} rules.
 * <p>
 * Used by {@link #assign_type_syntax_stmt_list}.
 */   
overlay_stmt   
   :   
      KW_OVERLAY^
      LPARENS! lvalue COMMA! expr (COMMA! expr (COMMA! expr)? )? RPARENS!
   ;   
   
/**   
 * Matches the <code>EXTENT</code> language statement. Calls {@link #lvalue}.
 * <p>
 * Used by {@link #assign_type_syntax_stmt_list}.
 */   
extent_stmt   
   :   
      KW_EXTENT^ LPARENS! lvalue RPARENS!
   ;   
   
/**   
 * Matches the <code>LENGTH</code> language statement. Calls {@link #lvalue}.
 * <p>
 * It turns out is NOT an Oracle-only statement, contrary to the
 * documentation.
 * <p>
 * Used by {@link #assign_type_syntax_stmt_list}.
 */   
length_stmt   
   :   
      KW_LENGTH^ LPARENS! lvalue RPARENS!
   ;   

/**   
 * Matches the <code>CURRENT-VALUE</code> language statement.  Calls 
 * {@link #sequence_ref} and {@link #database_name} rules.
 * <p>
 * Used by {@link #assign_type_syntax_stmt_list}.
 */   
current_value_stmt
   :
      KW_CUR_VAL^ 
      LPARENS! sequence_ref (COMMA! database_name[false])? RPARENS!
   ;

/**   
 * Matches the <code>DYNAMIC-CURRENT-VALUE</code> language statement.  Calls 
 * {@link #sequence_ref} and {@link #database_name} rules.
 * <p>
 * Used by {@link #assign_type_syntax_stmt_list}.
 */   
dynamic_current_value_stmt
   :
      KW_DYN_CURV^
      LPARENS! expr COMMA! expr RPARENS!
   ;

/**
 * Matches the <code>current-value</code>, <code>dynamic-current-value</code>,
 * <code>dynamic-next-value</code> and <code>next-value()</code> function
 * calls. The root node returned will be a <code>FUNC_INT</code> or a
 * <code>FUNC_INT64</code> and there will be 1 or 2 children which are
 * <code>STRING</code> nodes.
 * <p>
 * The exit action will save any original token type that rewriting may have
 * erased.  This will allow future processing to access this as an annotation
 * named 'oldtype'.
 * <p>
 * Built-in functions have a language keyword as the function name. This 
 * means that built-ins can support abbreviations.  It also means that some
 * built-in functions cannot be hidden by a user-defined function, if the
 * associated keyword is reserved. If the keyword is not reserved, then a
 * built-in function is hidden in the namespace by a user-defined function
 * of the same name.  This method's token type rewriting is dependent upon
 * this logic being implemented in the <code>SymbolResolver</code> class.
 * See {@link SymbolResolver#lookupFunction}.
 * <p>
 * Note that in the resulting AST, the parenthesis and commas used to define
 * the parameter list are discarded.  In other words, they do not appear in
 * the AST as nodes.  Instead, each parameter is created as a separate child
 * node of the function type token.  All parameters are maintained as child
 * nodes in the exact order of the function call itself, so this is
 * essentially the signature of the function.  It it also the proper AST
 * definition to enable simple evaluation of functions with any arbitrary
 * signature.
 */
sequence_funcs
   :
      (
           KW_CUR_VAL^
         | KW_NEXT_VAL^
      ) 
      LPARENS! sequence_ref (COMMA! database_name[false])? RPARENS!
      {
         // save the original token type, then rewrite the type
         ##.putAnnotation("oldtype", new Long(##.getType()));
         ##.setType(FUNC_INT64);
            
         // write our function-specific annotations
         sym.annotateFunction(##.getText(), ##, false);
      }
   ;

/**
 * Sequence references are unquoted text that must match a sequence name
 * defined in an accessible schema.  This rule matches a {@link #symbol} and
 * converts that token into a properly formatted <code>STRING</code>.
 */
sequence_ref
   :
      s:symbol
      {
         // convert unquoted sequence name to a quoted string
         #s.setType(STRING);
         #s.setText("\"" + #s.getText() + "\"");
      }
   ;
   
/**   
 * Matches an <code>EQUALS</code> token (the equals sign) and a following
 * {@link #expr}. The token type of the <code>EQUALS</code> is changed to
 * <code>ASSIGN</code> and the construct is rooted on that token.
 * <p>
 * Used by {@link #entry_stmt} and {@link #compile_stmt}.
 */
embedded_assign   
   :
      eq:EQUALS^  { #eq.setType(ASSIGN); }
      expr
   ;
   
/**   
 * Matches the optional <code>PAUSE</code> keyword prefix and the required
 * following numeric {@link #expr}.
 * <p>
 * Used by {@link #readkey_stmt} and {@link #wait_for_stmt} rules.
 * <p>
 * This is a separate rule to centralize logic and build the AST properly.
 */   
pause_clause   
   :
      KW_PAUSE^ expr
   ;
   
/**   
 * Matches the <code>VIEW</code> language statement.  Optionally calls the
 * {@link #lvalue} (for stream references), {@link #widget_phrase} and
 * {@link #in_window_clause} rules.
 * <p>
 * Progress has an undocumented feature where more than 1 widget phrase can
 * be matched instead of the documented single widget phrase. This rule
 * supports that undocumented mode.
 * <p>
 * Used by {@link #stmt_list}.
 */   
view_stmt   
   :
      KW_VIEW^
      (
           { LA(1) == KW_STREAM || LA(1) == KW_STRM_HND }? lvalue // used to be stream_reference
         | // empty alternative
      )
      (widget_phrase)*
      (in_window_clause)?
      DOT!
   ;
      
/**   
 * Matches the <code>UNDERLINE</code> language statement.  Calls the 
 * {@link #lvalue} (for an optional stream reference), {@link #lvalue_list}
 * for the list of widgets and {@link #frame_phrase} rules.
 * <p>
 * As an undocumented feature of Progress, the list of widgets can be
 * missing! In this case, the language statement is essentially a NOP.
 * <p>
 * Used by {@link #stmt_list}.
 */
underline_stmt
   :
      {
         Set<Aast> fields = new HashSet<>();
      }
      KW_UNDERLIN^
      (
           { LA(1) == KW_STREAM || LA(1) == KW_STRM_HND }? lvalue // used to be stream_reference
         | // empty alternative
      )
      (lvalue_list[fields])?
      (fp:frame_phrase[false, false])?
      DOT!
      {
         String frame = getFrameName(#fp);

         if (!fields.isEmpty())
         {
            sym.addFrameFields(frame, fields);
         }

         setUseDictExps(frame, #fp, ##);
      }
   ;
   
/**
 * Aggregates (in a tree building sense) a simple list of all fields under a specific language
 * statement.  Calls {@link #lvalue} and roots everything at a node of type
 * <code>CONTENT_ARRAY</code>. An undocumented feature of Progress is that each lvalue can
 * optionally be followed by a format phrase (using the {@link #format_phrase} rule).
 *
 * @param   fields
 *          Set into which all database field references should be stored.
 */
lvalue_list [Set<Aast> fields]
   :
      { astFactory.makeASTRoot(currentAST, #[CONTENT_ARRAY, ""]); }
      (
         l:lvalue
         (
            options { generateAmbigWarnings = false; }
            :
            format_phrase[null, null, false, false]
         )?
         {
            int type = #l.getType();
            if (type > BEGIN_FIELDTYPES && type < END_FIELDTYPES)
            {
               fields.add(#l);
            }
         }
      )+
   ;
   
/**
 * Matches the <code>UP</code> or <code>DOWN</code> language statements.  
 * Optionally calls the {@link #lvalue} (for stream references) and may
 * have {@link #expr} and {@link #frame_phrase} "options".
 * <p>
 * Options can be specified in any order.
 * <p>
 * Used by {@link #stmt_list}.
 */   
up_or_down_stmt   
   :
      (KW_UP^ | KW_DOWN^) 
      (
         options { generateAmbigWarnings = false; }
         :
           { LA(1) == KW_STREAM || LA(1) == KW_STRM_HND }?
           lvalue // used to be stream_reference
         | expr
         | frame_phrase[false, false]
      )*
      DOT!
   ;
      
/**   
 * Matches the <code>VALIDATE</code> language statement.  Calls the
 * {@link #record} rule.
 * <p>
 * Used by {@link #stmt_list}.
 */   
validate_stmt
   :
      KW_VALIDATE^ record[true, false, false] (KW_NO_ERROR)? DOT!
   ;
   
/**   
 * Matches the <code>TRIGGERS</code> keyword and the following trigger phrase
 * construct.  This is used in the {@link #def_var_stmt},
 * {@link #def_button_stmt} and {@link #menu_item_phrase} rules.
 * <p>
 * Calls the {@link #simple_on_stmt} rule to implement the actual event
 * matching and trigger block.  Uses {@link #end_stmt} to delimit the 
 * block.  This rule does NOT eat the trailing <code>DOT</code> since it is
 * called inline from other language statements.
 * <p>
 * Though it is not documented, it is possible to use either a 
 * <code>DOT or COLON</code> after the <code>TRIGGERS</code> keyword. This
 * is supported.  Note that although in many locations in Progress, 2
 * nearly identical keywords are both matched as the documented keyword
 * (e.g. {@link #buttons_clause} or {@link #field_list}),
 * the <code>TRIGGER</code> keyword does NOT work in Progress, thus we
 * provide no support for it here.
 * <p>
 * Although the 4GL documentation shows that at least one {@link #simple_on_stmt} is required
 * to be present, customer code has shown that it is optional. This undocumented feature is
 * supported.
 * <p>
 * Bogus ANTLR ambiguity warnings have been suppressed.
 */   
trigger_phrase   
   :
      KW_TRIGGERS^ (DOT! | COLON!) (simple_on_stmt)* end_stmt[false]
   ;
   
/**   
 * Matches a greatly simplified form of the <code>ON</code> statement which
 * is only seen in a {@link #trigger_phrase} Progress construct.  See
 * {@link #on_stmt} for details on the more complicated form.  This form
 * is not compatible with the language statement itself which is why
 * this is a separate rule.  This separation also allows the AST to be
 * built in a more natural manner.
 * <p>
 * This is the location in which we add and remove scopes to the symbol
 * dictionary for trigger blocks.
 * <p>
 * Bogus ANTLR ambiguity warnings have been suppressed.
 */
simple_on_stmt   
   :
      {
         boolean oldBufScope = setupTriggerScope();
      }
      KW_ON^ comma_event_list
      (
         options { generateAmbigWarnings = false; }
         :
         KW_ANYWHERE
      )?
      // we don't think the trigger block can be labeled, but if it can (and the label can be
      // malformed, then checkAndFixLabelWithReservedKeywordPrefix(1) will have to be called here
      // to properly engage the prediction logic for single_block
      (
           persistent_trigger_procedure 
         | single_block[true, true] // this is the "trigger block"
      )   
      {
         clearTriggerScope(oldBufScope);
      }
   ;
   
/**
 * Match the <code>TRANSACTION-MODE AUTOMATIC</code> statement. Used by
 * {@link #stmt_list}.
 */
transaction_mode_auto_stmt
   :
      t:KW_TRAN_MOD^  { #t.setType(TRANSACTION_MODE_AUTO); }
      KW_AUTOMATC! (KW_CHAINED)? DOT!
   ;
   
/**
 * Support the <code>TRIGGER PROCEDURE</code> language statement. All forms
 * (<code>CREATE</code>, <code>DELETE</code>, <code>FIND</code>, <code>WRITE</code> and 
 * <code>ASSIGN</code>, and the replication as well) are supported. Uses the
 * following rules to parse:
 * <p>
 * <ul>
 *    <li> {@link #record}
 *    <li> {@link #write_referencing_phrase}
 *    <li> {@link #lvalue}
 *    <li> {@link #assign_trigger_old_var_def}
 * </ul>
 * <p>
 * Called by {@link #stmt_list}.
 */
trigger_procedure_stmt
   :
      {
         Aast likeRef = null;
      }
      t:KW_TRIGGER^ { #t.setType(TRIGGER_PROCEDURE); }
      KW_PROC! KW_FOR!
      (
           // we have not confirmed that the force flag of record[] should be false
           (KW_CREATE | KW_DELETE | KW_FIND | KW_REPL_CRE | KW_REPL_DEL)
               KW_OF! record[true, false, false]
           
           // we have not confirmed that the force flag of record[] should be false
         | (KW_WRITE | KW_REPL_WRI) KW_OF! r:record[true, false, false]
            (
               // optional new reference clause
               options { generateAmbigWarnings = false; }
               :
               write_referencing_phrase[#r]
            )?
            (
               options { generateAmbigWarnings = false; }
               :
               // optional old reference clause
               write_referencing_phrase[#r]
            )?
           
         | KW_ASSIGN
           (
                KW_OF! lv:lvalue
              | vd:assign_trigger_new_var_def
           )
           {
              // the vd case is not exactly the "variable definition node", but it will
              // suffice for our downstream needs
              likeRef = (#lv != null) ? #lv : #vd.getImmediateChild(SYMBOL, null);
           }
           (
              options { generateAmbigWarnings = false; }
              :
              assign_trigger_old_var_def[likeRef]
           )?
      )
   ;
   
/**
 * Matches the <code>DISABLE TRIGGERS</code> statement, the qualifier for
 * <code>DUMP or LOAD</code> and the associated {@link #record}. Used by
 * {@link #stmt_list}.
 */
disable_triggers_stmt
   :
      d:KW_DISABLE^  { #d.setType(DISABLE_TRIGGERS); }
      KW_TRIGGERS!
      KW_FOR! (KW_DUMP | KW_LOAD)
      KW_OF! record[false, false, false]
      (KW_ALLW_REP)?
      DOT!
   ;
   
/**
 * Handles parsing of the NEW VALUE from the ASSIGN schema trigger procedure. The VALUE
 * token is optional and is dropped anyway.
 *
 * Called by {@link #trigger_procedure_stmt}.
 */
assign_trigger_new_var_def
   :
      {
         String symname = null;
      }
      KW_NEW^ (KW_VALUE!)? symname = sy:regular_parm[null, null]
      {
         sym.setVariableOptions(#sy.getText(), ##);
         sym.annotateVariableOptions(#sy.getText(), ##, true);
      }
   ;
   
/**
 * Process the OLD variable definition option for an assignment trigger procedure.
 * <p>
 * Matches only <code>OLD</code> branch  followed by an optional <code>VALUE</code> keyword 
 * (dropped) and then the variable definition which is handled by {@link #regular_parm}.
 * <p>
 * If the <code>likeRef</code> parameter is non-null AND this is a keyword <code>OLD</code>
 * where there is no explicit <code>AS</code> or <code>LIKE</code>, then this code will
 * implement an implicit <code>LIKE</code> clause.  This is an undocumented feature that
 * was found in customer code.
 * <p>
 * Called by {@link #trigger_procedure_stmt}, {@link #on_stmt}.
 *
 * @param    likeRef
 *           The variable or field definition to use for an implicit LIKE clause. This may be
 *           <code>null</code>.
 */
assign_trigger_old_var_def [Aast likeRef]
   :
      {
         String symname = null;
      }
      
      (
         (KW_OLD^) (KW_VALUE!)? symname = regular_parm[null, likeRef]
        
         {
            // ##.putAnnotation("define_buffer", new Boolean(false));
            
            // pass the AST with all options to the symbol resolver to set
            // those variable options accordingly
            sym.setVariableOptions(symname, ##);
            
            // write any non-standard options into annotations
            sym.annotateVariableOptions(symname, ##, true);
         }
      )?
   ;
   
/**   
 * Matches the <code>ON</code> language statement (which defines a trigger
 * block) and implements all possible variants.
 * <p>
 * Used by {@link #stmt_list} and {@link #trigger_phrase}.
 * <p>
 * The use of the <code>event</code> rule in the leftmost token position
 * is naturally ambiguous with the use of the <code>comma_event_list</code>  
 * since this also has an <code>event</code> in the leftmost position.
 * In this case, ANTLR does not detect any ambiguity since the first
 * alternative (<code>event</code>) must be followed by a match with the
 * very specific <code>key_function</code> rule, so the k = 2 is enough to
 * disambiguate this situation.
 * <p>
 * Uses:
 * <p>
 * <ul>
 *    <li> {@link #db_event}
 *    <li> {@link #record}
 *    <li> {@link #lvalue}
 *    <li> {@link #write_referencing_phrase}
 *    <li> {@link #assign_trigger_old_var_def}
 *    <li> {@link #event}
 *    <li> {@link #comma_event_list}
 *    <li> {@link #key_function}
 *    <li> {@link #comma_widget_list}
 *    <li> {@link #comma_events_of_widgets_clause}
 *    <li> {@link #persistent_trigger_procedure}
 *    <li> {@link #single_block}  (this is the actual implementation for
 *         the Progress trigger block)
 * </ul>
 * <p>
 * This is the location in which we add and remove scopes to the symbol
 * dictionary for trigger blocks.
 * <p>
 * Bogus ambiguity warnings by ANTLR have been disabled.  One of the warnings
 * started when the {@link #record_funcs} rule was added to 
 * {@link #primary_expr}, as it creates apparent ambiguity with the
 * <code>referencing_phrase</code>.  The ambiguity is not real.
 */   
on_stmt
   :
      {
         boolean bypass = false;
         boolean buf    = false;
         Aast    ref    = null;
         Aast    dbEvRef = null;
         Aast    recRef = null;
         Aast    fldRef = null;
         Aast    evtRef = null;
         int     eventType;

         boolean oldBufScope = setupTriggerScope();
      }
      
      KW_ON^

      // here is the event specification for what should cause the trigger to fire
      (
         options { generateAmbigWarnings = false; }
         :
           // dbevent names can be used with widgets instead of buffers (customer code shows
           // that at least DELETE can be used this way); so we must look ahead and only match
           // here if this is really a database case
           { sym.isTable(LT(3).getText()) || sym.isField(LT(3).getText()) }?
           (
              // database triggers will inline the db_event function
              evtRef:db_event KW_OF!
              (
                 { eventType = #evtRef.getChildAt(0).getType(); }
                 
                 (
                    { eventType == KW_CREATE || eventType == KW_DELETE || eventType == KW_FIND }?
                    ( record[true, false, false] )
                    
                    | { eventType == KW_WRITE }?
                       (
                          recRef:record[true, false, false]
                          (
                             options { generateAmbigWarnings = false; }
                             :
                             write_referencing_phrase[#recRef]
                          )*
                          (
                             options { generateAmbigWarnings = false; }
                             :
                             write_referencing_phrase[#recRef]
                          )*
                       )
                    
                    | { eventType == KW_ASSIGN }?
                    (
                       fldRef:lvalue
                       assign_trigger_old_var_def[#fldRef]
                    )
                 )
              )
              
              (
                 options { generateAmbigWarnings = false; }
                 :
                 KW_OVERRIDE
              )?
           )
           
           // this is the key_label key_function form
         | event key_function DOT!    { bypass = true; }
         
           // this is the standard form (that accepts lists and operates on
           // specific widgets
         | comma_event_list
           (
                KW_ANYWHERE
              | KW_OF comma_widget_list 
                (
                   options { generateAmbigWarnings = false; }
                   :
                     KW_ANYWHERE
                   | comma_events_of_widgets_clause
                )*
           )
      )
      
      // here is the trigger that fires
      (
           // we don't think the trigger block can be labeled, but if it can (and the label can
           // be malformed, then checkAndFixLabelWithReservedKeywordPrefix(1) will have to be
           // called here to properly engage the prediction logic for single_block
            
           // make this subrule optional if the key_label key_function
           // form was previously matched
           { bypass == false }?
           (
                KW_REVERT DOT!
              | persistent_trigger_procedure 
              | single_block[true, true] // this is the "trigger block"
              
                // this was found in customer code, it doesn't make any sense but the 4GL allows
                // it and silently does nothing
              | end_stmt_with_no_beginning
           )
           
         | // the empty alternative!
      )
      
      {
         clearTriggerScope(oldBufScope);
      }
   ;
   
/**
 * Match an end statement (and following DOT) that does not have an associated opening block.
 * This will manufacture an empty block to be returned, so that the downstream code can treat
 * this normally. 
 */
end_stmt_with_no_beginning!
   :
      end_stmt[true]
      {
         // build the tree
         ## = #([TRIGGER_BLOCK, "trigger block"], [KW_DO, "do"], [BLOCK, "block"]);
      }
   ;
   
/**
 * Implements the referencing phrase to modify a database event trigger,
 * matches on the <code>NEW or OLD</code> keywords and all following
 * optional or required constructs.
 * <p>
 * The 4GL documentation incorrectly specifies that for <code>WRITE</code>
 * triggers, there will be both a <code>NEW</code> and <code>OLD</code>
 * keyword, but it turns out that the <code>NEW</code> keyword is optional,
 * so the <code>OLD</code> keyword can appear by itself.
 * <p>
 * This is separated to make the AST build more natural.  Note that this
 * AST could be improved further.
 *
 * @param    ref
 *           The record or field being referenced. 
 */
write_referencing_phrase [Aast ref]
   :
      (KW_NEW^ | KW_OLD^)
      (
         options { generateAmbigWarnings = false; }
         :
         KW_BUFFER!
      )?
      b:symbol
      {
         ##.putAnnotation("define_buffer", new Boolean(true));
         defineBufferFromSymbol(#b, ref.getText(), null, null, false);
      }
   ;
   
/**   
 * Matches the list of all valid key function names and generates a node
 * of type <code>KEY_FUNCTION</code>. Used by {@link #on_stmt}.
 * <p>
 * <pre>
 * ABORT
 * APPEND-LINE (undocumented)
 * BACKSPACE
 * BACK-TAB
 * BELL
 * CLEAR
 * CURSOR-DOWN
 * CURSOR-LEFT
 * CURSOR-RIGHT
 * CURSOR-UP
 * DELETE-CHARACTER
 * DELETE-LINE (undocumented)
 * EDITOR-TAB (undocumented)
 * END
 * END-ERROR
 * ENDKEY
 * ENTER-MENUBAR
 * ERROR
 * GET (undocumented)
 * GO
 * HELP
 * HOME
 * INSERT-MODE
 * LEFT-END
 * NEW-LINE (undocumented)
 * NEXT-FRAME
 * PREV-FRAME
 * RECALL
 * RETURN
 * RIGHT-END
 * SCROLL-MODE
 * STOP
 * TAB
 * </pre>
 * <p>
 * <b>Page 785 of the Progress Language Reference lists these as the only
 * valid key functions that can be assigned actions.</b> This is the same
 * as the list on page 6-26 in the Progress Programming Handbook, which
 * makes the same statement about this being the list of possible actions.
 * However, page 6-19 in the Progress Programming Handbook, has a much
 * longer list of key functions and tests show that these other functions
 * ARE accepted in such an <code>ON</code> statement. These confusingly
 * documented additional keywords are supported here. Working 4GL code does
 * rely upon these events.  The keywords that are added (the other keywords
 * show up in both lists) are:
 * <p>
 * <pre>
 * BLOCK
 * BOTTOM-COLUMN
 * BREAK-LINE
 * CANCEL-PICK
 * CHOICES
 * CLOSE
 * COMPILE
 * COPY
 * CUT
 * DEFAULT-POP-UP
 * DELETE-COLUMN
 * DELETE-END-LINE
 * DELETE-FIELD
 * DOWN (found in customer source, not in documentation)
 * EDITOR-BACKTAB
 * EXIT
 * FIND
 * FIND-NEXT
 * FIND-PREVIOUS
 * GOTO
 * INSERT-COLUMN
 * INSERT-FIELD
 * INSERT-FIELD-DATA
 * INSERT-FIELD-LABEL
 * LEFT (found in customer source, not in documentation)
 * MAIN-MENU
 * MOVE
 * NEW
 * NEXT-ERROR
 * NEXT-WORD
 * OPEN-LINE-ABOVE
 * OPTIONS
 * PAGE-DOWN
 * PAGE-LEFT
 * PAGE-RIGHT
 * PAGE-UP
 * PASTE
 * PICK
 * PICK-AREA
 * PICK-BOTH
 * PREV-WORD
 * PUT
 * REPLACE
 * REPORTS
 * RESUME-DISPLAY
 * RIGHT (found in customer source, not in documentation)
 * SAVE-AS
 * SCROLL-LEFT
 * SCROLL-RIGHT
 * SETTINGS
 * STOP-DISPLAY
 * TOP-COLUMN
 * UNIX-END
 * UP (found in customer source, not in documentation)
 * </pre>
 */
key_function   
   :
      (
           // undocumented but fully working form
           s:STRING
      
           // well documented set of key functions that docs say can be
           // mapped as an action for a key
         | KW_ABORT    | KW_APPEND_L | KW_BACKSP   | KW_BACK_TAB | KW_BELL
         | KW_CLEAR    | KW_CUR_DOWN | KW_CUR_LEFT | KW_CUR_RGHT | KW_CUR_UP
         | KW_DEL_CHAR | KW_DEL_LINE | KW_EDIT_TAB | KW_END      | KW_END_ERR
         | KW_ENDKEY   | KW_ENTER_MB | KW_ERROR    | KW_GET      | KW_GO    
         | KW_HELP     | KW_HOME     | KW_INS_MODE | KW_LEFT_END | KW_NEW_LINE
         | KW_NEXT_FR  | KW_PREV_FR  | KW_RECALL   | KW_RETURN   | KW_RT_END
         | KW_SCR_MODE | KW_STOP     | KW_TAB
         
           // additional key functions from the extended list that are
           // confusingly left out of the list of things that can be remapped,
           // but which in fact can be remapped
         | KW_BLOCK    | KW_BOTTOM_C | KW_BREAK_L  | KW_CANCEL_P | KW_CHOICES
         | KW_CLOSE    | KW_COMPILE  | KW_COPY     | KW_CUT      | KW_DEF_POP
         | KW_DEL_COL  | KW_DEL_E_L  | KW_DEL_FLD  | KW_DEL_WORD | KW_EDIT_B_T
         | KW_EXIT     | KW_FIND     | KW_FIND_NXT | KW_FIND_PRV | KW_GOTO
         | KW_INS_COL  | KW_INS_FLD  | KW_INS_FLDD | KW_INS_FLDL | KW_MAINMENU
         | KW_MOVE     | KW_NEW      | KW_NEXT_ERR | KW_NEXTWORD | KW_OPEN_L_A
         | KW_OPTIONS  | KW_PAGE_DWN | KW_PAGE_LFT | KW_PAGE_RT  | KW_PAGE_UP
         | KW_PASTE    | KW_PICK     | KW_PICK_ARE | KW_PICK_BTH | KW_PREVWORD
         | KW_PUT      | KW_REPLACE  | KW_REPORTS  | KW_RESUME_D | KW_SAVE_AS
         | KW_SCR_LEFT | KW_SCR_RT   | KW_SETTINGS | KW_STOP_D   | KW_TOP_COL
         | KW_UNIX_END
         
           // found in customer source, not in the documentation
         | KW_DOWN     | KW_LEFT     | KW_RIGHT    | KW_UP
      )                 
      {
         // convert the string to a keyword (TODO: this does not check for the matching subset
         // of key function types)
         if (#s != null)
         {
            convertStringTo(#s, -1);
         }
         
         saveAndReplaceType(##, KEY_FUNCTION);
      }
   ;
      
/**   
 * Matches the <code>PERSISTENT</code> keyword and the following subset of
 * the {@link #run_stmt}. This matches more constructs than a normal
 * persistent trigger could specify but this should not matter in valid
 * Progress code.
 * <p>
 * Used by {@link #on_stmt}.
 */   
persistent_trigger_procedure   
   :
      KW_PERSIST^ run_stmt 
   ;
   
/**   
 * Matches the <code>APPLY</code> language statement.  Calls the
 * {@link #expr} and {@link #to_event_target} rules.
 * <p>
 * Used by {@link #stmt_list}.
 */   
apply_stmt   
   :
      KW_APPLY^ expr (to_event_target)? DOT!
   ;
   
/**   
 * Matches the <code>TO widget_phrase</code> construct of the
 * {@link #apply_stmt}.  Separated to improve the tree build.
 */   
to_event_target   
   :
      KW_TO^ widget_phrase
   ;
      
/**   
 * Matches the <code>WAIT-FOR</code> language statement.  Calls the
 * {@link #events_of_widgets_clause}, {@link #pause_clause} and
 * {@link #focus_clause} rules.
 * <p>
 * Options can be specified in any order.
 * <p>
 * ANTLR reports bogus ambiguity, the code doesn't really have a problem so
 * warnings have been disabled.
 * <p>
 * Note that there is a real ambiguous situation since the <code>FOCUS</code>
 * keyword that is an option is also a valid system handle which can be
 * matched as part of a widget phrase and thus it can appear in a widget
 * list after the <code>OF</code> keyword.  In the case where the
 * <code>FOCUS</code> option immediately follows a widget list, the
 * <code>FOCUS</code> keyword will be matched as a widget rather than as
 * a keyword.  This is because the widget lists are greedy loops and this is
 * quite standard behavior.  If this occurs, the tree will be different than
 * one might expect.  Watch for this situation.
 * <p>
 * Another real ambiguous situation (resolved by ordering of alternatives)
 * is that the <code>PAUSE, FOCUS or EXCLUSIVE-WEB-USER</code> will be
 * matched before the event list.  Otherwise, such constructs can not be
 * matched at all.  This means that a <code>PAUSE</code> event (if there is
 * such a thing) can't be matched.
 * <p>
 * Used by {@link #stmt_list}.
 */   
wait_for_stmt      
   :
      KW_WAIT_FOR^
      (
         options { generateAmbigWarnings = false; }
         :
           pause_clause
         | focus_clause
         | KW_EXCL_WEB
         | events_of_widgets_clause
      )+
      DOT!
   ;
   
/**   
 * Matches the .NET version of the <code>WAIT-FOR</code> language statement.
 * It calls {@link #expr} and {@link #set_lvalue}. This is ambiguous with
 * the {@link #wait_for_stmt} but that ambiguity must be resolved in the
 * caller using semantic predicates.
 * <p>
 * Used by {@link #stmt_list}.
 */   
wait_for_dotnet_stmt
   :
      KW_WAIT_FOR^ expr (set_lvalue)? DOT!
   ;
   
/**   
 * Matches the <code>SET handle</code> clause and a following {@link #lvalue}.
 * <p>
 * Used by {@link #wait_for_dotnet_stmt}. The subtree is rooted by the keyword
 * (which is the reason for using a separate rule for something so simple).
 */
set_lvalue   
   :
      KW_SET^ lvalue
   ;
   
/**   
 * Matches the <code>FOCUS</code> keyword and the required following
 * widget. Calls {@link #widget_phrase}.
 * <p>
 * Used by {@link #wait_for_stmt}.
 * <p>
 * This is a separate rule to centralize logic and build the AST properly.
 */   
focus_clause      
   :
      KW_FOCUS^ widget_phrase
   ;
      
/**   
 * Matches the optional <code>OR</code> keyword prefix and the required
 * event list, <code>OF</code> keyword and then a widget list.  Calls
 * {@link #event_list} and {@link #widget_list} rules.
 * <p>
 * Used by {@link #wait_for_stmt}.
 * <p>
 * This is a separate rule to centralize logic and build the AST properly.
 * <p>
 * Bogus ANTLR ambiguity warnings have been disabled.
 */   
events_of_widgets_clause      
   :
      (
         options { generateAmbigWarnings = false; }
         :
         KW_OR!
      )?
      event_list KW_OF^ widget_list
   ;
      
/**   
 * Matches the <code>OR</code> keyword, the required comma separated event
 * list, <code>OF</code> keyword and then a widget list.  Calls
 * {@link #comma_event_list} and {@link #comma_widget_list} rules.
 * <p>
 * Used by {@link #on_stmt}.
 * <p>
 * This is a separate rule to centralize logic and build the AST properly.
 */   
comma_events_of_widgets_clause      
   :
      KW_OR! comma_event_list KW_OF^ comma_widget_list
   ;      
      
/**
 * Matches the widget phrase construct of the Progress 4GL language which
 * consists of the optional <code>FIELD</code> keyword prefix and the
 * required following widget reference.  This optionally appears in a 
 * referencing rule but does not begin with a specific keyword.  All real 
 * processing is done using the {@link #widget} rule which really delegates
 * its work to {@link #lvalue}.
 * <p>
 * The <code>INPUT</code> keyword can be optionally inserted between the
 * <code>TO</code> and the widget. This is undocumented and has no discernable
 * semantic meaning so it is dropped from the tree. If the <code>INPUT</code>
 * keyword does exist, then it can be followed by an optional
 * {@link #frame_reference} which can't normally exist in front of the
 * widget.
 * <p>
 * Called from {@link #hide_stmt}, {@link #view_stmt}, {@link #widget_list}
 * {@link #comma_widget_list} and {@link #focus_clause}.
 */
widget_phrase   
   :
      (KW_FIELD!)?
      (
         KW_INPUT!
         (
            options { generateAmbigWarnings = false; }
            :
            frame_reference[false]
         )?
      )?
      widget
   ;
   
/**   
 * Matches the <code>QUERY</code> keyword and the required query name that
 * is coded as a {@link #malformed_symbol}. 
 * <p>
 * Used by {@link #def_browse_stmt}, {@link #reposition_stmt} and
 * {@link #get_stmt} rules.
 * <p>
 * This rule naturally duplicates a small subset of what can be matched
 * via {@link #lvalue}.  This is done to keep usage that is known to be
 * hard coded to this subset, unambiguous.
 * <p>
 * This is a separate rule to centralize logic and build the AST properly.
 */   
query_reference
   :
      (
         options { generateAmbigWarnings = false; }
         :
         KW_QUERY^
      )?
      q:malformed_symbol
      {
         #q.setType(QUERY);
         
         // TODO: need to check if explicit static/instance qualifiers can be used here
         // TODO: need to check implicit static/instance (based on the surrounding code)
         if (sym.lookupQuery(#q.getText()) != QUERY)
         {
            System.out.println(
               String.format("Warning: lookup of query '%s' failed at line %d column %d", 
                             #q.getText(),
                             #q.getLine(),
                             #q.getColumn()));
         }
      }
   ;
   
/**   
 * Matches a space or comma delimited list of widgets.  Calls
 * {@link #widget_phrase}. 
 * <p>
 * Used by {@link #wait_for_stmt} and this form has <code>WAIT-FOR</code>
 * specific logic so it may not be suitable for other use.
 * <p>
 * Bogus ANTLR ambiguity warnings have been suppressed.
 * <p>
 * <b>This rule uses a trick to create an artificial node as a root while
 * still allowing ANTLR to do normal tree building (no suppression of the
 * AST build).</b>  This is only useful in cases where there is no valid
 * token which can be used as a tree root AND where a loop is being used
 * such that the contents of the loop cannot be referenced in a manually
 * built #([,],,,) ANTLR construct. ANTLR doesn't support (as of 2.7.4) the
 * ability to label a loop such that the entire contents can be referenced.
 * If one labels the individual elements inside the loop, only the last one
 * is available when the loop concludes.  However, by leaving the normal
 * tree creation unchanged but manually creating an artificial node and
 * manually setting this node as the root of the returned tree BEFORE the
 * children are added, this problem can be solved.
 * <p>
 * This construct is dangerous because the loop must be manually closed.
 * The core issue is that a widget phrase can match a the <code>FOCUS</code>
 * system handle BUT <code>WAIT-FOR</code> also uses the <code>KW_FOCUS</code>
 * for its own purposes.  This means that the second and following widget 
 * phrases must not match a <code>FOCUS</code> system handle but instead must
 * manually close the loop and allow the calling code to match with a
 * {@link #focus_clause}.  This is only a problem because of the optional
 * commas that delimit widgets.  If the comma was mandatory, then the lack of
 * a following comma would end the list as is the case in 
 * {@link #comma_widget_list}.  A semantic predicate was added to close the
 * loop manually when a <code>KW_FOCUS</code> is encountered. 
 */   
widget_list   
   :
      { astFactory.makeASTRoot(currentAST, #[WIDGET_LIST, ""]); }   
      widget_phrase
      (
         options { generateAmbigWarnings = false; }
         :
         { LA(1) != KW_FOCUS }?
         (COMMA!)?
         widget_phrase
      )*
      { if (##.getNumberOfChildren() == 0) ## = null; }
   ;
   
/**   
 * Matches a comma delimited list of widgets.  Calls {@link #widget_phrase}. 
 * <p>
 * Used by ON.
 * <p>
 * <b>This rule uses a trick to create an artificial node as a root while
 * still allowing ANTLR to do normal tree building (no suppression of the
 * AST build).</b>  This is only useful in cases where there is no valid
 * token which can be used as a tree root AND where a loop is being used
 * such that the contents of the loop cannot be referenced in a manually
 * built #([,],,,) ANTLR construct. ANTLR doesn't support (as of 2.7.4) the
 * ability to label a loop such that the entire contents can be referenced.
 * If one labels the individual elements inside the loop, only the last one
 * is available when the loop concludes.  However, by leaving the normal
 * tree creation unchanged but manually creating an artificial node and
 * manually setting this node as the root of the returned tree BEFORE the
 * children are added, this problem can be solved.
 */   
comma_widget_list   
   :
      { astFactory.makeASTRoot(currentAST, #[WIDGET_LIST, ""]); }   
      widget_phrase (COMMA! widget_phrase)*
      { if (##.getNumberOfChildren() == 0) ## = null; }
   ;
     
   
/**   
 * Matches a space or comma delimited list of events.  Calls {@link #event}. 
 * <p>
 * Used by {@link #wait_for_stmt}.
 * <p>
 * <b>This rule uses a trick to create an artificial node as a root while
 * still allowing ANTLR to do normal tree building (no suppression of the
 * AST build).</b>  This is only useful in cases where there is no valid
 * token which can be used as a tree root AND where a loop is being used
 * such that the contents of the loop cannot be referenced in a manually
 * built #([,],,,) ANTLR construct. ANTLR doesn't support (as of 2.7.4) the
 * ability to label a loop such that the entire contents can be referenced.
 * If one labels the individual elements inside the loop, only the last one
 * is available when the loop concludes.  However, by leaving the normal
 * tree creation unchanged but manually creating an artificial node and
 * manually setting this node as the root of the returned tree BEFORE the
 * children are added, this problem can be solved.
 * <p>
 * This construct is dangerous because the loop must be manually closed.
 * The core issue is that an event can match a reserved keyword and thus 
 * even the <code>OF</code> keyword will be matched as an event rather than
 * the keyword that denotes the end of the event list.  This is only a
 * problem because of the optional commas that delimit events.  If the comma
 * was mandatory, then the lack of a following comma would end the list
 * as is the case in {@link #comma_event_list}.  A semantic predicate was
 * added to close the loop manually when the <code>OF</code> keyword is
 * encountered.
 */   
event_list   
   :
      { astFactory.makeASTRoot(currentAST, #[EVENT_LIST, ""]); }   
      event
      (
         options { generateAmbigWarnings = false; }
         :
         { LA(1) != KW_OF }? 
         (COMMA!)? event
      )*
      { if (##.getNumberOfChildren() == 0) ## = null; }
   ;
   
/**   
 * Matches a comma delimited list of events.  Calls {@link #event}. 
 * <p>
 * Used by ON.
 * <p>
 * <b>This rule uses a trick to create an artificial node as a root while
 * still allowing ANTLR to do normal tree building (no suppression of the
 * AST build).</b>  This is only useful in cases where there is no valid
 * token which can be used as a tree root AND where a loop is being used
 * such that the contents of the loop cannot be referenced in a manually
 * built #([,],,,) ANTLR construct. ANTLR doesn't support (as of 2.7.4) the
 * ability to label a loop such that the entire contents can be referenced.
 * If one labels the individual elements inside the loop, only the last one
 * is available when the loop concludes.  However, by leaving the normal
 * tree creation unchanged but manually creating an artificial node and
 * manually setting this node as the root of the returned tree BEFORE the
 * children are added, this problem can be solved.
 */   
comma_event_list   
   :
      { astFactory.makeASTRoot(currentAST, #[EVENT_LIST, ""]); }   
      event (COMMA! event)*
      { if (##.getNumberOfChildren() == 0) ## = null; }
   ;
   
/**   
 * Matches any valid Progress symbol or string returns an artificial token
 * <code>DB_EVENT</code> with a child that is the matched token. Matches
 * <code>CREATE, DELETE, FIND, WRITE and ASSIGN</code> keywords.
 * <p>
 * Used by {@link #on_stmt}.
 */   
db_event!   
   :
      (c:KW_CREATE | d:KW_DELETE | f:KW_FIND | w:KW_WRITE | a:KW_ASSIGN)
      { ## = #([DB_EVENT, "database event"], c, d, f, w, a); }
   ;   
   
/**   
 * Matches any valid Progress symbol or string or other key label and returns
 * an artificial token <code>EVENT</code> with a child that is the matched
 * token. Calls {@link #symbol} and {@link #other_single_char_tokens}. 
 * <p>
 * Used by {@link #event_list} and {@link #comma_event_list}.
 */   
event   
   :
      (s:STRING | o:other_single_char_tokens | e:reserved_or_symbol)
      {
         if (#s != null) convertStringTo(#s, EVENT);
         if (#o != null) #o.setType(EVENT);
         if (#e != null) #e.setType(EVENT);
      }         
   ;
      
/**   
 * Matches any valid Progress single char key label that would not be 
 * recognized as a symbol or string.  Note that 
 * <code>DOT, COLON and COMMA</code> are left out to allow parsing to
 * properly occur.  It is probable that to use these as key labels in
 * Progress, one would have to put the character in a string.
 * <p>
 * Matches:
 * <p>
 * <pre>
 * EQUALS
 * GT
 * LT
 * LPARENS 
 * RPARENS 
 * LBRACKET
 * RBRACKET 
 * PLUS
 * MINUS
 * MULTIPLY
 * AT
 * CARET 
 * UNKNOWN_VAL
 * NUM_LITERAL
 * UNKNOWN_TOKEN
 * </pre>
 * <p>
 * Called by {@link #event}.  <b>USE THIS SPARINGLY, IT IS NASTY!</b>
 */   
other_single_char_tokens
   :
      // these are all possible in an event! note that DOT, COLON, COMMA
      // and multi-char non-symbols (and non-strings) are not included
        EQUALS   | GT       | LT       | LPARENS  | RPARENS  | LBRACKET
      | RBRACKET | PLUS     | MINUS    | MULTIPLY | AT       | CARET 
      | UNKNOWN_VAL
      | NUM_LITERAL
      | UNKNOWN_TOKEN
   ;
      
/**   
 * Matches the <code>ALL</code> keyword and an optional list of fields
 * that should not be included. Calls the {@link #widget_array} rule.
 * <p>
 * Used by {@link #disable_stmt} and {@link #enable_stmt}.
 * <p>
 * This is a separate rule to centralize logic and build the AST properly.
 */   
all_clause
   :
      KW_ALL^ ( KW_EXCEPT widget_array )?
   ;
   
/**
 * Aggregates (in a tree building sense) a simple list of all fields under 
 * a specific language statement.  Calls {@link #widget} and roots everything
 * at a node of type <code>CONTENT_ARRAY</code>.
 */
widget_array
   :
      { astFactory.makeASTRoot(currentAST, #[CONTENT_ARRAY, ""]); }
      (widget)+
   ;
   
/**
 * Matches the <code>TEXT</code> keyword and an following field specification.
 * Calls the {@link #field_clause} rule.
 * <p>
 * Used by {@link #enable_stmt}.
 * <p>
 * This is a separate rule to centralize logic and build the AST properly.
 */   
text_clause
   :
      KW_TEXT^ LPARENS! (field_clause[false])+ RPARENS!
   ;
   
/**
 * Matches a field name and an optional <code>WHEN</code> condition.
 * Calls the {@link #widget} and {@link #format_phrase} rules.
 * <p>
 * Used by {@link #disable_stmt}, {@link #enable_stmt} and 
 * {@link #text_clause}.
 * <p>
 * The use of the format phrase naturally conflicts with the large set of
 * possible following tokens.  This is not really ambiguous so warnings
 * have been disabled.
 * <p>
 * This is a separate rule to centralize logic.
 *
 * @param   enableWidget
 *          <code>true</code> if this field clause represents a widget being enabled in a frame;
 *          else <code>false</code>.
 */
field_clause [boolean enableWidget]
   :
      {
         Aast   fld        = null;
         String symbolName = null;
      }
      (
         { allowSymbolMatch = true; }
         l:lvalue
         {
            // reset each loop iteration
            symbolName = null;
            fld        = null;
            
            if (#l.getType() == SYMBOL)
            {
               symbolName = #l.getText();
            }
            else
            {
               fld = findFieldNode(#l);
            }
            
            allowSymbolMatch = false;
         }
         (
            options { generateAmbigWarnings = false; }
            :
            f:format_phrase[symbolName, fld, true, enableWidget]
         )?
      )
   ;
   
/**
 * Matches the <code>IN WINDOW</code> clause seen in many Progress language
 * statements such as {@link #status_stmt} and {@link #pause_stmt}.
 * <p>
 * Following the keywords is a window reference which is coded here as a
 * reference to the {@link #expr} rule.  It is possible that there is some
 * window namespace processing that is needed such that named windows can
 * be referenced here. Any such processing is not handled at this time.
 * <p>
 * By making a separate rule for this, the logic is centralized and the
 * AST is kept simple because all tokens can be rooted at this option's
 * <code>IN</code> keyword.
 */
in_window_clause
   :
      KW_IN^ KW_WINDOW expr
   ;
   
/**
 * Matches the <code>FRAME frame</code> reference seen in many Progress
 * language statements such as {@link #clear_stmt} and the common
 * {@link #frame_phrase}.
 * <p>
 * If the single parameter is true, the frame is added to the frame namespace.
 * In any case, the following {@link #symbol} is rewritten to a
 * <code>WID_FRAME</code> token.
 * <p>
 * This rule is a simplified version of what can also be accessed directly
 * from {@link #lvalue} but this keeps certain references much less
 * ambiguous.
 * <p>
 * By making a separate rule for this, the logic is centralized and the
 * AST is kept simple because all tokens can be rooted at this option's
 * <code>FRAME</code> keyword.
 */
frame_reference [boolean define]
   :
      KW_FRAME^ frame:malformed_symbol
      {
         #frame.setType(WID_FRAME);
         
         if (define)
         {
            sym.addFrame(#frame.getText(), WID_FRAME);
         }
      }
   ;
   
/**
 * Matches a name which can be a simple symbol or optionally can be any number of tokens until
 * it finds whitespace, COLON, DOT, EQUALS, RPARENS, LPARENS, NOT_EQ, LT, LTE, GT or GTE.
 * <p>
 * Certain resource names are unusual symbols in that they can begin with a numeric digit,
 * minus sign and other unexpected characters that normally can only occur as the 2nd or
 * later character in a symbol.  In addition, other non-symbol tokens can appear inside
 * such symbols (e.g. the @ sign).  All of this is based on real cases found in customer
 * source code.
 * <p>
 * This special case is converted into a single <code>SYMBOL</code> token with the concatenated
 * text of all merged tokens.
 * <p>
 * Resource types that have this undocumented behavior:
 * <p>
 * <ol>
 *    <li> frame names
 *    <li> labels
 *    <li> query names
 *    <li> procedure names
 *    <li> function names
 * </ol>
 */
malformed_symbol
   :
      {
         // DOT and EOF are hard coded in the worker routine
         int[] exclude = new int[]
         {
            COLON,
            COMMA,
            EQUALS,
            LPARENS,
            RPARENS,
            NOT_EQ,
            LT,
            LTE,
            GT,
            GTE
         };
         
         boolean bad = mergeUntilWS(exclude);
         
         LT(1).setType(SYMBOL);
      }
      (
           SYMBOL^
         
           // this reference is for "pull" only, its corresponding type
           // is rewritten unconditionally at entry to this rule so it
           // will never get executed
         | ~DOT ~DOT ~DOT
      )
      {
         if (bad)
         {
            ##.putAnnotation("malformed", new Boolean(true));
         }
      }
   ;
   
/**   
 * Matches the <code>TERMINAL</code> language statement.  Calls the
 * {@link #expr} rule.
 * <p>
 * The <code>TERMINAL</code> keyword to the right side of the equals sign
 * is ambiguous with the <code>TERMINAL</code> function/variable that can 
 * be reached by the following expression. This isn't really a problem so
 * ambiguity warnings are disabled.
 * <p>
 * Used by {@link #stmt_list}.
 */
terminal_stmt   
   :
      KW_TERM^ EQUALS! 
      (
         options { generateAmbigWarnings = false; }
         :
           KW_TERM
         | expr
      )
      DOT!
   ;
   
/**   
 * Matches the <code>LOAD</code> language statement with a required following
 * {@link #expr}.  All options are supported and the following rules can
 * be called:
 * <p>
 * <ul>
 *    <li> {@link #dir_clause}
 *    <li> {@link #base_key_clause}          
 * </ul>
 * <p>
 * Used by {@link #stmt_list}.
 */
load_stmt   
   :
      KW_LOAD^ expr
      (
           dir_clause
         | KW_APPL
         | KW_NEW
         | base_key_clause
         | KW_NO_ERROR
      )*
      DOT!
   ;
   
/**
 * Matches the <code>DIR</code> keyword and a following {@link #expr}.
 * <p>
 * Used by {@link #load_stmt}.
 */
dir_clause
   :
      KW_DIR^ expr
   ;
   
/**
 * Matches the <code>BASE-KEY</code> keyword and a following
 * {@link #external_name_spec}.
 * <p>
 * Used by {@link #load_stmt}.
 */
base_key_clause
   :
      KW_BASE_KEY^ external_name_spec
   ;
   
/**   
 * Matches the <code>USE</code> language statement with a required following
 * {@link #expr}.
 * <p>
 * Used by {@link #stmt_list}.
 */
use_stmt   
   :
      KW_USE^ expr (KW_NO_ERROR)? DOT!
   ;
   
/**
 * Matches the <code>GET-KEY-VALUE</code> statement.  Calls the following:
 * <p>
 * <ul>
 *    <li> {@link #section_clause}
 *    <li> {@link #key_clause} 
 *    <li> {@link #value_clause}
 * </ul>
 * <p>
 * Used by {@link #stmt_list}.
 */   
get_key_value_stmt   
   :
      KW_GET_K_V^ section_clause key_clause value_clause DOT!
   ;
   
put_key_value_stmt
   :
      KW_PUT_K_V^
      (
           (section_clause key_clause value_clause)
         | color_or_font_clause
      )
      (KW_NO_ERROR)?
      DOT!
   ;
   
/**
 * Matches the <code>COLOR or FONT</code> clause.  Used by
 * {@link #put_key_value_stmt}.
 */
color_or_font_clause
   :
      (KW_COLOR^ | KW_FONT^) (expr | KW_ALL)
   ;
   
/**
 * Matches the <code>SECTION</code> clause and a required {@link #expr}.
 * Used by {@link #get_key_value_stmt} and {@link #put_key_value_stmt}.
 */
section_clause
   :
      KW_SECTION^ expr
   ;
   
/**
 * Matches the <code>KEY</code> clause and a required {@link #expr} or
 * <code>DEFAULT</code> keyword.  Used by {@link #get_key_value_stmt}
 * and {@link #put_key_value_stmt}.
 */
key_clause
   :
      KW_KEY^ (KW_DEFAULT | expr)
   ;
   
/**
 * Matches the <code>VALUE</code> clause and a required {@link #expr}.
 * Used by {@link #get_key_value_stmt} and {@link #put_key_value_stmt}.
 */
value_clause
   :
      KW_VALUE^ 
      
      // we use an expr here instead of an lvalue so that this rule
      // can be used by both get and put rules
      expr
   ;
   
/**   
 * Matches the <code>UNLOAD</code> language statement with a required
 * following {@link #expr}.
 * <p>
 * Used by {@link #stmt_list}.
 */
unload_stmt   
   :
      KW_UNLOAD^ expr (KW_NO_ERROR)? DOT!
   ;
   
/**
 * Matches the various permutations of language statements that start with
 * the <code>INPUT, INPUT-OUTPUT or OUTPUT</code> keywords.  It matches
 * these keywords and then switches into the correct alternative based on the
 * lookahead tokens.
 * <p>
 * This rule is called by {@link #stmt_list} and it handles the matching
 * unambiguously by left factoring the common keywords for multiple language
 * statements. The alternatives:  
 * <ul>
 *    <li> CLEAR (so simple that it is inlined)
 *    <li> CLOSE (so simple that it is inlined)
 *    <li> FROM {@link #io_from_to_stmt}
 *    <li> THROUGH/THRU {@link #io_through_stmt}
 *    <li> TO {@link #io_from_to_stmt}
 * </ul>
 * <p>
 * Certain alternatives can be followed by a reference to
 * {@link #io_common_options}.
 */      
io_stmt   
   :
      {
         boolean input  = true;
         boolean output = false;
         boolean thru   = false;
         Token t = LT(1);
         AST io_root = #([KW_INPUT,"i/o stmt"]);  // placeholder
         ((Aast) io_root).setLine(t.getLine());
         ((Aast) io_root).setColumn(t.getColumn());

         astFactory.makeASTRoot(currentAST, io_root);         
      }
      (
         (
              KW_INPUT!
            | KW_IN_OUT!   { input = false; }
            | KW_OUTPUT!   { input = false; output = true; }
         )
         (
              { input }?
              KW_CLEAR!    { io_root.setType(INPUT_CLEAR); }
            | (
                   { LA(1) == KW_STREAM || LA(1) == KW_STRM_HND }? lvalue
                 | // empty alternative
              )
              (
                   KW_CLOSE!
                   {
                      if (input)
                      {
                         io_root.setType(INPUT_CLOSE);
                      }
                      else if (output)
                      {
                         io_root.setType(OUTPUT_CLOSE);
                      }
                      else
                      {
                         io_root.setType(INPUT_OUTPUT_CLOSE);
                      }
                   }
                 | (
                      { input || output }?
                      io_from_to_stmt
                      {
                         if (input)
                         {
                            io_root.setType(INPUT_FROM);
                         }
                         else
                         {
                            io_root.setType(OUTPUT_TO);
                         }
                      }                   
                    | io_through_stmt
                      {
                         thru = true;
                         
                         if (input)
                         {
                            io_root.setType(INPUT_THRU);
                         }
                         else if (output)
                         {
                            io_root.setType(OUTPUT_THRU);
                         }
                         else
                         {
                            io_root.setType(INPUT_OUTPUT_THRU);
                         }
                      }                 
                   )
                   io_common_options[output, thru]
              )
         )
         DOT!
      )
   ;
   
/**   
 * Matches the <code>INPUT FROM and OUTPUT TO</code> language statements.
 * References the following:
 * <p>
 * <ul>
 *     <li> {@link #value}
 *     <li> {@link #os_dir_clause}
 *     <li> {@link #filename}
 * </ul>
 * <p>
 * Although it is undocumented, it is valid to use a string in place of a
 * file or device name (without using the <code>VALUE</code> construct).
 * The only documented use for this is in <code>OUTPUT TO</code> with the
 * &quot;CLIPBOARD&quot; but it turns out that it can be more generally
 * used.
 * <p>
 * This rule is called by {@link #io_stmt} which is what matches the
 * <code>INPUT and OUTPUT</code> keywords. This refactoring allows all I/O
 * statements to be supported without ambiguity.
 */
io_from_to_stmt   
   :
      {
         boolean output = false;
      }
      (
         (KW_FROM^ | KW_TO^ { output = true; } )
         (
              options { generateAmbigWarnings = false; }
              :
              KW_TERM
              
            | s:STRING
              {
                 // convert the Progress string to a Java string
                 String txt = character.progressToJavaString(#s.getText());
                 
                 // we won't be a STRING when this is done, so fix the text
                 #s.setText(txt);
                 
                 // determine the new type (this translates some special cases which the 4GL
                 // treats as keywords even though they were entered as string literals)
                 if (txt.equalsIgnoreCase("term") || txt.equalsIgnoreCase("terminal"))
                 {
                    #s.setType(KW_TERM);
                 }
                 else if (txt.equalsIgnoreCase("clipboard"))
                 {
                    #s.setType(KW_CLIP);
                 }
                 else if (txt.equalsIgnoreCase("web"))
                 {
                    #s.setType(KW_WEB);
                 }
                 else
                 {
                    // default
                    #s.setType(FILENAME);
                 }
              }
            
            | os_dir_clause
            
            | { output }?
              printer_clause
              
            | { output }?
              KW_WEB
              
            | LPARENS! expr RPARENS! // undocumented but found in cust code
                                     // (it is tolerated by Progress, but
                                     // just looks like someone forgot to
                                     // prepend the "value"! we don't know
                                     // if it actually works
            | value
            
              // MUST BE LAST to avoid ambiguity!
              // also handles the opsys-device parameter
              // which usually look just like filenames
              // but on some platforms can have a following
              // ':' as in lpt1:
            | filename
         )
      )
   ;   
   
/**   
 * Matches the <code>INPUT THROUGH/THRU, INPUT-OUTPUT THROUGH/THRU</code>
 * and <code>OUTPUT THROUGH/THRU</code> language statements.
 * <p>
 * References the following:
 * <p>
 * <ul>
 *     <li> {@link #external_name_spec}
 *     <li> {@link #command_token_list}
 * </ul>
 * <p>
 * Although it is undocumented, it is valid to use a string in place of a
 * file or device name (without using the <code>VALUE</code> construct).
 * <p>
 * This rule is called by {@link #io_stmt} which is what matches the
 * <code>INPUT, INPUT-OUTPUT and OUTPUT</code> keywords. This refactoring
 * allows all I/O statements to be supported without ambiguity.
 */
io_through_stmt   
   :
      KW_THROUGH^ external_name_spec command_token_list[true]
   ;    
   
/**   
 * Matches the <code>OS-DIR</code> construct with its following parenthesized
 * character {@link #expr} and options.
 * <p>
 * This rule is called by {@link #io_from_to_stmt}.
 */
os_dir_clause   
   :
      KW_OS_DIR^ LPARENS! expr RPARENS! (KW_NO_ATTRL)?
   ;
   
/**   
 * Matches the <code>PRINTER</code> keyword with its optional following
 * printer name as matched by {@link #filename}.
 * <p>
 * This rule is called by {@link #io_from_to_stmt}.
 */
printer_clause   
   :
      KW_PRINTER^
      (
         options { generateAmbigWarnings = false; }
         :
           STRING
         |
           // avoid collisions with the following options
           {
              LA(1) != KW_BINARY   && LA(1) != KW_ECHO     && 
              LA(1) != KW_NO_ECHO  && LA(1) != KW_UNBUF    &&
              LA(1) != KW_MAP      && LA(1) != KW_NO_MAP   &&
              LA(1) != KW_NO_CVT   && LA(1) != KW_CONVERT  &&
              LA(1) != KW_NUM_COPY && LA(1) != KW_PAGE_SZ  &&
              LA(1) != KW_COLLATE  && LA(1) != KW_LANDSCAP &&
              LA(1) != KW_PORTRAIT && LA(1) != KW_APPEND   &&
              LA(1) != KW_PAGED    && LA(1) != KW_KEEP_MSG
           }?
           filename
         | // the empty alternative
      )
   ;
   
/**   
 * Matches the set of all I/O options and roots them at an <code>IO_OPTIONS</code> node.
 * <p>
 * This rule is called by {@link #io_stmt}.
 *
 * @param    output
 *           If <code>true</code>, then this is an <code>OUTPUT</code> statement. Otherwise
 *           this is either an <code>INPUT</code> or <code>INPUT-OUTPUT</code> statement.
 * @param    thru
 *           If <code>true</code>, then this is an <code>INPUT THROUGH</code>,
 *           <code>INPUT-OUTPUT THROUGH</code> or <code>OUTPUT THROUGH</code> statement.
 *           Progress allows an undocumented <code>NO-ERROR</code> keyword in this
 *           case.
 */
io_common_options [boolean output, boolean thru]
   :
      { astFactory.makeASTRoot(currentAST, #[IO_OPTIONS,"options"]); }
      (io_options[output, thru])*
      { if (##.getNumberOfChildren() == 0) ## = null; }
   ;
   
/**
 * Matches a range of options used in I/O language statements.  Calls the
 * following rules:
 * <p>
 * <ul>
 *    <li> {@link #value} (this is a qualifier for MAP, undocumented feature of the 4GL)
 *    <li> {@link #filename}
 *    <li> {@link #source_or_target_clause}
 *    <li> {@link #literal}
 *    <li> {@link #value}
 * </ul>
 * <p>
 * The use of <code>literal and value</code> is made optional contrary to
 * the Progress documentation since customer code was found where this was
 * a valid construct.
 * <p>
 * Called by {@link #io_common_options}.
 *
 * @param    output
 *           If <code>true</code>, then this is an <code>OUTPUT</code>
 *           statement. Otherwise this is either an <code>INPUT</code> or 
 *           and <code>INPUT-OUTPUT</code> statement.
 * @param    thru
 *           If <code>true</code>, then this is an <code>INPUT THROUGH</code>,
 *           <code>INPUT-OUTPUT THROUGH</code> or <code>OUTPUT THROUGH</code> statement.
 *           Progress allows an undocumented <code>NO-ERROR</code> keyword in this
 *           case.
 */   
io_options [boolean output, boolean thru]
   :
        KW_BINARY
      | (KW_ECHO | KW_NO_ECHO)
      | KW_UNBUF
      | KW_MAP^
        (
           options { generateAmbigWarnings = false; }
           :
           value | filename
        )
      | KW_NO_MAP^
      | KW_LOB_DIR^ (literal | value)?
      | codepage_convert_phrase
      
        // INPUT THROUGH/INPUT-OUTPUT THROUGH/OUTPUT THROUGH allow an undocumented NO-ERROR 
      | { thru }?
        KW_NO_ERROR
      
        // although it is supposed to only be possible with output stmts, it is an undocumented
        // feature that this is allowed in INPUT FROM (but it seems meaningless so we drop it)
      | { !output && !thru }?
        KW_APPEND!
         
      | { output }?
        (
             (KW_NUM_COPY^ | KW_PAGE_SZ^) (literal | value)?
           | KW_COLLATE
           | KW_KEEP_MSG
           | (KW_LANDSCAP | KW_PORTRAIT)
           | KW_APPEND
           | KW_PAGED
        )
   ;
   
/**
 * Match the <code>NO-CONVERT</code> or a <code>CONVERT</code> followed by
 * two {@link #source_or_target_clause} references. Called by 
 * {@link #io_options} and {@link #copy_lob_stmt}.
 */
codepage_convert_phrase
   :
      KW_NO_CVT^ | KW_CONVERT^ (source_or_target_clause[true])*
   ;
   
/**   
 * Matches the <code>SOURCE or TARGET</code> and following codepage or field
 * construct which is matched as a {@link #expr}.
 * <p>
 * This rule is called by {@link #codepage_convert_phrase},  when it is
 * matching codepages and by {@link #dde_get_stmt} or
 * {@link #dde_request_stmt} or {@link #dde_send_stmt} when matching as a
 * source or target field.
 *
 * @param    cp
 *           <code>true</code> to allow the <code>CODEPAGE</code> keyword
 *           to be inserted between the <code>SOURCE</code> or 
 *           <code>TARGET</code> and the following expression.
 */
source_or_target_clause [boolean cp]   
   :
      (KW_SOURCE^ | KW_TARGET^)
      (
           { cp }?
           KW_CP
           
         | // empty alternative
      )
      expr
   ;
   
/**   
 * Matches the <code>MENU or SUB-MENU menu</code> reference seen in the
 * {@link #widget_qualifier}.  Other menu or sub-menu references (see
 * {@link #widget_phrase}) are resolved by common/master processing in the
 * {@link #lvalue} rule.
 * <p>
 * This is very similar to the {@link #submenu_descriptor} rule but it is
 * much simpler.  The two rules can't be used as alternatives because they
 * are ambiguous.
 * <p> 
 * By making a separate rule for this, the logic is centralized and the
 * AST is kept simple because all tokens can be rooted at this option's
 * <code>MENU or SUB-MENU</code> keyword.
 */
menu_reference   
   :
      {
         int ntype = WID_MENU;
      }
      (KW_MENU^ | KW_SUB_MENU^ { ntype = WID_SUB_MENU; } ) menu:symbol
      {
         #menu.setType(ntype);
      }
   ;
      
/**   
 * Matches the <code>VIEW-AS</code> phrase which is called from many locations
 * such as {@link #def_var_stmt}.  Uses the following:
 * <ul>
 *    <li> {@link #combo_box_phrase}
 *    <li> {@link #simple_widget_phrase}
 *    <li> {@link #editor_phrase}
 *    <li> {@link #radio_set_phrase}
 *    <li> {@link #selection_list_phrase}
 *    <li> {@link #slider_phrase}
 * </ul>
 * <p>
 * All options can be specified in any order.
 * <p>
 * This rule updates the widget namespace with a name and type of the
 * referenced symbol.  For variables (when called from 
 * <code>def_var_stmt</code>), this means that the variable has been added to
 * both namespaces.  This allows the same name to be used for both a
 * variable and a widget and the parser must recognize the usage by context.
 */
view_as_phrase [String symName]      
   :
      {
         int ntype = -1;
      }
      KW_VIEW_AS^
      (
           combo_box_phrase                    { ntype = WID_COMBO; }
         | ntype = simple_widget_phrase
         | editor_phrase                       { ntype = WID_EDITOR; }
         | radio_set_phrase                    { ntype = WID_RADIO; }
         | selection_list_phrase               { ntype = WID_SEL_LST; }
         | slider_phrase                       { ntype = WID_SLIDER; }
      )
      {
         if (symName != null)
            sym.addWidget(symName, ntype);
      }
   ;
      
/**   
 * Matches the <code>VIEW-AS COMBO-BOX</code> phrase which is called from a
 * {@link #view_as_phrase}.  Uses the following:
 * <p>
 * <ul>
 *    <li> {@link #list_items_clause}
 *    <li> {@link #inner_lines_clause}
 *    <li> {@link #size_phrase}
 *    <li> {@link #tooltip_clause}
 *    <li> {@link #max_chars_clause}
 *    <li> {@link #auto_completion_clause}
 * </ul>
 * <p>
 * All options can be specified in any order.
 * <p>
 * ANTLR detects ambiguity but there is no real problem. Warnings are
 * disabled.
 */
combo_box_phrase      
   :
      KW_COMBO_BX^
      (
         options { generateAmbigWarnings = false; }
         :
           list_items_clause
         | inner_lines_clause
         | size_phrase
         | KW_SORT
         | tooltip_clause
         | (KW_SIMPLE | KW_DROP_DWN | KW_DROP_LST)
         | max_chars_clause
         | auto_completion_clause
      )*
   ;
      
/**   
 * Matches the simplest form of <code>VIEW-AS</code> clauses for
 * <code>FILL-IN, TEXT and TOGGLE-BOX</code>.  See {@link #view_as_phrase}.
 * Calls {@link #size_phrase} and {@link #tooltip_clause} rules.
 * <p>
 * ANTLR detects ambiguity but there is no real problem. Warnings are
 * disabled.
 * <p>
 * This rule must return back a token type value to tell the caller which
 * widget was specified, since it can match 3 different widgets and the
 * caller must update the widget namespace.
 * <p>
 * By making a separate rule for this, the logic is centralized and the
 * AST is kept simple because all tokens can be rooted at this option's
 * keyword.
 * 
 * @return    The token type of the matched widget name.
 */
simple_widget_phrase returns [int ntype = -1] 
   :
      (
           KW_FILL_IN^      { ntype = WID_FILL_IN; }
         | KW_TEXT^         { ntype = WID_TEXT; }
         | KW_TOGGL_BX^     { ntype = WID_TOGGLE; }
      )
      (
         options { generateAmbigWarnings = false; }
         :
           KW_NATIVE
         | size_phrase
         | tooltip_clause
      )* 
   ;
      
/**   
 * Matches the <code>VIEW-AS EDITOR</code> phrase which is called from a {@link #view_as_phrase}.
 * <p>
 * Under common conditions, the 4GL requires there be a size phrase or inner clause. However,
 * there is an undocumented feature that allows the size phrase/inner clause to be missing
 * SO LONG AS THE WIDGET HAS THEM SPECIFIED ELSEWHERE (e.g. in a DEFINE VARIABLE). This was
 * found in customer source code.  In other words, there must have been a previous VIEW-AS
 * EDITOR clause with the proper minimum setup. We are not checking for that here.  Instead,
 * we are just loosening up the options processing to make it non-mandatory.
 * <p>
 * Uses the following:
 * <p>
 * <ul>
 *    <li> {@link #size_phrase}
 *    <li> {@link #inner_chars_and_lines_clause}
 *    <li> {@link #buffer_chars_or_lines_clause}
 *    <li> {@link #max_chars_clause}
 *    <li> {@link #tooltip_clause} 
 * </ul>
 * <p>
 * ANTLR detects ambiguity but there is no real problem. Warnings are disabled.
 * <p>
 * All options can be specified in any order.
 */
editor_phrase      
   :
      KW_EDITOR^ 
      ( 
         options { generateAmbigWarnings = false; }
         :
           buffer_chars_or_lines_clause
         | KW_LARGE
         | max_chars_clause
         | KW_NO_BOX
         | KW_NO_WRAP
         | KW_SCROLL_H
         | KW_SCROLL_V
         | tooltip_clause
         | (size_phrase | inner_chars_and_lines_clause)
      )*
   ;
      
/**   
 * Matches the <code>VIEW-AS RADIO-SET</code> phrase which is called from a
 * {@link #view_as_phrase}. Uses the following:
 * <p>
 * <ul>
 *    <li> {@link #size_phrase}
 *    <li> {@link #radio_buttons_clause}
 *    <li> {@link #tooltip_clause} 
 * </ul>
 * <p>
 * ANTLR detects ambiguity but there is no real problem. Warnings are
 * disabled.
 * <p>
* All options can be specified in any order.
 */
radio_set_phrase      
   :
      KW_RADIO_S^ 
      ( 
         options { generateAmbigWarnings = false; }
         :
           (
                KW_HORIZ 
                (
                   options { generateAmbigWarnings = false; }
                   :
                   KW_EXPAND
                )?
              | KW_VERT
           )
         | size_phrase
         | radio_buttons_clause
         | tooltip_clause
      )*
   ;
/**   
 * Matches the <code>VIEW-AS SELECTION-LIST</code> phrase which is called
 * from a {@link #view_as_phrase}.  Uses the following:
 * <p>
 * <ul>
 *    <li> {@link #list_items_clause}
 *    <li> {@link #size_phrase}
 *    <li> {@link #inner_chars_and_lines_clause} 
 *    <li> {@link #tooltip_clause}
 * </ul>
 * <p>
 * ANTLR detects ambiguity but there is no real problem. Warnings are
 * disabled.
 * <p>
 * All options can be specified in any order.
 */
selection_list_phrase      
   :
      KW_SEL_LST^
      (
         options { generateAmbigWarnings = false; }
         :
           (KW_SINGLE | KW_MULTIPLE)
         | KW_NO_DRAG
         | list_items_clause
         | KW_SCROLL_H
         | KW_SCROLL_V         
         | (size_phrase | inner_chars_and_lines_clause)         
         | KW_SORT
         | tooltip_clause
      )*
   ;
      
/**   
 * Matches the <code>VIEW-AS SLIDER</code> phrase which is called
 * from a {@link #view_as_phrase}.  Uses the following:
 * <p>
 * <ul>
 *    <li> {@link #max_and_min_value_clause}
 *    <li> {@link #tic_marks_clause}
 *    <li> {@link #size_phrase}
 *    <li> {@link #tooltip_clause}
 * </ul>
 * <p>
 * ANTLR detects ambiguity but there is no real problem. Warnings are
 * disabled.
 * <p>
 * All options can be specified in any order.
 */
slider_phrase      
   :
      KW_SLIDER^
      (
         options { generateAmbigWarnings = false; }
         :
           (KW_HORIZ | KW_VERT)
         | KW_NO_CUR_V
         | KW_LG_2_SM
         | max_and_min_value_clause
         | tic_marks_clause 
         | size_phrase         
         | tooltip_clause
      )*
   ;
      
/**   
 * Matches the <code>MAX-VALUE num</code> or <code>MIN-VALUE num</code>
 * clauses as used in a <code>VIEW-AS SLIDER</code> phrase.  See
 * {@link #view_as_phrase} and {@link #slider_phrase}.
 * <p>
 * By making a separate rule for this, the logic is centralized and the
 * AST is kept simple because all tokens can be rooted at this option's
 * keyword.
 */
max_and_min_value_clause      
   :
      (KW_MAX_VAL^ | KW_MIN_VAL^) NUM_LITERAL
   ;
         
/**   
 * Matches the <code>TIC-MARKS</code> clause as used in a 
 * <code>VIEW-AS SLIDER</code> phrase.  See {@link #view_as_phrase}
 * and {@link #slider_phrase}.
 * <p>
 * ANTLR detects ambiguity but there is no real problem. Warnings are
 * disabled.
 * <p>
 * By making a separate rule for this, the logic is centralized and the
 * AST is kept simple because all tokens can be rooted at this option's
 * keyword.
 */
tic_marks_clause         
   :
      KW_TIC_MARK^
      (KW_NONE | KW_TOP | KW_BOTTOM | KW_LEFT | KW_RIGHT | KW_BOTH)
      (
         options { generateAmbigWarnings = false; }
         :
         KW_FREQ NUM_LITERAL
      )?
   ;
   
/**   
 * Matches the <code>RADIO-BUTTONS</code> clause as used in a 
 * <code>VIEW-AS RADIO-SET</code> phrase.  See {@link #view_as_phrase}
 * and {@link #radio_set_phrase}.
 * <p>
 * By making a separate rule for this, the logic is centralized and the
 * AST is kept simple because all tokens can be rooted at this option's
 * keyword.
 */
radio_buttons_clause      
   :
      KW_RADIO_B^ STRING COMMA! literal
      (
         options { generateAmbigWarnings = false; }
         :
         COMMA! STRING COMMA! literal
      )*
   ;
      
/**   
 * Matches the <code>INNER-CHARS chars INNER-LINES lines</code> clause in an
 * arbitrary order (either keyword will be matched first, which is an
 * undocumented feature.  Although the documentation states both must be
 * present to complete the match, working 4GL code has also shown that it is
 * valid to match only one of the two. As another undocumented feature, the
 * second keyword and literal has been made optional.  This causes some
 * ambiguity to be detected, but the usage is fine.  For this reason, the
 * warning has been disabled.
 * <p>
 * See {@link #inner_chars_and_lines_core}
 */
inner_chars_and_lines_clause      
   :
      inner_chars_and_lines_core
      (
         options { generateAmbigWarnings = false; }
         :
         inner_chars_and_lines_core
      )?
   ;
      
/**   
 * Matches the <code>INNER-CHARS chars or INNER-LINES lines</code> clause as
 * used in a <code>VIEW-AS EDITOR</code> phrase.  See {@link #view_as_phrase}
 * and {@link #editor_phrase}.
 * <p>
 * By making a separate rule for this, the logic is centralized and the
 * AST is kept simple because all tokens can be rooted at this option's
 * keyword.
 */
inner_chars_and_lines_core      
   :
      (KW_INNER_C^ | KW_INNER_L^) NUM_LITERAL
   ;
      
      
/**   
 * Matches the <code>BUFFER-CHARS chars and BUFFER-LINES lines</code> clause
 * as used in a <code>VIEW-AS EDITOR</code> phrase.  See
 * {@link #view_as_phrase} and {@link #editor_phrase}.
 * <p>
 * By making a separate rule for this, the logic is centralized and the
 * AST is kept simple because all tokens can be rooted at this option's
 * keyword.
 */
buffer_chars_or_lines_clause      
   :
      (KW_BUFFER_C^ | KW_BUFFER_L^) NUM_LITERAL
   ;
      
/**   
 * Matches the <code>TOOLTIP</code> clause as used in a phrase from a
 * <code>VIEW-AS</code> construct.  See {@link #view_as_phrase}.
 * <p>
 * By making a separate rule for this, the logic is centralized and the
 * AST is kept simple because all tokens can be rooted at this option's
 * keyword.
 */
tooltip_clause      
   :
      // literal allows the unknown value
      KW_TOOLTIP^ literal
   ;
   
/**   
 * Matches the <code>LIST_ITEMS</code> and <code>LIST_ITEM_PAIRS</code> clauses as used in a
 * phrase from a <code>VIEW-AS COMBO-BOX</code> or <code>VIEW-AS SELECTION-LIST</code> construct.
 * See {@link #combo_box_phrase} or {@link #selection_list_phrase}.
 * <p>
 * By making a separate rule for this, the logic is centralized and the AST is kept simple
 * because all tokens can be rooted at this option's keyword.
 */
list_items_clause   
   :
      // the list-item-pairs form of this is supposed to only accept an even number of entries,
      // but the 4GL compiler silently accepts any number (including just a single empty string)
      // at runtime, the use of such a widget will generate an error 4041
      (KW_LIST_ITM^ | KW_LST_PAIR^) literal 
      (
         options { generateAmbigWarnings = false; }
         :
         COMMA! literal
      )*
   ;
   
/**   
 * Matches the <code>INNER-LINES</code> clause as used in a phrase from a
 * <code>VIEW-AS COMBO-BOX</code> construct.  See {@link #combo_box_phrase}.
 * <p>
 * By making a separate rule for this, the logic is centralized and the
 * AST is kept simple because all tokens can be rooted at this option's
 * keyword.
 */
inner_lines_clause   
   :
      KW_INNER_L^ NUM_LITERAL
   ;
   
/**   
 * Matches the <code>MAX-CHARS</code> clause as used in a phrase from a
 * <code>VIEW-AS COMBO-BOX</code> construct.  See {@link #combo_box_phrase}.
 * <p>
 * By making a separate rule for this, the logic is centralized and the
 * AST is kept simple because all tokens can be rooted at this option's
 * keyword.
 */
max_chars_clause   
   :
      KW_MAX_CHAR^ NUM_LITERAL
   ;
   
/**   
 * Matches the <code>AUTO-COMPLETION</code> clause as used in a phrase from a
 * <code>VIEW-AS COMBO-BOX</code> construct.  See {@link #combo_box_phrase}.
 * <p>
 * ANTLR detects ambiguity but there is no real problem. Warnings are
 * disabled.
 * <p>
 * By making a separate rule for this, the logic is centralized and the
 * AST is kept simple because all tokens can be rooted at this option's
 * keyword.
 */
auto_completion_clause   
   :
      KW_AUTO_COM^
      (
         options { generateAmbigWarnings = false; }
         :
         KW_UNIQ_MAT
      )?
   ;
   
   
/**   
 * Matches the <code>CLEAR</code> language statement.
 * <p>
 * Used by {@link #stmt_list}.
 */   
clear_stmt   
   :
      KW_CLEAR^ (frame_reference[false])? (KW_ALL)? (KW_NO_PAUSE)? DOT!
   ;
   
/**   
 * Matches the <code>NEXT and LEAVE</code> language statements.  Calls the
 * {@link #label_reference} to resolve the optional label.
 * <p>
 * Used by {@link #stmt_list}.
 */   
next_or_leave_stmt   
   :
      // label_reference may match nothing (empty alt) so it is non-optional
      // here even though it is really optional
      (KW_NEXT^ | KW_LEAVE^) label_reference DOT!
   ;   
   
/**
 * Matches a <code>DEFINE VARIABLE</code> Progress 4GL language statement.
 * This is the statement that defines or imports new user-named variables.  
 * The {@link #symbol} rule is used to force the token type for the variable
 * name to be a <code>SYMBOL</code>.  Then this name is passed as a parameter
 * to the {@link #as_clause} or {@link #like_clause} rules, which add this
 * variable to the variable dictionary associating it with the correct data 
 * type.
 * <p>
 * This rule is called by {@link #define_stmt} which is where the matching
 * on the <code>DEFINE</code> keyword and the optional shared scope for
 * variables, temp-tables and streams.  This rule supports shared, global
 * shared or local variables and the syntax for both declaring or defining
 * (<code>NEW</code> keyword) shared variables.  HOWEVER, all the matching
 * for these optional keywords is done in the <code>define_stmt</code> rule. 
 * This refactoring allows all <code>DEFINE</code> statements to be supported
 * without ambiguity.
 * <p>
 * Multiple options are supported in any order (which is how Progress handles
 * these), however this code does not check for a repeated definition of
 * the options.
 * <p>
 * This rule implements all logic needed to define and initialize array
 * variables (using the <code>EXTENT</code>, <code>INITIAL [ ,,, ]</code>
 * and <code>LABEL string, string...</code> syntax.  These option rules are
 * all implemented as separate rules to force a clean tree structure for 
 * each option rather than accepting a flat structure where all options and
 * any associated data are all created as siblings.  Instead, each option
 * subtree is created under its associated keyword's token and this is a
 * child of the <code>DEFINE</code> keyword.
 *
 * @param    allowColon
 *           Set to <code>true</code> on return to tell the calling code to match either
 *           a DOT or a COLON to end the statement (this will be set only if a triggers
 *           phrase is matched).
 *
 * @return   The variable name that is being defined.
 */
def_var_stmt [boolean[] allowColon] returns [String varname = null]
   :
      {
         boolean view_as_triggered = false;
         boolean handle            = false;
         boolean object            = false;
         int     ftype             = -1;
         int     size              = -1;
         boolean had_as_clause     = false;
      }
      KW_VAR!
      s:symbol  { varname = #s.getText(); }
      (
           a:as_clause[varname, false]
           { 
              ftype = #a.getFirstChild().getType();
              
              if (ftype == KW_HANDLE   ||
                  ftype == KW_WID_HAND ||
                  ftype == KW_COM_HNDL)
              {
                 handle = true;
              }
              
              if (ftype == CLASS_NAME)
              {
                 object = true;
              }
              
              if (had_as_clause)
              {
                 System.out.println(
                    String.format("Multiple AS clauses for the [%s] var definition at " + 
                                  "line %d column %d", 
                                  varname, #a.getLine(), #a.getColumn()));
              }
              else
              {
                 had_as_clause = true;
              }
           }
         | l:like_clause[varname, null, false]
           {
              ftype = #l.getFirstChild().getType();
              
              if (ftype == VAR_HANDLE     ||
                  ftype == VAR_COM_HANDLE ||
                  ftype == FIELD_HANDLE   ||
                  ftype == FIELD_COM_HANDLE)
              {
                 handle = true;
              }
              
              if (ftype == CLASS_NAME)
              {
                 object = true;
              }
           }
           
            // no annotation needs to be saved here, as the caller (define_stmt) will call
            // Variable.annotateVariableOptions which will take care of saving the extent, if
            // present
         | size = extent
         
           // the 4GL requires the AS/LIKE clause to appear before this, so the type will
           // be available
         | initializer[ftype]
         | label
         | ui_stuff
         | format_string[false]
         | column_label
         | decimals_clause
         | KW_DROP_TAR
         | case_sensitive
         | KW_NO_UNDO
         | view_as_phrase[varname]        { view_as_triggered = true; }
         | trigger_phrase  { allowColon[0] = true; }
      )*
      {
         // if no explicit override occurred AND if the type is not a handle AND not an object
         // add a fill-in widget to the widget namespace, with the same name (handle,
         // com-handle and object lvalues cannot be frame fields)
         // TODO: do we need to do this with RAW, MEMPTR types too?
         if (view_as_triggered == false && !handle && !object)
         {
            sym.addWidget(varname, WID_FILL_IN);
         }         
      }
   ;

/**
 * Matches a <code>DEFINE PROPERTY</code> Progress 4GL language statement.
 * This is the statement that defines new user-named properties for the
 * currently active class definition.
 * <p> 
 * The {@link #any_symbol_at_all} rule is used to force the token type for the
 * property name to be a <code>SYMBOL</code>. That means property names can
 * be reserved keywords. Then this name is passed as a parameter
 * to the {@link #as_clause} rule, which adds this property to the dictionary
 * for the class definition, associating it with the correct data type.
 * <p>
 * This rule is called by {@link #define_stmt} which is where the matching
 * on the <code>DEFINE</code> keyword and the optional shared scope for
 * variables, temp-tables and streams.  This rule supports an optional
 * {@link #access_mode}.  HOWEVER, all the matching for these optional
 * keywords is done in the <code>define_stmt</code> rule. This refactoring 
 * allows all <code>DEFINE</code> statements to be supported without
 * ambiguity.
 * <p>
 * Multiple options are supported in any order (which is how Progress handles
 * these), however this code does not check for a repeated definition of
 * the options.
 * <p>
 * The option sub-rules are all implemented as separate rules to force a
 * clean tree structure for each option rather than accepting a flat
 * structure where all options and any associated data are all created as
 * siblings.  Instead, each option subtree is created under its associated
 * keyword's token and this is a child of the <code>DEFINE</code> keyword.
 */
def_prop_stmt returns [String varname = null]
   :
      {
         int ftype = -1;
         int size  = -1;
      }
      KW_PROPERTY!
      s:any_symbol_at_all  { varname = #s.getText(); }
      (
           a:as_clause[varname, false]
           { 
              ftype = #a.getFirstChild().getType();
           }
           
           // no annotation needs to be saved here, as the caller (define_stmt) will call
           // Variable.annotateVariableOptions which will take care of saving the extent, if
           // present
         | size = extent
         | initializer[ftype]
         | KW_NO_UNDO
      )*
      getter_setter
      (
         options { generateAmbigWarnings = false; }
         :
         DOT! getter_setter
      )?
   ;
   
/**
 * Matches the <code>GET</code> or <code>SET</code> block of the
 * {@link #def_prop_stmt}. Uses {@link #access_mode}, {@link #param_list_definition},
 * {@link #block} and {@link #end_stmt}.
 */
getter_setter
   :
      (access_mode)?
      (KW_GET^ | KW_SET^)
      (
         options { generateAmbigWarnings = false; }
         :
         
         // parameters are optional
         (param_list_definition[false, false])?
         
         // DOT can't be used here (this solves a major ambiguity problem and is how the
         // 4GL does it)
         COLON!
         
         b:oo_block
         (
           options { generateAmbigWarnings = false; }
           :
           end_stmt[false]
         )?
      )?
   ;
   
/**
 * Matches a <code>DEFINE EVENT</code> Progress 4GL language statement.
 * This is the statement that defines new user-named events for the
 * currently active class definition.
 * <p> 
 * The {@link #any_symbol_at_all} rule is used to force the token type for the
 * event name to be a <code>SYMBOL</code>.  Then this name is passed added
 * to the dictionary for the class definition, associating it with the correct
 * data type.
 * <p>
 * This rule is called by {@link #define_stmt} which is where the matching
 * on the <code>DEFINE</code> keyword and the optional shared scope for
 * variables, temp-tables and streams.  This rule supports an optional
 * {@link #access_mode}.  HOWEVER, all the matching for these optional
 * keywords is done in the <code>define_stmt</code> rule. This refactoring 
 * allows all <code>DEFINE</code> statements to be supported without
 * ambiguity.
 * <p>
 * Multiple options are supported in any order (which is how Progress handles
 * these), however this code does not check for a repeated definition of
 * the options.
 * <p>
 * The option sub-rules are all implemented as separate rules to force a
 * clean tree structure for each option rather than accepting a flat
 * structure where all options and any associated data are all created as
 * siblings.  Instead, each option subtree is created under its associated
 * keyword's token and this is a child of the <code>DEFINE</code> keyword.
 */
def_event_stmt returns [String varname = null]
   :
      KW_EVENTS!
      s:any_symbol_at_all
      {
         varname = #s.getText();
         
         // here is where we enable the variable lookup for all subsequent
         // code in this scope, note that Progress only seems to ever have
         // 2 levels of scope: the external proc and any internal proc/
         // trigger defs, since you can only define shared or global shared
         // vars in the external proc scope, we should *not* need to
         // explicitly add vars to the global scope since adding to the
         // current scope should be equivalent
         sym.addLocalVariable(varname, CLASS_EVENT, null);
      }
      (
         options { generateAmbigWarnings = false; }
         :
         event_signature | dotnet_delegate
      ) 
   ;
   
/**
 * Match the specification of an event handler signature for a 4GL procedure or method that
 * will handle the associated event.  Called from {@link #def_event_stmt}.  Matches an optional
 * <code>KW_SIGNATUR</code>, then a <code>KW_VOID</code> followed by a
 * {@link #param_list_definition}.  The result is rooted at an <code>EVENT_SIGNATURE</code>
 * node. Separated from the main rule in order to build the tree properly.
 */
event_signature
   :
      { astFactory.makeASTRoot(currentAST, #[EVENT_SIGNATURE, "signature"]); }
      (KW_SIGNATUR!)? KW_VOID param_list_definition[false, false]
   ;

/**
 * Match the specification of an event handler as a .NET delegate class name
 * that will handle the associated event.  Called from {@link #def_event_stmt}.
 * Matches an optional <code>KW_DELEGATE</code>, then an optional
 * <code>KW_CLASS</code> followed by an {@link #user_defined_type_name}.  The
 * result is rooted at an <code>DOTNET_DELEGATE</code> node. Separated from the
 * main rule in order to build the tree properly.
 */
dotnet_delegate
   :
      { astFactory.makeASTRoot(currentAST, #[DOTNET_DELEGATE, "delegate"]); }
      (
         options { generateAmbigWarnings = false; }
         :
         KW_DELEGATE!
      )?
      (KW_CLASS!)? user_defined_type_name
   ;
   
/**
 * Matches the <code>EXTENT</code> keyword and the optional integer constant
 * that follows it.  Used by {@link #def_var_stmt} and the subtree is rooted
 * by the <code>EXTENT</code> keyword.
 */
extent returns [int size = -1]
   :
      KW_EXTENT^ { size = -1; }
      (
         options { generateAmbigWarnings = false; }
         :
         c:NUM_LITERAL { size = Integer.parseInt(c.getText()); }
      )?
   ;
   
/**
 * Matches the <code>INITIAL</code> keyword and the required single literal
 * or a left/right bracket bounded (and comma separated) list of literals.
 * Used by {@link #def_var_stmt} and the subtree is rooted by the 
 * <code>INITIAL</code> keyword.  Note that the commas are dropped from the
 * resulting AST to simplify processing.
 * <p>
 * A special case exists for the <code>TODAY</code> keyword. Normally,
 * an initializer would only accept a {@link #literal} but for variables
 * of type <code>DATE</code>, the <code>TODAY</code> 'function' can also
 * be used.  To accomodate this requirement, the {@link #initializer_constant} 
 * rule is called.
 * <p>
 * No testing is done to check the type(s) of the literal(s) against the 
 * expected type as set by the AS or LIKE clause.
 *
 * @param    vtype
 *           A token type that describes type of the variable being initialized. This is
 *           the type of the child of the <code>KW_AS</code> or <code>KW_LIKE</code> node.
 */
initializer [int vtype]   
   :
      KW_INIT^
      (
           initializer_constant[vtype]
         | LBRACKET!
           initializer_constant[vtype] (COMMA! initializer_constant[vtype])*
           RBRACKET!
      )   
   ;
   
/**
 * Matches a {@link #literal} or an {@link #lvalue} when the 
 * <code>TODAY</code> or <code>NOW</code> keywords are encountered. Handles a
 * special case for {@link #initializer} but instead of resolving to a 
 * <code>FUNC_DATE</code> or <code>FUNC_DATETIME</code> we override the type
 * to <code>DATE_LITERAL</code>, <code>DATETIME_LITERAL</code> or <code>DATETIME_TZ_LITERAL</code>
 * which is consistent with the usage of this as a constant.
 * <p>
 * Due to nastiness in Progress, there is an undocumented 'feature' that
 * allows one to specify a non-string literal as a string.  This is a problem
 * when one needs to find a token of one type and finds a <code>STRING</code>
 * instead.  For this reason, all callers now provide a state parameter that
 * determines if any possible fixups are needed.  If so, the init rule
 * rewrites the token properly.
 * <p>
 * <b>WARNING: in combination with the above feature, if a logical value is
 * initialized with a string, the format string (implicitly set to 'yes/no'
 * and 'true/false' or explicitly set via <code>LIKE or FORMAT</code>) will
 * determine how such strings will be converted to
 * <code>BOOL_TRUE or BOOL_FALSE</code>.  Since the format string may or may
 * not be known at this point in the parsing, the result cannot be determined  
 * in this rule.  In this case the original <code>STRING</code> is left
 * behind and will be handled by our 'boolean manhandling' processing later.
 * </b>
 *
 * @param    vtype
 *           A token type that describes type of the variable being initialized. This is
 *           the type of the child of the <code>KW_AS</code> or <code>KW_LIKE</code> node.
 */
initializer_constant [int vtype]   
   :
      {
         boolean string  = isCharType(vtype);
         boolean date    = isDateType(vtype);
         boolean numeric = isNumericType(vtype);
      
         // character need this manhandling too
         if (string)
         {
            Token   content = LT(1);
            if (content.getType() != STRING && content.getType() != UNKNOWN_VAL)
            {
               // wrap this in quotes
               String text = "\"" + content.getText() + "\"";
               content.setText(text);
               content.setType(STRING);
            }
         }
         else
         {
            Token   content = LT(1);
            int     oldtype = -1;
         
            // the SYMBOL case will happen in the "boolean manhandling" situation
            // only handle the STRING case here
            if (content.getType() == STRING)
            {
               // OK, we need to re-lex!
               Token fixed = relex(content.getText());
               
               // nulls should only occur if the input is the empty string
               if (fixed == null)
               {
                  // empty string is the same as unknown value in a 4GL date init
                  if (date)
                  {
                     content.setType(UNKNOWN_VAL);
                     content.setText("?");
                  }
                  else if (numeric)
                  {
                     // empty string is the same as the 0 value in a 4GL numeric init
                     content.setType(NUM_LITERAL);
                     content.setText("0");
                  }
               }
               else
               {
                  int t = fixed.getType();
                  
                  // only update the type and text if we found a new match,
                  // otherwise defer the fix to later...
                  if (t == UNKNOWN_VAL         || t == BOOL_TRUE    ||
                      t == BOOL_FALSE          || t == NUM_LITERAL  ||
                      t == DEC_LITERAL         || t == DATE_LITERAL ||
                      t == DATETIME_LITERAL    || t == KW_TODAY     ||
                      t == DATETIME_TZ_LITERAL || t == KW_NOW)
                  {               
                     content.setText(fixed.getText());
                     content.setType(fixed.getType());
                  }
               }
            }
         }
         
         // date types allow matching of the text that in other known contexts would
         // be a decimal literal, we must rewrite those cases here
         if (date)
         {
            Token next = LT(1);
            
            if (next.getType() == DEC_LITERAL)
            {
               next.setType(DATE_LITERAL);
            }
         }
      }
      (
           literal
         | { LA(1) == KW_TODAY || LA(1) == KW_NOW }?
           l:lvalue
           {
              // don't use saveAndReplaceType() here because we already have
              // the KW_TODAY/KW_NOW as our "oldtype" annotation and we are
              // just replacing the FUNC_* type that is the current type with
              // the literal type which eases our conversion effort in this
              // case since this is the only case where Progress treats these
              // "functions" as literals
              if (#l.getType() == FUNC_DATE)
              {
                 // today
                 #l.setType(DATE_LITERAL);
              }
              else if (#l.getType() == FUNC_DATETIME_TZ)
              {
                 // now
                 #l.setType(DATETIME_TZ_LITERAL);
              }
           }
      )
   ;
   
/**
 * Matches the <code>LABEL</code> keyword and the required string literal
 * or comma separated list of string literals. The subtree is rooted by the
 * <code>KW_LABEL</code> keyword.  Note that the commas are dropped from the
 * resulting AST to simplify processing.
 * <p>
 * Used by {@link #def_var_stmt}, {@link #def_parm_stmt} and
 * {@link #def_buf_stmt}.  Once added in all these locations, there was
 * a warning that there was ambiguity in the subrule, but this is a
 * bogus warning.  This warning has been disabled.
 * <p>
 * The list syntax is used to create different labels for each element in
 * an array.
 */
label   
   :
      KW_LABEL^ STRING
      (
         options { generateAmbigWarnings = false; }
         :
         COMMA! STRING
      )*
   ;
   
/**
 * Matches the <code>AS</code> language keyword that defines the data type
 * of a variable to be added to the variable dictionary. This rule 
 * encapsulates the logic of matching the correct variable data type and
 * inserting it into the variable dictionary in the correct scope. This is
 * called from {@link #def_var_stmt} and {@link #parameter}, both of which
 * support the same syntax.  It is also called from {@link #def_parm_stmt}  
 * which does allow for an optional <code>HANDLE TO</code> construct when being 
 * called for a <code>DEFINE PARAMETER</code>. 
 * <p>
 * The data type matching is implemented in {@link #var_type}.
 *
 * @param    varName
 *           Text which is the variable name that should be added to the
 *           current scope of the variable dictionary.  If <code>null</code>,
 *           no name will be added and this clause is essentially ignored.
 * @param    nativeTypes
 *           <code>true</code> to support extra type keywords for native
 *           shared library interfaces.
 */
as_clause [ String varName, boolean nativeTypes ]   
   :
      KW_AS^
      (
         // this is required to disambiguate when this rule is called
         // from the format_phrase since that phrase can have a
         // (KW_TO NUM_LITERAL) following an as_clause
         { (nativeTypes && LA(3) != NUM_LITERAL) }?
         ht:KW_HANDLE { #ht.setType(HANDLE_TO); } KW_TO!
      )?
      var_type[ varName, nativeTypes ]
      {
         // move the HANDLE_TO node to be the last child, since otherwise the downstream
         // tree processing will have to have many extra checks and conditional logic to
         // deal with this rare case (the common case is to just have a single child that
         // specifies the type)
         if (#ht != null)
         {
            #ht.move(null, -1);
         }
      }
   ;

/**
 * Matches the variable type language keyword in {@link #as_clause} or
 * directly in {@link #parameter}.
 * <p>
 * This design allows the logic to be centralized and reused from multiple 
 * callers.  To do this, this rule requires a single parameter which is the
 * variable name to be added to the dictionary. This is necessary since there
 * is no way to implement a standard match for the symbol name from all
 * language statements that call this rule.
 * <p>
 * The inclusion of the {@link #reserved_or_symbol} rule in the second token
 * position ensures that all possible symbols are considered a possible data
 * type and will be matched by this rule.  This means that the calling rules
 * include a list of reserved and unreserved keywords or generic symbols as
 * one of the options for its 2nd token, thus &quot;pulling&quot; the
 * recursive descent processing down into this rule.
 * <p>
 * It is important to note that in the end, it is not really valid to have
 * a token type of <code>SYMBOL</code>, random keyword match or anything
 * other than one of the valid <b>special abbreviations</b> for a data type.
 * What we really require is a token that is one of the following:
 * <pre>
 *    Data Type Keyword        Actual Minimum Abbrev      Lexer Abbrev
 *    -----------------        ---------------------      -------------
 *    KW_BLOB                  not supported as a variable
 *    KW_CHAR                  c                          char
 *    KW_CLASS                 n/a                        n/a
 *    KW_CLOB                  not supported as a variable
 *    KW_COM_HNDL              n/a                        n/a
 *    KW_DATE                  da                         n/a
 *    KW_DATETIME              n/a                        n/a
 *    KW_DATE_TZ               n/a                        n/a
 *    KW_DEC                   de                         dec
 *    KW_HANDLE                handle                     n/a
 *    KW_INT                   i                          int
 *    KW_INT64                 n/a                        n/a
 *    KW_LOGICAL               l                          n/a
 *    KW_LONGCHAR              n/a                        n/a
 *    KW_MEMPTR                m                          mem
 *    KW_RAW                   ra                         n/a
 *    KW_RECID                 re                         n/a
 *    KW_ROWID                 row                        n/a
 *    KW_WID_HAND              widg                       widget-h
 *    The following are supported in native interfaces:
 *    KW_BYTE                  n/a                        n/a
 *    KW_DOUBLE                n/a                        n/a
 *    KW_FLOAT                 n/a                        n/a
 *    KW_LONG                  n/a                        n/a
 *    KW_SHORT                 n/a                        n/a
 *    KW_UNS_SHRT              n/a                        n/a
 * </pre>
 * <p>
 * The difference between a <code>SYMBOL</code> and a <code>KW_CHAR</code>
 * can only be determined by context (the fact that the lookahead code has
 * identified a match with a <code>SYMBOL</code> token which is a construct 
 * that can occur in an this rule).  Thus <b>if</b> a match is found using
 * lookahead (in the rule that calls this one), then it <b>must</b> be
 * correct and necessary to convert the <code>SYMBOL</code> or other keyword
 * token type into the correct keyword representing the data type, using a
 * special abbreviation test.  This is done in an init action that is 
 * processed at the beginning of this rule. In this way, the rest of the rule
 * only sees the 2nd token as a data token type and never tries to
 * call the <code>reserved_or_symbol</code> rule since calling that rule 
 * would actually try to match (and consume from the token stream) a
 * <code>SYMBOL</code> or other keyword token type.  <b>Specifying a rule
 * reference to obtain its value from a lookahead perspective but then
 * changing the token's type such that the rule reference never gets called, 
 * is the critical trick that enables context sensitive symbol names!</b>
 * <p>
 * This use of a pull rule is only necessary because Progress provides
 * additional special abbreviation support in this construct. This is a set
 * of reduced abbreviations (see table above) beyond the normal (documented
 * or undocumented) abbreviations of a data type keyword as is possible 
 * elsewhere in Progress source code.  The <code>reserved_or_symbol</code> 
 * is needed because some possible reserved keyword matches could occur 
 * that would match one of these special abbreviations and thus the
 * <code>symbol</code> rule would not be sufficient to pull all possible
 * valid keyword matches into this rule.
 * <p>
 * The init action reads the text of the 2nd token and switches the token
 * type to that of the matching data type keyword, if any special
 * abbreviation is found.  All such processing is bypassed if the lexer
 * has already matched a valid data type keyword.
 * <p>
 * The logic of the special abbreviation processing is centralized in the
 * specialDataTypeKeywordAbbreviations() method.
 * <p>
 * Note that this rule does support the addition of scoped shared and global
 * shared variables.  No special steps are necessary to enable this due to
 * a quirk in how Progress handles such variables, they can only be defined
 * in the top-level context (not inside a procedure or function which is the
 * only block in which there is a different scoping level).  For this reason,
 * whenever shared variables are added to the dictionary, they can be added
 * to the current scope, since this is the same as the global scope in any
 * valid Progress code.
 *
 * @param    varName
 *           Text which is the variable name that should be added to the
 *           current scope of the variable dictionary.  If <code>null</code>,
 *           no name will be added and this clause is essentially ignored.
 * @param    nativeTypes
 *           <code>true</code> to support extra type keywords for native
 *           shared library interfaces.
 */
var_type [ String varName, boolean nativeTypes ]
   :
      {
         int vartype = -1;
         specialDataTypeKeywordAbbreviations(1);
      }   
      ( 
         options { generateAmbigWarnings = false; }
         :
           KW_CHAR                              { vartype = VAR_CHAR; }
         | KW_COM_HNDL                          { vartype = VAR_COM_HANDLE; }
         | KW_DATE                              { vartype = VAR_DATE; }
         | KW_DATETIME                          { vartype = VAR_DATETIME; }
         | KW_DATE_TZ                           { vartype = VAR_DATETIME_TZ; }                            
         | KW_DEC                               { vartype = VAR_DEC; }
         | (KW_HANDLE | KW_WID_HAND)            { vartype = VAR_HANDLE; }
         | KW_INT                               { vartype = VAR_INT; }
         | KW_INT64                             { vartype = VAR_INT64; }             
         | KW_LOGICAL                           { vartype = VAR_LOGICAL; }
         | KW_LONGCHAR                          { vartype = VAR_LONGCHAR; }
         | KW_RECID                             { vartype = VAR_RECID; }
         | KW_ROWID                             { vartype = VAR_ROWID; }
         | KW_RAW                               { vartype = VAR_RAW; }
         | KW_MEMPTR                            { vartype = VAR_MEMPTR; }
         
           // support the native types if necessary
         | { nativeTypes }?
           (
                KW_BYTE                         { vartype = VAR_BYTE; }
              | KW_DOUBLE                       { vartype = VAR_DOUBLE; }
              | KW_FLOAT                        { vartype = VAR_FLOAT; }
              | KW_LONG                         { vartype = VAR_LONG; }
              | KW_SHORT                        { vartype = VAR_SHORT; }
              | KW_UNS_LONG                     { vartype = VAR_ULONG; }
              | KW_UNS_SHRT                     { vartype = VAR_USHORT; }
           )
         
           // this is for "pull" and to match the class type (it will
           // pull for the LA2 and LA3 because of the optional CLASS
           // keyword
         | (
              options { generateAmbigWarnings = false; }
              :
              KW_CLASS!
           )?
           nam:user_defined_type_name           { vartype = VAR_CLASS; }
      )
      {
         if (varName != null)
         {
            String cls = (#nam == null) ? null
                                        : (String) #nam.getAnnotation("qualified");
         
            // here is where we enable the variable lookup for all subsequent
            // code in this scope, note that Progress only seems to ever have
            // 2 levels of scope: the external proc and any internal proc/
            // trigger defs, since you can only define shared or global shared
            // vars in the external proc scope, we should *not* need to
            // explicitly add vars to the global scope since adding to the
            // current scope should be equivalent
            sym.addLocalVariable(varName, vartype, cls);
         }
      }   
   ;   
   
/**
 * Matches the <code>LIKE</code> language keyword that defines the data type
 * of a variable to be added to the variable dictionary based on the type of
 * another variable, database field, temp-table/work-table field. This rule 
 * encapsulates the logic of matching the correct data type and inserting it
 * into the proper dictionary. This is called from many locations (see below)
 * which all handle the same syntax but for different purposes.  This allows
 * the matching logic to be centralized and reused from multiple callers.
 * <p>
 * If this is processing for a temp-table, then instead of <code>LIKE</code> the
 * <code>LIKE-SEQUENTIAL</code> keyword can be matched instead. 
 * <p>
 * To know what to do with each different type of dictionary add, this rule
 * requires 3 parameters.  The 1st is the is the variable, field or table
 * name to be added to the dictionary. This is necessary since there is no
 * way to implement a standard match for the symbol name from all language
 * statements that call this rule.
 * <p>
 * A 2nd parameter is <code>null</code> if this is a variable dictionary
 * add or it is an Object that represents a table in the schema dictionary
 * in which to add fields. 
 * <p>
 * A 3rd parameter provides the context of the type of symbol to be looked
 * up.  This is a simple boolean value that indicates if the name is a table
 * or a variable/field.  This context is known by the referencing rule and it 
 * eliminates any ambiguity regarding the type of lookup.
 * <p>
 * Locations from which this rule is referenced:
 * <ul>
 *    <li> as a variable or field (lvalue)
 *       <ul>
 *          <li> {@link #def_var_stmt}
 *          <li> {@link #def_parm_stmt}
 *          <li> {@link #format_phrase}
 *       </ul>
 *    <li> as a table name
 *       <ul>
 *          <li> {@link #def_temp_table_stmt} (for temp-tables and
 *               work-tables)
 *       </ul>
 *    <li> as a field name to be added to a temp-table or work-table
 *       <ul>
 *          <li> {@link #add_temp_table_field}
 *       </ul>
 * </ul>
 * <p>
 * The 2nd token found must be a valid name from the schema or as defined by
 * previous statements.  The corresponding token type is then used to set the
 * proper variable data type or to obtain the temp-table/work-table name.
 * For variables and fields, the type is returned by the {@link #lvalue} rule
 * which handles field level lookups.  For table names the {@link #record}
 * rule is used.
 * <p>
 * In temp-table and work-table adds, there is optional matching that is
 * additionally enabled by semantic predicate.  This matching is done in
 * the {@link #temp_table_use_index} rule.
 * <p>
 * Note that this rule DOES support the addition of scoped shared and global
 * shared variables.  No special steps are necessary to enable this due to
 * a quirk in how Progress handles such variables, they can only be defined
 * in the top-level context (not inside a procedure or function which is the
 * only block in which there is a different scoping level).  For this reason,
 * whenever shared variables are added to the dictionary, they can be added
 * to the current scope, since this is the same as the global scope in any
 * valid Progress code.
 * <p>
 * Bogus ANTLR ambiguity warnings have been disabled.
 *
 * @param    symName
 *           Text which is the variable, temp-table or work-table name that
 *           should be added to the current scope of the respective
 *           dictionary. If <code>null</code>, no name will be added and this
 *           clause is essentially ignored.
 * @param    table
 *           A reference to a table previously added to the dictionary. Must
 *           be <code>null</code> if this is a modification of the variable
 *           namespace.
 * @param    tablename
 *           If true, the name is to be interpreted as a tablename, otherwise
 *           the name is an lvalue (variable or field name).
 */
like_clause [String symName, Object table, boolean tablename]   
   :
      {
         boolean hasValidate = false;
         Aast field = null;
         Aast[] tableFields = null;
      }
      (
           { !tablename }?
           KW_LIKE_SEQ^
         | KW_LIKE^
      )
      {
         inLikeClause = true;
      }
      (
           { !tablename }?
           l:lvalue 
         | r:record[false, false, false]
      )
      {
         // this block is not an exit action, but does need to be executed
         // before the matching on temp_table_use_index 
         
         // this is only needed in lvalue annotations and subscript
         // processing (called from lvalue) so it can be "turned off" here
         inLikeClause = false;
         
         if (symName != null)
         {
            if (table != null)
            {
               if (tablename)
               {
                  // this is a "like table" form, so add all the fields
                  // from the referenced table into the target table
                  if (#r != null)
                  {
                     tableFields = sym.addFieldsFrom(table, #r.getText());
                  }
                  else
                  {
                     System.err.println("Missing required record for adding "
                                        + "'all fields from' table.");
                  }
               }
               else
               {
                  // this is a field to be added to a specific table def
                  // (temp-table or work-table)
                  if (#l != null)
                  {
                     field = sym.addFieldLike(table, symName, #l); 
                  }
                  else
                  {
                     String errtxt = "Missing required lvalue for adding " +
                                     symName +
                                     " field to table.";
                     System.err.println(errtxt);
                  }
               }
               
               // this works as long as this code is only called by things
               // like def_temp_table_stmt which is true at this time
               sym.promoteTable(table, true, false);
            }
            else
            {
               // here is where we enable the variable lookup for all
               // subsquent code in this scope, note that Progress only seems
               // to ever have 2 levels of scope: the external proc and any
               // internal proc/trigger defs, since you can only define shared
               // or global shared vars in the external proc scope, we should
               // *not* need to explicitly add vars to the global scope since
               // adding to the current scope should be equivalent
               if (#l != null)
               {
                  sym.addLocalVariableLike(symName, #l);
               }
               else
               {
                  System.err.println("Missing required lvalue for adding " +
                                     symName +
                                     " to local variable dictionary.");
               }
            }
         }
      }   
      (
         options { generateAmbigWarnings = false; }
         :
           { table != null }?
           kv:KW_VALIDATE
           {
              // TODO: do we need to record anything special in the field case?
              if (tablename)
                 sym.setTableOption(table, KW_VALIDATE, #kv);
                 
              hasValidate = true;
           }
           
         | { tablename }?
           temp_table_use_index[table]
      )*
      {
         if (field != null)
         {
            String sourceSchemaName = (String) #l.getAnnotation("schemaname");
            sym.processLikeField(field, hasValidate, sourceSchemaName);
         }
         
         if (tableFields != null)
         {
            String sourceSchemaName = (String) #r.getAnnotation("schemaname");
            for (int i = 0; i < tableFields.length; i++)
            {
               Aast child = tableFields[i];
               String schemaName = sourceSchemaName + "." + child.getText();
               sym.processLikeField(child, hasValidate, schemaName);
            }
         }
     }
   ;
   
/**
 * Matches the <code>LIKE</code> language keyword that defines the name of
 * a widget to pattern another widget upon.  In this case the type of the
 * widget is already known so this has no namespace implications.
 * This is called from {@link #def_button_stmt} and other locations which all
 * handle the same syntax.  This allows the logic to be centralized and
 * reused from multiple callers. This rule is only separate to properly
 * build the AST.
 */
like_widget_clause   
   :
      KW_LIKE^ widget
   ;
   
/**
 * Matches the <code>LIKE</code> language keyword that defines the name of
 * a menu to pattern another menu upon.  In this case the type of the
 * menu is already known so this has no namespace implications.
 * This is called from {@link #def_menu_stmt}.  This rule is only separate 
 * to properly build the AST.
 */
like_menu_clause
   :
      // placeholder for menu namespace resolution
      KW_LIKE^ symbol
   ;
   
/**
 * Matches all Progress 4GL widget names that exist in the variable 
 * namespace (which is the namespace in which Progress stores them).
 * <p>
 * This rule is simply a convenience method to use the prefer widgets
 * flag to bias the rule reference to {@link #lvalue} to first
 * pick a widget over a variable. When done, the prefer widgets flag
 * state is set back to the default of <code>false</code>.
 */
widget
   :
      {
         preferWidgets = true;
      }
      lvalue
      {
         preferWidgets = false;
      }
   ;
   
/**
 * Matches a set of UI oriented keywords and keywords and the required
 * following expression.  This is the list of keywords matched:
 * <pre>
 *    BGCOLOR
 *    DCOLOR
 *    FGCOLOR
 *    PFCOLOR
 *    FONT
 *    CONTEXT_HELP_ID
 *    COLUMN (allows COLUMNS as a synonym)
 *    ROW
 *    MOUSE-POINTER
 *    COLUMN-BGCOLOR
 *    COLUMN-DCOLOR
 *    COLUMN-FGCOLOR
 *    COLUMN-PFCOLOR
 *    COLUMN-FONT
 *    LABEL-BGCOLOR
 *    LABEL-DCOLOR
 *    LABEL-FGCOLOR
 *    LABEL-PFCOLOR
 *    LABEL-FONT
 *    ACCELERATOR
 *    WIDGET-ID
 * </pre>
 * <p>
 * This method has an override for the normal <code>KW_COL expr</code> that can be matched. 
 * There is a hard coded option in this method to match either <code>KW_COLUMNS</code> or 
 * <code>KW_COL</code> interchangeably (with a following expression) since the 4GL allows
 * this too.  This is an undocumented "feature" of the 4GL.
 * <p>
 * Used by {@link #def_var_stmt} and MANY other locations in the parser.
 * The subtree is rooted by the keyword (which is the reason for using a
 * separate rule for something so simple), but this also centralizes the
 * logic.
 */
ui_stuff
   :
      (
           KW_BGCOLOR^
         | KW_DCOLOR^
         | KW_FGCOLOR^
         | KW_PFCOLOR^ 
         | KW_FONT^         
         | KW_CTX_H_ID^
         | KW_COL^
         | c:KW_COLUMNS^ { #c.setType(KW_COL); }
         | KW_ROW^
         | KW_MOU_PTR^
         | KW_COL_BGC^  // browse only
         | KW_COL_DC^   // browse only
         | KW_COL_FGC^  // browse only
         | KW_COL_FONT^ // browse only
         | KW_COL_PFC^  // browse only
         | KW_LAB_BGC^  // browse only
         | KW_LAB_DC^   // browse and frame only
         | KW_LAB_FGC^  // browse only
         | KW_LAB_FONT^ // browse and frame only
         | KW_LAB_PFC^  // browse only
         | KW_ACCEL^
         | KW_WID_ID^
      )
      expr
   ;
   
/**   
 * Matches a set of image oriented keywords and keywords and the required
 * following {@link #image_phrase}.  This is the list of keywords matched:
 * <pre>
 * IMAGE
 * IMAGE-DOWN
 * IMAGE-INSENSITIVE
 * IMAGE-UP
 * </pre>
 * <p>
 * Used by {@link #def_button_stmt} and the subtree is rooted by the keyword
 * (which is the reason for using a separate rule for something so simple).
 */
image_stuff
   :
      (
           KW_IMAGE^
         | KW_IMG_DOWN^
         | KW_IMG_INS^ 
         | KW_IMG_UP^
      )
      image_phrase
   ;
   
/**   
 * Matches the image phrase which is called from a {@link #def_button_stmt}
 * or a {@link #def_image_stmt}. The first token must be the 
 * <code>KW_FILE</code> followed by a {@link #filename}. Also uses the
 * following:
 * <p>
 * <ul>
 *    <li> {@link #image_size_clause}
 *    <li> {@link #from_clause}
 * </ul>
 * <p>
 * Also allows a match to the <code>KW_FIL_NAME</code> token instead of the
 * <code>KW_FILE</code>.  This is an undocumented feature found in real 4GL
 * code. We normalize the parent to KW_FILE in this case to simplify downstream
 * use.
 */
image_phrase   
   :
      (
           (KW_FILE^ | f:KW_FIL_NAME^) filename { if (#f != null) #f.setType(KW_FILE); }
           (
              options { generateAmbigWarnings = false; }
              :
              image_size_clause
           )?
         | image_size_clause
      )
      (from_clause)?          
   ;
   
/**   
 * Matches the <code>IMAGE-SIZE, IMAGE-SIZE-CHARS and IMAGE-SIZE-PIXELS</code>
 * keywords and the required following width and height.  Used by
 * {@link #image_phrase} and the subtree is rooted by the keyword (which is
 * the reason for using a separate rule for something so simple).
 */
image_size_clause   
   :
      (KW_IMG_SZ^ | KW_IMG_SZ_C^ | KW_IMG_SZ_P^)
      numeric_literal KW_BY! numeric_literal
   ;
   
/**   
 * Matches the <code>FROM</code> keyword and the required following 
 * coordinates (x y or col row).  Used by the {@link #image_phrase} rule and
 * the subtree is rooted by the keyword (which is the reason for using a
 * separate rule for something so simple).
 */
from_clause   
   :
      KW_FROM^ location_spec location_spec
   ;
   
/**   
 * Matches the <code>NO-FOCUS</code> keyword and the optional following
 * <code>FLAT-BUTTON</code>. Used by {@link #def_button_stmt} and the subtree
 * is rooted by the keyword (which is the reason for using a separate rule
 * for something so simple).
 */
no_focus_clause   
   :
      KW_NO_FOCUS^ (KW_FLAT_BUT)?
   ;
   
/**   
 * Matches the <code>COLUMN-LABEL</code> keyword and the required following
 * string. Used by {@link #def_var_stmt} and the subtree is rooted by the
 * keyword (which is the reason for using a separate rule for something so
 * simple).
 */
column_label
   :
      KW_COL_LAB^ STRING
   ;
   
/**   
 * Matches the <code>FORMAT</code> keyword and the required following
 * string. Used by {@link #def_var_stmt} and {@link #format_phrase} rules.
 * The subtree is rooted by the keyword (which is the reason for using a
 * separate rule for something so simple).
 * <p>
 * Customer code has shown that Progress allows unquoted format 'strings'
 * to be used here.  For example 99/99/9999 instead of '99/99/9999'! This
 * means that we must use {@link #expr} instead of <code>STRING</code>
 * AND that this trick of using an expression may fail if their tolerance
 * for such ambiguous and arbitrary constructs is great.  For example,
 * if one were to code x(25) instead of 'x(25)', this would start to be
 * quite a problem since this would parse as a function call to a function
 * 'x' which probably doesn't exist and even if it did, it would be the
 * wrong tokens since we really should have a single required string token
 * following the <code>FORMAT</code> keyword.
 * <p>
 * To make this better, the exit action converts any non-quoted format to
 * a <code>STRING</code> node with the proper quoting.
 * <p>
 * It might be useful to merge this with {@link #ui_stuff} since both rules
 * have identical structure and they are <b><code>almost always</code> referenced
 * as peer alternatives.  However, there are a few places where this rule
 * is referenced without the other so this would add ambiguity where none
 * exists today.  More importantly, the risk associated with changes 
 * to this rule (because of the ridiculous ambiguity described above) means
 * it is safer to leave this separate so that changes can be made here as
 * needed.</b>
 *
 * @param    dynExpr
 *           flag indicating the format string can be any dynamic expression.
 */
format_string [ boolean dynExpr ]
   :
      KW_FORMAT^ e:expr
      {
         Aast child = #e.getChildAt(0);
      
         // convert any non-quoted format to a valid string
         if (!dynExpr && child.getType() != STRING)
         {
            child.setType(STRING);
            child.setText('"' + child.getText() + '"');
         }
      }
   ;
   
/**   
 * Matches the <code>HELP</code> keyword and the required following
 * string. Used by {@link #format_phrase} and the subtree is rooted by the
 * keyword (which is the reason for using a separate rule for something so
 * simple).
 */
help_string
   :
      KW_HELP^ STRING
   ;
   
/**   
 * Matches the <code>DECIMALS</code> keyword and the required following
 * integer literal. Used by {@link #def_var_stmt} and the subtree is rooted by
 * the keyword (which is the reason for using a separate rule for something so
 * simple).
 */
decimals_clause
   :
      KW_DECIMALS^ NUM_LITERAL
   ;
   
/**
 * Matches the <code>ON { ENDKEY | ERROR | QUIT | STOP } UNDO</code> 
 * construction used to modify block processing. This is a more efficient
 * coding than normally found in Progress 4GL documentation.  More
 * specifically, normally these 4 possible constructions are defined
 * separately, but in fact their structure is identical.
 * <p>
 * The core logic is shared with the <code>UNDO</code> statement and
 * implemented in a shared rule {@link #undo_core}. The exception to this
 * is the <code>KW_THROW</code> option which is different (and thus is matched
 * separately. 
 * <p>
 * Used from {@link #do_repeat_stmt} and {@link #for_stmt}.
 */
on_event_phrase
   :
      {
         boolean isQuit = false;
         boolean isErr  = false;
      }
      KW_ON^
      (
           KW_ENDKEY
         | KW_ERROR  { isErr = true; }
         | KW_QUIT   { isQuit = true; }
         | KW_STOP
      )
      (
           // UNDO keyword is optional with the QUIT directive, but ONLY with
           // the QUIT directive;  with all other directives, UNDO is NOT
           // optional
           { !isQuit || LA(1) == KW_UNDO }?
           undo_basic
         | // empty alternative
      )
      (undo_core | COMMA! KW_THROW)?
   ;
   
/**
 * Matches the top level <code>UNDO</code> statement, but some logic is
 * delegated to a shared rule {@link #undo_core} which is also used by
 * {@link #on_event_phrase}.  Since the {@link #throw_clause} is not
 * common with the <code>on_event_phrase</code>, that sub-rule is separately
 * referenced. 
 */
undo_stmt   
   :
      // label_reference may match nothing (empty alt) so it is non-optional
      // here even though it is really optional
      KW_UNDO^ label_reference (undo_core | throw_clause)? DOT!
   ;
   
/**   
 * Implements the shared core logic for <code>UNDO</code> (see 
 * {@link #undo_stmt}) and the {@link #on_event_phrase}.
 */
undo_core
   :
      COMMA!
      (
           other_on_event_actions
         | return_core
      )
   ;
   
/**
 * Matches the <code>KW_THROW</code> with a following {@link #expr} which
 * will be a child node.  Used by the {@link #undo_stmt}.
 */
throw_clause
   :
      COMMA! KW_THROW^ expr
   ;
   
/**
 * Implements the the <code>UNDO</code> action and the optional following
 * label references. This is done to improve the tree build.  This is only
 * called by {@link #on_event_phrase}.
 * <p>
 * Of particular interest is that this rule implements references to labels
 * (see {@link #label_reference}) and to a common form of <code>return</code>
 * keyword processing (see {@link #return_core} which is also accessed from
 * the <code>RETURN</code> statement rule {@link #return_stmt}).
 * <p>
 * Note that the use of expressions in the calling rules creates ambiguity
 * with the use of the label_reference. This rule is unambiguous in practice.
 * For this reason, ambiguity warnings have been disabled.
 */
undo_basic   
   :
      // label_reference may match nothing (empty alt) so it is non-optional
      // here even though it is really optional
      KW_UNDO^ label_reference
   ;
   
/**
 * Implements the other event actions and the optional following label
 * references. This is done to improve the tree build.
 * <p>
 * Of particular interest is that this rule implements references to labels
 * (see {@link #label_reference}) and to a common form of <code>return</code>
 * keyword processing (see {@link #return_core} which is also accessed from
 * the <code>RETURN</code> statement rule {@link #return_stmt}).
 * <p>
 * Note that the use of expressions in the calling rules creates ambiguity
 * with the use of the label_reference. This rule is unambiguous in practice.
 * For this reason, ambiguity warnings have been disabled.
 */
other_on_event_actions   
   :
      // label_reference may match nothing (empty alt) so it is non-optional
      // here even though it is really optional
      (KW_LEAVE^ | KW_NEXT^ | KW_RETRY^) label_reference
   ;   
   
/**
 * Matches a Progress 4GL label reference that refers to a label that was
 * previously defined in a block definition, see {@link #inner_block}.
 * <p>
 * The matching portion of this rule is defined as a single token that
 * matches the union of the label token type and the set of all possible
 * alternatives for symbol names.  Normally, symbols must start with an
 * alphabetic character or underscore, but labels are special in that they
 * can be malformed.  This means they can start with any valid symbol 
 * character (such as a minus, a number or other characters). To implement
 * this, a call to {@link #malformed_symbol} rule is used.
 * <p>
 * Since the <code>malformed_symbol</code> rule includes reserved keywords
 * in its possible token set, if the calling rule has following tokens that
 * map into this set, there will be a potentially inappropriate matching to
 * this rule.  For this reason, this rule may return nothing if it detects
 * that there is no label reference at all.
 * <p>
 * It is important to note that in the end, it is not really valid to have
 * a token type of <code>SYMBOL</code> since this could be anything.  What we
 * really require is a token that is a <code>LABEL</code>.  The result of
 * this rule is always rewritten to have the <code>LABEL</code> type.
 */
label_reference   
   :
      {
         // save off this token for error handling later
         Token lead  = LT(1);
         int   ttype = lead.getType(); 
         
         boolean proper = false;
         boolean mal    = false;
         
         // check if this a normal label
         if ((ttype > BEGIN_UNRESERVED && ttype < END_UNRESERVED) ||
             ttype == SYMBOL)
         {
            if (sym.lookupLabel(lead.getText()) != -1)
            {
               proper = true;
            }
         }
         
         if (!proper)
         {
            mal = isMalformedLabel(false, 1, null);
         }
      }
      (
         options { generateAmbigWarnings = false; }
         :
           LABEL
           
           // this represents the SYMBOL token and the full range of
           // unreserved keywords which exists to "pull" us into this
           // rule, but when we leave the type is always reset to LABEL
           // OR we throw an exception           
         | { proper || mal }?
           malformed_symbol
         
         | // empty alternative
      )
      {
         if (## != null)
         {
            ttype = ##.getType();
         
            if (ttype != LABEL)
            {
               int newtype = sym.lookupLabel(##.getText());
               
               // there should always be a match in valid Progress 4GL source
               // so we throw an exception if this is not the case
               if ( newtype != -1 )
               {
                  ##.setType( newtype );
               }
               else
               {
                  throw new NoViableAltException(lead, getFilename());
               }
            }
         }
      }
   ;
   
/**
 * Implements a <code>KW_CATCH</code> block which is part of the structured
 * error handling added in 10.2x. This uses the {@link #symbol}, 
 * {@link #as_clause}, {@link #block} and {@link #end_stmt} sub-rules to
 * match and build the tree.  The tree will be rooted at the
 * <code>KW_CATCH</code> node.  Variable and schema scopes are pushed and
 * popped in this statement.
 */
catch_stmt
   :
      {
         sym.addScope();
         sym.addSchemaScope(true);
      }
      (
         cat:KW_CATCH^ e:symbol as_clause[#e.getText(), false] COLON!
         {
            // this normally sets parsed options into the Variable instance
            // in the symbol resolver; but in this case it just ensures some
            // metadata is set properly
            sym.setVariableOptions(#e.getText(), #cat);
            
            // this will save some of the metadata state into the catch node
            // as annotations, but the normal options processing is not needed
            sym.annotateVariableOptions(#e.getText(), #cat, true);            
         }
         block end_stmt[true] 
      )
      {
         sym.deleteSchemaScope(true);
         sym.deleteScope();
      }      
   ;
   
/**
 * Implements a <code>KW_FINALLY</code> block which is part of the structured
 * error handling added in 10.2x. This uses the {@link #block} and 
 * {@link #end_stmt} sub-rules to match and build the tree.  The tree will be
 * rooted at the <code>KW_FINALLY</code> node.  Variable and schema scopes are
 * pushed and popped in this statement.
 */
finally_stmt
   :
      {
         sym.addScope();
         sym.addSchemaScope(true);
      }
      (
         KW_FINALLY^ COLON! block end_stmt[true]
      )
      {
         sym.deleteSchemaScope(true);
         sym.deleteScope();
      }      
   ;
   
/**
 * Matches the <code>KW_ROUTINEL</code> keyword with a following {@link #on_event_phrase}.
 */
routine_level_stmt
   :
     KW_ROUTINEL^ on_event_phrase DOT!
   ;
   
/**
 * Matches the <code>KW_BLK_LVL</code> keyword with a following {@link #on_event_phrase}.
 */
block_level_stmt
   :
     KW_BLK_LVL^ on_event_phrase DOT!
   ;
   
/**   
 * Implements the preselect phrase which is used to modify certain block
 * types to specify how to prequery the database to create a result set
 * that would not otherwise be available.  The tree is rooted at the
 * <code>PRESELECT</code> keyword and use rule references to implement
 * a more natural tree structure. See {@link #each_first_last_spec}, 
 * {@link #by_clause} and {@link #collate_clause} for more details.
 * <p>
 * Used by {@link #do_repeat_stmt} and similar code is inlined in the
 * {@link #for_stmt}.
 */
preselect_phrase   
   :
      // DO PRESELECT and REPEAT PRESELECT both force re-promotion
      KW_PRESEL^ each_first_last_spec[true, true]
   ;
   
/**   
 * Matches the construction used in the {@link #preselect_phrase},
 * {@link #for_or_preselect_clause} and in the {@link #for_stmt} for
 * specifying one or more {@link #record_phrase} specifications and
 * optionally prefixing the phrase with the <code>EACH, FIRST or LAST</code>
 * keywords.
 *
 * @param    force
 *           <code>true</code> to force promotion even if the record was
 *           already previously promoted.
 * @param    noProp                   
 *           <code>true</code> to force any promoted nodes to have their
 *           no propagate flag set. This disables propagation of these
 *           nodes across the current scope boundary.
 */
each_first_last_spec [boolean force, boolean noProp]
   :
      (
         (KW_EACH | KW_FIRST | KW_LAST)? record_phrase[force, noProp]
         (
            COMMA! (KW_EACH | KW_FIRST | KW_LAST)?
            record_phrase[force, noProp]
         )*
      )
   ;
   
/**   
 * Matches the <code>BY</code> keyword and the required following expression.
 * <p>
 * Used by {@link #aggregate_phrase} and {@link #loop_incr} rules.
 * <p>
 * The subtree is rooted by the keyword (which is the reason for using a 
 * separate rule for something so simple).
 */
by_clause   
   :
      KW_BY^ expr
   ;
   
/**   
 * Matches the <code>BY</code> keyword, the required following expression
 * and the optional <code>DESCENDING</code> keyword.
 * <p>
 * Used by {@link #open_query_stmt} and {@link #common_block_options} rules.
 * <p>
 * The subtree is rooted by the keyword (which is the reason for using a 
 * separate rule for something so simple).
 */
sort_order_clause   
   :
      KW_BY^ expr (KW_DESCEND)?
   ;
   
/**   
 * Matches the <code>COLLATE</code> keyword and the required following
 * 2 or 3 character expressions.  The tree is rooted on the 
 * <code>COLLATE</code> keyword and all parenthesis and commas are
 * dropped from the tree.  The resulting tree will have 2 or 3 children,
 * each an expression.
 * <p>
 * Used by {@link #preselect_phrase} and the subtree is rooted by the
 * keyword (which is the reason for using a separate rule for something so
 * simple).
 */
collate_clause   
   :
      KW_COLLATE^ LPARENS! expr COMMA! expr (COMMA! expr)? RPARENS!
      (KW_DESCEND)?
   ;
   
/**   
 * Implements the query tuning phrase including all possible options. In
 * some cases rule references are used to ensure that the tree is built
 * in the most useful manner.
 * <p>
 * Used by {@link #do_repeat_stmt} and {@link #for_stmt}.
 */
query_tuning_phrase
   :
      KW_QRY_TUNE^ LPARENS!
      (
           (KW_ARR_MSG | KW_NO_ARMSG)
         | (KW_BIND_WH | KW_NO_BIND)
         | query_tuning_cache_size
         | (query_tuning_debug | KW_NO_DEBUG)
         | KW_HINT
         | (KW_IDX_HINT | KW_NO_IDX_H)
         | (KW_JOIN_BY | KW_NO_JOIN)
         | (KW_LOOKAHD | KW_NO_LOOKA)
         | KW_ORD_JOIN
         | KW_REV_FROM
         | (KW_SEP_CONN | KW_NO_SEP_C)
      )*
      RPARENS!      
   ;
   
/**   
 * Matches the <code>CACHE-SIZE</code> keyword and the required following
 * integer literal.
 * <p>
 * Used by {@link #query_tuning_phrase} and the subtree is rooted by the
 * keyword (which is the reason for using a separate rule for something so
 * simple).
 */
query_tuning_cache_size   
   :
      KW_CACHE_SZ^ expr (KW_BYTE | KW_ROW)?
   ;
   
/**   
 * Matches the <code>DEBUG</code> keyword and the optional following
 * qualifier keyword (<code>SQL or EXTENDED</code>).
 * <p>
 * Used by {@link #query_tuning_phrase} and the subtree is rooted by the
 * keyword (which is the reason for using a separate rule for something so
 * simple).
 */
query_tuning_debug   
   :
      KW_DEBUG^ (KW_SQL | extended_diag_option)?
   ;
   
/**   
 * Matches the <code>EXTENDED</code> keyword and the optional following
 * qualifier keywords <code>CURSOR, DATA-BIND, PERFORMANCE or VERBOSE</code>.
 * <p>
 * Used by {@link #query_tuning_debug} and the subtree is rooted by the
 * keyword <code>KW_EXTENDED</code> (which is the reason for using a separate
 * rule for something so simple).
 */
extended_diag_option
   :
      KW_EXTENDED^ (KW_CURSOR | KW_DATA_BND | KW_PERF | KW_VERBOSE)
   ;
   
/**   
 * Implements the record phrase used in various block definitions. In its
 * most basic form, a table ({@link #record}) and optional include or exclude
 * list of fields is provided ({@link #field_list}).  Uses the 
 * {@link #literal} rule to match the optional constant that follows the
 * field list.  The {@link #join_clause}, {@link #of_clause}, 
 * {@link #where_clause}, {@link #using_clause} and {@link #use_index_clause}
 * rules are all implemented separately to generate the proper tree 
 * structure. Note that these last options can be specified in any order and
 * multiple times.
 * <p>
 * The join clause is only here for use by {@link #each_first_last_spec} and
 * not by {@link #find_stmt}.  Likewise the <code>NO-WAIT and NO-ERROR</code>
 * keywords are only here for <code>FIND</code>.  Otherwise this single
 * rule properly handles both cases.
 * <p>
 * Note that the use of unreserved keywords for <code>OUTER-JOIN</code> and
 * <code>LEFT</code> can cause ambiguity when k == 2 between the join_clause
 * and the of_clause (which includes the {@link #symbol} rule). In practice,
 * this is not really a problem because join_clause is a more specific rule.
 * For this reason, ambiguity warnings have been disabled.
 *
 * @param    force
 *           <code>true</code> to force promotion even if the record was
 *           already previously promoted.
 * @param    noProp                   
 *           <code>true</code> to force any promoted nodes to have their
 *           no propagate flag set. This disables propagation of these
 *           nodes across the current scope boundary.
 */
record_phrase [boolean force, boolean noProp]   
   :
      {
         astFactory.makeASTRoot(currentAST, #[RECORD_PHRASE,"record phrase"]);
         boolean old = sym.markPreferred(true);
      }
      record[true, force, noProp]
      {
         sym.markPreferred(old);
      }
      (field_list)?
      (
         options { generateAmbigWarnings = false; }
         :
         {
           LA(1) != KW_SH_LOCK && LA(1) != KW_EXC_LOCK &&
           LA(1) != KW_NO_LOCK && LA(1) != KW_NO_WAIT
         }?
         literal
      )?
      (
         options { generateAmbigWarnings = false; }
         :      
           join_clause
         | of_clause
         | where_clause
         | use_index_clause
         | using_clause
         | (KW_SH_LOCK | KW_EXC_LOCK | KW_NO_LOCK)
         | KW_NO_WAIT
         | KW_NO_PRE
         | KW_NO_ERROR
         
         // the v11.0 docs describe WHOLE-INDEX as an option instead of TABLE-SCAN (in the text
         // but not the syntax diagrams or keyword index), that option is untested at this time
         | KW_TAB_SCAN
      )*
   ;
   
/**
 * Matches a <code>CONNECT</code> language statement, the following
 * {@link #database_name} and arbitrary connect options.
 * <p>
 * Used by {@link #stmt_list}.
 */
connect_stmt   
   :
      KW_CONN^
      (
           { LT(1).getText().startsWith("-") == false }?
           database_name[true]
         | // empty alternative
      )
      connect_options
      (KW_NO_ERROR)?
      DOT!
   ;
   
/**
 * Matches a <code>DISCONNECT</code> language statement, and the following
 * {@link #database_name}. 
 * <p>
 * Used by {@link #stmt_list}.
 */
disconnect_stmt   
   :
      KW_DISCONN^ database_name[false] (KW_NO_ERROR)? DOT!
   ;
   
/**
 * Matches a <code>CREATE ALIAS</code> language statement, and the following
 * database names (see {@link #database_name}). 
 * <p>
 * Used by {@link #stmt_list}.
 */
create_alias_stmt   
   :
      c:KW_CREATE^  { #c.setType(CREATE_ALIAS); }
      KW_ALIAS! database_name[false]
      KW_FOR KW_DATABASE database_name[false]
      (KW_NO_ERROR)? 
      DOT!
   ;   

/**
 * Matches a <code>DELETE ALIAS</code> language statement, and the following
 * {@link #database_name}. 
 * <p>
 * Used by {@link #stmt_list}.
 */
delete_alias_stmt   
   :
      d:KW_DELETE^  { #d.setType(DELETE_ALIAS); }
      KW_ALIAS! database_name[false] DOT!
   ;
   
/**
 * Matches a <code>SAVE CACHE</code> language statement, and the following
 * {@link #current_or_complete_clause}, {@link #value} or {@link #filename}. 
 * <p>
 * Used by {@link #stmt_list}.
 */
save_cache_stmt   
   :
      s:KW_SAVE^  { #s.setType(SAVE_CACHE); } 
      KW_CACHE!
      current_or_complete_clause
      KW_TO!
      external_name_spec
      (KW_NO_ERROR)?
      DOT!
   ;   
   
/**
 * Matches a <code>CURRENT or COMPLETE</code> keywords and the following
 * {@link #database_name} or {@link #value} rule references. 
 * <p>
 * Used by {@link #save_cache_stmt}.
 */
current_or_complete_clause   
   :
      (KW_CURRENT^ | KW_COMPLETE^) database_name[false] 
   ;
   
/**
 * Matches a <code>OPEN QUERY</code> language statement and all possible
 * options.  Uses: 
 * <p>
 * <ul>
 *    <li> {@link #malformed_symbol}
 *    <li> {@link #for_or_preselect_clause}
 *    <li> {@link #max_rows_clause}
 *    <li> {@link #collate_clause}
 *    <li> {@link #by_clause}
 *    <li> {@link #query_tuning_phrase}
 * </ul>
 * <p>
 * Used by {@link #stmt_list}.                
 */
open_query_stmt   
   :
      o:KW_OPEN^  { #o.setType(OPEN_QUERY); }
      KW_QUERY! q:malformed_symbol { #q.setType(QUERY); }
      for_or_preselect_clause
      (
         options { generateAmbigWarnings = false; }
         :      
           KW_IDX_REPO
         | KW_BREAK
         | max_rows_clause 
         | (collate_clause | sort_order_clause)
         | query_tuning_phrase
      )*
      DOT!
   ;

/**
 * Matches a <code>CLOSE QUERY</code> language statement and the required
 * following query name (see {@link #malformed_symbol}).
 * <p>
 * Used by {@link #stmt_list}.
 */
close_query_stmt   
   :
      c:KW_CLOSE^  { #c.setType(CLOSE_QUERY); } 
      KW_QUERY! q:malformed_symbol { #q.setType(QUERY); } DOT!
   ;
   
/**
 * Matches a <code>REPOSITION</code> language statement, the required
 * following query name (see {@link #query_reference}) and repositioning
 * information.  Uses {@link #to_recid_or_rowid_clause} and 
 * {@link #reposition_spec} rules.
 * <p>
 * Used by {@link #stmt_list}.
 */
reposition_stmt
   :
      KW_REPOS^ query_reference
      (
           to_recid_or_rowid_clause
         | reposition_spec
      )
      (KW_NO_ERROR)?
      DOT!
   ;
   
/**
 * Matches a <code>GET</code> language statement, the required keywords and
 * query name (see {@link #query_reference}).  It supports all possible
 * options, in any order.
 * <p>
 * Used by {@link #stmt_list}.
 */
get_stmt   
   :
      KW_GET^
      (KW_FIRST | KW_LAST | KW_PREV | KW_NEXT | KW_CURRENT)
      query_reference
      (
           (KW_SH_LOCK | KW_EXC_LOCK | KW_NO_LOCK)
         | KW_NO_WAIT
      )*
      DOT!
   ;
   
/**   
 * Matches the <code>TO RECID, TO ROWID or TO ROW</code> keywords and the
 * following required expression of type <code>RECID, ROWID</code> or integer
 * (in the case of a <code>TO ROW</code>).  Calls {@link #expr}.
 * <p>
 * Used in the {@link #reposition_stmt}.
 * <p>
 * This is only separated to simplify the AST build.
 */
to_recid_or_rowid_clause   
   :
      KW_TO^ (KW_RECID | KW_ROWID | KW_ROW) expr (COMMA! expr)*
   ;
   
/**   
 * Matches a <code>FORWARDS or BACKWARDS</code> keyword and the
 * following required {@link #expr}.  Used in the {@link #reposition_stmt}.
 * <p>
 * This is only separated to simplify the AST build.
 */
reposition_spec
   :
      (KW_FORWARD^ | KW_BACKWARD^) expr
   ;   
   
/**   
 * Matches a <code>FOR or PRESELECT</code> keyword and the following required
 * {@link #each_first_last_spec} of the {@link #open_query_stmt}.
 * <p>
 * This is only separated to simplify the AST build.
 */
for_or_preselect_clause   
   :
      (f:KW_FOR^ | p:KW_PRESEL^) each_first_last_spec[false, false]
   ;
   
   
/**   
 * Matches a <code>MAX-ROWS</code> keyword and the following numeric expression of the
 * {@link #open_query_stmt}.  The Progress documentation is silent on whether the following
 * expression is required to be a constant.  Customer code has shown that this is a full
 * expression.
 * <p>
 * This is only separated to simplify the AST build.
 */
max_rows_clause
   :
      KW_MAX_ROWS^ expr
   ;
   
/**
 * Matches a <code>CREATE DATABASE</code> language statement, the required
 * following expressions and the optional <code>REPLACE or NO-ERROR</code>
 * keywords.
 * <p>
 * Uses {@link #expr} and {@link #from_database_clause} rules. 
 * <p>
 * Used by {@link #stmt_list}.
 */
create_database_stmt   
   :
      c:KW_CREATE^  { #c.setType(CREATE_DATABASE); }
      KW_DATABASE! expr (from_database_clause)?
      (
           KW_REPLACE
         | KW_NO_ERROR
      )*
      DOT!
   ;
   
/**   
 * Matches a <code>FROM</code> keyword and the following {@link #expr}
 * of the {@link #create_database_stmt}.
 * <p>
 * This is only separated to simplify the AST build.
 */
from_database_clause   
   :
      KW_FROM^ expr (KW_NEW_INST)?
   ;
   
/**
 * Matches a <code>CREATE QUERY</code> language statement, the required
 * following {@link #handle} and the optional {@link #in_widget_pool_clause}.
 * <p>
 * An undocumented feature was found where the <code>NO-ERROR</code> keyword
 * can be added at the end of this statement.
 * <p>
 * Used by {@link #stmt_list}.
 */
create_query_stmt   
   :
      c:KW_CREATE^  { #c.setType(CREATE_QUERY); }
      KW_QUERY! 
      handle 
      (in_widget_pool_clause)?
      (KW_NO_ERROR)?
      DOT!  
   ;
   
/**
 * Matches a <code>CREATE DATASET</code> language statement, the required
 * following {@link #handle} and the optional {@link #in_widget_pool_clause}.
 * <p>
 * Used by {@link #stmt_list}.
 */
create_dataset_stmt   
   :
      c:KW_CREATE^  { #c.setType(CREATE_DATASET); }
      KW_DATASET! 
      handle 
      (in_widget_pool_clause)?
      DOT!  
   ;
   
/**
 * Matches a <code>CREATE DATA-SOURCE</code> language statement, the required
 * following {@link #handle} and the optional {@link #in_widget_pool_clause}.
 * <p>
 * Used by {@link #stmt_list}.
 */
create_datasource_stmt   
   :
      c:KW_CREATE^  { #c.setType(CREATE_DATA_SOURCE); }
      KW_DATA_SRC! 
      handle 
      (in_widget_pool_clause)?
      DOT!  
   ;
   
/**   
 * This is a convenience method that matches {@link #lvalue} except a
 * validating semantic predicate is used to throw an exception if the
 * data type of the variable or field is not <code>HANDLE</code>.
 * <p>
 * It is an open question as to whether this rule should also have references
 * to <code>FUNC_HANDLE</code> or <code>METH_HANDLE</code>.  Generally the
 * answer would be NO if the caller would need to assign to this handle and YES
 * if the caller would just use the handle.  Most references to this rule are
 * in create/delete rules where this rule is safe. The 
 * {@link #from_handle_clause} does seem to use this differently. These two
 * types are provisionally supported for now until they can be checked in
 * real code.
 * <p>
 * To test for <code>ATTR_HANDLE</code>, <code>ATTR_COM_HANDLE</code>,
 * <code>METH_HANDLE</code> and <code>METH_COM_HANDLE</code>, a simple
 * check for a <code>COLON</code> operator is done.  This test should probably
 * be made more thorough.
 * <p>
 * Used in {@link #create_query_stmt} and in other locations.
 */
handle   
   :
      h:lvalue 
      {
         #h.getType() == VAR_HANDLE       ||
         #h.getType() == FIELD_HANDLE     ||
         #h.getType() == FUNC_HANDLE      ||
         #h.getType() == COLON            ||
         #h.getType() == SYS_HANDLE       ||
         #h.getType() == VAR_COM_HANDLE   ||
         #h.getType() == FIELD_COM_HANDLE ||
         #h.getType() == FUNC_COM_HANDLE
      }?
   ;
   
/**
 * Matches a <code>CREATE BUFFER</code> language statement, the required
 * following {@link #handle}, the required <code>FOR TABLE</code>
 * construct (see {@link #for_table_clause}) and the optional 
 * {@link #in_widget_pool_clause} and {@link #buffer_name_clause} rules.
 * <p>
 * An undocumented feature was found where the <code>NO-ERROR</code> keyword
 * can be added at the end of this statement.
 * <p>
 * Used by {@link #stmt_list}.
 */
create_buffer_stmt   
   :
      c:KW_CREATE^  { #c.setType(CREATE_BUFFER); }
      KW_BUFFER! 
      handle 
      for_table_clause
      (
           buffer_name_clause
         | in_widget_pool_clause
      )*
      (KW_NO_ERROR)?
      DOT!  
   ;
   
/**   
 * Matches a <code>FOR TABLE</code> construct and the following {@link #expr}
 * of the {@link #create_buffer_stmt}.
 * <p>
 * This is only separated to simplify the AST build.
 */
for_table_clause   
   :
      KW_FOR^ KW_TABLE expr
   ;
   
/**   
 * Matches a <code>BUFFER-NAME</code> keyword and the following {@link #expr}
 * of the {@link #create_buffer_stmt}.
 * <p>
 * This is only separated to simplify the AST build.
 */
buffer_name_clause   
   :
      KW_BUF_NAME^ expr
   ;
   
/**
 * Matches the <code>BUFFER-COMPARE and BUFFER-COPY</code> language statements
 * and all following constructs.
 * <p>
 * <ul>
 *    <li> {@link #record}
 *    <li> {@link #buffer_modification_list}
 *    <li> {@link #to_target}
 *    <li> {@link #save_result_in} (compare only)
 *    <li> {@link #explicit_compares} (compare only)
 *    <li> {@link #sub_assign_clause} (copy only)
 * </ul>
 * <p>
 * Used by {@link #stmt_list}.
 * <p>
 * Bogus ANTLR ambiguity warnings are disabled.
 */
buffer_compare_or_copy_stmt   
   :
      {
         boolean compare = true;
      }
      (
         (KW_BUF_COMP^ | KW_BUF_COPY^ { compare = false; } )
         record[true, true, false]
         (buffer_modification_list)?
         to_target
         (         
            options { generateAmbigWarnings = false; }
            :
              { compare }?
              (save_result_in | explicit_compares)
            | { !compare }?
              sub_assign_clause
            | KW_CASE_SEN
            | KW_BINARY
            | KW_NO_LOBS
         )*
         (KW_NO_ERROR)?
         DOT!
      )                       
   ;
   
/**
 * Matches the <code>EXCEPT or USING</code> keywords and the required list
 * of following {@link #lvalue}.
 * <p>
 * Used by {@link #buffer_compare_or_copy_stmt}.
 */
buffer_modification_list   
   :
      (KW_EXCEPT^ | KW_USING^) (lvalue)*
   ;   
   
/**
 * Matches the <code>TO</code> keyword and the following {@link #record}.
 * <p>
 * Used by {@link #buffer_compare_or_copy_stmt}.
 */
to_target   
   :
      KW_TO^ record[true, false, false]
   ;
   
/**
 * Matches the <code>SAVE</code> keyword (and optional <code>RESULT IN</code>)
 * and the following {@link #lvalue}.
 * <p>
 * Bogus ANTLR ambiguity warnings have been disabled.
 * <p>
 * Used by {@link #buffer_compare_or_copy_stmt} (only for the compare form).
 */
save_result_in   
   :
      KW_SAVE^
      (
         options { generateAmbigWarnings = false; }
         :
         KW_RESULT KW_IN
      )?
      lvalue
   ;
   
/**
 * Matches the <code>COMPARES</code> keyword (and optional 
 * <code>EXPLICIT</code> prefix), an optional <code>NO-ERROR</code> and
 * the {@link #compares_when_block}.
 * <p>
 * Used by {@link #buffer_compare_or_copy_stmt} (only for the compare form).
 */
explicit_compares   
   :
      ((KW_EXPLICIT)? KW_COMPARES^)?
      (KW_NO_ERROR)?
      compares_when_block
   ;
   
/**
 * Matches the <code>COLON</code> token followed by one or more 
 * {@link #when_clause} constructs and terminated by an <code>END</code>
 * keyword.
 * <p>
 * Used by {@link #explicit_compares}.
 */
compares_when_block   
   :
      COLON^ (when_clause)* KW_END! KW_COMPARES!
   ;
   
/**
 * Matches the <code>ASSIGN</code> keyword and a list of space delimited
 * assignment expressions.  This is similar to but more simplified version
 * of {@link #assign_stmt}.
 * <p>
 * Used by {@link #buffer_compare_or_copy_stmt} only in the copy form.
 * <p>
 * Bogus ANTLR ambiguity warnings are disabled.
 */
sub_assign_clause
   :
      KW_ASSIGN^
      (
         options { generateAmbigWarnings = false; }
         :
         assign (simple_when_clause)?
      )+
   ;
   
/**
 * Matches a <code>CREATE TEMP-TABLE</code> language statement, the required
 * following handle variable/field and the optional
 * {@link #in_widget_pool_clause}. The {@link #lvalue} rule and a validating 
 * semantic predicate are used to match the handle.
 * <p>
 * Used by {@link #stmt_list}.
 */
create_temp_table_stmt   
   :
      c:KW_CREATE^  { #c.setType(CREATE_TEMP_TABLE); }
      KW_TEMP_TAB! handle (in_widget_pool_clause)? (KW_NO_ERROR)? DOT!
   ;
   
/**
 * Matches an <code>EMPTY TEMP-TABLE</code> language statement and the
 * required following temp-table name as a {@link #record}.
 * <p>
 * Used by {@link #stmt_list}.
 */
empty_temp_table_stmt   
   :
      e:KW_EMPTY^  { #e.setType(EMPTY_TEMP_TABLE); }
      KW_TEMP_TAB! record[true, false, false] (KW_NO_ERROR)? DOT!
   ;
   
/**
 * Matches a <code>CREATE widget_type</code> or <code>CREATE automation-object</code>language
 * statement, and the following options.  The types of widgets supported are: 
 * <p>
 * <ul>
 *    <li> BUTTON
 *    <li> COMBO-BOX
 *    <li> CONTROL-FRAME
 *    <li> DIALOG-BOX
 *    <li> EDITOR
 *    <li> FILL-IN
 *    <li> FRAME
 *    <li> IMAGE
 *    <li> MENU
 *    <li> MENU-ITEM
 *    <li> RADIO-SET
 *    <li> RECTANGLE
 *    <li> SELECTION-LIST
 *    <li> SLIDER
 *    <li> SUB-MENU
 *    <li> TEXT
 *    <li> TOGGLE-BOX
 *    <li> WINDOW
 * </ul>
 * <p>
 * Uses:
 * <p>
 * <ul>
 *    <li> {@link #lvalue}
 *    <li> {@link #in_widget_pool_clause}
 *    <li> {@link #sub_assign_clause}
 *    <li> {@link #trigger_phrase}
 * </ul>
 * <p>
 * Used by {@link #stmt_list}. Allows the undocumented usage of 
 * <code>KW_NO_ERROR</code> at the end, which was found in working 4GL code.
 */
create_widget_or_object_stmt
   {
      boolean is_com_handle = false;
      boolean is_widget = false;
      boolean allowColon = false;
      boolean is_timer_extension = false;
   }
   :
      c:KW_CREATE^  { #c.setType(CREATE_WIDGET); }
      (
         options { generateAmbigWarnings = false; }
         :
           KW_BUTTON     { is_widget = true; }
         | KW_COMBO_BX   { is_widget = true; }
         | KW_CNTRL_FR   { is_widget = true; }
         | KW_DIALOG     { is_widget = true; }
         | KW_EDITOR     { is_widget = true; }
         | KW_FILL_IN    { is_widget = true; }
         | KW_FRAME      { is_widget = true; }
         | KW_IMAGE      { is_widget = true; }
         | KW_MENU       { is_widget = true; }
         | KW_MENU_ITM   { is_widget = true; }
         | KW_RADIO_S    { is_widget = true; }
         | KW_RECT       { is_widget = true; }
         | KW_SEL_LST    { is_widget = true; }
         | KW_SLIDER     { is_widget = true; }
         | KW_SUB_MENU   { is_widget = true; }
         | KW_TEXT       { is_widget = true; }
         | KW_TOGGL_BX   { is_widget = true; }
         | KW_WINDOW     { is_widget = true; }
         | KW_TIMER      { is_timer_extension = true; }
         | value
         | expr
      )
      (
         h:handle
         {
            if (#h.getType() == VAR_COM_HANDLE   ||
                #h.getType() == FIELD_COM_HANDLE ||
               (#h.getType() == SYS_HANDLE          &&
                #h.getAnnotation("oldtype") != null &&
                ((Long) #h.getAnnotation("oldtype")).intValue() == KW_COM_SELF))
            {
               if (is_widget)
               {
                  // do not allow com-handle vars for create widget cases
                  throw new NoViableAltException(LT(0), getFilename());
               }
               #c.setType(CREATE_OBJECT);
               is_com_handle = true;
            }
            
            if (is_timer_extension)
            {
               #c.setType(CREATE_TIMER);
            }
         }
      )
      (
           { is_com_handle }? com_connect_option
         | // empty alternative
      )
      (
           { !is_com_handle && !is_timer_extension }? in_widget_pool_clause
         | // empty alternative
      )
      (KW_NO_ERROR)?
      (
           { !is_com_handle || is_timer_extension }? attribute_assign_clause
         | // empty alternative
      )
      (
           { !is_com_handle || is_timer_extension }? trigger_phrase { allowColon = true; }
         | // empty alternative
      )
      
      (
           { allowColon }?
           COLON!
         | DOT!
      )
   ;   

/**
 * Matches the <code>ASSIGN</code> keyword and a list of space delimited
 * <code>ATTRIBUTE</code> assignment expressions.  This is similar to but
 * a more simplified version of {@link #assign_stmt}.  Please note that
 * this is a special form for use in the {@link #create_widget_or_object_stmt} AND
 * that the use of the {@link #attribute_assign} rule for the core logic
 * especially since that use is in a loop) MUST be protected with a
 * semantic predicate to ensure that the rule reference doesn't consume
 * a possible following <code>TRIGGERS</code> keyword (since the leftmost
 * token of the called rule can include ANY reserved or other symbol.
 * <p>
 * Also used in the {@link #create_browse_stmt} and {@link #create_server_stmt}.
 * <p>
 * Bogus ANTLR ambiguity warnings are disabled.
 */
attribute_assign_clause
   :
      KW_ASSIGN^
      (
         options { generateAmbigWarnings = false; }
         :
         { LA(1) != KW_TRIGGERS }?
         attribute_assign
      )+
   ;
   
/**   
 * Matches an {@link #attribute_or_method}, an <code>EQUALS</code> token (the 
 * equals sign) and a following {@link #expr}. The token type of the 
 * <code>EQUALS</code> is changed to <code>ASSIGN</code> and the entire
 * construct is rooted on that token.  The token type of the first child
 * of the <code>ASSIGN</code> node must be an attribute, though this is not
 * enforced.
 * <p>
 * Any use of this rule especially if that use is in a loop) MUST be
 * protected with a semantic predicate to ensure that the rule reference 
 * doesn't consume possible following keywords (since the leftmost
 * token of the called rule can include ANY reserved or other symbol.
 * <p>
 * Used by {@link #attribute_assign_clause}.
 */   
attribute_assign   
   :
      attribute_or_method
      eq:EQUALS^   { #eq.setType(ASSIGN); }   
      expr
   ;
   
/**
 * Matches the <code>CONNECT</code> option for the {@link #create_widget_or_object_stmt}
 * with than optional <code>TO expression</code>.
 */
com_connect_option
   :
      KW_CONN^ (KW_TO! expr)?
   ;
   
/**
 * Matches a <code>RELEASE OBJECT</code> language statement, and the
 * following required {@link #expr}.  The Progress documentation states that
 * it should be a <code>COM-HANDLE</code>, but does not state whether this is
 * allowed to be any expression that evaluates to a <code>COM-HANDLE</code>.
 * Customer source code shows that any expression will work. 
 * <p>
 * Used by {@link #stmt_list}.
 */
release_object_stmt   
   :
      r:KW_RELEASE^  { #r.setType(RELEASE_OBJECT); }
      KW_OBJECT! expr (KW_NO_ERROR)? DOT!
   ;   
   
/**   
 * This is a convenience method that matches {@link #lvalue} except a
 * validating semantic predicate is used to throw an exception if the
 * data type of the variable or field is not a <code>COM-HANDLE</code>.
 * <p>
 * It is an open question as to whether this rule should also have references
 * to <code>FUNC_COM_HANDLE</code>.  Generally the answer would
 * be NO if the caller would need to assign to this handle and YES if the
 * caller would just use the handle.  Right now, references to this rule are
 * in create/release rules where this rule is safe.
 * <p>
 * Used in {@link #create_widget_or_object_stmt} and {@link #release_object_stmt}.
 */
com_handle
   :
      h:lvalue 
      {
         #h.getType() == VAR_COM_HANDLE   ||
         #h.getType() == FIELD_COM_HANDLE ||
         (#h.getType() == SYS_HANDLE          &&
          #h.getAnnotation("oldtype") != null &&
          ((Long) #h.getAnnotation("oldtype")).intValue() == KW_COM_SELF)
      }?
   ;
   
/**
 * Matches a <code>DELETE OBJECT</code> language statement, and the
 * following required {@link #expr} which must resolve to a handle or an object
 * reference. The fact that this is an expression is undocumented, the docs
 * only state that this is a handle or an object, but customer code has shown
 * it can be any expression.
 * <p>
 * Used by {@link #stmt_list}.
 */
delete_object_stmt
   {
      boolean parens = false;
   }
   :
      d:KW_DELETE^ { #d.setType(DELETE_OBJECT); } KW_OBJECT!
      expr
      (KW_NO_ERROR)?
      DOT!
   ;   
      
/**
 * Matches a <code>DELETE WIDGET</code> language statement, and the
 * following required {@link #handle} (or list thereof). 
 * <p>
 * Used by {@link #stmt_list}.
 */
delete_widget_stmt   
   :
      d:KW_DELETE^  { #d.setType(DELETE_WIDGET); }
      KW_WIDGET! (handle)+ DOT!
   ;   
   
/**
 * Matches the <code>DDE ADVISE</code> language statement. There is always
 * a following {@link #expr} and {@link #item_clause}.  There is an optional
 * use of {@link #time_clause}. Used by {@link #stmt_list}.
 */
dde_advise_stmt
   :
      d:KW_DDE^  { #d.setType(DDE_ADVISE); }
      KW_ADVISE!
      expr
      (KW_START | KW_STOP)
      item_clause
      (time_clause)?
      (KW_NO_ERROR)?
      DOT!
   ;

/**
 * Matches the <code>DDE EXECUTE</code> language statement. There is always
 * a following {@link #expr} and {@link #command_clause}.  There is an
 * optional use of {@link #time_clause}. Used by {@link #stmt_list}.
 */
dde_execute_stmt
   :
      d:KW_DDE^  { #d.setType(DDE_EXECUTE); }
      KW_EXECUTE!
      expr
      command_clause
      (time_clause)?
      (KW_NO_ERROR)?
      DOT!
   ;

/**
 * Matches the <code>DDE GET</code> language statement. There is always
 * a following {@link #expr}, {@link #source_or_target_clause} and
 * {@link #item_clause}. There is an optional use of {@link #time_clause}.
 * Used by {@link #stmt_list}.
 */
dde_get_stmt
   :
      d:KW_DDE^  { #d.setType(DDE_GET); }
      KW_GET!
      expr
      source_or_target_clause[false]
      item_clause
      (time_clause)?
      (KW_NO_ERROR)?
      DOT!
   ;

/**
 * Matches the <code>DDE INITIATE</code> language statement. There is always
 * a following {@link #expr}, {@link #frame_handle_clause},
 * {@link #application_clause} and {@link #topic_clause}. Used by
 * {@link #stmt_list}.
 */
dde_initiate_stmt
   :
      d:KW_DDE^  { #d.setType(DDE_INITIATE); }
      KW_INITIATE!
      expr
      frame_handle_clause
      application_clause
      topic_clause
      (KW_NO_ERROR)?
      DOT!
   ;

/**
 * Matches the <code>DDE REQUEST</code> language statement. There is always
 * a following {@link #expr}, {@link #source_or_target_clause} and
 * {@link #item_clause}. There is an optional use of {@link #time_clause}.
 * Used by {@link #stmt_list}.
 */
dde_request_stmt
   :
      d:KW_DDE^  { #d.setType(DDE_REQUEST); }
      KW_REQUEST!
      expr
      source_or_target_clause[false]
      item_clause
      (time_clause)?
      (KW_NO_ERROR)?
      DOT!
   ;

/**
 * Matches the <code>DDE SEND</code> language statement. There is always
 * a following {@link #expr}, {@link #source_or_target_clause} and
 * {@link #item_clause}. There is an optional use of {@link #time_clause}.
 * Used by {@link #stmt_list}.
 */
dde_send_stmt
   :
      d:KW_DDE^  { #d.setType(DDE_SEND); }
      KW_SEND!
      expr
      source_or_target_clause[false]
      item_clause
      (time_clause)?
      (KW_NO_ERROR)?
      DOT!
   ;

/**
 * Matches the <code>DDE TERMINATE</code> language statement. There is always
 * a following {@link #expr}. Used by {@link #stmt_list}.
 */
dde_terminate_stmt
   :
      d:KW_DDE^  { #d.setType(DDE_TERMINATE); }
      KW_TERMINAT!
      expr
      (KW_NO_ERROR)?
      DOT!
   ;
   
/**
 * Matches <code>TIME</code> clause followed by an {@link #expr}. Used by
 * {@link #dde_advise_stmt}, {@link #dde_execute_stmt}, {@link #dde_get_stmt},
 * {@link #dde_request_stmt} and {@link #dde_send_stmt}.
 */
time_clause
   :
      KW_TIME^ expr
   ;
   
/**
 * Matches <code>ITEM</code> clause followed by an {@link #expr}. Used by
 * {@link #dde_advise_stmt}, {@link #dde_get_stmt}, {@link #dde_request_stmt}
 * and {@link #dde_send_stmt}.
 */
item_clause
   :
      KW_ITEM^ expr
   ;
   
/**
 * Matches <code>COMMAND</code> clause followed by an {@link #expr}. Used by
 * {@link #dde_execute_stmt}.
 */
command_clause
   :
      KW_COMMAND^ expr
   ;
   
/**
 * Matches <code>FRAME</code> clause followed by a {@link #handle}.
 * Used by {@link #dde_initiate_stmt}.
 */
frame_handle_clause
   :
      KW_FRAME^ handle
   ;

/**
 * Matches <code>APPLICATION</code> clause followed by an {@link #expr}.
 * Used by {@link #dde_initiate_stmt}.
 */
application_clause
   :
      KW_APPL^ expr
   ;
   
/**
 * Matches <code>TOPIC</code> clause followed by an {@link #expr}.
 * Used by {@link #dde_initiate_stmt}.
 */
topic_clause
   :
      KW_TOPIC^ expr
   ;
   
/**
 * Matches the <code>SYSTEM-DIALOG COLOR</code> statement and all following
 * options. Uses {@link #expr}, {@link #update_logical_clause} and
 * {@link #in_window_handle_clause}.  Called by {@link #stmt_list}.
 */
system_dialog_color_stmt
   :
      sd:KW_SYS_DLG^  { #sd.setType(SYSTEM_DIALOG_COLOR); }
      KW_COLOR! expr
      (
           update_logical_clause
         | in_window_handle_clause
      )*
      DOT!
   ;

/**
 * Matches the <code>SYSTEM-DIALOG FONT</code> statement and all following
 * options. Uses {@link #expr}, {@link #update_logical_clause},
 * {@link #min_max_size_clause} and {@link #in_window_handle_clause}.
 * <p>
 * Called by {@link #stmt_list}.
 */
system_dialog_font_stmt
   :
      sd:KW_SYS_DLG^  { #sd.setType(SYSTEM_DIALOG_FONT); }
      KW_FONT! expr
      (
           KW_ANSI_ONL
         | KW_FIXD_ONL
         | min_max_size_clause
         | update_logical_clause
         | in_window_handle_clause
      )*
      DOT!
   ;

/**
 * Matches the <code>SYSTEM-DIALOG GET-DIR</code> statement and all following options. Uses
 * {@link #lvalue}, {@link #initial_dir_clause} and {@link #title_clause}.
 * <p>
 * This can also match an {@link #update_logical_clause} and/or {@link #in_window_handle_clause}
 * which are both undocumented features which were found in customer applications. 
 * <p>
 * Called by {@link #stmt_list}.
 */
system_dialog_get_dir_stmt
   :
      sd:KW_SYS_DLG^  { #sd.setType(SYSTEM_DIALOG_GET_DIR); }
      KW_GET_DIR! lvalue
      (
           KW_RET_2SD
         | initial_dir_clause
         | title_clause
         | update_logical_clause   // undocumented
         | in_window_handle_clause // undocumented
      )*
      DOT!
   ;
   
/**
 * Matches the <code>SYSTEM-DIALOG GET-FILE</code> statement and all
 * following options. Uses:
 * <p>
 * <ul>
 *   <li> {@link #lvalue}
 *   <li> {@link #update_logical_clause},
 *   <li> {@link #filters_clause}
 *   <li> {@link #title_clause}
 *   <li> {@link #default_extension_clause}
 *   <li> {@link #initial_dir_clause}
 *   <li> {@link #in_window_handle_clause}
 * </ul>
 * <p>
 * Called by {@link #stmt_list}.
 */
system_dialog_get_file_stmt
   :
      sd:KW_SYS_DLG^  { #sd.setType(SYSTEM_DIALOG_GET_FILE); }
      KW_GET_FILE! lvalue
      (
           KW_ASK_OVER
         | KW_CREAT_TF
         | KW_MUST_EXI
         | KW_RET_2SD
         | KW_SAVE_AS
         | KW_USE_FIL
         | default_extension_clause
         | initial_dir_clause
         | filters_clause
         | title_clause
         | update_logical_clause
         | in_window_handle_clause
      )*
      DOT!
   ;

/**
 * Matches the <code>SYSTEM-DIALOG PRINTER-SETUP</code> statement and all
 * following options. Uses {@link #expr}, {@link #update_logical_clause},
 * {@link #num_copies_clause} and {@link #in_window_handle_clause}.
 * <p>
 * Called by {@link #stmt_list}.
 */
system_dialog_printer_setup_stmt
   :
      sd:KW_SYS_DLG^  { #sd.setType(SYSTEM_DIALOG_PRINTER_SETUP); }
      KW_PRT_SET!
      (
           KW_LANDSCAP
         | KW_PORTRAIT 
         | num_copies_clause
         | update_logical_clause
         | in_window_handle_clause
      )*
      DOT!
   ;
   
/**
 * Matches <code>UPDATE</code> clause followed by an {@link #lvalue}.
 * Used by {@link #system_dialog_color_stmt}, 
 * {@link #system_dialog_font_stmt}, {@link #system_dialog_get_file_stmt} and
 * {@link #system_dialog_printer_setup_stmt}.
 */
update_logical_clause
   :
      KW_UPDATE^ lvalue
   ;
   
/**   
 * Matches the <code>IN WINDOW</code> clause followed by a {@link #handle}.
 * Used by {@link #system_dialog_color_stmt}, 
 * {@link #system_dialog_font_stmt}, {@link #system_dialog_get_file_stmt} and
 * {@link #system_dialog_printer_setup_stmt}.
 */
in_window_handle_clause   
   :
      KW_IN^ KW_WINDOW handle
   ;
   
/**   
 * Matches the <code>MAX-SIZE or MIN-SIZE</code> clause followed by an
 * {@link #expr}.  Used by {@link #system_dialog_font_stmt}.
 */
min_max_size_clause
   :
      (KW_MAX_SZ^ | KW_MIN_SZ^) expr
   ;
   
/**   
 * Matches the <code>FILTERS</code> clause followed by a list of pairs of
 * character expressions matched by {@link #filter_spec}. Optionally uses
 * {@link #initial_filter_num}. Used by {@link #system_dialog_get_file_stmt}.
 */
filters_clause
   :
      KW_FILTERS^ filter_spec (COMMA! filter_spec)* (initial_filter_num)?
   ;
   
/**
 * Matches a pair of {@link #expr} and roots both of them at a node of type
 * <code>FILTER_SPEC</code>. Used by {@link #filters_clause}. 
 */
filter_spec
   :
      {
         astFactory.makeASTRoot(currentAST, #[FILTER_SPEC,""]);
      }
      expr expr
   ;

/**
 * Matches <code>INITIAL-FILTER</code> clause followed by an {@link #expr}.
 * Used by {@link #filters_clause}. 
 */
initial_filter_num
   :
      KW_INIT_FLT^ expr
   ;
   
/**   
 * Matches the <code>DEFAULT-EXTENSION</code> clause followed by an 
 * {@link #expr}. Used by {@link #system_dialog_get_file_stmt}.
 */
default_extension_clause   
   :
      KW_DEF_EXTN^ expr
   ;
   
/**   
 * Matches the <code>INITIAL-DIR</code> clause followed by an {@link #expr}. 
 * Used by {@link #system_dialog_get_file_stmt}.
 */
initial_dir_clause
   :
      KW_INIT_DIR^ expr
   ;
   
/**   
 * Matches the <code>TITLE</code> clause followed by an {@link #expr}.  Used
 * by {@link #system_dialog_get_file_stmt}.
 */
title_clause   
   :
      KW_TITLE^ expr
   ;

/**   
 * Matches the <code>NUM-COPIES</code> clause followed by an {@link #expr}.
 * Used by {@link #system_dialog_printer_setup_stmt}.
 */
num_copies_clause
   :
      KW_NUM_COPY^ expr
   ;
   
/**
 * Matches the <code>SYSTEM-HELP</code> statement and all following options.
 */
system_help_stmt   
   :
      KW_SYS_HELP^ expr (window_name_clause)? help_command_clause DOT!
   ;
   
/**
 * Matches <code>WINDOW-NAME</code> clause followed by an {@link #expr}.
 * Used by {@link #system_help_stmt}.
 */
window_name_clause
   :
      KW_WIN_NAME^ expr
   ;
   
/**
 * Matches all the various commands that can be passed to the {@link #system_help_stmt}. Uses
 * {@link #expr}, {@link #position_clause_arg_x}, {@link #position_clause_arg_y},
 * {@link #position_clause_arg_width} and {@link #position_clause_arg_height}.
 */
help_command_clause
   :
        KW_CONTENTS^
      | KW_FINDER^
      | KW_POS^ KW_MAXIMIZE
      | KW_FORCE_F^
      | KW_HELP^
      | KW_QUIT^
      | (
             KW_CTX^
           | KW_SET_CONT^
           | KW_CTX_POP^
           | KW_KEY^
           | KW_PART_KEY^
           | KW_COMMAND^
           | KW_HELP_TOP^
           | KW_ALT_KEY^
        )
        expr
      | KW_MULT_KEY^ expr KW_TEXT expr
      | KW_POS^ position_clause_arg_x
                position_clause_arg_y
                position_clause_arg_width
                position_clause_arg_height
   ;
   
/**
 * Matches {@code X} and a following {@link #expr}.  Used by {@link #help_command_clause}.
 */
position_clause_arg_x
   :
      KW_X^ expr
   ;
   
/**
 * Matches {@code Y} and a following {@link #expr}.  Used by {@link #help_command_clause}.
 */
position_clause_arg_y
   :
      KW_Y^ expr
   ;
   
/**
 * Matches {@code WIDTH} and a following {@link #expr}.  Used by {@link #help_command_clause}.
 */
position_clause_arg_width
   :
      KW_WIDTH^ expr
   ;
   
/**
 * Matches {@code HEIGHT} and a following {@link #expr}.  Used by {@link #help_command_clause}.
 */
position_clause_arg_height
   :
      KW_HEIGHT_C^ expr
   ;
   
   
/**
 * Matches a <code>CREATE BROWSE</code> language statement, and the
 * following options. 
 * <p>
 * Uses:
 * <p>
 * <ul>
 *    <li> {@link #handle}
 *    <li> {@link #in_widget_pool_clause}
 *    <li> {@link #attribute_assign_clause}
 *    <li> {@link #trigger_phrase}
 * </ul>
 * <p>
 * Used by {@link #stmt_list}.
 */
create_browse_stmt
   {
      boolean allowColon = false;
   }
   :
      c:KW_CREATE^  { #c.setType(CREATE_BROWSE); }
      KW_BROWSE! handle
      (in_widget_pool_clause)?
      (attribute_assign_clause)?
      (trigger_phrase { allowColon = true; })?
      (
           { allowColon }?
           COLON!
         | DOT!
      )
   ;
   
/**
 * Matches a <code>CREATE SOAP-HEADER</code> language statement, and the
 * following options. 
 * <p>
 * Uses:
 * <p>
 * <ul>
 *    <li> {@link #handle}
 *    <li> {@link #in_widget_pool_clause}
 * </ul>
 * <p>
 * Used by {@link #stmt_list}.
 */
create_soap_header_stmt
   :
      c:KW_CREATE^  { #c.setType(CREATE_SOAP_HEADER); }
      KW_SOAP_HDR! handle
      (in_widget_pool_clause)?
      DOT!
   ;
   
/**
 * Matches a <code>CREATE SOAP-HEADER</code> language statement, and the
 * following options. 
 * <p>
 * Uses:
 * <p>
 * <ul>
 *    <li> {@link #handle}
 *    <li> {@link #in_widget_pool_clause}
 * </ul>
 * <p>
 * Used by {@link #stmt_list}.
 */
create_soap_header_entryref_stmt
   :
      c:KW_CREATE^  { #c.setType(CREATE_SOAP_HEADER_ENTRYREF); }
      KW_SOAP_HER! handle
      (in_widget_pool_clause)?
      DOT!
   ;
   
/**
 * Matches a <code>CREATE SERVER</code> language statement, the following
 * {@link #handle} and the optional {@link #attribute_assign_clause}.
 * <p>
 * Used by {@link #stmt_list}.
 */
create_server_stmt
   :
      c:KW_CREATE^  { #c.setType(CREATE_SERVER); }
      KW_SERVER! handle (attribute_assign_clause)? DOT!
   ;
   
/**
 * Matches a <code>CREATE SERVER-SOCKET or CREATE SOCKET</code> language
 * statements and the required following {@link #handle}. 
 * <p>
 * Used by {@link #stmt_list}.
 */
create_socket_stmt
   :
      c:KW_CREATE^
      (
           KW_SRV_SOCK!   { #c.setType(CREATE_SRV_SOCKET); }
         | KW_SOCKET!     { #c.setType(CREATE_SOCKET); }
      )
      handle (KW_NO_ERROR)? DOT!
   ;
   
/**
 * Matches a <code>CREATE CLIENT-PRINCIPAL</code> language statement and the
 * required following {@link #handle}. 
 * <p>
 * Used by {@link #stmt_list}.
 */
create_client_principal_stmt
   :
      c:KW_CREATE^
      KW_CLNT_PRL! { #c.setType(CREATE_CLIENT_PRINCIPAL); }
      handle DOT!
   ;
   
/**
 * Matches a <code>CREATE X-DOCUMENT or CREATE X-NODEREF</code> language
 * statements and the required following {@link #handle}.  Working code has
 * been found that allows a <code>KW_NO_ERROR</code> to optionally be present
 * (in both statements) and this undocumented feature is supported here.
 * Also uses {@link #in_widget_pool_clause}.
 * <p>
 * Used by {@link #stmt_list}.
 */
create_xml_stmt
   :
      c:KW_CREATE^
      (
           KW_X_DOC!   { #c.setType(CREATE_X_DOCUMENT); }
         | KW_X_NODE!  { #c.setType(CREATE_X_NODEREF); }
      )
      handle (in_widget_pool_clause)? (KW_NO_ERROR)? DOT!
   ;
   
/**
 * Matches a <code>CREATE SAX-READER</code>, <code>CREATE SAX-WRITER</code>
 * and <code>CREATE SAX-ATTRIBUTES</code> language statements and the required
 * following {@link #handle}. Also uses {@link #in_widget_pool_clause}. 
 * <p>
 * Used by {@link #stmt_list}.
 */
create_sax_stmt   
   :
      c:KW_CREATE^
      (
           KW_SAX_ATTR!   { #c.setType(CREATE_SAX_ATTRIBUTES); }
         | KW_SAX_READ!   { #c.setType(CREATE_SAX_READER); }
         | KW_SAX_WRIT!   { #c.setType(CREATE_SAX_WRITER); }
      )
      handle (in_widget_pool_clause)? (KW_NO_ERROR)? DOT!
   ;
   
/**
 * Matches a <code>CREATE CALL</code> language statement, the required
 * following {@link #handle} and the optional {@link #in_widget_pool_clause}.
 * <p>
 * Used by {@link #stmt_list}.
 */
create_call_stmt   
   :
      c:KW_CREATE^  { #c.setType(CREATE_CALL); }
      KW_CALL! 
      handle
      (in_widget_pool_clause)?
      (KW_NO_ERROR)?
      DOT!  
   ;
   
/**
 * Matches a <code>DELETE PROCEDURE</code> language statement, and the
 * following required {@link #expr}. 
 * <p>
 * Used by {@link #stmt_list}.
 */
delete_procedure_stmt   
   :
      d:KW_DELETE^  { #d.setType(DELETE_PROCEDURE); }
      KW_PROC! expr (KW_NO_ERROR)? DOT!
   ;      
   
/**
 * Matches a <code>CREATE WIDGET-POOL</code> language statement, and the
 * following optional {@link #expr} and/or <code>KW_PERSIST</code>. It is
 * undocumented, but working 4GL code shows that it is possible to match the
 * <code>KW_PERSIST</code> for an unnamed pool (where there is no preceding 
 * <code>expr</code>).
 * <p>
 * Used by {@link #stmt_list}.
 */
create_widget_pool_stmt   
   :
      c:KW_CREATE^  { #c.setType(CREATE_WIDGET_POOL); }
      KW_WID_POOL! (expr)? (KW_PERSIST)? (KW_NO_ERROR)? DOT!
   ;      
   
/**
 * Matches a <code>DELETE WIDGET-POOL</code> language statement, and the
 * following optional {@link #expr}. 
 * <p>
 * Used by {@link #stmt_list}.
 */
delete_widget_pool_stmt   
   :
      d:KW_DELETE^  { #d.setType(DELETE_WIDGET_POOL); }
      KW_WID_POOL! (expr)? (KW_NO_ERROR)? DOT!
   ;   
   
/**   
 * Matches the <code>IN WIDGET-POOL</code> keywords and the required following
 * pool name as a character expression (see {@link #expr}). 
 * <p>
 * Used by {@link #create_query_stmt}, {@link #create_temp_table_stmt},
 * {@link #create_buffer_stmt}, {@link #create_widget_or_object_stmt} and
 * {@link #create_browse_stmt}.
 * <p>
 * This is a separate rule to centralize logic and build the AST properly.
 */   
in_widget_pool_clause   
   :
      i:KW_IN^  { #i.setType(IN_WIDGET_POOL); }
      KW_WID_POOL! expr
   ;
   
/**
 * Matches a <code>DEFINE TEMP-TABLE or DEFINE WORK-TABLE</code> Progress 4GL
 * language statement (note that <code>WORKFILE</code> is a synonym for
 * <code>WORK-TABLE</code> and is handled as such. This is the statement that
 * defines or imports new user-named temp-tables and work-tables for a
 * procedure (internal or external).  
 * <p>
 * This rule is called by {@link #define_stmt} which is where the matching
 * on the <code>DEFINE</code> keyword and the optional shared scope for
 * variables, temp-tables and streams.  This rule supports shared, global
 * shared or local variables and the syntax for both declaring or defining
 * (<code>NEW</code> keyword) shared variables.  HOWEVER, all the matching
 * for these optional keywords is done in the <code>define_stmt</code> rule. 
 * This refactoring allows all <code>DEFINE</code> statements to be supported
 * without ambiguity.
 * <p> 
 * This rule matches the <code>TEMP-TABLE, WORK-TABLE or WORKFILE</code>
 * keywords and then calls other rules to implement the different
 * alternatives.
 * <p>
 * Since field name references within this statement can be unambiguous
 * even when they are ambiguous outside of this statement, a schema scope
 * is added and removed at the beginning and end of this rule.  Then anytime
 * fields are added to the table, the definition in this scope is refreshed
 * which makes the latest definition available to subsequent matching in
 * this rule.
 * <p>
 * Calls the following:
 * <p>
 * <ul>
 *    <li> {@link #symbol}
 *    <li> {@link #like_clause}
 *    <li> {@link #add_temp_table_field}
 *    <li> {@link #index_clause}
 * </ul>
 *
 * @param    tt_new
 *           <code>true</code> if this is a new shared or new global shared temp-table.
 * @param    tt_global
 *           <code>true</code> if this is a global shared temp-table.
 * @param    tt_shared
 *           <code>true</code> if this is a shared temp-table.
 * @param    am
 *           Access modifiers if not null. Only will be non-null if this is a resource defined
 *           as a member of a class definition.
 * @param    st
 *           Static specifier if not null. Only will be non-null if this is a resource defined
 *           as a member of a class definition.
 */
def_temp_table_stmt [boolean tt_new, boolean tt_global, boolean tt_shared, Aast am, Aast st] 
                    returns [String tablename = null]
   :
      {
         Object table   = null;
         int    tabType = TEMP_TABLE;
         String b4table = null;
         
         sym.addSchemaScope(false);
      }
      (
         (tt:KW_TEMP_TAB^ | wt:KW_WORK_TAB^ { tabType = WORK_TABLE; })
         t:symbol
         {                                              
            Aast node = (#tt == null) ? #wt : #tt;
            if (tt_shared)
            {
               node.putAnnotation("tt_shared", true);
               if (tt_global)
               {
                  node.putAnnotation("tt_global", true);
               }
               if (tt_new)
               {
                  node.putAnnotation("tt_new", true);
               }
            }
            // always add temp-tables/work-tables into the global scope
            table = sym.addTable(#t.getText(), tabType, true, node, am, st);
         }
         
         (
            options { generateAmbigWarnings = false; }
            :
              ku:KW_NO_UNDO  { sym.setTableOption(table, KW_NO_UNDO, #ku); }
            | KW_UNDO // documented "feature" found in customer code
            | label // undocumented "feature" found in customer code
            | kr:KW_RCOD_INF { sym.setTableOption(table, KW_RCOD_INF, #kr); }
            | KW_REF_ONLY
            | xml_node_name
            | serialize_name
            | namespace_clause
            | b4table=bt:before_table_clause
            | like_clause[#t.getText(), table, true]
            | add_temp_table_field[table]
            | { tabType == TEMP_TABLE }?
              index_clause[table]
         )*
      )
      {
         sym.finalizeTableDefine(table);
         sym.deleteSchemaScope(false);
         
         tablename = #t.getText();
         
         // handle the before-table definition which was deferred because it must only be done
         // after the main table is fully defined (otherwise the fields, indexes and so forth
         // may not be present)
         if (b4table != null)
         {
            Object before = sym.addTable(b4table, TEMP_TABLE, true, #bt, null, null);
            Aast[] fields = sym.addFieldsFrom(before, tablename);
            
            for (int i = 0; i < fields.length; i++)
            {
               sym.processLikeField(fields[i], false, tablename + "." + fields[i].getText());
            }
            
            sym.finalizeTableDefine(before);
         }
      }
   ;
   
/**
 * Matches the <code>BEFORE-TABLE</code> clause with the following {@link #symbol}. Used
 * from {@link #def_temp_table_stmt}.
 *
 * @return   The before-table name to define.
 */
before_table_clause returns [String tname = null]
   :
      KW_B4_TABLE^ t:symbol { tname = #t.getText(); }
   ;
   
/**
 * Matches the <code>SERIALIZE-NAME</code> clause with the following string.
 * Used from {@link #def_temp_table_stmt}.
 */
serialize_name
   :
      KW_SERIALZN^ STRING
   ;
   
/**   
 * Matches <code>VALIDATE</code> or <code>USE-INDEX</code> keywords and (in
 * the second case) the following index name plus optional
 * <code>AS PRIMARY</code> construct. Used by the {@link #like_clause} on
 * behalf of the {@link #def_temp_table_stmt} rule.
 * <p>
 * Bogus ANTLR ambiguity warnings have been disabled.
 * <p>
 * Separated to properly build the AST.
 *
 * @param    table
 *           The temp-table object that is being modified.
 */
temp_table_use_index [Object table]  
   :
      {
         boolean primary = false; 
      }
      u:use_index_clause
      (
         options { generateAmbigWarnings = false; }
         :
         KW_AS! KW_PRIMARY   { primary = true; } 
      )?
      {
         Aast child = (Aast) #u.getFirstChild();
         sym.addIndexFrom(table, #u, child.getText(), primary);
      }
   ;   
   
/**
 * Matches a field definition when creating a temp-table using the 
 * <code>DEFINE TEMP-TABLE or DEFINE WORK-TABLE</code> statements. Called
 * by {@link #def_temp_table_stmt} rules.
 * <p>
 * Calls {@link #symbol}, {@link #as_field_clause}, {@link #like_clause}
 * and {@link #temp_table_field_options}.  As an undocumented feature, the
 * <code>temp_table_field_options</code> can be matched before the
 * <code>AS</code> or <code>LIKE</code> clause. This was found in working
 * 4GL code and is supported.
 * <p>
 * This rule parses the field name, type and any options.  The table object
 * that the field is associated with is passed in as a parameter. This is
 * used with the name and type to add the field to the schema namespace.
 * <p>
 * <b>An artificial root node is created to contain all tokens under a single
 * root with type <code>DEFINE_FIELD</code>. This rule uses a trick to create
 * an artificial node as a root while still allowing ANTLR to do normal tree
 * building (no suppression of the AST build).</b> This is only useful in
 * cases where there is no valid token which can be used as a tree root AND
 * where a loop is being used such that the contents of the loop cannot be 
 * referenced in a manually built #([,],,,) ANTLR construct. ANTLR doesn't
 * support (as of 2.7.4) the ability to label a loop such that the entire
 * contents can be referenced. If one labels the individual elements inside
 * the loop, only the last one is available when the loop concludes. 
 * However, by leaving the normal tree creation unchanged but manually 
 * creating an artificial node and manually setting this node as the root
 * of the returned tree BEFORE the children are added, this problem can be
 * solved.
 * <p>
 * Bogus ANTLR ambiguity warnings have been disabled.
 *
 * @param    table
 *           The table object to associate this field with in the schema
 *           namespace.
 */
add_temp_table_field [Object table]   
   :
      {
         astFactory.makeASTRoot(currentAST, #[DEFINE_FIELD,"define field"]);
         
         String fname = null;
         int    ftype = -1;
      }
      
      (KW_FIELD!)?
      f:symbol  { fname = #f.getText(); }
      // undocumented feature that allows some options to be placed before
      // the AS or LIKE clause (the fname and string parms will not be set
      // yet, but for the majority of downstream rules that is OK)
      (
         options { generateAmbigWarnings = false; }
         :
         temp_table_field_options[fname, ftype]
      )?
      (
           a:as_field_clause[fname, table]
           {
              ftype = #a.getFirstChild().getType();
           }
         | l:like_clause[fname, table, false]
           {
              ftype = #l.getFirstChild().getType();
           }
      )
      (
         options { generateAmbigWarnings = false; }
         :
         temp_table_field_options[fname, ftype]
      )?
      {
         // this should be able to find the right field because the temp-tab
         // should be in the closest scope, thus the options will be forced
         // into the field just defined
         sym.setFieldOptions(fname, ##);
      }
   ;
   
/**
 * Matches the <code>AS</code> language keyword that defines the data type
 * of a field to be added to the a temp-table. This rule encapsulates the
 * logic of matching the correct variable data type and adding this field
 * into the temp-table or work-table being defined.
 * <p>
 * This is called from {@link #add_temp_table_field}.
 * <p>
 * The inclusion of the {@link #reserved_or_symbol} rule in the second token
 * position ensures that all possible symbols are considered a possible field
 * type and will be matched by this rule.  This means that the calling rules
 * include a list of reserved and unreserved keywords or generic symbols as
 * one of the options for its 2nd token, thus &quot;pulling&quot; the
 * recursive descent processing down into this rule.
 * <p>
 * It is important to note that in the end, it is not really valid to have
 * a token type of <code>SYMBOL</code>, random keyword match or anything
 * other than one of the valid <b>special abbreviations</b> for a field type.
 * What we really require is a token that is one of the following:
 * <pre>
 *    Data Type Keyword        Actual Minimum Abbrev      Lexer Abbrev
 *    -----------------        ---------------------      -------------
 *    KW_BLOB                  n/a                        n/a
 *    KW_CHAR                  c                          char
 *    KW_CLASS                 n/a                        n/a
 *    KW_CLOB                  n/a                        n/a
 *    KW_COM_HNDL              n/a                        n/a
 *    KW_DATE                  da                         n/a
 *    KW_DATETIME              n/a                        n/a
 *    KW_DATE_TZ               n/a                        n/a
 *    KW_DEC                   de                         dec
 *    KW_HANDLE                handle                     n/a
 *    KW_INT                   i                          int
 *    KW_INT64                 n/a                        n/a
 *    KW_LOGICAL               l                          n/a
 *    KW_LONGCHAR              NOT SUPPORTED AS A FIELD!!!!!!!!!!!!!!!
 *    KW_MEMPTR                NOT SUPPORTED AS A FIELD!!!!!!!!!!!!!!!
 *    KW_RAW                   ra                         n/a
 *    KW_RECID                 re                         n/a
 *    KW_ROWID                 row                        n/a
 *    KW_WID_HAND              widg                       widget-h
 * </pre>
 * <p>
 * The difference between a <code>SYMBOL</code> and a <code>KW_CHAR</code>
 * can only be determined by context (the fact that the lookahead code has
 * identified a match with a <code>SYMBOL</code> token which is a construct 
 * that can occur in an this rule).  Thus <b>if</b> a match is found using
 * lookahead (in the rule that calls this one), then it <b>must</b> be
 * correct and necessary to convert the <code>SYMBOL</code> or other keyword
 * token type into the correct keyword representing the field type, using a
 * special abbreviation test.  This is done in an init action that is 
 * processed at the beginning of this rule. In this way, the rest of the rule
 * only sees the 2nd token as a field token type and never tries to
 * call the <code>reserved_or_symbol</code> rule since calling that rule 
 * would actually try to match (and consume from the token stream) a
 * <code>SYMBOL</code> or other keyword token type.  <b>Specifying a rule
 * reference to obtain its value from a lookahead perspective but then
 * changing the token's type such that the rule reference never gets called, 
 * is the critical trick that enables context sensitive symbol names!</b>
 * <p>
 * This use of a pull rule is only necessary because Progress provides
 * additional special abbreviation support in this construct. This is a set
 * of reduced abbreviations (see table above) beyond the normal (documented
 * or undocumented) abbreviations of a field type keyword as is possible 
 * elsewhere in Progress source code.  The <code>reserved_or_symbol</code> 
 * is needed because some possible reserved keyword matches could occur 
 * that would match one of these special abbreviations and thus the
 * <code>symbol</code> rule would not be sufficient to pull all possible
 * valid keyword matches into this rule.
 * <p>
 * The init action reads the text of the 2nd token and switches the token
 * type to that of the matching field type keyword, if any special
 * abbreviation is found.  All such processing is bypassed if the lexer
 * has already matched a valid field type keyword.
 * <p>
 * The logic of the special abbreviation processing is centralized in the
 * specialDataTypeKeywordAbbreviations() method.
 *
 * @param   symName
 *          The name of the field to add.
 * @param   table
 *          The table in which the field should be added.
 */
as_field_clause [String symName, Object table]   
   :
      {
         int ftype = -1;
         specialDataTypeKeywordAbbreviations(2);
      }
      (
         KW_AS^ 
         ( 
            options { generateAmbigWarnings = false; }
            :
              KW_BLOB                        { ftype = FIELD_BLOB; }
            | KW_CHAR                        { ftype = FIELD_CHAR; }
            | KW_CLOB                        { ftype = FIELD_CLOB; }
            | KW_COM_HNDL                    { ftype = FIELD_COM_HANDLE; }
            | KW_DATE                        { ftype = FIELD_DATE; }
            | KW_DATETIME                    { ftype = FIELD_DATETIME; }
            | KW_DATE_TZ                     { ftype = FIELD_DATETIME_TZ; }
            | KW_DEC                         { ftype = FIELD_DEC; }
            | KW_HANDLE                      { ftype = FIELD_HANDLE; }
            | KW_INT                         { ftype = FIELD_INT; }
            | KW_INT64                       { ftype = FIELD_INT64; }
            | KW_LOGICAL                     { ftype = FIELD_LOGICAL; }
            | KW_RAW                         { ftype = FIELD_RAW; }
            | KW_RECID                       { ftype = FIELD_RECID; }
            | KW_ROWID                       { ftype = FIELD_ROWID; }
            | KW_WID_HAND                    { ftype = FIELD_HANDLE; }
            
            
              // this is for "pull" and to match the class type (it will
              // pull for the LA2 and LA3 because of the optional CLASS
              // keyword
            | (
                 options { generateAmbigWarnings = false; }
                 :
                 KW_CLASS!
              )?
              user_defined_type_name         { ftype = FIELD_CLASS; }
         )
      )
      {
         sym.addField(table, symName, ftype, ##);
         sym.promoteTable(table, true, false);
      }
   ;   
   
/**   
 * Matches the field options that can be specified when creating a temp-table
 * using the <code>DEFINE TEMP-TABLE or DEFINE WORK-TABLE</code> statements.
 * Called by {@link #add_temp_table_field}.
 * <p>
 * All options can be specified in any order.
 * <p>
 * Calls:
 * <p>
 * <ul>
 *    <li> {@link #extent}
 *    <li> {@link #initializer}
 *    <li> {@link #label}
 *    <li> {@link #ui_stuff}
 *    <li> {@link #format_string}
 *    <li> {@link #help_string}
 *    <li> {@link #column_label}
 *    <li> {@link #decimals_clause}
 *    <li> {@link #view_as_phrase}
 *    <li> {@link #case_sensitive}
 * </ul>
 * <p>
 * Bogus ANTLR ambiguity warnings have been disabled.
 *
 * @param    symText
 *           The name of the symbol being added which is passed to the
 *           view-as phrase so that the widget namespace can be maintained
 *           properly.
 * @param    ftype
 *           A token type that describes type of the field being created.  This is the type
 *           of the child of the <code>KW_AS</code> or <code>KW_LIKE</code> node.
 */
temp_table_field_options [String symText, int ftype]   
   :
      {
         int size = -1;
      }
      (
         options { generateAmbigWarnings = false; }
         :
           KW_TTCP
         | column_codepage 
         | size = extent
         | initializer[ftype]
         | label
         | ui_stuff
         | format_string[false]
         | help_string
         | column_label
         | decimals_clause
         | case_sensitive
         | KW_SERIALZH
         | serialize_name
         | xml_data_type
         | xml_node_type
         | xml_node_name
         | view_as_phrase[symText]
      )*
   ;
   
/**
 * Matches the <code>COLUMN-CODEPAGE</code> option with a following
 * {@link #expr}.  Used by {@link #temp_table_field_options}.
 */
column_codepage
   :
      KW_COL_CP^ expr
   ;
   
/**
 * Matches <code>XML-NODE-TYPE</code> with a following <code>STRING</code>. Used by 
 * {@link #temp_table_field_options} and {@link #def_dataset_stmt}.
 */
xml_node_type
   :
      KW_XML_NTYP^ STRING
   ;
   
/**
 * Matches <code>XML-DATA-TYPE</code> with a following <code>STRING</code>. Used by 
 * {@link #temp_table_field_options}.
 */
xml_data_type
   :
      KW_XML_DTYP^ STRING
   ;
   
/**
 * Matches <code>XML-NODE-NAME</code> and a following <code>STRING</code>. Used by
 * {@link #def_temp_table_stmt}.
 */
xml_node_name
   :
      KW_XML_NNAM^ STRING
   ;
/**
 * Matches an index definition of a <code>DEFINE TEMP-TABLE</code> language
 * statement.  Calls {@link #symbol}, {@link #is_clause} and
 * {@link #index_field_clause}.
 * <p>
 * Bogus ANTLR ambiguity warnings have been disabled.
 * <p>
 * Used by {@link #def_temp_table_stmt}.
 *
 * @param    table
 *           The temp-table object that is being modified. 
 */
index_clause [Object table]
   :
      i:KW_INDEX^
      s:any_non_reserved_symbol!
      {
         #i.setType(INDEX); 
         #i.setText(#s.getText());
      }
      (is_clause)*
      (
         options { generateAmbigWarnings = false; }
         :
         index_field_clause[table]
      )+
      {
         sym.addIndex(table, #s.getText(), ##);
      }
   ;
   
/**
 * Matches the <code>IS</code> keyword of a <code>DEFINE TEMP-TABLE</code>
 * language statement and handles the optional index qualifiers.  Note that
 * <b>Progress also allows the <code>AS</code> keyword to be substituted for
 * <code>IS</code> although this is undocumented!</b>  This rule supports
 * both forms but the token type is always set to <code>KW_IS</code> to
 * simplify tree processing.
 * <p>
 * Bogus ANTLR ambiguity warnings have been disabled.
 * <p>
 * Used by {@link #index_clause}.
 */
is_clause
   :
      (
           i:KW_IS^ { #i.setType(PROPERTIES); }
         | k:KW_AS^ { #k.setType(PROPERTIES); }
      ) 
      (
         options { generateAmbigWarnings = false; }
         :
           KW_UNIQUE
         | KW_PRIMARY
         | KW_WORD_IDX
      )*
   ;
   
/**   
 * Matches the index field clause of a <code>DEFINE TEMP-TABLE</code> language statement.  Calls
 * {@link #symbol} and optionally {@link #sql_direction_spec} and roots everything under an
 * artificial token <code>INDEX_FIELD</code>.
 * <p>
 * We must NOT call {@link #lvalue} here, because the 4GL has very tight restrictions on what can
 * match and using lvalue does a full schema-namespace lookup that can have ambiguity which the
 * 4GL would not detect.  The symbol is rewritten and annotated to be a field, so that the result
 * is the same as if we has used <code>lvalue</code>. The 4GL restrictions:
 * <p>
 * <ul>
 *   <li> must be a field in the current temp-table/work-table definition
 *   <li> cannot be a qualified name
 *   <li> cannot be abbreviated
 *   <li> cannot reference an unsubscripted extent field
 *   <li> cannot reference a subscripted extent field
 * </ul>
 * <p>
 * Used by {@link #index_clause}.
 *
 * @param    table
 *           The temp-table object that is being modified. 
 */
index_field_clause [Object table]   
   :
      { astFactory.makeASTRoot(currentAST, #[INDEX_FIELD,"index field"]); }
      
      s:symbol
      {
         sym.annotateField(#s, table, false, false);
      }
      
      (
         options { generateAmbigWarnings = false; }
         :
         sql_direction_spec
      )?
   ;
   
/**
 * Matches a Progress 4GL database reference that refers to a valid database
 * or an alias to a valid database.
 * <p>
 * The matching portion of this rule is defined as a single token that
 * matches the union of the <code>DATABASE</code> token type and the set of
 * all possible alternatives for symbol names (a reference to the 
 * {@link #symbol} rule).
 * <p>
 * Note that if this rule is used as an alternative to a rule that includes a
 * leftmost token defined by the <code>symbol</code> rule, there will be an
 * ambiguity that is only resolved by lookahead <b>and</b> the proper
 * ordering of the rules.
 * <p>
 * It is required that where this rule is referenced as one alternative that
 * conflicts with the definitions of other alternatives, <b>the order of the 
 * alternatives must be more specific (in terms of matching criteria) to less
 * specific</b>.
 * <p>
 * The inclusion of the <code>symbol</code> rule ensures that all valid
 * symbols are considered a database and will be matched by this rule.
 * <p>
 * It is important to note that in the end, it is not really valid to have
 * a token type of <code>SYMBOL</code> since this could be anything.  What we
 * really require is a token that is type <code>DATABASE</code>.
 * <p>
 * The difference between a <code>SYMBOL</code> and a <code>DATABASE</code>
 * can only be determined by context (the fact that the lookahead code has
 * identified a match with a <code>SYMBOL</code> token which is a construct
 * that can occur in a <code>database</code> call.  Thus <b>if</b> a match is
 * found using lookahead (in the rule that calls this one), then it
 * <b>must</b> be correct and necessary to convert the <code>SYMBOL</code> 
 * token type into the correct type. This is done in an init action that is 
 * processed at the beginning of this rule. In this way, the rest of the rule 
 * only sees the first token as a valid <code>DATABASE</code> and never tries
 * to call the <code>symbol</code> rule since calling that rule would actually 
 * try to match (and consume from the token stream) a <code>SYMBOL</code>
 * token type.  <b>Specifying a rule reference to obtain its value from a 
 * lookahead perspective but then changing the token's type such that the
 * rule reference never gets called, is the critical trick that enables
 * context sensitive symbol names!</b>
 * <p>
 * This rule can only be used in cases where it is non-optional OR there is
 * no ambiguity (between the set of unreserved keywords referenced herein and 
 * any following keywords that would be improperly matched by this rule).
 * In this case a semantic predicate could be used to isolate the conflicting
 * token types and bypass this rule reference in those cases.
 */
database 
   :
      {
         if (LA(1) != DATABASE)
         {
            int newtype = sym.lookupDatabase(LT(1).getText());
            
            // there should always be a match in valid Progress 4GL source
            // so we throw an exception if this is not the case
            if ( newtype != -1 )
            {
               LT(1).setType( newtype );
            }
            else
            {
               throw new NoViableAltException(LT(1), getFilename());
            }
         }
      }
      (
           DATABASE
           
           // this represents the SYMBOL token and the full range of
           // unreserved keywords which exists to "pull" us into this
           // rule, but when we leave the type is always reset to DATABASE
           // OR we throw an exception (note that symbol can pull in a           
           // DB_SYMBOL type which is NEVER a database name and will
           // generate an exception --> this is OK and correct!)
         | symbol
      )
   ;
   
/**
 * Matches a database name token that can be a string, a character expression
 * that is evaluated via the <code>VALUE( expr )</code> construct, or it
 * can be a symbol that represents a physical filename or a logical database
 * name.  These last two options are differentiated by a boolean passed from
 * the caller.  Used by {@link #connect_stmt}, {@link #create_alias_stmt}
 * and {@link #delete_alias_stmt} rules.
 *
 * @param    physical
 *           If true, the name could be a physical filename, otherwise it is
 *           a logical database name.
 */
database_name [boolean physical]   
   :
      (
         options { generateAmbigWarnings = false; }
         :
           STRING^
         | value
         | s:filename
           {
              if (!physical)
              {
                 // convert unquoted database name to a quoted string
                 #s.setType(STRING);
                 #s.setText("\"" + #s.getText() + "\"");
              }
           }
      )
   ;
   
/**
 * Matches any token except <code>NO-ERROR or DOT</code>.  In the
 * {@link #connect_stmt}, database connect options are provided which can
 * conform to any arbitrary syntax.  The options end with either of the
 * above tokens.
 * <p>
 * This rule is nasty because it has to match a list of any kind of token.
 * This is inherently ambiguous, and made worse by the loop.  ANTLR  
 * detects ambiguity but the semantic predicate does resolve it.  As such
 * warnings have been disabled.
 */
connect_options
   :
      {
         astFactory.makeASTRoot(currentAST, #[CONNECT_OPTIONS,"options"]);
         Token tok;
      }
      (
         options { generateAmbigWarnings = false; }
         :
           { tok = LT(1); }
           v:value
           { processHiddenText(tok, #v, true); }
         |
           { (LA(1) != KW_NO_ERROR && LA(1) != DOT) }?
           t:.
           {
              processHiddenText(t, #t, true);
              #t.setType(CONNECT_TEXT);
           }         
      )*   
      { if (##.getNumberOfChildren() == 0) ## = null; }
   ;
   
/**
 * Matches a Progress 4GL record reference that refers to a table in the
 * database schema, or to a buffer/temp-table/work-table that was previously
 * defined or created in the source file.  Any of these references can often
 * be substituted for each other in 4GL source.
 * <p>
 * The matching portion of this rule is defined as a single token that
 * matches the union of the set of all possible record types and the set of
 * all possible alternatives for symbol names (a reference to the 
 * {@link #symbol} rule).
 * <p>
 * Note that if this rule is used as an alternative to a rule that includes a
 * leftmost token defined by the <code>symbol</code> rule, there will be an
 * ambiguity that is only resolved by lookahead <b>and</b> the proper
 * ordering of the rules.
 * <p>
 * It is required that where this rule is referenced as one alternative that
 * conflicts with the definitions of other alternatives, <b>the order of the 
 * alternatives must be more specific (in terms of matching criteria) to less
 * specific</b>.
 * <p>
 * The inclusion of the <code>symbol</code> rule ensures that all valid
 * symbols are considered a record and will be matched by this rule.
 * <p>
 * It is important to note that in the end, it is not really valid to have
 * a token type of <code>SYMBOL</code> since this could be anything.  What we
 * really require is a token that is one of the following:
 * <pre>
 *    TABLE
 *    BUFFER
 *    TEMP_TABLE
 *    WORK_TABLE
 * </pre>
 * <p>
 * The difference between a <code>SYMBOL</code> and a <code>TABLE</code> can
 * only be determined by context (the fact that the lookahead code has
 * identified a match with a <code>SYMBOL</code> token which is a construct
 * that can occur in a <code>record</code> call.  Thus <b>if</b> a match is
 * found using lookahead (in the rule that calls this one), then it
 * <b>must</b> be correct and necessary to convert the <code>SYMBOL</code> 
 * token type into the correct type. This is done in an init action that is 
 * processed at the beginning of this rule. In this way, the rest of the rule 
 * only sees the first token as a valid <code>TABLE</code> and never tries to
 * call the <code>symbol</code> rule since calling that rule would actually 
 * try to match (and consume from the token stream) a <code>SYMBOL</code>
 * token type.  <b>Specifying a rule reference to obtain its value from a 
 * lookahead perspective but then changing the token's type such that the
 * rule reference never gets called, is the critical trick that enables
 * context sensitive symbol names!</b>
 * <p>
 * The exit action saves the fully qualified schema name that was matched
 * in the schema dictionary. This allows downstream processing to access this
 * as an annotation named 'schemaname'.  In addition, the exit action stores
 * the backing construct for this record (this will be the token type TABLE,
 * TEMP_TABLE or WORK_TABLE).  By definition, this will be the same as the
 * token type for every record type except buffers.  This annotation is
 * called 'recordtype'.
 * <p>
 * This rule can only be used in cases where it is non-optional OR there is
 * no ambiguity (between the set of unreserved keywords referenced herein and 
 * any following keywords that would be improperly matched by this rule).
 * In this case a semantic predicate could be used to isolate the conflicting
 * token types and bypass this rule reference in those cases.
 * <p>
 * All record references in the parser call this rule.  This makes it the
 * ideal location to implement the promotion of a table into the current
 * scope, since when this rule is called the record has been referenced.
 * This is done at the end of the init action, when we know that no exception
 * is going to be thrown.  <b>At this time we ALWAYS promote each table 
 * no matter the manner in which we are called and without regard to the
 * type of reference (strong, weak or free reference). This seems to resolve
 * the implicit lookup requirements for nearly all Progress cases. What makes
 * this 'trick' work is that the Progress compiler itself refuses to allow
 * ambiguous names to be referenced (code with ambiguous name references
 * will fail to compile). This means that the 4GL programmer is required
 * to resolve ambiguity.  Those places that look ambiguous are made
 * unambiguous because the table name has been given preference in the name
 * lookups (in the P2J terminology, it was 'promoted'). By always promoting
 * this code resolves these issues, but it can create one problem that a
 * more descriminating promotion policy (which is probably how Progress is
 * internally written) would not have: it is possible that there is a very
 * obscure case where the existence of additional (unexpected from a Progress
 * perspective) promoted tables can make some names ambiguous that otherwise
 * would be unambiguous.</b>  Full details of this obscure case are in the 
 * <a href="package-summary.html#too_much_promotion">Design Document</a>.
 *
 * @param    promote
 *           If <code>true</code>, promote the record's table for future
 *           name resolution.  If <code>false</code>, do not promote the
 *           table.
 * @param    force
 *           If <code>true</code>, forces re-promotion of records that have
 *           already been promoted. <code>false</code> will promote but only
 *           if it hasn't ever been promoted. This flag is only honored if
 *           the <code>promote</code> parameter is <code>true</code>.
 * @param    noProp                   
 *           <code>true</code> to force any promoted nodes to have their
 *           no propagate flag set. This disables propagation of these
 *           nodes across the current scope boundary.
 */
record [boolean promote, boolean force, boolean noProp]
   :
      {
         String tablename = LT(1).getText();
         
         if (LA(1) < BEGIN_RECORDTYPES || LA(1) > END_RECORDTYPES)
         {
            // TODO: honor forceTemp here!
            int newtype = sym.lookupTable(tablename);
            
            // there should always be a match in valid Progress 4GL source
            // so we throw an exception if this is not the case
            if ( newtype != -1 )
            {
               LT(1).setType( newtype );
            }
            else
            {
               // sym.dumpSchema();
               throw new NoViableAltException(LT(1), getFilename());
            }
         }
         
         // if we have gotten this far, then the tablename is valid so we
         // can add this table to the current schema scope using this string
         if (promote)
            sym.promoteTableName(tablename, force, noProp);
      }
      (
           BEGIN_RECORDTYPES..END_RECORDTYPES
           
           // this token reference represents all partially or fully qualified
           // schema names, its token type will be rewritten on entry to this
           // method so this reference is here to "pull" the top level
           // prediction logic into this method --> note that this token type
           // is defined outside of the BEGIN_RECORDTYPES and END_RECORDTYPES
           // range which causes it to be rewritten every time
         | DB_SYMBOL
           
           // this represents the SYMBOL token and the full range of
           // unreserved keywords which exists to "pull" us into this
           // rule (in addition to the <code>DB_SYMBOL</code> type, but when
           // we leave the type is always reset to TABLE OR we throw an
           // exception
         | symbol
      )
      {
         // now store an annotation with the fully qualified name and the
         // shorted unabbreviated unique buffer name
         String fullname = sym.lookupFullTableName(tablename);
         String bufname  = sym.lookupBufferName(tablename, false);
         String dbname   = sym.lookupDatabaseName(tablename);
                  
         if (fullname != null)         
            ##.putAnnotation("schemaname", fullname);
            
         if (bufname != null)
            ##.putAnnotation("bufname", bufname);
            
         if (dbname != null)
            ##.putAnnotation("dbname", dbname);
            
         // store the backing construct (this will be the same as the token
         // type for every record type except buffers, in which case it will
         // be TABLE, TEMP_TABLE or WORK_TABLE)
         int backing = sym.lookupTableRecordType(tablename);
         
         if (backing != -1)
         {
            ##.putAnnotation("recordtype", new Long(backing));
         }
      }
   ;
   
/**   
 * Matches the <code>FIELDS</code> or <code>EXCEPT</code> keywords and the
 * optional following field list which is used to specify the list of fields
 * to include or exclude from a particular table.  Note that due to the
 * natural ambiguity of <code>FIELDS</code> being optionally abbreviated as
 * <code>FIELD</code> and the fact that there is a separate <code>FIELD</code>
 * keyword used for different purposes, the <code>FIELDS</code> keyword has
 * been dropped and only a <code>FIELD</code> keyword will be found.
 * <p>
 * Used by {@link #record_phrase} and the subtree is rooted by the keyword
 * (which is the reason for using a separate rule for something so simple).
 */
field_list
   :
      (
           KW_FIELD^
         | KW_EXCEPT^
      )
      (
         options { generateAmbigWarnings = false; }
         :
         LPARENS! (lvalue)* RPARENS!
      )?
   ;
   
/**   
 * Matches the <code>OUTER-JOIN</code> keyword and the optional prefix keyword
 * <code>LEFT</code> which are used to specify an outer join or left outer
 * join.  The <code>LEFT</code> token (if it exists) is made a child of 
 * <code>OUTER-JOIN</code>.
 * <p>
 * Used by {@link #record_phrase} and the subtree is rooted by the keyword
 * (which is the reason for using a separate rule for something so simple).
 */
join_clause   
   :
      (KW_LEFT)? KW_OUT_JOIN^
   ;
   
/**   
 * Matches the <code>OF</code> keyword and the required following record
 * which is used to specify the table to join on.
 * <p>
 * Used by {@link #record_phrase} and the subtree is rooted by the keyword
 * (which is the reason for using a separate rule for something so simple).
 */
of_clause   
   :
      KW_OF^ record[true, false, false]
   ;
   
/**   
 * Matches the <code>WHERE</code> keyword and the required following
 * expression.  Note that the expression processing that is exposed via the
 * {@link #expr} rule has been modified to support the <code>CONTAINS</code>
 * operator. This operator is specific in usage to the where clause and
 * follows a form of:
 * <pre>
 *    field CONTAINS expression
 * </pre>
 * <p>
 * This support is implemented in {@link #compare_expr} which means that
 * the <code>CONTAINS</code> operator has the same precedence as operators
 * <code>BEGINS, MATCHES</code> and other comparisons. For more details on
 * how <code>CONTAINS</code> works, please see page 960 in the Progress 4GL
 * language reference.  <b>Note that placing the operator support inside the
 * general purpose expression processing was required to resolve 
 * ambiguity!</b> This makes sense since <code>CONTAINS</code> <b>is</b> an 
 * operator and thus its operands are naturally conflicting with other
 * operators.  The ambiguity would be a problem in certain cases (invoving
 * the <code>LBRACKET</code> array syntax in {@link #lvalue} if we attempted
 * to separate the <code>CONTAINS</code> processing into this location.
 * <p>
 * Though undocumented, customer code was found in which the following
 * expression was optional!  This compiled without problem though it is 
 * nonsense (in this case, the <code>WHERE</code> must be ignored). This
 * feature causes ambiguity warnings which have been disabled.
 * <p>
 * Used by {@link #record_phrase} and the subtree is rooted by the keyword
 * (which is the reason for using a separate rule for something so simple).
 */
where_clause   
   :
      KW_WHERE^
      (
         options { generateAmbigWarnings = false; }
         :
         {
           LA(1) != KW_SH_LOCK && LA(1) != KW_EXC_LOCK &&
           LA(1) != KW_NO_LOCK && LA(1) != KW_NO_WAIT
         }?
         expr
      )?
   ;
   
/**   
 * Matches the <code>USE-INDEX</code> keyword and the required following
 * index.  The index is matched as a valid user-defined symbol using the
 * {@link #symbol} rule, however the result's token type is rewritten to be
 * <code>INDEX</code>.
 * <p>
 * Used by {@link #record_phrase} and the subtree is rooted by the keyword
 * (which is the reason for using a separate rule for something so simple).
 */
use_index_clause   
   :
      KW_USE_IDX^ s:any_non_reserved_symbol { #s.setType(INDEX); }
   ;
   
/**   
 * Matches the <code>USING</code> keyword and the required following list
 * of {@link #lvalue} (fields) and optional matching frames.
 * <p>
 * Each field can be optionally prefaced with the <code>INPUT</code> keyword
 * which seems to have no effect (it is silently dropped). This is an
 * undocumented feature of Progress. This parser drops this keyword if it
 * is found.
 * <p>
 * Used by {@link #record_phrase} and the subtree is rooted by the keyword
 * (which is the reason for using a separate rule for something so simple).
 */
using_clause
   :
      KW_USING^
      (KW_INPUT!)?
      (
         options { generateAmbigWarnings = false; }
         :
         frame_reference[false]
      )?
      lvalue
      (
         KW_AND!
         (KW_INPUT!)?
         (
            options { generateAmbigWarnings = false; }
            :
            frame_reference[false]
         )?
         lvalue
      )*
   ;
   
   
/**
 * Matches the <code>DISPLAY</code> language statement and builds the associated AST.
 * <p>
 * The following list of rules are called to properly implement the tree building:
 * <p>
 * <ul>
 *    <li> {@link #lvalue} (for stream references)
 *    <li> {@link #space_or_skip} 
 *    <li> {@link #expr}
 *    <li> {@link #format_phrase}
 *    <li> {@link #aggregate_phrase}
 *    <li> {@link #simple_when_clause} 
 *    <li> {@link #at_base_field_clause}
 *    <li> {@link #frame_phrase}
 * </ul>
 * <p> 
 * The use of the format phrase following an expression is ambiguous since it includes rule
 * references that include options that include expressions.  Thus, the lookahead set for the
 * format phrase is quite large and this naturally can conflict with following constructs in the
 * form item's calling rule(s).  A review of the code does not show any problems and no false
 * positives have been encountered yet. Ambiguity warnings for this case have been disabled but
 * this location bears watching.  Similar ambiguities exist for the <code>UNLESS-HIDDEN</code> 
 * and aggregate phrase.  These have been reviewed and the ambiguity warnings suppressed.
 * <p>
 * Another tricky part of this statement is resolving the ambiguity between the record and field
 * specifications.  The 4GL behavior is that records take precedence when there is a name
 * conflict EXCEPT in the case where there are other form items specified.  This means that
 * records are disambiguated from fields by trying to do a table lookup, if it fails the record
 * is not matched (see {@link #record_spec}) OR even when it is matched, we use lookahead to
 * determine if there are any following form items (in which case it the table form is not
 * matched).  If the table is not matched, the {@link #expr} is checked.  The careful ordering
 * of these options and using a semantic predicate is needed to resolve the ambiguity.  For this
 * reason, ambiguity warnings have been disabled.
 * <p>
 * There is a bad hack in this rule to allow a various options that are supposed to follow the
 * expression, be parsed in any order.  Normally this is done by looping and making these peer
 * alternatives BUT in this case, the format_phrase rule uses a loop internally and this
 * generates hangs in the resulting parser.  So, in this case, we only allow the format phrase 
 * to be executed once each time through the loop.  This bypasses the hang.  Nasty but it works.
 * <p>
 * Called by {@link #stmt_list}.
 */
display_stmt
   :
      {
         String    symbolName  = null;
         Aast      fld         = null;
         Set<Aast> fields      = new HashSet<>();
         String    frame       = "";
         boolean   isTableName = false;
         boolean   isFieldName = false;
         boolean   moreFollows = false;
         boolean   first       = true;
      }
      KW_DISP^
      (
           { LA(1) == KW_STREAM || LA(1) == KW_STRM_HND }? lvalue // used to be stream_reference
         | // empty alternative
      )
      (
         options { generateAmbigWarnings = false; }
         :
         KW_UNL_HID
      )?
      {
         String nxt = LT(1).getText();
         int    flw = LA(2);
      
         // disambiguate the special case of a table name that conflicts with a field name
         // anything (even a literal) other than a very short list of reserved keywords that
         // follows the next token means that it cannot be a table
         
         // this works because the record will be a single token and the other possible
         // following non-form-items are reserved keywords which can be easily checked
         // and can't be used as an lvalue or field name
         moreFollows = (flw != KW_EXCEPT   &&
                        flw != KW_IN       &&
                        flw != KW_WITH     &&
                        flw != KW_NO_ERROR &&
                        flw != DOT);
         isTableName = isRecord(nxt);
         isFieldName = sym.isField(nxt);
      }
      (
           space_or_skip { first = false; }
           
         | { first && isTableName && (!isFieldName || !moreFollows) }?
           record_spec[fields]
           { first = false; }
           
         | (
                // is there a problem here like the var def issue in
                // form_item?
                { LA(2) == KW_AS || LA(2) == KW_LIKE }?
                s:symbol
              | e:expr
           )
           {
              first = false;
              
              int numFP = 0;
              
              // make sure these are reset every time through the loop
              symbolName = null;
              fld        = null;
              
              if (#s != null)
              {
                 symbolName = #s.getText();
                 int type = #s.getType();
                 if (type > BEGIN_FIELDTYPES && type < END_FIELDTYPES)
                 {
                    fields.add(#s);
                 }
              }
              if (#e != null)
              {
                 fld = findFieldNode(#e);
                 if (fld != null)
                 {
                    fields.add(fld);
                 }
              }
              
              #s = null;
              #e = null;
           }
           // WARNING: since format_phrase may be empty and this rule is an
           //          optional loop, the prediction logic for format_phrase
           //          includes all possible following constructs (e.g. DOT)
           //          even though these can't be matched in a format phrase;
           //          this leads to an infinite loop which we avoid by not
           //          allowing more than 2 matches to a format phrase (YES,
           //          there can be 2 format phrases, one before the aggregate
           //          phrase and one after --> just like in the 4GL)
           (
              options { generateAmbigWarnings = false; }
              :
                LPARENS! aggregate_phrase RPARENS!
              | { numFP < 2 }?
                format_phrase[symbolName, fld, true, false]  { numFP++; }
           )*
      )*
      (in_window_clause)?
      (
         fp:frame_phrase[false, false]
         {
            if (frame.length() == 0)
            {
               String name = getFrameName(#fp);
               if (name.length() > 0)
               {
                  frame = name;
               }
            }
         }
      )*
      (KW_NO_ERROR)?
      DOT!
      {
         if (!fields.isEmpty())
         {
            sym.addFrameFields(frame, fields);
         }

         setUseDictExps(frame, #fp, ##);
      }
   ;
   
/**
 * Matches the <code>SCROLL</code> language statement.  Calls the
 * {@link #frame_phrase}.
 * <p>
 * Used by {@link #stmt_list}.
 */
scroll_stmt
   :
      KW_SCROLL^ (KW_FROM_CUR)? (KW_UP | KW_DOWN)? 
      (frame_phrase[false, false])? DOT!
   ;
      
/**   
 * Matches the <code>CHOOSE</code> language statement, the required following
 * {@link #choose_row_or_field_list} and all possible options (which can be
 * specified in any order).
 * <p>
 * Calls:
 * <p>
 * <ul>
 *    <li> {@link #choose_keys_clause}
 *    <li> {@link #go_on_clause}
 *    <li> {@link #pause_clause}
 *    <li> {@link #simple_color_clause}
 *    <li> {@link #frame_phrase}
 * </ul>
 * <p>
 * Bogus ANTLR ambiguity warnings have been disabled.
 * <p>
 * Used by {@link #choose_stmt}.
 */
choose_stmt      
   :
      KW_CHOOSE^ choose_row_or_field_list
      (
         options { generateAmbigWarnings = false; }
         :
           KW_AUTO_RET
         | KW_NO_ERROR
         | choose_keys_clause
         | go_on_clause
         | pause_clause
         | simple_color_clause
         | frame_phrase[false, false]
      )*
      DOT!
   ;
   
/**   
 * Matches the <code>ROW or FIELD</code> keyword and the required following
 * {@link #choose_field_spec}.
 * <p>
 * Bogus ANTLR ambiguity warnings have been disabled.
 * <p>
 * Used by {@link #choose_stmt}.
 */
choose_row_or_field_list   
   :
      (
         options { generateAmbigWarnings = false; }
         :
           KW_ROW^ choose_field_spec
         | KW_FIELD^ (choose_field_spec)+
      )
   ;
      
/**   
 * Matches an {@link #lvalue} and an optional following 
 * {@link #choose_help_clause}.
 * <p>
 * Used by {@link #choose_row_or_field_list}.
 */
choose_field_spec
   :
      lvalue (choose_help_clause)?
   ;
   
/**   
 * Matches the <code>HELP</code> keyword and the required following
 * <code>STRING</code>.
 * <p>
 * Used by {@link #choose_field_spec}.
 */
choose_help_clause   
   :
      KW_HELP^ STRING
   ;
   
/**   
 * Matches the <code>KEYS</code> keyword and the required following
 * {@link #lvalue} of <code>CHARACTER</code> data type.
 * <p>
 * Used by {@link #choose_stmt}.
 */
choose_keys_clause   
   :
      KW_KEYS^ 
      c:lvalue { #c.getType() == VAR_CHAR || #c.getType() == FIELD_CHAR }?
      // TODO: should we support VAR_LONGCHAR and FIELD_CLOB here?
   ;
   
/**
 * Matches the <code>PUT</code> language statement and builds the
 * associated AST.
 * <p>
 * The following list of rules are called to properly implement the tree
 * building:
 * <p>
 * <ul>
 *    <li> {@link #lvalue} (for stream references)
 *    <li> {@link #space_or_skip} 
 *    <li> {@link #null_clause}
 *    <li> {@link #put_field}
 * </ul>
 * <p> 
 * Called by {@link #stmt_list}.
 */
put_stmt
   :
      KW_PUT^
      (
           { LA(1) == KW_STREAM || LA(1) == KW_STRM_HND }? lvalue // used to be stream_reference
         | // empty alternative
      )                                  
      (KW_UNFORMAT | KW_CONTROL)?
      (
           runtime_space_or_skip
         | null_clause
         | put_field
      )*
      DOT!
   ;
      
/**      
 * Implements proper tree building for the most common field specification
 * in the {@link #put_stmt}.  Uses {@link #expr}, {@link #format_string} and 
 * {@link #to_or_at_clause}.  Roots everything at a node of type
 * <code>PUT_FIELD</code>.
 */
put_field      
   :
      { astFactory.makeASTRoot(currentAST, #[PUT_FIELD,"output field"]); }
      (
         expr
         (
              format_string[true]
            | to_or_at_clause  
         )*
      )
   ;
      
/**      
 * Matches the <code>AT or TO</code> keywords and a required following
 * expression.  This is a variant of some constructs in the 
 * {@link #format_phrase} however this rule is not compatible with those
 * rules.  This calls {@link #expr} and is used by {@link #put_stmt}.
 */
to_or_at_clause      
   :
      (KW_AT^ | KW_TO^) expr
   ;
      
/**
 * Matches the <code>PUT CURSOR</code> language statement and builds the
 * associated AST.  Calls {@link #ui_stuff} for the ROW and COL options.
 * <p>
 * Called by {@link #stmt_list}.
 */
put_cursor_stmt
   :
      p:KW_PUT^  { #p.setType(PUT_CURSOR); }
      KW_CURSOR!
      (
           KW_OFF
         | { LA(1) == KW_ROW || LA(1) == KW_COL || LA(1) == KW_COLUMNS }?
           ui_stuff
           (
                { LA(1) == KW_ROW || LA(1) == KW_COL || LA(1) == KW_COLUMNS }?
                ui_stuff
              | // empty alternative
           )
      )
      DOT!
   ;
      
/**      
 * Matches the <code>PUBLISH</code> language statement, the required
 * following character {@link #expr} and all options.  Uses the
 * {@link #from_handle_clause} and {@link #param_passing_list} rules.
 * <p>
 * Called from {@link #stmt_list}.
 */
publish_stmt      
   :
      KW_PUBLISH^ expr (from_handle_clause)?
      (param_passing_list[false])? DOT!
   ;
      
/**      
 * Matches the <code>SUBSCRIBE</code> language statement, the required
 * following character {@link #expr} and all options.  Uses the
 * {@link #proc_handle_clause}, {@link #in_procedure_clause} and
 * {@link #run_proc_clause} rules.
 * <p>
 * Called from {@link #stmt_list}.
 */
subscribe_stmt      
   :
      KW_SUBSCRIB^
      (
         options { generateAmbigWarnings = false; }
         :
         proc_handle_clause
      )?
      
      (KW_TO!)? expr
      (in_procedure_clause | KW_ANYWHERE)
      (run_proc_clause)?
      (KW_NO_ERROR)?
      DOT!
   ;
      
/**      
 * Matches the <code>UNSUBSCRIBE</code> language statement, the required
 * following character {@link #expr} and all options.  Uses the
 * {@link #proc_handle_clause} and {@link #in_procedure_clause}
 * rules.
 * <p>
 * Called from {@link #stmt_list}.
 */
unsubscribe_stmt      
   :
      KW_UNSUBSCR^
      (
         options { generateAmbigWarnings = false; }
         :
         proc_handle_clause
      )?
      
      (KW_TO!)? (KW_ALL | expr)
      (in_procedure_clause)?
      DOT!
   ;
      
/**   
 * Matches a <code>FROM</code> keyword and the following {@link #handle}.
 * Used in the {@link #publish_stmt}.
 * <p>
 * This is only separated to simplify the AST build.
 */
from_handle_clause   
   :
      KW_FROM^ handle
   ;      
      
/**   
 * Matches a <code>PROCEDURE</code> keyword and the following {@link #handle}.
 * Used in the {@link #subscribe_stmt}.
 * <p>
 * This is only separated to simplify the AST build.
 */
proc_handle_clause   
   :
      KW_PROC^ handle
   ;      
   
/**   
 * Matches a <code>RUN-PROCEDURE</code> keyword and the following 
 * character {@link #expr}. Used in the {@link #subscribe_stmt}.
 * <p>
 * This is only separated to simplify the AST build.
 */
run_proc_clause   
   :
      KW_RUN_PROC^ expr
   ;      
   
/**   
 * Matches the <code>WHEN</code> keyword and the required following
 * expression. This is a simpler form of the {@link #when_clause} which
 * requires a <code>THEN</code> following construct.
 * <p>
 * Used by {@link #display_stmt}, {@link #assign_stmt} and
 * {@link #field_clause}.
 * <p>
 * The subtree is rooted by the keyword (which is the reason for using a 
 * separate rule for something so simple).
 */
simple_when_clause   
   :
      KW_WHEN^ expr
   ;
   
/**   
 * Matches the <code>@</code> sign and the required following variable or
 * field name (see {@link #lvalue}).
 * <p>
 * Used by {@link #display_stmt}.
 * <p>
 * The subtree is rooted by the keyword (which is the reason for using a 
 * separate rule for something so simple).
 *
 * @param    matchSym
 *           <code>true</code> if this clause should allow matches to
 *           a <code>SYMBOL</code> node in place of the lvalue.
 *
 * @return   The symbol name if a <code>SYMBOL</code> was matched, otherwise
 *           <code>null</code>.
 */
at_base_field_clause [boolean matchSym] returns [String name]
   {
      name = null;
      
      if (matchSym)
      {
         allowSymbolMatch = true;
      }
   }
   :
      AT^ base:lvalue
      {
         if (matchSym)
         {
            allowSymbolMatch = false;
            
            if (#base.getType() == SYMBOL)
            {
               name = #base.getText(); 
            }
         }
      }
   ;
   
/**
 * Matches the Progress <code>MESSAGE</code> language statement and builds
 * the resulting AST nodes.  The string parameter is expressed as a call
 * to the {@link #expr} rule.
 * <p>
 * The following list of rules are called to properly implement the tree
 * building:
 * <p>
 * <ul>
 *    <li> {@link #simple_color_clause}
 *    <li> {@link #content_array}
 *    <li> {@link #view_as_alert_box_clause}
 *    <li> {@link #set_or_update_clause}
 *    <li> {@link #in_window_clause} 
 * </ul>
 * <p> 
 * These options can be in any order.  Contrary to the Progress documentation,
 * it is also possible to have no expression as a message (this has been
 * found in customer code and it works fine).
 * <p>
 * Called by {@link #stmt_list}.
 */
msg_stmt
   :
      KW_MSG^ (simple_color_clause)?
      content_array
      (
           view_as_alert_box_clause
         | set_or_update_clause
      )*
      (in_window_clause)?
      DOT!
   ;
   
/**
 * Matches the list of {@link #expr} or {@link #space_or_skip} input for 
 * the {@link #msg_stmt}.  The input can be made in any order and with any
 * number of items.  Everything will be rooted at a <code>CONTENT_ARRAY</code>
 * node.
 */
content_array   
   :
      { astFactory.makeASTRoot(currentAST, #[CONTENT_ARRAY, ""]); }
      (space_or_skip | expr)*
   ;
   
/**   
 * Matches the <code>COLOR</code> keyword and the required following color 
 * phrase.  Uses {@link #color_phrase}.  This is the simplified version of
 * {@link #color_spec}.
 * <p>
 * Used by {@link #put_screen_stmt} and {@link #choose_stmt}.
 * <p>
 * The subtree is rooted by the keyword (which is the reason for using a 
 * separate rule for something so simple).
 */
simple_color_clause   
   :
      KW_COLOR^ color_phrase
   ;
   
/**   
 * Matches the <code>VIEW-AS ALERT-BOX</code> construct and the optional
 * following definitions.  Calls {@link #buttons_clause} and 
 * {@link #simple_title_string} rules.  It is unclear if the options need
 * to be processed in any order.  At the current time, they are only
 * processed in a fixed order.
 * <p>
 * Used by {@link #msg_stmt}.
 * <p>
 * Bogus ANTLR ambiguity warnings have been disabled.
 * <p>
 * The subtree is rooted by the keyword (which is the reason for using a 
 * separate rule for something so simple).
 */
view_as_alert_box_clause   
   :
      KW_VIEW_AS^ KW_ALERT_BX
      (
         options { generateAmbigWarnings = false; }
         :
         KW_MSG | KW_QUEST | KW_INFO | KW_ERROR | KW_WARN
      )?
      (
         options { generateAmbigWarnings = false; }
         :
         buttons_clause
      )?
      (title_clause)?
   ;
   
/**   
 * Matches the <code>TITLE</code> keyword and the required following string. 
 * This is the simplified version of {@link #title_phrase}.
 * <p>
 * Used by {@link #view_as_alert_box_clause}.
 * <p>
 * The subtree is rooted by the keyword (which is the reason for using a 
 * separate rule for something so simple).
 */
simple_title_string   
   :
      KW_TITLE^ STRING
   ;
   
/**   
 * Matches the <code>BUTTONS</code>keyword and the required following 
 * button set.  Note that the <code>BUTTON</code> keyword is a valid
 * abbreviation for <code>BUTTONS</code> but it is a separate keyword, used
 * elsewhere in Progress.  For this reason, this rule matches on both
 * keywords and the token type is set to <code>BUTTONS</code> either way.
 * <p>
 * Used by {@link #view_as_alert_box_clause}.
 * <p>
 * The subtree is rooted by the keyword (which is the reason for using a 
 * separate rule for something so simple).
 */
buttons_clause   
   :
      KW_BUTTON^ 
      (KW_YES_NO | KW_YES_NO_C | KW_OK | KW_OK_CAN | KW_RETRY_C)
   ;
   
/**   
 * Matches the <code>SET field or UPDATE field</code> construct and the
 * required following as or like clause. Note that by using the 
 * {@link #as_clause} and {@link #like_clause} rules, the variable
 * namespace is naturally updated which should be correct. Also uses
 * {@link #lvalue} and {@link #format_string} rules.  The options can
 * be specified in any order.
 * <p>
 * Experience shows that a variable that is already defined can be
 * redefined (or Progress ignores the second definition) using this
 * phrase.  For this reason, it is possible to encounter an 
 * <code>AS or LIKE</code> construct after an <code>lvalue</code>. However,
 * this implementation is written to pass <code>null</code> as the symbol
 * name which makes either called rule bypass the variable addition to
 * the namespace.  If it turns out that Progress actually redefines the
 * variable, a change will be needed here.  Normally, it is not possible
 * to redefine a variable, so this behavior is unusual.
 * <p>
 * Used by {@link #msg_stmt}.
 * <p>
 * The subtree is rooted by the keyword (which is the reason for using a 
 * separate rule for something so simple).
 */
set_or_update_clause   
   :
      {
         boolean newVar = false;
      }
      (
         (KW_SET^ | KW_UPDATE^) 
         (
              { sym.lookupVariable(LT(1).getText()) == -1 }?
              f:symbol
              (
                   as_clause[#f.getText(), false]
                 | like_clause[#f.getText(), null, false]
              )
              { newVar = true; }
            | lvalue
              (as_clause[null, false] | like_clause[null, null, false])?           
         )
         (format_string[false] | KW_AUTO_RET)*
      )
      {
         if (newVar)
         {
            // pass the AST with all options to the symbol resolver to set
            // those variable options accordingly (really only applies to the
            // format_string and KW_AUTO_RET)
            sym.setVariableOptions(#f.getText(), ##);      
         
            // write any non-standard options into annotations (this will
            // really only be the tempIdx since there are no real options
            // in this form of variable definition)
            sym.annotateVariableOptions(#f.getText(), ##, true);            
         }
      }
   ;
   
/**
 * Matches the Progress <code>RUN</code> language statement with all of its
 * options and builds the resulting AST.  This rule relies heavily upon
 * lexer support in the {@link ProgressLexer#mSYMBOL} rule and upon the
 * {@link #filename} rule for handling the required 2nd token (the external
 * procedure name, internal procedure name or library reference).  The
 * {@link #value} rule is also called in the case of the user providing
 * an expression to specify the name.
 * <p>
 * Although the Progress documentation does not say so, it is possible to
 * specify the target to run as a string without using the <code>VALUE</code>
 * construct.  This is supported.
 * <p>
 * It is possible for the parser to identify something as a 
 * <code>FILENAME</code> when it really is a <code>INT_PROC</code> in the 
 * case where the internal procedure is defined after the run statement is 
 * encountered.  This is perfectly valid in a Progress source file.  This
 * problem can be fixed in a subsequent processing of the parser-generated
 * tree.
 * <p>
 * Most of the optional constructs are matched in the following rule
 * references: {@link #persist_clause}, {@link #on_server_clause},
 * {@link #in_procedure_clause}, {@link #trans_distinct_clause},
 * {@link #set_handle} or {@link #asynch_clause}.  Doing this allows the AST
 * to be built cleanly.
 * <p>
 * Note that the manner in which the <code>PERSISTENT, ON SERVER and IN</code>
 * clauses are processed allows them to appear in any order (which is how
 * Progress handles it) however there is no checking to ensure that multiple
 * instances of the same clause do not occur.  Due to defects in ANTLR's
 * code generation, the <code>ASYNCHRONOUS</code> clause, 
 * <code>TRANSACTION DISTINCT</code> and parameters have been moved into
 * this options processing to get the code to generate properly.  The result
 * is the same with valid Progress code and in fact, this may even have been
 * ultimately necessary.  Bogus ANTLR ambiguity warnings have been disabled.
 * <p>
 * As an undocumented feature, Progress allows any number of 
 * <code>NO-ERROR</code> keywords to be placed at the end of the statement,
 * with more than 1 being equivalent to having just 1 keyword.
 * <p>
 * In a <code>VALUE()</code> construct there is no checking of the type 
 * of the result (whether it is in the procedure dictionary because it is
 * an internal procedure or whether it is a filename).  In addition, no 
 * checking of the filesystem (using <code>PROPATH</code>) is done to 
 * ensure that the filename is valid.  Internal procedures specified via
 * a name (i.e. not using <code>VALUE()</code>) are detected and the
 * subsequent token type is set to <code>INT_PROC</code>. 
 * <p>
 * Another undocumented feature is that procedure names can be reserved
 * keywords, even if it is an internal procedure name. So the procedure name
 * can also match a {@link #reserved_or_symbol}.
 */
run_stmt
   :
      {
         CommonToken next = (CommonToken) LT(2);
      
         // the run stmt can match a simple string but it can also match a string that prefixes
         // other non-whitespace characters and it is the unquoted contents concatenated to the
         // following text that is matched; to duplicate this we must get rid of the quotes here
         // and duplicate any keyword processing that would otherwise have occurred; then we let
         // the normal matching occur below
         if (next.getType() == STRING)
         {
            boolean unix = Configuration.getParameter("unix-escapes", true);
            
            // convert the Progress string to a Java string (includes escape processing and
            // removal of quotes)
            String txt = character.progressToJavaString(next.getText(), !unix,  true);
            
            // no-transform approach to simply remove quotes; it is not clear if the above
            // transformations should or should not be applied
            // txt = txt.substring(1, txt.length() - 1);
            
            // save cleaned up string
            next.setText(txt);
            
            // simulate the keyword processing that would have happened if this wasn't a string
            Keyword found = sym.lookupKeyword(txt);
            
            next.setType((found != null) ? found.getTokenType() : SYMBOL);
         }
      }
      KW_RUN^
      (
         options { generateAmbigWarnings = false; }
         :
           LIBRARY_REF
           
         | v:value
         
           // undocumented feature: reserved keywords can be proc names 
           // this rule should be last to ensure that ambiguity is properly
           // avoided
         | f:filename
           {
              int ptype = sym.lookupProcedure(#f.getText()); 
           
              if (ptype != -1)
              {
                 // generally this is only triggered if the name exists
                 // as an internal procedure
                 saveAndReplaceType(#f, ptype);
              }
              else
              {
                 // by default we assume it is an external procedure (it
                 // won't be in the dictionary so it can't be looked up)
                 #f.setType(FILENAME);
              }
           }
      )
      (
         options { generateAmbigWarnings = false; }
         :
           persist_clause
         | set_handle
         | on_server_clause
         | in_procedure_clause
           {
              if (#f != null && #f.getType() == FILENAME)
                 #f.setType(INT_PROC);
           }
         
           // the next 2 rule references were originally split into separate
           // optional rules, BUT ANTLR misgenerated the prediction code 
           // (in an unnecessary manner), this is avoided by moving these
           // here and semantically it is the same (actually it is more
           // flexible and things can be in any order which may have been
           // necessary ultimately)
         | asynch_clause
         | param_passing_list[true]
      )*
      (preproc_arguments)?   
      ( KW_NO_ERROR (KW_NO_ERROR!)* )?      
      DOT!
   ;
   
/**
 * Matches a list of expressions as valid preprocessor/compiler arguments
 * in the <code>RUN</code> language statement and builds a single AST of
 * the entire list, rooted at an artificial <code>ARGUMENTS</code> node.
 * <p>
 * As an undocumented 4GL feature, {@link #value} can be referenced instead of
 * an expression.
 * <p>
 * <b>This rule uses a trick to create an artificial node as a root while
 * still allowing ANTLR to do normal tree building (no suppression of the
 * AST build).</b>  This is only useful in cases where there is no valid
 * token which can be used as a tree root AND where a loop is being used
 * such that the contents of the loop cannot be referenced in a manually
 * built #([,],,,) ANTLR construct. ANTLR doesn't support (as of 2.7.4) the
 * ability to label a loop such that the entire contents can be referenced.
 * If one labels the individual elements inside the loop, only the last one
 * is available when the loop concludes.  However, by leaving the normal
 * tree creation unchanged but manually creating an artificial node and
 * manually setting this node as the root of the returned tree BEFORE the
 * children are added, this problem can be solved.
 * <p>
 * Though undocumented, customer source has shown that it is valid to
 * insert a comma in between each preprocessor argument.  These are
 * optionally matched and discarded.
 * <p>
 * Another undocumented feature that is supported is that any text except
 * a <code>DOT</code> or <code>KW_NO_ERROR</code> is matched even through
 * only a value() construct or valid expression are documented as valid.
 * See {@link #broken_preproc_args}.
 * <p>
 * Called by {@link #run_stmt}.  This rule only exists to create the proper
 * AST, otherwise it would be much simpler to inline since the match is
 * trivial.
 */
preproc_arguments  
   :
      { astFactory.makeASTRoot(currentAST, #[ARGUMENTS,"arguments"]); }
      (
         options { generateAmbigWarnings = false; }
         :
           value
           
           // exclude some stuff that has no business matching an expression
           // but which CAN match the broken_preproc_args
         | { LA(1) != UNKNOWN_TOKEN && LA(1) != COMMA }?
           expr
           
           // match anything except DOT and KW_NO_ERROR (MUST be last)
         | broken_preproc_args
      )
      (
         (
            options { generateAmbigWarnings = false; }
            :
            COMMA!
         )?
         (
            options { generateAmbigWarnings = false; }
            :
              value
              
              // exclude some stuff that has no business matching an expression
              // but which CAN match the broken_preproc_args
            | { LA(1) != UNKNOWN_TOKEN && LA(1) != COMMA }?
              expr
              
              // match anything except DOT and KW_NO_ERROR (MUST be last)
            | broken_preproc_args
         )
      )*
      { if (##.getNumberOfChildren() == 0) ## = null; }
   ;
   
/**
 * Match anything except a <code>DOT</code> or <code>KW_NO_ERROR</code>.
 * Progress seems to match any text as a preprocessor argument, even though
 * only a value() construct or valid expression are documented.
 */
broken_preproc_args
   :
      ~(KW_NO_ERROR | DOT)
   ;
   
/**   
 * Matches the <code>PERSISTENT</code> keyword and the optional following
 * <code>SET</code> construct.
 * <p>
 * <b>Note that there is no procedure handle processing implemented at this
 * time. The matching is done as a symbol only, no namespace is
 * maintained.</b>
 * <p>
 * Used by {@link #run_stmt} and the subtree is rooted by the keyword
 * (which is the reason for using a separate rule for something so simple).
 */
persist_clause   
   :
      // there is no procedure handle processing at this time!
      KW_PERSIST^
      (
         options { generateAmbigWarnings = false; }
         :
         set_handle
      )?
   ;
   
/**   
 * Matches the <code>SET handle</code> clause and a following {@link #lvalue}
 * which must be a handle type.
 * <p>
 * Used by {@link #persist_clause}, {@link #run_stmt} and 
 * {@link #asynch_clause}.  The subtree is rooted by the keyword (which is
 * the reason for using a separate rule for something so simple).
 */
set_handle   
   :
      // TODO: we probably should remove functions and system handles from
      //       the possibilities here since this is something that is
      //       assigned
      
      // tests seem to show that this must be a simple lvalue and cannot be a chained sub-expr
      KW_SET^ l:lvalue { isHandleType(#l) }?
   ;
   
/**   
 * Matches the <code>ON</code> keyword and the optional following 
 * <code>SERVER handle</code> construct.
 * <p>
 * <b>Note that there is no session or server handle processing implemented 
 * at this time.  The matching is done as a symbol only, no namespace
 * is maintained.</b>
 * <p>
 * Used by {@link #run_stmt} and the subtree is rooted by the keyword
 * (which is the reason for using a separate rule for something so simple).
 * <p>
 * Ambiguity warnings are disabled because although ANTLR detects that the
 * optional <code>SERVER</code> keyword could conflict with the following
 * symbol, in fact Progress (and our implementation) both give preference
 * to the keyword.
 */
on_server_clause   
   :
      // there is no session or server handle processing at this time!
      KW_ON^ 
      (
         options { generateAmbigWarnings = false; }
         :
         KW_SERVER
      )?
      expr
      (
         options { generateAmbigWarnings = false; }
         :
         trans_distinct_clause
      )?
   ;
   
/**   
 * Matches the <code>IN</code> keyword and the required following expression
 * that resolves to the procedure handle construct in which the internal
 * procedure is to be run.
 * <p>
 * Used by {@link #run_stmt}, {@link #dynamic_function_func}, {@link #event_proc_clause}, 
 * {@link #subscribe_stmt} and {@link #unsubscribe_stmt}, the subtree is rooted by the keyword 
 * (which is the reason for using a separate rule for something so simple).
 */
in_procedure_clause   
   :
      KW_IN^ expr
   ;
      
/**   
 * Matches the <code>TRANSACTION DISTINCT</code> clause. Although official documentation states 
 * that both keywords are needed to define the clause, practice showed the DISTINCT keyword is
 * optional.
 * <p>
 * Used by {@link #run_stmt} and the subtree is rooted by the keyword
 * (which is the reason for using a separate rule for something so simple).
 */
trans_distinct_clause   
   :
      t:KW_TRANS^ { #t.setType(TRANSACTION_DISTINCT); } 
      (
         options { generateAmbigWarnings = false; }
         :
         KW_DISTINCT!
      )?
   ;
   
/**   
 * Matches the <code>ASYNCHRONOUS</code> clause and following options.
 * Calls {@link #set_handle} and {@link #event_proc_clause}.
 * <p>
 * Used by {@link #run_stmt} and the subtree is rooted by the keyword
 * (which is the reason for using a separate rule for something so simple).
 */
asynch_clause   
   :
      KW_ASYNC^
      (
         options { generateAmbigWarnings = false; }
         :
         set_handle
      )?
      (
         options { generateAmbigWarnings = false; }
         :
         event_proc_clause
      )?
   ;
   
/**   
 * Matches the <code>EVENT-PROCEDURE</code> clause and its required
 * expression and optional {@link #in_procedure_clause}.
 * <p>
 * Used by {@link #asynch_clause} and the subtree is rooted by the keyword
 * (which is the reason for using a separate rule for something so simple).
 * <p>
 * Bogus ANTLR ambiguity warnings have been disabled.
 */
event_proc_clause   
   :
      KW_EVT_PROC^ expr
      (
         options { generateAmbigWarnings = false; }
         :
         in_procedure_clause
      )?
   ;
   
/**
 * Matches the <code>SUPER</code> statement with an optional following
 * parameter list (see {@link #param_passing_list}). The resulting tree will be
 * rooted at a <code>KW_SUPER</code> node. Called from {@link #stmt_list}.
 * <p>
 * This must be the first statement used in a {@link #constructor}.
 */
super_stmt
   :
      KW_SUPER^ param_passing_list[false]
   ;
   
/**
 * Matches the <code>RUN SUPER</code> statement with an optional following
 * parameter list (see {@link #param_passing_list}). The resulting tree will be
 * rooted at a <code>RUN_SUPER</code> node. Called from {@link #stmt_list}.
 */
run_super_stmt
   :
      r:KW_RUN^  { #r.setType(RUN_SUPER); }
      KW_SUPER!
      (param_passing_list[false])?
      (KW_NO_ERROR)?
      DOT!
   ;
   
/**
 * Matches the <code>THIS-OBJECT</code> statement with an optional following
 * parameter list (see {@link #param_passing_list}). The resulting tree will be
 * rooted at a <code>KW_THIS_OBJ</code> node. Called from {@link #stmt_list}.
 * <p>
 * This must be the first statement used in a {@link #constructor}.
 */
this_object_stmt
   :
      KW_THIS_OBJ^ param_passing_list[false]
   ;
   
/**   
 * Matches a parenthesized list of comma-separated parameters used in a call-site and uses a
 * rule reference to {@link #parameter} to properly build the AST.  The parameters are optional
 * as some customer code has been found that has an empty set of parenthesis.
 * <p>
 * Called from {@link #run_stmt}, {@link #function}, {@link #super_function},
 * {@link #super_stmt}, {@link #run_super_stmt}, {@link #new_phrase},
 * {@link #run_stored_proc_stmt} {@link #this_object_stmt} and {@link #publish_stmt}.  
 * <p>
 * This rule is used from many places, but when called from the RUN statement it will consume an
 * unlimited number of unmatched ending parenthesis characters. This is an undocumented "feature"
 * of Progress that will compile and work.
 *
 * @param    isRunStmt                      
 *           <code>true</code> if the caller is the RUN statement. 
 */
param_passing_list [boolean isRunStmt]   
   :
      LPARENS^ (parameter (COMMA! parameter)*)? RPARENS!
      (
           { isRunStmt }?
           (
              options { generateAmbigWarnings = false; }
              :
              RPARENS!
           )*
         | // empty alternative
      )
   ;
   
/**   
 * Matches a parenthesized list of comma-separated parameter definitions and uses a rule
 * reference to {@link #parameter_spec} to properly build the AST.  Note that the parameters
 * are optional as some customer code has been found (in function definitions) that has an
 * empty set of parenthesis.
 * <p>
 * Called from {@link #constructor_stmt}, {@link #method_stmt}, {@link #func_stmt},
 * {@link #getter_setter} and {@link #event_signature}.
 *
 * @param    defineBuffer
 *           <code>true</code> if the caller supports inline buffer
 *           definitions (versus just passing a buffer parameter which
 *           doesn't create a new buffer).
 * @param    funcDef
 *           <code>true</code> if the caller is a FUNCTION definition. 
 */
param_list_definition [boolean defineBuffer, boolean funcDef]   
   :
      LPARENS^ 
      (
         parameter_spec[defineBuffer, funcDef] (COMMA! parameter_spec[defineBuffer, funcDef])*
      )?
      RPARENS!
   ;
   
/**
 * Matches a single parameter being passed at a call-site. The preceding <code>INPUT</code>,
 * <code>INPUT-OUTPUT</code> and <code>OUTPUT</code> keywords are optional. The core parameter is
 * matched using the {@link #expr} rule.
 * <p>
 * The <code>INPUT</code> keyword can also be found as the leftmost token in a conditional
 * expression (there is a function by the same name that takes a single non-parenthesized
 * parameter). This conflicts with the optional keyword here.  Warnings has been disabled for that
 * subrule as a result.
 * <p>
 * To actually use the <code>INPUT</code> built-in function as a parameter, one must paranthesize
 * the call OR add the following FRAME frame-name qualifier between the KW_INPUT and the widget
 * name.  This is how it works in Progress!  Otherwise the presence of the KW_INPUT is treated as
 * an ignored mode qualifier.
 * <p>
 * The subtree returned from this rule will be rooted at an artificial node of type
 * <code>PARAMETER</code>.
 * <p>
 * Called from {@link #param_passing_list}.
 */
parameter
   :
      // this should be merged with the func_call_parm rule (BUT that one doesn't create a root
      // node and there we match KW_INPUT as a builtin based on the flag passed in) 
      {
         astFactory.makeASTRoot(currentAST, #[PARAMETER,"parameter"]);
         
         boolean isInputFunc = (LA(1) == KW_INPUT &&
                                LA(2) == KW_FRAME &&
                                resolveLvalueCoreType(4, false) != -1);
      }
      (
         options { generateAmbigWarnings = false; }
         :
           { !isInputFunc }?
           KW_INPUT
         | KW_OUTPUT
         | KW_IN_OUT
      )?
      expr
      (KW_APPEND | KW_BY_VALUE | KW_BY_REF | KW_BIND)*
   ;
   
/**
 * Matches a single statement parameter definition for methods, constructors, destructors,
 * property getters/setters and user-defined function definitions. This is not used at
 * call-sites but only to define the interface for the given callable unit.
 * <p>
 * Optionally matches <code>INPUT, INPUT-OUTPUT and OUTPUT</code> keywords or a
 * <code>BUFFER</code> definition as well as a range of possible table parameter types.
 * The {@link #table_parm}, {@link #def_buf_stmt} and {@link #as_clause} rules are used.
 * <p>
 * This rule will define new variables in the variable dictionary. For normal variables it
 * calls the {@link #as_clause} rule to add this variable name into the variable dictionary
 * with the correct data type.  Note that this rule passes the variable name as a
 * <code>String</code> argument to the <code>as_clause</code> rule since this current rule has
 * the knowledge of this name but doesn't have the knowledge of how to deal with the data types
 * and variable dictionary.
 * <p>
 * Instead of an AS clause, this will also match the undocumented {@link #like_clause}. 
 * <p>
 * For buffers and other table parameters, these are each handled by separate rules.
 * <p>
 * Note that it seems that the 2nd alternative (a symbol followed by the
 * <code>AS</code> keyword) should be detected as ambiguous with the
 * 3rd alternative (an expression).  However ANTLR automatically inserts
 * a test for the 2nd token being equivalent to <code>AS</code> which 
 * disambiguates these alternatives.  This may be due to the use of a rule
 * reference rather than an inline keyword implementation.  <b>If this coding
 * changes, a semantic predicate may have to be used to manually disambiguate.
 * There will be no ANTLR warnings since these have been disabled, so you
 * MUST exercise great care in modifying this rule.</b>
 * <p>
 * The <code>INPUT</code> keyword can also be found as the leftmost token
 * in a conditional expression (there is a function by the same name that 
 * takes a single non-parenthesized parameter). This conflicts with the
 * optional keyword here, but we will match first on this keyword and this
 * is the correct behavior.  Warnings has been disabled for that subrule as
 * a result.
 * <p>
 * The subtree returned from this rule will be rooted at an artificial node
 * of type <code>PARAMETER</code>.
 * <p>
 * Called from {@link #param_list_definition}.
 *
 * @param    defineBuffer
 *           <code>true</code> if the caller supports inline buffer
 *           definitions (versus just passing a buffer parameter which
 *           doesn't create a new buffer).
 * @param    funcDef
 *           <code>true</code> if the caller is a FUNCTION definition. 
 */
parameter_spec [boolean defineBuffer, boolean funcDef]
   :
      {
         String  varname = null;
         boolean abbrev  = false;
         int     size    = -1;
         
         // this should really be changed to a different token type to differentiate it from
         // the parameter passing syntax in the parameter rule
         astFactory.makeASTRoot(currentAST, #[PARAMETER,"parameter"]);
      }
      
      // this is optional for 2 reasons:
      // - KW_INPUT is optional in the RUN and FUNCTION forms
      // - KW_BUFFER can be in the 1st token position in the RUN form (is this still needed?)
      (
         options { generateAmbigWarnings = false; }
         :
         k1:KW_INPUT | k2:KW_OUTPUT | k3:KW_IN_OUT
      )?
      ( 
         options { generateAmbigWarnings = false; }
         :
           varname = tp:table_parm
           
         | { defineBuffer }?
           buf:def_buf_stmt[false, funcDef, null, null]
             
           // at this time we don't need to insert a semantic predicate
           // { LA(2) == KW_AS }? because ANTLR generates this test already
           // for the current encoding; IF THIS CHANGES YOU MAY NEED TO 
           // INSERT THIS PREDICATE OTHERWISE THE AMBIGUITY WILL BE REAL
           // --> YOU WILL HAVE NO OTHER WARNING SINCE WARNINGS HAVE BEEN
           //     DISABLED
         | param:symbol { varname = #param.getText(); }
           (as_clause[varname, false] | like_clause[varname, null, false])
           // no annotation needs to be saved here, as this rule will call
           // Variable.annotateVariableOptions which will take care of saving the extent, 
           // if present
           (size = extent)?
           
           // this is here to support some undocumented syntax in the
           // function stmt, where there is no parameter name and the AS
           // keyword is missing: this can be done on IN HANDLE and
           // FORWARD definitions; we exclude non-definitions so that there
           // is no ambiguity with the following expr (which is a runtime
           // case so this should work)
         | vt:var_type[null, false]   { abbrev = true; }
      )         
      {
         // this should only occur during function definitions and only for
         // table handles or symbols followed by the AS keyword
         if (varname != null)
         {
            // the only option to set is the parameter type
            sym.setVariableOptions(varname, ##);
         
            // write any non-standard options into annotations (this probably
            // only means the tempidx since no other options are really
            // possible)
            sym.annotateVariableOptions(varname, ##, true);
            
            // check if the type was specified
            if (#k1 == null && #k2 == null && #k3 == null)
            {
               // default to KW_INPUT
               ##.putAnnotation("parmtype", new Long(KW_INPUT));
            }
         }
         else
         {
            // in the parameter definition case where this is not a full
            // variable definition (it is the abbreviated var def, a table,
            // table-handle, buffer), create those annotations that are
            // possible
            
            Aast mode  = (#k1 == null) ? ((#k2 == null) ? #k3 : #k2) : #k1;
            int  ptype = (mode == null) ? KW_INPUT : mode.getType();
            ##.putAnnotation("parmtype", new Long(ptype));
            
            Aast typ = (#tp == null) ? ((#buf == null) ? #vt : #buf) : #tp;
            if (typ != null)
               ##.putAnnotation("type", new Long(typ.getType()));
         }
      }
   ;
   
/**
 * Matches the <code>CALL</code> language statement, the required
 * {@link #filename} of the invoked HLC (host language call) function and
 * and optional list of {@link #expr} (arguments). Used by {@link #stmt_list}.
 */
call_stmt
   :
      KW_CALL^ filename (expr)* DOT!
   ;
   
/**
 * Matches the <code>RUN STORED-PROCEDURE</code> language statement, the
 * following {@link #reserved_or_symbol}, the optional {@link #assign}
 * and optional {@link #param_passing_list}. The result will be rooted at a
 * <code>RUN_STORED_PROCEDURE</code> node.  Called from {@link #stmt_list}.
 * <p>
 * Warning: this is an early implementation which does not yet support the
 * notion that a stored procedure is like a special kind of table type. The
 * stored procedure is a schema object created in the database supported
 * by a 4GL data server.  That schema object is known to the 4GL and it
 * is called with optional parameters (and must be closed when done). But
 * to get the data out of the results, Progress creates a buffer of the same
 * name as the procedure, which can be used in record phrases to read the
 * data. Likewise, there are properties defined for the results which are
 * treated like fields.  The schema dictionary, symbol resolver and parser
 * all need updates to fully support this new resource. It is likely that
 * real 4GL code using these constructs will not parse properly at this time.
 * Also note that there are special purpose names <code>proc-text</code> and
 * <code>proc-text-buffer</code> which can be used as a field (actually it
 * is a text string of the entire current record's contents) and as a buffer
 * respectively. That means that definitions of these need to be created
 * during this method. Following code may rely upon this to parse. Finally,
 * there may be a special <code>closeallprocs</code> procedure that is
 * the equivalent to calling {@link #close_stored_proc_stmt} on all currently
 * open stored procedures.
 */
run_stored_proc_stmt   
   :
      r:KW_RUN^  { #r.setType(RUN_STORED_PROCEDURE); } 
      KW_STORPROC!
      reserved_or_symbol 
      (assign_proc_handle)?
      (KW_NO_ERROR)?
      (param_passing_list[false])?
      DOT!
   ;
   
/**
 * Match the <code>integer-field = PROC-HANDLE</code> clause of the
 * {@link #run_stored_proc_stmt} rule. This matches an {@link #lvalue}
 * followed by an <code>EQUALS</code> and then another <code>lvalue</code>.
 * The result is rooted at an <code>ASSIGN</code> node. This is separated
 * to allow the tree to be built properly.  Also called from the
 * {@link #close_stored_proc_stmt}.
 */
assign_proc_handle
   :
      lvalue eq:EQUALS^ { #eq.setType(ASSIGN); } lvalue
   ;
   
/**
 * Matches the <code>CLOSE STORED-PROCEDURE</code> language statement, the
 * following {@link #reserved_or_symbol}, the optional {@link #assign}
 * and optional {@link #where_clause}. The result will be rooted at a
 * <code>CLOSE_STORED_PROCEDURE</code> node.  Called from {@link #stmt_list}.
 * <p>
 * Warning: this is an early implementation which does not yet support the
 * notion that a stored procedure is like a special kind of table type. The
 * stored procedure is a schema object created in the database supported
 * by a 4GL data server.  That schema object is known to the 4GL and it
 * is called with optional parameters (and must be closed when done). But
 * to get the data out of the results, Progress creates a buffer of the same
 * name as the procedure, which can be used in record phrases to read the
 * data. Likewise, there are properties defined for the results which are
 * treated like fields.  The schema dictionary, symbol resolver and parser
 * all need updates to fully support this new resource. It is likely that
 * real 4GL code using these constructs will not parse properly at this time.
 */
close_stored_proc_stmt
   :
      r:KW_CLOSE^  { #r.setType(CLOSE_STORED_PROCEDURE); } 
      KW_STORPROC!
      reserved_or_symbol
      (assign_proc_handle)?
      (where_clause)?
      DOT!
   ;
   
/**      
 * Matches all valid filenames (relative and absolute). This rule can match
 * any lexer token in the first position and that matching will continue until
 * whitespace is encountered or if one of a specific list of excluded token
 * types is matched. Since this matching is so broad, it is highly ambiguous
 * and this rule must be placed as the last alternative in callers such that 
 * it doesn't overmatch. This means that it can conflict with literals,
 * reserved keywords, operators and any other unexpected text in a 4GL
 * program.
 * <p>
 * Token types that end the merging process:
 * <p>
 * <pre>
 *   DOT
 *   EOF
 *   LPARENS
 *   RPARENS
 * </pre>
 * <p>
 * The lexer has a <code>FILENAME</code> token but it does not have enough
 * context to differentiate the full range of filenames that are possible.
 * In Progress, when a filename is matched, the lexing changes to an approach
 * that seems to be based on whitespace. Any text at all can be matched until
 * a separate excluded token type or whitespace is encountered. This has been
 * duplicated in the init action of this rule. The <code>mergeUntilWS</code>
 * method is used to merge all the different tokens using this same approach.
 * It is called from the init action and the resulting single merged token
 * will have its type set to <code>FILENAME</code>.
 * <p>
 * The matching portion of this rule is defined as a single token that
 * matches the union of the <code>FILENAME</code> token type and the set of
 * all tokens that are not a <code>DOT</code>.
 */
filename   
   :
      {
         // DOT and EOF are hard coded in the worker routine
         int[] exclude = new int[]
         {
            LPARENS,
            RPARENS
         };
         
         mergeUntilWS(exclude);
         LT(1).setType(FILENAME);
      }
      (
           FILENAME^
         
           // this reference is for "pull" only, its corresponding type
           // is rewritten unconditionally at entry to this rule so it
           // will never get executed
         | ~DOT ~DOT ~DOT
      )
   ;
   
/**
 * Matches the Progress <code>RETURN</code> statement using a rule reference
 * to {@link #return_core} and builds the resulting AST.  The core logic is in
 * the rule reference rather than here.  The separation was made to allow
 * the following <code>DOT</code> to be visible as a standalone language
 * statement when called from this rule.  The <code>return_core</code> rule
 * can also be called as a component of another language statement rather than 
 * a statement that requires such a separator.
 */
return_stmt 
   :
      return_core DOT!
   ;
   
/**
 * Matches the Progress <code>RETURN</code> keyword and builds the
 * resulting AST.  This is used in multiple locations including the
 * {@link #return_stmt} and the {@link #on_event_phrase}.
 */
return_core
   :
      KW_RETURN^ 
      (
         options { generateAmbigWarnings = false; }
         :
         KW_ERROR | KW_NO_APPLY
      )?
      (
         options { generateAmbigWarnings = false; }
         :
         expr
      )?   
   ;
   
/**
 * An empty statement (a perfectly legal Progress 4GL construct). Note that
 * one can consider this a <code>NOP</code>. Strangely, an optional 
 * <code>NO-LOCK</code>, <code>SHARE-LOCK</code> or 
 * <code>EXCLUSIVE-LOCK</code> keyword can be present. If a locking keyword
 * is present, then <code>NO-ERROR</code> can also be optionally specified
 * (after the locking keyword).
 */
empty_stmt
   :
      ( (KW_NO_LOCK! | KW_SH_LOCK! | KW_EXC_LOCK!) (KW_NO_ERROR!)? )?
      DOT!
   ;
   
/**
 * Matches the P2J-extension language statement <code>P2J-OPEN-URL</code> and a mandatory
 * following character expression.
 */
p2j_open_url_stmt
   :
      KW_OPEN_URL^ expr DOT!
   ;
   
/**
 * Parses any valid Progress 4GL expression and creates an AST that is
 * suitable for evaluation.  This is the main "entry point" for all Progress
 * 4GL expression parsing.  Progress expressions can contain sub-expressions
 * and this entry point is properly recursive without causing any ambiguity.
 * <p>
 * The AST is structured with operator nodes as the root nodes and each
 * operator (unary or binary) has the matching number of child nodes (1 or
 * 2 respectively).  Each child node is an operand and the semantics of
 * left and right operands (for binary operators) is maintained.
 * <p>
 * The objective of all expression evaluation is to generate a single result
 * of one of the possible Progress data types: integer, decimal, logical,
 * date or character.  The key point is that unless a Progress construct 
 * defines that it requires a literal (e.g. the <code>AS</code> clause for
 * <code>DEFINE VARIABLE</code>), it is usually the case that an expression
 * is allowed.  It is always the case that there is some limit to the 
 * resulting data types that are allowed for the expression.  Another way to
 * say this is that many Progress constructs will require an expression of
 * a specific type (e.g. the <code>MESSAGE</code> statement requires a
 * character expression as its main data to be displayed).  Since all valid
 * expressions can be evaluated or resolved down to a specific data type,
 * this is <b>the key</b> to building a consistent syntax for the Progress
 * language.  Since Progress expressions can contain lvalues, function calls,
 * literals by themselves as well as part of a larger expression, none of
 * the Progress constructs that reference expressions of a specific type
 * need to worry about what is included or how it is parsed.  As long as
 * expressions are used, all of the variety of possible combinations are
 * obtained &quot;for free&quot; because a common expression implementation
 * is used.  For this reason, this top level entry point is probably the
 * most heavily referenced rule in this entire grammar.  <b>Changes must be
 * made very carefully!</b>
 * <p>
 * All of the Progress operators are supported (except for the undocumented
 * bitwise operators) and the precedence order is as follows:
 * <pre>
 * lowest  OR
 *    |    AND
 *    |    NOT
 *    |    =, EQ, &lt;&gt;, NE, &lt;, LT, &gt;, GT, &le;, LE, &ge;, GE, MATCHES, BEGINS, CONTAINS
 *    |    binary +, binary - 
 *    |    *, /, MODULO
 *    |    unary +, unary -
 *    |    DB_REF_NON_STATIC
 *    |    COLON (used as an invocation mechanism)
 * highest ( )
 * </pre>
 * <p>
 * Note that CONTAINS is only documented for use within a <code>WHERE</code>
 * clause, see {@link #record_phrase} and {@link #where_clause}.
 * <p>
 * The key problems that had to be solved for this implementation included: 
 * <ul>
 *    <li> Structuring the rules to implement the proper precedence AND to
 *         properly build a tree that could be correctly evaluated.
 *    <li> Ensuring that expressions properly nest (through recursion).
 *    <li> Eliminating ambiguity that would defeat the parser's lookahead 
 *         mechanisms.
 * </ul>
 * <p>
 * To handle the precedence issue, each precedence level (see the table
 * above) is handled in a separate rule.  Starting at in this top level
 * entry point, the lowest precedence rule is referenced. This reference is
 * made with an optional <code>KW_OR</code> operator.  If thus operator is
 * present, it is the root of the resulting tree.  This means it is the
 * last operator to be evaluated since its operands will be evaluated
 * first.  This is the proper definition of lowest precedence.  If this
 * operator is present, then a second operand is expected.
 * <p>
 * Each precedence level is similarly constructed, the left operand being
 * required, all operators being optional and each operator choice an equal 
 * alternative. If any of the operators exists, then it is the root of the
 * subtree and a right operand is expected.  This hold true for all rules
 * that process binary operators.    Please note that this left-mandatory
 * and optional right side (operator + right operand) is critical for
 * allowing an expression to be as simple as a single lvalue (variable or
 * field name), function call or literal. Alternatively, it can expand into
 * a highly nested set of recursively parsed expressions that all evaluate
 * to a single scalar result of one of the known Progress data types.
 * <p>
 * At the bottom of the hierarchy of rules are unary operators, these
 * are implemented as an optional prefix operator and a mandatory right
 * operand.
 * <p>
 * The final level is the primary expression.  This is where the precedence
 * operator parenthesis is handled.  It also is where we allow a single
 * lvalue, literal or function reference.  While the literals are fairly
 * easily defined, lvalues and functions are complicated by the fact that
 * they are used defined symbols and there is inherent ambiguity in the
 * definitions for each of these.  See {@link #lvalue} and {@link #func_call}
 * for more details on how this is resolved.
 * <p>
 * This rule is designed to allow recursion (the highest precedence levels
 * {@link #func_call} and {@link #primary_expr} reference this rule again,
 * allowing infinite nesting of sub-expressions, with proper parsing,
 * evaluation and tree building.
 * <p>
 * Ambiguity is primary caused by the potential overlap in the rules
 * referenced from {@link #primary_expr}.  This is caused by the fact that
 * multiple rule references made there must all allow for a user-defined
 * symbol to appear in the token stream.  Multiple rules all requiring
 * the same definition to appear is the definition of <b>ambiguity</b>!
 * See that rule for more details on the solution.
 * <p>
 * To the extent that is feasible, syntax checking is implemented in the
 * expression processing, however there is one important area of checking
 * that cannot be properly specified using the structure of the parser:
 * operand data type checking.  For this reason, such checking (when it is
 * implemented) will be built using validating semantic predicates.  The need
 * for this checking is limited based on the assumption that all input is
 * a valid Progress 4GL source file.
 * <p>
 * <code>ExpressionEvaluator</code> is dependent upon this entry
 * point.
 * <p>
 * Note that there is an ambiguity caused between the <code>OR</code> keyword
 * implemented in <code>expr</code> and by the use of the <code>OR</code>
 * keyword in a <code>WHEN</code> clause (additional WHEN conditions can
 * be added to the same statement using an <code>OR WHEN condition</code>
 * syntax). Since the following condition is an expression, one must detect
 * the use of the following <code>WHEN</code> as the next token after the
 * <code>OR</code> to disambiguate. The <code>expr</code> rule which 
 * implements the logical <code>OR</code> processing thus uses a semantic
 * predicate to terminate expression processing (and not match on the
 * <code>OR</code> in the case where lookahead 2 is the reserved keyword
 * <code>WHEN</code>.  Ambiguity warnings are disabled since this trick  
 * resolves the issue.
 * <p>
 * When this is called from a parser that is invoked with the top level
 * entry point {@link #external_proc}, this rule will root the resulting
 * subtree in an artificial token <code>EXPRESSION</code>.  Otherwise, this
 * method is used directly and no artificial root is created.
 * <p>
 * The {link #log_and_expr} is called as the only rule reference.
 */
expr
   :
      {
         // tracks the re-entrancy of this rule, if this is 0 on entry, then
         // this has been no recursion yet...
         exprLvl++;         
      }
      log_and_expr
      (
         options { warnWhenFollowAmbig = false; }
         :
         { LA(2) != KW_WHEN }?
         kwo:KW_OR^ { checkExtraSpace(#kwo); } log_and_expr
      )*
      {
         // only make an artificial root at the first entry into this rule
         // (as opposed to subsequent recursive entries) and if the parser
         // is operating from the top level entry point
         if (topLevelEntry && exprLvl == 1)
         {
            ## = #([EXPRESSION,"expression"], ##);
         }
      
         exprLvl--;
      }
   ;

/**
 * Implements the 2nd lowest precedence level of Progress 4GL expressions
 * <code>KW_AND</code> (the logical AND) operator.
 * <p>
 * This is only called by the top level {@link #expr} rule.  This method
 * calls the {@link #log_not_expr} rule.
 * <p>
 * There is a bogus ambiguity warning generated by this rule (which occurs
 * because of the addition of the special <code>IF THEN ELSE</code> function
 * (see {@link #if_func}) in the {@link #primary_expr} rule. No real ambiguity
 * seems to exist and no errors can be found, so warnings have been disabled.
 * <p>
 * See the top level expression parsing rule {@link #expr} for more details.
 */
log_and_expr 
   :
      log_not_expr
      (
         options { warnWhenFollowAmbig = false; }
         :
         kwa:KW_AND^ { checkExtraSpace(#kwa); } log_not_expr
      )*
   ; 

/**
 * Implements the 3rd lowest precedence level of Progress 4GL expressions
 * <code>KW_NOT</code> (the logical NOT) operator.
 * <p>
 * This is only called by the 2nd level {@link #log_and_expr} rule.
 * This method calls the {@link #compare_expr} rule.
 * <p>
 * See the top level expression parsing rule {@link #expr} for more details.
 */
log_not_expr 
   :
      (KW_NOT^)? compare_expr
   ;

/**
 * Implements the 4th lowest precedence level of Progress 4GL expressions
 * - the comparison operators.
 * <p>
 * This is only called by the 3nd level {@link #log_not_expr} rule. This  
 * method subsequently calls the {@link #sum_expr} rule.
 * <p>
 * Provides the following operators:
 * <pre>
 *    EQUALS or KW_EQ 
 *    NOT_EQ or KW_NE
 *    GT     or KW_GT
 *    LT     or KW_LT
 *    GTE    or KW_GTE
 *    LTE    or KW_LTE
 *    KW_BEGINS
 *    KW_MATCHES
 *    KW_CONTAINS
 * </pre>
 * <p>
 * When in an embedded SQL statement, the following additional operators
 * are supported:
 * <pre>
 *    KW_BETWEEN
 *    NOT_BETWEEN
 *    KW_LIKE
 *    NOT_LIKE
 *    KW_IN
 *    NOT_IN
 *    IS_NULL
 *    IS_NOT_NULL
 * </pre>
 * <p>
 * If this is called from within a <code>WHERE</code> clause that is in
 * an SQL {@link #select_stmt}, then all forms of sub-select are supported
 * in this rule. Uses {@link #sub_select_stmt} and
 * {@link #in_operator_sub_select_stmt} to provide this support.
 * <p>
 * Note that to eliminate ambiguity with the unreserved keywords listed
 * above, if any of these keywords are found in the proper operator position 
 * we reset the token type to the associated non-keyword operator's type.
 * <p>
 * The other two alternative approaches each cause problems:
 * <ul>
 *    <li> If the token type is not forced to the non-keyword type, then
 *         expression evaluation would be more complicated because more
 *         than one token type would generate the same action.
 *    <li> If the token type of the keyword was forced to the non-keyword
 *         corresponding operator by the lexer's keyword lookup mechanism, 
 *         then the non-reserved keywords could never be used as user-defined 
 *         symbols since the token type would always be set to the operators
 *         instead.  This would not match how Progress works.
 * </ul>
 * <p>
 * The approach taken allows for the lexer to <b>always</b> set the token
 * type to the unreserved keyword (e.g. the text <code>EQ</code> is set to
 * the <code>KW_EQ</code> token type.  Then it is the parser's context that
 * determines if that token matches into this rule or not.  If so, the token
 * type is rewritten to the correct matching operator. In this case it would
 * be set to the token type <code>EQUALS</code> which is the same as if the
 * lexer had encountered the '=' character.  This generates a simple tree
 * where the operators are well known.  However, if the keyword is used as
 * a symbol name (e.g. a variable), this will work unchanged as the lexer
 * won't hard code the type as the operator.  See {@link #lvalue} for more
 * details.
 * <p>
 * See the top level expression parsing rule {@link #expr} for more details.
 * <p>
 * Note that ANTLR reports ambiguity between this rule and the {@link #assign}
 * rule.  The ambiguity is resolved by the ordering of the top level rule
 * placing the <code>ASSIGN</code> language statement (which is unambiguous)
 * with higher precedence than the <code>assignment</code> rule. As long as
 * the <code>assign</code> rule is only used by the <code>assign_stmt</code>
 * rule, there is no real ambiguity.  For this reason, ambiguity warnings
 * are suppressed for this rule.
 */
compare_expr
   :
      {
         boolean inBetween = false;
         boolean inLike    = false;
         boolean inIn      = false;
         boolean isNull    = false;
         boolean avoid     = false;
      }
      
      sum_expr
      
      {
         // additional code to bypass the operator checking in the case
         // where were ARE NOT in embedded SQL mode and there is a following
         // keyword that might be confused with an embedded SQL operator
         // (for example, in def_var_stmt where there is a like_clause that
         // follows a format_string)
         
         int ttype = LA(1);
         
         avoid = !embeddedSql && (ttype == KW_LIKE    ||
                                  ttype == KW_NOT     ||
                                  ttype == KW_IN      ||
                                  ttype == KW_BETWEEN ||
                                  ttype == KW_IS); 
      }
      
      (
           { !avoid }?
           ( 
              options { generateAmbigWarnings = false; }
              :
              (
                  EQUALS^ | eq:KW_EQ^   { checkExtraSpace(#eq); #eq.setType(EQUALS); }
                | NOT_EQ^ | ne:KW_NE^   { checkExtraSpace(#ne); #ne.setType(NOT_EQ); }
                | GT^     | gt:KW_GT^   { checkExtraSpace(#gt); #gt.setType(GT); }
                | LT^     | lt:KW_LT^   { checkExtraSpace(#lt); #lt.setType(LT); }
                | GTE^    | ge:KW_GTE^  { checkExtraSpace(#ge); #ge.setType(GTE); }
                | LTE^    | le:KW_LTE^  { checkExtraSpace(#le); #le.setType(LTE); }
                | kwb:KW_BEGINS^ { checkExtraSpace(#kwb); }
                | kwm:KW_MATCHES^ { checkExtraSpace(#kwm); }
                | kwc:KW_CONTAINS^ { checkExtraSpace(#kwc); }
                
                | { embeddedSql }?
                  (
                       nb:KW_NOT^ KW_BETWEEN!
                       {
                          #nb.setType(NOT_BETWEEN);
                          inBetween = true;
                       }
                     | nl:KW_NOT^ KW_LIKE!
                       {
                          #nb.setType(NOT_LIKE);
                          inLike = true;
                       }
                     | ni:KW_NOT^ KW_IN!
                       {
                          #ni.setType(NOT_IN);
                          inIn = true;
                       }
                     | inn:KW_IS^ KW_NOT! KW_NULL!
                       {
                          #in.setType(IS_NOT_NULL);
                          isNull = true;
                       }
                     | in:KW_IS^ KW_NULL!
                       {
                          #in.setType(IS_NULL);
                          isNull = true;
                       }
                     | KW_BETWEEN^     { inBetween = true; }
                     | KW_LIKE^        { inLike = true; }
                     | KW_IN^          { inIn = true; }
                  )
              )
              
              (
                   { inIn }?
                   in_operator_parms
                   
                 | { embeddedSql && selectNestLevel > 1 }?
                   sub_select_stmt
                   
                   // common case
                 | { !isNull }?
                   sum_expr
              )
              
              (
                   { inBetween }?
                   and_expr
                   
                 | { inLike }?
                   escape_char
                   
                 | { embeddedSql && selectNestLevel > 1 }?
                   in_operator_sub_select_stmt
                 
                 | // empty alt
              )
           )*
         
         | // empty alternative
      )
   ;
   
/**
 * Qualifying 3rd expression for the <code>BETWEEN</code> operator.
 */
and_expr
   :
      KW_AND^ sum_expr
   ;
   
/**
 * Matches <code>ESCAPE</code> keyword and a following string literal
 * for the <code>LIKE</code> operator.
 */
escape_char
   :
      KW_ESCAPE^ STRING
   ;
   
/**
 * Implements the 5th precedence level of Progress 4GL expressions, the
 * <code>PLUS</code> and <code>MINUS</code> operators (for all data types,
 * even the non-numeric uses for dates or string concatenation).
 * <p>
 * This is only called by the 4th level {@link #compare_expr} rule. This  
 * method subsequently calls the {@link #prod_expr} rule. 
 * <p> 
 * There is a bogus ambiguity warning generated by this rule (which occurs
 * because of the addition of the special <code>IF THEN ELSE</code> function
 * (see {@link #if_func}) in the {@link #primary_expr} rule. No real ambiguity
 * seems to exist and no errors can be found, so warnings have been disabled.
 * <p>
 * See the top level expression parsing rule {@link #expr} for more details.
 */
sum_expr
   :
      prod_expr
      (
         options { warnWhenFollowAmbig = false; }
         :
         (PLUS^ | MINUS^) prod_expr
      )*
   ; 

/**
 * Implements the 6th precedence level of Progress 4GL expressions, the
 * <code>MULTIPLY</code>, <code>DIVIDE</code> and <code>KW_MOD</code>
 * operators.
 * <p>
 * This is only called by the 5th level {@link #sum_expr} rule. This  
 * method subsequently calls the {@link #un_type} rule. 
 * <p>
 * See the top level expression parsing rule {@link #expr} for more details.
 * <p>
 * Note that ANTLR reports ambiguity between this rule and the {@link #assign}
 * rule.  The ambiguity is resolved by the ordering of the top level rule
 * placing the <code>ASSIGN</code> language statement (which is unambiguous)
 * with higher precedence than the <code>assignment</code> rule. As long as
 * the <code>assign</code> rule is only used by the <code>assign_stmt</code>
 * rule, there is no real ambiguity.  For this reason, ambiguity warnings
 * are suppressed for this rule.
 */
prod_expr
   :
      un_type
      (
         options { generateAmbigWarnings = false; }
         :
         (MULTIPLY^ | DIVIDE^ | KW_MOD^) un_type
      )* 
   ;

/**
 * Implements the 7th precedence level of Progress 4GL expressions, the
 * unary operators <code>PLUS</code> and <code>MINUS</code>. This rule is
 * different in structure from the previous levels, as the operator is
 * leftmost and optional.  There is always a right operand.
 * <p>
 * In any case where the programmer did not use intervening whitespace
 * between the operator and a literal value (integer or decimal), the 
 * lexer will have already included that sign character in the text of
 * the literal's token.  So this rule would not apply.  This rule can only
 * be encountered when there is a separate operator token in the token stream
 * which only occurs in the absence of whitespace. Since postfixed unary
 * operators can only occur on literals and then only without intervening
 * whitespace, this rule never has to match on a postfixed unary operator.
 * <p>
 * This is only called by the 6th level {@link #prod_expr} rule. This  
 * method subsequently calls the {@link #primary_expr} rule. 
 * <p>
 * See the top level expression parsing rule {@link #expr} for more details.
 */
un_type
   :
      (
           {
              !(allowStringMatch &&
                LA(2) == MINUS   &&
                !followedByWhitespace(LT(1)))
           }?
           m:MINUS^ { #m.setType( UN_MINUS ); }
           
         | {
              !(allowStringMatch &&
                LA(2) == PLUS   &&
                !followedByWhitespace(LT(1)))
           }?
           p:PLUS^  { #p.setType( UN_PLUS ); }
      )?
      chained_object_members
   ;
   
/**
 * Implements the 8th precedence level of Progress 4GL expressions, the chaining of object
 * and handle references. A {@link #primary_expr} is always matched as the first component. If
 * the node(s) from the primary expression are an object reference (a reference to a Progress
 * user defined type/OO 4GL class) OR if the node(s) are a HANDLE or COM-HANDLE type, then the
 * code checks for a following <code>COLON</code> operator without any following white space.
 * Under these conditions, one of the {@link #downstream_chained_reference} (for OO 4GL refs),
 * {@link #com_property_or_method} (COM-HANDLE refs) or  {@link #attribute_or_method} (anything
 * else) will be called.
 * <p>
 * This rule enables chaining of these constructs on a completely mix and match basis which can
 * continue indefinitely so long as the most recently returned node is an object, com-handle or
 *  handle reference and there is a following <code>COLON</code> without any following white
 * space.
 * <p>
 * If the preceding node is an object type, the exact fully qualified class name being referenced
 * is read from the "qualified" annotation in the node.  That class name is used to resolve the
 * method or data member being referenced via chaining. The rule
 * {@link #downstream_chained_reference} provides the matching logic for both of those cases,
 * but the chaining itself is done here. The operator used for chaining is a <code>COLON</code>
 * but this will be converted to an <code>OBJECT_INVOCATION</code> before return.
 * <p>
 * If the given node is a COM-HANDLE type, if it is then the rule {@link #com_property_or_method}
 * provides the matching logic but the downstream chaining itself is done here. The operator used
 * for chaining is a <code>COLON</code> but this will be converted to a
 * <code>COM_INVOCATION</code> before return.
 * <p>
 * Supports the <code>DB_REF_NON_STATIC</code> operator. This is the syntax added in 10.2x which
 * allows a database table name or table + field name to be directly referenced off a handle using
 * the <code>::</code> operator.  This includes chaining support and mix/match chaining.
 * <p>
 * This is only called by the 7th level {@link #un_type} rule.
 * <p>
 * See the top level expression parsing rule {@link #expr} for more details.
 */
chained_object_members
   :
      {
         boolean active_x = false;
         boolean object   = false;
         boolean isStatic = false;
         String  cls      = null;
         Aast    refnode  = null;
      }
      pe:primary_expr  { refnode = #pe; }
      (
         options { generateAmbigWarnings = false; }
         :
           // we MUST NOT match a COLON here unless there is NO following whitespace
           { !followedByWhitespace(LT(1)) }?
           col:COLON^
           {
              if (refnode == null)
              {
                 throw genExc("Missing refnode for COLON operator!", LT(1));
              }
           
              // COM-HANDLE reference (or possible reference)
              active_x = isComHandleType(refnode);
              
              // determine if the primary expression is an object or class reference
              object = isObjectSubtree(refnode);
              
              if (object)
              {
                 isStatic = isStaticObject(refnode);
              
                 // determine to what specific class that reference refers
                 cls = getObjectClassName(refnode);
              }
           }
           
           (
                { object }?
                dr:downstream_chained_reference[cls, isStatic]
                {
                   #col.setType(OBJECT_INVOCATION);
                   refnode = #dr;
                }
             
              | { active_x }?
                com_property_or_method { #col.setType(COM_INVOCATION); }
              
                // we don't test whether something is a handle type here because there are a
                // range of use cases that allow dereferencing non-handle types; for example,
                // there is TEMP-TABLE tt:HANDLE
              | am:attribute_or_method  { refnode = #am; }
                (
                  options { generateAmbigWarnings = false; }
                  :
                  // this is needed for constructs such as var:ATTR IN FRAME frame
                  widget_qualifier
                )?
           )
           
         | DB_REF_NON_STATIC^ sym1:reserved_or_symbol { #sym1.setType(SYMBOL); }
           (
              options { generateAmbigWarnings = false; }
              :
              DB_REF_NON_STATIC^ sym2:reserved_or_symbol { #sym2.setType(SYMBOL); }
           )?
      )*
   ;
      
/**
 * Helper rule to match a {@link #method_call}, {@link #class_event} or an 
 * {@link #object_data_member} in a {@link #chained_object_members}. This
 * rule will match one or the other and will disambiguate the two based on
 * a lookup of the method name in the given class' namespace. This method
 * MUST ONLY be used from the 3rd-most token from the left of the expression
 * (or further right than that) since it uses {@link #any_symbol_at_all} to
 * pull even reserved keywords into this rule and then it rewrites the token
 * types before lookahead picks the subrules. This allows the subrules to
 * be common code that are also called from {@link #primary_expr} (where they
 * are not allowed to match with reserved keywords, even in Progress, and doing
 * so would hose the parser bigtime).
 *
 * @param    cls
 *           Class name containing the members for which to search.
 * @param    isStatic    
 *           <code>true</code> if this is a static reference.
 */
downstream_chained_reference [String cls, boolean isStatic]
   :
      {
         CommonToken tok      = (CommonToken) LT(1);
         String      txt      = tok.getText();
         int         type     = tok.getType();
         boolean     unquoted = false;
         int         newtype  = -1;
         
         // special unquoted method name matching case
         if (insideClassEvent && LA(2) != LPARENS)
         {
            if (sym.lookupObjectMethod(null, txt, isStatic) != -1)
            {
               newtype  = SYMBOL;
               unquoted = true;
            }
         }
         
         if (newtype == -1)
         {
            // try the common cases
            if (LA(2) == LPARENS)
            {
               if (!isStatic)
               {
                  // instance-qualified members can be static or instance
                  newtype = sym.lookupObjectMethod(cls, txt);
               }
               else
               {
                  // class-qualified members can only be static
                  newtype = sym.lookupObjectMethod(cls, txt, true);
               }
            }
            else
            {
               if (!isStatic)
               {
                  // instance-qualified members can be static or instance
                  newtype = sym.lookupDataMember(cls, txt);
               }
               else
               {
                  // class-qualified members can only be static
                  newtype = sym.lookupDataMember(cls, txt, true);
               }
            }
         }
         
         // if we have to rewrite, do it here
         if ((newtype >= BEGIN_METH && newtype <= END_METH)         ||
             (newtype >= BEGIN_VARTYPES && newtype <= END_VARTYPES) ||
             newtype == CLASS_EVENT                                 ||
             newtype == SYMBOL)
         {
            tok.setType(newtype);
         }
         else
         {
            // must be an unreserved keyword, reserved keyword, DB_SYMBOL,
            // FILENAME, BOOL_FALSE or BOOL_TRUE that is not recognized as a
            // class member
            throw new NoViableAltException(tok, getFilename());
         }
      }
      (
         options { generateAmbigWarnings = false; }
         :
           // special case of recursion from within class_event, where some
           // nasty ambiguity requires matching on an unquoted method name to
           // be done here
           { unquoted }?
           SYMBOL
           
           // this matches a method call which can only occur inside the
           // referenced class definition (or its parent hierarchy); note
           // that we must avoid matching an lvalue followed by an aggregate
           // phrase or some other silly Progress construct that looks
           // exactly like a method call... this is safe to do anyway
         | { LA(1) >= BEGIN_METH && LA(1) <= END_METH }? 
           method_call[cls, isStatic]
           
           // disambiguation logic (like method calls above)
         | { LA(1) == CLASS_EVENT }? 
           class_event[cls, isStatic]
           
           // this matches all data members and properties defined in the
           // referenced class definition (or its parent hierarchy)
         | object_data_member[cls, isStatic]
         
           // the inclusion of the any_symbol_at_all rule reference exists to
           // "pull" us into this rule, but when we leave the type is always
           // reset to one of the valid types OR we throw an exception
         | any_symbol_at_all
      )
   ;
   
/**      
 * Matches all Progress 4GL {@link #lvalue} references that exist in the
 * namespace of the given class name.
 * <p>
 * This rule is simply a convenience method to use the current class name 
 * instance variable to bias the <code>lvalue</code> to lookup data members
 * within the given namespace. When done, the flag is set back to its prior
 * state.
 *
 * @param    cls
 *           Class name containing the members for which to search.
 * @param    isStatic    
 *           <code>true</code> if this is a static reference.
 */
object_data_member [String cls, boolean isStatic]
   :
      {
         String oldClassName = currentClassName; 
         currentClassName = cls;
         
         boolean oldStaticFlag = currentStaticFlag;
         currentStaticFlag = isStatic;
      }
      lvalue
      {
         currentClassName  = oldClassName;
         currentStaticFlag = oldStaticFlag;
      }
   ;
   
/**
 * Implements the 9th and highest precendence level of Progress 4GL
 * expressions, the parenthesis precedence operators and any all of the
 * common expression constructs.  The following are supported (in this
 * order):
 * <p>
 * <ol>
 *    <li> {@link #literal}
 *    <li> {@link #new_phrase}
 *    <li> {@link #sql_aggregate_funcs}
 *    <li> {@link #seek_func}
 *    <li> {@link #cast_func}
 *    <li> {@link #sequence_funcs}
 *    <li> {@link #dynamic_function_func}
 *    <li> {@link #frame_funcs}
 *    <li> {@link #record_funcs}
 *    <li> {@link #func_call}
 *    <li> {@link #method_call}
 *    <li> {@link #class_event}
 *    <li> {@link #user_defined_type_name}
 *    <li> {@link #accum_function}
 *    <li> {@link #can_find_function}
 *    <li> {@link #lvalue}
 *    <li> {@link #postfix_funcs}
 *    <li> {@link #if_func}
 *    <li> {@link #super_function}
 * </ol>
 * <p>
 * Progress has several unusual functions that don't parenthesize parameters
 * (e.g. available or ambiguous) and sometimes the function names are even
 * postfixed after the first parameter (entered or not entered)! This
 * design requires that such functions have hard coded matching rules that
 * are called from here.
 * <p>
 * This rule is different in structure from the previous levels, as the
 * there is just a set of alternatives and one of them allows direct recursion
 * back to the top level <code>expr</code> expression entry point.
 * <p>
 * Due to the nature of the <code>func_call</code> and <code>lvalue</code>
 * rules, there is an inherent ambiguity since either of these can be
 * comprised of a user-defined symbol.  We use 2 tricks to resolve this
 * ambiguity:
 * <ul>
 *    <li> Both rule references must specify a match with the full set of
 *         possible token types that can be seen as user-defined symbols.
 *         This is done by referencing a common {@link #symbol} rule. This
 *         rule reference generates the proper lookahead code in the top
 *         level parser rules, such that these rules will be triggered
 *         when necessary.  By placing the more specific of the two rules
 *         <code>func_call</code> first, the lack of a match on the
 *         subsequent parenthesis causes processing to drop through to
 *         <code>lvalue</code>, resolving the ambiguity generated by the
 *         need to have a common reference to the <code>symbol</code> rule.
 *         It is most important to understand that without this common rule
 *         reference, the generated lookahead code would be incorrect and
 *         these rules would not work.
 *    <li> When these rules are properly matched by lookahead, the symbol
 *         token types are rewritten to match the function or variable
 *         types that are required by the respective rules.  This is critical
 *         to allowing expression evaluation to occur, otherwise there would
 *         be no real record of the proper token type (be it an lvalue or
 *         function call).  In addition, these token types are data type
 *         specific (e.g. there is a type for <code>VAR_INT</code>, 
 *         <code>VAR_LOGICAL</code> and all other variable data types).
 * </ul>
 * <p>
 * <code>method_call</code> is similar to <code>func_call</code> and is
 * disambiguated from <code>lvalue</code> and <code>func_call</code> with a
 * semantic predicate that does a method name lookup.
 * <p>
 * Similar to the problems with <code>lvalue</code> and <code>func_call</code>
 * <code>record_funcs</code> are also ambiguous since there is a
 * {@link #record} which is required in the leftmost token when dealing with
 * a postfixed function name.  This ambiguity is resolved by a query to the
 * schema namespace in a semantic predicate. If the symbol is a record, the
 * record rule is matched, otherwise it is bypassed.
 * <p>
 * This is only called by the 8th level {@link #chained_object_members} rule.
 * <p>
 * See the top level expression parsing rule {@link #expr} for more details.
 * <p>
 * ANTLR will detect ambiguity between many of these  alternatives.  This has
 * been determined to be non-ambiguous for the Progress domain so ambiguity
 * warnings have been disabled. 
 */
primary_expr
   :
      {
         boolean connected = false;
         boolean isStatic  = false;
         String  nextTxt   = LT(1).getText();
         
         // fixup the LT(1) by discarding the crap and putting the last token
         // back together
         if (isMalformedSymbolFuncCall())
         {
            nextTxt = reassembleMalformedSymbol();
         }
         
         // the connected built-in can match an unquoted dbname that can
         // appear as a symbol (but we can't JUST match on that since it
         // can also be a var which appears here as a symbol token), so we
         // temporarily enable symbol matching in the lvalue rule
         if (LA(1) == KW_CONN_ED && LA(2) == LPARENS)
         {
            connected = true;
            allowSymbolMatch = true;
         }
      }
      (
         options { generateAmbigWarnings = false; }
         :
           literal
           
           // must be before record_funcs; the 2nd check is necessary in case that there
           // is a class name that is also a table name (canLoadClass is expensive so we
           // try to avoid calling it by doing a quick check on the table names)
         | {
              LA(2) != LPARENS &&
              (!sym.isTable(LT(2).getText()) || sym.canLoadClass(LT(2).getText()))
           }?
           new_phrase
           
           // these next functions have specialized signature processing, but
           // they otherwise work normally (from a syntax perspective), so
           // they must occur before func_call since it would otherwise
           // conflict
         | { embeddedSql }?
           sql_aggregate_funcs
         | seek_func
         | cast_func
         | sequence_funcs
         | dynamic_function_func
         | frame_funcs
         
           // the prediction logic can't properly differentiate the only case where we have the
           // non-reserved ERROR keyword (which must have a following LPARENS), so we must avoid
           // this path if there is no LPARENS
         | { LA(1) != KW_ERROR || LA(2) == LPARENS }?
           record_funcs
           
           // avoid matching an lvalue followed by an aggregate phrase!
           // or some other silly Progress construct that looks exactly
           // like a function call...  this is safe to do anyway
         | { 
              (LA(1) >= BEGIN_FUNCTYPES && LA(1) <= END_FUNCTYPES) ||
              (sym.lookupFunction(nextTxt) != -1)
           }? 
           func_call
           
           // this matches a "standalone" method call which can only occur
           // inside a class definition; note that we must avoid matching an
           // lvalue followed by an aggregate phrase or some other silly
           // Progress construct that looks exactly like a method call...
           // this is safe to do anyway; reserved keywords CANNOT be matched via
           // this location 
         | { 
              (LA(1) >= BEGIN_METH && LA(1) <= END_METH)                              ||
              (inStaticCtxt  && (sym.lookupObjectMethod(null, nextTxt, true) != -1))  ||
              (!inStaticCtxt && (sym.lookupObjectMethod(null, nextTxt) != -1))
           }?
           { isStatic = sym.isStaticMethod(null, nextTxt); }
           method_call[null, isStatic]
           
           // standalone event reference, this is disambiguation logic (like
           // method calls above); reserved keywords CANNOT be matched via
           // this location
           // TODO: does inStaticCtxt need to be used here?
         | { 
              LA(1) == CLASS_EVENT ||
              (sym.lookupDataMember(null, nextTxt) == CLASS_EVENT)
           }?
           { isStatic = sym.isStaticDataMember(null, nextTxt); }
           class_event[null, isStatic]
           
           // standalone data member reference must match here instead of the "bare bones"
           // lvalue below
         | { 
              (inStaticCtxt  && (sym.lookupDataMember(null, nextTxt, true) != -1)) ||
              (!inStaticCtxt && (sym.lookupDataMember(null, nextTxt) != -1))
           }?
           { isStatic = sym.isStaticDataMember(null, nextTxt); }
           object_data_member[null, isStatic]
           
           // class name references (used for static member access); the STRING version cannot
           // be matched here (which is probably fine) since any string is matched in the
           // literal rule above; we MUST exclude anything that matches BOTH a variable name and
           // a classname since the 4GL will prefer a var to a static reference in this case
           // TODO: check if unqualified object data members are preferred here
         | { sym.lookupVariable(nextTxt) == -1 && sym.canLoadClass(nextTxt) }?
           non_reserved_user_defined_type_name
           
         | accum_function
         | can_find_function
         
           // avoid matches to the super object reference which is an lvalue
         | { LA(2) != COLON }?
           super_function
           
         | q:lvalue 
           (
                { 
                   (#q.getType() == KW_FRAME || #q.getType() == KW_BROWSE) &&
                   ((sym.lookupVariable(LT(1).getText()) != -1) ||
                    (sym.isField(LT(1).getText())))
                }?
                l:lvalue
                (
                     // ANTLR's linear approximation approach makes this
                     // situation ambiguous when it shouldn't be; by adding
                     // this semantic predicate, the "connection" between the
                     // NOT and ENTERED is restored
                     {
                        (LA(1) == KW_NOT && LA(2) == KW_ENTERED) || 
                        (LA(1) == KW_ENTERED)              
                     }?
                     p1:postfix_funcs
                  | // empty alternative
                )
              | // ANTLR's linear approximation approach makes this situation
                // ambiguous when it shouldn't be; by adding this semantic
                // predicate, the "connection" between the NOT and ENTERED
                // is restored
                {
                   (LA(1) == KW_NOT && LA(2) == KW_ENTERED) || 
                   (LA(1) == KW_ENTERED)              
                }?
                p2:postfix_funcs
                
              | // empty alternative           
           )
         | if_func
         
         | LPARENS^ expr RPARENS!   
      )
      {
         if (connected)
         {
            allowSymbolMatch = false;
         }
      
         // rewrite entered/not entered to match normal func call 
         if (#p1 != null || #p2 != null)
         {
            Aast post = (#p1 != null ? #p1 : #p2);
            
            // attach the main lvalue
            #q.setNextSibling(null);
            post.setFirstChild(#q);

            // avoid infinite loops in tree building ops, so we "do it 
            // ourselves"
            if (#l != null)
            {
               #l.setNextSibling(null);
               #q.setNextSibling(#l);
            }
            
            ## = post;
         }
      }
   ;
   
/**
 * Matches either a <code>NUM_LITERAL</code> or a <code>DEC_LITERAL</code>.
 */
numeric_literal
   :
      (NUM_LITERAL | DEC_LITERAL)
   ;

/**  
 * Matches all variations of Progress literals (constants).  This includes:
 * <pre>
 *    NUM_LITERAL
 *    DEC_LITERAL
 *    STRING
 *    BOOL_TRUE
 *    BOOL_FALSE 
 *    DATE_LITERAL
 *    DATETIME_LITERAL
 *    DATETIME_TZ_LITERAL
 *    UNKNOWN_VAL
 *    KW_FIND_NO  (compiler constant == 1)
 *    KW_FIND_PO  (compiler constant == 2)
 *    KW_FIND_CS  (compiler constant == 4)
 *    KW_FIND_GLO (compiler constant == 8)
 *    KW_FIND_WA  (compiler constant == 16)
 *    KW_FIND_SEL (compiler constant == 32)
 *    KW_READ_AVL (compiler constant == 1)
 *    KW_READ_E_N (compiler constant == 2)
 *    KW_ROW_UMOD (ROW-STATE value, compiler constant == 0)
 *    KW_ROW_DELD (ROW-STATE value, compiler constant == 1)
 *    KW_ROW_MODD (ROW-STATE value, compiler constant == 2)
 *    KW_ROW_CRTD (ROW-STATE value, compiler constant == 3)
 *    KW_SAX_COMP (compiler constant == 3)
 *    KW_SAX_PARE (compiler constant == 4)
 *    KW_SAX_RUNN (compiler constant == 2)
 *    KW_SAX_UNIN (compiler constant == 1)
 *    KW_SAX_WBEG (compiler constant == 2)
 *    KW_SAX_WCOM (compiler constant == 6)
 *    KW_SAX_WCON (compiler constant == 5)
 *    KW_SAX_WELM (compiler constant == 4)
 *    KW_SAX_WERR (compiler constant == 7)
 *    KW_SAX_WIDL (compiler constant == 1)
 *    KW_SAX_WTAG (compiler constant == 3)
 *    KW_WIN_DMIN (undocumented keyword which is a compiler constant == 4)
 *    KW_WIN_MAX (compiler constant == 1)
 *    KW_WIN_MIN (compiler constant == 2)
 *    KW_WIN_NORM (compiler constant == 3)
 *    KW_SEAR_SLF (SEARCH-SELF mode for the add-super-procedure() parameter, compiler constant == 1)
 *    KW_SEAR_TRG (SEARCH-TARGET mode for the add-super-procedure() parameter, compiler constant == 2)
 *    KW_HOST_B_O (compiler constant == 1)
 *    KW_B_ENDIAN (compiler constant == 2)
 *    KW_L_ENDIAN (compiler constant == 3)
 * </pre>
 * <p>
 * Note that the documentation states that there is no such thing as a date
 * constant in Progress 4GL, however experience shows that this is not true.
 * <p>
 * There are no literal representations for RECID, ROWID or HANDLE data
 * types since these are internal data structures of the Progress language.
 * One can only obtain an instance of such a value from Progress itself so
 * there is no way to hard-code such a reference as a constant.
 * <p>
 * There is no literal representation for RAW data type.
 * <p>
 * Certain upstream usage of expressions (see {@link #form_item}) enable a
 * special mode of parsing where 2 or more unquoted '-' or '+' characters
 * are matched as a string literal. Note that the characters cannot be
 * intermixed without intervening whitespace and no other non-alphanumeric
 * characters on the U.S. keyboard have this same behavior. 
 */
literal
   :
      {
         StringBuilder sb = new StringBuilder("\'");
      }
      (
           NUM_LITERAL
         | DEC_LITERAL
         | STRING
         | BOOL_TRUE
         | BOOL_FALSE 
         | DATE_LITERAL
         | DATETIME_LITERAL
         | DATETIME_TZ_LITERAL
         | UNKNOWN_VAL
         | nl:KW_NO_LOCK { #nl.setType(NO_LOCK_LITERAL); }
         | sl:KW_SH_LOCK { #sl.setType(SHARE_LOCK_LITERAL); }
         | el:KW_EXC_LOCK { #el.setType(EXCLUSIVE_LOCK_LITERAL); }
         | nw:KW_NO_WAIT { #nw.setType(NO_WAIT_LITERAL); }
         | ra:KW_READ_AVL { #ra.setType(READ_AVAILABLE_LITERAL); }
         | re:KW_READ_E_N { #re.setType(READ_EXACT_NUM_LITERAL); }
         | KW_FIND_NO 
         | KW_FIND_SEL 
         | KW_FIND_PO 
         | KW_FIND_CS 
         | KW_FIND_GLO
         | KW_FIND_WA 
         | KW_ROW_UMOD
         | KW_ROW_DELD
         | KW_ROW_MODD
         | KW_ROW_CRTD
         | KW_SAX_COMP
         | KW_SAX_PARE
         | KW_SAX_RUNN
         | KW_SAX_UNIN
         | KW_SAX_WBEG
         | KW_SAX_WCOM
         | KW_SAX_WCON
         | KW_SAX_WELM
         | KW_SAX_WERR
         | KW_SAX_WIDL
         | KW_SAX_WTAG
         | KW_WIN_DMIN
         | KW_WIN_MAX
         | KW_WIN_MIN
         | KW_WIN_NORM
         | ss:KW_SEAR_SLF { #ss.setType(SEARCH_SELF_LITERAL); }
         | st:KW_SEAR_TRG { #st.setType(SEARCH_TARGET_LITERAL); }
         | KW_HOST_B_O
         | KW_B_ENDIAN
         | KW_L_ENDIAN
         | { allowStringMatch && !followedByWhitespace(LT(1)) }?
           MINUS! { sb.append('-'); }
           (
              options { generateAmbigWarnings = false; }
              :
              { !prefacedByWhitespace(LT(1)) }?
              m:MINUS!
              {
                 sb.append('-');
              }
           )+
         | { allowStringMatch && !followedByWhitespace(LT(1)) }?
           PLUS!  { sb.append('+'); }
           (
              options { generateAmbigWarnings = false; }
              :
              { !prefacedByWhitespace(LT(1)) }?
              p:PLUS!
              {
                 sb.append('+');
              }
           )+
      )
      {
         if (sb.length() > 1)
         {
            // close off the contents
            sb.append('\'');
         
            // we must manufacture a STRING node here
            ## = #([STRING, sb.toString()]);
         }
         
         ##.putAnnotation("is-literal", new Boolean(true));
      }
   ;   
      
/**      
 * Matches all Progress 4GL function calls that take parenthesized parameter
 * lists.  Built-in Progress functions that don't take arguments also do not
 * take parenthesized parameters.  These special cases are handled as
 * predefined variables in the symbol dictionary, since they essentially are
 * read-only variables even though they are called functions.
 * <p>
 * The matching portion of this rule comes in 2 parts:
 * <ul>
 *    <li> The leftmost token is the function name.  This is defined as the
 *         union of the set of all function token types (one for each return 
 *         data type, e.g. <code>FUNC_INT</code>) and the set of all possible
 *         alternatives for symbol names (a reference to the {@link #symbol}
 *         for function names). Both <code>lvalue</code> and
 *         <code>func_call</code> refer to a leftmost token that includes the
 *         symbol text.  This creates an ambiguity that is only resolved by
 *         lookahead <b>and</b> the proper ordering of the rules (since both
 *         are equal alternatives in the <code>primary_expr</code> rule).
 *    <li> The next part is the {@link #func_call_parameters}.
 * </ul>
 * <p>
 * It is required that where this rule is referenced as one alternative that
 * conflicts with the definitions of other alternatives (see
 * {@link #primary_expr} and {@link #lvalue}), <b>the order of the 
 * alternatives must be more specific (in terms of matching criteria) to less
 * specific</b>.  Since the <code>func_call</code> rule contains the very
 * specific <code>symbol</code> followed by <code>LPARENS</code> pattern,
 * it is more specific than a simple <code>symbol</code> by itself (which is
 * the definition of an <code>lvalue</code>.  For this reason, 
 * <code>func_call</code> must always be placed ahead of <code>lvalue</code>
 * in alternative lists.
 * <p>
 * The inclusion of the <code>malformed_symbol</code> rule ensures that all
 * symbols that are followed by an <code>LPARENS</code> are considered a
 * function call and will be matched by this rule.  This means that the 
 * top level expression processing will include a symbol as one of the options
 * for its leftmost token, thus &quot;pulling&quot; the recursive descent
 * processing down the <code>expr</code> rule.
 * <p>
 * It is important to note that in the end, it is not really valid to have
 * a token type of <code>SYMBOL</code> since this could be anything.  What we
 * really require is a leftmost token that is one of the valid function
 * tokens:
 * <pre>
 *   FUNC_CHAR
 *   FUNC_CLASS
 *   FUNC_COM_HANDLE
 *   FUNC_DATE
 *   FUNC_DATETIME
 *   FUNC_DATETIME_TZ
 *   FUNC_DEC
 *   FUNC_INT
 *   FUNC_INT64
 *   FUNC_HANDLE
 *   FUNC_LOGICAL
 *   FUNC_LONGCHAR
 *   FUNC_MEMPTR
 *   FUNC_RAW
 *   FUNC_RECID
 *   FUNC_ROWID
 *   FUNC_POLY
 * </pre>
 * The difference between a <code>SYMBOL</code> and a <code>FUNC_INT</code>
 * can only be determined by context (the fact that the lookahead code has
 * identified a match with a <code>SYMBOL</code> token followed by an
 * <code>LPARENS</code> token which is a construct that can <b>only</b> occur
 * in a function call.  Thus <b>if</b> a match is found using lookahead (in 
 * the rule that calls this one), then it <b>must</b> be correct and 
 * necessary to convert the <code>SYMBOL</code> token type into the correct 
 * function type, using a lookup into the function namespace. This is done in
 * an init action that is processed at the beginning of this rule. In this
 * way, the rest of the rule only sees the first token as a valid function
 * token type and never tries to call the <code>symbol</code> rule since
 * calling that rule would actually try to match (and consume from the
 * token stream) a <code>SYMBOL</code> token type.  <b>Specifying a rule
 * reference to obtain its value from a lookahead perspective but then
 * changing the token's type such that the rule reference never gets called,
 * is the critical trick that enables context sensitive symbol names!</b>
 * <p>
 * All functions have a token type associated with their respective return
 * value.  Some built-in functions may have the special type of 
 * <code>FUNC_POLY</code>. This special type indicates that the function can
 * return multiple different return types. The built-in functions typically
 * implement this return type based on the type of the parameters passed. 
 * User-defined functions cannot provide the same feature, this can only 
 * occur in built-in functions.
 * <p>
 * The exit action will save any original token type that rewriting may have
 * erased.  This will allow future processing to access this as an annotation
 * named 'oldtype'.
 * <p>
 * Built-in functions have a language keyword as the function name. This 
 * means that built-ins can support abbreviations.  It also means that some
 * built-in functions cannot be hidden by a user-defined function, if the
 * associated keyword is reserved. If the keyword is not reserved, then a
 * built-in function is hidden in the namespace by a user-defined function
 * of the same name.  This method's token type rewriting is dependent upon
 * this logic being implemented in the <code>SymbolResolver</code> class.
 * See {@link SymbolResolver#lookupFunction}.
 * <p>
 * To ensure this method sees tokens matching the list of functions with
 * reserved names, the {@link #reserved_functions} method is included as
 * an alternative.  This is the same 'pull' trick that is being used for
 * generic symbols above.  The reason that the generic symbol technique is
 * not enough to pull reserved functions is that only unreserved keywords
 * are pulled by the <code>symbol</code> method.  This can't be changed
 * without negative impacts on other code, since <code>symbol</code> is 
 * used in many locations in the parser.  Instead a custom method was
 * created to generate pull.
 */
func_call   
   :
      {
         int     oldtype  = -1;
         int     current  = LA(1);
         String  funcname = LT(1).getText();
         boolean builtin  = false;
         
         if (current < BEGIN_FUNCTYPES || current > END_FUNCTYPES)
         {
            Function func    = sym.lookupFunctionDef(funcname);
            int      newtype = sym.lookupWrappedType(func);
            
            // there should always be a match in valid Progress 4GL source
            // so we throw an exception if this is not the case
            if (newtype != -1)
            {
               builtin = func.isBuiltin();
               oldtype = LA(1);
               LT(1).setType(newtype);
            }
            else
            {
               throw new NoViableAltException(LT(1), getFilename());
            }
         }
      }
      (
         (
              FUNC_CHAR^
            | FUNC_CLASS^
            | FUNC_COM_HANDLE^
            | FUNC_DATE^
            | FUNC_DATETIME^
            | FUNC_DATETIME_TZ^
            | FUNC_DEC^
            | FUNC_HANDLE^
            | FUNC_INT^
            | FUNC_INT64^
            | FUNC_LOGICAL^ 
            | FUNC_LONGCHAR^
            | FUNC_MEMPTR^
            | FUNC_RAW^
            | FUNC_RECID^
            | FUNC_ROWID^
            | FUNC_POLY^ 
              
            // the inclusion of the reserved_variables rule reference exists
            // to "pull" us into the func_call rule, but when we leave the
            // type is always reset one of the valid datatypes OR we throw an
            // exception
            | reserved_functions
         
            // the inclusion of the symbol rule reference exists to "pull" us
            // into this rule, but when we leave the type is always reset
            // one of the valid datatypes OR we throw an exception
            | symbol
            
            // malformed function call names can start with one of these
            // characters, this is just for "pull", the tokens will be
            // rewritten in the primary_expr rule
            | UNKNOWN_TOKEN
         )
         func_call_parameters[builtin]
      )
      {
         // save any original token type that rewriting may have erased
         if (oldtype != -1)
            ##.putAnnotation("oldtype", new Long(oldtype));
            
         // write our function-specific annotations
         sym.annotateFunction(funcname, ##, false);            
      }
   ;
   
/**      
 * Matches all Progress 4GL method calls and their parenthesized parameter
 * lists.
 * <p>
 * The matching portion of this rule comes in 2 parts:
 * <ul>
 *    <li> The leftmost token is the method name.  This is defined as the
 *         union of the set of all method token types (one for each return 
 *         data type, e.g. <code>METH_INT</code>) and the set of all possible
 *         alternatives for symbol names (a reference to the
 *         {@link #any_non_reserved_symbol}. Although all reserved keywords can
 *         be used as method names, this use of all possible keywords would
 *         create huge ambiguity if allowed in the leftmost token of an
 *         expression. This rule can be called in such a location for a
 *         "standalone" method call (within the class in which it is
 *         defined). However, in that case Progress disallows matching with
 *         reserved keywords, so one must use the class name (for static
 *         events) or THIS-OBJECT (for instance events) as a qualifier.
 *         That is why this rule doesn't "pull" the reserved keywords. The
 *         {@link #downstream_chained_reference} does include reserved
 *         keywords and it does call this rule.  However, it rewrites the
 *         token types up front so that the lookahead won't break.
 *    <li> The next part is the {@link #func_call_parameters}.
 * </ul>
 * <p>
 * It is required that where this rule is referenced as one alternative that
 * conflicts with the definitions of other alternatives (see
 * {@link #primary_expr} and {@link #lvalue}), <b>the order of the 
 * alternatives must be more specific (in terms of matching criteria) to less
 * specific</b>.  Since the <code>method_call</code> rule contains the very
 * specific <code>symbol</code> followed by <code>LPARENS</code> pattern,
 * it is more specific than a simple <code>symbol</code> by itself (which is
 * the definition of an <code>lvalue</code>.  For this reason, 
 * <code>func_call</code> must always be placed ahead of <code>lvalue</code>
 * in alternative lists.
 * <p>
 * The inclusion of the <code>any_symbol_at_all</code> rule ensures that all
 * valid symbols that are followed by an <code>LPARENS</code> are considered a
 * method call and will be matched by this rule.  This means that the 
 * top level expression processing will include a symbol as one of the options
 * for its leftmost token, thus &quot;pulling&quot; the recursive descent
 * processing down the <code>expr</code> rule.
 * <p>
 * It is important to note that in the end, it is not really valid to have
 * a token type of <code>SYMBOL</code> since this could be anything.  What we
 * really require is a leftmost token that is one of the valid function
 * tokens:
 * <pre>
 *   METH_CHAR
 *   METH_CLASS
 *   METH_COM_HANDLE
 *   METH_DATE
 *   METH_DATETIME
 *   METH_DATETIME_TZ
 *   METH_DEC
 *   METH_INT
 *   METH_INT64
 *   METH_HANDLE
 *   METH_LOGICAL
 *   METH_LONGCHAR
 *   METH_MEMPTR
 *   METH_RAW
 *   METH_RECID
 *   METH_ROWID
 *   METH_POLY
 *   METH_VOID
 * </pre>
 * <p>
 * There is ambiguity between this rule and the {@link #func_call} rule
 * which is resolved by semantic predicates that use the next token's text
 * to determine if there is a match to a function or method call.
 * <p>
 * All methods have a token type associated with their respective return
 * value.  Some built-in method may have the special type of 
 * <code>METH_POLY</code>. This special type indicates that the method can
 * return multiple different return types. <code>METH_VOID</code> is a
 * special type which returns no value.
 * <p>
 * The exit action will save any original token type that rewriting may have
 * erased.  This will allow future processing to access this as an annotation
 * named 'oldtype'.
 *
 * @param    cls
 *           Fully qualified name of the class being referenced or
 *           <code>null</code> if this is within the current class
 *           definition.
 * @param    isStatic    
 *           <code>true</code> if this is a static reference.
 */
method_call [String cls, boolean isStatic]   
   :
      {
         int     oldtype  = -1;
         int     current  = LA(1);
         String  methname = LT(1).getText();
      
         if (current < BEGIN_METH || current > END_METH)
         {
            int newtype = sym.lookupObjectMethod(cls, methname, isStatic);
            
            // there should always be a match in valid Progress 4GL source
            // so we throw an exception if this is not the case
            if (newtype != -1)
            {
               oldtype = LA(1);
               LT(1).setType(newtype);
            }
            else
            {
               throw new NoViableAltException(LT(1), getFilename());
            }
         }
      }
      (
           METH_CHAR^
         | METH_CLASS^
         | METH_COM_HANDLE^
         | METH_DATE^
         | METH_DATETIME^
         | METH_DATETIME_TZ^
         | METH_DEC^
         | METH_HANDLE^
         | METH_INT^
         | METH_INT64^
         | METH_LOGICAL^ 
         | METH_LONGCHAR^
         | METH_MEMPTR^
         | METH_RAW^
         | METH_RECID^
         | METH_ROWID^
         | METH_POLY^
         | METH_VOID^
           
         // the inclusion of the any_non_reserved_symbol rule reference exists
         // to "pull" us into this rule, but when we leave the type is always
         // reset one of the valid datatypes OR we throw an exception
         | any_non_reserved_symbol
      )
      func_call_parameters[false]
      {
         // save any original token type that rewriting may have erased
         if (oldtype != -1)
            ##.putAnnotation("oldtype", new Long(oldtype));
            
         // this leaves notes behind regarding the called method
         sym.annotateObjectMethod(cls, methname, isStatic, ##);
         
         if (##.getType() == METH_CLASS)
         {
            String rcls = sym.lookupObjectMethodClass(cls, methname, isStatic);
            
            if (rcls == null)
            {
               String err = String.format("Missing class name for %s!",
                                          ##.getDescriptiveTokenText());
               throw genExc(err, ##);
            }
            
            // this leaves different annotations than above; the annotations are about
            // the returned reference NOT about the called method
            sym.annotateClassRef(rcls, ##);
         }
      }
   ;
    
/**
 * Matches a parenthesized, comma-delimited list of {@link #func_call_parm}
 * parameters.
 * <p>
 * This matches a set of 2 or more tokens where the first is always
 * <code>LPARENS</code> and the last is always <code>RPARENS</code>. In between is an
 * optional expression, optionally followed by any number of
 * [<code>COMMA</code> func_call_parm] constructs.  This is a parameter list
 * and it can contain any kind of expression.  This rule calls the top level
 * expression entry point allows the full variety of all possible expressions
 * to be specified as parameters.
 * <p> 
 * Note that in the resulting AST, the parenthesis and commas used to define
 * the parameter list are discarded.  In other words, they do not appear in
 * the AST as nodes.  Instead, each parameter (which can be a complete
 * expression subtree or a simple single node like an lvalue) is created
 * as a separate child node of the function type token.  All parameters are 
 * maintained as child nodes in the exact order of the function call itself,
 * so this is essentially the signature of the function.  It it also the
 * proper AST definition to enable simple evaluation of functions with any
 * arbitrary signature.
 * <p>
 * There is an undocumented feature of the 4GL where <code>INPUT</code>,
 * <code>OUTPUT</code> and <code>INPUT-OUTPUT</code> are all ignored if
 * present in a function call parameter. This can be seen in the
 * {@link #func_call_parm}.
 *
 * @param    builtin
 *           <code>true</code> if this is a built-in function.
 */
func_call_parameters [boolean builtin]
   :
      LPARENS!
      ( func_call_parm[builtin] (COMMA! func_call_parm[builtin])* )?
      RPARENS!
   ;
   
/**
 * Represents a function call parameter which is the equivalent of an
 * {@link #expr} BUT it can be optionally prefaced with any of the keywords
 * <code>INPUT, INPUT-OUTPUT or OUTPUT</code>.  These are ignored but seem
 * to be tolerated since it looks the same as <code>RUN</code> statement
 * or <code>PUBLISH</code> statement parameters (as well as the various
 * procedure and function declarations).  The only exception to this is
 * if the function call is to a built-in function that has not been overridden
 * by a user-defined function. In such a case, the <code>INPUT</code> keyword
 * will be treated as the built-in <code>INPUT</code> function instead of
 * being ignored.                                
 * <p>
 * To actually use the <code>INPUT</code> built-in function as a parameter,
 * (in a call to a user-defined function) one must paranthesize the call OR add the
 * following FRAME frame-name qualifier between the KW_INPUT and the widget name.
 * This is how it works in Progress!  Otherwise the presence of the KW_INPUT is
 * treated as an ignored mode qualifier. 
 *
 * @param    builtin
 *           <code>true</code> if this is a parameter to a builtin function.
 */
func_call_parm [boolean builtin]
   {
      boolean isInputFunc = (LA(1) == KW_INPUT &&
                             LA(2) == KW_FRAME &&
                             resolveLvalueCoreType(4, false) != -1);
   }
   :
      // this should be merged with the parameter rule (this one doesn't create a root
      // node and here we match KW_INPUT as a builtin based on the flag passed in) 
      (                             
         options { generateAmbigWarnings = false; }
         :
           { !builtin && !isInputFunc }?
           KW_INPUT                              
         | KW_OUTPUT
         | KW_IN_OUT
      )?
      expr                             
      (KW_APPEND | KW_BY_VALUE | KW_BY_REF | KW_BIND)*
   ;
   
/**      
 * Matches all Progress 4GL class event references and the usage of the publish,
 * subscribe and unsubscribe methods.
 * <p>
 * The matching portion of this rule comes in 3 parts:
 * <ul>
 *    <li> The leftmost token is the event name.  This is defined as the
 *         {@link #any_non_reserved_symbol}. Although all reserved keywords can
 *         be used as class events, this use of all possible keywords would
 *         create huge ambiguity if allowed in the leftmost token of an
 *         expression. This rule can be called in such a location for a
 *         "standalone" event reference (within the class in which it is
 *         defined). However, in that case Progress disallows matching with
 *         reserved keywords, so one must use the class name (for static
 *         events) or THIS-OBJECT (for instance events) as a qualifier.
 *         That is why this rule doesn't "pull" the reserved keywords. The
 *         {@link #downstream_chained_reference} does include reserved
 *         keywords and it does call this rule.  However, it rewrites the
 *         token types up front so that the lookahead won't break.
 *    <li> The next part is the <code>COLON</code> followed by one of the
 *         keywords <code>KW_PUBLISH</code>, <code>KW_SUBSCRIB</code> or
 *         <code>KW_UNSUBSCR</code>. 
 *    <li> The <code>KW_PUBLISH</code> case is followed by
 *         {@link #func_call_parameters}.
 *    <li> The <code>KW_SUBSCRIB</code> and <code>KW_UNSUBSCR</code> cases have
 *         common matching logic.  They are always followed by an
 *         <code>LPARENS</code>. Inside the parenthesis can be an
 *         optional {@link #expr} plus <code>COLON</code> and then an
 *         <code>any_symbol_at_all</code> unquoted method name reference.  
 *         The other alternative inside the parenthesis is an <code>expr</code>
 *         followed by an optional <code>COMMA</code> plus <code>expr</code>.
 *         At the end is always an <code>RPARENS</code>.
 * </ul>
 * <p>
 * It is required that where this rule is referenced as one alternative that
 * conflicts with the definitions of other alternatives (see
 * {@link #primary_expr} and {@link #lvalue}), <b>the order of the 
 * alternatives must be more specific (in terms of matching criteria) to less
 * specific</b>.  For this reason, this rule must always be placed ahead of
 * <code>lvalue</code> in alternative lists.
 * <p>
 * The inclusion of the <code>any_symbol_at_all</code> rule ensures that all
 * valid symbols that are followed by an <code>LPARENS</code> are considered a
 * method call and will be matched by this rule.  This means that the 
 * top level expression processing will include a symbol as one of the options
 * for its leftmost token, thus &quot;pulling&quot; the recursive descent
 * processing down the <code>expr</code> rule.
 * <p>
 * It is important to note that in the end, it is not really valid to have
 * a token type of <code>SYMBOL</code> since this could be anything.  What we
 * really require is a leftmost token that is <code>CLASS_EVENT</code>.
 * <p>
 * The exit action will save any original token type that rewriting may have
 * erased.  This will allow future processing to access this as an annotation
 * named 'oldtype'.
 *
 * @param    cls
 *           Fully qualified name of the class being referenced or
 *           <code>null</code> if this is within the current class
 *           definition.
 * @param    isStatic    
 *           <code>true</code> if this is a static reference.
 */
class_event [String cls, boolean isStatic]   
   :
      {
         int     oldtype  = -1;
         int     current  = LA(1);
         String  evtname = LT(1).getText();
      
         if (current != CLASS_EVENT)
         {
            int newtype = sym.lookupDataMember(cls, evtname, isStatic);
            
            // there should always be a match in valid Progress 4GL source
            // so we throw an exception if this is not the case
            if (newtype == CLASS_EVENT)
            {
               oldtype = LA(1);
               LT(1).setType(newtype);
            }
            else
            {
               throw new NoViableAltException(LT(1), getFilename());
            }
         }
      }
      (
           evt:CLASS_EVENT
           
           // the inclusion of the any_non_reserved_symbol rule reference exists
           // to "pull" us into this rule, but when we leave the type is always
           // reset one of the valid datatypes OR we throw an exception
         | any_non_reserved_symbol
      )
      c:COLON^  { #c.setType(OBJECT_INVOCATION); }
      (
           KW_PUBLISH func_call_parameters[false]
         | (
              (KW_SUBSCRIB | KW_UNSUBSCR) LPARENS!
              (
                   // optimally the unquoted method name form would look
                   // like this:
                   // (expr COLON)? any_symbol_at_all
                   // and the procedure reference form would look like this:
                   // expr (COMMA expr)?
                   // but the above is far too ambiguous, so it is coded
                   // differently to left-factor common matches and use
                   // semantic predicates to avoid overbroad matching
                   
                   { 
                      (LA(1) >= BEGIN_METH && LA(1) <= END_METH) ||
                      (sym.lookupObjectMethod(null, LT(1).getText(), isStatic) != -1)
                   }?
                   any_symbol_at_all
                   
                   // in this case the any_symbol_at_all will be matched
                   // inside a nested call to downstream_chained_reference
                   // from the first expr rule reference instead of being
                   // explicitly coded here; 2nd form is matched pretty
                   // directly here
                 | { insideClassEvent = true; }
                   expr
                   { insideClassEvent = false; }
                   (COMMA! expr)?
              )
              RPARENS!
           )
      )
      {
         // save any original token type that rewriting may have erased
         if (oldtype != -1)
            #evt.putAnnotation("oldtype", new Long(oldtype));
      }
   ;
   
/**      
 * Matches the <code>seek()</code> function call.  The root node returned
 * will be a <code>FUNC_INT</code> and there will be one child which may
 * be <code>KW_INPUT</code>, <code>KW_OUTPUT</code> or an expression that
 * represents a stream.
 * <p>
 * The exit action will save any original token type that rewriting may have
 * erased.  This will allow future processing to access this as an annotation
 * named 'oldtype'.
 * <p>
 * Built-in functions have a language keyword as the function name. This 
 * means that built-ins can support abbreviations.  It also means that some
 * built-in functions cannot be hidden by a user-defined function, if the
 * associated keyword is reserved. If the keyword is not reserved, then a
 * built-in function is hidden in the namespace by a user-defined function
 * of the same name.  This method's token type rewriting is dependent upon
 * this logic being implemented in the <code>SymbolResolver</code> class.
 * See {@link SymbolResolver#lookupFunction}.
 * <p>
 * Note that in the resulting AST, the parenthesis and commas used to define
 * the parameter list are discarded.  In other words, they do not appear in
 * the AST as nodes.  Instead, each parameter (which can be a complete
 * expression subtree or a simple single node like an lvalue) is created
 * as a separate child node of the function type token.  All parameters are 
 * maintained as child nodes in the exact order of the function call itself,
 * so this is essentially the signature of the function.  It it also the
 * proper AST definition to enable simple evaluation of functions with any
 * arbitrary signature.
 * <p>
 * The <code>SEEK</code> function can reference the unnamed input or unnamed
 * output streams using only the keywords <code>INPUT</code> and
 * <code>OUTPUT</code>. This is honored.
 */
seek_func   
   :
      s:KW_SEEK^ { #s.setType(FUNC_INT); } 
      LPARENS!
      (
         options { generateAmbigWarnings = false; }
         :
           KW_INPUT
         | KW_OUTPUT
         | expr
      )
      RPARENS!
      {
         // save the original token type that rewriting erased
         ##.putAnnotation("oldtype", new Long(KW_SEEK));
            
         // write our function-specific annotations
         sym.annotateFunction(#s.getText(), ##, false);            
      }
   ;
   
/**      
 * Matches the <code>cast()</code> function call.  The root node returned
 * will be a <code>FUNC_CLASS</code> and there will be an {@link #expr} child
 * and a second child of {@link #user_defined_type_name}.
 * <p>
 * The exit action will save any original token type that rewriting may have
 * erased.  This will allow future processing to access this as an annotation
 * named 'oldtype'.
 * <p>
 * Built-in functions have a language keyword as the function name. This 
 * means that built-ins can support abbreviations.  It also means that some
 * built-in functions cannot be hidden by a user-defined function, if the
 * associated keyword is reserved. If the keyword is not reserved, then a
 * built-in function is hidden in the namespace by a user-defined function
 * of the same name.  This method's token type rewriting is dependent upon
 * this logic being implemented in the <code>SymbolResolver</code> class.
 * See {@link SymbolResolver#lookupFunction}.
 * <p>
 * Note that in the resulting AST, the parenthesis and commas used to define
 * the parameter list are discarded.  In other words, they do not appear in
 * the AST as nodes.  Instead, each parameter (which can be a complete
 * expression subtree or a simple single node like an lvalue) is created
 * as a separate child node of the function type token.  All parameters are 
 * maintained as child nodes in the exact order of the function call itself,
 * so this is essentially the signature of the function.  It it also the
 * proper AST definition to enable simple evaluation of functions with any
 * arbitrary signature.
 */
cast_func   
   :
      c:KW_CAST^ { #c.setType(FUNC_CLASS); } 
      LPARENS! expr COMMA! udt:user_defined_type_name RPARENS!
      {
         // save the original token type that rewriting erased
         ##.putAnnotation("oldtype", new Long(KW_CAST));
            
         // write our function-specific annotations
         sym.annotateFunction(#c.getText(), ##, false);
         
         // the above probably did not set the class name so we force it here
         String qname = (String) #udt.getAnnotation("qualified");
         sym.annotateClassRef(qname, ##);
      }
   ;
   
/**      
 * Matches the <code>dynamic-function()</code> function call.  The root node
 * returned will be a <code>FUNC_POLY</code> and there will be at least one
 * child which would be an expression that names the function to be called.
 * This mandatory first parameter many optionally be followed by a
 * {@link #in_procedure_clause} and then any number of comma-separated
 * parameter expressions.
 * <p>
 * The exit action will save any original token type that rewriting may have
 * erased.  This will allow future processing to access this as an annotation
 * named 'oldtype'.
 * <p>
 * Built-in functions have a language keyword as the function name. This 
 * means that built-ins can support abbreviations.  It also means that some
 * built-in functions cannot be hidden by a user-defined function, if the
 * associated keyword is reserved. If the keyword is not reserved, then a
 * built-in function is hidden in the namespace by a user-defined function
 * of the same name.  This method's token type rewriting is dependent upon
 * this logic being implemented in the <code>SymbolResolver</code> class.
 * See {@link SymbolResolver#lookupFunction}.
 * <p>
 * Note that in the resulting AST, the parenthesis and commas used to define
 * the parameter list are discarded.  In other words, they do not appear in
 * the AST as nodes.  Instead, each parameter (which can be a complete
 * expression subtree or a simple single node like an lvalue) is created
 * as a separate child node of the function type token.  All parameters are 
 * maintained as child nodes in the exact order of the function call itself,
 * so this is essentially the signature of the function.  It it also the
 * proper AST definition to enable simple evaluation of functions with any
 * arbitrary signature.
 */
dynamic_function_func   
   :
      d:KW_DYN_FUNC^
      {
         #d.setType(FUNC_POLY);
      } 
      LPARENS!
      (
         func_call_parm[false] (in_procedure_clause)?
         (COMMA! func_call_parm[false])*
      )
      RPARENS!
      {
         // save the original token type that rewriting erased
         ##.putAnnotation("oldtype", new Long(KW_DYN_FUNC));
            
         // write our function-specific annotations
         sym.annotateFunction(#d.getText(), ##, false);            
      }
   ;
   
/**      
 * Matches the frame related function calls that can take a frame name as
 * a parameter. The following are supported:
 * <p>
 * <pre>
 * Function      Type
 * ------------  -------------
 * FRAME-COL     FUNC_DEC
 * FRAME-DOWN    FUNC_INT
 * FRAME-LINE    FUNC_INT
 * FRAME-ROW     FUNC_DEC
 * </pre>
 * <p>
 * The root node returned will be a <code>FUNC_INT</code> or 
 * <code>FUNC_DEC</code> and there will be one child node which is a frame
 * reference (see {@link #lvalue}) OR it is a <code>SYMBOL</code> that is an
 * invalid frame name.  Progress silently tolerates such an invalid frame
 * name and just returns 0 at runtime. This method is needed because of this
 * behavior since an invalid frame reference would not be possible in a
 * regular <code>lvalue</code>.
 * <p>
 * The exit action will save any original token type that rewriting may have
 * erased.  This will allow future processing to access this as an annotation
 * named 'oldtype'.
 * <p>
 * Built-in functions have a language keyword as the function name. This 
 * means that built-ins can support abbreviations.  It also means that some
 * built-in functions cannot be hidden by a user-defined function, if the
 * associated keyword is reserved. If the keyword is not reserved, then a
 * built-in function is hidden in the namespace by a user-defined function
 * of the same name.  This method's token type rewriting is dependent upon
 * this logic being implemented in the <code>SymbolResolver</code> class.
 * See {@link SymbolResolver#lookupFunction}.
 * <p>
 * Note that in the resulting AST, the parenthesis and commas used to define
 * the parameter list are discarded.  In other words, they do not appear in
 * the AST as nodes.  Instead, each parameter (which can be a complete
 * expression subtree or a simple single node like an lvalue) is created
 * as a separate child node of the function type token.  All parameters are 
 * maintained as child nodes in the exact order of the function call itself,
 * so this is essentially the signature of the function.  It it also the
 * proper AST definition to enable simple evaluation of functions with any
 * arbitrary signature.
 */
frame_funcs   
   :
      (
           fc:KW_FR_COL^  { saveAndReplaceType(#fc, FUNC_DEC); }
         | fd:KW_FR_DOWN^ { saveAndReplaceType(#fd, FUNC_INT); }
         | fl:KW_FR_LINE^ { saveAndReplaceType(#fl, FUNC_INT); }
         | fr:KW_FR_ROW^  { saveAndReplaceType(#fr, FUNC_DEC); }
      )
      LPARENS!
      (
           { LA(1) != KW_FRAME }?
           frame:malformed_symbol
           {
              // it is possible to pass anything (even a name that isn't 
              // a frame) so we only convert this if the frame name is valid
              if (sym.lookupFrame(#frame.getText()) != -1)
              {
                 #frame.setType(WID_FRAME);
              }
           }
           
         | lvalue
      )
      RPARENS!
      {
         // write our function-specific annotations
         sym.annotateFunction(##.getText(), ##, false);            
      }
   ;

/**
 * This rule provides generic method and attribute support in a centralized
 * manner.  This is used from {@link #lvalue}.  The result of this rule
 * is either an <code>ATTR_*</code> node or a <code>METH_*</code> node
 * with method parameters (usually an {@link #expr}) as immediate children.
 * <p>
 * All method references are resolved to a value between the range of
 * <code>BEGIN_METH</code> and <code>END_METH</code> types and all attribute
 * references are resolved to a value between the range of 
 * <code>BEGIN_ATTR</code> and <code>END_ATTR</code> types.  A namespace of
 * the method and attribute names and associated types is used to resolve   
 * these values.  <b>This namespace is partial at this time and thus it only
 * represents a small portion of all possible methods and attributes.</b>
 * <p>
 * List of possible resulting token types:
 * <p>
 * <pre>
 *   ATTR_CHAR
 *   ATTR_CLASS
 *   ATTR_COM_HANDLE
 *   ATTR_DATE
 *   ATTR_DATETIME
 *   ATTR_DATETIME_TZ
 *   ATTR_DEC
 *   ATTR_HANDLE
 *   ATTR_INT
 *   ATTR_INT64
 *   ATTR_LOGICAL
 *   ATTR_LONGCHAR
 *   ATTR_MEMPTR
 *   ATTR_RAW
 *   ATTR_RECID
 *   ATTR_ROWID
 *   ATTR_POLY
 *   END_ATTR
 *   BEGIN_METH
 *   METH_CHAR
 *   METH_CLASS
 *   METH_COM_HANDLE
 *   METH_DATE
 *   METH_DATETIME
 *   METH_DATETIME_TZ
 *   METH_DEC
 *   METH_HANDLE
 *   METH_INT
 *   METH_INT64
 *   METH_LOGICAL
 *   METH_LONGCHAR
 *   METH_MEMPTR
 *   METH_RAW
 *   METH_RECID
 *   METH_ROWID
 *   METH_POLY
 *   METH_VOID
 * </pre>
 * <p>
 * This rule uses a {@link #reserved_or_symbol} to create a "pull" for all
 * possible attribute or method names.  In the init action, the 2nd token
 * is inspected to determine if it is an <code>LPARENS</code> or
 * <code>LBRACKET</code> (yes, this is an undocumented "feature").  In either
 * of these cases, this is determined to be a method.  The optional method
 * parameters are then parsed.
 * <p>
 * Please note that this use of reserved keywords is required since
 * attributes and methods can use reserved names.
 * <p>
 * Some parameters are keywords (e.g. <code>NO-LOCK</code>) rather than
 * valid expressions. For this reason, the {@link #reserved_or_symbol} rule
 * is the last-ditch alternative for a parameter.
 * <p>
 * WARNING: starting in Progress v10.x attributes and methods can be chained
 * using the <code>COLON</code> operator. This means that this rule will
 * recursively call itself to implement such support. 
 */
attribute_or_method
   :
      {
         int     oldtype  = -1;
         int     current  = LA(1);
         boolean oldpref = preferWidgets;
      
         // temporarily disable widget preference so the expression will
         // parse without widget refs
         if (oldpref)
         {
            preferWidgets = false;
         }
         
         if (current < BEGIN_ATTR_METH || current > END_ATTR_METH)
         {
            // fixup for stupid/unnecessary Progress ambiguity
            if (current == KW_WIDTH)
            {
               current = KW_WIDTH_C;
            }
               
            // do a lookup, it better find something!
            int newtype = sym.lookupAttributeOrMethod(current);
            
            // save off old type
            oldtype = current;
            
            // handle a fixup due to ambiguity in Progress (it is also an               
            // ATTR_HANDLE and that is how it is returned from the lookup)
            if (current == KW_BUF_FLD)
            {
               if (LA(2) == LPARENS || LA(2) == LBRACKET)
               {
                  newtype = METH_HANDLE;
               }
            }
            
            // safety first
            if ( newtype != -1 )
            {
               // set the type
               LT(1).setType( newtype );
            }
            else
            {
               // unknown method or attribute!
               throw new NoViableAltException(LT(1), getFilename());
            }
         }
      }
      (
         // the reserved_or_symbol rule reference is for pull only, the
         // init action rewrites the token type so this never will be
         // called at runtime
         (
              ATTR_CHAR^
            | ATTR_CLASS^
            | ATTR_COM_HANDLE^
            | ATTR_DATE^
            | ATTR_DATETIME^
            | ATTR_DATETIME_TZ^
            | ATTR_DEC^
            | ATTR_HANDLE^
            | ATTR_INT^
            | ATTR_INT64^
            | ATTR_LOGICAL^
            | ATTR_LONGCHAR^
            | ATTR_RAW^
            | ATTR_RECID^
            | ATTR_ROWID^
            | ATTR_POLY^
            | METH_CHAR^
            | METH_CLASS^
            | METH_COM_HANDLE^
            | METH_DATE^
            | METH_DATETIME^
            | METH_DATETIME_TZ^
            | METH_DEC^
            | METH_HANDLE^
            | METH_INT^
            | METH_INT64^
            | METH_LOGICAL^
            | METH_LONGCHAR^
            | METH_RAW^
            | METH_RECID^
            | METH_ROWID^
            | METH_POLY^
            | METH_VOID^
            | reserved_or_symbol
         )
         
         (
            options { generateAmbigWarnings = false; }
            :
            (LPARENS! | LBRACKET!)
            (
               (
                  options { generateAmbigWarnings = false; }
                  :
                    expr
                  | reserved_or_symbol
               )
               (
                  COMMA!
                  (
                     options { generateAmbigWarnings = false; }
                     :
                       expr
                     | reserved_or_symbol
                  )
               )*
            )?
            (RPARENS! | RBRACKET!)
         )?
      )
      {
         // reset back to state at entrance to this method
         if (oldpref)
         {
            preferWidgets = true;
         }
         
         int ttype = ##.getType();
         
         // store the fully qualified class name for atts/meths that return
         // an object reference
         if (ttype == ATTR_CLASS || ttype == METH_CLASS)
         {
            String cls = sym.lookupAttributeOrMethodClass(current);
            
            if (cls == null)
            {
               String err = String.format("Missing class name for %s!",
                                          ##.getDescriptiveTokenText());
               throw genExc(err, ##);
            }
            
            sym.annotateClassRef(cls, ##);
         }
         
         // save off original type from before token rewriting
         if (oldtype != -1)
            ##.putAnnotation("oldtype", new Long(oldtype));
      }
   ;   
   
/**
 * Matches COM method and property references, their optional parenthesized
 * parameters list. Uses recursion to support chaining.  The result will be
 * rooted at a <code>COM_METHOD</code> or <code>COM_PROPERTY</code> node.
 * Any parameters will be direct children using the {@link #com_parameter},
 * the parenthesis and commas will be dropped.  This rule rewrites the token
 * type of the first token, as it will be a symbol or reserved keyword on
 * entry. That first token is the method or property name and the original
 * token type will be saved in the <code>oldtype</code> annotation.
 * <p>
 * The {@link #reserved_or_symbol} rule is used to "pull" processing down
 * this path but it never will be matched due to the token rewriting. If
 * a 2nd token of <code>LPARENS</code> is seen, then the result will be
 * a <code>COM_METHOD</code>, even if this is really an indexed property
 * which uses a parenthesized list of integer expressions as an index, one
 * per array dimension.  This is a flaw which cannot be resolved without
 * information not available in the source code.  So those indexed properties
 * will have to be disambiguated from methods by some later process.
 * <p>
 * There are COM property and method names that are documented in the
 * language reference. These have some special meaning or importance and
 * thus they are given better support than the unknown interfaces in the
 * controls that are used (such interfaces are only known on systems in
 * which those controls are registered). In our case one can find the
 * following token types set into the <code>oldtype</code> annotation of
 * a <code>COM_PROPERTY</code> or <code>COM_METHOD</code>. The following
 * are supported:
 * <p>
 * <pre>
 * Token Type        Name                        Type
 * ---------------   --------------------------  --------------------
 * KW_CNTRL_NM       Control-Name                Property
 * KW_CONTROLS       Controls                    Property
 * KW_HEIGHT_C       Height                      Property
 * KW_HONOR_PK       HonorProKeys                Property
 * KW_HONOR_RK       HonorReturnKey              Property
 * KW_LEFT           Left                        Property
 * KW_LOADCTRL       LoadContols                 Method
 * KW_NAME           Name                        Property
 * KW_TAG            Tag                         Property
 * KW_TOP            Top                         Property
 * KW_WID_HAND       Widget-Handle               Property
 * KW_WIDTH          Width                       Property
 * </pre>
 */
com_property_or_method
   :
      {
         int oldtype  = -1;
         int current  = LA(1);
         
         if (current != COM_METHOD && current != COM_PROPERTY)
         {
            int newtype = (LA(2) == LPARENS) ? COM_METHOD : COM_PROPERTY;
            
            // save off old type
            oldtype = current;
            
            // set the type
            LT(1).setType(newtype);
         }
      }
      (
         // the reserved_or_symbol rule reference is for pull only, the
         // init action rewrites the token type so this never will be
         // called at runtime
         (
              COM_METHOD^
            | COM_PROPERTY^
            | reserved_or_symbol
         )
         
         (
            options { generateAmbigWarnings = false; }
            :
            LPARENS!
            (
               options { generateAmbigWarnings = false; }
               :
               com_parameter (COMMA! com_parameter)*
            )? 
            RPARENS!
         )?
      )
      {
         // save off original type from before token rewriting
         if (oldtype != -1)
            ##.putAnnotation("oldtype", new Long(oldtype));
      }
   ;
   
/**
 * Matches a single COM parameter which is an {@link #expr} and which may 
 * have optional modifiers for overriding data types (see 
 * {@link #com_as_clause}) or for specifying passing parameters by pointer
 * (<code>BY-POINTER</code> or <code>BY-VARIANT-POINTER</code>).
 * <p>
 * The tree will always be rooted by <code>COM_PARAMETER</code>. Even if
 * there is no parameter expression, this rule will return a single node
 * of this type so that the null parameter construct of COM can be properly
 * supported.  In other words, the expression is optional.
 * <p>
 * <code>KW_IN_OUT</code> and <code>KW_OUTPUT</code> are matched as
 * optionally prefacing this construct.
 * <p>
 * Called by {@link #com_property_or_method}.
 */   
com_parameter
   :
      {
         astFactory.makeASTRoot(currentAST, #[COM_PARAMETER,""]);
      }
      (
         (KW_IN_OUT | KW_OUTPUT)?
         expr (com_as_clause)? (KW_BY_PTR | KW_BY_VAR_P)?
      )?
   ;
   
/**
 * Matches the construct following a COM parameter which overrides the
 * default Progress type conversion with a more specific COM type conversion.
 * The first token must be a <code>KW_AS</code> and this will root the
 * sub-tree.
 * <p>
 * The following types are known to be possible (and the normal ones in
 * {@link #as_clause} are known to be impossible to use here):
 * <p>
 * <pre>
 * CURRENCY
 * FLOAT
 * ERROR-CODE
 * IUNKNOWN
 * SHORT
 * UNSIGNED-BYTE
 * </pre>
 * <p>
 * Called by {@link #com_parameter}.
 */
com_as_clause
   :
      KW_AS^ 
      (
           KW_CURRENCY
         | KW_FLOAT
         | KW_ERR_CODE
         | KW_IUNKNOWN
         | KW_SHORT 
         | KW_UNS_BYTE
      )
   ;
   
/**   
 * Matches an optional widget qualifier using the keyword <code>IN</code>.
 *  Calls the {@link #frame_reference}, {@link #lvalue} (for browse) and
 * {@link #menu_reference} rules.  Roots the tree at the <code>IN</code>.
 * <p>
 * Used by <code>lvalue</code> and the recursive call to <code>lvalue</code>
 * is protected by a semantic predicate keyed off of <code>KW_BROWSE</code>.
 * <p>
 * Uses a flag to allow a match for the <code>IN prochandle</code> using
 * {@link #expr} for matching when called from rules which can ambiguously
 * match both this clause and a following expression.
 */
widget_qualifier
   :
      KW_IN^ 
      (
         options { generateAmbigWarnings = false; }
         :
           frame_reference[false]                       
         | menu_reference
         
           // used to be browse_reference but that rule was removed and this
           // was consolidated as a recursive call to lvalue
         | { LA(1) == KW_BROWSE }? lvalue
      )
   ;
   
/**   
 * Matches Progress 4GL built-in function names that are <b>reserved</b> 
 * keywords. This only includes functions that take parameters. Built-in
 * functions that take no parameters are treated like variables.  See
 * {@link #reserved_variables}.
 * <p>
 * This is generally only used to generate the proper lookahead predictions
 * in the top level parser methods, such that calling rules would see tokens
 * of this type.  It is expected that the calling method would rewrite the
 * token type of such tokens such that this alternative is never actually
 * entered at runtime.  Thus this method really just provides a 'pull' into
 * the calling method.  See {@link #func_call}.
 * <p> 
 * Note that there are some built-in functions that can act like variables
 * OR as a normal parenthesized function.  In other words, for the following
 * list, the parenthesis are optional:
 * <p>
 * <ul>
 *    <li>ETIME
 *    <li>FRAME-COL
 *    <li>FRAME-DOWN
 *    <li>FRAME-LINE
 *    <li>FRAME-ROW
 *    <li>LINE-COUNTER
 *    <li>PAGE-NUMBER
 *    <li>USERID
 * </ul>
 * <p>
 * Such keywords are included in both this rule and in
 * <code>reserved_variables</code>. 
 */
reserved_functions
   :
      (
           KW_U_MSG
         | KW_ALIAS
         | KW_ASC
         | KW_CAN_DO
         | KW_CAST
         | KW_CHR
         | KW_COUNT_OF
         | KW_CONN_ED                 
         | KW_DBCP
         | KW_DBCOLL
         | KW_DBPARAM
         | KW_DBREST
         | KW_DBTASKID
         | KW_DBTYPE
         | KW_DBVERS
         | KW_DYN_FUNC
         | KW_DYN_INVK
         | KW_ENCODE
         | KW_ENTRY
         | KW_ETIME
         | KW_FILL
         | KW_FIRST
         | KW_FIRST_OF
         | KW_FR_COL
         | KW_FR_DOWN
         | KW_FR_LINE
         | KW_FR_ROW
         | KW_GET_BYTE
         | KW_GET_COLL
         | KW_INDEX
         | KW_IS_LEAD
         | KW_KBLABEL
         | KW_KEYCODE
         | KW_KEYFUNC
         | KW_KEYLAB
         | KW_KW
         | KW_LAST
         | KW_LAST_OF
         | KW_LDBNAME
         | KW_LIB
         | KW_LINE_CNT
         | KW_LOOKUP
         | KW_MEMBER
         | KW_NUM_ENT
         | KW_PAGE_NUM
         | KW_PDBNAME
         | KW_PROGNAME
         | KW_QRY_OFF
         | KW_R_INDEX
         | KW_SDBNAME
         | KW_SEARCH
         | KW_SETUSER
         | KW_TO_ROWID
         | KW_TRIM
         | KW_USERID
      )
   ;
   
/**      
 * Matches all Progress 4GL variable, field, stream, buffer or widget names
 * that exist in the respective namespaces, including all forms of built-in
 * variables such as system handles or 'functions' that take no parameters 
 * and essentially behave as variables.
 * <p>
 * The actual lvalue portion of this rule is defined as a single token that
 * matches the union of the set of all token types for variables, fields,
 * widgets, buffers or <code>SYS_HANDLE or STREAM</code> (this means that 
 * there is one token type for each possible data type such as 
 * <code>VAR_INT</code> or <code>FIELD_LOGICAL</code> or
 * <code>WID_FILL_IN</code>) and the set of all possible alternatives for 
 * symbol names (a reference to the {@link #symbol} rule) or the special
 * <code>DB_SYMBOL</code>.  <code>DB_SYMBOL</code> is the type assigned to
 * all partially or fully qualified database schema names.  It is created by
 * the {@link ProgressLexer#mSYMBOL} method and it may refer to either a
 * table (in dbname.tablename format) or a field (in dbname.tablename.field
 * format or in tablename.field format).  In this rule, we only expect to
 * ever be processing a field name (table names are right out).
 * <p>
 * Due to widget, buffer and query references (especially when attributes 
 * and methods are used) it is possible to see certain prefixing keywords 
 * (some reserved and some not) in front of the lvalue itself.  In these
 * cases, the lvalue will not be found in one of the standard lookup
 * processing, but instead the keyword will force the token type to be set
 * accordingly (e.g. <code>BUFFER record</code>).  The list of such keywords:
 * <p>
 * <pre>
 *    BROWSE
 *    BUFFER
 *    DATASET
 *    DATA-SOURCE
 *    DATA-RELATION
 *    FRAME
 *    MENU
 *    SUB-MENU
 *    MENU-ITEM
 *    QUERY
 *    TEMP-TABLE
 *    STREAM
 *    STREAM-HANDLE
 *    TABLE-HANDLE
 *    DATASET-HANDLE
 *    TABLE
 * </pre>
 * <p>
 * It is important to note that a {@link #record} can be matched as a
 * BUFFER or TEMP-TABLE.
 * <p>
 * The operation of this processing is VERY similar if not identical to
 * some other rules (see {@link #frame_reference} and
 * {@link #query_reference}).  Keeping these other rules allows more
 * specific usage to be implemented without calling this rule.  This can 
 * reduce ambiguity.
 * <p>
 * Note that both <code>lvalue</code> and <code>func_call</code> refer to a
 * leftmost token that includes the <code>symbol</code> rule.  This creates
 * an ambiguity that is only resolved by lookahead <b>and</b> the proper
 * ordering of the rules (since both are equal alternatives in the
 * <code>primary_expr</code> rule).
 * <p>
 * It is required that where this rule is referenced as one alternative that
 * conflicts with the definitions of other alternatives (see
 * {@link #primary_expr} and {@link #func_call}), <b>the order of the 
 * alternatives must be more specific (in terms of matching criteria) to less
 * specific</b>.  Since the <code>func_call</code> rule contains the very
 * specific <code>symbol</code> followed by <code>LPARENS</code> pattern,
 * it is more specific than a simple <code>symbol</code> by itself (which is
 * one definition of an <code>lvalue</code>.  For this reason, 
 * <code>func_call</code> must always be placed ahead of <code>lvalue</code>
 * in alternative lists.
 * <p>
 * The inclusion of the <code>symbol</code> rule ensures that all valid
 * symbols are considered an lvalue and will be matched by this rule.  This
 * means that the top level expression processing will include a symbol as
 * one of the options for its leftmost token, thus &quot;pulling&quot; the
 * recursive descent processing down the <code>expr</code> rule.
 * <p>
 * It is important to note that in the end, it is not really valid to have
 * a token type of <code>SYMBOL</code> or <code>DB_SYMBOL</code> since this 
 * could be anything.  What we really require is a token that is one of the
 * valid variable, field, stream, widget or system handle tokens:
 * <pre>
 *   VAR_CHAR
 *   VAR_CLASS
 *   VAR_COM_HANDLE
 *   VAR_DATE
 *   VAR_DATETIME
 *   VAR_DATETIME_TZ
 *   VAR_DEC
 *   VAR_HANDLE
 *   VAR_INT
 *   VAR_INT64
 *   VAR_LOGICAL
 *   VAR_LONGCHAR
 *   VAR_MEMPTR
 *   VAR_RAW
 *   VAR_RECID
 *   VAR_ROWID
 *   SYS_HANDLE
 *   CLASS_NAME (static OO 4GL class reference)
 *   VAR_BYTE
 *   VAR_DOUBLE
 *   VAR_FLOAT
 *   VAR_LONG
 *   VAR_SHORT
 *   VAR_USHORT
 *   STREAM
 *   FIELD_BLOB
 *   FIELD_CHAR
 *   FIELD_CLASS
 *   FIELD_CLOB
 *   FIELD_COM_HANDLE
 *   FIELD_DATE
 *   FIELD_DATETIME
 *   FIELD_DATETIME_TZ
 *   FIELD_DEC
 *   FIELD_INT
 *   FIELD_INT64
 *   FIELD_HANDLE
 *   FIELD_LOGICAL
 *   FIELD_RAW
 *   FIELD_RECID
 *   FIELD_ROWID
 *   WID_BROWSE
 *   WID_BUTTON
 *   WID_COMBO
 *   WID_DIALOG
 *   WID_EDITOR
 *   WID_FILL_IN
 *   WID_FRAME
 *   WID_LITERAL
 *   WID_MENU
 *   WID_MENU_ITM
 *   WID_RADIO
 *   WID_RECT
 *   WID_SEL_LST
 *   WID_SLIDER
 *   WID_SUB_MENU
 *   WID_TEXT
 *   WID_TOGGLE
 *   WID_WINDOW
 *   BUFFER
 *   TABLE
 *   TEMP_TABLE
 *   WORK_TABLE
 * </pre>
 * <p>
 * The difference between a <code>SYMBOL</code> (or <code>DB_SYMBOL</code>)
 * and a <code>VAR_INT</code> can only be determined by context (the fact
 * that the lookahead code has identified a match with a <code>SYMBOL</code> 
 * or <code>DB_SYMBOL</code> token which is a construct that can occur in an
 * lvalue call.  Thus <b>if</b> a match is found using lookahead (in the rule
 * that calls this one), then it <b>must</b> be correct and necessary to 
 * convert the <code>SYMBOL</code> or <code>DB_SYMBOL</code> token type into 
 * the correct variable or field type, using a lookup into the variable and
 * field namespaces. This is done in an init action that is processed at 
 * the beginning of this rule. In this way, the rest of the rule only sees
 * the first token as a valid variable or field token type and never tries to
 * call the <code>symbol</code> rule since calling that rule would actually 
 * try to match (and consume from the token stream) a <code>SYMBOL</code>
 * token type.  <b>Specifying a rule reference to obtain its value from a 
 * lookahead perspective but then changing the token's type such that the
 * rule reference never gets called, is the critical trick that enables
 * context sensitive symbol names!</b>
 * <p>
 * The exit action saves the fully qualified schema name that was matched
 * in the schema dictionary. This allows downstream processing to access this
 * as an annotation named 'schemaname'.  In addition, the exit action stores
 * the backing construct for this field (this will be the token type TABLE,
 * TEMP_TABLE or WORK_TABLE). This annotation is called 'recordtype'.
 * <p>
 * The exit action will save any original token type that rewriting may have
 * erased.  This will allow future processing to access this as an annotation
 * named 'oldtype'.
 * <p>
 * This rule supports an undocumented qualified schema naming quirk in which
 * there is whitespace between the record and the .field portion of a schema
 * name.  In that case Progress matches "record    .field" as the single
 * qualified field name "record.field".  This quirk is matched when a record
 * is followed by whitespace, a <code>DOT</code> and a token for which the
 * combined text resolves as a field.  In that case, the 1st two tokens are
 * discarded and their text is concatenated into the 3rd token as if it had
 * been tokenized properly originally. The lexer can't solve this problem as
 * the syntax is inherently ambiguous with many other constructs. So instead
 * we use the knowledge and context of the parser to reassemble the tokens 
 * when needed.  WARNING: this code requires that the parser's lookahead
 * depth be at least 3 (k=3).  It is interesting to note that no whitespace
 * can ever fall between the dot and the field name.  It can only occur
 * before the dot.
 * <p>
 * This class has a flag to prefer widgets which allows callers of this rule
 * to bias it towards matching widgets in preference to its default policy
 * of matching a variable.  The vast majority of locations where this rule
 * is called do not want a match with a widget and so the default is to match
 * variables first.
 * <p>
 * Built-in variables have a language keyword as the variable name. This 
 * means that built-ins can support abbreviations.  It also means that some
 * built-in variables cannot be hidden by a user-defined variable, if the
 * associated keyword is reserved. If the keyword is not reserved, then a
 * built-in variable is hidden in the namespace by a user-defined variable
 * of the same name.  This method's token type rewriting is dependent upon
 * this logic being implemented in the <code>SymbolResolver</code> class.
 * See {@link SymbolResolver#lookupVariable}.
 * <p>
 * To ensure this method sees tokens matching the list of variables with
 * reserved names, the {@link #reserved_variables} method is included as
 * an alternative.  This is the same 'pull' trick that is being used for
 * generic symbols above.  The reason that the generic symbol technique is
 * not enough to pull reserved variables is that only unreserved keywords
 * are pulled by the <code>symbol</code> method.  This can't be changed
 * without negative impacts on other code, since <code>symbol</code> is 
 * used in many locations in the parser.  Instead a custom method was
 * created to generate pull.
 * <p>
 * Note that Progress 4GL built-in functions that take no parameters are
 * actually implemented in this parser as built-in variables.  This
 * accurately matches the syntax and semantics of how Progress is implemented
 * even if the name itself differs.
 * <p>
 * This rule also handles all language statements that act like assignable
 * variables and/or look like functions that can be assigned.  The first
 * type are simple enough that they are simply handled as global variables.
 * The second kind are handled by the {@link #assign_type_syntax_stmt_list}
 * which is called from the main body of this rule.
 * <p>
 * Since Progress 4GL variable names take precedence (i.e. hide) unqualified
 * field names that conflict, the token rewriting code in the method does
 * the variable namespace lookup first and if a result is found, this is
 * used.  This effectively hides the matching field name since the schema
 * dictionary is never consulted.  If no match is found to a variable name,
 * the schema dictionary is used to result the name.
 * <p>
 * Variables and fields in Progress can be defined as arrays using the
 * <code>EXTENT</code> keyword (see {@link #def_var_stmt}). The 
 * <code>lvalue</code> rule includes an optional postfix reference to the 
 * rule that implements the subscripting operator [ ].  Only one level of
 * subscripting is possible because there is no support for multi-dimensional
 * arrays in Progress 4GL.  See {@link #subscript} for details.
 * <p>
 * The use of the <code>subscript</code> rule is detected as ambiguous by
 * ANTLR, but this is bogus.  Warnings have been disabled for this
 * sub-rule.
 * <p>
 * An immediate following <code>COLON</code> which is not itself followed
 * by whitespace, will trigger the matching of Progress methods and
 * attributes (see {@link #attribute_or_method} or if the referent (to the
 * left of the <code>COLON</code> is a <code>VAR_COM_HANDLE</code>, a
 * <code>FIELD_COM_HANDLE</code> or a <code>SYS_HANDLE</code> with an
 * <code>oldtype</code> annotation of <code>KW_COM_SELF</code>, then the
 * following processing will match COM (Active-X) methods and properties
 * via the {@link #com_property_or_method} rule. Progress methods and
 * attributes will be rooted at the <code>COLON</code> node and COM methods
 * and properties are rooted at that same node except the type will be
 * rewritted to be <code>COM_INVOCATION</code>.
 * <p>
 * An immediate following <code>DB_REF_NON_STATIC</code> token (the
 * <code>::</code> "operator" is supported here.  This is the syntax added in
 * 10.2x which allows a database reference in the form handle::table_name or
 * handle::table_name::field_name.  These are non-static database names directly
 * referenced off a handle using the :: operator. 
 * <br><strong>Note</strong>At this moment only same-type of chaining can be parsed
 * correctly (ie: only <code>:</code> (COLON) or only <code>::</code> DB_REF_NON_STATIC).
 * The mixing of both punctuation elements (which is valid in the 4GL) is not supported yet.
 
 * <p>
 * Widget references can have an optional qualifier that is matched by the
 * {@link #widget_qualifier} rule.
 */
lvalue
   :
      {
         int    oldtype = -1;
         String clsName = null;
                              
         // it is possible to have an optional prefix for widgets, queries and other resources
         // this "qualifier" changes the downstream processing
         boolean qualifier = false;
         int     qualType  = -1;
         
         switch (LA(1))
         {
            case KW_FRAME:
               // this is a reserved keyword, so we cheat here because we
               // know it MUST be followed by a FRAME (we could look this
               // up in the frame namespace, but why?)
               qualType  = WID_FRAME;
               break;
            case KW_QUERY:
               // this is a reserved keyword, so we cheat here because we
               // know it MUST be followed by a QUERY
               qualType  = QUERY;
               
               // TODO: need to check if explicit static/instance qualifiers can be used here
               // TODO: need to check implicit static/instance (based on the surrounding code)
               if (sym.lookupQuery(LT(2).getText()) != QUERY)
               {
                  CommonToken tok = (CommonToken) LT(2);
                  System.out.println(
                     String.format("Warning: lookup of query '%s' failed at line %d column %d", 
                                   tok.getText(),
                                   tok.getLine(),
                                   tok.getColumn()));
               }
               break;
            case KW_STREAM:
               // this is a reserved keyword, so we cheat here because we
               // know it MUST be followed by a STREAM 
               qualType = STREAM;
               break;
            case KW_BROWSE:
               // not reserved, we have to look it up (it shares a namespace
               // with variable but we have split the namespace to allow
               // the same symbol to refer to both a WID_ and VAR_ value)
               // with browse, we only need to check the widget namespace
               qualType = sym.lookupWidget(LT(2).getText()); 
               break;
            case KW_MENU: 
            case KW_SUB_MENU:
               // not reserved, we have to look it up (it shares a namespace
               // with frame)
               qualType = sym.lookupMenu(LT(2).getText());
               break;
            case KW_MENU_ITM:
               // not reserved, we have to look it up
               qualType = sym.lookupMenuItem(LT(2).getText());
               break;
            case KW_TAB_HAND:
            case KW_DSET_HND:
            case KW_STRM_HND:
               // table-handle is not reserved, dataset-handle and stream-handle are reserved
               // but we will recursively call ourselves below to match the required following
               // handle (field, var...) so just force qualType to be a valid token type and
               // that will properly bypass the other init action logic below; this can't just
               // lookup vars because it can reference a field
               qualType = LA(1);
               break;
            case KW_BUFFER:
            case KW_TABLE:
            case KW_TEMP_TAB:
               // buffer and temp-table are not reserved, we have to look it
               // up in those cases, we'll just do it for table too; this is
               // only used as a check to prove that there is a following 
               // record reference, the actual value is only used to bypass
               // the following init action logic below
               qualType = sym.lookupTable(LT(2).getText());
               break;
            case KW_DATASET:
               // TODO: need to check if explicit static/instance qualifiers can be used here
               // TODO: need to check implicit static/instance (based on the surrounding code)
               // it is reserved but we look it up anyway
               qualType = sym.lookupDataSet(LT(2).getText());
               break;
            case KW_DATA_SRC:
               // TODO: need to check if explicit static/instance qualifiers can be used here
               // TODO: need to check implicit static/instance (based on the surrounding code)
               // it is reserved but we look it up anyway
               qualType = sym.lookupDataSource(LT(2).getText());
               break;
            case KW_DATA_REL:
               // not reserved, we have to look it up
               qualType = sym.lookupDataRelation(LT(2).getText());
               break;
         }
         
         // if we match one of the optional prefixes AND the next token is
         // found in the associated namespace, then we know that we have
         // one of the special widget/query qualifier forms --> change all
         // downstream processing to handle this!
         if (qualType != -1)
         {
            qualifier = true;         
         }
         else if (((LA(1) == KW_CUR_VAL  || LA(1) == KW_DYN_CURV ||
                    LA(1) == KW_ENTRY    || LA(1) == KW_EXTENT   ||
                    LA(1) == KW_FIX_CP   || LA(1) == KW_LENGTH   ||
                    LA(1) == KW_OVERLAY  || LA(1) == KW_PUT_BITS ||
                    LA(1) == KW_PUT_BYTE || LA(1) == KW_PUT_BYTS ||
                    LA(1) == KW_PUT_DBL  || LA(1) == KW_PUT_FLT  ||
                    LA(1) == KW_PUT_I64  || LA(1) == KW_PUT_LONG || 
                    LA(1) == KW_PUT_SHT  || LA(1) == KW_PUT_STR  || 
                    LA(1) == KW_PUT_UL   || LA(1) == KW_PUT_USHT ||
                    LA(1) == KW_RAW      || LA(1) == KW_SET_B_OR ||
                    LA(1) == KW_SET_PTR  || LA(1) == KW_SET_SZ   ||
                    LA(1) == KW_SUBSTR)  && LA(2) == LPARENS) ||
                  LA(1) == KW_DYN_NEW)
         {
            // nothing to do here, just avoid the else portion of this block
            // which would try to rewrite our leftmost token
            
            // these are the assignment-style language statements
         }
         else if (LA(1) == KW_THIS_OBJ || LA(1) == KW_SUPER)
         {
            oldtype = LA(1);        
            LT(1).setType(VAR_CLASS);
         
            if (oldtype == KW_THIS_OBJ)
            {
               // special this-object reference
               clsName = sym.getCurrentClassName();
            }
            else
            {
               // special super object reference
               clsName = sym.getCurrentParentName();
            }
         }
         else
         {         
            // this is a normal lvalue! 
            
            // this works because the VARTYPES, FIELDTYPES and WIDGETS
            // sections are kept in a contiguous "super" block called LVALUE
            if ( LA(1) < BEGIN_LVALUE || LA(1) > END_LVALUE )
            {
               String text = null;
            
               // check variables, widgets, class members and fields
               int newtype = resolveLvalueCoreType(1, true);
               
               // lookup a frame if the field reference wasn't found
               if (newtype == -1)
               {
                  text = LT(1).getText();
               
                  // WARNING: this matches with a frame name that does NOT
                  // have the preceding KW_FRAME qualifying keyword, since
                  // there is a quirk with frame names (they can start with
                  // any combination of NUM_LITERAL, MINUS and UNKNOWN_TOKEN
                  // characters in front of the real symbol and for all other
                  // frame references we use the malformed_symbol rule to 
                  // re-assemble the parts; this code will not work if
                  // such a frame name is used here
               
                  // this may be a frame name
                  newtype = sym.lookupFrame(text);
               }
               
               // last chance! it may be a stream
               if (newtype == -1)
               {
                  // this may be a stream name
                  newtype = sym.lookupStream(text);
               }

               // last last chance: it may be a menu/sub-menu reference.
               if (newtype == -1)
               {
                  // this may be a menu name
                  newtype = sym.lookupMenu(text);
               }
               
               // at this point, there should always be a match in valid
               // Progress 4GL source so we throw an exception if this is
               // not the case
               if (newtype != -1)
               {
                  oldtype = LA(1);  
                  LT(1).setType(newtype);
               }
               else
               {
                  // trigger this exception unless we are being processed
                  // for expression evaluation or are allowed to match a
                  // symbol, in which case this search is expected to fail
                  // gracefully
                  if (!evaluatingExpression && !allowSymbolMatch)
                  {
                     throw new NoViableAltException(LT(1), getFilename());
                  }
               }
            }
         }
      }
      (
         options { generateAmbigWarnings = false; }
         :
           // widget, buffer, temp-table or query qualifier path
           { qualifier }?
           (
                (
                     KW_QUERY^
                   | KW_BROWSE^ 
                   | KW_MENU^
                   | KW_SUB_MENU^ 
                   | KW_MENU_ITM^
                   | KW_DATASET^
                   | KW_DATA_SRC^
                   | KW_DATA_REL^
                ) 
                w:symbol                       { #w.setType(qualType); }
                
              | KW_STREAM^ strm:malformed_symbol  { #strm.setType(qualType); }
              | KW_FRAME^ fr:malformed_symbol     { #fr.setType(qualType); }
              | (KW_BUFFER^ | KW_TABLE^ | KW_TEMP_TAB^) 
                record[true, false, false]
                (
                   options { generateAmbigWarnings = false; }
                   :
                   KW_APPEND
                )?
              
                // this is recursive to parse the handle var/field that is required to follow
              | (KW_TAB_HAND^ | KW_DSET_HND^ | KW_STRM_HND^) lvalue
           )
           
           // handle assignment-type language statements (except the 4 simple
           // ones that are treated like vars: CURRENT-LANGUAGE, FRAME-VALUE,
           // PROMSGS and PROPATH)
         | assign_type_syntax_stmt_list
              
           // normal lvalue path
         | (                
                VAR_CHAR^        
              | VAR_CLASS^
              | VAR_COM_HANDLE^
              | VAR_DATE^
              | VAR_DATETIME^
              | VAR_DATETIME_TZ^ 
              | VAR_DEC^
              | VAR_HANDLE^
              | VAR_INT^
              | VAR_INT64^
              | VAR_LOGICAL^
              | VAR_LONGCHAR^
              | VAR_MEMPTR^
              | VAR_RAW^
              | VAR_RECID^   
              | VAR_ROWID^
              
              // native interface types
              | VAR_BYTE^      | VAR_DOUBLE^  | VAR_FLOAT^   | VAR_LONG^
              | VAR_SHORT^     | VAR_USHORT^
              
              | SYS_HANDLE^
              
              | STREAM^
              
              | FIELD_BLOB^
              | FIELD_CHAR^
              | FIELD_CLASS^
              | FIELD_CLOB^
              | FIELD_COM_HANDLE^
              | FIELD_DATE^
              | FIELD_DATETIME^
              | FIELD_DATETIME_TZ^
              | FIELD_DEC^
              | FIELD_HANDLE^
              | FIELD_INT^
              | FIELD_INT64^ 
              | FIELD_LOGICAL^
              | FIELD_RAW^
              | FIELD_RECID^
              | FIELD_ROWID^
              
              | WID_BROWSE^  | WID_BUTTON^   | WID_COMBO^    | WID_DIALOG^
              | WID_EDITOR^  | WID_FILL_IN^  | WID_FRAME^    | WID_IMAGE^    
              | WID_LITERAL^ | WID_MENU^     | WID_MENU_ITM^ | WID_RADIO^
              | WID_RECT^    | WID_SEL_LST^  | WID_SLIDER^   | WID_SUB_MENU^
              | WID_TEXT^    | WID_TOGGLE^   | WID_WINDOW^
                            
                // these are here to match the global variables that are 
                // really "functions" but have no parenthesis... we set the
                // token type to the FUNC_ version for consistency
              | FUNC_CHAR^        | FUNC_CLASS^       | FUNC_COM_HANDLE^
              | FUNC_DATE^        | FUNC_DATETIME^    | FUNC_DATETIME_TZ^
              | FUNC_DEC^         | FUNC_HANDLE^      | FUNC_INT^
              | FUNC_INT64^       | FUNC_LOGICAL^     | FUNC_LONGCHAR^
              | FUNC_MEMPTR^      | FUNC_RAW^         | FUNC_RECID^
              | FUNC_ROWID^       | FUNC_POLY^ 
                            
                // this token reference represents all partially or fully
                // qualified schema names, its token type will be rewritten
                // on entry to this method so this reference is here to "pull"
                // the top level prediction logic into this method --> note
                // that this token type is defined outside of the BEGIN_LVALUE
                // and END_LVALUE range which causes it to be rewritten every
                // time
              | DB_SYMBOL^
              
                // the inclusion of the reserved_variables rule reference
                // exists to "pull" us into the lvalue rule, but when we leave
                // the type is always reset one of the valid datatypes OR we
                // throw an exception
              | reserved_variables
              
                // some symbol matching cases will be provided from the lexer
                // as a filename, if our special expression processing mode or
                // symbol-matching mode is active, we allow this
              | { evaluatingExpression || allowSymbolMatch }?
                FILENAME
              
                // the inclusion of the symbol rule reference exists to "pull"
                // us into the lvalue rule, but when we leave the type is
                // always reset one of the valid datatypes OR we throw an
                // exception UNLESS we are allowed to match on a symbol, which
                // is only done as a special mode for expression processing
                // enabled by a specific parser flag; note that the optional
                // "DOT symbol" is there to pull in the database field refs
                // that have whitespace (see isQualifiedFieldNameQuirk())
              | symbol { evaluatingExpression || allowSymbolMatch }?
                (
                   options { generateAmbigWarnings = false; }
                   :
                   DOT symbol
                )?
           )
           (
              options { generateAmbigWarnings = false; }
              :
              subscript
           )?           
      )
      (                                                                
         options { generateAmbigWarnings = false; }
         :
         // avoid matching the IN WIDGET-POOL expr construct as this rule matches on the
         // KW_IN clause, we need to disambiguate between a KW_IN used in a widget qualifier
         // context and a KW_IN used with a IN proc-handle clause; thus, this rule will be 
         // called only if we are in a widget-qualifier case
         { LA(2) != KW_WID_POOL && LA(2) != KW_WINDOW && 
           (LA(2) == KW_FRAME || LA(2) == KW_MENU || LA(2) == KW_SUB_MENU || LA(2) == KW_BROWSE) }?
         widget_qualifier
      )?      
      {
         // assume the AST node to annotate is the root node (we need this
         // to be the variable reference itself)
         Aast core  = ##;
         int  ctype = core.getType();
         
         // browse references will be rooted at a different node
         if (ctype == KW_BROWSE)
         {
            // walk down to get the variable reference as the AST node
            core = (Aast) core.getFirstChild();
         }
         
         int    ftype = core.getType();
         String ftext = core.getText();
         
         // after the lvalue has been fully parsed (including subscripts and so forth), it is
         // easy to see the next token (before now it was an arbitrary number of tokens away);
         // if the next token is a COLON w/o whitespace, then we would have wanted to parse the
         // lvalue with preferWidgets set to true; simulate this now
         if (ftype >= BEGIN_VARTYPES &&
             ftype <= END_VARTYPES   &&
             LA(1) == COLON          &&
             !followedByWhitespace(LT(1)))
         {
            int override = sym.lookupWidget(ftext);
            
            if (override != -1)
            {
               ftype = override;
               core.setType(override);
            }
         }
                  
         // save any original token type that rewriting may have erased
         if (oldtype != -1)
            core.putAnnotation("oldtype", new Long(oldtype));
            
         // fixup deferred qualified class name
         // TODO: this doesn't account for all cases (the use of lookupDataMember() above
         //       does not set this var; this means that these variable references have
         //       no refid, nor class name and cannot be properly used downstream
         if (clsName != null)
            sym.annotateClassRef(clsName, core);
            
         // field post-processing
         if (ftype > BEGIN_FIELDTYPES && ftype < END_FIELDTYPES)
         {
            sym.annotateField(core, null, false, !inLikeClause);
         }
         
         // function post-processing
         if (ftype > BEGIN_FUNCTYPES && ftype < END_FUNCTYPES)
         {
            // write our function-specific annotations
            sym.annotateFunction(ftext, core, false);            
         }            
            
         // variable and widget post-processing
         if ((ftype > BEGIN_VARTYPES && ftype < END_VARTYPES) ||
             (ftype > BEGIN_WIDGETS  && ftype < END_WIDGETS))
         {
            annotateVariable(ftext, ftype, core);
         }
      }
   ; 
   
/**
 * Matches the array subscripting operator [ ] in all of its forms. The most
 * basic version requires a single integer expression (which can be as simple
 * as an integer constant or it can include a full expression including other
 * variables or fields).  Since Progress 4GL allows whitespace to be placed
 * between the symbol name and the left bracket, as well as used arbitrarily
 * inside the brackets, this subscripting support is best implemented in 
 * the parser.  Note that the Progress documentation states that the subscript
 * must be specified only as integer constants, experience proves that 
 * the real requirement is for it to be an integer expression.
 * <p>
 * In Progress it is possible to reference a range of array elements in 
 * some cases (e.g. the <code>DISPLAY</code> language statement) however it
 * is not possible to do this in traditional expressions (i.e. Progress does
 * not have integrated support for vector operations).  To use this range
 * subscript, the initial expression is followed by the <code>FOR</code>
 * keyword and then an integer constant to specify the length of the range. 
 * This rule handles both forms, although there is no checking to confirm
 * that the range form is only used in valid locations.
 * <p>
 * The resulting tree is rooted under the left bracket token and the right
 * bracket is discarded.
 * <p>
 * When called during the processing of {@link #like_clause}, the expression
 * being matched is always converted into the equivalent of a 
 * <code>NUM_LITERAL</code> with a value of 1. Strangely, this is how 
 * Progress does it.  The bizarre thing here is that complete garbage can
 * be inserted in this expression and if it is close enough to something that
 * would be a valid expression it will be allowed.  Here is a list of 
 * subscripts that can work:
 * <p>
 * <pre>
 * def var array  as int extent 10 init [1, 2, 3, 4, 5, 6, 7, 8, 9, 10].
 *
 * def var scalar1 like array[3].
 * def var scalar2 like array[1].
 * def var scalar3 like array[0].            -&gt; normally invalid since 0 is out of bounds!
 * def var scalar4 like array[17].           -&gt; normally invalid since 17 is out of bounds!
 * def var scalar5 like array[trash].        -&gt; normally invalid since trash is not a var!
 * def var scalar6 like array[trash + junk]. -&gt; normally invalid since trash and junk are not vars!
 * def var scalar7 like array[1 + 2].
 * </pre>
 * <p>
 * Additionally, with the advent of .NET object support in 10.2x, the index
 * property concept of .NET is supported.  This means that any kind of
 * expression (not just numeric) can be used to index into .NET object arrays
 * if that .NET class has the proper support.
 * <p>
 * In Progress, all of the above are treated as a <code>like array[1]</code>
 * though this is undocumented and non-intuitive.  Either way, we do the
 * same thing.
 * <p>
 * This rule is only called by the {@link #lvalue} rule.
 */
subscript
   :
      {
         boolean oldpref = preferWidgets;
      
         // temporarily disable widget preference so the expression will
         // parse without widget refs
         if (oldpref)
         {
            preferWidgets = false;
         }
      }
      (
         LBRACKET^
         (
              { !inLikeClause }?
              expr (KW_FOR NUM_LITERAL)?
            | likeSubscript
         )
         RBRACKET!
      )
      {
         // reset back to state at entrance to this method
         if (oldpref)
         {
            preferWidgets = true;
         }         
      }
   ; 
   
/**
 * Read the following tokens until the <code>RBRACKET</code>, discard these
 * and build an <code>EXPRESSION -&gt; NUM_LITERAL</code> subtree with the
 * value of 1.  This is needed to support bizarre behavior of subscripts in 
 * a {@link #like_clause}.  It is called from {@link #subscript}.
 */
likeSubscript
   :
      (
         options { greedy=true; }
         :
         { LA(1) != RBRACKET }?
         .!
      )+
      {
         ## = #([EXPRESSION, "expression"], #([NUM_LITERAL, "1"]));
      }
   ;
   
/**   
 * Matches Progress 4GL built-in variable names that are <b>reserved</b> 
 * keywords. This includes functions that take NO parameters. Built-in
 * functions that take no parameters are treated like variables.  For more
 * on functions that take parameters and which have reserved names, see
 * {@link #reserved_functions}.
 * <p>
 * This rule also lists those system handles that are reserved keywords,
 * since such handles are essentially global variables and thus are
 * implemented here as such.
 * <p>
 * This is generally only used to generate the proper lookahead predictions
 * in the top level parser methods, such that calling rules would see tokens
 * of this type.  It is expected that the calling method would rewrite the
 * token type of such tokens such that this alternative is never actually
 * entered at runtime.  Thus this method really just provides a 'pull' into
 * the calling method.  See {@link #lvalue}.
 * <p>
 * Note that there are some built-in functions that can act like variables
 * OR as a normal parenthesized function.  In other words, for the following
 * list, the parenthesis are optional:
 * <p>
 * <ul>
 *    <li>ETIME
 *    <li>FRAME-COL
 *    <li>FRAME-DOWN
 *    <li>FRAME-LINE
 *    <li>FRAME-ROW
 *    <li>LINE-COUNTER
 *    <li>PAGE-NUMBER
 *    <li>USERID
 * </ul>
 * <p>
 * Such keywords are included in both this rule and in
 * <code>reserved_functions</code>.
 */
reserved_variables
   :
      (
           KW_U_CTRL
         | KW_U_MSG
         | KW_U_SERIAL
         | KW_U_PCTRL
         | KW_ACT_FORM
         | KW_ACT_WIN
         | KW_AUD_CTRL
         | KW_AUD_POL
         | KW_CLIP
         | KW_CODEBASE
         | KW_COMPILER 
         | KW_CUR_LANG
         | KW_CUR_WIN  
         | KW_DATASRV
         | KW_DBNAME
         | KW_DEBUGGER 
         | KW_DEF_WIN  
         | KW_ERR_STAT
         | KW_ETIME
         | KW_FIL_INFO
         | KW_FOCUS
         | KW_FR_COL
         | KW_FR_DB
         | KW_FR_DOWN
         | KW_FR_FIELD
         | KW_FR_FILE
         | KW_FR_INDEX
         | KW_FR_LINE
         | KW_FR_NAME
         | KW_FR_ROW
         | KW_FR_VAL
         | KW_GW
         | KW_GET_CP
         | KW_GO_PEND
         | KW_IS_ATTR
         | KW_LASTKEY
         | KW_LAST_EVT
         | KW_LINE_CNT
         | KW_LOG_MGR                                             
         | KW_MSG_LINE
         | KW_NOW
         | KW_NUM_ALIA
         | KW_NUM_DBS
         | KW_OPSYS
         | KW_PAGE_NUM
         | KW_PROC_HND
         | KW_PROC_ST
         | KW_PROFILER
         | KW_PROGRESS
         | KW_PROMSGS
         | KW_PROPATH
         | KW_PROVER
         | KW_RCOD_INF
         | KW_RETRY
         | KW_SCRN_LNS
         | KW_SECUR_P
         | KW_SELF
         | KW_SESSION
         | KW_SUPER
         | KW_TERM
         | KW_THIS_OBJ
         | KW_THIS_PRC
         | KW_TIME
         | KW_TRANS
         | KW_USERID
      )
   ;
   
/**   
 * Matches all possible user-defined symbol names. A very important element
 * of this approach is that the set <b>must include unreserved keywords</b>.
 * This is important since this rule is referenced in many different parser
 * rules in order to add this set to the lookahead list for various
 * alternatives.  Since the lexer defaults all symbols to the more specific
 * token type (if they match an unreserved keyword), one will only encounter
 * a <code>SYMBOL</code> token type if there is no overlap with the
 * unreserved keyword namespace.  In other words, since the text
 * &quot;char&quot; matches both an unreserved keyword and a possible valid
 * user-defined symbol, the lexer will default the token type to 
 * <code>KW_CHAR</code>.  This is critical since the number of places in the
 * parser that expect a real <code>SYMBOL</code> type are few and the token
 * type can be overridden based on context.  If the reverse bias was made in
 * the lexer, then it would be virtually impossible to implement the parser
 * because Progress is both very keyword driven and highly context sensitive.
 * This means that the number of places in the parser that depend upon a
 * keyword match is significantly higher than those that expect a symbol
 * match.  Thus we bias the token types toward keywords and then override
 * the type in the few places where we <b>know</b> (based on context) that
 * the next token is a <code>SYMBOL</code> rather than a keyword.
 * <p>
 * This rule enables those explicit overrides and it is only used in the
 * following cases:
 * <ul>
 *    <li> The calling rule expects the next token to be a <code>SYMBOL</code>
 *         token type rather than a keyword.  An example is a 
 *         <code>DEFINE VARIABLE</code> statement, where must be a symbol
 *         type and cannot be a keyword, even though in any keyword conflict
 *         the lexer defaults to the keyword type.  This is handled by the
 *         init code of this method which rewrites the token type to be a
 *         <code>SYMBOL</code> if all of the following are true (in all other
 *         cases an exception is thrown):
 *       <ul>
 *          <li> The token type is not already <code>SYMBOL</code>.
 *          <li> There is a keyword conflict.
 *          <li> The keyword is unreserved.
 *       </ul>
 *    <li> The calling rule requires that its lookahead prediction include
 *         the set of all possible user-defined symbols, in order to ensure
 *         that the top level parser rules will properly select and invoke
 *         the calling rule <b>but the calling rule itself rewrites the
 *         token type and is designed to never actually invoke the
 *         <code>symbol</code> rule itself</b>.  In this case, we refer to
 *         this as using <code>symbol</code> for &quot;pull&quot; purposes
 *         only.  This trick is critical for making the {@link #lvalue} and       
 *         {@link #func_call} rules work!
 * </ul>
 * <p>
 * One more trick is used here to ensure that the set of all possible input
 * token types is matched. Since all of the unreserved keywords are
 * assigned artificial token types from a list maintained in the parser,
 * it is designed such that all of these token types are created in a
 * contiguous set that is bounded by <code>BEGIN_UNRESERVED</code> and
 * <code>END_UNRESERVED</code> token types.  This trick allows this rule to 
 * add the set of all unreserved keywords to the rule by specifying a
 * range starting and ending with these token types.  Note that ANTLR does
 * generate a <code>matchRange</code> call from this grammar but it does not
 * actually generate or provide the backing method!  Such a method was added
 * manually to satisfy this requirement.
 * <p>
 * It is very important to note that any actual invocation of this rule
 * (in cases where this isn't only used in pull mode) will cause the token
 * type to be rewritten as <code>SYMBOL</code> or an exception to be
 * thrown.  No other alternative will result, no matter what the actual
 * token type is on input.  Only call this rule if this is the expected
 * behavior or if pull mode is being used.  In the case of pull mode, it
 * is critical that the calling rule rewrite the token type in an init action
 * to ensure that the token type never causes an actual invocation of this
 * rule.
 */
symbol
   :
      {
         if (LA(1) != SYMBOL)
         {
            Keyword found = sym.lookupKeyword(LT(1).getText());
            if (found != null && !found.isReserved())
            {
               LT(1).setType(SYMBOL);
            }
            else
            {
               throw new NoViableAltException(LT(1), getFilename());
            }
         }
      }
      ( 
           SYMBOL
           
           // this range of unreserved keywords exists to "pull" us into this
           // rule, but when we leave the type is always reset to SYMBOL
           // OR we throw an exception
         | BEGIN_UNRESERVED..END_UNRESERVED
      )
   ;
   
/**
 * This rule implements the special built-in <code>IF THEN ELSE</code>
 * function, which can be used in an expression (which is what differentiates
 * it from the language statement of the same name). Note that this construct
 * inherently conflicts with the language statement and the parser resolves
 * this just as Progress does: by giving the language statement 
 * {@link #if_stmt} precedence and thus matching on it first (in the
 * top-level {@link #statement} rule rather than in {@link #assignment}
 * which is how this current rule gets matched.
 * <p>
 * This rule is described as special because it doesn't follow the normal
 * patterns of function calls, even for built-in functions. In particular,
 * it has 3 reserved keywords that are all requried (the <code>ELSE</code>
 * is not optional) and in between each keyword is an expression. Since this
 * does not match the standard (generic) function call syntax, this function
 * was implemented as a standalone alternative in {@link #primary_expr},
 * which makes it a peer of the generic {@link #func_call} rule.
 * <p>
 * Ambiguity warnings were generated in the {@link #log_and_expr} and in
 * {@link #sum_expr} when this rule was added to the {@link #primary_expr}
 * rule. As far as can be found, there is no real ambiguity in practice and
 * those rules have had ambiguity warnings suppressed as a result.
 */
if_func   
   :
      i:KW_IF^ { saveAndReplaceType(#i, FUNC_POLY); }
      expr KW_THEN! expr KW_ELSE! expr
      {
         // write our function-specific annotations
         sym.annotateFunction(##.getText(), ##, false);            
      }
   ;
   
/**
 * This rule implements the special built-in <code>ACCUM</code> function,
 * which can be used in an expression (which is what differentiates
 * it from the language statement of the same name). Note that this construct
 * inherently conflicts with the language statement and the parser resolves
 * this just as Progress does: by giving the language statement 
 * {@link #accumulate_stmt} precedence and thus matching on it first (in the
 * top-level {@link #statement} rule rather than in {@link #assignment}
 * which is how this current rule gets matched.
 * <p>
 * This rule is described as special because it doesn't follow the normal
 * patterns of function calls, even for built-in functions. In particular,
 * it has 2 unparenthesized parameters that are actually either a
 * multi-keyword {@link #aggregate_phrase} or an {@link #expr}. Since this
 * does not match the standard (generic) function call syntax, this function
 * was implemented as a standalone alternative in {@link #primary_expr},
 * which makes it a peer of the generic {@link #func_call} rule.
 * <p>
 * Ambiguity warnings were generated in the <code>aggregate_phrase</code>
 * when this rule was added to the {@link #primary_expr}. As far as can be
 * found, there is no real ambiguity in practice and that rule has had
 * ambiguity warnings suppressed as a result.
 */
accum_function   
   :
      a:KW_ACCUM^ { saveAndReplaceType(#a, FUNC_POLY); }
      aggregate_phrase expr
      {
         // write our function-specific annotations
         sym.annotateFunction(##.getText(), ##, false);            
      }      
   ;
   
/**
 * This rule implements the special built-in <code>CAN-FIND</code> function,
 * which can be used in an expression.
 * <p>
 * This rule is described as special because it doesn't follow the normal
 * patterns of function calls, even for built-in functions. In particular,
 * its parameters are an exact match for the {@link #find_stmt} excluding
 * the <code>FIND</code> keyword itself! Since this does not match the
 * standard (generic) function call syntax, this function was implemented
 * as a standalone alternative in {@link #primary_expr}, which makes it a
 * peer of the generic {@link #func_call} rule.
 */
can_find_function   
   :
      {
         // push a scope, where the target buffers will be added
         sym.addSchemaScope(false);
      }
      c:KW_CAN_FIND^  { saveAndReplaceType(#c, FUNC_LOGICAL); }
      LPARENS!
      (
         options { generateAmbigWarnings = false; }
         :      
         KW_FIRST | KW_LAST | KW_NEXT | KW_PREV | KW_CURRENT
      )?
      record_phrase[true, false]      
      RPARENS!
      {
         // write our function-specific annotations
         sym.annotateFunction(##.getText(), ##, false);

         // remove the scope, but do not propagate
         sym.deleteSchemaScope(false);
      }
   ;
   
/**
 * This rule implements the special built-in <code>SUPER</code> function,
 * which can be used in an expression.
 * <p>
 * This rule is described as special because it doesn't follow the normal
 * patterns of function calls, even for built-in functions. In particular,
 * its optional parameters are an exact match for the {@link #parameter}
 * rule. Since this does not match the standard (generic) function call
 * syntax, this function was implemented as a standalone alternative in
 * {@link #primary_expr}, which makes it a peer of the generic 
 * {@link #func_call} rule.
 * <p>
 * ANTLR's bogus ambiguity warnings have been disabled.
 */
super_function   
   :
      s:KW_SUPER^  { saveAndReplaceType(#s, FUNC_POLY); }
      (
         options { generateAmbigWarnings = false; }
         :
         param_passing_list[false]
      )?
      {
         // write our function-specific annotations
         sym.annotateFunction(##.getText(), ##, false);            
      }
   ;
   
/**
 * This rule implements the special built-in functions that can be used
 * used in an expression for processing {@link #record} or
 * {@link #primary_expr} (a field) but which don't take normal parenthesized
 * parameters (the single parameter for each of these is postfixed) or the
 * function takes parenthesized parameters but the parameters include a
 * <code>record</code> (which is not normally allowed in an expression).  The
 * <code>INPUT</code> form can have an optional prefixed 
 * {@link #frame_reference}.
 * <p>
 * The following are matched:
 * <p>
 * <pre>
 *       Function Syntax           Optional Parens            Notes
 * ------------------------------  ---------------  ---------------------------
 * AVAILABLE record                       Y
 * AMBIGUOUS record                       Y
 * CURRENT-CHANGED record                 Y
 * ERROR( record )                        N
 * INPUT field                            N*        Parens can be matched in the
 *                                                  primary_expr rule that is
 *                                                  used to match the field, but
 *                                                  it doesn't operate the same
 *                                                  way as function parens.
 * INPUT FRAME frame field                N*        See INPUT field notes above.
 * LOCKED record                          Y
 * NEW record                             Y
 * RECID( record )                        N
 * ROWID( record )                        N
 * ROW-STATE( record )                    N
 * </pre>
 * <p>
 * This rule is described as special because it doesn't follow the normal
 * patterns of function calls, even for built-in functions.  Since this
 * does not match the standard (generic) function call syntax, this function
 * was implemented as a standalone alternative in {@link #primary_expr},
 * which makes it a peer of the generic {@link #func_call} rule.
 * <p>
 * Ambiguity warnings were generated in the {@link #parameter} and 
 * {@link #on_stmt} rules when this rule was added to {@link #primary_expr}.
 * As far as can be found, there is no real ambiguity in practice and those
 * rules have had ambiguity warnings suppressed as a result.
 * <p>
 * Parenthesis can be optionally used in some of these functions as noted
 * in the above table.
 */
record_funcs
   :
      {
         boolean matched_set = false;
         int     ftype       = -1;
         String  fname       = LT(1).getText();
      }
      (
           (
                a:KW_AVAIL^    { saveAndReplaceType(#a, FUNC_LOGICAL); }
              | b:KW_AMBIG^    { saveAndReplaceType(#b, FUNC_LOGICAL); }
              | c:KW_CUR_CHG^  { saveAndReplaceType(#c, FUNC_LOGICAL); }
              | d:KW_LOCKED^   { saveAndReplaceType(#d, FUNC_LOGICAL); }
              | e:KW_NEW^      { saveAndReplaceType(#e, FUNC_LOGICAL); }
           )           
           (LPARENS! { matched_set = true; } )?
           record[true, false, false]
           (
                // only match this if there was a matching LPARENS above
                { matched_set }?
                RPARENS!
              | // the empty alternative
           )
           
         | i:KW_INPUT^ { saveAndReplaceType(#i, FUNC_POLY); }
           primary_expr  // this allows the frame ref prefix AND also an
                         // optional set of parens around the parameter
                         // which is an undocumented option found in customer
                         // source
           
         | (
                er:KW_ERROR^      { saveAndReplaceType(#er, FUNC_LOGICAL); }
              | re:KW_RECID^      { saveAndReplaceType(#re, FUNC_RECID); }
              | ro:KW_ROWID^      { saveAndReplaceType(#ro, FUNC_ROWID); }
              | rs:KW_ROW_STAT^   { saveAndReplaceType(#rs, FUNC_INT); }
           )
           LPARENS! record[true, false, false] RPARENS!
      )
      {
         // write our function-specific annotations
         sym.annotateFunction(fname, ##, false);            
      }
   ;

/**
 * Matches the <code>NEW</code> phrase in an expression, allowing new object
 * instances to be instantiated. Uses {@link #user_defined_type_name} and
 * {@link #param_passing_list} and is called from {@link #primary_expr}.
 */
new_phrase
   :
      KW_NEW^ udt:user_defined_type_name param_passing_list[false]
      {
         // save the original token type, then rewrite the type
         ##.putAnnotation("oldtype", new Long(##.getType()));
         ##.setType(FUNC_CLASS);
            
         // write our function-specific annotations
         sym.annotateFunction(##.getText(), ##, false);
         
         // the above probably did not set the class name so we force it here
         String qname = (String) #udt.getAnnotation("qualified");
         sym.annotateClassRef(qname, ##);
      }
   ;
   
/**
 * Matches the <code>DYNAMIC-NEW</code> statement which looks like an
 * assignment, allowing new object instances to be instantiated using a
 * runtime (dynamic) choice of the class. Uses {@link #func_call_parameters}
 * and is called from {@link #assign_type_syntax_stmt_list} which means
 * that the leftmost <code>lvalue EQUALS</code> is factored out.
 */
dynamic_new_stmt
   :
      KW_DYN_NEW^ expr func_call_parameters[true]
      (
         options { generateAmbigWarnings = false; }
         :
         KW_NO_ERROR
      )?
      {
         // save the original token type, then rewrite the type
         ##.putAnnotation("oldtype", new Long(##.getType()));
         ##.setType(FUNC_CLASS);
            
         // write our function-specific annotations
         sym.annotateFunction(##.getText(), ##, false);
         
         // the above probably did not set the class name so we force it here
         sym.annotateClassRef("Progress.Lang.Object", ##);
      }
   ;
   
/**
 * This rule implements the special built-in functions that can be used
 * used in an expression but which don't take normal parenthesized parameters
 * AND for which the single {@link #lvalue} parameter for each of these is
 * prefixed in front of the function name.  Note that the <code>lvalue</code>
 * is not matched here, but rather in {@link #primary_expr}. 
 * <p>
 * The following are matched:
 * <p>
 * <pre>
 * lvalue ENTERED
 * lvalue NOT ENTERED
 * </pre>
 * <p>
 * This rule is described as special because it doesn't follow the normal
 * patterns of function calls, even for built-in functions.  Since this
 * does not match the standard (generic) function call syntax, this function
 * was implemented as an option after the <code>lvalue</code> alternative in
 * in the calling rule.
 */
postfix_funcs!
   :
      {
         boolean entered = true;
      }
      (KW_NOT { entered = false; } )? KW_ENTERED
      {
         String funcName = "entered";
         int    oldtype  = KW_ENTERED;
         
         if (!entered)
         {
            funcName = "not entered";
            oldtype  = NOT_ENTERED;
         }
         
         ## = #([FUNC_LOGICAL, funcName]);
         
         // write our function-specific annotations
         sym.annotateFunction(funcName, ##, false);            
         
         // add the oldtype annotation
         ##.putAnnotation("oldtype", new Long(oldtype));
      }
   ;
   
/**
 * This rule implements the reference-point construct used in at-phrase.
 * It is called from {@link #at_phrase} rule only.
 * <p>
 * ANTLR detects ambiguity due to the use of non-reserved keywords 
 * and the potential to use symbols and expressions in the k=1 construct in
 * an upper level rule {@link #frame_phrase}.  To eliminate the ambiguity
 * warning, ANTLR's greedy option had to be explicitly enabled.   
 * Since greedy is the default, this doesn't change the code generation, it
 * only suppresses the warning.  <b>Note that this rule checks the k=2 
 * lookahead before matching the <code>PLUS or MINUS</code> keywords.  This 
 * may be OK but it may also bypass the match in a case where something
 * unexpected follows the optional keyword.  Since this cannot be proven
 * for all cases, watch this code!</b> 
 */                                              
ref_point
   :  lvalue
      (
         options { greedy = true; }
         :
         (PLUS^ | MINUS^)
         numeric_literal
      )?
   ;
   
/**
 * This rule implements the at-phrase construct. It is called from
 * {@link #frame_phrase} and {@link #form_item} rules.
 * <p>
 * This rule defines the AT phrase of Progress 4GL. The definition here is
 * less strict than in 4GL, but for the sake of parsing a valid 4GL source it
 * is adequate.
 * <p> 
 * Ambiguity warnings were generated in the {@link #def_frame_stmt} and in
 * this rule because the optional alignment nonreserved keywords conflicted
 * with the symbols in the frame definition.
 * As far as can be found, there is no real ambiguity in practice and
 * those rules have hade ambiguity warning suppressed as a result.
 */                                              
at_phrase
   :  KW_AT^
      (
           location_spec location_spec
           (
              options { generateAmbigWarnings = false; }
              :
              KW_COLON_AL | KW_LEFT_AL | KW_RIGHT_AL
           )?
         | numeric_literal
      )
   ;
   
/**
 * This rule implements the column/row/x/y coordinate processing for the
 * <code>AT phrase</code> ({@link #at_phrase}).
 * <p>
 * The subtree is rooted by the keyword (which is the reason for using a 
 * separate rule for something so simple).
 */                                              
location_spec   
   :
      (
           (KW_COL^ | KW_ROW^ | KW_X^ | KW_Y^) numeric_literal
         | (KW_COL_OF^ | KW_ROW_OF^ | KW_X_OF^ | KW_Y_OF^) ref_point
      )   
   ;
   
/**
 * Matches the <code>OS-DELETE</code> language statement. Calls {@link #value}
 * and {@link #filename} rules.
 * <p>
 * The use of the <code>filename</code> rule in a loop is in conflict with
 * the non-reserved <code>RECURSIVE</code> keyword.  For this reason, the
 * loop has be made non-greedy and the bogus ambiguity warnings have been
 * disabled.
 * <p>
 * Supports an undocumented <code>NO-ERROR</code> keyword at the end.
 * <p>
 * Used by {@link #stmt_list}.
 */
os_delete_stmt
   :
      KW_OS_DEL^ 
      (
         options 
         {
            greedy = false;
            generateAmbigWarnings = false;
         }
         :
         external_name_spec
      )+
      (KW_RECURSE)?
      (KW_NO_ERROR)?
      DOT!
   ;
   
/**
 * Matches the <code>OS-CREATE-DIR</code> language statement. Calls 
 * {@link #external_name_spec}.
 * <p>
 * Used by {@link #stmt_list}.
 */
os_create_dir_stmt   
   :
      KW_OS_MKDIR^ (external_name_spec)+ DOT!
   ;
   
/**
 * Matches the <code>OS-APPEND, OS-COPY and OS-RENAME</code> language
 * statements. Calls {@link #external_name_spec}.
 * <p>
 * Used by {@link #stmt_list}.
 */
os_misc_stmt   
   :
      (KW_OS_APPND^ | KW_OS_COPY^ | KW_OS_REN^)
      external_name_spec
      external_name_spec
      DOT!
   ;
   
   
/**
 * Matches any operating system specific shell command (
 * <code>UNIX, DOS, OS2, BTOS, CTOS and VMS</code>) and the generic shell
 * command <code>OS-COMMAND</code> constructs including matching all possible
 * command tokens up to a following <code>DOT</code>.
 * <p>
 * This rule is unusual because it accepts an arbitrary list of any possible
 * or arbitrary token except the <code>DOT</code>.  Any of these constructs 
 * can be a {@link #value} rule which allows Progress expressions to be
 * evaluated and the results used as part of the command token list.
 * <p>
 * This approach is inherently ambiguous because the optional keywords (that
 * can precede the command token list) can appear in the subsequent list.
 * This ambiguity is resolved by greedily matching these keywords first.
 * Ambiguity warnings have been disabled since this is not a real issue.
 * <p>
 * Limitations of this rule include:
 * <p>
 * <ul>
 *    <li> Unmatched double or single quotes cannot be used. Insert these 
 *         inside a valid Progress string instead.
 *    <li> Whitespace that must be retained must be placed inside a valid 
 *         Progress string.
 *    <li> <code>DOT</code> is used as the termination character for the
 *         command token list so one can only pass a '.' inside a valid 
 *         Progress string.
 * </ul>
 * <p>
 * Progress shared these same limitations except it can accept an unmatched
 * double quotes character (which is probably just automatically ended at
 * the end of the file).  Progress has an additional limitation against 
 * using a colon followed by whitespace, which is an error.  This limitation
 * is not checked for by this rule.
 * <p>
 * Called by {@link #stmt_list}.
 */
shell_command_stmt
   :
      (
           KW_UNIX^
         | KW_DOS^
         | KW_OS_CMD^
         | KW_OS2^
         | KW_BTOS^
         | KW_CTOS^
         | KW_VMS^
      )
      (
         options { generateAmbigWarnings = false; }
         :
         KW_SILENT | KW_NO_WAIT
      )?
      (
         options { generateAmbigWarnings = false; }
         :
         KW_NO_CONS
      )?
      command_token_list[false]
      DOT!
   ;
   
/**
 * This rule implements the command token list processing for the shell
 * command statements ({@link #shell_command_stmt}).
 * <p>
 * <b>This rule uses a trick to create an artificial node as a root while
 * still allowing ANTLR to do normal tree building (no suppression of the
 * AST build).</b>  This is only useful in cases where there is no valid
 * token which can be used as a tree root AND where a loop is being used
 * such that the contents of the loop cannot be referenced in a manually
 * built #([,],,,) ANTLR construct. ANTLR doesn't support (as of 2.7.4) the
 * ability to label a loop such that the entire contents can be referenced.
 * If one labels the individual elements inside the loop, only the last one
 * is available when the loop concludes.  However, by leaving the normal
 * tree creation unchanged but manually creating an artificial node and
 * manually setting this node as the root of the returned tree BEFORE the
 * children are added, this problem can be solved.
 * <p>
 * All children of the root node are one of the following types:
 * <p>
 * <pre>
 * STRING
 * COMMAND_TEXT
 * KW_VALUE
 * </pre>
 * <p>
 * The <code>COMMAND_TEXT</code> node is set as the type of any matched
 * token which is not of the other two types.  Since any token except
 * <code>DOT</code> can be included here, these tokens MUST be rewritten
 * otherwise downstream processing of the resulting tree would be quite
 * difficult.
 * <p>
 * This rule only exists to create the proper AST, otherwise it would be much
 * simpler to inline since the match is trivial.
 *
 * @param    io_options
 *           <code>true</code> to avoid matching keywords that can be used
 *           as options in one of the I/O through statements.
 */
command_token_list [boolean io_options]   
   :
      {
         astFactory.makeASTRoot(currentAST, #[COMMAND_TOKENS,"tokens"]);
         Token tok;
      }
      (
         options
         {
            generateAmbigWarnings = false;
         }                                            
         :
           { tok = LT(1); }
           v:value
           { processHiddenText(tok, #v, true); }
           
           // only a DOT that is followed by whitespace will trigger the end
           // of this loop, but something like "filename.*" is valid and will
           // appear as a SYMBOL DOT MULTIPLY, so it must be manually
           // disambiguated
         | { !followedByWhitespace(LT(1)) }?
           d:DOT
           {
              processHiddenText(d, #d, true);
              #d.setType(COMMAND_TEXT);
           }
         
         | 
           // avoid collisions with the following options when called from
           // io_through_stmt
           {
              !io_options ||
              (LA(1) != KW_ECHO     && LA(1) != KW_NO_ECHO  &&
               LA(1) != KW_UNBUF    && LA(1) != KW_MAP      &&
               LA(1) != KW_NO_MAP   && LA(1) != KW_NO_CVT   &&
               LA(1) != KW_CONVERT  && LA(1) != KW_PAGE_SZ  &&
               LA(1) != KW_PAGED)
           }?
           t:~DOT
           {
              processHiddenText(t, #t, true);
              
              if (#t.getType() != STRING) 
              {
                 #t.setType(COMMAND_TEXT);
              }
           }         
      )*   
      { if (##.getNumberOfChildren() == 0) ## = null; }
   ;
   
/**
 * Tokenizes a Progress 4GL source file (input as a stream of characters)
 * into a stream of tokens suitable for the <code>ProgressParser</code>.
 * Each token has an integer token type by which the parser references and
 * matches tokens.  This is done to make a high performance parser. To make
 * this work, the lexer must create a token object when it matches a top
 * level rule, this token object includes the token type and the actual text
 * found in the original character stream.  The lexer's main job is to
 * lookahead into the character stream, switch into the top level rule that
 * matches some set of characters and when this is complete, to create a
 * valid token object and return this to the parser.  The lexer thus
 * implements a token stream interface from the parser's perspective.
 * <p>
 * <b>Please note that this is a generated file using ANTLR 2.7.4 and a
 * grammar specified in <code>progress.g</code>.</b>
 * <p>
 * The Progress 4GL language is highly context sensitive, particularly in
 * the area of user-defined symbols and language keywords. Examples of this 
 * attribute include:
 * <ul>
 *    <li> The polymorphic usage of the same language keywords for different
 *         meaning depending on which language statement is being processed.
 *    <li> The very large number of language keywords means that many or
 *         most of these must be left as unreserved keywords (the user may
 *         define symbols that hide these keywords).
 *    <li> The same symbol may appear in multiple different namespaces in
 *         the same procedure, yet these different uses are all
 *         properly differentiated.
 * </ul>
 * <p>
 * Such constructions cannot be disambiguated by looking ahead on a character 
 * basis.  Instead, the context depends upon a lookahead at the token level.
 * This means that such items must be handled by the parser rather than the
 * lexer.  For this reason, the lexer is only partially able to participate
 * in the process of symbol resolution.
 * <p>
 * There are many different types of symbols.  Language keywords are the
 * most obvious symbol that are visible in all Progress 4GL source files.
 * Other types include variables, procedure names, function names, labels,
 * temp table names, etc.  All of these symbols follow the same basic
 * rules by which they are matched in the lexer, for details on this, see
 * <code>{@link #mSYMBOL}</code>.  As a result, the basic symbol rule in the
 * lexer matches all of these types and cannot set the token type properly
 * since it doesn't have enough information to do so at the time it is
 * matching.
 * <p>
 * Symbol resolution is the process by which the proper token type is  
 * assigned to a token created from a match to the generic symbol rule.
 * In the lexer, the only symbol resolution that can be done is to match
 * with language keywords.  Language keywords are given preference over 
 * other symbol types in any case where there is a conflict.  This is done
 * because there are significantly more locations in the parser that are 
 * matching on language keywords than there are locations where a user
 * defined symbol can be in use.  Thus it makes sense to handle the conflicts
 * at the smaller number of locations in the parser, and assume that a
 * keyword match is first a keyword.  This has a great simplifying effect
 * on the parser design.  The result is selected locations in the parser
 * that override (reassign) the token type based on context. 
 * <p>
 * It seems that the parser could provide feedback to the lexer at certain
 * points in its parsing.  By storing shared state with the lexer, the
 * lexer could then assign the token type properly instead of taking the 
 * &quot;error on the side of keywords&quot; approach which it takes now.
 * However, due to the manner in which the parser does lookahead, this shared
 * state implementation is not possible.  The lookahead can be done at a level
 * in the parser at which the context is not yet known (outside of the rules
 * that actually handle the matching). If the lookahead depth (the 
 * &quot;k&quot;) is greater than 1, then symbolic tokens (with the default
 * token type set) can be returned before the parser can set the state
 * variables that would properly control the assignment of those same
 * tokens.  <b>The decoupling of parser lookahead from the rules that know
 * enough to set the state causes it to be infeasible to implement shared 
 * state between the parser and the lexer.</b>
 * <p>
 * The hundreds (thousands?) of keywords supported by Progress 4GL naturally
 * drive the use of symbol resolution and token type rewriting rather than
 * hard coding a rule for every keyword.  This is also made necessary by the
 * need to implement abbreviation support.  All of this is handled in the
 * lexer by the relatively small rule <code>mSYMBOL</code>. This efficient
 * approach to keyword processing makes the lexer significantly smaller and
 * much simpler to implement and maintain that it otherwise might have been
 * given the keyword-richness of Progress 4GL.
 * <p>  
 * Other key design points:
 * <ul>
 *    <li> Progress 4GL is generally case-insensitive and for this reason
 *         this grammar has been designed to ignore case.  All symbol
 *         definitions use lowercase characters and ANTLR generates code to
 *         match both the uppercase and lowercase version of each character.
 *         This is enabled using the lexer options section.
 *    <li> The lexer is defined to create a single token of any string or
 *         comment, including nested comments (which are all provided to the
 *         parser in one single token).  All logic for matching comments and
 *         string literals (both single quoted and double quoted) is embedded
 *         in the lexer.  This <b>greatly simplifies</b> the parser since 
 *         Progress allows these to be placed in such a wide range of
 *         locations (e.g. comments can be inserted anywhere in between
 *         keywords in any language statement) that it would be impossible to
 *         properly manage the state in the parser due to the essentially
 *         infinite number of permutations.  
 *    <li> At the top level of the lexer, all tokens are one of the
 *         following types: 
 *       <ul>
 *          <li> comments
 *          <li> string literals
 *          <li> whitespace (spaces, tabs, carriage returns, line feeds)
 *          <li> operators
 *             <ul>
 *                <li> relational
 *                <li> comparisons
 *                <li> arithmetic
 *                <li> assignment
 *                <li> precedence (parenthesis)
 *                <li> array subscripting
 *             </ul>
 *          <li> statement and parameter separators (period, colon, comma,
 *               and at sign)
 *          <li> integer and decimal literals
 *          <li> date literals
 *          <li> the unknown value
 *          <li> language keywords 
 *          <li> user-defined symbols
 *          <li> filenames (relative or absolute)
 *          <li> r-code library references
 *          <li> partially or fully qualified database symbols
 *          <li> the ignore character for set/update statements (the caret)
 *          <li> tildes 
 *          <li> other special characters that are invalid Progress 4GL 
 *               but which may occur in arbitrary token lists used in
 *               things such as shell commands that are embedded in Progress
 *               source code (these are "unknown tokens")
 *       </ul>
 *    <li> The lexer's top level entry point is <code>nextToken</code>. This
 *         method has a for-loop that uses up to 3 characters of lookahead
 *         to identify which of the top level token rules to call. To do
 *         this, each token rule's left-most match characters are "rolled up"
 *         and tested in this top level rule. 
 *    <li> Since comments and strings are at the same precedence level, each 
 *         can be embedded in the other.
 *    <li> Comments and whitespace tokens are both artificially set to the
 *         special "skip" token type, which drops them from the token
 *         stream seen by the parser.  This is an important feature that
 *         simplifies the parser design.  At some future time, it is possible
 *         to capture the comments into another hidden token stream, but at 
 *         this time it is not done.
 *    <li> To simplify the parser design, an assumption was made that every
 *         Progress source file comes with the dot after the last statement.
 *         Progress allows the last dot to be omitted. The lexer is
 *         responsible for injection of a fake dot character if this is the
 *         case. it is done by using a filtered input stream provided by
 *         DotKludgeStream and DotKludgeReader classes. Those classes append
 *         a dot character to the regular input if necessary thus providing
 *         a consistent input to the lexer.
 * </ul>
 * <p>
 * All token types are defined as integer constants in 
 * <code>{@link ProgressParserTokenTypes}</code> which is an Interface that
 * the parser, lexer and other related classes all implement.  This allows 
 * all of these classes to directly refer to the token types and share this
 * common set of definitions.  All top-level lexer rules generate a token
 * of the same name in the <code>ProgressParserTokenTypes</code> Interface. 
 * There is a tokens { } section in the parser where artificial tokens are
 * defined (tokens that are not backed by a lexer rule).  These tokens are
 * also added to the <code>ProgressParserTokenTypes</code> Interface and can
 * thus be directly referenced by the lexer and parser.
 * <p>
 * Two of the most important lexer rules are {@link #mSYMBOL} and
 * {@link #mNUM_LITERAL}.  Each of these rules process highly ambiguous
 * constructs requiring a rule that recognizes multiple token types and
 * overrides the default type based on the internal matching.  Both rules
 * have to disambiguate the '.' character and the <code>mSYMBOL</code> rule
 * also has to handle the '/' character.  <b>Changes to these rules must be
 * handled with the utmost care!</b>
 * <p>
 * Due to the near total overlap between schema processing and Progress 4GL
 * source code, this lexer has been updated to support a mode with schema- 
 * specific keywords such that it can lex the standard schema dump-file
 * (.df format).  The only difference is the addition of a small set of
 * schema-specific keywords that are activated by the method
 * {@link #activateSchemaProcessing}.  Once activated it cannot be removed!
 */
class ProgressLexer extends Lexer;

options
{
   k = 4;                                 // lookahead depth is 4 chars
   charVocabulary = '\u0000'..'\ufffe';   // specifies range 0x0000-0xFFFE
                                          // (0xFFFF is EOF)
   testLiterals = false;                  // do our own keyword processing
   caseSensitive = false;                 // Progress 4GL is case-insensitive
   caseSensitiveLiterals = false;   
}

// helper members and methods
{
   /** Stores all keyword definitions in alphabetical order. */
   private static Keyword[] keywords = null;

   /** Stores all ignored keyword definitions. */
   private static Keyword[] ignored = null;

   /** Debug mode, defaults to <code>false</code>. */
   private boolean debug = false;
   
   /** Schema mode, defaults to <code>false</code>. */
   private boolean schema = false;
   
   /** Export mode, defaults to <code>false</code>. */
   private boolean export = false;
   
   /** Hidden mode, defaults to <code>false</code>. */
   private boolean hidden = false;
   
   /** Raw string processing mode, defaults to <code>false</code>. */
   private boolean raw = false;
   
   /** Honor backslashes as escape characters. */
   private boolean unixEscapes = true;
   
   /**
    * Enables language keyword resolution.  This instance is shared with the
    * parser but no real time state sharing is done.
    */
   private SymbolResolver sym = null;

   /**
    * Creates a <code>Keyword</code> array and initializes that array to
    * the list of all recognized Progress 4GL language keywords. The
    * initialization is done in-line and the source file is organized in
    * an easy to maintain column format.
    * <p>
    * The following values are needed for each <code>Keyword</code> object:
    * <ul>
    *    <li> full keyword text for the keyword
    *    <li> the minimum number of characters for matching ( > 0 means that
    *         abbreviations are possible, 0 means no abbreviations)
    *    <li> the integer token type to assign to the keyword on a match
    *         (see below)
    *    <li> whether it is a reserved keyword or not (true means it is
    *         reserved)
    * </ul>
    * <p>
    * See <code>{@link Keyword}</code> for more details.
    * <p>
    * The core idea here is to populate a facility where the symbol text (in
    * the mSYMBOL method) can be submitted to the keyword dictionary for a
    * lookup. If a match is found, then the token type of the token is set
    * based on the token type field in the <code>Keyword</code> object
    * found in the dictionary.  This returned token type will override the
    * default token type of <code>SYMBOL</code>.  Since all of these keywords
    * are not encoded in the matching rules of the grammar itself, there are
    * no actual token definitions in the lexer that match these keywords.
    * For this reason, we take advantage of the feature in the parser in
    * which we can create (reserve) a unique integer value and associate it 
    * with a specific artificial token name (e.g. KW_DEFINE).  These
    * artificial token types are used as the values when constructing
    * <code>Keyword</code> objects.
    * <p>
    * When adding <code>Keyword</code> objects to this array, make sure to
    * add the associated artificial token type to the tokens { } section
    * of the parser.
    * <p>
    * Some entries have inherent conflicts due to ambiguity in the Progress
    * language.  For example, <code>WIDTH</code> and <code>WIDTH-CHARS</code>
    * can both be abbreviated as <code>WIDTH</code> even though they are
    * used for completely separate purposes and are described as separate
    * keywords.  Of course, one cannot use <code>WIDTH-</code> or
    * <code>WIDTH-C</code> etc as a replacement for <code>WIDTH</code>. So
    * these really are 2 different keywords but there are fixups inside
    * Progress that makes it tolerant of this issue.  In our case, we detect
    * these possible overlaps in the parser and rewrite the token to the
    * expected version so that downstream processing is easier.  In this
    * list, some changes have been made to ensure that conflicts don't
    * occur (e.g. <code>WIDTH-CHARS</code> is stated to have a minimum
    * abbreviation of 6 chars instead of the correct value of 5).
    * <p>
    * The following keywords exist in the keyword list but they have no
    * references in the parser.  There are two reasons for this. First,
    * some of the keywords are used for stored procedure support which is
    * currently incomplete in the parser.  When that support is completed,
    * these keywords should be removed from this list.  Second, there are
    * several keywords which are listed in the keyword index and in the notes
    * their purpose is stated.  However, no other reference to these keywords
    * exists in the language reference.  So these are listed as "possible
    * undocumented ..." features.  In this case, when there is enough
    * information available to determine how they must be supported, those
    * changes to the parser can be made and the tokens should then be removed
    * from this list.
    * <pre>
    *                                        Exists
    *                                        in 4GL
    *                                        Keyword
    * Keyword Text              Token        Index    Issue 
    * ------------------------  -----------  -------  ----------------------------------------------------------------------
    * CLOSEALLPROCS             KW_CLOSE_AP  no       special stored-procedure-related name, not fully implemented in parser
    * MARGIN-HEIGHT-CHARS       KW_MARG_H_C  yes      possible undocumented attribute
    * MARGIN-HEIGHT-PIXELS      KW_MARG_H_P  yes      possible undocumented attribute
    * MARGIN-WIDTH-CHARS        KW_MARG_W_C  yes      possible undocumented attribute
    * MARGIN-WIDTH-PIXELS       KW_MARG_W_P  yes      possible undocumented attribute
    * OCTET-LENGTH              KW_OCT_LEN   yes      possible undocumented FIND option
    * PROC-TEXT                 KW_PROC_TXT  yes      special stored-procedure-related name, not fully implemented in parser
    * PROC-TEXT-BUFFER          KW_PROC_T_B  yes      special stored-procedure-related name, not fully implemented in parser
    * RULE-ROW                  KW_RULE_ROW  yes      possible undocumented attribute
    * RULE-Y                    KW_RULE_Y    yes      possible undocumented attribute
    * SCROLL-DELTA              KW_SCR_DELT  yes      possible undocumented attribute
    * SCROLL-OFFSET             KW_SCR_OFFS  yes      possible undocumented attribute
    * SCROLLED-ROW-POSITION     KW_SCR_RPOS  yes      possible undocumented attribute
    * SEND-SQL-STATEMENT        KW_SEND_SQL  yes      special stored-procedure-related name, not fully implemented in parser
    * TRAILING                  KW_TRAILING  yes      possible undocumented built-in function and option for FOR/DO/REPEAT
    * </pre>
    * <p>
    * In addition to the above keywords, there are many other keywords that
    * may appear in the keyword list below, but which are not otherwise
    * referenced.  There are 3 common reasons for this:
    * <ul>
    *   <li> an event name will be used as the <code>oldtype</code>
    *        annotation of an EVENT node
    *   <li> an constant name may be used in certain built-in functions or
    *        method calls
    *   <li> an built-in function (or built-in global variable) name will be
    *        used as the <code>oldtype</code> annotation
    * </ul>
    * There is no problem with the parser/lexer in this regard.  Do not be
    * concerned!
    * <p>
    * With this noted, there are some cases where the keyword list appears
    * to be missing some keywords that are listed in the Progress 4GL Language
    * Reference Keyword Index.
    * <p>
    * The most common case is a "duplicated" keyword that is really a synonym
    * for another longer keyword.  The longer keyword has an abbreviation
    * which makes it match the same text as the shorter keyword.  From the
    * Progress compiler's perspective both forms are treated the same. This
    * is different from the <code>WIDTH</code> and <code>WIDTH-</code> case
    * noted above, since in that case the keywords are not synonyms. In
    * the duplicate case, the shorter keyword is considered a duplicate of
    * the longer one. WE ALWAYS INCLUDE THE LONGER KEYWORD WITH ITS
    * ABBREVIATION AND THE SHORT KEYWORD IS NOT NEEDED BECAUSE IT IS HANDLED
    * BY THE OTHER ONE THAT IS ALREADY THERE.  In other words, this short
    * keyword never needs to be added to the keyword list because the support
    * is already sufficient.  It is unknown why Progress documented "both"
    * keywords, because the longer one covers both usages.  In all such cases
    * the short keyword will be missing and that is OK! Here is the list:
    * <p>
    * <pre>
    * Short Keyword Text  Long Keyword Text    Abbreviation
    * ------------------  -------------------- ------------
    * ACCUM               ACCUMULATE           5
    * ASC                 ASCENDING            3
    * BUTTON              BUTTONS              6
    * DESC                DESCENDING           4
    * FIELD               FIELDS               5
    * FORM                FORMAT               4
    * FORWARD             FORWARDS             7
    * MAX                 MAXIMUM              3
    * MIN                 MINIMUM              3
    * REPOSITION-BACKWARD REPOSITION-BACKWARDS 19
    * REPOSITION-FORWARD  REPOSITION-FORWARDS  18
    * </pre>
    * <p>
    * The following are keywords that are documented in the Progress 4GL
    * Language Reference Keyword Index but which are not really language
    * keywords at all:
    * <p>
    * <ul>
    *    <li><code>DEFINED</code> - this is a symbol that is a preprocessor
    *        function but it is not something that has any meaning in a 4GL
    *        program so the parser should not have any knowledge of it
    *    <li><code>ENABLED-FIELDS</code> - the keyword index displays this as
    *        "ENABLED-FIELDS" (with the double quotes) and this is noted as
    *        an option to the validate() method; testing has shown that this
    *        is a string literal NOT a real keyword (if you take the double
    *        quotes away you get a compiler error so this is not really a
    *        keyword)
    *    <li><code>SET-CURRENT-VALUE</code> - this is a possible value
    *        returned inside a string from the <code>DBRESTRICTIONS</code>
    *        built-in function but this can't appear in directly (unquoted)
    *        in 4GL source so this is NOT a real keyword
    * </ul>
    * <p>
    * Progress is so over-the-top keyword intensive that they sometimes have
    * multiple forms of the same keyword (e.g. USER and USERID, WAIT and
    * WAIT-FOR, GETBYTE and GET-BYTE). These are not abbreviations but just
    * synonyms that require separate entries in the keyword index. The key
    * point is that they will both resolve to the same token type (e.g.
    * WAIT and WAIT-FOR both resolve to KW_WAIT_FOR). The order of these
    * in the list will determine what "standard text" will be found when
    * using a reverse text lookup (lookup of the text for a keyword based on
    * its token type). In particular, the LAST entry to be added will be the
    * "canonical" text returned from a reverse lookup. This affects reports
    * and other features. Always order these with the most common form of
    * the keyword as the last entry in the keyword index. So WAIT-FOR must
    * follow WAIT in the list to ensure that the reverse lookup of 
    * KW_WAIT_FOR would return the text WAIT-FOR instead of WAIT.
    */
   static
   {
      keywords = new Keyword[]
      {
         /* format:  Full Keyword Text (lowercased)    Abr  Token Type   Reserved */
         new Keyword("_cbit"                          ,  0, KW_CBIT    , false),  // undocumented
         new Keyword("_control"                       ,  0, KW_U_CTRL  , true ),  // undocumented
         new Keyword("_msg"                           ,  0, KW_U_MSG   , true ),  // undocumented
         new Keyword("_pcontrol"                      ,  0, KW_U_PCTRL , true ),  // undocumented
         new Keyword("_serial-num"                    ,  7, KW_U_SERIAL, true ),  // undocumented
         new Keyword("abort"                          ,  0, KW_ABORT   , false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("absolute"                       ,  3, KW_ABS     , false),
         new Keyword("abstract"                       ,  0, KW_ABSTRACT, false),
         new Keyword("accelerator"                    ,  0, KW_ACCEL   , false),
         new Keyword("accept-changes"                 ,  0, KW_ACC_CHG , false),  // missing from keyword index and UNTESTED at this time
         new Keyword("accept-row-changes"             ,  0, KW_ACC_RCHG, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("accumulate"                     ,  5, KW_ACCUM   , true ),  // handles "separate keyword" accum too
         new Keyword("active"                         ,  0, KW_ACTIVE  , false),  // missing from keyword index and UNTESTED at this time
         new Keyword("active-form"                    ,  0, KW_ACT_FORM, true ),
         new Keyword("active-window"                  ,  0, KW_ACT_WIN , true ),
         new Keyword("actor"                          ,  0, KW_ACTOR   , false),  // missing from keyword index and UNTESTED at this time
         new Keyword("add"                            ,  0, KW_ADD     , true ),
         new Keyword("add-buffer"                     ,  0, KW_ADD_BUF , false),
         new Keyword("add-calc-column"                ,  0, KW_ADD_C_C , false),
         new Keyword("add-columns-from"               ,  0, KW_ADD_C_F , false),
         new Keyword("add-events-procedure"           ,  0, KW_ADD_EVTP, false),
         new Keyword("add-fields-from"                ,  0, KW_ADD_F_F , false),
         new Keyword("add-first"                      ,  0, KW_ADD_1ST , false),
         new Keyword("add-header-entry"               ,  0, KW_ADD_HENT, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("add-index-field"                ,  0, KW_ADD_IDXF, false),
         new Keyword("add-interval"                   ,  0, KW_ADD_INVL, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("add-last"                       ,  0, KW_ADD_LAST, false),
         new Keyword("add-like-column"                ,  0, KW_ADD_L_C , false),
         new Keyword("add-like-field"                 ,  0, KW_ADD_LIKF, false),
         new Keyword("add-like-index"                 ,  0, KW_ADD_LIKI, false),
         new Keyword("add-new-field"                  ,  0, KW_ADD_NEWF, false),
         new Keyword("add-new-index"                  ,  0, KW_ADD_NEWI, false),
         new Keyword("add-parent-id-relation"         ,  0, KW_ADD_PREL, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("add-relation"                   ,  0, KW_ADD_REL , false),  // missing from keyword index and UNTESTED at this time
         new Keyword("add-schema-location"            ,  0, KW_ADD_SLOC, false),
         new Keyword("add-source-buffer"              ,  0, KW_ADD_SRCB, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("add-super-procedure"            ,  0, KW_ADD_SUP , false),
         new Keyword("adm-data"                       ,  0, KW_ADM_DATA, false),
         new Keyword("adm-timing-out"                 ,  0, KW_ADM_TO  , false),  // not a real keyword, but is normally defined in the PSC webspeed 4GL code
         new Keyword("advise"                         ,  0, KW_ADVISE  , false),
         new Keyword("after-buffer"                   ,  0, KW_AFT_BUFF, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("after-fill"                     ,  0, KW_AFT_FILL, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("after-row-fill"                 ,  0, KW_AFT_R_F , false),  // missing from keyword index and UNTESTED at this time
         new Keyword("after-rowid"                    ,  0, KW_AFT_ROID, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("after-table"                    ,  0, KW_AFT_TBL , false),  // missing from keyword index and UNTESTED at this time
         new Keyword("alert-box"                      ,  0, KW_ALERT_BX, false),
         new Keyword("alias"                          ,  0, KW_ALIAS   , true ),
         new Keyword("all"                            ,  0, KW_ALL     , true ),
         new Keyword("allow-column-searching"         ,  0, KW_ALLW_C_S, false),
         new Keyword("allow-replication"              ,  0, KW_ALLW_REP, false),
         new Keyword("alter"                          ,  0, KW_ALTER   , true ),
         new Keyword("alternate-key"                  ,  0, KW_ALT_KEY , false),
         new Keyword("always-on-top"                  ,  0, KW_ALW_ON_T, false),
         new Keyword("ambiguous"                      ,  5, KW_AMBIG   , true ),
         new Keyword("analyze"                        ,  6, KW_ANALYZ  , true ),  // totally undocumented
         new Keyword("and"                            ,  0, KW_AND     , true ),
         new Keyword("ansi-only"                      ,  0, KW_ANSI_ONL, false),
         new Keyword("any"                            ,  0, KW_ANY     , true ),
         new Keyword("any-key"                        ,  0, KW_ANY_KEY , false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("any-printable"                  ,  0, KW_ANY_PRT , false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("anywhere"                       ,  0, KW_ANYWHERE, false),
         new Keyword("append"                         ,  0, KW_APPEND  , false),
         new Keyword("append-child"                   ,  0, KW_APPEND_C, false),
         new Keyword("append-line"                    ,  0, KW_APPEND_L, false),  // missing in keyword index, found in customer source
         new Keyword("appl-alert-boxes"               , 10, KW_APPL_A_B, false),
         new Keyword("appl-context-id"                ,  0, KW_APPL_CID, false),
         new Keyword("application"                    ,  0, KW_APPL    , false),
         new Keyword("apply"                          ,  0, KW_APPLY   , true ),
         new Keyword("apply-callback"                 ,  0, KW_APPL_CBK, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("appserver-info"                 ,  0, KW_APP_INFO, false),
         new Keyword("appserver-password"             ,  0, KW_APP_PW  , false),
         new Keyword("appserver-userid"               ,  0, KW_APP_UID , false),
         new Keyword("array-message"                  ,  0, KW_ARR_MSG , false),
         new Keyword("as"                             ,  0, KW_AS      , true ),
         new Keyword("ascending"                      ,  3, KW_ASC     , true ),  // ASC is documented as a separate non-reserved keyword, but in fact it is reserved AND ASCENDING can be used as a complete synonym
         new Keyword("asciitohtml"                    ,  0, KW_ASCII_2H, false),  // not a real keyword, but is normally defined in the PSC webspeed 4GL code
         new Keyword("ask-overwrite"                  ,  0, KW_ASK_OVER, false),
         new Keyword("assembly"                       ,  0, KW_ASSEMBLY, false),
         new Keyword("assign"                         ,  0, KW_ASSIGN  , true ),
         new Keyword("asynchronous"                   ,  0, KW_ASYNC   , true ),
         new Keyword("async-request-count"            ,  0, KW_ASYNC_RC, false),
         new Keyword("async-request-handle"           ,  0, KW_ASYNC_RH, false),
         new Keyword("at"                             ,  0, KW_AT      , true ),
         new Keyword("attach-data-source"             ,  0, KW_ATT_DSRC, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("attached-pairlist"              ,  0, KW_ATT_PLST, false),
         new Keyword("attr-space"                     ,  4, KW_ATTR    , true ),
         new Keyword("attribute-names"                ,  0, KW_ATTR_NAM, false),
         new Keyword("audit-control"                  ,  0, KW_AUD_CTRL, true ),
         new Keyword("audit-enabled"                  ,  0, KW_AUD_ENAB, false),
         new Keyword("audit-event-context"            ,  0, KW_AUD_EV_C, false),
         new Keyword("audit-policy"                   ,  0, KW_AUD_POL , true ),
         new Keyword("authentication-failed"          ,  0, KW_AUTHEN_F, false),
         new Keyword("authorization"                  ,  0, KW_AUTHORZN, true ),
         new Keyword("auto-completion"                ,  9, KW_AUTO_COM, false),
         new Keyword("auto-delete"                    ,  0, KW_AUTO_DEL, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("auto-delete-xml"                ,  0, KW_AUTO_D_X, false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("auto-end-key"                   ,  0, KW_AUTO_END, false),
         new Keyword("auto-endkey"                    ,  0, KW_AUTO_END, false),
         new Keyword("auto-go"                        ,  0, KW_AUTO_GO , false),
         new Keyword("auto-indent"                    ,  8, KW_AUTO_IND, false),
         new Keyword("auto-resize"                    ,  0, KW_AUTO_RES, false),
         new Keyword("auto-return"                    ,  8, KW_AUTO_RET, true ),
         new Keyword("auto-synchronize"               ,  0, KW_AUTO_SYN, false),
         new Keyword("auto-validate"                  ,  8, KW_AUTO_VAL, false),
         new Keyword("auto-zap"                       ,  6, KW_AUTO_ZAP, false),
         new Keyword("automatic"                      ,  0, KW_AUTOMATC, false),
         new Keyword("available"                      ,  5, KW_AVAIL   , true ),
         new Keyword("available-formats"              ,  0, KW_AVL_FMTS, false),
         new Keyword("available-messages"             ,  0, KW_AVL_MSGS, false),  // not a real keyword, but is normally defined in the PSC webspeed 4GL code 
         new Keyword("average"                        ,  3, KW_AVERAGE , false),
         new Keyword("avg"                            ,  0, KW_AVG     , false),
         new Keyword("background"                     ,  4, KW_BACKGRND, true ),
         new Keyword("backspace"                      ,  0, KW_BACKSP  , false),
         new Keyword("backwards"                      ,  8, KW_BACKWARD, false),
         new Keyword("back-tab"                       ,  0, KW_BACK_TAB, false),
         new Keyword("basic-logging"                  ,  0, KW_BAS_LOGG, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("base64-decode"                  ,  0, KW_BASE64_D, false),
         new Keyword("base64-encode"                  ,  0, KW_BASE64_E, false),
         new Keyword("base-ade"                       ,  0, KW_BASE_ADE, false),
         new Keyword("base-key"                       ,  0, KW_BASE_KEY, false),
         new Keyword("batch-mode"                     ,  5, KW_BATCH_MO, false),
         new Keyword("batch-size"                     ,  0, KW_BATCH_SZ, false),
         new Keyword("before-buffer"                  ,  0, KW_B4_BUFF , false),  // missing from keyword index and UNTESTED at this time
         new Keyword("before-fill"                    ,  0, KW_B4_FILL , false),  // missing from keyword index and UNTESTED at this time
         new Keyword("before-hide"                    ,  8, KW_B4_HIDE , true ),
         new Keyword("before-row-fill"                ,  0, KW_B4_R_F  , false),  // missing from keyword index and UNTESTED at this time
         new Keyword("before-rowid"                   ,  0, KW_B4_ROWID, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("before-table"                   ,  0, KW_B4_TABLE, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("begin-event-group"              ,  0, KW_BEG_EV_G, false),
         new Keyword("begins"                         ,  0, KW_BEGINS  , true ),
         new Keyword("bell"                           ,  0, KW_BELL    , true ),
         new Keyword("between"                        ,  0, KW_BETWEEN , true ),
         new Keyword("bgcolor"                        ,  3, KW_BGCOLOR , false),
         new Keyword("big-endian"                     ,  0, KW_B_ENDIAN, true ),
         new Keyword("binary"                         ,  0, KW_BINARY  , false),
         new Keyword("bind"                           ,  0, KW_BIND    , false),
         new Keyword("bind-where"                     ,  0, KW_BIND_WH , false),
         new Keyword("blank"                          ,  0, KW_BLANK   , true ),
         new Keyword("blob"                           ,  0, KW_BLOB    , false),  // missing from keyword index and UNTESTED at this time
         new Keyword("block"                          ,  0, KW_BLOCK   , false),  // missing in keyword index, found in prog handbook as a key function
         new Keyword("block-iteration-display"        ,  0, KW_BLK_IT_D, false),
         new Keyword("block-level"                    ,  9, KW_BLK_LVL , false),
         new Keyword("border-bottom-chars"            ,  8, KW_BORD_B_C, false),
         new Keyword("border-bottom-pixels"           , 15, KW_BORD_B_P, false),
         new Keyword("border-left-chars"              ,  8, KW_BORD_L_C, false),
         new Keyword("border-left-pixels"             , 13, KW_BORD_L_P, false),
         new Keyword("border-right-chars"             ,  8, KW_BORD_R_C, false),
         new Keyword("border-right-pixels"            , 14, KW_BORD_R_P, false),
         new Keyword("border-top-chars"               ,  8, KW_BORD_T_C, false),
         new Keyword("border-top-pixels"              , 12, KW_BORD_T_P, false),
         new Keyword("both"                           ,  0, KW_BOTH    , false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("bottom"                         ,  0, KW_BOTTOM  , false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("bottom-column"                  ,  0, KW_BOTTOM_C, false),  // missing in keyword index, found in prog handbook as a key function
         new Keyword("box"                            ,  0, KW_BOX     , false),
         new Keyword("box-selectable"                 , 10, KW_BOX_SEL , false),
         new Keyword("break"                          ,  0, KW_BREAK   , true ),
         new Keyword("break-line"                     ,  0, KW_BREAK_L , false),  // missing in keyword index, found in prog handbook as a key function
         new Keyword("browse"                         ,  0, KW_BROWSE  , false),
         new Keyword("btos"                           ,  0, KW_BTOS    , true ),  // undocumented
         new Keyword("buffer"                         ,  0, KW_BUFFER  , false),
         new Keyword("buffer-chars"                   ,  0, KW_BUFFER_C, false),
         new Keyword("buffer-compare"                 , 11, KW_BUF_COMP, true ),  // this is reserved and has an abbreviation even though the documentation states it should be otherwise 
         new Keyword("buffer-copy"                    ,  0, KW_BUF_COPY, true ),  // this is reserved even though the documentation states it should be otherwise
         new Keyword("buffer-create"                  ,  0, KW_BUF_CREA, false),
         new Keyword("buffer-delete"                  ,  0, KW_BUF_DEL , false),
         new Keyword("buffer-field"                   ,  0, KW_BUF_FLD , false),
         new Keyword("buffer-handle"                  ,  0, KW_BUF_HNDL, false),
         new Keyword("buffer-lines"                   ,  0, KW_BUFFER_L, false),
         new Keyword("buffer-name"                    ,  0, KW_BUF_NAME, false),
         new Keyword("buffer-release"                 ,  0, KW_BUF_REL , false),
         new Keyword("buffer-validate"                ,  0, KW_BUF_VLID, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("buffer-value"                   ,  0, KW_BUF_VAL , false),
         new Keyword("buttons"                        ,  6, KW_BUTTON  , false),  // handles "separate keyword" button too
         new Keyword("by"                             ,  0, KW_BY      , true ),
         new Keyword("by-pointer"                     ,  0, KW_BY_PTR  , false),
         new Keyword("by-reference"                   ,  0, KW_BY_REF  , false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("by-value"                       ,  0, KW_BY_VALUE, false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("by-variant-pointer"             ,  0, KW_BY_VAR_P, false),
         new Keyword("byte"                           ,  0, KW_BYTE    , false),  // missing in keyword index, found elsewhere in external program interface ref
         new Keyword("bytes-read"                     ,  0, KW_BYTES_R , false),
         new Keyword("bytes-written"                  ,  0, KW_BYTES_W , false),
         new Keyword("cache"                          ,  0, KW_CACHE   , false),
         new Keyword("cache-size"                     ,  0, KW_CACHE_SZ, false),
         new Keyword("call"                           ,  0, KW_CALL    , true ),
         new Keyword("call-name"                      ,  0, KW_CALL_NAM, false),
         new Keyword("call-type"                      ,  0, KW_CALL_TYP, false),
         new Keyword("cancel-break"                   ,  0, KW_CANC_BRK, false),
         new Keyword("callback"                       ,  0, KW_CALLBACK, false),  // FWD TIMER extension
         new Keyword("cancel-button"                  ,  0, KW_CANCEL_B, false),
         new Keyword("cancel-pick"                    ,  0, KW_CANCEL_P, false),  // missing in keyword index, found in prog handbook as a key function
         new Keyword("cancel-requests"                ,  0, KW_CANCEL_R, false),
         new Keyword("cancelled"                      ,  0, KW_CANCELLD, false),
         new Keyword("can-create"                     ,  0, KW_CAN_CREA, false),
         new Keyword("can-delete"                     ,  0, KW_CAN_DEL , false),
         new Keyword("can-do"                         ,  0, KW_CAN_DO  , true ),
         new Keyword("can-dump"                       ,  0, KW_CAN_DUMP, true ),
         new Keyword("can-find"                       ,  0, KW_CAN_FIND, true ),
         new Keyword("can-load"                       ,  0, KW_CAN_LOAD, true ),
         new Keyword("can-query"                      ,  0, KW_CAN_QRY , false),
         new Keyword("can-read"                       ,  0, KW_CAN_READ, false),
         new Keyword("can-set"                        ,  0, KW_CAN_SET , false),
         new Keyword("can-write"                      ,  0, KW_CAN_WRT , false),
         new Keyword("upper"                          ,  0, KW_CAPS    , false),  // missing in keyword index, found elsewhere in SQL ref
         new Keyword("caps"                           ,  0, KW_CAPS    , false),
         new Keyword("careful-paint"                  ,  0, KW_CARE_PNT, false),
         new Keyword("case"                           ,  0, KW_CASE    , true ),
         new Keyword("case-sensitive"                 ,  8, KW_CASE_SEN, true ),
         new Keyword("cast"                           ,  0, KW_CAST    , true ),
         new Keyword("catch"                          ,  0, KW_CATCH   , false),
         new Keyword("cdecl"                          ,  0, KW_CDECL   , false),
         new Keyword("cease"                          ,  0, KW_CEASE   , false),  // FWD TIMER extension
         new Keyword("centered"                       ,  6, KW_CENTER  , true ),
         new Keyword("chained"                        ,  0, KW_CHAINED , false),
         new Keyword("character"                      ,  4, KW_CHAR    , false),
         new Keyword("character_length"               ,  0, KW_CHAR_LEN, false),  // possible SQL keyword in the keyword index but no reference found in the SQL 89 reference 
         new Keyword("charset"                        ,  0, KW_CHARSET , false),
         new Keyword("checked"                        ,  0, KW_CHECKED , false),
         new Keyword("check"                          ,  0, KW_CHECK   , true ),
         new Keyword("check-agent-mode"               ,  0, KW_CHECK_AM, false),  // not a real keyword, but is normally defined in the PSC webspeed 4GL code
         new Keyword("child-buffer"                   ,  0, KW_CHLD_BUF, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("choices"                        ,  0, KW_CHOICES , false),  // missing in keyword index, found in prog handbook as a key function
         new Keyword("choose"                         ,  0, KW_CHOOSE  , false),
         new Keyword("chr"                            ,  0, KW_CHR     , true ),
         new Keyword("child-num"                      ,  0, KW_CHLD_NUM, false),
         new Keyword("class"                          ,  0, KW_CLASS   , false),
         new Keyword("class-type"                     ,  0, KW_CLS_TYPE, false),
         new Keyword("clipboard"                      ,  0, KW_CLIP    , true ),
         new Keyword("clear"                          ,  0, KW_CLEAR   , true ),
         new Keyword("clear-appl-context"             ,  0, KW_CLR_AP_C, false),
         new Keyword("clear-log"                      ,  0, KW_CLR_LOG , false),
         new Keyword("clear-selection"                , 12, KW_CLR_SEL , false),
         new Keyword("clear-sort-arrows"              , 16, KW_CLR_S_AR, false),
         new Keyword("client-connection-id"           ,  0, KW_CLNT_C_I, false),
         new Keyword("client-principal"               ,  0, KW_CLNT_PRL, false),
         new Keyword("client-tty"                     ,  0, KW_CLNT_TTY, false),
         new Keyword("client-type"                    ,  0, KW_CLNT_TYP, false),
         new Keyword("client-workstation"             ,  0, KW_CLNT_WS , false),
         new Keyword("clob"                           ,  0, KW_CLOB    , false),  // missing from keyword index and UNTESTED at this time
         new Keyword("clone-node"                     ,  0, KW_CLONE_ND, false),
         new Keyword("close"                          ,  0, KW_CLOSE   , false),
         new Keyword("close-log"                      ,  0, KW_CLOSE_LG, false),
         new Keyword("closeallprocs"                  ,  0, KW_CLOSE_AP, false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("cancel-requests-after"          ,  0, KW_CNCL_R_A, false),
         new Keyword("code"                           ,  0, KW_CODE    , false),
         new Keyword("codebase-locator"               ,  0, KW_CODEBASE, true ),
         new Keyword("codepage-convert"               ,  0, KW_CP_CVT  , false),
         new Keyword("codepage"                       ,  0, KW_CP      , false),
         new Keyword("col-of"                         ,  0, KW_COL_OF  , false),
         new Keyword("collate"                        ,  0, KW_COLLATE , false),
         new Keyword("colon"                          ,  0, KW_COLON   , true ),
         new Keyword("colon-aligned"                  , 11, KW_COLON_AL, false),
         new Keyword("color"                          ,  0, KW_COLOR   , true ),
         new Keyword("color-table"                    ,  0, KW_COLR_TAB, false),
         new Keyword("column"                         ,  3, KW_COL     , false),
         new Keyword("column-bgcolor"                 ,  0, KW_COL_BGC , false),
         new Keyword("column-codepage"                ,  0, KW_COL_CP  , false),  // missing from keyword index and UNTESTED at this time
         new Keyword("column-dcolor"                  ,  0, KW_COL_DC  , false),
         new Keyword("column-fgcolor"                 ,  0, KW_COL_FGC , false),
         new Keyword("column-font"                    ,  0, KW_COL_FONT, false),
         new Keyword("column-label"                   , 10, KW_COL_LAB , true ),
         new Keyword("column-movable"                 ,  0, KW_COL_MOV , false),
         new Keyword("column-of"                      ,  0, KW_COL_OF  , false),
         new Keyword("column-pfcolor"                 ,  0, KW_COL_PFC , false),
         new Keyword("column-read-only"               ,  0, KW_COL_R_O , false),
         new Keyword("column-resizable"               ,  0, KW_COL_RES , false),
         new Keyword("column-scrolling"               ,  0, KW_COL_SCR , false),
         new Keyword("columns"                        ,  0, KW_COLUMNS , true ),
         new Keyword("command"                        ,  0, KW_COMMAND , false),
         new Keyword("com-handle"                     ,  0, KW_COM_HNDL, false),  // missing in keyword index, found elsewhere in external program interface ref
         new Keyword("com-self"                       ,  0, KW_COM_SELF, false),
         new Keyword("compare"                        ,  0, KW_COMPARE , false),
         new Keyword("compares"                       ,  0, KW_COMPARES, false),
         new Keyword("combo-box"                      ,  0, KW_COMBO_BX, false),
         new Keyword("compile"                        ,  0, KW_COMPILE , false),
         new Keyword("compiler"                       ,  0, KW_COMPILER, true ),
         new Keyword("complete"                       ,  0, KW_COMPLETE, false),
         new Keyword("component-handle"               ,  0, KW_COM_HNDL, false),
         new Keyword("config-name"                    ,  0, KW_CFG_NAME, false),
         new Keyword("connect"                        ,  0, KW_CONN    , false),
         new Keyword("connected"                      ,  0, KW_CONN_ED , true ),
         new Keyword("constructor"                    ,  0, KW_CONSTRUC, false),
         new Keyword("context"                        ,  0, KW_CTX     , false),
         new Keyword("context-help"                   ,  0, KW_CTX_H   , false),
         new Keyword("context-help-file"              ,  0, KW_CTX_H_F , false),
         new Keyword("context-help-id"                ,  0, KW_CTX_H_ID, false),
         new Keyword("context-popup"                  ,  0, KW_CTX_POP , false),
         new Keyword("contains"                       ,  0, KW_CONTAINS, false),
         new Keyword("contents"                       ,  0, KW_CONTENTS, false),
         new Keyword("control"                        ,  0, KW_CONTROL , true ),
         new Keyword("control-box"                    ,  0, KW_CNTRL_BX, false),
         new Keyword("control-container"              ,  0, KW_CNTRL_FR, false),  // missing in keyword index (and other API ref), found in customer code as a synonym 
         new Keyword("control-frame"                  ,  0, KW_CNTRL_FR, false),
         new Keyword("control-name"                   ,  0, KW_CNTRL_NM, false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("controls"                       ,  0, KW_CONTROLS, false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("convert"                        ,  0, KW_CONVERT , false),
         new Keyword("convert-datetime"               ,  0, KW_CVT_DT  , false),  // not a real keyword, but is normally defined in the PSC webspeed 4GL code
         new Keyword("convert-to-offset"              , 15, KW_CVT_2OFF, false),
         new Keyword("convert-3d-colors"              ,  0, KW_CVT_3D_C, false),
         new Keyword("cookie-date"                    ,  0, KW_COOKIE_D, false),  // not a real keyword, but is normally defined in the PSC webspeed 4GL code
         new Keyword("copy"                           ,  0, KW_COPY    , false),  // missing in keyword index, found in prog handbook as a key function
         new Keyword("copy-dataset"                   ,  0, KW_CPY_DSET, false),
         new Keyword("copy-lob"                       ,  0, KW_CPY_LOB , true ),
         new Keyword("copy-sax-attributes"            ,  0, KW_CPY_SATR, false),
         new Keyword("copy-temp-table"                ,  0, KW_CPY_TTBL, false),
         new Keyword("count"                          ,  0, KW_COUNT   , false),
         new Keyword("count-of"                       ,  0, KW_COUNT_OF, true ),
         new Keyword("coverage"                       ,  0, KW_COVERAGE, false),  // missing from keyword index, found in customer code and UNTESTED at this time
         new Keyword("cpcase"                         ,  0, KW_CPCASE  , false),
         new Keyword("cpcoll"                         ,  0, KW_CPCOLL  , false),
         new Keyword("cpinternal"                     ,  0, KW_CPINTERN, false),
         new Keyword("cplog"                          ,  0, KW_CPLOG   , false),
         new Keyword("cpprint"                        ,  0, KW_CPPRINT , false),
         new Keyword("cprcodein"                      ,  0, KW_CPRCODEI, false),
         new Keyword("cprcodeout"                     ,  0, KW_CPRCODEO, false),
         new Keyword("cpstream"                       ,  0, KW_CPSTREAM, false),
         new Keyword("cpterm"                         ,  0, KW_CPTERM  , false),
         new Keyword("crc-value"                      ,  0, KW_CRC_VAL , false),
         new Keyword("create"                         ,  0, KW_CREATE  , true ),
         new Keyword("create-like"                    ,  0, KW_CREAT_LK, false),
         new Keyword("create-on-add"                  ,  0, KW_CREAT_OA, false),  // missing from keyword index, found in customer code and UNTESTED at this time
         new Keyword("create-node"                    ,  0, KW_CREAT_ND, false),
         new Keyword("create-node-namespace"          ,  0, KW_CREAT_NN, false),
         new Keyword("create-result-list-entry"       ,  0, KW_CREAT_RL, false),
         new Keyword("create-test-file"               ,  0, KW_CREAT_TF, false),
         new Keyword("ctos"                           ,  0, KW_CTOS    , true ),  // undocumented
         new Keyword("currency"                       ,  0, KW_CURRENCY, false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("current"                        ,  0, KW_CURRENT , true ),
         new Keyword("current-changed"                ,  0, KW_CUR_CHG , true ),
         new Keyword("current-column"                 ,  0, KW_CUR_COL , false),
         new Keyword("current-environment"            , 11, KW_CUR_ENV , false),
         new Keyword("current-iteration"              ,  0, KW_CUR_ITER, false),
         new Keyword("current-language"               , 12, KW_CUR_LANG, true ),
         new Keyword("current-result-row"             ,  0, KW_CUR_RES , false),
         new Keyword("current-row-modified"           ,  0, KW_CUR_R_M , false),
         new Keyword("current-window"                 ,  0, KW_CUR_WIN , true ),
         new Keyword("current-value"                  ,  0, KW_CUR_VAL , false),
         new Keyword("current_date"                   ,  0, KW_CUR_DATE, false),  // possible SQL keyword in the keyword index but no reference found in the SQL 89 reference
         new Keyword("cursor"                         ,  4, KW_CURSOR  , true ),
         new Keyword("cursor-char"                    ,  0, KW_CUR_CHAR, false),
         new Keyword("cursor-down"                    ,  0, KW_CUR_DOWN, false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("cursor-left"                    ,  0, KW_CUR_LEFT, false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("cursor-line"                    ,  0, KW_CUR_LINE, false),
         new Keyword("cursor-offset"                  ,  0, KW_CUR_OFF , false),
         new Keyword("cursor-right"                   ,  0, KW_CUR_RGHT, false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("cursor-up"                      ,  0, KW_CUR_UP  , false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("cut"                            ,  0, KW_CUT     , false),  // missing in keyword index, found in prog handbook as a key function
         new Keyword("data-bind"                      ,  0, KW_DATA_BND, false),
         new Keyword("data-entry-return"              , 14, KW_DATA_E_R, false),
         new Keyword("data-relation"                  ,  0, KW_DATA_REL, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("data-source"                    ,  0, KW_DATA_SRC, true ),
         new Keyword("data-source-complete-map"       ,  0, KW_DATA_SCM, false),
         new Keyword("data-source-modified"           ,  0, KW_DATA_SM,  false),
         new Keyword("data-type"                      ,  6, KW_DATATYPE, false),
         new Keyword("database"                       ,  0, KW_DATABASE, true ),
         new Keyword("dataservers"                    ,  0, KW_DATASRV , true ),
         new Keyword("dataset"                        ,  0, KW_DATASET , true ),
         new Keyword("dataset-handle"                 ,  0, KW_DSET_HND, true ),
         new Keyword("date"                           ,  0, KW_DATE    , false),
         new Keyword("datetime"                       ,  0, KW_DATETIME, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("datetime-tz"                    ,  0, KW_DATE_TZ , false),  // missing from keyword index and UNTESTED at this time
         new Keyword("date-format"                    ,  6, KW_DATE_FMT, false),
         new Keyword("day"                            ,  0, KW_DAY     , false),
         new Keyword("dbcodepage"                     ,  0, KW_DBCP    , true ),
         new Keyword("dbcollation"                    ,  0, KW_DBCOLL  , true ),
         new Keyword("dbname"                         ,  0, KW_DBNAME  , true ),
         new Keyword("dbparam"                        ,  0, KW_DBPARAM , true ),
         new Keyword("db-context"                     ,  0, KW_DB_CTXT , false),
         new Keyword("db-references"                  ,  0, KW_DB_REF  , false),
         new Keyword("db-remote-host"                 ,  0, KW_DB_REM_H, false), // missing in keyword index, found elsewhere in lang ref
         new Keyword("dbrestrictions"                 ,  6, KW_DBREST  , true ),
         new Keyword("dbtaskid"                       ,  0, KW_DBTASKID, true ),
         new Keyword("dbtype"                         ,  0, KW_DBTYPE  , true ),
         new Keyword("dbversion"                      ,  6, KW_DBVERS  , true ),
         new Keyword("dcolor"                         ,  0, KW_DCOLOR  , false),
         new Keyword("dde"                            ,  0, KW_DDE     , true ),
         new Keyword("dde-error"                      ,  0, KW_DDE_ERR , false),
         new Keyword("dde-id"                         ,  5, KW_DDE_ID  , false),
         new Keyword("dde-item"                       ,  0, KW_DDE_ITEM, false),
         new Keyword("dde-name"                       ,  0, KW_DDE_NAME, false),
         new Keyword("dde-notify"                     ,  0, KW_DDE_NOTI, false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("dde-topic"                      ,  0, KW_DDE_TOPI, false),
         new Keyword("deblank"                        ,  0, KW_DEBLANK , true ),
         new Keyword("debug"                          ,  4, KW_DEBUG   , false),
         new Keyword("debugger"                       ,  0, KW_DEBUGGER, true ),
         new Keyword("debug-alert"                    ,  0, KW_DBG_ALRT, false),
         new Keyword("debug-list"                     ,  0, KW_DBG_LST , true ),
         new Keyword("decimal"                        ,  3, KW_DEC     , false),
         new Keyword("decimals"                       ,  0, KW_DECIMALS, true ),
         new Keyword("declare"                        ,  0, KW_DECLARE , true ),
         new Keyword("declare-namespace"              ,  0, KW_DECL_NSP, false),
         new Keyword("decrypt"                        ,  0, KW_DECRYPT , false),
         new Keyword("default"                        ,  0, KW_DEFAULT , true ),
         new Keyword("default-action"                 ,  0, KW_DEF_ACTN, false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("default-buffer-handle"          ,  0, KW_DEF_BUFH, false),
         new Keyword("default-button"                 ,  9, KW_DEFLT_BN, false),
         new Keyword("default-commit"                 ,  0, KW_DEF_COMM, false),
         new Keyword("default-extension"              , 10, KW_DEF_EXTN, false),
         new Keyword("default-noxlate"                , 12, KW_DEF_NOXL, true ),
         new Keyword("default-pop-up"                 ,  0, KW_DEF_POP , false),  // missing in keyword index, found in prog handbook as a key function
         new Keyword("default-string"                 ,  0, KW_DEF_STR , false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("default-value"                  ,  0, KW_DEF_VAL , false),
         new Keyword("default-window"                 ,  0, KW_DEF_WIN , true ),
         new Keyword("define"                         ,  3, KW_DEFINE  , true ),
         new Keyword("define-user-event-manager"      ,  0, KW_DEF_UEVM, false),
         new Keyword("delegate"                       ,  0, KW_DELEGATE, false),
         new Keyword("delete"                         ,  0, KW_DELETE  , true ),
         new Keyword("delete-character"               , 11, KW_DEL_CHAR, false),
         new Keyword("delete-column"                  ,  0, KW_DEL_COL , false),  // missing in keyword index, found in prog handbook as a key function
         new Keyword("deletecookie"                   ,  0, KW_DEL_COOK, false),  // not a real keyword, but is normally defined in the PSC webspeed 4GL code
         new Keyword("delete-cookie"                  ,  0, KW_DEL_COOK, false),  // not a real keyword, but is normally defined in the PSC webspeed 4GL code
         new Keyword("delete-current-row"             ,  0, KW_DEL_C_R , false),
         new Keyword("delete-end-line"                ,  0, KW_DEL_E_L , false),  // missing in keyword index, found in prog handbook as a key function
         new Keyword("delete-field"                   ,  0, KW_DEL_FLD , false),  // missing in keyword index, found in prog handbook as a key function
         new Keyword("delete-header-entry"            ,  0, KW_DEL_H_EN, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("delete-line"                    ,  0, KW_DEL_LINE, false),
         new Keyword("delete-node"                    ,  0, KW_DEL_NODE, false),
         new Keyword("delete-result-list-entry"       ,  0, KW_DEL_R_L , false),
         new Keyword("delete-selected-row"            ,  0, KW_DEL_S_R , false),
         new Keyword("delete-selected-rows"           ,  0, KW_DEL_S_RS, false),
         new Keyword("delete-word"                    ,  0, KW_DEL_WORD, false),  // missing in keyword index, found in prog handbook as a key function
         new Keyword("descending"                     ,  4, KW_DESCEND , true ),  // handles "separate keyword" desc too
         new Keyword("description"                    ,  0, KW_DESCR   , false),  // missing in keyword index, found in customer code
         new Keyword("deselect-focused-row"           ,  0, KW_DESEL_FR, false),
         new Keyword("deselect-rows"                  ,  0, KW_DESEL_R , false),
         new Keyword("deselect-selected-row"          ,  0, KW_DESEL_SR, false),
         new Keyword("deselection"                    ,  0, KW_DESELCTN, false),
         new Keyword("destructor"                     ,  0, KW_DESTRUCT, false),
         new Keyword("detach-data-source"             ,  0, KW_DET_DSRC, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("dialog-box"                     ,  0, KW_DIALOG  , false),
         new Keyword("dictionary"                     ,  4, KW_DICT    , true ),
         new Keyword("dir"                            ,  0, KW_DIR     , false),
         new Keyword("directory"                      ,  0, KW_DIRECTRY, false),  // missing in keyword index, found in customer code
         new Keyword("delimiter"                      ,  0, KW_DELIMIT , true ),
         new Keyword("disable"                        ,  0, KW_DISABLE , true ),
         new Keyword("disable-auto-zap"               ,  0, KW_DIS_A_ZA, true ),
         new Keyword("disable-connections"            ,  0, KW_DISABL_C, false),
         new Keyword("disable-dump-triggers"          ,  0, KW_DIS_D_TR, false),
         new Keyword("disable-load-triggers"          ,  0, KW_DIS_L_TR, false),
         new Keyword("disable-redraw"                 ,  0, KW_DIS_REDR, false), // FWD extension, not real 4GL!
         new Keyword("disabled"                       ,  0, KW_DISABLED, false),
         new Keyword("disconnect"                     ,  6, KW_DISCONN , true ),
         new Keyword("dispatch"                       ,  0, KW_DISPATCH, false),  // not a real keyword, but is normally defined in the PSC webspeed 4GL code
         new Keyword("display"                        ,  4, KW_DISP    , true ),
         new Keyword("display-message"                ,  0, KW_DISP_MSG, false),
         new Keyword("display-timezone"               ,  0, KW_DISP_TZ , false),  // missing from keyword index and UNTESTED at this time
         new Keyword("display-type"                   ,  9, KW_DISP_TYP, false),
         new Keyword("distinct"                       ,  0, KW_DISTINCT, true ),
         new Keyword("do"                             ,  0, KW_DO      , true ),
         new Keyword("domain-description"             ,  0, KW_DOMAIN_D, false),
         new Keyword("domain-name"                    ,  0, KW_DOMAIN_N, false),
         new Keyword("domain-type"                    ,  0, KW_DOMAIN_T, false),
         new Keyword("dos"                            ,  0, KW_DOS     , true ),
         new Keyword("double"                         ,  0, KW_DOUBLE  , false),
         new Keyword("down"                           ,  0, KW_DOWN    , true ),
         new Keyword("drag-enabled"                   ,  0, KW_DRAG_EN , false),
         new Keyword("drop"                           ,  0, KW_DROP    , true ),
         new Keyword("drop-down"                      ,  0, KW_DROP_DWN, false),
         new Keyword("drop-down-list"                 ,  0, KW_DROP_LST, false),
         new Keyword("drop-file-notify"               ,  0, KW_DROP_FN , false),
         new Keyword("drop-target"                    ,  0, KW_DROP_TAR, false),
         new Keyword("dslog-manager"                  ,  0, KW_DSL_MGR , false),
         new Keyword("dump"                           ,  0, KW_DUMP    , false),
         new Keyword("dump-logging-now"               ,  0, KW_DUMP_LGN, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("dynamic-cast"                   ,  0, KW_DYN_CAST, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("dynamic-current-value"          ,  0, KW_DYN_CURV, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("dynamic-function"               , 12, KW_DYN_FUNC, true ),  // undocumented abbreviation found in customer code
         new Keyword("dynamic-invoke"                 ,  0, KW_DYN_INVK, true ),  // missing from keyword index and UNTESTED at this time
         new Keyword("dynamic-new"                    ,  0, KW_DYN_NEW , false),  // missing from keyword index and UNTESTED at this time
         new Keyword("dynamic-next-value"             ,  0, KW_DYN_NEXV, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("dynamic"                        ,  0, KW_DYNAMIC , false),
         new Keyword("each"                           ,  0, KW_EACH    , true ),
         new Keyword("echo"                           ,  0, KW_ECHO    , false),
         new Keyword("edge-chars"                     ,  4, KW_EDGE_C  , false),
         new Keyword("edge-pixels"                    ,  6, KW_EDGE_P  , false),
         new Keyword("edit-can-paste"                 ,  0, KW_EDIT_C_P, false),
         new Keyword("edit-can-undo"                  ,  0, KW_EDIT_C_U, false),
         new Keyword("edit-clear"                     ,  0, KW_EDIT_CLR, false),
         new Keyword("edit-copy"                      ,  0, KW_EDIT_CPY, false),
         new Keyword("edit-cut"                       ,  0, KW_EDIT_CUT, false),
         new Keyword("edit-paste"                     ,  0, KW_EDIT_PAS, false),
         new Keyword("edit-undo"                      ,  0, KW_EDIT_UND, false),
         new Keyword("editing"                        ,  0, KW_EDITING , true ),
         new Keyword("editor"                         ,  0, KW_EDITOR  , false),
         new Keyword("editor-backtab"                 ,  0, KW_EDIT_B_T, false),  // missing in keyword index, found in prog handbook as a key function
         new Keyword("editor-tab"                     ,  0, KW_EDIT_TAB, false),  // missing in keyword index, found elsewhere in prog handbook
         new Keyword("else"                           ,  0, KW_ELSE    , true ),
         new Keyword("empty"                          ,  0, KW_EMPTY   , false),
         new Keyword("empty-dataset"                  ,  0, KW_EMPTY_DS, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("empty-selection"                ,  0, KW_EMPTY_SN, false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("empty-temp-table"               ,  0, KW_EMPTY_TT, false),
         new Keyword("enable"                         ,  0, KW_ENABLE  , true ),
         new Keyword("enabled"                        ,  0, KW_ENABLED , false),  // FWD TIMER extension
         new Keyword("enable-connections"             ,  0, KW_ENABLE_C, false),
         new Keyword("enable-events"                  ,  0, KW_ENABLE_E, false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("encode"                         ,  0, KW_ENCODE  , true ),
         new Keyword("encoding"                       ,  0, KW_ENCODING, false),
         new Keyword("encrypt"                        ,  0, KW_ENCRYPT , false),
         new Keyword("encrypt-audit-mac-key"          ,  0, KW_ENC_AMK , false),
         new Keyword("encryption-salt"                ,  0, KW_ENC_SALT, false),
         new Keyword("end-box-selection"              ,  0, KW_END_B_SN, false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("end-document"                   ,  0, KW_END_DOC , false),
         new Keyword("end-element"                    ,  0, KW_END_ELEM, false),
         new Keyword("end-event-group"                ,  0, KW_END_EV_G, false),
         new Keyword("end-search"                     ,  0, KW_END_SEAR, false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("end-user-prompt"                ,  0, KW_END_USER, false),
         new Keyword("end"                            ,  0, KW_END     , true ),
         new Keyword("end-error"                      ,  0, KW_END_ERR , false),
         new Keyword("end-file-drop"                  ,  0, KW_END_F_D , false),
         new Keyword("endkey"                         ,  0, KW_ENDKEY  , false),
         new Keyword("end-key"                        ,  0, KW_ENDKEY  , false),
         new Keyword("end-move"                       ,  0, KW_END_MOV , false),
         new Keyword("end-resize"                     ,  0, KW_END_RESZ, false),
         new Keyword("end-row-resize"                 ,  0, KW_END_RRES, false),
         new Keyword("entered"                        ,  0, KW_ENTERED , false),
         new Keyword("enter-menubar"                  ,  0, KW_ENTER_MB, false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("entry"                          ,  0, KW_ENTRY   , true ),
         new Keyword("entry-types-list"               ,  0, KW_ENT_TLST, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("error"                          ,  0, KW_ERROR   , false),
         new Keyword("error-code"                     ,  0, KW_ERR_CODE, false),  // missing in keyword index, found elsewhere in external program interface ref
         new Keyword("error-column"                   ,  9, KW_ERR_COL , false),
         new Keyword("error-object-detail"            ,  0, KW_ERR_OBJD, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("error-row"                      ,  0, KW_ERR_ROW,  false),
         new Keyword("error-stack-trace"              ,  0, KW_ERR_S_T ,  false),
         new Keyword("error-status"                   , 10, KW_ERR_STAT, true ),
         new Keyword("error-string"                   ,  0, KW_ERR_STR,  false),  // missing from keyword index and UNTESTED at this time
         new Keyword("eq"                             ,  0, KW_EQ      , false),
         new Keyword("escape"                         ,  0, KW_ESCAPE  , true ),
         new Keyword("etime"                          ,  0, KW_ETIME   , true ),
         new Keyword("events"                         ,  5, KW_EVENTS  , false),
         new Keyword("event-group-id"                 ,  0, KW_EVT_GRID, false),
         new Keyword("event-procedure"                ,  0, KW_EVT_PROC, true ),
         new Keyword("event-procedure-context"        ,  0, KW_E_PROC_C, false),
         new Keyword("event-type"                     ,  7, KW_EVT_TYPE, false),
         new Keyword("except"                         ,  0, KW_EXCEPT  , true ),
         new Keyword("exclusive-id"                   ,  0, KW_EXCL_ID , false),
         new Keyword("exclusive-lock"                 ,  9, KW_EXC_LOCK, true ),
         new Keyword("exclusive-web-user"             ,  0, KW_EXCL_WEB, false),
         new Keyword("execute"                        ,  0, KW_EXECUTE , false),
         new Keyword("execution-log"                  ,  0, KW_EXEC_LOG, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("exists"                         ,  0, KW_EXISTS  , true ),
         new Keyword("exit"                           ,  0, KW_EXIT    , false),  // missing in keyword index, found in prog handbook as a key function
         new Keyword("exp"                            ,  0, KW_EXP     , false),
         new Keyword("expand"                         ,  0, KW_EXPAND  , false),
         new Keyword("expandable"                     ,  0, KW_EXPANDBL, false),
         new Keyword("explicit"                       ,  0, KW_EXPLICIT, false),
         new Keyword("export"                         ,  0, KW_EXPORT  , true ),
         new Keyword("export-principal"               ,  0, KW_EXPORT_P, false),
         new Keyword("extended"                       ,  0, KW_EXTENDED, false),
         new Keyword("extent"                         ,  0, KW_EXTENT  , false),
         new Keyword("external"                       ,  0, KW_EXTERN  , false),
         new Keyword("false"                          ,  0, BOOL_FALSE , true ),
         new Keyword("fetch"                          ,  0, KW_FETCH   , true ),
         new Keyword("fetch-selected-row"             ,  0, KW_FETCH_SR, false),
         new Keyword("fields"                         ,  5, KW_FIELD   , true ),  // handles "separate keyword" field too
         new Keyword("file"                           ,  0, KW_FILE    , false),
         new Keyword("file-create-date"               ,  0, KW_FIL_C_D , false),
         new Keyword("file-create-time"               ,  0, KW_FIL_C_T , false),
         new Keyword("file-information"               ,  9, KW_FIL_INFO, true ),
         new Keyword("file-mod-date"                  ,  0, KW_FIL_M_D , false),
         new Keyword("file-mod-time"                  ,  0, KW_FIL_M_T , false),
         new Keyword("file-name"                      ,  0, KW_FIL_NAME, false),
         new Keyword("file-offset"                    ,  8, KW_FIL_OFF,  false),
         new Keyword("file-size"                      ,  0, KW_FIL_SIZE, false),
         new Keyword("file-type"                      ,  0, KW_FIL_TYPE, false),
         new Keyword("filename"                       ,  0, KW_FIL_NAME, false),
         new Keyword("fill"                           ,  0, KW_FILL    , true ),
         new Keyword("fill-in"                        ,  0, KW_FILL_IN , false),
         new Keyword("fill-mode"                      ,  0, KW_FILL_MOD, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("fill-where-string"              ,  0, KW_FILL_WST, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("filled"                         ,  0, KW_FILLED  , false),
         new Keyword("filters"                        ,  0, KW_FILTERS , false),
         new Keyword("final"                          ,  0, KW_FINAL   , false),
         new Keyword("finally"                        ,  0, KW_FINALLY , false),
         new Keyword("find"                           ,  0, KW_FIND    , true ),
         new Keyword("find-by-rowid"                  ,  0, KW_FIND_BR , false),
         new Keyword("find-case-sensitive"            ,  0, KW_FIND_CS , true ),
         new Keyword("find-current"                   ,  0, KW_FIND_CUR, false),
         new Keyword("find-failed"                    ,  0, KW_FIND_FD , false),  // missing from keyword index and UNTESTED at this time
         new Keyword("find-first"                     ,  0, KW_FIND_1ST, false),
         new Keyword("find-global"                    ,  0, KW_FIND_GLO, true ),
         new Keyword("find-last"                      ,  0, KW_FIND_LST, false),
         new Keyword("find-next"                      ,  0, KW_FIND_NXT, false),  // missing in keyword index, found in prog handbook as a key function
         new Keyword("find-next-occurrence"           ,  0, KW_FIND_NO , true ),
         new Keyword("find-previous"                  ,  0, KW_FIND_PRV, false),  // missing in keyword index, found in prog handbook as a key function
         new Keyword("find-prev-occurrence"           ,  0, KW_FIND_PO , true ),
         new Keyword("find-select"                    ,  0, KW_FIND_SEL, true ),
         new Keyword("find-unique"                    ,  0, KW_FIND_UNI, false),
         new Keyword("find-wrap-around"               ,  0, KW_FIND_WA , true ),
         new Keyword("finder"                         ,  0, KW_FINDER  , false),
         new Keyword("first"                          ,  0, KW_FIRST   , true ),
         new Keyword("first-async-request"            ,  0, KW_FIRST_AR, false),
         new Keyword("first-buffer"                   ,  0, KW_FIRST_BU, false),
         new Keyword("first-child"                    ,  0, KW_FIRST_CH, false),
         new Keyword("first-column"                   ,  0, KW_FIRST_CO, false),
         new Keyword("first-data-source"              ,  0, KW_FIRST_DS, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("first-dataset"                  ,  0, KW_FIR_DSET, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("first-form"                     ,  0, KW_FIRST_FM, false),
         new Keyword("first-of"                       ,  0, KW_FIRST_OF, true ),
         new Keyword("first-object"                   ,  0, KW_FIRST_OB, false),
         new Keyword("first-procedure"                , 10, KW_FIRST_PR, false),
         new Keyword("first-query"                    ,  0, KW_FIRST_QR, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("first-server"                   ,  0, KW_FIRST_SR, false),
         new Keyword("first-server-socket"            ,  0, KW_FIRST_SS, false),
         new Keyword("first-socket"                   ,  0, KW_FIRST_SO, false),
         new Keyword("first-tab-item"                 , 11, KW_FIRST_TI, false),
         new Keyword("fit-last-column"                ,  0, KW_FIT_LCOL, false),
         new Keyword("fix-codepage"                   ,  0, KW_FIX_CP  , false),  // missing from keyword index and UNTESTED at this time
         new Keyword("fixed-only"                     ,  0, KW_FIXD_ONL, false),
         new Keyword("fgcolor"                        ,  3, KW_FGCOLOR , false),
         new Keyword("flat-button"                    ,  0, KW_FLAT_BUT, false),
         new Keyword("float"                          ,  0, KW_FLOAT   , false),
         new Keyword("focus"                          ,  0, KW_FOCUS   , true ),
         new Keyword("focused-row"                    ,  0, KW_FOCUS_R , false),
         new Keyword("focused-row-selected"           ,  0, KW_FOCUS_RS, false),
         new Keyword("font"                           ,  0, KW_FONT    , true ),
         new Keyword("font-table"                     ,  0, KW_FONT_TAB, false),
         new Keyword("for"                            ,  0, KW_FOR     , true ),
         new Keyword("force-file"                     ,  0, KW_FORCE_F , false),
         new Keyword("foreground"                     ,  4, KW_FORE    , false),
         new Keyword("foreign-key-hidden"             ,  0, KW_F_KEY_H , false),  // missing from keyword index, found elsewhere in lang ref, TESTED
         new Keyword("form-input"                     ,  0, KW_FORM_INP, false),
         new Keyword("form-long-input"                ,  0, KW_FORM_LIN, false),
         new Keyword("format"                         ,  4, KW_FORMAT  , true ),  // handles "separate keyword" form too
         new Keyword("format-datetime"                ,  0, KW_FMT_DT  , false),  // not a real keyword, but is normally defined in the PSC webspeed 4GL code
         new Keyword("formatted"                      ,  8, KW_FORMATTE, false),
         new Keyword("forward-only"                   ,  0, KW_FWD_ONLY, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("forwards"                       ,  7, KW_FORWARD , false),  // handles "separate keyword" forward too
         new Keyword("fragment"                       ,  7, KW_FRAGMENT, false),
         new Keyword("frame"                          ,  4, KW_FRAME   , true ),
         new Keyword("frame-col"                      ,  0, KW_FR_COL  , true ),
         new Keyword("frame-db"                       ,  0, KW_FR_DB   , true ),
         new Keyword("frame-down"                     ,  0, KW_FR_DOWN , true ),
         new Keyword("frame-field"                    ,  0, KW_FR_FIELD, true ),
         new Keyword("frame-file"                     ,  0, KW_FR_FILE , true ),
         new Keyword("frame-index"                    , 10, KW_FR_INDEX, true ),
         new Keyword("frame-line"                     ,  0, KW_FR_LINE , true ),
         new Keyword("frame-name"                     ,  0, KW_FR_NAME , true ),
         new Keyword("frame-row"                      ,  0, KW_FR_ROW  , true ),
         new Keyword("frame-spacing"                  ,  9, KW_FR_SPACE, false),
         new Keyword("frame-value"                    ,  9, KW_FR_VAL  , true ),
         new Keyword("frame-x"                        ,  0, KW_FR_X    , false),
         new Keyword("frame-y"                        ,  0, KW_FR_Y    , false),
         new Keyword("from"                           ,  0, KW_FROM    , true ),
         new Keyword("from-current"                   ,  8, KW_FROM_CUR, false),
         new Keyword("from-chars"                     ,  6, KW_FROM_CHR, true ),
         new Keyword("from-pixels"                    ,  6, KW_FROM_PIX, true ),
         new Keyword("frequency"                      ,  0, KW_FREQ    , false),
         new Keyword("full-height-chars"              , 11, KW_FULL_H_C, false),
         new Keyword("full-height-pixels"             , 13, KW_FULL_H_P, false),
         new Keyword("full-width-chars"               , 10, KW_FULL_W_C, false),
         new Keyword("full-width-pixels"              , 12, KW_FULL_W_P, false),
         new Keyword("full-pathname"                  , 10, KW_FULLPATH, false),
         new Keyword("function"                       ,  0, KW_FUNCT   , false),
         new Keyword("function-call-type"             ,  0, KW_FUNC_C_T, true ),
         new Keyword("gateways"                       ,  7, KW_GW      , true ),
         new Keyword("ge"                             ,  0, KW_GTE     , false),
         new Keyword("generate-md5"                   ,  0, KW_GEN_MD5 , false),
         new Keyword("generate-pbe-key"               ,  0, KW_GEN_PBEK, false),
         new Keyword("generate-pbe-salt"              ,  0, KW_GEN_PBES, false),
         new Keyword("generate-random-key"            ,  0, KW_GEN_RNDK, false),
         new Keyword("generate-uuid"                  ,  0, KW_GEN_UUID, false),
         new Keyword("get"                            ,  0, KW_GET     , false),
         new Keyword("get-attr-call-type"             ,  0, KW_GET_A_CT, true ),
         new Keyword("getattribute"                   ,  0, KW_GETATTRI, false),  // not a real keyword, but is normally defined in the PSC webspeed 4GL code
         new Keyword("get-attribute"                  ,  0, KW_GET_ATTR, false),
         new Keyword("get-attribute-node"             ,  0, KW_GET_A_N , false),
         new Keyword("get-binary-data"                ,  0, KW_GET_BDAT, false),
         new Keyword("get-bits"                       ,  0, KW_GET_BITS, false),
         new Keyword("get-blue-value"                 ,  8, KW_GET_BLUE, false),
         new Keyword("get-browse-column"              ,  0, KW_GET_BR_C, false),
         new Keyword("get-buffer-handle"              ,  0, KW_GET_BUFH, true ),
         new Keyword("getbyte"                        ,  0, KW_GET_BYTE, true ),
         new Keyword("get-byte"                       ,  0, KW_GET_BYTE, true ),
         new Keyword("get-byte-order"                 ,  0, KW_GET_B_OR, false),
         new Keyword("get-bytes"                      ,  0, KW_GET_BYTS, false),
         new Keyword("get-bytes-available"            ,  0, KW_GET_B_A , false),
         new Keyword("get-callback-proc-context"      ,  0, KW_GET_CBPC, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("get-callback-proc-name"         ,  0, KW_GET_CBPN, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("getcgi"                         ,  0, KW_GET_CGI , false),  // not a real keyword, but is normally defined in the PSC webspeed 4GL code
         new Keyword("get-cgi"                        ,  0, KW_GET_CGI , false),  // not a real keyword, but is normally defined in the PSC webspeed 4GL code
         new Keyword("get-cgi-list"                   ,  0, KW_GET_CGIL, false),
         new Keyword("get-cgi-long"                   ,  0, KW_GET_CGL , false),  // not a real keyword, but is normally defined in the PSC webspeed 4GL code
         new Keyword("get-cgi-long-value"             ,  0, KW_GET_CGLV, false),
         new Keyword("get-cgi-value"                  ,  0, KW_GET_CGIV, false),
         new Keyword("get-changes"                    ,  0, KW_GET_CHG , false),  // missing from keyword index and UNTESTED at this time
         new Keyword("get-child"                      ,  0, KW_GET_CHLD, false),
         new Keyword("get-child-relation"             , 13, KW_GET_CREL, false),  // missing from keyword index, manually tested abbrev and reserved status
         new Keyword("get-codepage"                   ,  0, KW_GET_CODP, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("get-codepages"                  ,  0, KW_GET_CP  , true ),
         new Keyword("get-collation"                  ,  0, KW_GET_CLL , false),  // missing from keyword index and UNTESTED at this time
         new Keyword("get-collations"                 ,  0, KW_GET_COLL, true ),
         new Keyword("getcookie"                      ,  0, KW_GET_COOK, false),  // not a real keyword, but is normally defined in the PSC webspeed 4GL code
         new Keyword("get-cookie"                     ,  0, KW_GET_COOK, false),  // not a real keyword, but is normally defined in the PSC webspeed 4GL code
         new Keyword("get-config"                     ,  0, KW_GET_CFG , false),  // not a real keyword, but is normally defined in the PSC webspeed 4GL code
         new Keyword("get-config-value"               ,  0, KW_GET_CFGV, false),
         new Keyword("get-current"                    ,  0, KW_GET_CUR , false),
         new Keyword("get-dataset-buffer"             ,  0, KW_GET_DS_B, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("get-dir"                        ,  0, KW_GET_DIR , false),  // missing from keyword index and UNTESTED at this time
         new Keyword("get-document-element"           ,  0, KW_GET_D_E , false),
         new Keyword("get-double"                     ,  0, KW_GET_DBL , false),
         new Keyword("get-dropped-file"               ,  0, KW_GET_D_F , false),
         new Keyword("get-dynamic"                    ,  0, KW_GET_DYN , false),
         new Keyword("get-error-column"               ,  0, KW_GET_ERRC, true ),
         new Keyword("get-error-row"                  ,  0, KW_GET_ERRR, true ),
         new Keyword("getfield"                       ,  0, KW_GET_FLD , false),  // not a real keyword, but is normally defined in the PSC webspeed 4GL code
         new Keyword("get-field"                      ,  0, KW_GET_FLD , false),  // not a real keyword, but is normally defined in the PSC webspeed 4GL code
         new Keyword("get-file"                       ,  0, KW_GET_FILE, false),
         new Keyword("get-file-name"                  ,  0, KW_GET_FNAM, true ),
         new Keyword("get-file-offset"                , 14, KW_GET_FOFF, true ),
         new Keyword("get-first"                      ,  0, KW_GET_1ST , false),
         new Keyword("get-float"                      ,  0, KW_GET_FLT , false),
         new Keyword("get-green-value"                ,  9, KW_GET_GRN , false),
         new Keyword("get-header-entry"               ,  0, KW_GET_HD_E, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("get-index-by-namespace-name"    ,  0, KW_GET_IBNN, false),
         new Keyword("get-index-by-qname"             ,  0, KW_GET_IBQN, false),
         new Keyword("get-int64"                      ,  0, KW_GET_I64 , false),
         new Keyword("get-iteration"                  ,  0, KW_GET_ITER, false),
         new Keyword("get-key-value"                  , 11, KW_GET_K_V , true ),
         new Keyword("get-last"                       ,  0, KW_GET_LAST, false),
         new Keyword("get-localname-by-index"         ,  0, KW_GET_LNBI, false),
         new Keyword("get-long"                       ,  0, KW_GET_LONG, false),
         new Keyword("get-long-value"                 ,  0, KW_GET_L_V , false),  // not a real keyword, but is normally defined in the PSC webspeed 4GL code
         new Keyword("get-mouse-position"             ,  0, KW_GET_MOP,  false),  // FWD extension, not a real 4GL
         new Keyword("get-message"                    ,  0, KW_GET_MSG , false),
         new Keyword("get-message-groups"             ,  0, KW_GET_MSGG, false),  // not a real keyword, but is normally defined in the PSC webspeed 4GL code
         new Keyword("get-messages"                   ,  0, KW_GET_MSGS, false),  // not a real keyword, but is normally defined in the PSC webspeed 4GL code
         new Keyword("get-next"                       ,  0, KW_GET_NEXT, false),
         new Keyword("get-node"                       ,  0, KW_GET_NODE, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("get-number"                     ,  0, KW_GET_NUM , false),
         new Keyword("get-parent"                     ,  0, KW_GET_PAR , false),
         new Keyword("get-pointer-value"              ,  0, KW_GET_PTR , false),
         new Keyword("get-prev"                       ,  0, KW_GET_PREV, false),
         new Keyword("get-printers"                   ,  0, KW_GET_PRT , false),
         new Keyword("get-property"                   ,  0, KW_GET_PROP, false),
         new Keyword("get-qname-by-index"             ,  0, KW_GET_QNBI, false),
         new Keyword("get-red-value"                  ,  7, KW_GET_RED , false),
         new Keyword("get-relation"                   ,  0, KW_GET_REL , false),  // missing from keyword index and UNTESTED at this time
         new Keyword("get-repositioned-row"           ,  0, KW_GET_R_R , false),
         new Keyword("get-rgb-value"                  ,  0, KW_GET_RGB , false),
         new Keyword("get-short"                      ,  0, KW_GET_SHRT, false),
         new Keyword("get-selected-widget"            , 12, KW_GET_SELW, false),
         new Keyword("get-serialized"                 ,  0, KW_GET_SER , false),  // missing from keyword index and UNTESTED at this time
         new Keyword("get-signature"                  ,  0, KW_GET_SIG , false),
         new Keyword("get-size"                       ,  0, KW_GET_SZ  , false),
         new Keyword("get-socket-option"              ,  0, KW_GET_S_O , false),
         new Keyword("get-source-buffer"              ,  0, KW_GET_SRCB, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("get-string"                     ,  0, KW_GET_STR , false),
         new Keyword("get-tab-item"                   ,  0, KW_GET_TI  , false),
         new Keyword("get-text-height-chars"          , 15, KW_GET_THCH, false),
         new Keyword("get-text-height-pixels"         , 17, KW_GET_THPX, false),
         new Keyword("get-text-width-chars"           , 14, KW_GET_TWCH, false),
         new Keyword("get-text-width-pixels"          , 16, KW_GET_TWPX, false),
         new Keyword("get-top-buffer"                 ,  0, KW_GET_TOPB, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("get-transaction-state"          ,  0, KW_GET_T_S , false),  // not a real keyword, but is normally defined in the PSC webspeed 4GL code
         new Keyword("get-type-by-index"              ,  0, KW_GET_TBI , false),
         new Keyword("get-type-by-namespace-name"     ,  0, KW_GET_TBNN, false),
         new Keyword("get-type-by-qname"              ,  0, KW_GET_TBQN, false),
         new Keyword("get-unsigned-short"             ,  0, KW_GET_USHT, false),
         new Keyword("get-unsigned-long"              ,  0, KW_GET_UL  , false),
         new Keyword("get-uri-by-index"               ,  0, KW_GET_UBI , false),
         new Keyword("get-user-field"                 ,  0, KW_GET_U_F , false),  // not a real keyword, but is normally defined in the PSC webspeed 4GL code
         new Keyword("get-value"                      ,  0, KW_GET_VAL , false),  // not a real keyword, but is normally defined in the PSC webspeed 4GL code
         new Keyword("get-value-by-index"             ,  0, KW_GET_VBI , false),
         new Keyword("get-value-by-namespace-name"    ,  0, KW_GET_VBNN, false),
         new Keyword("get-value-by-qname"             ,  0, KW_GET_VBQN, false),
         new Keyword("get-wait-state"                 ,  0, KW_GET_WAIT, false),
         new Keyword("global"                         ,  0, KW_GLOBAL  , true ),
         new Keyword("gt"                             ,  0, KW_GT      , false),
         new Keyword("go"                             ,  0, KW_GO      , false),
         new Keyword("go-on"                          ,  0, KW_GO_ON   , true ),
         new Keyword("go-pending"                     ,  7, KW_GO_PEND , true ),
         new Keyword("go-to"                          ,  0, KW_GOTO    , false),  // missing in keyword index, found in prog handbook as a key function
         new Keyword("grant"                          ,  0, KW_GRANT   , true ),
         new Keyword("graphic-edge"                   ,  9, KW_GRAPHIC , true ),
         new Keyword("grid-factor-horizontal"         , 13, KW_GRD_F_H , false),
         new Keyword("grid-factor-vertical"           , 13, KW_GRD_F_V , false),
         new Keyword("grid-snap"                      ,  0, KW_GRD_SNAP, false),
         new Keyword("grid-unit-height-chars"         , 16, KW_GRD_UHC , false),
         new Keyword("grid-unit-height-pixels"        , 18, KW_GRD_UHP , false),
         new Keyword("grid-unit-width-chars"          , 15, KW_GRD_UWC , false),
         new Keyword("grid-unit-width-pixels"         , 17, KW_GRD_UWP , false),
         new Keyword("grid-visible"                   ,  0, KW_GRD_VIS , false),
         new Keyword("group"                          ,  0, KW_GROUP   , true ),
         new Keyword("group-box"                      ,  0, KW_GROUP_BX, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("guid"                           ,  0, KW_GUID    , false),
         new Keyword("handle"                         ,  0, KW_HANDLE  , false),
         new Keyword("handler"                        ,  0, KW_HANDLER , false),
         new Keyword("has-lobs"                       ,  0, KW_HAS_LOBS, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("has-records"                    ,  0, KW_HAS_REC , false),
         new Keyword("having"                         ,  0, KW_HAVING  , true ),
         new Keyword("header"                         ,  0, KW_HEADER  , true ),
         new Keyword("height-chars"                   ,  6, KW_HEIGHT_C, false),
         new Keyword("height-pixels"                  ,  8, KW_HEIGHT_P, false),
         new Keyword("help"                           ,  0, KW_HELP    , true ),
         new Keyword("help-topic"                     ,  0, KW_HELP_TOP, false),
         new Keyword("hex-decode"                     ,  0, KW_HEX_DECD, false),
         new Keyword("hex-encode"                     ,  0, KW_HEX_ENCD, false),
         new Keyword("hide"                           ,  0, KW_HIDE    , true ),
         new Keyword("hidden-field"                   ,  0, KW_HID_FLD , false),  // not a real keyword, but is normally defined in the PSC webspeed 4GL code
         new Keyword("hidden-field-list"              ,  0, KW_HID_FLDL, false),  // not a real keyword, but is normally defined in the PSC webspeed 4GL code
         new Keyword("hidden"                         ,  0, KW_HIDDEN  , false),
         new Keyword("hint"                           ,  0, KW_HINT    , false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("home"                           ,  0, KW_HOME    , false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("horizontal"                     ,  4, KW_HORIZ   , false),
         new Keyword("host-byte-order"                ,  0, KW_HOST_B_O, true ),
         new Keyword("html-charset"                   ,  0, KW_HTML_CHS, false),
         new Keyword("html-encode"                    ,  0, KW_HTML_ENC, false),  // not a real keyword, but is normally defined in the PSC webspeed 4GL code
         new Keyword("html-end-of-line"               ,  0, KW_HTML_EOL, false),
         new Keyword("html-end-of-page"               ,  0, KW_HTML_EOP, false),
         new Keyword("html-frame-begin"               ,  0, KW_HTML_FRB, false),
         new Keyword("html-frame-end"                 ,  0, KW_HTML_FRE, false),
         new Keyword("html-header-begin"              ,  0, KW_HTML_H_B, false),
         new Keyword("html-header-end"                ,  0, KW_HTML_H_E, false),
         new Keyword("html-title-begin"               ,  0, KW_HTML_T_B, false),
         new Keyword("html-title-end"                 ,  0, KW_HTML_T_E, false),
         new Keyword("htmlerror"                      ,  0, KW_HTMLERR , false),  // not a real keyword, but is normally defined in the PSC webspeed 4GL code
         new Keyword("hwnd"                           ,  0, KW_HWND    , false),
         new Keyword("honorprokeys"                   ,  0, KW_HONOR_PK, false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("honorreturnkey"                 ,  0, KW_HONOR_RK, false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("icon"                           ,  0, KW_ICON    , false),
         new Keyword("ide-parent-hwnd"                ,  0, KW_IDEPHWND, false),
         new Keyword("ide-window-type"                ,  0, KW_IDEWNTYP, false),
         new Keyword("ignore-current-modified"        ,  0, KW_IGN_CMOD, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("if"                             ,  0, KW_IF      , true ),
         new Keyword("icfparameter"                   ,  8, KW_ICFPARM , false),  // undocumented feature: can be abbreviated
         new Keyword("image"                          ,  0, KW_IMAGE   , false),
         new Keyword("image-down"                     ,  0, KW_IMG_DOWN, false),
         new Keyword("image-insensitive"              ,  0, KW_IMG_INS , false),
         new Keyword("image-size"                     ,  0, KW_IMG_SZ  , false),
         new Keyword("image-size-chars"               , 12, KW_IMG_SZ_C, false),
         new Keyword("image-size-pixels"              , 12, KW_IMG_SZ_P, false),
         new Keyword("image-up"                       ,  0, KW_IMG_UP  , false),
         new Keyword("immediate-display"              ,  0, KW_IMM_DISP, false),
         new Keyword("implements"                     ,  0, KW_IMPLEMTS, false),
         new Keyword("import"                         ,  0, KW_IMPORT  , true ),
         new Keyword("import-principal"               ,  0, KW_IMP_PRNC, false),
         new Keyword("import-node"                    ,  0, KW_IMP_NODE, false),
         new Keyword("in"                             ,  0, KW_IN      , true ),
         new Keyword("in-handle"                      ,  0, KW_IN_HNDL , false),
         new Keyword("increment-exclusive-id"         ,  0, KW_INC_EX_I, false),
         new Keyword("index"                          ,  0, KW_INDEX   , true ),
         new Keyword("index-hint"                     ,  0, KW_IDX_HINT, false),
         new Keyword("index-information"              ,  0, KW_IDX_INFO, false),
         new Keyword("indexed-reposition"             ,  0, KW_IDX_REPO, false),
         new Keyword("indicator"                      ,  0, KW_INDICAT , true ),
         new Keyword("information"                    ,  4, KW_INFO    , false),
         new Keyword("inherits"                       ,  0, KW_INHERITS, false),
         new Keyword("inherit-bgcolor"                ,  0, KW_INH_BGC , false),  // missing from keyword index and UNTESTED at this time
         new Keyword("inherit-fgcolor"                ,  0, KW_INH_FGC , false),  // missing from keyword index and UNTESTED at this time
         new Keyword("init-session"                   ,  0, KW_INIT_SES, false),  // not a real keyword, but is normally defined in the PSC webspeed 4GL code
         new Keyword("initial"                        ,  4, KW_INIT    , false),
         new Keyword("initial-dir"                    ,  0, KW_INIT_DIR, false),
         new Keyword("initial-filter"                 ,  0, KW_INIT_FLT, false),
         new Keyword("initialize-document-type"       ,  0, KW_INIT_D_T, false),
         new Keyword("initiate"                       ,  0, KW_INITIATE, false),
         new Keyword("inner-chars"                    ,  0, KW_INNER_C , false),
         new Keyword("inner-lines"                    ,  0, KW_INNER_L , false),
         new Keyword("inner"                          ,  0, KW_INNER   , false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("input"                          ,  0, KW_INPUT   , true ),
         new Keyword("input-output"                   ,  7, KW_IN_OUT  , true ),
         new Keyword("input-value"                    ,  0, KW_INPT_VAL, false),
         new Keyword("insert"                         ,  0, KW_INSERT  , true ),
         new Keyword("insert-attribute"               ,  0, KW_INS_ATTR, false),
         new Keyword("insert-backtab"                 ,  8, KW_INS_BTAB, false),
         new Keyword("insert-before"                  ,  0, KW_INS_B4  , false),
         new Keyword("insert-column"                  ,  0, KW_INS_COL , false),  // missing in keyword index, found in prog handbook as a key function
         new Keyword("insert-file"                    ,  0, KW_INS_FILE, false),
         new Keyword("insert-field"                   ,  0, KW_INS_FLD , false),  // missing in keyword index, found in prog handbook as a key function
         new Keyword("insert-field-data"              ,  0, KW_INS_FLDD, false),  // missing in keyword index, found in prog handbook as a key function
         new Keyword("insert-field-label"             ,  0, KW_INS_FLDL, false),  // missing in keyword index, found in prog handbook as a key function
         new Keyword("insert-mode"                    ,  0, KW_INS_MODE, false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("insert-row"                     ,  0, KW_INS_ROW , false),
         new Keyword("insert-string"                  ,  0, KW_INS_STR , false),
         new Keyword("insert-tab"                     ,  8, KW_INS_TAB , false),
         new Keyword("instantiating-procedure"        ,  0, KW_INST_PRC, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("int64"                          ,  0, KW_INT64   , false),
         new Keyword("integer"                        ,  3, KW_INT     , false),
         new Keyword("interface"                      ,  0, KW_INTERFAC, false),
         new Keyword("internal-entries"               ,  0, KW_INT_ENT , false),
         new Keyword("interval"                       ,  0, KW_INTERVAL, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("into"                           ,  0, KW_INTO    , true ),
         new Keyword("invoke"                         ,  0, KW_INVOKE  , false),
         new Keyword("is"                             ,  0, KW_IS      , true ),
         new Keyword("is-attr-space"                  ,  7, KW_IS_ATTR , true ),
         new Keyword("is-class"                       ,  7, KW_IS_CLASS, false),
         new Keyword("is-codepage-fixed"              ,  0, KW_IS_CP_FX, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("is-column-codepage"             ,  0, KW_IS_COLCP, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("is-lead-byte"                   ,  0, KW_IS_LEAD , true ),
         new Keyword("is-open"                        ,  0, KW_IS_OPEN , false),
         new Keyword("is-parameter-set"               ,  0, KW_IS_P_SET, false),
         new Keyword("is-row-selected"                ,  0, KW_IS_R_SEL, false),
         new Keyword("is-selected"                    ,  0, KW_IS_SEL  , false),
         new Keyword("is-xml"                         ,  0, KW_IS_XML  , false),  // missing from keyword index, found elsewhere in lang ref
         new Keyword("iso-date"                       ,  0, KW_ISO_DATE, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("item"                           ,  0, KW_ITEM    , false),
         new Keyword("items-per-row"                  ,  0, KW_ITEM_ROW, false),
         new Keyword("iteration-changed"              ,  0, KW_ITER_CHG, false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("iunknown"                       ,  0, KW_IUNKNOWN, false),  // missing in keyword index, found elsewhere in external program interface ref
         new Keyword("join"                           ,  0, KW_JOIN    , true ),
         new Keyword("join-by-sqldb"                  ,  0, KW_JOIN_BY , false),
         new Keyword("kblabel"                        ,  0, KW_KBLABEL , true ),
         new Keyword("keep-connection-open"           ,  0, KW_KEEP_CON, false),
         new Keyword("keep-frame-z-order"             , 12, KW_KEEP_ZOR, false),
         new Keyword("keep-messages"                  ,  0, KW_KEEP_MSG, false),
         new Keyword("keep-security-cache"            ,  0, KW_KEEP_SEC, false),
         new Keyword("keep-tab-order"                 ,  0, KW_KEEP_TAB, false),
         new Keyword("keycode"                        ,  0, KW_KEYCODE , true ),
         new Keyword("key-code"                       ,  0, KW_KEYCODE , true ),
         new Keyword("keyfunction"                    ,  7, KW_KEYFUNC , true ),
         new Keyword("key-function"                   ,  8, KW_KEYFUNC , true ),
         new Keyword("keylabel"                       ,  0, KW_KEYLAB  , true ),
         new Keyword("key-label"                      ,  0, KW_KEYLAB  , true ),
         new Keyword("key"                            ,  0, KW_KEY     , false),
         new Keyword("keys"                           ,  0, KW_KEYS    , true ),
         new Keyword("keyword"                        ,  0, KW_KW      , true ),
         new Keyword("keyword-all"                    ,  0, KW_KW_ALL  , false),
         new Keyword("label"                          ,  0, KW_LABEL   , true ),
         new Keyword("label-bgcolor"                  ,  9, KW_LAB_BGC , false),
         new Keyword("label-dcolor"                   ,  8, KW_LAB_DC  , false),
         new Keyword("label-fgcolor"                  ,  9, KW_LAB_FGC , false),
         new Keyword("label-font"                     ,  0, KW_LAB_FONT, false),
         new Keyword("label-pfcolor"                  ,  9, KW_LAB_PFC , false),
         new Keyword("labels"                         ,  0, KW_LABELS  , false),
         new Keyword("landscape"                      ,  0, KW_LANDSCAP, false),
         new Keyword("languages"                      ,  8, KW_LANGUAGE, false),
         new Keyword("large"                          ,  0, KW_LARGE   , false),
         new Keyword("large-to-small"                 ,  0, KW_LG_2_SM , false),
         new Keyword("last"                           ,  0, KW_LAST    , true ),
         new Keyword("last-batch"                     ,  0, KW_LAST_BAT, false),
         new Keyword("last-form"                      ,  0, KW_LAST_FRM, false),
         new Keyword("last-object"                    ,  0, KW_LAST_OBJ, false),
         new Keyword("last-of"                        ,  0, KW_LAST_OF , true ),
         new Keyword("lastkey"                        ,  0, KW_LASTKEY , true ),
         new Keyword("last-key"                       ,  0, KW_LASTKEY , true ),
         new Keyword("last-async-request"             ,  0, KW_LAST_AR , false),
         new Keyword("last-child"                     ,  0, KW_LAST_CH , false),
         new Keyword("last-event"                     ,  9, KW_LAST_EVT, true ),
         new Keyword("last-procedure"                 , 10, KW_LAST_PRC, false),
         new Keyword("last-server"                    ,  0, KW_LAST_SRV, false),
         new Keyword("last-server-socket"             ,  0, KW_LAST_SS , false),
         new Keyword("last-socket"                    ,  0, KW_LAST_SOC, false),
         new Keyword("last-tab-item"                  , 10, KW_LAST_TI , false),
         new Keyword("ldbname"                        ,  0, KW_LDBNAME , true ),
         new Keyword("lower"                          ,  0, KW_LC      , false),  // missing in keyword index, found elsewhere in SQL reference
         new Keyword("lc"                             ,  0, KW_LC      , false),
         new Keyword("le"                             ,  0, KW_LTE     , false),
         new Keyword("leave"                          ,  0, KW_LEAVE   , true ),
         new Keyword("left"                           ,  0, KW_LEFT    , false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("left-aligned"                   , 10, KW_LEFT_AL , false),
         new Keyword("left-end"                       ,  0, KW_LEFT_END, false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("left-mouse-click"               ,  0, KW_LEFT_MC , false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("left-mouse-dblclick"            ,  0, KW_LEFT_MDC, false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("left-mouse-down"                ,  0, KW_LEFT_MD , false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("left-mouse-up"                  ,  0, KW_LEFT_MU , false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("left-trim"                      ,  0, KW_L_TRIM  , false),
         new Keyword("length"                         ,  0, KW_LENGTH  , false),
         new Keyword("library"                        ,  0, KW_LIB     , true ),
         new Keyword("like"                           ,  0, KW_LIKE    , true ),
         new Keyword("like-sequential"                ,  0, KW_LIKE_SEQ, true ),
         new Keyword("line"                           ,  0, KW_LINE    , false),
         new Keyword("line-counter"                   , 10, KW_LINE_CNT, true ),
         new Keyword("list-events"                    ,  0, KW_LST_EVNT, false),
         new Keyword("listing"                        ,  5, KW_LISTING , true ),
         new Keyword("listings"                       ,  0, KW_LISTINGS, false),  // missing from keyword index, found in customer code and UNTESTED at this time
         new Keyword("list-item-pairs"                ,  0, KW_LST_PAIR, false),
         new Keyword("list-items"                     ,  0, KW_LIST_ITM, false),
         new Keyword("list-property-names"            ,  0, KW_LIST_PNM, false),
         new Keyword("list-query-attrs"               ,  0, KW_LST_QRY , false),
         new Keyword("list-set-attrs"                 ,  0, KW_LST_SET , false),
         new Keyword("list-widgets"                   ,  0, KW_LST_WID , false),
         new Keyword("literal-question"               ,  0, KW_LIT_QSTN, false),
         new Keyword("little-endian"                  ,  0, KW_L_ENDIAN, true ),
         new Keyword("lt"                             ,  0, KW_LT      , false),
         new Keyword("load"                           ,  0, KW_LOAD    , false),
         new Keyword("loadcontrols"                   ,  0, KW_LOADCTRL, false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("load-domains"                   ,  0, KW_LOAD_DMN, false),
         new Keyword("load-icon"                      ,  0, KW_LOAD_ICO, false),
         new Keyword("load-image"                     ,  0, KW_LOAD_IMG, false),
         new Keyword("load-image-down"                ,  0, KW_LOAD_I_D, false),
         new Keyword("load-image-insensitive"         ,  0, KW_LOAD_I_I, false),
         new Keyword("load-image-up"                  ,  0, KW_LOAD_I_U, false),
         new Keyword("load-mouse-pointer"             , 12, KW_LOAD_M_P, false),
         new Keyword("load-picture"                   ,  0, KW_LOAD_PIC, false),
         new Keyword("load-small-icon"                ,  0, KW_LOAD_S_I, false),
         new Keyword("lob-dir"                        ,  0, KW_LOB_DIR , false),  // missing from keyword index and UNTESTED at this time
         new Keyword("local-host"                     ,  0, KW_LOC_HOST, false),
         new Keyword("local-name"                     ,  0, KW_LOC_NAME, false),
         new Keyword("local-port"                     ,  0, KW_LOC_PORT, false),
         new Keyword("locator-column-number"          ,  0, KW_LOC_C_N , false),
         new Keyword("locator-line-number"            ,  0, KW_LOC_L_N , false),
         new Keyword("locator-public-id"              ,  0, KW_LOC_P_ID, false),
         new Keyword("locator-system-id"              ,  0, KW_LOC_S_ID, false),
         new Keyword("locator-type"                   ,  0, KW_LOC_TYPE, false),
         new Keyword("locked"                         ,  0, KW_LOCKED  , true ),
         new Keyword("lock-registration"              ,  0, KW_LOCK_REG, false),
         new Keyword("log"                            ,  0, KW_LOG     , false),
         new Keyword("log-audit-event"                ,  0, KW_LOG_A_EV, false),
         new Keyword("log-entry-types"                ,  0, KW_LOG_EN_T, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("log-id"                         ,  0, KW_LOG_ID  , false),  // missing from keyword index and found in customer code
         new Keyword("log-manager"                    ,  0, KW_LOG_MGR , true ),
         new Keyword("log-threshold"                  ,  0, KW_LOG_THRS, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("logfile-name"                   ,  0, KW_LOGF_NAM, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("logging-level"                  ,  0, KW_LOGG_LEV, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("logical"                        ,  4, KW_LOGICAL , false),
         new Keyword("login-expiration-timestamp"     ,  0, KW_LOG_E_TS, false),
         new Keyword("login-host"                     ,  0, KW_LOG_HOST, false),
         new Keyword("login-state"                    ,  0, KW_LOG_STAT, false),
         new Keyword("logout"                         ,  0, KW_LOGOUT  , false),
         new Keyword("lookahead"                      ,  0, KW_LOOKAHD , false),
         new Keyword("lookup"                         ,  0, KW_LOOKUP  , true ),
         new Keyword("long"                           ,  0, KW_LONG    , false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("longchar"                       ,  0, KW_LONGCHAR, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("longchar-to-node-value"         ,  0, KW_LCHR_2NV, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("machine-class"                  ,  0, KW_MACH_CLS, true ),
         new Keyword("main-menu"                      ,  0, KW_MAINMENU, false),  // missing in keyword index, found in prog handbook as a key function
         new Keyword("mandatory"                      ,  0, KW_MAND    , false),
         new Keyword("manual-highlight"               ,  0, KW_MAN_HIGH, false),
         new Keyword("margin-extra"                   ,  0, KW_MARG_EX , false),  // found in keyword index, this is an undocumented DEFINE BUTTON option
         new Keyword("margin-height-chars"            , 13, KW_MARG_H_C, false),
         new Keyword("margin-height-pixels"           , 15, KW_MARG_H_P, false),
         new Keyword("margin-width-chars"             , 12, KW_MARG_W_C, false),
         new Keyword("margin-width-pixels"            , 14, KW_MARG_W_P, false),
         new Keyword("maximum"                        ,  3, KW_MAX     , false),  // handles "separate keyword" max too
         new Keyword("maximize"                       ,  0, KW_MAXIMIZE, false),
         new Keyword("max-button"                     ,  0, KW_MAX_BTN , false),
         new Keyword("max-chars"                      ,  0, KW_MAX_CHAR, false),
         new Keyword("max-data-guess"                 ,  0, KW_MAX_D_G , false),
         new Keyword("max-height"                     ,  0, KW_MAX_HT  , false),
         new Keyword("max-height-chars"               , 12, KW_MAX_H_C , false),
         new Keyword("max-height-pixels"              , 12, KW_MAX_H_P , false),
         new Keyword("max-width-chars"                ,  9, KW_MAX_W_C , false),
         new Keyword("max-width-pixels"               , 11, KW_MAX_W_P , false),
         new Keyword("max-rows"                       ,  0, KW_MAX_ROWS, false),
         new Keyword("max-size"                       ,  0, KW_MAX_SZ  , false),
         new Keyword("max-value"                      ,  7, KW_MAX_VAL , false),
         new Keyword("max-width"                      ,  0, KW_MAX_WID , false),
         new Keyword("map"                            ,  0, KW_MAP     , true ),
         new Keyword("matches"                        ,  0, KW_MATCHES , false),
         new Keyword("md5-value"                      ,  0, KW_MD5_VAL , false),  // missing from keyword index and UNTESTED at this time
         new Keyword("md5-digest"                     ,  0, KW_MD5_DIG , false),
         new Keyword("member"                         ,  0, KW_MEMBER  , true ),
         new Keyword("memptr"                         ,  3, KW_MEMPTR  , false),
         new Keyword("memptr-to-node-value"           ,  0, KW_MPTR_2NV, false),
         new Keyword("menu"                           ,  0, KW_MENU    , false),
         new Keyword("menu-bar"                       ,  0, KW_MENU_BAR, false),
         new Keyword("menu-drop"                      ,  0, KW_MENU_DRP, false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("menu-key"                       ,  6, KW_MENU_KEY, false),
         new Keyword("menu-mouse"                     ,  6, KW_MENU_MOU, false),
         new Keyword("menubar"                        ,  0, KW_MENU_BAR, false),
         new Keyword("menu-item"                      ,  0, KW_MENU_ITM, false),
         new Keyword("merge-by-field"                 ,  0, KW_MERGE_BF, false),
         new Keyword("merge-changes"                  ,  0, KW_MERGE_CH, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("merge-row-changes"              ,  0, KW_MERGE_RC, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("message"                        ,  0, KW_MSG     , true ),
         new Keyword("messages"                       ,  0, KW_MESSAGES, false),
         new Keyword("message-area"                   ,  0, KW_MSG_AREA, false),
         new Keyword("message-area-font"              ,  0, KW_MSG_AFNT, false),
         new Keyword("message-digest"                 ,  0, KW_MSG_DIG , false),  // missing from keyword index and UNTESTED at this time
         new Keyword("message-lines"                  ,  0, KW_MSG_LINE, true ),
         new Keyword("method"                         ,  0, KW_METHOD  , false),
         new Keyword("middle-mouse-click"             ,  0, KW_MDL_MC  , false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("middle-mouse-dblclick"          ,  0, KW_MDL_MDC , false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("middle-mouse-down"              ,  0, KW_MDL_MD  , false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("middle-mouse-up"                ,  0, KW_MDL_MU  , false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("minimum"                        ,  3, KW_MIN     , false),  // handles "separate keyword" min too
         new Keyword("min-button"                     ,  0, KW_MIN_BTN , false),
         new Keyword("min-column-width-chars"         ,  0, KW_MIN_CWCH, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("min-column-width-pixels"        ,  0, KW_MIN_CWPX, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("min-height-chars"               , 10, KW_MIN_H_C , false),
         new Keyword("min-height-pixels"              , 12, KW_MIN_H_P , false),
         new Keyword("min-schema-marshal"             ,  0, KW_MIN_SCHM, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("min-width-chars"                ,  9, KW_MIN_W_C , false),
         new Keyword("min-width-pixels"               , 11, KW_MIN_W_P , false),
         new Keyword("min-size"                       ,  0, KW_MIN_SZ  , false),
         new Keyword("min-value"                      ,  7, KW_MIN_VAL , false),
         new Keyword("mnemonic"                       ,  0, KW_MNEMON  , false), // FWD extension, not real 4GL!
         new Keyword("modulo"                         ,  3, KW_MOD     , false),
         new Keyword("modified"                       ,  0, KW_MODIFIED, false),
         new Keyword("month"                          ,  0, KW_MONTH   , false),
         new Keyword("mouse"                          ,  0, KW_MOUSE   , true ),
         new Keyword("mouse-extend-click"             ,  0, KW_MOU_XC  , false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("mouse-extend-dblclick"          ,  0, KW_MOU_XDC , false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("mouse-extend-down"              ,  0, KW_MOU_XD  , false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("mouse-extend-up"                ,  0, KW_MOU_XU  , false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("mouse-menu-click"               ,  0, KW_MOU_MC  , false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("mouse-menu-dblclick"            ,  0, KW_MOU_MDC , false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("mouse-menu-down"                ,  0, KW_MOU_MD  , false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("mouse-menu-up"                  ,  0, KW_MOU_MU  , false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("mouse-move-click"               ,  0, KW_MOU_MVC , false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("mouse-move-dblclick"            ,  0, KW_MOU_MVDC, false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("mouse-move-down"                ,  0, KW_MOU_MVD , false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("mouse-move-up"                  ,  0, KW_MOU_MVU , false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("mouse-pointer"                  ,  7, KW_MOU_PTR , false),
         new Keyword("mouse-select-click"             ,  0, KW_MOU_SC  , false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("mouse-select-dblclick"          ,  0, KW_MOU_SDC , false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("mouse-select-down"              ,  0, KW_MOU_SD  , false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("mouse-select-up"                ,  0, KW_MOU_SU  , false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("movable"                        ,  0, KW_MOVABLE , false),
         new Keyword("move"                           ,  0, KW_MOVE    , false),  // missing in keyword index, found in prog handbook as a key function
         new Keyword("move-after-tab-item"            , 10, KW_MOV_A_T , false),
         new Keyword("move-before-tab-item"           , 10, KW_MOV_B_T , false),
         new Keyword("move-column"                    ,  8, KW_MOV_COL , false),
         new Keyword("move-to-bottom"                 ,  9, KW_MOV_2_B , false),
         new Keyword("move-to-eof"                    ,  0, KW_MOV_2EOF, false),
         new Keyword("move-to-top"                    ,  9, KW_MOV_2_T , false),
         new Keyword("mpe"                            ,  0, KW_MPE     , true ),
         new Keyword("mtime"                          ,  0, KW_MTIME   , false),  // missing from keyword index and UNTESTED at this time
         new Keyword("multi-compile"                  ,  0, KW_MULT_CMP, false),
         new Keyword("multiple"                       ,  0, KW_MULTIPLE, false),
         new Keyword("multiple-key"                   ,  0, KW_MULT_KEY, false),
         new Keyword("multitasking-interval"          ,  0, KW_MULTI_IN, false),
         new Keyword("must-exist"                     ,  0, KW_MUST_EXI, false),
         new Keyword("must-understand"                ,  0, KW_MUST_UND, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("name"                           ,  0, KW_NAME    , false),
         new Keyword("namespace-prefix"               ,  0, KW_NAMESP_P, false),
         new Keyword("namespace-uri"                  ,  0, KW_NAMESP_U, false),
         new Keyword("native"                         ,  0, KW_NATIVE  , false),
         new Keyword("ne"                             ,  0, KW_NE      , false),
         new Keyword("needs-appserver-prompt"         ,  0, KW_NEEDS_AP, false),
         new Keyword("needs-prompt"                   ,  0, KW_NEEDS_PR, false),
         new Keyword("nested"                         ,  0, KW_NESTED  , false),  // missing from keyword index and UNTESTED at this time
         new Keyword("new"                            ,  0, KW_NEW     , true ),
         new Keyword("new-instance"                   ,  0, KW_NEW_INST, false),
         new Keyword("new-line"                       ,  0, KW_NEW_LINE, false),  // missing in keyword index, found in customer source code
         new Keyword("new-row"                        ,  0, KW_NEW_ROW , false),
         new Keyword("next"                           ,  0, KW_NEXT    , true ),
         new Keyword("next-column"                    ,  0, KW_NEXT_COL, false),
         new Keyword("next-error"                     ,  0, KW_NEXT_ERR, false),  // missing in keyword index, found in prog handbook as a key function
         new Keyword("next-frame"                     ,  0, KW_NEXT_FR , false),
         new Keyword("next-prompt"                    ,  0, KW_NEXT_PMT, true ),
         new Keyword("next-rowid"                     ,  0, KW_NEXT_RID, false),
         new Keyword("next-sibling"                   ,  0, KW_NEXT_SIB, false),
         new Keyword("next-tab-item"                  , 10, KW_NEXT_TAB, false),
         new Keyword("next-value"                     ,  0, KW_NEXT_VAL, false),
         new Keyword("next-word"                      ,  0, KW_NEXTWORD, false),  // missing in keyword index, found in prog handbook as a key function
         new Keyword("no"                             ,  0, BOOL_FALSE , true ),
         new Keyword("no-apply"                       ,  0, KW_NO_APPLY, false),
         new Keyword("no-array-message"               ,  0, KW_NO_ARMSG, false),
         new Keyword("no-assign"                      ,  0, KW_NO_ASSGN, false),
         new Keyword("no-attr-list"                   ,  7, KW_NO_ATTRL, true ),
         new Keyword("no-attr-space"                  ,  7, KW_NO_ATTR , true ),
         new Keyword("no-auto-validate"               ,  0, KW_NO_AUTOV, false),
         new Keyword("no-bind-where"                  ,  0, KW_NO_BIND , false),
         new Keyword("no-box"                         ,  0, KW_NO_BOX  , false),
         new Keyword("no-column-scrolling"            ,  0, KW_NO_COLS , false),  // missing in keyword index, found in customer source code.
         new Keyword("no-console"                     ,  0, KW_NO_CONS , false),
         new Keyword("no-convert"                     ,  0, KW_NO_CVT  , false),
         new Keyword("no-convert-3d-colors"           ,  0, KW_NO_CV_3D, false),
         new Keyword("no-current-value"               ,  0, KW_NO_CUR_V, false),
         new Keyword("no-debug"                       ,  0, KW_NO_DEBUG, false),
         new Keyword("no-drag"                        ,  0, KW_NO_DRAG , false),
         new Keyword("no-echo"                        ,  0, KW_NO_ECHO , false),
         new Keyword("no-empty-space"                 ,  0, KW_NO_EM_SP, false),
         new Keyword("no-error"                       ,  0, KW_NO_ERROR, true ),
         new Keyword("no-fill"                        ,  4, KW_NO_FILL , true ),
         new Keyword("no-focus"                       ,  0, KW_NO_FOCUS, true ),
         new Keyword("no-help"                        ,  0, KW_NO_HELP , false),
         new Keyword("no-hide"                        ,  0, KW_NO_HIDE , true ),
         new Keyword("no-index-hint"                  ,  0, KW_NO_IDX_H, false),
         new Keyword("no-inherit-bgcolor"             , 14, KW_NO_INHBG, false),
         new Keyword("no-inherit-fgcolor"             , 14, KW_NO_INHFG, false),
         new Keyword("no-join-by-sqldb"               ,  0, KW_NO_JOIN , false),
         new Keyword("no-labels"                      ,  8, KW_NO_LABEL, true ),
         new Keyword("no-lobs"                        ,  0, KW_NO_LOBS , true ),
         new Keyword("no-lock"                        ,  0, KW_NO_LOCK , true ),
         new Keyword("no-lookahead"                   ,  0, KW_NO_LOOKA, false),
         new Keyword("no-map"                         ,  0, KW_NO_MAP  , true ),
         new Keyword("no-message"                     ,  6, KW_NO_MSG  , true ),
         new Keyword("no-pause"                       ,  0, KW_NO_PAUSE, true ),
         new Keyword("no-prefetch"                    ,  8, KW_NO_PRE  , true ),
         new Keyword("no-row-markers"                 ,  0, KW_NO_ROW_M, false),
         new Keyword("no-return-value"                , 13, KW_NO_RET_V, true ),  // missing in keyword index, found elsewhere in external program interface ref
         new Keyword("no-schema-marshal"              ,  0, KW_NO_SCH_M, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("no-scrollbar-vertical"          ,  0, KW_NO_SCR_V, false),
         new Keyword("no-separators"                  ,  0, KW_NO_SEPS , false),
         new Keyword("no-separate-connection"         ,  0, KW_NO_SEP_C, false),
         new Keyword("no-tab-stop"                    ,  0, KW_NO_TAB_S, false),
         new Keyword("no-underline"                   ,  6, KW_NO_UNDL , false),
         new Keyword("no-undo"                        ,  0, KW_NO_UNDO , true ),
         new Keyword("no-validate"                    ,  6, KW_NO_VALID, true ),
         new Keyword("no-wait"                        ,  0, KW_NO_WAIT , true ),
         new Keyword("no-word-wrap"                   ,  0, KW_NO_WRAP , false),
         new Keyword("node-value"                     ,  0, KW_NODE_VAL, false),
         new Keyword("node-value-to-memptr"           ,  0, KW_NODE_V2M, false),
         new Keyword("node-value-to-longchar"         ,  0, KW_NODV_2LC, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("nonamespace-schema-location"    ,  0, KW_NNMSP_SL, false),
         new Keyword("none"                           ,  0, KW_NONE    , false),
         new Keyword("normal"                         ,  3, KW_NORMAL  , false),
         new Keyword("normalize"                      ,  0, KW_NORMALZE, false),
         new Keyword("not"                            ,  0, KW_NOT     , true ),
         new Keyword("not-active"                     ,  0, KW_NOT_ACTV, false),
         new Keyword("now"                            ,  0, KW_NOW     , true ),
         new Keyword("null"                           ,  0, KW_NULL    , true ),
         new Keyword("num-aliases"                    ,  7, KW_NUM_ALIA, true ),
         new Keyword("num-buffers"                    ,  0, KW_NUM_BUFF, false),
         new Keyword("num-buttons"                    ,  7, KW_NUM_BUTT, false),
         new Keyword("num-children"                   ,  0, KW_NUM_CHLN, false),
         new Keyword("num-child-relations"            ,  0, KW_NUM_CH_R, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("num-copies"                     ,  0, KW_NUM_COPY, false),
         new Keyword("num-columns"                    ,  7, KW_NUM_COL , false),
         new Keyword("num-dbs"                        ,  0, KW_NUM_DBS , true ),
         new Keyword("num-dropped-files"              ,  0, KW_NUM_DROP, false),
         new Keyword("num-entries"                    ,  0, KW_NUM_ENT , true ),
         new Keyword("num-fields"                     ,  0, KW_NUM_FLD , false),
         new Keyword("num-formats"                    ,  0, KW_NUM_FMTS, false),
         new Keyword("num-header-entries"             ,  0, KW_NUM_HD_E, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("num-items"                      ,  0, KW_NUM_ITMS, false),
         new Keyword("num-iterations"                 ,  0, KW_NUM_ITER, false),
         new Keyword("num-lines"                      ,  0, KW_NUM_LNS , false),
         new Keyword("num-locked-columns"             , 14, KW_NUM_LK_C, false),
         new Keyword("num-log-files"                  ,  0, KW_NUM_LOGF, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("num-messages"                   ,  0, KW_NUM_MSG , false),
         new Keyword("num-parameters"                 ,  0, KW_NUM_PARM, false),
         new Keyword("num-references"                 ,  0, KW_NUM_REF , false),
         new Keyword("num-relations"                  ,  0, KW_NUM_REL , false),  // missing from keyword index and UNTESTED at this time
         new Keyword("num-replaced"                   ,  0, KW_NUM_REPL, false),
         new Keyword("num-results"                    ,  0, KW_NUM_RES , false),
         new Keyword("num-selected-rows"              ,  0, KW_NUM_SR  , false),
         new Keyword("num-selected-widgets"           , 12, KW_NUM_SW  , false),
         new Keyword("num-source-buffers"             ,  0, KW_NUM_SRCB, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("num-top-buffers"                ,  0, KW_NUM_TOPB, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("num-tabs"                       ,  0, KW_NUM_TABS, false),
         new Keyword("num-to-retain"                  ,  0, KW_NUM_2RTN, false),
         new Keyword("num-visible-columns"            ,  0, KW_NUM_V_C , false),
         new Keyword("numeric"                        ,  0, KW_NUMERIC , false),
         new Keyword("numeric-decimal-point"          , 15, KW_NUM_D_P , false),
         new Keyword("numeric-format"                 ,  9, KW_NUM_FMT , false),
         new Keyword("numeric-separator"              ,  0, KW_NUM_SEP , false),
         new Keyword("object"                         ,  0, KW_OBJECT  , false),
         new Keyword("octet-length"                   ,  0, KW_OCT_LEN , false),
         new Keyword("of"                             ,  0, KW_OF      , true ),
         new Keyword("off-end"                        ,  0, KW_OFF_END , false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("off-home"                       ,  0, KW_OFF_HOME, false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("off"                            ,  0, KW_OFF     , true ),
         new Keyword("ok"                             ,  0, KW_OK      , false),
         new Keyword("ok-cancel"                      ,  0, KW_OK_CAN  , false),
         new Keyword("old"                            ,  0, KW_OLD     , true ),
         new Keyword("ole-invoke-locale"              ,  0, KW_OLE_INVL, false),
         new Keyword("ole-names-locale"               ,  0, KW_OLE_NAML, false),
         new Keyword("on"                             ,  0, KW_ON      , true ),
         new Keyword("on-frame-border"                ,  8, KW_ON_FR_B , false),
         new Keyword("only"                           ,  0, KW_ONLY    , false),
         new Keyword("open"                           ,  0, KW_OPEN    , true ),
         new Keyword("open-line-above"                ,  0, KW_OPEN_L_A, false),  // missing in keyword index, found in prog handbook as a key function
         new Keyword("opsys"                          ,  0, KW_OPSYS   , true ),
         new Keyword("option"                         ,  0, KW_OPTION  , true ),
         new Keyword("options"                        ,  0, KW_OPTIONS , false),  // missing in keyword index, found in prog handbook as a key function
         new Keyword("or"                             ,  0, KW_OR      , true ),
         new Keyword("ordinal"                        ,  0, KW_ORDINAL , false),
         new Keyword("order"                          ,  0, KW_ORDER   , false),  // missing in keyword index, found in SQL reference
         new Keyword("ordered-join"                   ,  0, KW_ORD_JOIN, false),
         new Keyword("origin-handle"                  ,  0, KW_ORG_HAND, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("origin-rowid"                   ,  0, KW_ORG_ROID, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("os2"                            ,  0, KW_OS2     , true ),  // undocumented
         new Keyword("os-append"                      ,  0, KW_OS_APPND, true ),
         new Keyword("os-copy"                        ,  0, KW_OS_COPY , true ),
         new Keyword("os-command"                     ,  0, KW_OS_CMD  , true ),
         new Keyword("os-create-dir"                  ,  0, KW_OS_MKDIR, true ),
         new Keyword("os-drives"                      ,  8, KW_OS_DRV  , false),
         new Keyword("os-delete"                      ,  0, KW_OS_DEL  , true ),
         new Keyword("os-dir"                         ,  0, KW_OS_DIR  , true ),
         new Keyword("os-error"                       ,  0, KW_OS_ERR  , false),
         new Keyword("os-getenv"                      ,  0, KW_OS_G_ENV, false),
         new Keyword("os-rename"                      ,  0, KW_OS_REN  , true ),
         new Keyword("otherwise"                      ,  0, KW_OTHER   , true ),
         new Keyword("outer-join"                     ,  0, KW_OUT_JOIN, false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("outer"                          ,  0, KW_OUTER   , false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("output"                         ,  0, KW_OUTPUT  , true ),
         new Keyword("outputcontenttype"              ,  0, KW_OUTPUTCT, false),  // not a real keyword, but is normally defined in the PSC webspeed 4GL code
         new Keyword("output-content-type"            ,  0, KW_OUTPUTCT, false),  // not a real keyword, but is normally defined in the PSC webspeed 4GL code
         new Keyword("outputheaders"                  ,  0, KW_OUT_HDR , false),  // not a real keyword, but is normally defined in the PSC webspeed 4GL code
         new Keyword("outputheader"                   ,  0, KW_OUT_HDR , false),  // not a real keyword, but is normally defined in the PSC webspeed 4GL code
         new Keyword("outputhttpheader"               ,  0, KW_OUT_HH  , false),  // not a real keyword, but is normally defined in the PSC webspeed 4GL code
         new Keyword("output-http-header"             ,  0, KW_OUT_HH  , false),  // not a real keyword, but is normally defined in the PSC webspeed 4GL code
         new Keyword("output-messages"                ,  0, KW_OUT_MSGS, false),  // not a real keyword, but is normally defined in the PSC webspeed 4GL code
         new Keyword("overlay"                        ,  0, KW_OVERLAY , true ),
         new Keyword("override"                       ,  0, KW_OVERRIDE, false),
         new Keyword("owner"                          ,  0, KW_OWNER   , false),
         new Keyword("owner-document"                 ,  0, KW_OWN_DOC , false),
         new Keyword("p2j-open-url"                   ,  0, KW_OPEN_URL, false),  // P2J-extension, not real 4GL!
         new Keyword("p2j-remote-call"                ,  0, KW_P2J_RC  , false),  // P2J-extension, not real 4GL!
         new Keyword("page"                           ,  0, KW_PAGE    , true ),
         new Keyword("paged"                          ,  0, KW_PAGED   , false),
         new Keyword("page-bottom"                    ,  8, KW_PAGE_B  , true ),
         new Keyword("page-down"                      ,  0, KW_PAGE_DWN, false),  // missing in keyword index, found in prog handbook as a key function
         new Keyword("page-left"                      ,  0, KW_PAGE_LFT, false),  // missing in keyword index, found in prog handbook as a key function
         new Keyword("page-number"                    ,  8, KW_PAGE_NUM, true ),
         new Keyword("page-right"                     ,  0, KW_PAGE_RT , false),  // missing in keyword index, found in prog handbook as a key function
         new Keyword("page-size"                      ,  0, KW_PAGE_SZ , false),
         new Keyword("page-top"                       ,  0, KW_PAGE_T  , true ),
         new Keyword("page-up"                        ,  0, KW_PAGE_UP , false),  // missing in keyword index, found in prog handbook as a key function
         new Keyword("page-width"                     ,  8, KW_PAGE_WID, false),
         new Keyword("parameter"                      ,  5, KW_PARM    , true ),
         new Keyword("parent-buffer"                  ,  0, KW_PAR_BUFF, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("parent-fields-after"            ,  0, KW_PAR_FLDA, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("parent-fields-before"           ,  0, KW_PAR_FLDB, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("parent-id-field"                ,  0, KW_PAR_IFLD, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("parent-id-relation"             ,  0, KW_PAR_IREL, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("parent-relation"                ,  0, KW_PAR_REL , false),  // missing from keyword index and UNTESTED at this time
         new Keyword("parent-window-close"            ,  0, KW_PAR_W_C , false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("parent"                         ,  0, KW_PARENT  , false),
         new Keyword("parse-status"                   ,  0, KW_PARSE_ST, false),
         new Keyword("partial-key"                    ,  0, KW_PART_KEY, false),
         new Keyword("pathname"                       ,  0, KW_PATHNAME, false),
         new Keyword("pascal"                         ,  0, KW_PASCAL  , false),
         new Keyword("password-field"                 ,  0, KW_PASSWD_F, true ),
         new Keyword("paste"                          ,  0, KW_PASTE   , false),  // missing in keyword index, found in prog handbook as a key function
         new Keyword("pause"                          ,  0, KW_PAUSE   , true ),
         new Keyword("pbe-hash-algorithm"             , 12, KW_PBE_H_AL, false),
         new Keyword("pbe-key-rounds"                 ,  0, KW_PBE_KEYR, false),
         new Keyword("pdbname"                        ,  0, KW_PDBNAME , true ),
         new Keyword("performance"                    ,  0, KW_PERF    , false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("persistent"                     ,  7, KW_PERSIST , true ),
         new Keyword("persistent-procedure"           ,  0, KW_PRS_PROC, false),
         new Keyword("persistent-cache-disabled"      ,  0, KW_PRS_C_D , false),
         new Keyword("pfcolor"                        ,  3, KW_PFCOLOR , false),
         new Keyword("pick"                           ,  0, KW_PICK    , false),  // missing in keyword index, found in prog handbook as a key function
         new Keyword("pick-area"                      ,  0, KW_PICK_ARE, false),  // missing in keyword index, found in prog handbook as a key function
         new Keyword("pick-both"                      ,  0, KW_PICK_BTH, false),  // missing in keyword index, found in prog handbook as a key function
         new Keyword("pixels"                         ,  0, KW_PIXELS  , true ),
         new Keyword("pixels-per-column"              , 14, KW_PIX_COL , false),
         new Keyword("pixels-per-row"                 ,  0, KW_PIX_ROW , false),
         new Keyword("portrait"                       ,  0, KW_PORTRAIT, false),
         new Keyword("position"                       ,  7, KW_POS     , false),
         new Keyword("popup-menu"                     ,  7, KW_POP_MENU, false),
         new Keyword("popup-only"                     ,  7, KW_POP_ONLY, false),
         new Keyword("precision"                      ,  0, KW_PRECISN , false),
         new Keyword("prefer-dataset"                 ,  0, KW_PREF_DS , false),
         new Keyword("prepare-string"                 ,  0, KW_PREP_STR, false),
         new Keyword("prepared"                       ,  0, KW_PREPARED, false),
         new Keyword("preprocess"                     ,  7, KW_PREPROC , true ),
         new Keyword("preprocessed-label"             ,  0, KW_PRE_LBL , false), // FWD extension, not real 4GL!
         new Keyword("preselect"                      ,  6, KW_PRESEL  , false),
         new Keyword("prev"                           ,  0, KW_PREV    , false),
         new Keyword("prev-column"                    ,  0, KW_PREV_COL, false),
         new Keyword("prev-frame"                     ,  0, KW_PREV_FR , false),
         new Keyword("prev-sibling"                   ,  0, KW_PREV_SIB, false),
         new Keyword("prev-tab-item"                  , 10, KW_PREV_T_I, false),
         new Keyword("prev-word"                      ,  0, KW_PREVWORD, false),  // missing in keyword index, found in prog handbook as a key function
         new Keyword("primary"                        ,  0, KW_PRIMARY , false),
         new Keyword("printer"                        ,  0, KW_PRINTER , false),
         new Keyword("printer-control-handle"         ,  0, KW_PRT_C_H , false),
         new Keyword("printer-hdc"                    ,  0, KW_PRT_HDC , false),
         new Keyword("printer-name"                   ,  0, KW_PRT_NAME, false),
         new Keyword("printer-port"                   ,  0, KW_PRT_PORT, false),
         new Keyword("printer-setup"                  ,  0, KW_PRT_SET , false),
         new Keyword("private"                        ,  0, KW_PRIVATE , false),
         new Keyword("private-data"                   ,  9, KW_PRIV_DAT, false),
         new Keyword("privileges"                     ,  0, KW_PRIVILEG, true ),
         new Keyword("procedure-name"                 ,  0, KW_PROCNAME, false),
         new Keyword("procedure-call-type"            ,  0, KW_PROC_C_T, true ),
         new Keyword("procedure-complete"             ,  0, KW_PROC_COM, false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("procedure"                      ,  5, KW_PROC    , false),
         new Keyword("process-web-request"            ,  0, KW_PROC_W_R, false),  // not a real keyword, but is normally defined in the PSC webspeed 4GL code
         new Keyword("process"                        ,  0, KW_PROCESS , true ),
         new Keyword("proc-handle"                    ,  7, KW_PROC_HND, true ),
         new Keyword("proc-status"                    ,  7, KW_PROC_ST , true ),
         new Keyword("proc-text"                      ,  0, KW_PROC_TXT, false),
         new Keyword("proc-text-buffer"               ,  0, KW_PROC_T_B, false),
         new Keyword("prodataset"                     ,  0, KW_PRODATAS, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("profiler"                       ,  0, KW_PROFILER, true ),
         new Keyword("profiling"                      ,  0, KW_PROFILNG, false),  // missing from keyword index, found in customer code and UNTESTED at this time
         new Keyword("program-name"                   ,  0, KW_PROGNAME, true ),
         new Keyword("progress"                       ,  0, KW_PROGRESS, true ),
         new Keyword("progress-source"                , 10, KW_PROG_SRC, false),
         new Keyword("prompt"                         ,  0, KW_PROMPT  , false),
         new Keyword("prompt-for"                     ,  8, KW_PRMT_FOR, true ),
         new Keyword("promsgs"                        ,  0, KW_PROMSGS , true ),
         new Keyword("propath"                        ,  0, KW_PROPATH , true ),
         new Keyword("property"                       ,  0, KW_PROPERTY, false),
         new Keyword("protected"                      ,  0, KW_PROTECTD, false),
         new Keyword("proversion"                     ,  7, KW_PROVER  , true ),
         new Keyword("proxy"                          ,  0, KW_PROXY   , false),
         new Keyword("proxy-password"                 ,  0, KW_PROX_PWD, false),
         new Keyword("proxy-userid"                   ,  0, KW_PROX_UID, false),
         new Keyword("public-id"                      ,  0, KW_PUB_ID  , false),
         new Keyword("public"                         ,  0, KW_PUBLIC  , false),
         new Keyword("publish"                        ,  0, KW_PUBLISH , false),
         new Keyword("published-events"               ,  0, KW_PUB_EVTS, false),
         new Keyword("put"                            ,  0, KW_PUT     , true ),
         new Keyword("put-bits"                       ,  0, KW_PUT_BITS, false),
         new Keyword("putbyte"                        ,  0, KW_PUT_BYTE, true ),
         new Keyword("put-byte"                       ,  0, KW_PUT_BYTE, true ),
         new Keyword("put-int64"                      ,  0, KW_PUT_I64 , false),
         new Keyword("put-bytes"                      ,  0, KW_PUT_BYTS, false),
         new Keyword("put-double"                     ,  0, KW_PUT_DBL , false),
         new Keyword("put-float"                      ,  0, KW_PUT_FLT , false),
         new Keyword("put-key-value"                  , 11, KW_PUT_K_V , true ),
         new Keyword("put-long"                       ,  0, KW_PUT_LONG, false),
         new Keyword("put-short"                      ,  0, KW_PUT_SHT , false),
         new Keyword("put-string"                     ,  0, KW_PUT_STR , false),
         new Keyword("put-unsigned-long"              ,  0, KW_PUT_UL  , false),
         new Keyword("put-unsigned-short"             ,  0, KW_PUT_USHT, false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("query"                          ,  0, KW_QUERY   , true ),
         new Keyword("query-close"                    ,  0, KW_QRY_CLOS, true ),
         new Keyword("query-off-end"                  ,  0, KW_QRY_OFF , true ),
         new Keyword("query-open"                     ,  0, KW_QRY_OPEN, false),
         new Keyword("query-prepare"                  ,  0, KW_QRY_PREP, false),
         new Keyword("query-tuning"                   ,  0, KW_QRY_TUNE, true ),
         new Keyword("question"                       ,  0, KW_QUEST   , false),
         new Keyword("queue-message"                  ,  0, KW_QUE_MSG , false),  // not a real keyword, but is normally defined in the PSC webspeed 4GL code
         new Keyword("quit"                           ,  0, KW_QUIT    , true ),
         new Keyword("quoter"                         ,  0, KW_QUOTER  , false),
         new Keyword("r-index"                        ,  0, KW_R_INDEX , true ),
         new Keyword("radio-buttons"                  ,  0, KW_RADIO_B , false),
         new Keyword("radio-set"                      ,  0, KW_RADIO_S , false),
         new Keyword("random"                         ,  0, KW_RANDOM  , false),
         new Keyword("raw"                            ,  0, KW_RAW     , false),
         new Keyword("raw-transfer"                   ,  0, KW_RAW_TRAN, false),
         new Keyword("rcode-information"              , 10, KW_RCOD_INF, true ),
         new Keyword("readkey"                        ,  0, KW_READKEY , true ),
         new Keyword("real"                           ,  0, KW_REAL    , false),
         new Keyword("read"                           ,  0, KW_READ    , false),
         new Keyword("read-available"                 ,  0, KW_READ_AVL, true ),
         new Keyword("read-exact-num"                 ,  0, KW_READ_E_N, true ),
         new Keyword("read-file"                      ,  0, KW_READ_FIL, false),
         new Keyword("read-json"                      ,  0, KW_READ_JSN, false),
         new Keyword("read-only"                      ,  0, KW_READ_ONL, false),
         new Keyword("read-response"                  ,  0, KW_READ_RES, false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("read-xml"                       ,  0, KW_READ_XML, false),
         new Keyword("read-xmlschema"                 ,  0, KW_READ_XSC, false),
         new Keyword("recall"                         ,  0, KW_RECALL  , false),
         new Keyword("recid"                          ,  0, KW_RECID   , true ),
         new Keyword("record-length"                  ,  0, KW_REC_LEN , false),
         new Keyword("rectangle"                      ,  4, KW_RECT    , true ),
         new Keyword("recursive"                      ,  0, KW_RECURSE , false),
         new Keyword("reference-only"                 ,  0, KW_REF_ONLY, false),
         new Keyword("refresh"                        ,  0, KW_REFRESH , false),
         new Keyword("refresh-audit-policy"           ,  0, KW_REFR_A_P, false),
         new Keyword("refreshable"                    ,  0, KW_REFRABLE, false),
         new Keyword("register-domain"                ,  0, KW_REG_DMN , false),
         new Keyword("reject-changes"                 ,  0, KW_REJ_CHGS, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("reject-row-changes"             ,  0, KW_REJ_RCHG, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("rejected"                       ,  0, KW_REJECTED, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("relation-fields"                ,  0, KW_REL_FLDS, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("relations-active"               ,  0, KW_RELS_ACT, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("release"                        ,  0, KW_RELEASE , true ),
         new Keyword("remote-host"                    ,  0, KW_REM_HOST, false),
         new Keyword("remote-port"                    ,  0, KW_REM_PORT, false),
         new Keyword("remote"                         ,  0, KW_REMOTE  , false),
         new Keyword("remove-attribute"               ,  0, KW_REM_ATTR, false),
         new Keyword("remove-child"                   ,  0, KW_REM_CHLD, false),
         new Keyword("remove-events-procedure"        ,  0, KW_REM_EVTP, false),
         new Keyword("remove-super-procedure"         ,  0, KW_REM_SUP , false),
         new Keyword("repeat"                         ,  0, KW_REPEAT  , true ),
         new Keyword("replace"                        ,  0, KW_REPLACE , false),
         new Keyword("replace-child"                  ,  0, KW_REP_CHLD, false),
         new Keyword("replace-selection-text"         ,  0, KW_REP_STXT, false),
         new Keyword("replication-create"             ,  0, KW_REPL_CRE, false),
         new Keyword("replication-delete"             ,  0, KW_REPL_DEL, false),
         new Keyword("replication-write"              ,  0, KW_REPL_WRI, false),
         new Keyword("reports"                        ,  0, KW_REPORTS , false),  // missing in keyword index, found in prog handbook as a key function
         new Keyword("reposition"                     ,  0, KW_REPOS   , true ),
         new Keyword("reposition-backwards"           , 19, KW_REPOS_B , true ),  // the documented keyword is reposition-backward with no abbrev, but customer code shows reposition-backwards as a valid alternative 
         new Keyword("reposition-forwards"            , 18, KW_REPOS_F , true ),  // the documented keyword is reposition-forward with no abbrev, but customer code shows reposition-forwards as a valid alternative
         new Keyword("reposition-mode"                ,  0, KW_REPOS_M , false),
         new Keyword("reposition-to-row"              ,  0, KW_REPOS_2R, true ),
         new Keyword("reposition-to-rowid"            ,  0, KW_REPOS_2I, true ),
         new Keyword("request"                        ,  0, KW_REQUEST , false),
         new Keyword("reset"                          ,  0, KW_RESET   , false),
         new Keyword("resize"                         ,  0, KW_RESIZE  , false),
         new Keyword("resizable"                      ,  6, KW_RESIZABL, false),
         new Keyword("restart-rowid"                  ,  0, KW_REST_RID, false),
         new Keyword("result"                         ,  0, KW_RESULT  , false),
         new Keyword("resume-display"                 ,  0, KW_RESUME_D, false),  // missing in keyword index, found in prog handbook as a key function
         new Keyword("retain"                         ,  0, KW_RETAIN  , true ),
         new Keyword("retain-shape"                   ,  0, KW_RET_SHAP, false),
         new Keyword("retry"                          ,  0, KW_RETRY   , true ),
         new Keyword("retry-cancel"                   ,  0, KW_RETRY_C , false),
         new Keyword("return"                         ,  0, KW_RETURN  , true ),
         new Keyword("return-value-data-type"         ,  0, KW_RET_VDT , false),
         new Keyword("returns"                        ,  0, KW_RETURNS , true ),
         new Keyword("return-inserted"                , 10, KW_RET_INS , false),
         new Keyword("return-value"                   , 10, KW_RET_VAL , false),
         new Keyword("return-to-start-dir"            , 18, KW_RET_2SD , false),
         new Keyword("reverse-from"                   ,  0, KW_REV_FROM, false),
         new Keyword("revert"                         ,  0, KW_REVERT  , true ),
         new Keyword("revoke"                         ,  0, KW_REVOKE  , true ),
         new Keyword("rgb-value"                      ,  0, KW_RGB_VAL , false),
         new Keyword("right"                          ,  0, KW_RIGHT   , false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("right-aligned"                  , 11, KW_RIGHT_AL, false),
         new Keyword("right-end"                      ,  0, KW_RT_END  , false),  // missing from keyword index, found elsewhere in lang ref
         new Keyword("right-mouse-click"              ,  0, KW_RT_MC   , false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("right-mouse-dblclick"           ,  0, KW_RT_MDC  , false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("right-mouse-down"               ,  0, KW_RT_MD   , false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("right-mouse-up"                 ,  0, KW_RT_MU   , false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("right-trim"                     ,  0, KW_R_TRIM  , false),
         new Keyword("roles"                          ,  0, KW_ROLES   , false),
         new Keyword("round"                          ,  0, KW_ROUND   , false),
         new Keyword("rounded"                        ,  0, KW_ROUNDED , false),  // missing from keyword index and UNTESTED at this time
         new Keyword("routine-level"                  ,  0, KW_ROUTINEL, false),
         new Keyword("row"                            ,  0, KW_ROW     , false),
         new Keyword("row-create"                     ,  0, KW_ROW_CRT , false),  // missing from keyword index and UNTESTED at this time
         new Keyword("row-created"                    ,  0, KW_ROW_CRTD, true ),  // missing from keyword index, found elsewhere in lang ref, TESTED
         new Keyword("row-delete"                     ,  0, KW_ROW_DEL , false),  // missing from keyword index and UNTESTED at this time
         new Keyword("row-deleted"                    ,  0, KW_ROW_DELD, true ),  // missing from keyword index, found elsewhere in lang ref, TESTED
         new Keyword("row-display"                    ,  0, KW_ROW_DISP, false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("row-entry"                      ,  0, KW_ROW_ENTR, false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("row-height-chars"               , 10, KW_ROW_H_C , false),
         new Keyword("row-height-pixels"              , 12, KW_ROW_H_P , false),
         new Keyword("row-leave"                      ,  0, KW_ROW_LEAV, false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("row-markers"                    ,  6, KW_ROW_MARK, false),  // the abbreviation here is undocumented, but found in customer source, TESTED
         new Keyword("row-modified"                   ,  0, KW_ROW_MODD, true ),  // missing from keyword index, found elsewhere in lang ref, TESTED
         new Keyword("row-resizable"                  ,  0, KW_ROW_RESZ, false),
         new Keyword("row-state"                      ,  0, KW_ROW_STAT, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("row-unmodified"                 ,  0, KW_ROW_UMOD, true ),  // missing from keyword index, found elsewhere in lang ref, TESTED
         new Keyword("row-update"                     ,  0, KW_ROW_UPD , false),  // missing from keyword index and UNTESTED at this time
         new Keyword("rowid"                          ,  0, KW_ROWID   , false ), // documented as reserved in the keyword index, but found to be unreserved
         new Keyword("row-of"                         ,  0, KW_ROW_OF  , false),
         new Keyword("rule"                           ,  0, KW_RULE    , false),
         new Keyword("rule-row"                       ,  0, KW_RULE_ROW, false),
         new Keyword("rule-y"                         ,  0, KW_RULE_Y  , false),
         new Keyword("run-procedure"                  ,  0, KW_RUN_PROC, false),
         new Keyword("run-web-object"                 ,  0, KW_RUN_W_O , false),  // not a real keyword, but is normally defined in the PSC webspeed 4GL code
         new Keyword("run"                            ,  0, KW_RUN     , true ),
         new Keyword("save"                           ,  0, KW_SAVE    , true ),
         new Keyword("save-as"                        ,  0, KW_SAVE_AS , false),
         new Keyword("save-file"                      ,  0, KW_SAVE_FIL, false),
         new Keyword("save-row-changes"               ,  0, KW_SAVE_RCH, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("save-where-string"              ,  0, KW_SAVE_WST, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("sax-attributes"                 ,  0, KW_SAX_ATTR, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("sax-complete"                   , 10, KW_SAX_COMP, true ),
         new Keyword("sax-parse"                      ,  0, KW_SAX_PARS, false),
         new Keyword("sax-parser-error"               ,  0, KW_SAX_PARE, true ),
         new Keyword("sax-parse-first"                ,  0, KW_SAX_PARF, false),
         new Keyword("sax-parse-next"                 ,  0, KW_SAX_PARN, false),
         new Keyword("sax-reader"                     ,  0, KW_SAX_READ, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("sax-running"                    ,  0, KW_SAX_RUNN, true ),
         new Keyword("sax-uninitialized"              ,  0, KW_SAX_UNIN, true ),
         new Keyword("sax-write-begin"                ,  0, KW_SAX_WBEG, true ),
         new Keyword("sax-write-complete"             ,  0, KW_SAX_WCOM, true ),
         new Keyword("sax-write-content"              ,  0, KW_SAX_WCON, true ),
         new Keyword("sax-write-element"              ,  0, KW_SAX_WELM, true ),
         new Keyword("sax-write-error"                ,  0, KW_SAX_WERR, true ),
         new Keyword("sax-write-idle"                 ,  0, KW_SAX_WIDL, true ),
         new Keyword("sax-write-tag"                  ,  0, KW_SAX_WTAG, true ),
         new Keyword("sax-writer"                     ,  0, KW_SAX_WRIT, false),
         new Keyword("schema"                         ,  0, KW_SCHEMA  , true ),
         new Keyword("schema-change"                  ,  0, KW_SCH_CHG , false),  // missing from keyword index and UNTESTED at this time
         new Keyword("schema-location"                ,  0, KW_SCH_LOC , false),
         new Keyword("schema-marshal"                 ,  0, KW_SCH_MARS, false),
         new Keyword("schema-path"                    ,  0, KW_SCH_PATH, false),
         new Keyword("screen"                         ,  0, KW_SCREEN  , true ),
         new Keyword("screen-io"                      ,  0, KW_SCRN_IO , true ),
         new Keyword("screen-lines"                   ,  0, KW_SCRN_LNS, true ),
         new Keyword("screen-value"                   ,  9, KW_SCRN_VAL, true ),
         new Keyword("scroll"                         ,  0, KW_SCROLL  , true ),
         new Keyword("scroll-bars"                    ,  0, KW_SCROLLBA, false),
         new Keyword("scroll-delta"                   ,  0, KW_SCR_DELT, false),
         new Keyword("scroll-left"                    ,  0, KW_SCR_LEFT, false),  // missing in keyword index, found in prog handbook as a key function
         new Keyword("scroll-mode"                    ,  0, KW_SCR_MODE, false),  // missing from keyword index, found elsewhere in lang ref
         new Keyword("scroll-notify"                  ,  0, KW_SCR_NOT , false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("scroll-offset"                  ,  0, KW_SCR_OFFS, false),
         new Keyword("scroll-right"                   ,  0, KW_SCR_RT  , false),  // missing in keyword index, found in prog handbook as a key function
         new Keyword("scroll-to-current-row"          ,  0, KW_SCR_2CR , false),
         new Keyword("scroll-to-item"                 , 11, KW_SCR_2ITM, false),
         new Keyword("scroll-to-selected-row"         ,  0, KW_SCR_2SR , false),
         new Keyword("scrollable"                     ,  0, KW_SCROLLBL, false),
         new Keyword("scrollbar-horizontal"           , 11, KW_SCROLL_H, false),
         new Keyword("scrollbar-vertical"             , 11, KW_SCROLL_V, false),
         new Keyword("scrolled-row-position"          , 16, KW_SCR_RPOS, false),
         new Keyword("scrolling"                      ,  0, KW_SCROLLIN, false),
         new Keyword("sdbname"                        ,  0, KW_SDBNAME , true ),
         new Keyword("seal"                           ,  0, KW_SEAL    , false),
         new Keyword("seal-timestamp"                 ,  0, KW_SEAL_TST, false),
         new Keyword("search"                         ,  0, KW_SEARCH  , true ),
         new Keyword("search-self"                    ,  0, KW_SEAR_SLF, true ),
         new Keyword("search-target"                  ,  0, KW_SEAR_TRG, true ),
         new Keyword("section"                        ,  0, KW_SECTION , false),
         new Keyword("security-policy"                ,  0, KW_SECUR_P , true ),
         new Keyword("seek"                           ,  0, KW_SEEK    , true ),
         new Keyword("select"                         ,  0, KW_SELECT  , true ),
         new Keyword("select-all"                     ,  0, KW_SEL_ALL , false),
         new Keyword("select-focused-row"             ,  0, KW_SEL_FOCR, false),
         new Keyword("select-next-row"                ,  0, KW_SEL_NEXT, false),
         new Keyword("select-prev-row"                ,  0, KW_SEL_PREV, false),
         new Keyword("select-row"                     ,  0, KW_SEL_ROW , false),
         new Keyword("selectable"                     ,  0, KW_SELECTBL, false),
         new Keyword("selected"                       ,  0, KW_SELECTED, false),
         new Keyword("selection"                      ,  0, KW_SELECTN , false),
         new Keyword("selection-end"                  ,  0, KW_SEL_END , false),
         new Keyword("selection-list"                 ,  0, KW_SEL_LST , false),
         new Keyword("selection-start"                ,  0, KW_SEL_STRT, false),
         new Keyword("selection-text"                 ,  0, KW_SEL_TXT , false),
         new Keyword("self"                           ,  0, KW_SELF    , true ),
         new Keyword("send"                           ,  0, KW_SEND    , false),
         new Keyword("send-sql-statement"             ,  8, KW_SEND_SQL, false),
         new Keyword("sensitive"                      ,  0, KW_SENSITIV, false),
         new Keyword("separate-connection"            ,  0, KW_SEP_CONN, false),
         new Keyword("separator-fgcolor"              ,  0, KW_SEP_FGC , false),
         new Keyword("separators"                     ,  0, KW_SEPS    , false),
         new Keyword("serialize-hidden"               ,  0, KW_SERIALZH, false),
         new Keyword("serialize-name"                 ,  0, KW_SERIALZN, false),
         new Keyword("serialize-row"                  ,  0, KW_SERIALZR, false), // missing in keyword index, found elsewhere in lang ref, UNTESTED
         new Keyword("server"                         ,  0, KW_SERVER  , false),
         new Keyword("server-connection-bound"        ,  0, KW_SRV_C_B , false),
         new Keyword("server-connection-bound-request",  0, KW_SRV_C_BR, false),
         new Keyword("server-connection-context"      ,  0, KW_SRV_C_C , false),
         new Keyword("server-connection-id"           ,  0, KW_SRV_C_I , false),
         new Keyword("server-operating-mode"          ,  0, KW_SRV_OP_M, false),
         new Keyword("server-socket"                  ,  0, KW_SRV_SOCK, false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("session"                        ,  0, KW_SESSION , true ),
         new Keyword("session-end"                    ,  0, KW_SESS_END, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("session-id"                     ,  0, KW_SESSN_ID, false),
         new Keyword("set"                            ,  0, KW_SET     , true ),
         new Keyword("settings"                       ,  0, KW_SETTINGS, false),  // missing in keyword index, found in prog handbook as a key function
         new Keyword("setuserid"                      ,  7, KW_SETUSER , true ),
         new Keyword("set-actor"                      ,  0, KW_SET_ACTR, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("set-appl-context"               ,  0, KW_SET_ACTX, false),
         new Keyword("set-attr-call-type"             ,  0, KW_SET_A_CT, true ),
         new Keyword("set-attribute"                  ,  0, KW_SET_ATTR, false),
         new Keyword("set-attribute-list"             ,  0, KW_SET_A_L , false),  // not a real keyword, but is normally defined in the PSC webspeed 4GL code
         new Keyword("set-attribute-node"             ,  0, KW_SET_A_N , false),
         new Keyword("set-blue-value"                 ,  8, KW_SET_BLUE, false),
         new Keyword("set-break"                      ,  0, KW_SET_BRK , false),
         new Keyword("set-buffers"                    ,  0, KW_SET_BUF , false),
         new Keyword("set-byte-order"                 ,  0, KW_SET_B_OR, false),
         new Keyword("set-callback"                   ,  0, KW_SET_CBAC, false),
         new Keyword("set-callback-procedure"         ,  0, KW_SET_CB_P, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("set-client"                     ,  0, KW_SET_CLNT, false),
         new Keyword("set-commit"                     ,  0, KW_SET_COMM, false),
         new Keyword("set-connect-procedure"          ,  0, KW_SET_C_P , false),
         new Keyword("set-contents"                   ,  0, KW_SET_CONT, false),
         new Keyword("setcookie"                      ,  0, KW_SET_COOK, false),  // not a real keyword, but is normally defined in the PSC webspeed 4GL code
         new Keyword("set-cookie"                     ,  0, KW_SET_COOK, false),  // not a real keyword, but is normally defined in the PSC webspeed 4GL code
         new Keyword("set-db-client"                  ,  0, KW_SET_DBCL, false),
         new Keyword("set-dynamic"                    ,  0, KW_SET_DYN , false),
         new Keyword("set-event-manager-option"       ,  0, KW_SET_EVMO, false),
         new Keyword("set-green-value"                ,  9, KW_SET_GRN , false),
         new Keyword("set-input-source"               ,  0, KW_SET_ISRC, false),
         new Keyword("set-must-understand"            ,  0, KW_SET_M_UN, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("set-node"                       ,  0, KW_SET_NODE, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("set-numeric-format"             ,  0, KW_SET_N_F , false),
         new Keyword("set-option"                     ,  0, KW_SET_OPT , false),
         new Keyword("set-output-destination"         ,  0, KW_SET_ODST, false),
         new Keyword("set-parameter"                  ,  0, KW_SET_PARM, false),
         new Keyword("set-property"                   ,  0, KW_SET_PROP, false),
         new Keyword("set-pointer-value"              ,  0, KW_SET_PTR , false),
         new Keyword("set-read-response-procedure"    ,  0, KW_SET_RRP , false),
         new Keyword("set-red-value"                  ,  7, KW_SET_RED , false),
         new Keyword("set-repositioned-row"           ,  0, KW_SET_RPOS, false),
         new Keyword("set-rgb-value"                  ,  0, KW_SET_RGB , false),
         new Keyword("set-rollback"                   ,  0, KW_SET_ROLL, false),
         new Keyword("set-selection"                  ,  0, KW_SET_SEL , false),
         new Keyword("set-serialized"                 ,  0, KW_SET_SERD, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("set-size"                       ,  0, KW_SET_SZ  , false),
         new Keyword("set-socket-option"              ,  0, KW_SET_S_O , false),
         new Keyword("set-sort-arrow"                 ,  0, KW_SET_S_AR, false),
         new Keyword("set-transaction-state"          ,  0, KW_SET_T_S , false),  // not a real keyword, but is normally defined in the PSC webspeed 4GL code
         new Keyword("set-user-field"                 ,  0, KW_SET_U_F , false),  // not a real keyword, but is normally defined in the PSC webspeed 4GL code
         new Keyword("set-wait-state"                 ,  8, KW_SET_WAIT, false),  // undocumented abbreviation found in customer source
         new Keyword("setwebstate"                    ,  0, KW_SET_W_S , false),  // not a real keyword, but is normally defined in the PSC webspeed 4GL code
         new Keyword("sha1-digest"                    ,  0, KW_SHA1_DIG, false),
         new Keyword("share-lock"                     ,  5, KW_SH_LOCK , true ),
         new Keyword("shared"                         ,  0, KW_SHARED  , true ),
         new Keyword("short"                          ,  0, KW_SHORT   , false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("show-in-taskbar"                ,  0, KW_SHOW_ITB, false),
         new Keyword("show-stats"                     ,  9, KW_SHOW_ST , true ),
         new Keyword("side-labels"                    ,  8, KW_SIDE_L  , false),
         new Keyword("side-label-handle"              , 12, KW_SIDE_L_H, false),
         new Keyword("signature"                      ,  0, KW_SIGNATUR, false),
         new Keyword("silent"                         ,  0, KW_SILENT  , false),
         new Keyword("simple"                         ,  0, KW_SIMPLE  , false),
         new Keyword("single"                         ,  0, KW_SINGLE  , false),
         new Keyword("size"                           ,  0, KW_SIZE    , false),
         new Keyword("size-chars"                     ,  6, KW_SIZE_C  , false),
         new Keyword("size-pixels"                    ,  6, KW_SIZE_P  , false),
         new Keyword("skip"                           ,  0, KW_SKIP    , true ),
         new Keyword("skip-deleted-record"            ,  0, KW_SKIP_D_R, true ),
         new Keyword("slider"                         ,  0, KW_SLIDER  , false),
         new Keyword("small-icon"                     ,  0, KW_SMAL_ICO, false),
         new Keyword("small-title"                    ,  0, KW_SMAL_TTL, false),
         new Keyword("smallint"                       ,  0, KW_SMALLINT, false),
         new Keyword("soap-header"                    ,  0, KW_SOAP_HDR, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("soap-header-entryref"           ,  0, KW_SOAP_HER, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("soap-fault"                     ,  0, KW_SOAP_F  , false),  // missing from keyword index and UNTESTED at this time
         new Keyword("soap-fault-actor"               ,  0, KW_SOAP_F_A, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("soap-fault-code"                ,  0, KW_SOAP_F_C, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("soap-fault-detail"              ,  0, KW_SOAP_F_D, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("soap-fault-string"              ,  0, KW_SOAP_F_S, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("socket"                         ,  0, KW_SOCKET  , false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("some"                           ,  0, KW_SOME    , true ),
         new Keyword("sort"                           ,  0, KW_SORT    , false),
         new Keyword("sort-ascending"                 ,  0, KW_SORT_ASC, false),
         new Keyword("sort-number"                    ,  0, KW_SORT_NUM, false),
         new Keyword("source"                         ,  0, KW_SOURCE  , false),
         new Keyword("source-procedure"               ,  0, KW_SRC_PROC, false),
         new Keyword("space"                          ,  0, KW_SPACE   , true ),
         new Keyword("sql"                            ,  0, KW_SQL     , false),
         new Keyword("sqrt"                           ,  0, KW_SQRT    , false),
         new Keyword("ssl-server-name"                ,  0, KW_SSL_SRVN, false),
         new Keyword("standalone"                     ,  0, KW_STANDALN, false),
         new Keyword("start"                          ,  0, KW_START   , false),
         new Keyword("start-box-selection"            ,  0, KW_START_BS, false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("start-move"                     ,  0, KW_START_MV, false),
         new Keyword("start-document"                 ,  0, KW_START_DC, false),
         new Keyword("start-element"                  ,  0, KW_START_EL, false),
         new Keyword("start-resize"                   ,  0, KW_START_RS, false),
         new Keyword("start-row-resize"               ,  0, KW_START_RR, false),
         new Keyword("start-search"                   ,  0, KW_START_SC, false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("starting"                       ,  0, KW_STARTING, false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("startup-parameters"             ,  0, KW_STUP_PAR, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("state-detail"                   ,  0, KW_STAT_DET, false),
         new Keyword("static"                         ,  0, KW_STATIC  , false),
         new Keyword("status"                         ,  0, KW_STATUS  , true ),
         new Keyword("status-area"                    ,  0, KW_STATUS_A, false),
         new Keyword("status-area-font"               ,  0, KW_STAT_A_F, false),
         new Keyword("stdcall"                        ,  0, KW_STDCALL , false),
         new Keyword("stop"                           ,  0, KW_STOP    , false),
         new Keyword("stop-after"                     ,  0, KW_STOP_AFT, true),
         new Keyword("stop-display"                   ,  0, KW_STOP_D  , false),  // missing in keyword index, found in prog handbook as a key function
         new Keyword("stop-parsing"                   ,  0, KW_STOP_PRS, false),
         new Keyword("stopped"                        ,  6, KW_STOPPED , false),
         new Keyword("stored-procedure"               , 11, KW_STORPROC, false),
         new Keyword("stream"                         ,  0, KW_STREAM  , true ),
         new Keyword("stream-handle"                  ,  0, KW_STRM_HND, true ),
         new Keyword("stream-io"                      ,  0, KW_STRM_IO , true ),
         new Keyword("stretch-to-fit"                 ,  0, KW_ST_2_FIT, false),
         new Keyword("strict"                         ,  0, KW_STRICT  , false),
         new Keyword("string-value"                   ,  0, KW_STR_VAL , false),
         new Keyword("string-xref"                    ,  0, KW_STR_XREF, false),
         new Keyword("string"                         ,  0, KW_STRING  , false),
         new Keyword("sub-average"                    ,  7, KW_SUB_AVG , false),
         new Keyword("sub-count"                      ,  0, KW_SUB_CNT , false),
         new Keyword("sub-maximum"                    ,  7, KW_SUB_MAX , false),
         new Keyword("sub-menu"                       ,  4, KW_SUB_MENU, false),
         new Keyword("sub-menu-help"                  ,  0, KW_SUB_M_H , false),
         new Keyword("sub-minimum"                    ,  7, KW_SUB_MIN , false),
         new Keyword("sub-total"                      ,  0, KW_SUB_TOT , false),
         new Keyword("subscribe"                      ,  0, KW_SUBSCRIB, false),
         new Keyword("subtype"                        ,  0, KW_SUBTYPE , false),
         new Keyword("sum"                            ,  0, KW_SUM     , false),
         new Keyword("substitute"                     ,  5, KW_SUBSTIT , false),
         new Keyword("substring"                      ,  6, KW_SUBSTR  , false),
         new Keyword("super"                          ,  0, KW_SUPER   , true ),  // not documented as reserved, but in practice it is treated as such
         new Keyword("super-procedures"               ,  0, KW_SUP_PROC, false),
         new Keyword("suppress-namespace-processing"  ,  0, KW_SUP_NS_P, false),
         new Keyword("suppress-warnings"              , 10, KW_SUP_WARN, false),
         new Keyword("symmetric-encryption-algorithm" ,  0, KW_SYM_EN_A, false),
         new Keyword("symmetric-encryption-iv"        ,  0, KW_SYM_EN_I, false),
         new Keyword("symmetric-encryption-key"       ,  0, KW_SYM_EN_K, false),
         new Keyword("symmetric-support"              ,  0, KW_SYM_SUPP, false),
         new Keyword("synchronize"                    ,  0, KW_SYNCHRON, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("system-alert-boxes"             , 12, KW_SYS_A_B , false),
         new Keyword("system-dialog"                  ,  0, KW_SYS_DLG , true ),
         new Keyword("system-help"                    ,  0, KW_SYS_HELP, false),
         new Keyword("system-id"                      ,  0, KW_SYS_ID  , false),
         new Keyword("tab"                            ,  0, KW_TAB     , false),
         new Keyword("tab-position"                   ,  0, KW_TAB_POS , false),
         new Keyword("tab-stop"                       ,  0, KW_TAB_STOP, false),
         new Keyword("table"                          ,  0, KW_TABLE   , true ),
         new Keyword("table-crc-list"                 ,  0, KW_TAB_CRCL, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("table-list"                     ,  0, KW_TAB_LIST, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("table-handle"                   ,  0, KW_TAB_HAND, false),
         new Keyword("table-number"                   ,  0, KW_TAB_NUM , true ),
         new Keyword("table-scan"                     ,  0, KW_TAB_SCAN, false),
         new Keyword("tag"                            ,  0, KW_TAG     , false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("target"                         ,  0, KW_TARGET  , false),
         new Keyword("target-procedure"               ,  0, KW_TAR_PROC, false),
         new Keyword("temp-directory"                 ,  8, KW_TEMP_DIR, false),
         new Keyword("temp-table"                     ,  0, KW_TEMP_TAB, false),
         new Keyword("temp-table-prepare"             ,  0, KW_TT_PREP , false),
         new Keyword("term"                           ,  0, KW_TERM    , true ),  // non-abbreviated synonym for TERMINAL
         new Keyword("terminal"                       ,  0, KW_TERM    , true ),  // the keyword index incorrectly states that TERMINAL can be abbreviatated down to a minimum TERM, this is not correct, there is no abbreviation possible (just the synonym TERM)
         new Keyword("terminate"                      ,  0, KW_TERMINAT, false),
         new Keyword("text"                           ,  0, KW_TEXT    , true ),
         new Keyword("text-cursor"                    ,  0, KW_TXT_CURS, true ),
         new Keyword("text-seg-grow"                  ,  0, KW_TXT_SEG , false),
         new Keyword("text-selected"                  ,  0, KW_TXT_SEL , false),
         new Keyword("then"                           ,  0, KW_THEN    , true ),
         new Keyword("this-object"                    ,  0, KW_THIS_OBJ, true ),
         new Keyword("this-procedure"                 ,  0, KW_THIS_PRC, true ),
         new Keyword("three-d"                        ,  0, KW_3D      , false),
         new Keyword("thru"                           ,  0, KW_THROUGH , false),
         new Keyword("through"                        ,  0, KW_THROUGH , false),
         new Keyword("throw"                          ,  0, KW_THROW   , false),
         new Keyword("tic-marks"                      ,  0, KW_TIC_MARK, false),
         new Keyword("time"                           ,  0, KW_TIME    , true ),
         new Keyword("timer"                          ,  0, KW_TIMER   , false),  // FWD TIMER extension
         new Keyword("timezone"                       ,  0, KW_TIMEZONE, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("time-source"                    ,  0, KW_TIME_SRC, false),
         new Keyword("title"                          ,  0, KW_TITLE   , true ),
         new Keyword("title-bgcolor"                  ,  9, KW_TITL_BGC, false),
         new Keyword("title-dcolor"                   ,  8, KW_TITL_DC , false),
         new Keyword("title-fgcolor"                  ,  9, KW_TITL_FGC, false),
         new Keyword("title-font"                     ,  8, KW_TITL_FON, false),
         new Keyword("to"                             ,  0, KW_TO      , true ),
         new Keyword("today"                          ,  0, KW_TODAY   , false),
         new Keyword("toggle-box"                     ,  0, KW_TOGGL_BX, false),
         new Keyword("tooltip"                        ,  0, KW_TOOLTIP , false),
         new Keyword("tooltips"                       ,  0, KW_TOOLTIPS, false),
         new Keyword("top"                            ,  0, KW_TOP     , false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("top-column"                     ,  0, KW_TOP_COL , false),  // missing in keyword index, found in prog handbook as a key function
         new Keyword("top-nav-query"                  ,  0, KW_TOP_NAVQ, false),
         new Keyword("top-only"                       ,  0, KW_TOP_ONLY, true ),
         new Keyword("topic"                          ,  0, KW_TOPIC   , false),
         new Keyword("total"                          ,  0, KW_TOTAL   , false),
         new Keyword("to-rowid"                       ,  0, KW_TO_ROWID, true ),
         new Keyword("trace-filter"                   ,  0, KW_TRC_FILT, false),  // missing from keyword index, found in customer code and UNTESTED at this time
         new Keyword("tracking-changes"               ,  0, KW_TRAC_CHG, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("trailing"                       ,  0, KW_TRAILING, false),
         new Keyword("trans-init-procedure"           ,  0, KW_TRAN_I_P, false),
         new Keyword("trans"                          ,  0, KW_TRANS   , true ), // undocumented feature transa and transac are unreserved (just symbols) so we exclude them here by using 2 keyword defs to leave a loophole like in the 4GL
         new Keyword("transaction"                    ,  8, KW_TRANS   , true ),
         new Keyword("transaction-mode"               ,  0, KW_TRAN_MOD, false),
         new Keyword("transparent"                    ,  0, KW_TRANSPAR, false),
         new Keyword("trigger"                        ,  0, KW_TRIGGER , true ),
         new Keyword("triggers"                       ,  0, KW_TRIGGERS, true ),
         new Keyword("trim"                           ,  0, KW_TRIM    , true ),
         new Keyword("truncate"                       ,  5, KW_TRUNC   , false),
         new Keyword("true"                           ,  0, BOOL_TRUE  , true ),
         new Keyword("ttcodepage"                     ,  0, KW_TTCP    , false),  // missing from keyword index and UNTESTED at this time
         new Keyword("type"                           ,  0, KW_TYPE    , false),
         new Keyword("type-of"                        ,  0, KW_TYPE_OF , false),
         new Keyword("unbox"                          ,  6, KW_UNBOX   , false),
         new Keyword("unbuffered"                     ,  6, KW_UNBUF   , false),
         new Keyword("undo"                           ,  0, KW_UNDO    , true ),
         new Keyword("unique"                         ,  0, KW_UNIQUE  , true ),
         new Keyword("unix"                           ,  0, KW_UNIX    , true ),
         new Keyword("unix-end"                       ,  0, KW_UNIX_END, false),  // missing in keyword index, found in prog handbook as a key function
         new Keyword("unless-hidden"                  ,  0, KW_UNL_HID , false),
         new Keyword("unload"                         ,  0, KW_UNLOAD  , false),
         new Keyword("underline"                      ,  6, KW_UNDERLIN, true ),
         new Keyword("unformatted"                    ,  6, KW_UNFORMAT, true ),
         new Keyword("union"                          ,  0, KW_UNION   , true ),
         new Keyword("unique-id"                      ,  0, KW_UNIQ_ID , false),
         new Keyword("unique-match"                   ,  0, KW_UNIQ_MAT, false),
         new Keyword("unsigned-byte"                  ,  0, KW_UNS_BYTE, false),  // missing in keyword index, found elsewhere in external program interface ref
         new Keyword("unsigned-short"                 ,  0, KW_UNS_SHRT, false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("unsigned-long"                  ,  0, KW_UNS_LONG, false),
         new Keyword("unsubscribe"                    ,  0, KW_UNSUBSCR, false),
         new Keyword("up"                             ,  0, KW_UP      , true ),
         new Keyword("update"                         ,  0, KW_UPDATE  , true ),
         new Keyword("update-attribute"               ,  0, KW_UPD_ATTR, false),
         new Keyword("url"                            ,  0, KW_URL     , false),
         new Keyword("urldecode"                      ,  0, KW_URL_DECO, false),  // not a real keyword, but is normally defined in the PSC webspeed 4GL code
         new Keyword("url-decode"                     ,  0, KW_URL_DECO, false),
         new Keyword("urlencode"                      ,  0, KW_URL_ENCO, false),  // not a real keyword, but is normally defined in the PSC webspeed 4GL code
         new Keyword("url-encode"                     ,  0, KW_URL_ENCO, false),
         new Keyword("url-field"                      ,  0, KW_URL_FLD , false),  // not a real keyword, but is normally defined in the PSC webspeed 4GL code
         new Keyword("url-field-list"                 ,  0, KW_URL_FLDL, false),  // not a real keyword, but is normally defined in the PSC webspeed 4GL code
         new Keyword("url-format"                     ,  0, KW_URL_FMT , false),  // not a real keyword, but is normally defined in the PSC webspeed 4GL code
         new Keyword("url-password"                   ,  0, KW_URL_PW  , false),
         new Keyword("url-userid"                     ,  0, KW_URL_UID , false),
         new Keyword("user"                           ,  0, KW_USERID  , true ),
         new Keyword("userid"                         ,  0, KW_USERID  , true ),
         new Keyword("user-id"                        ,  0, KW_USER_ID , false),
         new Keyword("user-data"                      ,  0, KW_USR_DATA, false),  // missing in keyword reference, found in customer code
         new Keyword("use"                            ,  0, KW_USE     , false),
         new Keyword("use-dict-exps"                  ,  0, KW_USE_DCT , false),
         new Keyword("use-filename"                   ,  0, KW_USE_FIL , false),
         new Keyword("use-index"                      ,  0, KW_USE_IDX , true ),
         new Keyword("use-revvideo"                   ,  0, KW_USE_REV , false),
         new Keyword("use-text"                       ,  0, KW_USE_TXT , false),
         new Keyword("use-underline"                  ,  0, KW_USE_UND , false),
         new Keyword("use-widget-pool"                ,  0, KW_USE_WIDP, false),
         new Keyword("using"                          ,  0, KW_USING   , true ),
         new Keyword("utc-offset"                     ,  0, KW_UTC_OFF , false),
         new Keyword("v6display"                      ,  0, KW_V6DISP  , false),
         new Keyword("v6frame"                        ,  0, KW_V6FRAME , true ),
         new Keyword("validate-expression"            ,  0, KW_VAL_EXPR, false),
         new Keyword("validate-message"               ,  0, KW_VAL_MSG , false),
         new Keyword("validate-seal"                  ,  0, KW_VAL_SEAL, false),
         new Keyword("validate-xml"                   ,  0, KW_VAL_XML , false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("validate"                       ,  0, KW_VALIDATE, false),
         new Keyword("validation-enabled"             ,  0, KW_VAL_ENAB, false),
         new Keyword("valid-event"                    ,  0, KW_VAL_EVT , false),
         new Keyword("valid-handle"                   ,  0, KW_VAL_HND , false),
         new Keyword("valid-object"                   ,  0, KW_VAL_OBJ , false),
         new Keyword("value"                          ,  0, KW_VALUE   , true ),
         new Keyword("value-changed"                  ,  0, KW_VAL_CHG , true ),
         new Keyword("values"                         ,  0, KW_VALUES  , true ),
         new Keyword("variable"                       ,  3, KW_VAR     , false),
         new Keyword("verbose"                        ,  0, KW_VERBOSE , false),
         new Keyword("vertical"                       ,  4, KW_VERT    , false),
         new Keyword("version"                        ,  0, KW_VERSION , false),
         new Keyword("view"                           ,  0, KW_VIEW    , true ),
         new Keyword("view-as"                        ,  0, KW_VIEW_AS , true ),
         new Keyword("view-first-column-on-reopen"    ,  0, KW_VIEW_FCR, false),
         new Keyword("visible"                        ,  0, KW_VISIBLE , false),
         new Keyword("virtual-height-chars"           , 14, KW_VIRT_HC , false),
         new Keyword("virtual-height-pixels"          , 16, KW_VIRT_HP , false),
         new Keyword("virtual-width-chars"            , 13, KW_VIRT_WC , false),
         new Keyword("virtual-width-pixels"           , 15, KW_VIRT_WP , false),
         new Keyword("vms"                            ,  0, KW_VMS     , true ),  // undocumented
         new Keyword("void"                           ,  0, KW_VOID    , false),
         new Keyword("wait"                           ,  0, KW_WAIT_FOR, false),  // non-abbreviated synonym for WAIT-FOR
         new Keyword("wait-for"                       ,  0, KW_WAIT_FOR, true ),
         new Keyword("warning"                        ,  0, KW_WARN    , false),
         new Keyword("web"                            ,  0, KW_WEB     , false),  // missing in keyword index, found in WebSpeed reference
         new Keyword("web-context"                    ,  0, KW_WEB_CTX , false),
         new Keyword("web.output"                     ,  0, KW_WEB_OUT , false),  // not a real keyword, but is normally defined in the PSC webspeed 4GL code
         new Keyword("weekday"                        ,  0, KW_WEEK    , false),
         new Keyword("when"                           ,  0, KW_WHEN    , true ),
         new Keyword("where"                          ,  0, KW_WHERE   , true ),
         new Keyword("where-string"                   ,  0, KW_WHERE_ST, false),  // missing from keyword index and UNTESTED at this time
         new Keyword("while"                          ,  0, KW_WHILE   , true ),
         new Keyword("widget"                         ,  0, KW_WIDGET  , false),
         new Keyword("widget-enter"                   ,  8, KW_WID_ENT , false),
         new Keyword("widget-handle"                  ,  8, KW_WID_HAND, false),
         new Keyword("widget-id"                      ,  0, KW_WID_ID  , false),
         new Keyword("widget-leave"                   ,  8, KW_WID_LEAV, false),
         new Keyword("widget-pool"                    ,  0, KW_WID_POOL, false),
         new Keyword("width"                          ,  0, KW_WIDTH   , false),
         new Keyword("width-chars"                    ,  6, KW_WIDTH_C , false),
         new Keyword("width-pixels"                   ,  7, KW_WIDTH_P , false),
         new Keyword("window"                         ,  0, KW_WINDOW  , true ),
         new Keyword("window-close"                   ,  0, KW_WIN_CLOS, false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("window-delayed-minimize"        , 17, KW_WIN_DMIN, true ),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("window-maximized"               , 12, KW_WIN_MAX , true ),
         new Keyword("window-minimized"               , 12, KW_WIN_MIN , true ),
         new Keyword("window-name"                    ,  0, KW_WIN_NAME, false),
         new Keyword("window-normal"                  ,  0, KW_WIN_NORM, true ),
         new Keyword("window-resized"                 ,  0, KW_WIN_RESI, false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("window-restored"                ,  0, KW_WIN_REST, false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("window-state"                   , 10, KW_WIN_STAT, false),
         new Keyword("window-system"                  ,  0, KW_WIN_SYS , false),
         new Keyword("with"                           ,  0, KW_WITH    , true ),
         new Keyword("word-index"                     ,  0, KW_WORD_IDX, false),
         new Keyword("word-wrap"                      ,  0, KW_WORD_WRP, false),
         new Keyword("work-area-height-pixels"        ,  0, KW_WA_H_P  , false),
         new Keyword("work-area-width-pixels"         ,  0, KW_WA_W_P  , false),
         new Keyword("work-area-x"                    ,  0, KW_WA_X    , false),
         new Keyword("work-area-y"                    ,  0, KW_WA_Y    , false),
         new Keyword("workfile"                       ,  0, KW_WORK_TAB, true ),
         new Keyword("work-table"                     ,  8, KW_WORK_TAB, true ),
         new Keyword("write"                          ,  0, KW_WRITE   , true ),
         new Keyword("write-cdata"                    ,  0, KW_WR_CDATA, false),
         new Keyword("write-characters"               ,  0, KW_WR_CHARS, false),
         new Keyword("write-comment"                  ,  0, KW_WR_CMNT , false),
         new Keyword("write-data"                     ,  0, KW_WR_DATA,  false),  // missing in keyword reference, found in customer code
         new Keyword("write-data-element"             ,  0, KW_WR_D_ELM, false),
         new Keyword("write-empty-element"            ,  0, KW_WR_E_ELM, false),
         new Keyword("write-entity-ref"               ,  0, KW_WR_ENT_R, false),
         new Keyword("write-external-dtd"             ,  0, KW_WR_EXDTD, false),
         new Keyword("write-fragment"                 ,  0, KW_WR_FRAGM, false),
         new Keyword("write-json"                     ,  0, KW_WR_JSON , false),
         new Keyword("write-message"                  ,  0, KW_WR_MSG  , false),
         new Keyword("write-processing-instruction"   ,  0, KW_WR_PRINS, false),
         new Keyword("write-status"                   ,  0, KW_WR_STAT , false),
         new Keyword("write-xml"                      ,  0, KW_WR_XML  , false),
         new Keyword("write-xmlschema"                ,  0, KW_WR_XMLSC, false),
         new Keyword("x"                              ,  0, KW_X       , false),
         new Keyword("xcode"                          ,  0, KW_XCODE   , true ),
         new Keyword("xml-data-type"                  ,  0, KW_XML_DTYP, false),
         new Keyword("xml-node-name"                  ,  0, KW_XML_NNAM, false),  // missing in keyword index, found elsewhere in lang ref, TESTED
         new Keyword("xml-node-type"                  ,  0, KW_XML_NTYP, false),
         new Keyword("xml-schema-path"                ,  0, KW_XML_SCHP, false),
         new Keyword("xml-suppress-namespace-processing",0, KW_XML_SNSP, false),
         new Keyword("xref"                           ,  0, KW_XREF    , true ),
         new Keyword("xref-xml"                       ,  0, KW_XREF_XML, true ),
         new Keyword("x-document"                     ,  0, KW_X_DOC   , false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("x-noderef"                      ,  0, KW_X_NODE  , false),  // missing in keyword index, found elsewhere in lang ref
         new Keyword("x-of"                           ,  0, KW_X_OF    , false),
         new Keyword("year"                           ,  0, KW_YEAR    , false),
         new Keyword("year-offset"                    ,  0, KW_YEAR_OFF, false),
         new Keyword("yes"                            ,  0, BOOL_TRUE  , true ),
         new Keyword("yes-no"                         ,  0, KW_YES_NO  , false),
         new Keyword("yes-no-cancel"                  ,  0, KW_YES_NO_C, false),
         new Keyword("y"                              ,  0, KW_Y       , false),
         new Keyword("y-of"                           ,  0, KW_Y_OF    , false) 
      };
   }
   
   /**
    * Constructs a lexer using a reader as input and associating a 
    * <code>SymbolResolver</code> with the instance. The reader will be
    * wrapped in a {@link DotKludgeReader} instance.
    *
    * @param    in
    *           The reader representing the Progress 4GL source code.
    * @param    sr
    *           Provides keyword dictionary storage and lookup for symbol
    *           resolution in the lexer.
    */
   public ProgressLexer(Reader in, SymbolResolver sr) 
   {
      this(in, sr, true);
   }
   
   /**
    * Constructs a lexer using a reader as input and associating a 
    * <code>SymbolResolver</code> with the instance.
    *
    * @param    in
    *           The reader representing the Progress 4GL source code.
    * @param    sr
    *           Provides keyword dictionary storage and lookup for symbol
    *           resolution in the lexer.
    * @param    kludge
    *           <code>true</code> to force the reader to be wrapped in a
    *           {@link DotKludgeReader} instance.
    */
   public ProgressLexer(Reader in, SymbolResolver sr, boolean kludge) 
   {
      this(new CharBuffer(kludge ? new DotKludgeReader(in) : in));
      sym = sr;
      initializeKeywordDictionary();
   }
   
   /**
    * Constructs a lexer using stream as input and associating a 
    * <code>SymbolResolver</code> with the instance.  The stream will be
    * wrapped in a {@link DotKludgeStream} instance.
    *
    * @param    in
    *           The stream representing the Progress 4GL source code.
    * @param    sr
    *           Provides keyword dictionary storage and lookup for symbol
    *           resolution in the lexer.
    */
   public ProgressLexer(InputStream in, SymbolResolver sr) 
   {
      this(in, sr, true);
   }
   
   /**
    * Constructs a lexer using stream as input and associating a 
    * <code>SymbolResolver</code> with the instance.
    *
    * @param    in
    *           The stream representing the Progress 4GL source code.
    * @param    sr
    *           Provides keyword dictionary storage and lookup for symbol
    *           resolution in the lexer.
    * @param    kludge
    *           <code>true</code> to force the stream to be wrapped in a
    *           {@link DotKludgeStream} instance.
    */
   public ProgressLexer(InputStream in, SymbolResolver sr, boolean kludge) 
   {
      this(new ByteBuffer(kludge ? new DotKludgeStream(in) : in));
      sym = sr;
      initializeKeywordDictionary();
   }
   
   /**
    * Get the keyword list.
    *
    * @return   The list of all recognized keywords.
    */
   public static Keyword[] getKeywords()
   {
      return keywords;
   }
   
   /**
    * Get the keyword ignore list.
    *
    * @return   The list of all ignored keywords.
    */
   public static Keyword[] getIgnoredKeywords()
   {
      return ignored;
   }
   
   /**
    * Remove each of the given keywords from the keyword dictionary. This
    * matches the behavior of the Progress "keyword forget" list (the -k
    * parameter on startup).
    *
    * @param   names
    *          Keywords to remove.
    */
   public void ignoreKeywords(String[] names)
   {
      ArrayList<Keyword> list = new ArrayList<Keyword>();
   
      for (int i = 0; i < names.length; i++)
      {
         Keyword found = sym.lookupKeyword(names[i]);
         
         if (found != null)
         {
            list.add(found);
            sym.removeKeyword(found);
         }
         else
         {
            String err = String.format("Cannot ignore unknown keyword '%s'.",
                                       names[i]);
         
            // TODO: log this?
            System.out.println(err);
         }
      }
      
      if (!list.isEmpty())
      {
         ignored = list.toArray(new Keyword[0]);
      }
   }
   
   /** 
    * Reports on the state of debug in the lexer.
    *
    * @return    Debug state.
    */
   public boolean isDebug()
   {
      return debug;
   }
   
   /** 
    * Sets the state of debug in the lexer.
    *
    * @param    dbg
    *           Debug state.
    */
   public void setDebug( boolean dbg )
   {
      debug = dbg;
   }
   
   /**
    * Reports on whether schema-specific processing is activated in the lexer.
    *
    * @return    Schema-specific processing activation state.
    */
   public boolean isSchema()
   {
      return schema;
   }
   
   /**
    * Activates schema-specific processing in the lexer.  At this time the
    * only difference in output is in the recognized keywords.  This is a
    * 'one way' state change and cannot be reversed!
    */
   public void activateSchemaProcessing()
   {
      schema = true;
      initializeSchemaKeywords();
   }   
   
   /**
    * Reports on whether export-specific processing is activated in the lexer.
    *
    * @return    Export-specific processing activation state.
    */
   public boolean isExport()
   {
      return export;
   }
   
   /**
    * Activates export-specific processing in the lexer.  This is used
    * specifically by semantic predicate logic in the lexer DSTRING rule
    * to handle export mode-specific processing which differs from the
    * standard string rules.  This is a 'one way' state change and cannot
    * be reversed!
    */
   public void activateExportMode()
   {
      export = true;
   }   
   
   /**
    * Reports on whether hidden stream processing is activated in the lexer.
    *
    * @return    Hidden stream processing activation state.
    */
   public boolean isHidden()
   {
      return hidden;
   }
   
   /**
    * Activates hidden stream processing in the lexer. This sets the returned  
    * token class to <code>TokenStreamHiddenTokenFilter</code> and disables
    * the lexer's normal 'skip' processing of certain token types.
    * This is a 'one way' state change and cannot be reversed!
    */
   public void activateHiddenProcessing()
   {
      hidden = true;
      setTokenObjectClass("antlr.CommonHiddenStreamToken");
   }   
   
   /**
    * Turns on/off raw string processing mode.  The mode may be changed an
    * arbitrary number of times.
    *
    * @param     raw
    *            Raw processing mode:  <code>true</code> to turn on,
    *            <code>false</code> to turn off.
    */
   public void setRaw(boolean raw)
   {
      this.raw = raw;
   }
   
   /**
    * Reports on whether the lexer is currently in raw string processing
    * mode.
    *
    * @return    <code>true</code> if in raw string mode, else
    *            <code>false</code>.
    */
   public boolean isRaw()
   {
      return raw;
   }
   
   /**
    * Sets whether the lexer is currently honoring the backslash as
    * an escape character.
    *
    * @param     unixEscapes
    *            <code>true</code> if in UNIX escape mode, else
    *            <code>false</code>.
    */
   public void setUnixEscapes(boolean unixEscapes)
   {
      this.unixEscapes = unixEscapes;
   }
   
   /**
    * Reports on whether the lexer is currently honoring the backslash as
    * an escape character.
    *
    * @return    <code>true</code> if in UNIX escape mode, else
    *            <code>false</code>.
    */
   public boolean isUnixEscapes()
   {
      return unixEscapes;
   }
   
   /**
    * Allow this instance to cleanup any resources that were allocated.
    */
   public void close()
   {
      // nothing to do
   }
   
   /**
    * Process the keyword array to add each <code>Keyword</code> into the
    * keyword dictionary. This method is called during construction so that
    * all lexer processing can have keyword lookup facilities available
    * from the first token.  The keyword array is a static member initialized
    * during class load.
    */
   private void initializeKeywordDictionary()
   {
      for (int i = 0; i < keywords.length; i++)
      {
         sym.addKeyword(keywords[i]);
      }
   }
   
   /**
    * Creates a <code>Keyword</code> array and initializes that array to
    * a the list of all Progress schema-specific keywords. The
    * initialization is done in-line and the source file is organized in
    * an easy to maintain column format.  A loop then processes this array
    * and adds each <code>Keyword</code> into the keyword dictionary. This
    * method is called during construction so that all lexer processing can
    * have keyword lookup facilities available from the first token.
    * <p>
    * The following values are needed for each <code>Keyword</code> object:
    * <ul>
    *    <li> full keyword text for the keyword
    *    <li> the minimum number of characters for matching (abbreviations
    *         will always be disabled so this value is set to 0)
    *    <li> the integer token type to assign to the keyword on a match
    *         (see below)
    *    <li> whether it is a reserved keyword or not (these keywords are
    *         never reserved so this value is always <code>false</code>)
    * </ul>
    * <p>
    * See <code>{@link Keyword}</code> for more details.
    * <p>
    * The core idea here is to populate a facility where the symbol text (in
    * the mSYMBOL method) can be submitted to the keyword dictionary for a
    * lookup. If a match is found, then the token type of the token is set
    * based on the token type field in the <code>Keyword</code> object
    * found in the dictionary.  This returned token type will override the
    * default token type of <code>SYMBOL</code>.  Since all of these keywords
    * are not encoded in the matching rules of the grammar itself, there are
    * no actual token definitions in the lexer that match these keywords.
    * For this reason, we take advantage of the feature in the parser in
    * which we can create (reserve) a unique integer value and associate it 
    * with a specific artificial token name (e.g. KW_DEFINE).  These
    * artificial token types are used as the values when constructing
    * <code>Keyword</code> objects.
    * <p>
    * When adding <code>Keyword</code> objects to this array, make sure to
    * add the associated artificial token type to the tokens { } section
    * of the parser.
    */
   private void initializeSchemaKeywords()
   {
      Keyword[] schemaKeywords = 
      {
         /* format:  Full Keyword Text   , Abr, Token Type     , Reserved */
         new Keyword("abbreviated"       ,   0, KW_ABBV        , false),
         new Keyword("area"              ,   0, KW_AREA        , false),
         new Keyword("bigint"            ,   0, KW_BIGINT      , false),
         new Keyword("buffer-pool"       ,   0, KW_BUFRPOOL    , false),
         new Keyword("bufpool"           ,   0, KW_BUFPOOL     , false),
         new Keyword("clob-codepage"     ,   0, KW_CLOB_CP     , false),
         new Keyword("clob-collation"    ,   0, KW_CLOB_COL    , false),
         new Keyword("clob-type"         ,   0, KW_CLOB_TYP    , false),
         new Keyword("crc"               ,   0, KW_CRC         , false),
         new Keyword("cycle-on-limit"    ,   0, KW_CYCLE       , false),
         new Keyword("dump-name"         ,   0, KW_DMP_NAME    , false),
         new Keyword("field-trigger"     ,   0, KW_FLD_TRG     , false),
         new Keyword("fixchar"           ,   0, KW_FIXCHAR     , false),
         new Keyword("frozen"            ,   0, KW_FROZEN      , false),
         new Keyword("index-field"       ,   0, KW_IDX_FLD     , false),
         new Keyword("inactive"          ,   0, KW_INACTIVE    , false),
         new Keyword("increment"         ,   0, KW_INCR        , false),
         new Keyword("lob-area"          ,   0, KW_LOB_AREA    , false),
         new Keyword("lob-bytes"         ,   0, KW_LOB_BYTE    , false),
         new Keyword("lob-size"          ,   0, KW_LOB_SIZE    , false),
         new Keyword("no-override"       ,   0, KW_NO_OVRRD    , false),
         new Keyword("not-case-sensitive",   0, KW_NOT_CS      , false),
         new Keyword("null-allowed"      ,   0, KW_NULLALWD    , false),
         new Keyword("psc"               ,   0, KW_PSC         , false),
         new Keyword("sequence"          ,   0, KW_SEQUENCE    , false),
         new Keyword("sql-width"         ,   0, KW_SQL_WID     , false),
         new Keyword("table-trigger"     ,   0, KW_TAB_TRG     , false),
         new Keyword("timestamp"         ,   0, KW_TIMESTMP    , false),
         new Keyword("word"              ,   0, KW_WORD_IDX    , false),
         new Keyword("valexp"            ,   0, KW_VALEXP      , false),
         new Keyword("valmsg"            ,   0, KW_VALMSG      , false)
      };
      
      for (int i = 0; i < schemaKeywords.length; i++)
      {
         sym.addKeyword(schemaKeywords[i]);
      }
   }
   
   /**
    * Provides a command line interface for an end user to drive and/or test
    * the ProgressLexer class.
    * <p>
    * Syntax:
    * <pre>
    *    java ProgressLexer &lt;Progress_4GL_source_file&gt;
    * </pre>
    * Where:
    * <ul>
    *   <li> &lt;Progress_4GL_source_filename&gt; is the filename of a
    *        preprocessed Progress 4GL source file (enclose the name in
    *        double quotes if spaces or other reserved command line
    *        characters are part of the filename)
    * </ul>
    *
    * @param   args 
    *          List of command line arguments.
    */
   public static void main(String[] args)  
   {
      String syntax = "Syntax: java ProgressLexer <Progress_4GL_source_file>";
      
      if (args.length != 1)
      {
         System.err.println(syntax);
         System.exit(-1);
      }
      
      try
      {
         // setup
         SymbolResolver    sym    = new SymbolResolver(true);
         FileReader        fr     = new FileReader(args[0]);
         BufferedReader    br     = new BufferedReader(fr);
         LexerDumpFilter   test   = new LexerDumpFilter(br, sym);
         
         test.setOutput(new PrintWriter(System.out));
         test.activateHiddenProcessing();
         test.setBypassSkip(true);
                           
         // start the lexing and print a record for each token                  
         test.dump();
      }
       
      catch (Exception excpt)
      {
         excpt.printStackTrace();
      }
   }   
}

/**
 * Matches any amount of whitespace in a program and sets the token type to
 * &quot;skip&quot;.  In addition, each newline character causes the lexer's
 * line counter to be incremented in order to properly maintain each token's 
 * line information.
 * <p>
 * Spaces, tabs and carriage returns and line feeds (newlines) are all
 * matched.  Although the preprocessor generally removes carriage returns,  
 * if the input file was to include a ~015 (escaped carriage return), it
 * would be placed in the output file.  For safety, we must leave the
 * carriage return processing in this rule to ensure that any such matches
 * can be consumed as whitespace.  Normally such escaped carriage returns 
 * would only be found inside strings, but in complex preprocessor
 * replacements, it is very possible that a coding error could leave such
 * characters outside a string in some cases.  Progress 4GL is tolerant of
 * such escaped carriage returns and thus our whitespace rule must also
 * handle this condition.
 * <p>
 * Non-visible ASCII codes decimal 127 and below are matched here as well. 
 * <p>
 * This is a top level lexer rule which means that there is an associated
 * <code>WS</code> token.
 * <p>
 * If the {@link #activateHiddenProcessing} is called, then tokens generated
 * using this rule will not be skipped, but instead will flow through to
 * the caller.
 */
WS
   :
      ( ' ' | '\t' | '\r' | '\n' { newline(); } | JUNK )+
      { if (!hidden) $setType( Token.SKIP ); }
   ;

/**
 * Matches any tilde character '~' (that is not in a string or comment) and
 * sets the token type to <code>skip</code> which drops such characters from
 * the token stream. This is consistent with how Progress handles such 
 * characters.
 * <p>
 * This is a top level lexer rule which means that there is an associated
 * <code>TILDE</code> token, however these will never be seen by the parser.
 * <p>
 * If the {@link #activateHiddenProcessing} is called, then tokens generated
 * using this rule will not be skipped, but instead will flow through to
 * the caller.
 */
TILDE
   :
      '~'
      { if (!hidden) $setType( Token.SKIP ); }
   ;

/**
 * Matches an entire comment, including any nested comments.  The resulting
 * token is set to &quot;skip&quot; so that the parser will not ever see
 * comments in the token stream.
 * <p>
 * The lexer's internal line counter is updated whenever a newline character
 * is encountered.
 * <p>
 * Of great importance is the fact that the closure subrule for the contents
 * of a comment <b>must not operate on a greedy basis</b>.  Being greedy
 * means that the loop will consume all possible matches to its rule
 * without regard to the end of loop condition.  This is the default way
 * that the lexer works, but since this loop can match any character (we
 * use the wildcard character because any character can be embedded in a
 * comment), a greedy loop would not ever be able to detect its end
 * condition.  Instead we turn off greedy mode and this makes the loop
 * watch for the matching comment close sequence of characters.  Note that
 * nested comments are possible by recursively referencing this same rule
 * inside the comment contents subrule.
 * <p>
 * Whitespace is maintained inside the resulting comment text and the lexer's
 * line counter is properly incremented upon encountering a newline.
 * <p>
 * This is a top level lexer rule which means that there is an associated 
 * <code>COMMENT</code> token. 
 * <p>
 * If the {@link #activateHiddenProcessing} is called, then tokens generated
 * using this rule will not be skipped, but instead will flow through to
 * the caller.
 */
COMMENT
   :
      CMT_OPEN
      (
         options 
         {
            greedy = false;
            generateAmbigWarnings = false;
         }
         :
           '\n'               { newline(); }
         | COMMENT
         | .
      )*
      CMT_CLOSE
       
      { if (!hidden) $setType( Token.SKIP ); }
   ;

/**
 * Matches various string types depending upon the processing mode:
 * <ul>
 *    <li>in <b>raw</b> processing mode, matches a double-quoted string
 *        encoded string which represents encoded binary (i.e., raw) data,
 *        setting the token type to <code>RAW_STRING</code>;  otherwise
 *    <li>in <b>export</b> processing mode, matches a double-quoted string
 *        which represents a Progress string as it would appear at runtime,
 *        except that embedded double quotes are doubled up;  otherwise
 *    <li>matches a single or double quoted string followed by any (optional) 
 *        string options.
 * </ul>
 * <p>
 * This is a top level lexer rule which means that there is an associated 
 * <code>STRING</code> token. 
 *
 * @see    "mRSTRING"
 * @see    "mXSTRING"
 * @see    "mSSTRING"
 * @see    "mDSTRING"
 * @see    "mSTR_OPTIONS" 
 */
STRING
   :
      (
           { raw }? RSTRING { $setType(RAW_STRING); }
         | { export }? XSTRING
         | ( SSTRING | DSTRING ) (STR_OPTIONS)?
      )
   ;

/**
 * Match any valid symbol name in the Progress 4GL language or the standalone
 * <code>DIVIDE</code> operator '/'.  This includes the following:
 * <ul>
 *    <li> divide sign '/'
 *    <li> underscore '_' (does not have a token type of its own, it is a
 *         generic symbol if found on its own)
 *    <li> language keywords (reserved and unreserved)
 *    <li> variables (global shared, scoped shared and local)
 *    <li> database, table and field names
 *    <li> user-defined functions
 *    <li> built-in functions of the Progress 4GL
 *    <li> internal procedure names
 *    <li> labels
 *    <li> path and filenames (which include any valid symbol character
 *         as well as '/' or '\' or '.' or '..') which supports absolute
 *         and relative filenames including the special references to
 *         the current directory and the parent directory
 *    <li> single '\' is matched but then discarded (it is never seen by
 *         the parser because its type is set to <code>SKIP</code>
 *    <li> r-code library references in the form 
 *         <code>pathname&lt;&lt;membername&gt;&gt;</code>
 *    <li> other user-defined names (e.g. frames, temp-tables etc.)
 * </ul>
 * <p>
 * All symbols must start with an alphabetic character.  Starting in the
 * second character and continuing to a maximum of 32 characters (including
 * the initial character), can be any alphabetic and numeric characters or
 * one of the special symbol characters (<code> # $ % &amp; - _ </code>).  
 * <b>Symbols are matched case-insensitively.</b>
 * <p>
 * An exception to the previous rule that a symbol must start with an
 * alphabetic character is for special meta-data entries in the Progress 
 * schema dictionary, whose table and field names start with an underscore.
 * For this reason, such entries are allowed more generally such that the
 * rule as implemented is: the first letter of every symbol must be an 
 * alphabetic character or an underscore.
 * <p>
 * Another exception to the previous rule that a symbol must start with an
 * alphabetic character is for a filename. Filenames <b>may also</b> start
 * with the '/' or '\' characters (in which case they are absolute filenames).
 * The second and following characters in filenames can include '/', '\' or
 * '.' in addition to all the special symbol characters listed above.
 * <p>
 * In addition, DOS and Windows filenames can start with a drive letter
 * followed by a ':' as in 'c:'.  This construct is matched as a symbol
 * followed by a colon.  This is needed to allow single character variable
 * names + attribute or method calls to properly work in the parser. The
 * difficulty with putting the whole thing together again into a single
 * filename is left to the {@link ProgressParser#filename} rule.
 * <p>
 * <b>Note that there are many characters that are not allowed in filenames
 * at this time.  This is an artificial limitation designed to keep the 
 * current lexer simple and compatible with the known/common cases of
 * filenames while not introducing too many potential ambiguities.  
 * Examples of characters that are not matched include: '~', '!', '@', '^',
 * '+', '='.  The general rule: if the character is not a valid symbol char,
 * '/', '\', '.' or '..' then it is not matched.</b>  This can be changed if
 * necessary, but the decision was made to defer such handling until the
 * need arises. When a filename is detected the token type is set to
 * <code>FILENAME</code>.  Both relative and absolute filenames are matched 
 * by this rule (except for the drive letter component as noted above).
 * <p>
 * Since the '/' character is matched in this rule, an inherent ambiguity
 * was created with any top level <code>DIVIDE</code> rule that matches '/'.
 * For this reason, the ambiguity was resolved by matching on a standalone
 * '/' in this rule and setting the resulting token type to 
 * <code>DIVIDE</code>.  The top level <code>DIVIDE</code> rule was 
 * removed.
 * <p>
 * The '\' is a special case in Progress.  This character can be inserted
 * anywhere and it is simply ignored.  This feature is supported in this
 * rule as well, but matching a standalone backslash and setting it's type
 * to <code>SKIP</code>.  Note that a backslash that is part of a larger
 * filename, is NOT skipped.
 * <p>
 * To avoid lexer non-determinism, any filename that starts with './', '../',
 * '.\' or '..\' is actually matched at the top level by 
 * {@link #mNUM_LITERAL}.  This is needed because the <code>DOT</code> in the
 * first position conflicts.  So instead, once <code>mNUM_LITERAL</code>
 * matches one of these strings, it includes this rule optionally.  This
 * allows all normal filename processing to occur as needed but doesn't  
 * require it since the previous string can legitimately stand alone.
 * <p>
 * R-code library references are matched with the format 
 * <code>pathname&lt;&lt;membername&gt;&gt;</code>.  This allows any filename
 * to be specified as a prefix and then the library member name followed
 * as a suffix, but contained in the unambiguous double angle brackets. 
 * When this construction is matched, the token type is set to
 * <code>LIBRARY_REF</code>.
 * <p>
 * A special case exists for partially or fully qualified database symbols.
 * Such symbols are comprised of 2 or 3 regular symbols, each separated
 * by a '.' and no whitespace.  This method recognizes such optional
 * constructions and sets the token type to <code>DB_SYMBOL</code>.  In
 * order to make this extra check work, the optional section (following the
 * first or main symbol definition) is differentiated using a semantic
 * predicate. This only allows the database symbol matching if the following
 * conditions hold:
 * <ul>
 *    <li> a <code>DOT</code> character is found after the symbol without
 *         any intervening whitespace
 *    <li> the next character after the <code>DOT</code> is an alphabetic 
 *         character (it can't be numeric or a special symbol because each
 *         part of the qualified database symbol name must follow the
 *         standard rules (see above) for symbol naming
 * </ul>
 * <p>
 * There is a second optional section nested inside the first one, which
 * handles the condition where there is a fully qualified field name.  The
 * resulting possible constructions are:
 * <pre>
 * user-defined symbol or unqualified db/table/field name    --&gt; sym
 * fully qualified table or partially qualified field name   --&gt; sym.sym
 * fully qualified field name                                --&gt; sym.sym.sym
 * </pre>
 * <p>
 * Any unqualified database, table or field names will be lexed as a standard
 * symbol (which is the default token type). But in any case where a simple
 * symbol is followed by a '.' AND an alphabetic character with no 
 * intervening whitespace, this is known to be a qualified database name and
 * it <b>cannot</b> be a generic symbol or language keyword.  <b>If there is
 * ever a testcase found that does not follow this rule, this method will
 * break, badly!</b>
 * <p>
 * It is also important to note that any relative filename that does not
 * include any path separator ('/' or '\') and which does not have any more
 * than 2 '.' characters will be matched as a <code>DB_SYMBOL</code> due to
 * inherent ambiguity in the Progress decision to use the '.' as the
 * database qualifier.  This means that the parser must be aware that some
 * filenames can appear as database symbols.
 * <p>
 * Most importantly, once a symbol has been found, a keyword lookup occurs
 * and the token type is overridden from the default (<code>SYMBOL</code>)
 * to the artificial token type associated with the keyword (see
 * <code>initializeKeywordDictionary</code>).  The parser may subsequently
 * override this token type yet again, but this is the extent of the lexer's
 * symbol resolution.  <b>Keyword token types take precedence over the
 * default token type <code>SYMBOL</code>, however only <code>SYMBOL</code>
 * types are overridden since there are no circumstances in which there are
 * language keywords that can match database symbols, filenames or library
 * references.</b>  All
 * <code>DIVIDE, DB_SYMBOL, FILENAME and LIBRARY_REF</code> token types
 * bypass the keyword lookup and override.
 * <p>
 * See the overview for a discussion of how and why the symbol resolution
 * in this method is limited to keywords and the 
 * <code>SYMBOL, DIVIDE, DB_SYMBOL, FILENAME and LIBRARY_REF</code> token
 * types.
 * <p>
 * This is a top level lexer rule which means that there is an associated 
 * <code>SYMBOL</code> token. The
 * <code>DIVIDE, DB_SYMBOL, FILENAME and LIBRARY_REF</code> tokens are
 * artifically created in the parser so that they can be manually set
 * here as an override.
 * <p>
 * Standalone backslashes are set to <code>skip</code> by default.  However,
 * if the {@link #activateHiddenProcessing} is called, then tokens generated
 * using this rule will not be skipped, but instead will flow through to
 * the caller.
 *
 * @see    "mLETTER"
 * @see    "mDIGIT"
 * @see    "mSYM_CHAR"
 * @see     Keyword
 * @see     KeywordDictionary
 * @see     SymbolResolver
 */
SYMBOL
   :
      {
         int stype = SYMBOL;
      }         
      (
         options { generateAmbigWarnings = false; }
         :
           '/' VALID_SYM_CHAR    { stype = FILENAME; }
         | '/'                   { stype = DIVIDE; }
         | '\\' VALID_SYM_CHAR   { stype = FILENAME; }
         | '\\'                  { stype = BACKSLASH; }         
         | LETTER
         | '_'
      )
      (
         options { generateAmbigWarnings = false; }
         :
           // primary alternative for the most common case
           VALID_SYM_CHAR
           
           // a backslash followed by a slash is invalid
         | { stype != BACKSLASH }?
           '/'                   { stype = FILENAME; }
         | "./"                  { stype = FILENAME; }
         | "../"                 { stype = FILENAME; }
         | '\\'                  { stype = FILENAME; }
         | ".\\"                 { stype = FILENAME; }
         | "..\\"                { stype = FILENAME; }
         
           // this alternative MUST use this disambiguating semantic
           // predicate to resolve the ambiguity between a symbol followed by
           // a '.' and a partially or fully qualified database symbol; by
           // putting this predicate in, the optional section is bypassed
           // unless any following dot is also followed by a valid char
           // --> ANTLR would not check for this otherwise before entering
           // this block, at which point we are already matching and can't
           // back out (it would be a false positive)
           // valid chars are currently set to all possible matched chars in
           // this loop, note that this excludes many valid filename chars
           // so this may need to be improved or changed later
         | {
              (LA(2) >= 'a' && LA(2) <= 'z') ||
              (LA(2) >= '0' && LA(2) <= '9') ||
              LA(2)  == '#' || LA(2) == '$'  ||
              LA(2)  == '%' || LA(2) == '&'  ||
              LA(2)  == '-' || LA(2) == '_'
           }?
           '.'                   { if (stype == SYMBOL) stype = DB_SYMBOL; }
      )*
      (
           // this library reference symbol alternative MUST use this
           // disambiguating semantic predicate to resolve the ambiguity
           // between a symbol followed by a '<' (possibly an expression) and
           // a symbol followed by "<<" which indicates the library reference
           // form of path-name<<member-name>> to specify a r-code library
           // and a specific member of that library; by putting this predicate
           // in, the optional section is bypassed unless there are 2
           // sequential '<' characters --> ANTLR would not check for this
           // otherwise before entering this block, at which point we are
           // already matching and can't back out (it would be a false
           // positive)
           { LA(2) == '<' }?
           "<<" LETTER (VALID_SYM_CHAR | '.')* ">>"
           { stype = LIBRARY_REF; }
      )?
      
      {
         // standalone backslash chars must be ignored (they can be 
         // arbitrarily inserted in Progress code and make no difference)
         if ( stype == BACKSLASH )
         {
            if (getText().equals("\\"))
            {
               if (!hidden)
               {
                  // ignore the standalone backslash
                  stype = Token.SKIP;
               }
            }
            else
            {
               // backslash is not a valid token to be seen in the parser
               // in this case, it must really be a filename
               stype = FILENAME;
            }
         }
      
         // in order to handle filenames (which can include the '.' char)
         // we had to make the DB_SYMBOL matching more generic since there
         // is no way to know in the lexer, whether a simple "hello.p" is
         // a DB_SYMBOL or a FILENAME; for this reason we add this test to
         // force something to a FILENAME token type based on finding
         // a value that couldn't be in a DB_SYMBOL
         if (stype == DB_SYMBOL)
         {
            int     count   = 0;
            int     index   = 0;
            boolean invalid = false;
            String  txt     = $getText;
            
            while ( index != -1 )
            {
               index = txt.indexOf('.', index+1);
               if (index != -1) 
               {
                  // this next code works because we know that there will 
                  // never be a '.' as the last char of a symbol (not legal
                  // in Progress because this would be ambiguous with '.'
                  // as an end of statement delimiter) --> if this ever
                  // changes, this code will throw an 
                  // java.lang.StringIndexOutOfBoundsException
                  char ch = (char) txt.charAt(index+1);
                  if (Character.isLetter(ch) == false && ch != '_')
                     invalid = true;
                  count++;
               }
            }

            if (invalid || count > 2)
               stype = FILENAME;
         }
         
         // only generic symbols need to be overridden by possible keyword
         // matches; partially/fully qualified database object (table or
         // field) names OR filenames OR the DIVIDE operator OR r-code library
         // references cannot be keywords, so we bypass this logic
         if (stype == SYMBOL)
         {
            // non-database path which could be a keyword, we bias toward
            // this outcome due to the keyword-rich design of Progress 4GL
            Keyword found = sym.lookupKeyword(getText());
            
            if (found != null)
            {
               // we will *assume* we are a keyword (we may actually be a
               // user-defined symbol, but we will handle that in the parser
               stype = found.getTokenType();
            }
         }
         $setType(stype);
      }
   ;   
   
/**
 * Matches the ':' character. This is a top level rule, creating a
 * <code>COLON</code> token type.
 * <p>
 * The hidden token stream can be used to determine whether the colon was
 * followed by whitespace or not.  This is needed in the parser to
 * differentiate between labels (which must always be followed by whitespace)
 * and attributes/methods where the colon cannot be followed by whitespace.
 */
COLON    : ':' ;

/**
 * Matches the '::' non-static database reference "operator". This is a top
 * level rule, creating a <code>DB_REF_NON_STATIC</code> token type.
 * <p>
 * The hidden token stream can be used to determine whether the colon was
 * followed by whitespace or not.  This is needed in the parser to
 * differentiate between labels (which must always be followed by whitespace)
 * and attributes/methods where the colon cannot be followed by whitespace.
 */
DB_REF_NON_STATIC : "::" ;

/**
 * Matches the ',' character. This is a top level rule, creating a
 * <code>COMMA</code> token type.
 */
COMMA    : ',' ;

/**
 * Matches the '=' character. This is a top level rule, creating a
 * <code>EQUALS</code> token type.
 */
EQUALS   : '=' ;

/**
 * Matches the '&lt;&gt;' character sequence. This is a top level rule, creating a
 * <code>NOT_EQ</code> token type.
 */
NOT_EQ   : "<>" ;

/**
 * Matches the '&gt;' character. This is a top level rule, creating a
 * <code>GT</code> token type.
 */
GT       : '>' ;

/**
 * Matches the '&lt;' character. This is a top level rule, creating a
 * <code>LT</code> token type.
 */
LT       : '<' ;

/**
 * Matches the '&gt;=' character sequence. This is a top level rule, creating
 * a <code>GTE</code> token type.
 */
GTE      : ">=" ;

/**
 * Matches the '&lt;=' character sequence. This is a top level rule, creating
 * a <code>LTE</code> token type.
 */
LTE      : "<=" ;

/**
 * Matches the '(' character. This is a top level rule, creating a
 * <code>LPARENS</code> token type.
 */
LPARENS  : '(' ;

/**
 * Matches the ')' character. This is a top level rule, creating a
 * <code>RPARENS</code> token type.
 */
RPARENS  : ')' ;

/**
 * Matches the '[' character. This is a top level rule, creating a
 * <code>LBRACKET</code> token type.
 */
LBRACKET : '[' ;

/**
 * Matches the ']' character. This is a top level rule, creating a
 * <code>RBRACKET</code> token type.
 */
RBRACKET : ']' ;

/**
 * Matches the '*' character. This is a top level rule, creating a
 * <code>MULTIPLY</code> token type.
 */
MULTIPLY : '*' ;

/**
 * Matches the '@' character.  This is a top level rule, creating an 
 * <code>AT</code> token type.
 */
AT       : '@' ;

/**
 * Matches the '^' character.  This is a top level rule, creating an 
 * <code>CARET</code> token type.
 */
CARET    : '^' ;

/**
 * Matches the unknown value '?'. This is a top level rule, creating a
 * <code>UNKNOWN_VAL</code> token type.
 */
UNKNOWN_VAL : '?' ;

/**
 * Matches all forms of valid integer literals, decimal literals and due to
 * ambiguity, the period character (<code>DOT</code> token), all date
 * literals, the minus sign (<code>MINUS</code> token), the plus sign
 * (<code>PLUS</code>) and all <code>FILENAME</code> tokens that begin with
 * '.' or '..' must be matched here as well.
 * <p>
 * This is a very special method that encapsulates and resolves all ambiguity
 * caused by the different usage of '.' as a decimal point and also as a
 * statement separator.  Likewise it handles the ambiguity presented by the
 * traditional minus/plus operators and the prefixing/postfixing of a
 * negative or positive sign on a numeric or decimal literal.  Finally, it
 * also handles the ambiguity associated with the use of '-' as a separator
 * in a date literal.
 * <p>
 * Scenarios that are matched:
 * <ul>
 *    <li> Matches the '.' character creating a <code>DOT</code> token type.
 *    <li> Matches the following strings as a <code>FILENAME</code> token
 *         type (the following {@link #mSYMBOL} is optional) :
 *       <ul>
 *          <li> './' <code>mSYMBOL</code>
 *          <li> '../' <code>mSYMBOL</code>
 *          <li> '.\' <code>mSYMBOL</code>
 *          <li> '..\' <code>mSYMBOL</code>
 *       </ul>
 *    <li> Matches the '-' character creating a <code>MINUS</code> token type
 *         (if this is not followed by digits or a decimal point and digits).
 *    <li> Matches the '-' character followed by digits as a
 *         <code>NUM_LITERAL</code>.
 *    <li> Matches the '-' character followed by digits and then a '.' and
 *         more digits as a <code>DEC_LITERAL</code>.
 *    <li> Matches the '-' character followed by a '.' and digits as a
 *         <code>DEC_LITERAL</code> token type.
 *    <li> Matches the '+' character creating a <code>PLUS</code> token type
 *         (if this is not followed by digits or a decimal point and digits).
 *    <li> Matches the '+' character followed by digits as a
 *         <code>NUM_LITERAL</code>. The '+' is discarded as syntactic sugar.
 *    <li> Matches the '+' character followed by digits and then a '.' and
 *         more digits as a <code>DEC_LITERAL</code>. The '+' is discarded as
 *         syntactic sugar.
 *    <li> Matches the '+' character followed by a '.' and digits as a
 *         <code>DEC_LITERAL</code> token type. The '+' is discarded as
 *         syntactic sugar.
 *    <li> Matches a sequence of one or more digits (without a decimal point
 *         and additional digits) as an integer literal with a
 *         <code>NUM_LITERAL</code> token type.  There can be an optional
 *         leading negative/positive sign (the cases noted above). If no
 *         leading sign is present, then an optional postfixed sign (negative
 *         or positive) can be present.
 *    <li> Matches a sequence of one or more digits followed by a '.' but
 *         no additional digits, as an integer literal (it does not consume
 *         the following '.' in this case), with a <code>NUM_LITERAL</code> 
 *         token type. There can be an optional leading negative/positive
 *         sign (the cases noted above).
 *    <li> Matches one or more numeric digits followed by a decimal point
 *         and another sequence of numeric digits as a decimal constant 
 *         with a <code>DEC_LITERAL</code> token type.  There can be an
 *         optional leading negative/positive sign (the cases noted above).
 *         If no leading sign is present, then an optional postfixed sign
 *         (negative or positive) can be present.
 *    <li> Matches a decimal point followed by a sequence of numeric digits
 *         as a decimal constant with a <code>DEC_LITERAL</code> token type.
 *         There can be an optional leading negative/positive sign (the cases
 *         noted above). If no leading sign is present, then an optional
 *         postfixed sign (negative or positive) can be present.
 *    <li> Matches digits followed by a '-', '/' or '.' and one or two digits 
 *         as a <code>DATE_LITERAL</code>. If there are no following
 *         digits after the separator, then it is not matched as a date,
 *         instead it is a integer literal followed by an operator.  The
 *         second separator for the year portion is optional, as is the year
 *         itself and there is an optional negative sign for BC dates. The
 *         second separator can also be any of the '-', '/' or '.' characters
 *         and the first and second separators can be any combination of
 *         these characters (the two separators don't have to be the same
 *         character in a single date literal).
 * </ul>
 * <p>
 * The key to this rule is that there cannot be any intervening whitespace
 * between digits and decimal points or between digits and the date separators
 * '-' or '/'.  Likewise, there cannot be any intervening whitespace between
 * a leading or trailing '-' or '+' and the numeric or decimal literal it
 * precedes or follows respectively. In such a case the '-' is lexed as a
 * <code>MINUS</code> token, the '+' is lexed as a <code>PLUS</code> token
 * and the parser must use context to determine if it is a unary or binary
 * operator.
 * <p>
 * By handling the case of '.' (by itself) here instead of in a (more
 * intuitive) dedicated top level rule, the ambiguity between decimal literals
 * and the standalone dot can be properly managed.  This is also true of the
 * '-' processing which would be easier and more intuitive in a top level
 * rule, but which needs to be here to resolve ambiguity.
 * <p>
 * In the case of a decimal with no digits on the right side of the decimal
 * point, this is always interpreted as an integer literal AND a subsequent
 * 2nd token of a <code>DOT</code>.  This oddity cannot be easily resolved
 * BUT interestingly enough this is <b>exactly</b> how Progress itself
 * handles this situation AND this is a perfectly valid match since it
 * essentially is an integer.
 * <p>
 * Date literals are generally matched in mm/dd/yyyy or mm-dd-yyyy formats,
 * although the '-' and '/' separators can even be mixed in the same literal
 * (mm-dd/yyyy or mm/dd-yyyy). Both months and days can be specified in 1 or 
 * 2 digits (actually this rule matches any number of digits for months). 
 * Either separator character can be used, although in Progress one cannot
 * use the '-' between the mm and dd if the constant is being used in an
 * expression.  This lexer has no such limitation.  In Progress the month
 * must be an integer between 1 and 12 but this limit is not implemented 
 * here. In Progress the day must be between 1 and the number of days in that
 * month, though this limit is not implemented here. The second separator and
 * the year digits are optional. There can be an optional negative sign
 * preceding any year digits and one can specify a dd and yyyy separator
 * without specifying any year digits at all.  Any number of year digits are
 * valid, though in Progress the year 0 is treated specially, as follows:
 * <ul>
 *    <li> Valid (all resolve to the year 2000)
 *    <ul>
 *       <li> 0
 *       <li> -0
 *       <li> 00
 *    </ul>
 *    <li> Invalid (all complain about specifying the year 0)
 *    <ul>
 *       <li> 000
 *       <li> 0000
 *       <li> 00000 (and so on...)
 *       <li> -00
 *       <li> -000 (and so on...)
 *    </ul>
 * </ul>
 * <p>
 * So y, yy, yyy, yyyy, yyyyy... and -y, -yy, -yyy, -yyyy... all are valid
 * year specifications, except for the specific year zero issue noted above.
 * These limitations on how the year digits work are not implemented here.
 * Over all, this rule supports a superset of the possible date literal
 * formats in Progess and no checking of the values is done.
 * <p>
 * WARNING: at this time there is no support for a quirk of Progress date
 * literals in which constructs like 1/1--- are valid dates equivalent
 * to 01/01/00.
 * <p>
 * This is a top level rule.  The token type defaults to 
 * <code>NUM_LITERAL</code> and is overridden by specific actions depending
 * on the scenario matched.
 */
NUM_LITERAL
   :
      {
         int     ntype = NUM_LITERAL;
         int     digs  = 0;
         boolean sign  = false;
      }
      (  
           // decimal number with no digits (or sign char) on left of decimal
           // point and with an optional postfixed sign
           '.' (DIGIT)+ ('-' | '+'!)?             { ntype = DEC_LITERAL; }
           
           // these next rules must be here because otherwise there is
           // lexer nondeterminism between this rule and mSYMBOL (warning:
           // the rules that match a slash will potentially also conflict
           // with the top level COMMENT rule, so this is checked and takes
           // precedence --> in this case we just match DOT below)
         | { LA(3) != '*' }?
           "./" (SYMBOL)?                         { ntype = FILENAME; }
         | { LA(4) != '*' }?
           "../" (SYMBOL)?                        { ntype = FILENAME; }
         | ".\\" (SYMBOL)?                        { ntype = FILENAME; }
         | "..\\" (SYMBOL)?                       { ntype = FILENAME; }
         
           // the all important DOT!
         | '.'                                    { ntype = DOT; }
          
           // datetime-tz literal
         | {
             (LA(1) >= '0' && LA(1) <= '9')   &&
             (LA(2) >= '0' && LA(2) <= '9')   &&
             (LA(3) >= '0' && LA(3) <= '9')   &&
             (LA(4) >= '0' && LA(4) <= '9')   &&
             LA(5) == '-'                     &&
             (LA(6) >= '0' && LA(6) <= '9')   &&
             (LA(7) >= '0' && LA(7) <= '9')   &&
             LA(8) == '-'                     &&
             (LA(9) >= '0' && LA(9) <= '9')   &&
             (LA(10) >= '0' && LA(10) <= '9') &&
             (LA(11) == 't' || LA(11) == 'T') &&
             (LA(12) >= '0' && LA(12) <= '9') &&
             (LA(13) >= '0' && LA(13) <= '9') &&
             LA(14) == ':' &&
             (LA(15) >= '0' && LA(15) <= '9') &&
             (LA(16) >= '0' && LA(16) <= '9')
           }?
           DIGIT DIGIT DIGIT DIGIT '-' DIGIT DIGIT '-' DIGIT DIGIT
           't' // implies 'T' also, as caseSensitive = false
           DIGIT DIGIT ':' DIGIT (DIGIT)?
           (':' DIGIT (DIGIT)? 
                  (
                     // avoid matching a following "lone" dot (no digits after it)
                     { LA(1) >= '.' && LA(2) >= '0' && LA(2) <= '9' }? ('.' (DIGIT)+ )
                     | // empty alternative
                  )
           )? // seconds and millis are optional
           { ntype = DATETIME_LITERAL; }
           (
            {
              (LA(1) == '+' || LA(1) == '-') &&
              (LA(2) >= '0' && LA(2) <= '9')
            }? // if timezome offset is present then it's a DATETIME_TZ_LITERAL 
            (('+'|'-') DIGIT (DIGIT)? (':' DIGIT (DIGIT)? )?  { ntype = DATETIME_TZ_LITERAL; })
           )?
         
           // num literal with a preceding unary plus sign or a decimal
           // literal with a preceding positive sign character 
         | '+'! (DIGIT)+
           (
                // avoid matching a following "lone" dot (no digits after it)
                { (LA(2) >= '0' && LA(2) <= '9') }?
                '.' (DIGIT)+                      { ntype = DEC_LITERAL; }
              | // empty alternative
           )
         
           // all the rest of the num literals, all date literals and the
           // common case for a decimal literal with digits on both sides of
           // the decimal point; handles an optional prefixed negative sign;
           // this is the most complicated part of this rule and is required
           // to be so complicated due to the ambiguity it resolves
         | {
             (LA(1) >= '0' && LA(1) <= '9') ||
             (LA(1) == '-' && (LA(2) >= '0' && LA(2) <= '9')) 
           }?
           
           // handle the optional prefixed negative sign
           ( '-' { sign = true; } )?
           
           // core num literal matching (or left side of the decimal point
           // for decimal literals or leftmost date portion for date literals) 
           (DIGIT)+ 
           
           // optional decimal literal or date literal processing
           (
                // BOTH the 1st two alternatives MUST use this disambiguating
                // semantic predicate EVEN THOUGH IT IS THE SAME FOR BOTH!
                // The ambiguity that is resolved is not related to the
                // choice of one alternative versus the other, rather it
                // differentiates between the valid cases of matching a
                // '.' or '/' or '-' char with following digits versus the
                // false positive whenever one or more contiguous digits are
                // followed by a lone '.' or '/' or '-'
                
                // this can be the back-end of a decimal literal or it can be
                // the middle portion of a date literal
                { (LA(2) >= '0' && LA(2) <= '9') }?
                '.' ( DIGIT { digs++; } )+         { ntype = DEC_LITERAL; }
                      
                (
                     // year portion of a date literal that has a '.' as
                     // its first separator (between the month and day
                     // portions)
                     {
                        digs < 3 &&
                        (LA(1) != '.' || 
                         (LA(2) == '-' || (LA(2) >= '0' && LA(2) <= '9')))
                     }?
                     (
                          '/'   { ntype = DATE_LITERAL; }
                        | '-'   { ntype = DEC_LITERAL; }
                        | '.'   { ntype = DATE_LITERAL; }
                     )
                     ('-')? (DIGIT { ntype = DATE_LITERAL; } )*
                     
                     // this can be a DEC_LITERAL when the 2nd "separator" is '-' and
                     // there are no following year digits; this is how the 4GL also
                     // treats this case (e.g. 999.99- or 12.31-) EXCEPT in a variable
                     // definition initializer of a date type, then these same forms
                     // (if their month and day values are valid) are treated as a date; 
                     // in that case, the parser will rewrite the type so we don't try
                     // to handle that here
                     
                   | ('-' | '+'!)
                )?
                
              | { (LA(2) >= '0' && LA(2) <= '9') }?
              
                // middle portion of a date literal
                ('/' | '-') DIGIT (DIGIT)?
                
                // year portion of a date literal
                (
                   {
                      LA(1) != '.' || 
                      (LA(2) == '-' || (LA(2) >= '0' && LA(2) <= '9'))
                   }?
                   (
                        '/'
                      | '-'
                      | '.'
                   )
                   ('-')? (DIGIT)*
                )?
                
                { ntype = DATE_LITERAL; }
                
                // handle a postfixed sign char but only if no leading sign
                // was present and we aren't a decimal or date literal
              | { !sign }?
                ('-' | '+'!)
              | // empty alternative
           )
           
           // must be included to avoid ambiguity at the top level, handles
           // both the minus and plus operators in addition to the case of
           // a decimal literal with no leading dignits but which has a
           // prefixed sign char (this must follow the date literal 
           // processing due to issues with '-' ambiguity
         | (
               '-'                                { ntype = MINUS; }
             | '+'                                { ntype = PLUS; }
           )
           (
                // avoid matching a following "lone" dot (no digits after it)
                { (LA(2) >= '0' && LA(2) <= '9') }?
                '.' (DIGIT)+                      { ntype = DEC_LITERAL; }
              | // empty alternative
           )
      )
      {
         // this is a case where we can't drop the '+' until after we know it
         // is part of a decimal literal, since it might just be a plus sign
         if (ntype == DEC_LITERAL)
         {
            String txt = getText();
            
            if (txt.charAt(0) == '+')
            {
               // get rid of the useless unary plus
               setText(txt.substring(1));
            }
         }
         
         $setType(ntype);
      }
   ;
   
/**
 * Matches other unexpected characters and generates a token rather than
 * throwing an exception (this rule must be changed when matched characters
 * are moved to another top level rule).  Note that this rule is designed to
 * match the unusual characters that can't be normally used in a Progress 4GL
 * source program, but might appear in a command token list for a shell 
 * command that is embedded in a procedure. There are many other possible 
 * characters in the extended ASCII range that are not handled here. Likewise
 * the other control characters (other than whitespace) are not matched. The
 * list of matched characters:
 * <p>
 * <pre>
 * '`' 
 * '!' 
 * '#'
 * '$'
 * '%' 
 * '&amp;'
 * '{'
 * '}'
 * '|' 
 * ';'
 * </pre>
 * <p>
 * Anything that is not matched here will still trigger an exception. Note
 * that ANTLR's filter feature seems to be a better approach to resolve this
 * problem (of wanting to generate an <code>UNKNOWN_TOKEN</code> rather
 * than throw an exception), however it can only handle the consumption of
 * such characters.  The filter rule that is called when there is no top level
 * match cannot generate tokens because of the way that <code>nextToken</code>
 * is generated by ANTLR.  Such filters are not allowed to generate a token
 * so this less-optimal approach of specifying the special characters in
 * advance is the result.
 */
UNKNOWN_TOKEN
   :
      '`' | '!' | '#' | '$' | '%' | '&' | '{' | '}' | '|' | ';'
   ;
   
/**   
 * Matches any single character that can appear in the 2nd or later position 
 * in a symbol (matches any <code>LETTER, DIGIT or SYM_CHAR</code>). This is
 * simply a helper rule to make references easier.
 */
protected   
VALID_SYM_CHAR
   :
      LETTER | DIGIT | SYM_CHAR
   ;
   
/**
 * Matches an opening single quote, arbitrary contents and an ending single
 * quote.  Two contiguous single quote characters and an escape prefixed
 * single quote character are accepted as contents (they do not terminate
 * the string).
 * <p>
 * Any newlines inside the string are identified and the lexer's internal
 * newline counter is properly maintained.  Note that all such newlines
 * are maintained in the output string because these are the escaped chars
 * that have been left behind by the preprocessor.  If such characters are not
 * escaped, then the preprocessor removes such chars (carriage returns and
 * line feeds).  This is how Progress 4GL handles these characters.  This
 * means that in a raw source file that has not been preprocessed, a string
 * literal can be split across any number of lines and the Progress 
 * preprocessor will put the string back together, ignoring the carriage
 * returns and newlines.
 * <p>
 * Tilde and the backslash can BOTH be Progress escape characters and as such,
 * they need special attention. Backslash is only honored if UNIX escapes
 * mode is set on. In particular, this rule must separately match 3 or 6
 * constructs as part of a string depending on which escape sequences are
 * honored.
 * <p>
 * <pre>
 *    ~'
 *    \'
 *    ~~
 *    ~\
 *    \\
 *    \~
 *    ~
 *    \
 * </pre>
 * <p>
 * The first 2 are a way of embedding a quote in a string.  The next 4 are
 * important because if this rule were to encounter '~~', '\\', '~\' or '\~'
 * (which are valid strings) the rule would consume the first escape and then
 * encounter an escaped quote which would not terminate the string properly,
 * leading to a non-ending string.  So this rule matches on any escaped escape
 * char to eliminate this situation.  Finally, a single ~ or \ that is not
 * followed by a &quot; or a duplicate escape char is matched as a single
 * character.  This is required (one may think that the closure rule
 * should handle this case) because when ANTLR sees a ~ or \ in the leftmost
 * position of the alternatives, it DOES NOT include it in the list of
 * 'everything that is not a closing quote'.  Likewise \n is not included.
 * Of course, these tilde constructions must be placed in a specific order
 * and if so, the ambiguity warnings that ANTLR reports can be disabled.
 * <p>
 * The greedy option does not need to be disabled here as the closure rule
 * termination is built into the subrule itself: it accepts anything that
 * isn't an unescaped single quote character (see above).
 * <p>
 * Tabs and spaces are maintained inside strings.
 */
protected
SSTRING
   :
      (
         '\''                              // opening '
              (
                 options { generateAmbigWarnings = false; }
                 :
                   "~\'"                   // allow ~'
                 | "~~"                    // allow ~~ (consume as one)
                 | "~\\"                   // allow ~\ (consume as one)
                 | '~'                     // allow ~ by itself
                 | { unixEscapes }?
                   "\\\'"                  // allow \'
                 | { unixEscapes }?
                   "\\\\"                  // allow \\ (consume as one)
                 | { unixEscapes }?
                   "\\~"                   // allow \~ (consume as one)
                 | '\\'                    // allow \ by itself
                 | "\'\'"                  // allow ''
                 
                 | { schema }?             // convert CRNL to a space char
                   "\r\n"!                 // when lexing a DF file
                   {
                      append(' ');
                      newline();
                   }
                 | { schema }?             // convert NL or CR to a space
                   ('\n'! | '\r'!)         // char when lexing a DF file
                   {
                      append(' ');
                      newline();
                   }
                 | '\n'     { newline(); } // otherwise allow NL
                 
                 | ~'\''                   // anything else except closing '
              )*
         '\''                              // closing '
      )
   ;

/**
 * Matches an opening double quote, arbitrary contents and an ending double
 * quote.  Two contiguous double quote characters and an escape prefixed
 * double quote character are accepted as contents (they do not terminate
 * the string).
 * <p>
 * Any newlines inside the string are identified and the lexer's internal
 * newline counter is properly maintained.  Note that all such newlines
 * are maintained in the output string because these are the escaped chars
 * that have been left behind by the preprocessor.  If such characters are not
 * escaped, then the preprocessor removes such chars (carriage returns and
 * line feeds).  This is how Progress 4GL handles these characters.  This
 * means that in a raw source file that has not been preprocessed, a string
 * literal can be split across any number of lines and the Progress 
 * preprocessor will put the string back together, ignoring the carriage
 * returns and newlines.
 * <p>
 * Tilde and the backslash can BOTH be Progress escape characters and as such,
 * they need special attention. Backslash is only honored if UNIX escapes
 * mode is set on. In particular, this rule must separately match 3 or 6
 * constructs as part of a string depending on which escape sequences are
 * honored.
 * <p>
 * <pre>
 *    ~&quot;
 *    \&quot;
 *    ~~
 *    ~\
 *    \\
 *    \~
 *    ~
 *    \
 * </pre>
 * <p>
 * The first 2 are a way of embedding a quote in a string.  The next 4 are
 * important because if this rule were to encounter '~~', '\\', '~\' or '\~'
 * (which are valid strings) the rule would consume the first escape and then
 * encounter an escaped quote which would not terminate the string properly,
 * leading to a non-ending string.  So this rule matches on any escaped escape
 * char to eliminate this situation.  Finally, a double ~ or \ that is not
 * followed by a &quot; or a duplicate escape char is matched as a double
 * character.  This is required (one may think that the closure rule
 * should handle this case) because when ANTLR sees a ~ or \ in the leftmost
 * position of the alternatives, it DOES NOT include it in the list of
 * 'everything that is not a closing quote'.  Likewise \n is not included.
 * Of course, these tilde constructions must be placed in a specific order
 * and if so, the ambiguity warnings that ANTLR reports can be disabled.
 * <p>
 * The greedy option does not need to be disabled here as the closure rule
 * termination is built into the subrule itself: it accepts anything that
 * isn't an unescaped double quote character (see above).
 * <p>
 * Tabs and spaces are maintained inside strings.
 */
protected
DSTRING
   :
      (
         '\"'                              // opening "
              (
                 options { generateAmbigWarnings = false; }
                 :
                   "~\""                   // allow ~"
                 | "~~"                    // allow ~~ (consume as one)
                 | "~\\"                   // allow ~\ (consume as one)
                 | '~'                     // allow ~ by itself
                 | { unixEscapes }?
                   "\\\""                  // allow \"
                 | { unixEscapes }?
                   "\\\\"                  // allow \\ (consume as one)
                 | { unixEscapes }?
                   "\\~"                   // allow \~ (consume as one)
                 | '\\'                    // allow \ by itself
                 | "\"\""                  // allow ""
                 
                 | { schema }?             // convert CRNL to a space char
                   "\r\n"!                 // when lexing a DF file
                   {
                      append(' ');
                      newline();
                   }
                 | { schema }?             // convert NL or CR to a space
                   ('\n'! | '\r'!)         // char when lexing a DF file
                   {
                      append(' ');
                      newline();
                   }
                 | '\n'     { newline(); } // otherwise allow NL
                 
                 | ~'\"'                   // anything else except closing "
              )*
         '\"'                              // closing "
      )
   ;

/**
 * This rule is only invoked when in export processing mode. Export strings
 * are essentially runtime strings which have no escape sequences, except
 * an embedded double quote (") is doubled up ("") to avoid confusion with
 * the enclosing delimiter.
 * <p>
 * Matches an opening double quote, arbitrary contents and an ending double
 * quote.  Two contiguous double quote characters are accepted as contents
 * (they do not terminate the string), but they are collapsed into a single
 * quote character.  Opening and closing double quotes are discarded.
 * <p>
 * Any newlines inside the string are identified and the lexer's internal
 * newline counter is properly maintained.
 * <p>
 * Tabs and spaces and all non-printing characters are maintained inside
 * strings (including embedded binary control characters).
 */
protected
XSTRING
   :
      (
         '\"'!                             // opening " (discarded)
              (
                 options { generateAmbigWarnings = false; }
                 :
                   "\"\""!
                   { append("\""); }       // convert "" to "
                 | '\n'   { newline(); }   // allow NL; preserve counter
                 | ~'\"'                   // anything else except closing "
              )*
         '\"'!                             // closing " (discarded)
      )
   ;

/**
 * This rule is only invoked when in raw processing mode. Raw strings
 * represent encoded binary data.
 * <p>
 * Matches an opening double quote, arbitrary contents and an ending double
 * quote.  Opening and closing double quotes are discarded.
 * <p>
 * Currently, no assumptions are made about the contents of the string,
 * except that it will not contain an instance of a double quote, since this
 * is the end delimiter.  All other data is accepted exactly as is.
 * Currently, no decoding is performed.
 */
protected
RSTRING
   :
      (
         '\"'! ( ~'\"' )* '\"'!
      )
   ;

/**
 * Matches the options that can be appended to a Progress string literal.
 * These options consist of 4 parts, some of which are optional.  String 
 * options always start with a ':' which must follow the string's ending 
 * single or double quote with no intervening whitespace.  After that
 * colon, there must be one or more of a justification, size allocation in
 * characters and a flag noting if this string is translatable or not.
 * <pre>
 *    justification          'r' | 'l' | 'c' | 't'
 *    translatable           'u' | 'x'
 *    size                   integer literal (one or more contiguous digits)
 * </pre>
 * <p>
 * The use of 'X' is an undocumented feature of the 4GL and it can only appear
 * if 'U' is not present. The meaning of this is not yet known.
 * <p>
 * These options can come in any order and at least one of them must appear.
 * This causes some problems since there is no simple syntax to specify
 * such conditions in the grammar.  Whitespace cannot be placed between any
 * of the options.
 * <p>
 * In addition, just as the lookahead for the <code>mDEC_LITERAL</code> rule
 * could incorrectly 'pull' processing down into that rule as a false match, 
 * so too can this happen here. For this reason, the possible alternatives
 * are specified and then the empty alternative is used to detect a false
 * match and the lexer's stream input is manually reset to backup and restart
 * processing at the top-level.
 * <p>
 * None of this would be possible if comments or whitespace were allowed in
 * string options.  They are not.
 */   
protected
STR_OPTIONS
   :
      { 
         // save state *before* matching of ':' so we can reset if it is
         // really supposed to be a statement separator and not string
         // options as we first thought
         boolean commitFlag = true;
         int     back       = mark();
         String  save       = getText();
      }
      (
         ':'
         (
            options { generateAmbigWarnings = false; }:
               ('r' | 'l' | 'c' | 't') ('u' | 'x') (DIGIT)*
            |  ('r' | 'l' | 'c' | 't') (DIGIT)+ ('u' | 'x')?
            |  ('r' | 'l' | 'c' | 't')
            |  ('u' | 'x') ('r' | 'l' | 'c' | 't') (DIGIT)*
            |  ('u' | 'x') (DIGIT)+ ('r' | 'l' | 'c' | 't')?
            |  ('u' | 'x')
            | (DIGIT)+
              (
                   ( ('r' | 'l' | 'c' | 't') ('u' | 'x')? )
                 | ( ('u' | 'x') ('r' | 'l' | 'c' | 't')? )
              )?
            | {
                 // We match the EMPTY alternative if one of the valid
                 // string options is not found above!!!!!
                 // Abort! Reset state to remove the already matched ':'.
                 commitFlag = false;
                 setText( save );
                 rewind( back );
              }
         )
      )
      {
         // if we did not rewind, we need to reset our mark without rewinding
         if ( commitFlag )
            commit();
      }
   ;

/**
 * Matches any alphabetic character: <code>a - z</code>.
 */   
protected   
LETTER    : 'a'..'z' ;
   
/**
 * Matches any numeric digit: <code>0 - 9</code>.
 */
protected
DIGIT     : '0'..'9' ;

/**
 * Matches all characters that can be made part of a valid user-defined
 * symbol (except alphabetic and numeric characters):
 * <code> # $ % &amp; - _ </code>
 */
protected
SYM_CHAR  : '#' | '$' | '%' | '&' | '-' | '_' ;

/**
 * Matches the open comment sequence of characters: <code>/&#042;</code>
 */
protected
CMT_OPEN  : "/*" ;

/**
 * Matches the close comment sequence of characters: <code>&#042;/</code>
 */
protected
CMT_CLOSE : "*/" ;

/**
 * Matches non-visible ASCII codes that should be silently ignored. Extended
 * ASCII characters are NOT matched here. This is text that should not appear
 * in 4GL outside of a string or comment.
 */
protected
JUNK      : '\000'..'\010' | '\013'..'\014' | '\016'..'\037' | '\177' ;
